Type: Package
Package: loo
Title: Efficient Leave-One-Out Cross-Validation and WAIC for Bayesian
        Models
Version: 2.10.0
Date: 2026-06-24
Authors@R: c(
    person("Aki", "Vehtari", email = "Aki.Vehtari@aalto.fi", role = "aut"),
    person("Jonah", "Gabry", email = "jgabry@gmail.com", role = c("cre", "aut")),
    person("Måns", "Magnusson", role = "aut"),
    person("Yuling", "Yao", role = "aut"),
    person("Paul-Christian", "Bürkner", role = "aut"),
    person("Topi", "Paananen", role = "aut"),
    person("Andrew", "Gelman", role = "aut"),
    person("Ben", "Goodrich", role = "ctb"),
    person("Juho", "Piironen", role = "ctb"),
    person("Bruno", "Nicenboim", role = "ctb"),
    person("Leevi", "Lindgren", role = "ctb"),
    person("Visruth", "Srimath Kandali", role = "ctb"),
    person("Florence", "Bockting", role = "ctb")
  )
Maintainer: Jonah Gabry <jgabry@gmail.com>
Description: Efficient approximate leave-one-out cross-validation (LOO)
    for Bayesian models fit using Markov chain Monte Carlo, as described
    in Vehtari, Gelman, and Gabry (2017) <doi:10.1007/s11222-016-9696-4>.
    The approximation uses Pareto smoothed importance sampling (PSIS), a
    new procedure for regularizing importance weights.  As a byproduct of
    the calculations, we also obtain approximate standard errors for
    estimated predictive errors and for the comparison of predictive
    errors between models. The package also provides methods for using
    stacking and other model weighting techniques to average Bayesian
    predictive distributions.
License: GPL (>= 3)
URL: https://mc-stan.org/loo/, https://discourse.mc-stan.org
BugReports: https://github.com/stan-dev/loo/issues
Depends: R (>= 3.5)
Imports: checkmate, matrixStats (>= 0.52), parallel, posterior (>=
        1.7.0), stats
Suggests: bayesplot (>= 1.7.0), brms (>= 2.10.0), ggplot2, graphics,
        knitr, rmarkdown, rstan, rstanarm (>= 2.19.0), rstantools,
        spdep, testthat (>= 3.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/testthat/parallel: true
Config/testthat/start-first: loo_subsampling_cases, loo_subsampling
Encoding: UTF-8
LazyData: TRUE
SystemRequirements: pandoc (>= 1.12.3), pandoc-citeproc
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-24 16:15:31 UTC; jgabry
Author: Aki Vehtari [aut],
  Jonah Gabry [cre, aut],
  Måns Magnusson [aut],
  Yuling Yao [aut],
  Paul-Christian Bürkner [aut],
  Topi Paananen [aut],
  Andrew Gelman [aut],
  Ben Goodrich [ctb],
  Juho Piironen [ctb],
  Bruno Nicenboim [ctb],
  Leevi Lindgren [ctb],
  Visruth Srimath Kandali [ctb],
  Florence Bockting [ctb]
Repository: CRAN
Date/Publication: 2026-06-26 08:20:08 UTC

# Generated by roxygen2: do not edit by hand

S3method("$",loo)
S3method("[",loo)
S3method("[[",loo)
S3method(.compute_point_estimate,default)
S3method(.compute_point_estimate,matrix)
S3method(.ndraws,default)
S3method(.ndraws,matrix)
S3method(.thin_draws,default)
S3method(.thin_draws,matrix)
S3method(.thin_draws,numeric)
S3method(E_loo,default)
S3method(E_loo,matrix)
S3method(E_loo_khat,default)
S3method(E_loo_khat,matrix)
S3method(ap_psis,array)
S3method(ap_psis,default)
S3method(ap_psis,matrix)
S3method(as.psis_loo,psis_loo)
S3method(as.psis_loo,psis_loo_ss)
S3method(as.psis_loo_ss,psis_loo)
S3method(as.psis_loo_ss,psis_loo_ss)
S3method(crps,matrix)
S3method(crps,numeric)
S3method(dim,importance_sampling)
S3method(dim,kfold)
S3method(dim,loo)
S3method(dim,psis_loo)
S3method(dim,waic)
S3method(elpd,array)
S3method(elpd,matrix)
S3method(importance_sampling,array)
S3method(importance_sampling,default)
S3method(importance_sampling,matrix)
S3method(loo,"function")
S3method(loo,array)
S3method(loo,matrix)
S3method(loo_approximate_posterior,"function")
S3method(loo_approximate_posterior,array)
S3method(loo_approximate_posterior,matrix)
S3method(loo_compare,default)
S3method(loo_compare,psis_loo_ss_list)
S3method(loo_crps,matrix)
S3method(loo_model_weights,default)
S3method(loo_moment_match,default)
S3method(loo_predictive_metric,matrix)
S3method(loo_scrps,matrix)
S3method(loo_subsample,"function")
S3method(nobs,psis_loo_ss)
S3method(plot,loo)
S3method(plot,psis)
S3method(plot,psis_loo)
S3method(pointwise,loo)
S3method(print,compare.loo)
S3method(print,compare.loo_ss)
S3method(print,importance_sampling)
S3method(print,importance_sampling_loo)
S3method(print,loo)
S3method(print,pareto_k_table)
S3method(print,pseudobma_bb_weights)
S3method(print,pseudobma_weights)
S3method(print,psis)
S3method(print,psis_loo)
S3method(print,psis_loo_ap)
S3method(print,stacking_weights)
S3method(print,waic)
S3method(print_dims,elpd_generic)
S3method(print_dims,importance_sampling)
S3method(print_dims,importance_sampling_loo)
S3method(print_dims,kfold)
S3method(print_dims,psis_loo)
S3method(print_dims,psis_loo_ss)
S3method(print_dims,waic)
S3method(psis,array)
S3method(psis,default)
S3method(psis,matrix)
S3method(psis_n_eff,default)
S3method(psis_n_eff,matrix)
S3method(relative_eff,"function")
S3method(relative_eff,array)
S3method(relative_eff,default)
S3method(relative_eff,importance_sampling)
S3method(relative_eff,matrix)
S3method(scrps,matrix)
S3method(scrps,numeric)
S3method(sis,array)
S3method(sis,default)
S3method(sis,matrix)
S3method(tis,array)
S3method(tis,default)
S3method(tis,matrix)
S3method(update,psis_loo_ss)
S3method(waic,"function")
S3method(waic,array)
S3method(waic,matrix)
S3method(weights,importance_sampling)
export(.compute_point_estimate)
export(.ndraws)
export(.thin_draws)
export(E_loo)
export(compare)
export(crps)
export(elpd)
export(example_loglik_array)
export(example_loglik_matrix)
export(extract_log_lik)
export(find_model_names)
export(gpdfit)
export(is.kfold)
export(is.loo)
export(is.psis)
export(is.psis_loo)
export(is.sis)
export(is.tis)
export(is.waic)
export(kfold)
export(kfold_split_grouped)
export(kfold_split_random)
export(kfold_split_stratified)
export(loo)
export(loo.array)
export(loo.function)
export(loo.matrix)
export(loo_approximate_posterior)
export(loo_approximate_posterior.array)
export(loo_approximate_posterior.function)
export(loo_approximate_posterior.matrix)
export(loo_compare)
export(loo_crps)
export(loo_i)
export(loo_model_weights)
export(loo_model_weights.default)
export(loo_moment_match)
export(loo_moment_match.default)
export(loo_predictive_metric)
export(loo_scrps)
export(loo_subsample)
export(loo_subsample.function)
export(mcse_loo)
export(nlist)
export(obs_idx)
export(pareto_k_ids)
export(pareto_k_influence_values)
export(pareto_k_table)
export(pareto_k_values)
export(pointwise)
export(print_dims)
export(pseudobma_weights)
export(psis)
export(psis_n_eff_values)
export(psislw)
export(relative_eff)
export(scrps)
export(sis)
export(stacking_weights)
export(tis)
export(waic)
export(waic.array)
export(waic.function)
export(waic.matrix)
export(weights.importance_sampling)
importFrom(matrixStats,colLogSumExps)
importFrom(matrixStats,colMaxs)
importFrom(matrixStats,colSums2)
importFrom(matrixStats,colVars)
importFrom(matrixStats,logSumExp)
importFrom(parallel,makePSOCKcluster)
importFrom(parallel,mclapply)
importFrom(parallel,parLapply)
importFrom(parallel,stopCluster)
importFrom(stats,constrOptim)
importFrom(stats,nobs)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,rgamma)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,update)
importFrom(stats,var)
importFrom(stats,weights)

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----more-knitr-ops, include=FALSE--------------------------------------------
knitr::opts_chunk$set(
  cache=TRUE,
  message=FALSE, 
  warning=FALSE
)

## ----lpdf, eval=FALSE---------------------------------------------------------
# /**
#  * Normal log-pdf for spatially lagged responses
#  *
#  * @param y Vector of response values.
#  * @param mu Mean parameter vector.
#  * @param sigma Positive scalar residual standard deviation.
#  * @param rho Positive scalar autoregressive parameter.
#  * @param W Spatial weight matrix.
#  *
#  * @return A scalar to be added to the log posterior.
#  */
# real normal_lagsar_lpdf(vector y, vector mu, real sigma,
#                         real rho, matrix W) {
#   int N = rows(y);
#   real inv_sigma2 = 1 / square(sigma);
#   matrix[N, N] W_tilde = -rho * W;
#   vector[N] half_pred;
# 
#   for (n in 1:N) W_tilde[n,n] += 1;
# 
#   half_pred = W_tilde * (y - mdivide_left(W_tilde, mu));
# 
#   return 0.5 * log_determinant(crossprod(W_tilde) * inv_sigma2) -
#          0.5 * dot_self(half_pred) * inv_sigma2;
# }

## ----setup, cache=FALSE-------------------------------------------------------
library("loo")
library("brms")
library("bayesplot")
library("ggplot2")
color_scheme_set("brightblue")
theme_set(theme_default())


SEED <- 10001 
set.seed(SEED) # only sets seed for R (seed for Stan set later)

# loads COL.OLD data frame and COL.nb neighbor list
data(oldcol, package = "spdep") 

## ----data---------------------------------------------------------------------
str(COL.OLD[, c("CRIME", "HOVAL", "INC")])

## ----fit, results="hide"------------------------------------------------------
fit <- brm(
  CRIME ~ INC + HOVAL + sar(COL.nb, type = "lag"), 
  data = COL.OLD,
  data2 = list(COL.nb = COL.nb),
  chains = 4,
  seed = SEED
)

## ----plot-lagsar, message=FALSE-----------------------------------------------
lagsar <- as.matrix(fit, pars = "lagsar")
estimates <- quantile(lagsar, probs = c(0.25, 0.5, 0.75))
mcmc_hist(lagsar) + 
  vline_at(estimates, linetype = 2, size = 1) +
  ggtitle("lagsar: posterior median and 50% central interval")

## ----approx-------------------------------------------------------------------
posterior <- as.data.frame(fit)
y <- fit$data$CRIME
N <- length(y)
S <- nrow(posterior)
loglik <- yloo <- sdloo <- matrix(nrow = S, ncol = N)

for (s in 1:S) {
  p <- posterior[s, ]
  eta <- p$b_Intercept + p$b_INC * fit$data$INC + p$b_HOVAL * fit$data$HOVAL
  W_tilde <- diag(N) - p$lagsar * spdep::nb2mat(COL.nb)
  Cinv <- t(W_tilde) %*% W_tilde / p$sigma^2
  g <- Cinv %*% (y - solve(W_tilde, eta))
  cbar <- diag(Cinv)
  yloo[s, ] <- y - g / cbar
  sdloo[s, ] <- sqrt(1 / cbar)
  loglik[s, ] <- dnorm(y, yloo[s, ], sdloo[s, ], log = TRUE)
}

# use loo for psis smoothing
log_ratios <- -loglik
psis_result <- psis(log_ratios)

## ----plot, cache = FALSE------------------------------------------------------
plot(psis_result, label_points = TRUE)

## ----checklast, cache = FALSE-------------------------------------------------
yloo_sub <- yloo[S, ]
sdloo_sub <- sdloo[S, ]
df <- data.frame(
  y = y, 
  yloo = yloo_sub,
  ymin = yloo_sub - sdloo_sub * 2,
  ymax = yloo_sub + sdloo_sub * 2
)
ggplot(data=df, aes(x = y, y = yloo, ymin = ymin, ymax = ymax)) +
  geom_errorbar(
    width = 1, 
    color = "skyblue3", 
    position = position_jitter(width = 0.25)
  ) +
  geom_abline(color = "gray30", size = 1.2) +
  geom_point()

## ----psisloo------------------------------------------------------------------
(psis_loo <- loo(loglik))

## ----fit_dummy, cache = TRUE--------------------------------------------------
# see help("mi", "brms") for details on the mi() usage
fit_dummy <- brm(
  CRIME | mi() ~ INC + HOVAL + sar(COL.nb, type = "lag"), 
  data = COL.OLD,
  data2 = list(COL.nb = COL.nb),
  chains = 0
)

## ----exact-loo-cv, results="hide", message=FALSE, warning=FALSE, cache = TRUE----
S <- 500
res <- vector("list", N)
loglik <- matrix(nrow = S, ncol = N)
for (i in seq_len(N)) {
  dat_mi <- COL.OLD
  dat_mi$CRIME[i] <- NA
  fit_i <- update(fit_dummy, newdata = dat_mi, 
                  # just for vignette
                  chains = 1, iter = S * 2)
  posterior <- as.data.frame(fit_i)
  yloo <- sdloo <- rep(NA, S)
  for (s in seq_len(S)) {
    p <- posterior[s, ]
    y_miss_i <- y
    y_miss_i[i] <- p$Ymi
    eta <- p$b_Intercept + p$b_INC * fit_i$data$INC + p$b_HOVAL * fit_i$data$HOVAL
    W_tilde <- diag(N) - p$lagsar * spdep::nb2mat(COL.nb)
    Cinv <- t(W_tilde) %*% W_tilde / p$sigma^2
    g <- Cinv %*% (y_miss_i - solve(W_tilde, eta))
    cbar <- diag(Cinv);
    yloo[s] <- y_miss_i[i] - g[i] / cbar[i]
    sdloo[s] <- sqrt(1 / cbar[i])
    loglik[s, i] <- dnorm(y[i], yloo[s], sdloo[s], log = TRUE)
  }
  ypred <- rnorm(S, yloo, sdloo)
  res[[i]] <- data.frame(y = c(posterior$Ymi, ypred))
  res[[i]]$type <- rep(c("pp", "loo"), each = S)
  res[[i]]$obs <- i
}
res <- do.call(rbind, res)

## ----yplots, cache = FALSE, fig.width=10, out.width="95%", fig.asp = 0.3------
res_sub <- res[res$obs %in% 1:4, ]
ggplot(res_sub, aes(y, fill = type)) +
  geom_density(alpha = 0.6) +
  facet_wrap("obs", scales = "fixed", ncol = 4)

## ----loo_exact, cache=FALSE---------------------------------------------------
log_mean_exp <- function(x) {
  # more stable than log(mean(exp(x)))
  max_x <- max(x)
  max_x + log(sum(exp(x - max_x))) - log(length(x))
}
exact_elpds <- apply(loglik, 2, log_mean_exp)
exact_elpd <- sum(exact_elpds)
round(exact_elpd, 1)

## ----compare, fig.height=5----------------------------------------------------
df <- data.frame(
  approx_elpd = psis_loo$pointwise[, "elpd_loo"],
  exact_elpd = exact_elpds
)
ggplot(df, aes(x = approx_elpd, y = exact_elpd)) +
  geom_abline(color = "gray30") +
  geom_point(size = 2) +
  geom_point(data = df[4, ], size = 3, color = "red3") +
  xlab("Approximate elpds") +
  ylab("Exact elpds") +
  coord_fixed(xlim = c(-16, -3), ylim = c(-16, -3))

## ----pt4----------------------------------------------------------------------
without_pt_4 <- c(
  approx = sum(psis_loo$pointwise[-4, "elpd_loo"]),
  exact = sum(exact_elpds[-4])  
)
round(without_pt_4, 1)

## ----brms-stan-code, eval=FALSE-----------------------------------------------
# // generated with brms 2.2.0
# functions {
# /**
#  * Normal log-pdf for spatially lagged responses
#  *
#  * @param y Vector of response values.
#  * @param mu Mean parameter vector.
#  * @param sigma Positive scalar residual standard deviation.
#  * @param rho Positive scalar autoregressive parameter.
#  * @param W Spatial weight matrix.
#  *
#  * @return A scalar to be added to the log posterior.
#  */
#   real normal_lagsar_lpdf(vector y, vector mu, real sigma,
#                           real rho, matrix W) {
#     int N = rows(y);
#     real inv_sigma2 = 1 / square(sigma);
#     matrix[N, N] W_tilde = -rho * W;
#     vector[N] half_pred;
#     for (n in 1:N) W_tilde[n, n] += 1;
#     half_pred = W_tilde * (y - mdivide_left(W_tilde, mu));
#     return 0.5 * log_determinant(crossprod(W_tilde) * inv_sigma2) -
#            0.5 * dot_self(half_pred) * inv_sigma2;
#   }
# }
# data {
#   int<lower=1> N;  // total number of observations
#   vector[N] Y;  // response variable
#   int<lower=0> Nmi;  // number of missings
#   int<lower=1> Jmi[Nmi];  // positions of missings
#   int<lower=1> K;  // number of population-level effects
#   matrix[N, K] X;  // population-level design matrix
#   matrix[N, N] W;  // spatial weight matrix
#   int prior_only;  // should the likelihood be ignored?
# }
# transformed data {
#   int Kc = K - 1;
#   matrix[N, K - 1] Xc;  // centered version of X
#   vector[K - 1] means_X;  // column means of X before centering
#   for (i in 2:K) {
#     means_X[i - 1] = mean(X[, i]);
#     Xc[, i - 1] = X[, i] - means_X[i - 1];
#   }
# }
# parameters {
#   vector[Nmi] Ymi;  // estimated missings
#   vector[Kc] b;  // population-level effects
#   real temp_Intercept;  // temporary intercept
#   real<lower=0> sigma;  // residual SD
#   real<lower=0,upper=1> lagsar;  // SAR parameter
# }
# transformed parameters {
# }
# model {
#   vector[N] Yl = Y;
#   vector[N] mu = Xc * b + temp_Intercept;
#   Yl[Jmi] = Ymi;
#   // priors including all constants
#   target += student_t_lpdf(temp_Intercept | 3, 34, 17);
#   target += student_t_lpdf(sigma | 3, 0, 17)
#     - 1 * student_t_lccdf(0 | 3, 0, 17);
#   // likelihood including all constants
#   if (!prior_only) {
#     target += normal_lagsar_lpdf(Yl | mu, sigma, lagsar, W);
#   }
# }
# generated quantities {
#   // actual population-level intercept
#   real b_Intercept = temp_Intercept - dot_product(means_X, b);
# }

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----stancode-----------------------------------------------------------------
# Note: some syntax used in this Stan program requires RStan >= 2.26 (or CmdStanR)
# To use an older version of RStan change the line declaring `y` to: int y[N];
stancode <- "
data {
  int<lower=1> K;
  int<lower=1> N;
  matrix[N,K] x;
  array[N] int y;
  vector[N] offset_; // offset is reserved keyword in Stan so use offset_

  real beta_prior_scale;
  real alpha_prior_scale;
}
parameters {
  vector[K] beta;
  real intercept;
}
model {
  y ~ poisson(exp(x * beta + intercept + offset_));
  beta ~ normal(0,beta_prior_scale);
  intercept ~ normal(0,alpha_prior_scale);
}
generated quantities {
  vector[N] log_lik;
  for (n in 1:N) {
    log_lik[n] = poisson_lpmf(y[n] | exp(x[n] * beta + intercept + offset_[n]));
  }
}
"

## ----setup, message=FALSE-----------------------------------------------------
library("rstan")
library("loo")
seed <- 9547
set.seed(seed)

## ----modelfit, message=FALSE--------------------------------------------------
# Prepare data
data(roaches, package = "rstanarm")
roaches$roach1 <- sqrt(roaches$roach1)
y <- roaches$y
x <- roaches[, c("roach1", "treatment", "senior")]
offset <- log(roaches[, "exposure2"])
n <- dim(x)[1]
k <- dim(x)[2]

standata <- list(
  N = n,
  K = k,
  x = as.matrix(x),
  y = y,
  offset_ = offset,
  beta_prior_scale = 2.5,
  alpha_prior_scale = 5.0
)

# Compile
stanmodel <- stan_model(model_code = stancode)

# Fit model
fit <- sampling(stanmodel, data = standata, seed = seed, refresh = 0)
print(fit, pars = "beta")

## ----loo1---------------------------------------------------------------------
loo1 <- loo(fit)
loo1

## ----loo_moment_match---------------------------------------------------------
# available in rstan >= 2.21
loo2 <- loo(fit, moment_match = TRUE)
loo2

## ----stanfitfuns--------------------------------------------------------------
# create a named list of draws for use with rstan methods
.rstan_relist <- function(x, skeleton) {
  out <- utils::relist(x, skeleton)
  for (i in seq_along(skeleton)) {
    dim(out[[i]]) <- dim(skeleton[[i]])
  }
  out
}

# rstan helper function to get dims of parameters right
.create_skeleton <- function(pars, dims) {
  out <- lapply(seq_along(pars), function(i) {
    len_dims <- length(dims[[i]])
    if (len_dims < 1) {
      return(0)
    }
    return(array(0, dim = dims[[i]]))
  })
  names(out) <- pars
  out
}

# extract original posterior draws
post_draws_stanfit <- function(x, ...) {
  as.matrix(x)
}

# compute a matrix of log-likelihood values for the ith observation
# matrix contains information about the number of MCMC chains
log_lik_i_stanfit <- function(x, i, parameter_name = "log_lik", ...) {
  loo::extract_log_lik(x, parameter_name, merge_chains = FALSE)[,, i]
}

# transform parameters to the unconstraint space
unconstrain_pars_stanfit <- function(x, pars, ...) {
  skeleton <- .create_skeleton(x@sim$pars_oi, x@par_dims[x@sim$pars_oi])
  upars <- apply(pars, 1, FUN = function(theta) {
    rstan::unconstrain_pars(x, .rstan_relist(theta, skeleton))
  })
  # for one parameter models
  if (is.null(dim(upars))) {
    dim(upars) <- c(1, length(upars))
  }
  t(upars)
}

# compute log_prob for each posterior draws on the unconstrained space
log_prob_upars_stanfit <- function(x, upars, ...) {
  apply(
    upars,
    1,
    rstan::log_prob,
    object = x,
    adjust_transform = TRUE,
    gradient = FALSE
  )
}

# compute log_lik values based on the unconstrained parameters
log_lik_i_upars_stanfit <- function(
  x,
  upars,
  i,
  parameter_name = "log_lik",
  ...
) {
  S <- nrow(upars)
  out <- numeric(S)
  for (s in seq_len(S)) {
    out[s] <- rstan::constrain_pars(x, upars = upars[s, ])[[parameter_name]][i]
  }
  out
}

## ----loo_moment_match.default, message=FALSE----------------------------------
loo3 <- loo::loo_moment_match.default(
  x = fit,
  loo = loo1,
  post_draws = post_draws_stanfit,
  log_lik_i = log_lik_i_stanfit,
  unconstrain_pars = unconstrain_pars_stanfit,
  log_prob_upars = log_prob_upars_stanfit,
  log_lik_i_upars = log_lik_i_upars_stanfit
)
loo3

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----eval=FALSE---------------------------------------------------------------
# library("rstan")
# 
# # Prepare data
# url <- "http://stat.columbia.edu/~gelman/arm/examples/arsenic/wells.dat"
# wells <- read.table(url)
# wells$dist100 <- with(wells, dist / 100)
# X <- model.matrix(~ dist100 + arsenic, wells)
# standata <- list(y = wells$switch, X = X, N = nrow(X), P = ncol(X))
# 
# # Fit model
# fit_1 <- stan("logistic.stan", data = standata)
# print(fit_1, pars = "beta")

## ----eval=FALSE---------------------------------------------------------------
# library("loo")
# 
# # Extract pointwise log-likelihood
# # using merge_chains=FALSE returns an array, which is easier to
# # use with relative_eff()
# log_lik_1 <- extract_log_lik(fit_1, merge_chains = FALSE)
# 
# # as of loo v2.0.0 we can optionally provide relative effective sample sizes
# # when calling loo, which allows for better estimates of the PSIS effective
# # sample sizes and Monte Carlo error
# r_eff <- relative_eff(exp(log_lik_1), cores = 2)
# 
# # preferably use more than 2 cores (as many cores as possible)
# # will use value of 'mc.cores' option if cores is not specified
# loo_1 <- loo(log_lik_1, r_eff = r_eff, cores = 2)
# print(loo_1)

## ----eval=FALSE---------------------------------------------------------------
# standata$X[, "arsenic"] <- log(standata$X[, "arsenic"])
# fit_2 <- stan(fit = fit_1, data = standata)
# 
# log_lik_2 <- extract_log_lik(fit_2, merge_chains = FALSE)
# r_eff_2 <- relative_eff(exp(log_lik_2))
# loo_2 <- loo(log_lik_2, r_eff = r_eff_2, cores = 2)
# print(loo_2)

## ----eval=FALSE---------------------------------------------------------------
# # Compare
# comp <- loo_compare(loo_1, loo_2)

## ----eval=FALSE---------------------------------------------------------------
# print(comp) # can set simplify=FALSE for more detailed print output

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----stancode-----------------------------------------------------------------
# Note: some syntax used in this Stan program requires RStan >= 2.26 (or CmdStanR)
# To use an older version of RStan change the line declaring `y` to: int y[N];
stancode <- "
data {
  int<lower=1> K;
  int<lower=1> N;
  matrix[N,K] x;
  array[N] int y;
  vector[N] offset_; // offset is reserved keyword in Stan so use offset_

  real beta_prior_scale;
  real alpha_prior_scale;
}
parameters {
  vector[K] beta;
  real intercept;
}
model {
  y ~ poisson(exp(x * beta + intercept + offset_));
  beta ~ normal(0,beta_prior_scale);
  intercept ~ normal(0,alpha_prior_scale);
}
generated quantities {
  vector[N] log_lik;
  for (n in 1:N)
    log_lik[n] = poisson_lpmf(y[n] | exp(x[n] * beta + intercept + offset_[n]));
}
"

## ----setup, message=FALSE-----------------------------------------------------
library("rstan")
library("loo")
seed <- 9547
set.seed(seed)

## ----modelfit-holdout, message=FALSE------------------------------------------
# Prepare data
data(roaches, package = "rstanarm")
roaches$roach1 <- sqrt(roaches$roach1)
roaches$offset <- log(roaches[,"exposure2"])
# 20% of the data goes to the test set:
roaches$test <- 0
roaches$test[sample(.2 * seq_len(nrow(roaches)))] <- 1
# data to "train" the model
data_train <- list(y = roaches$y[roaches$test == 0],
                   x = as.matrix(roaches[roaches$test == 0,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$test == 0,]),
                   K = 3,
                   offset_ = roaches$offset[roaches$test == 0],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )
# data to "test" the model
data_test <- list(y = roaches$y[roaches$test == 1],
                   x = as.matrix(roaches[roaches$test == 1,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$test == 1,]),
                   K = 3,
                   offset_ = roaches$offset[roaches$test == 1],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )

## ----fit-train----------------------------------------------------------------
# Compile
stanmodel <- stan_model(model_code = stancode)
# Fit model
fit <- sampling(stanmodel, data = data_train, seed = seed, refresh = 0)

## ----gen-test-----------------------------------------------------------------
gen_test <- gqs(stanmodel, draws = as.matrix(fit), data= data_test)
log_pd <- extract_log_lik(gen_test)

## ----elpd-holdout-------------------------------------------------------------
(elpd_holdout <- elpd(log_pd))

## ----prepare-folds, message=FALSE---------------------------------------------
# Prepare data
roaches$fold <- kfold_split_random(K = 10, N = nrow(roaches))

## -----------------------------------------------------------------------------
# Prepare a matrix with the number of post-warmup iterations by number of observations:
log_pd_kfold <- matrix(nrow = 4000, ncol = nrow(roaches))
# Loop over the folds
for(k in 1:10){
  data_train <- list(y = roaches$y[roaches$fold != k],
                   x = as.matrix(roaches[roaches$fold != k,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$fold != k,]),
                   K = 3,
                   offset_ = roaches$offset[roaches$fold != k],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )
  data_test <- list(y = roaches$y[roaches$fold == k],
                   x = as.matrix(roaches[roaches$fold == k,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$fold == k,]),
                   K = 3,
                   offset_ = roaches$offset[roaches$fold == k],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )
  fit <- sampling(stanmodel, data = data_train, seed = seed, refresh = 0)
  gen_test <- gqs(stanmodel, draws = as.matrix(fit), data= data_test)
  log_pd_kfold[, roaches$fold == k] <- extract_log_lik(gen_test)
}

## ----elpd-kfold---------------------------------------------------------------
(elpd_kfold <- elpd(log_pd_kfold))

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----setup, message=FALSE-----------------------------------------------------
library("rstan")
library("loo")
set.seed(4711)

## ----llfun_logistic-----------------------------------------------------------
# we'll add an argument log to toggle whether this is a log-likelihood or 
# likelihood function. this will be useful later in the vignette.
llfun_logistic <- function(data_i, draws, log = TRUE) {
  x_i <- as.matrix(data_i[, which(grepl(colnames(data_i), pattern = "X")), drop=FALSE])
  logit_pred <- draws %*% t(x_i)
  dbinom(x = data_i$y, size = 1, prob = 1/(1 + exp(-logit_pred)), log = log)
}

## ----eval=FALSE---------------------------------------------------------------
# # Prepare data
# url <- "http://stat.columbia.edu/~gelman/arm/examples/arsenic/wells.dat"
# wells <- read.table(url)
# wells$dist100 <- with(wells, dist / 100)
# X <- model.matrix(~ dist100 + arsenic, wells)
# standata <- list(y = wells$switch, X = X, N = nrow(X), P = ncol(X))
# 
# # Compile
# stan_mod <- stan_model("logistic.stan")
# 
# # Fit model
# fit_1 <- sampling(stan_mod, data = standata, seed = 4711)
# print(fit_1, pars = "beta")

## ----eval=FALSE---------------------------------------------------------------
# # used for draws argument to loo_i
# parameter_draws_1 <- extract(fit_1)$beta
# 
# # used for data argument to loo_i
# stan_df_1 <- as.data.frame(standata)
# 
# # compute relative efficiency (this is slow and optional but is recommended to allow
# # for adjusting PSIS effective sample size based on MCMC effective sample size)
# r_eff <- relative_eff(llfun_logistic,
#                       log = FALSE, # relative_eff wants likelihood not log-likelihood values
#                       chain_id = rep(1:4, each = 1000),
#                       data = stan_df_1,
#                       draws = parameter_draws_1,
#                       cores = 2)
# 
# loo_i(i = 1, llfun_logistic, r_eff = r_eff, data = stan_df_1, draws = parameter_draws_1)

## ----eval=FALSE---------------------------------------------------------------
# set.seed(4711)
# loo_ss_1 <-
#   loo_subsample(
#     llfun_logistic,
#     observations = 100, # take a subsample of size 100
#     cores = 2,
#     # these next objects were computed above
#     r_eff = r_eff,
#     draws = parameter_draws_1,
#     data = stan_df_1
#   )
# print(loo_ss_1)

## ----eval=FALSE---------------------------------------------------------------
# set.seed(4711)
# loo_ss_1b <-
#   update(
#     loo_ss_1,
#     observations = 200, # subsample 200 instead of 100
#     r_eff = r_eff,
#     draws = parameter_draws_1,
#     data = stan_df_1
#   )
# print(loo_ss_1b)

## ----eval=FALSE---------------------------------------------------------------
# set.seed(4711)
# loo_ss_1c <-
#   loo_subsample(
#     x = llfun_logistic,
#     r_eff = r_eff,
#     draws = parameter_draws_1,
#     data = stan_df_1,
#     observations = 100,
#     estimator = "hh_pps", # use Hansen-Hurwitz
#     loo_approximation = "lpd", # use lpd instead of plpd
#     loo_approximation_draws = 100,
#     cores = 2
#   )
# print(loo_ss_1c)

## ----eval=FALSE---------------------------------------------------------------
# fit_laplace <- optimizing(stan_mod, data = standata, draws = 2000,
#                           importance_resampling = TRUE)
# parameter_draws_laplace <- fit_laplace$theta_tilde # draws from approximate posterior
# log_p <- fit_laplace$log_p # log density of the posterior
# log_g <- fit_laplace$log_g # log density of the approximation

## ----eval=FALSE---------------------------------------------------------------
# set.seed(4711)
# loo_ap_1 <-
#   loo_approximate_posterior(
#     x = llfun_logistic,
#     draws = parameter_draws_laplace,
#     data = stan_df_1,
#     log_p = log_p,
#     log_g = log_g,
#     cores = 2
#   )
# print(loo_ap_1)

## ----eval=FALSE---------------------------------------------------------------
# set.seed(4711)
# loo_ap_ss_1 <-
#   loo_subsample(
#     x = llfun_logistic,
#     draws = parameter_draws_laplace,
#     data = stan_df_1,
#     log_p = log_p,
#     log_g = log_g,
#     observations = 100,
#     cores = 2
#   )
# print(loo_ap_ss_1)

## ----eval=FALSE---------------------------------------------------------------
# standata$X[, "arsenic"] <- log(standata$X[, "arsenic"])
# fit_2 <- sampling(stan_mod, data = standata)
# parameter_draws_2 <- extract(fit_2)$beta
# stan_df_2 <- as.data.frame(standata)
# 
# # recompute subsampling loo for first model for demonstration purposes
# 
# # compute relative efficiency (this is slow and optional but is recommended to allow
# # for adjusting PSIS effective sample size based on MCMC effective sample size)
# r_eff_1 <- relative_eff(
#   llfun_logistic,
#   log = FALSE, # relative_eff wants likelihood not log-likelihood values
#   chain_id = rep(1:4, each = 1000),
#   data = stan_df_1,
#   draws = parameter_draws_1,
#   cores = 2
# )
# 
# set.seed(4711)
# loo_ss_1 <- loo_subsample(
#   x = llfun_logistic,
#   r_eff = r_eff_1,
#   draws = parameter_draws_1,
#   data = stan_df_1,
#   observations = 200,
#   cores = 2
# )
# 
# # compute subsampling loo for a second model (with log-arsenic)
# 
# r_eff_2 <- relative_eff(
#   llfun_logistic,
#   log = FALSE, # relative_eff wants likelihood not log-likelihood values
#   chain_id = rep(1:4, each = 1000),
#   data = stan_df_2,
#   draws = parameter_draws_2,
#   cores = 2
# )
# loo_ss_2 <- loo_subsample(
#   x = llfun_logistic,
#   r_eff = r_eff_2,
#   draws = parameter_draws_2,
#   data = stan_df_2,
#   observations = 200,
#   cores = 2
# )
# 
# print(loo_ss_2)

## ----eval=FALSE---------------------------------------------------------------
# # Compare
# comp <- loo_compare(loo_ss_1, loo_ss_2)
# print(comp)

## ----eval=FALSE---------------------------------------------------------------
# loo_ss_2 <-
#   loo_subsample(
#     x = llfun_logistic,
#     r_eff = r_eff_2,
#     draws = parameter_draws_2,
#     data = stan_df_2,
#     observations = loo_ss_1,
#     cores = 2
#   )

## ----eval=FALSE---------------------------------------------------------------
# idx <- obs_idx(loo_ss_1)
# loo_ss_2 <- loo_subsample(
#   x = llfun_logistic,
#   r_eff = r_eff_2,
#   draws = parameter_draws_2,
#   data = stan_df_2,
#   observations = idx,
#   cores = 2
# )

## ----eval=FALSE---------------------------------------------------------------
# comp <- loo_compare(loo_ss_1, loo_ss_2)
# print(comp)

## ----eval=FALSE---------------------------------------------------------------
# # use loo() instead of loo_subsample() to compute full PSIS-LOO for model 2
# loo_full_2 <- loo(
#   x = llfun_logistic,
#   r_eff = r_eff_2,
#   draws = parameter_draws_2,
#   data = stan_df_2,
#   cores = 2
# )
# loo_compare(loo_ss_1, loo_full_2)

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----warnings=FALSE, message=FALSE--------------------------------------------
library("rstan")
library("loo")
library("matrixStats")
options(mc.cores = parallel::detectCores(), parallel=FALSE)
set.seed(24877)

## ----stancode_horseshoe-------------------------------------------------------
# Note: some syntax used in this program requires RStan >= 2.26 (or CmdStanR)
# To use an older version of RStan change the line declaring `y` to:
#    int<lower=0,upper=1> y[N];
stancode_horseshoe <- "
data {
  int <lower=0> N;
  int <lower=0> P;
  array[N] int <lower=0, upper=1> y;
  matrix [N,P] X;
  real <lower=0> scale_global;
  int <lower=0,upper=1> mixis;
}
transformed data {
  real<lower=1> nu_global=1; // degrees of freedom for the half-t priors for tau
  real<lower=1> nu_local=1;  // degrees of freedom for the half-t priors for lambdas
                             // (nu_local = 1 corresponds to the horseshoe)
  real<lower=0> slab_scale=2;// for the regularized horseshoe
  real<lower=0> slab_df=100; // for the regularized horseshoe
}
parameters {
  vector[P] z;                // for non-centered parameterization
  real <lower=0> tau;         // global shrinkage parameter
  vector <lower=0>[P] lambda; // local shrinkage parameter
  real<lower=0> caux;
}
transformed parameters {
  vector[P] beta;
  { 
    vector[P] lambda_tilde;   // 'truncated' local shrinkage parameter
    real c = slab_scale * sqrt(caux); // slab scale
    lambda_tilde = sqrt( c^2 * square(lambda) ./ (c^2 + tau^2*square(lambda)));
    beta = z .* lambda_tilde*tau;
  }
}
model {
  vector[N] means=X*beta;
  vector[N] log_lik;
  target += std_normal_lpdf(z);
  target += student_t_lpdf(lambda | nu_local, 0, 1);
  target += student_t_lpdf(tau | nu_global, 0, scale_global);
  target += inv_gamma_lpdf(caux | 0.5*slab_df, 0.5*slab_df);
  for (n in 1:N) {
    log_lik[n]= bernoulli_logit_lpmf(y[n] | means[n]);
  }
  target += sum(log_lik);
  if (mixis) {
    target += log_sum_exp(-log_lik);
  }
}
generated quantities {
  vector[N] means=X*beta;
  vector[N] log_lik;
  for (n in 1:N) {
    log_lik[n] = bernoulli_logit_lpmf(y[n] | means[n]);
  }
}
"

## ----results='hide', warning=FALSE, message=FALSE, error=FALSE----------------
data(voice)
y <- voice$y
X <- voice[2:length(voice)]
n <- dim(X)[1]
p <- dim(X)[2]
p0 <- 10
scale_global <- 2*p0/(p-p0)/sqrt(n-1)
standata <- list(N = n, P = p, X = as.matrix(X), y = c(y), scale_global = scale_global, mixis = 0)

## ----results='hide', warning=FALSE--------------------------------------------
chains <- 4
n_iter <- 2000
warm_iter <- 1000
stanmodel <- stan_model(model_code = stancode_horseshoe)
fit_post <- sampling(stanmodel, data = standata, chains = chains, iter = n_iter, warmup = warm_iter, refresh = 0)
loo_post <-loo(fit_post)

## -----------------------------------------------------------------------------
print(loo_post)

## ----results='hide', warnings=FALSE-------------------------------------------
standata$mixis <- 1
fit_mix <- sampling(stanmodel, data = standata, chains = chains, iter = n_iter, warmup = warm_iter, refresh = 0, pars = "log_lik")
log_lik_mix <- extract(fit_mix)$log_lik

## -----------------------------------------------------------------------------
l_common_mix <- rowLogSumExps(-log_lik_mix)
log_weights <- -log_lik_mix - l_common_mix
elpd_mixis <- logSumExp(-l_common_mix) - rowLogSumExps(t(log_weights))

## -----------------------------------------------------------------------------
data(voice_loo)
elpd_loo <- voice_loo$elpd_loo

## -----------------------------------------------------------------------------
elpd_psis <- loo_post$pointwise[,1]
print(paste("RMSE(PSIS) =",round( sqrt(mean((elpd_loo-elpd_psis)^2)) ,2)))
print(paste("RMSE(MixIS) =",round( sqrt(mean((elpd_loo-elpd_mixis)^2)) ,2)))

## -----------------------------------------------------------------------------
elpd_psis <- loo_post$pointwise[,1]
print(paste("ELPD (PSIS)=",round(sum(elpd_psis),2)))
print(paste("ELPD (MixIS)=",round(sum(elpd_mixis),2)))
print(paste("ELPD (brute force)=",round(sum(elpd_loo),2)))

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----more-knitr-ops, include=FALSE--------------------------------------------
knitr::opts_chunk$set(
  cache = TRUE,
  message = FALSE, 
  warning = FALSE
)

## ----pkgs, cache=FALSE--------------------------------------------------------
library("brms")
library("loo")
library("bayesplot")
library("ggplot2")
color_scheme_set("brightblue")
theme_set(theme_default())

CHAINS <- 4
SEED <- 5838296
set.seed(SEED)

## ----hurondata----------------------------------------------------------------
N <- length(LakeHuron)
df <- data.frame(
  y = as.numeric(LakeHuron),
  year = as.numeric(time(LakeHuron)),
  time = 1:N
)

ggplot(df, aes(x = year, y = y)) + 
  geom_point(size = 1) +
  labs(
    y = "Water Level (ft)", 
    x = "Year",
    title = "Water Level in Lake Huron (1875-1972)"
  ) 

## ----fit, results = "hide"----------------------------------------------------
fit <- brm(
  y ~ ar(time, p = 4), 
  data = df, 
  prior = prior(normal(0, 0.5), class = "ar"),
  control = list(adapt_delta = 0.99), 
  seed = SEED, 
  chains = CHAINS
)

## ----plotpreds, cache = FALSE-------------------------------------------------
preds <- posterior_predict(fit)
preds <- cbind(
  Estimate = colMeans(preds), 
  Q5 = apply(preds, 2, quantile, probs = 0.05),
  Q95 = apply(preds, 2, quantile, probs = 0.95)
)

ggplot(cbind(df, preds), aes(x = year, y = Estimate)) +
  geom_smooth(aes(ymin = Q5, ymax = Q95), stat = "identity", linewidth = 0.5) +
  geom_point(aes(y = y)) + 
  labs(
    y = "Water Level (ft)", 
    x = "Year",
    title = "Water Level in Lake Huron (1875-1972)",
    subtitle = "Mean (blue) and 90% predictive intervals (gray) vs. observed data (black)"
  ) 

## ----setL---------------------------------------------------------------------
L <- 20

## ----loo1sap, cache = FALSE---------------------------------------------------
loo_cv <- loo(log_lik(fit)[, (L + 1):N])
print(loo_cv)

## ----exact_loglik, results="hide"---------------------------------------------
loglik_exact <- matrix(nrow = ndraws(fit), ncol = N)
for (i in L:(N - 1)) {
  past <- 1:i
  oos <- i + 1
  df_past <- df[past, , drop = FALSE]
  df_oos <- df[c(past, oos), , drop = FALSE]
  fit_i <- update(fit, newdata = df_past, recompile = FALSE)
  loglik_exact[, i + 1] <- log_lik(fit_i, newdata = df_oos, oos = oos)[, oos]
}

## ----helpers------------------------------------------------------------------
# some helper functions we'll use throughout

# more stable than log(sum(exp(x))) 
log_sum_exp <- function(x) {
  max_x <- max(x)  
  max_x + log(sum(exp(x - max_x)))
}

# more stable than log(mean(exp(x)))
log_mean_exp <- function(x) {
  log_sum_exp(x) - log(length(x))
}

# compute log of raw importance ratios
# sums over observations *not* over posterior samples
sum_log_ratios <- function(loglik, ids = NULL) {
  if (!is.null(ids)) loglik <- loglik[, ids, drop = FALSE]
  rowSums(loglik)
}

# for printing comparisons later
rbind_print <- function(...) {
  round(rbind(...), digits = 2)
}

## ----exact1sap, cache = FALSE-------------------------------------------------
exact_elpds_1sap <- apply(loglik_exact, 2, log_mean_exp)
exact_elpd_1sap <- c(ELPD = sum(exact_elpds_1sap[-(1:L)]))

rbind_print(
  "LOO" = loo_cv$estimates["elpd_loo", "Estimate"],
  "LFO" = exact_elpd_1sap
)

## ----setkthresh---------------------------------------------------------------
k_thres <- 0.7

## ----refit_loglik, results="hide"---------------------------------------------
approx_elpds_1sap <- rep(NA, N)

# initialize the process for i = L
past <- 1:L
oos <- L + 1
df_past <- df[past, , drop = FALSE]
df_oos <- df[c(past, oos), , drop = FALSE]
fit_past <- update(fit, newdata = df_past, recompile = FALSE)
loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
approx_elpds_1sap[L + 1] <- log_mean_exp(loglik[, oos])

# iterate over i > L
i_refit <- L
refits <- L
ks <- NULL
for (i in (L + 1):(N - 1)) {
  past <- 1:i
  oos <- i + 1
  df_past <- df[past, , drop = FALSE]
  df_oos <- df[c(past, oos), , drop = FALSE]
  loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
  
  logratio <- sum_log_ratios(loglik, (i_refit + 1):i)
  psis_obj <- suppressWarnings(psis(logratio))
  k <- pareto_k_values(psis_obj)
  ks <- c(ks, k)
  if (k > k_thres) {
    # refit the model based on the first i observations
    i_refit <- i
    refits <- c(refits, i)
    fit_past <- update(fit_past, newdata = df_past, recompile = FALSE)
    loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
    approx_elpds_1sap[i + 1] <- log_mean_exp(loglik[, oos])
  } else {
    lw <- weights(psis_obj, normalize = TRUE)[, 1]
    approx_elpds_1sap[i + 1] <- log_sum_exp(lw + loglik[, oos])
  }
} 

## ----plot_ks------------------------------------------------------------------
plot_ks <- function(ks, ids, thres = 0.6) {
  dat_ks <- data.frame(ks = ks, ids = ids)
  ggplot(dat_ks, aes(x = ids, y = ks)) + 
    geom_point(aes(color = ks > thres), shape = 3, show.legend = FALSE) + 
    geom_hline(yintercept = thres, linetype = 2, color = "red2") + 
    scale_color_manual(values = c("cornflowerblue", "darkblue")) + 
    labs(x = "Data point", y = "Pareto k") + 
    ylim(-0.5, 1.5)
}

## ----refitsummary1sap, cache=FALSE--------------------------------------------
cat("Using threshold ", k_thres, 
    ", model was refit ", length(refits), 
    " times, at observations", refits)

plot_ks(ks, (L + 1):(N - 1))

## ----lfosummary1sap, cache = FALSE--------------------------------------------
approx_elpd_1sap <- sum(approx_elpds_1sap, na.rm = TRUE)
rbind_print(
  "approx LFO" = approx_elpd_1sap,
  "exact LFO" = exact_elpd_1sap
)

## ----plot1sap, cache = FALSE--------------------------------------------------
dat_elpd <- data.frame(
  approx_elpd = approx_elpds_1sap,
  exact_elpd = exact_elpds_1sap
)

ggplot(dat_elpd, aes(x = approx_elpd, y = exact_elpd)) +
  geom_abline(color = "gray30") +
  geom_point(size = 2) +
  labs(x = "Approximate ELPDs", y = "Exact ELPDs")

## ----diffs1sap, cache=FALSE---------------------------------------------------
max_diff <- with(dat_elpd, max(abs(approx_elpd - exact_elpd), na.rm = TRUE))
mean_diff <- with(dat_elpd, mean(abs(approx_elpd - exact_elpd), na.rm = TRUE))

rbind_print(
  "Max diff" = round(max_diff, 2), 
  "Mean diff" =  round(mean_diff, 3)
)

## ----exact_loglikm, results="hide"--------------------------------------------
M <- 4
loglikm <- matrix(nrow = ndraws(fit), ncol = N)
for (i in L:(N - M)) {
  past <- 1:i
  oos <- (i + 1):(i + M)
  df_past <- df[past, , drop = FALSE]
  df_oos <- df[c(past, oos), , drop = FALSE]
  fit_past <- update(fit, newdata = df_past, recompile = FALSE)
  loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
  loglikm[, i + 1] <- rowSums(loglik[, oos])
}

## ----exact4sap, cache = FALSE-------------------------------------------------
exact_elpds_4sap <- apply(loglikm, 2, log_mean_exp)
(exact_elpd_4sap <- c(ELPD = sum(exact_elpds_4sap, na.rm = TRUE)))

## ----refit_loglikm, results="hide"--------------------------------------------
approx_elpds_4sap <- rep(NA, N)

# initialize the process for i = L
past <- 1:L
oos <- (L + 1):(L + M)
df_past <- df[past, , drop = FALSE]
df_oos <- df[c(past, oos), , drop = FALSE]
fit_past <- update(fit, newdata = df_past, recompile = FALSE)
loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
loglikm <- rowSums(loglik[, oos])
approx_elpds_4sap[L + 1] <- log_mean_exp(loglikm)

# iterate over i > L
i_refit <- L
refits <- L
ks <- NULL
for (i in (L + 1):(N - M)) {
  past <- 1:i
  oos <- (i + 1):(i + M)
  df_past <- df[past, , drop = FALSE]
  df_oos <- df[c(past, oos), , drop = FALSE]
  loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
  
  logratio <- sum_log_ratios(loglik, (i_refit + 1):i)
  psis_obj <- suppressWarnings(psis(logratio))
  k <- pareto_k_values(psis_obj)
  ks <- c(ks, k)
  if (k > k_thres) {
    # refit the model based on the first i observations
    i_refit <- i
    refits <- c(refits, i)
    fit_past <- update(fit_past, newdata = df_past, recompile = FALSE)
    loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
    loglikm <- rowSums(loglik[, oos])
    approx_elpds_4sap[i + 1] <- log_mean_exp(loglikm)
  } else {
    lw <- weights(psis_obj, normalize = TRUE)[, 1]
    loglikm <- rowSums(loglik[, oos])
    approx_elpds_4sap[i + 1] <- log_sum_exp(lw + loglikm)
  }
} 

## ----refitsummary4sap, cache = FALSE------------------------------------------
cat("Using threshold ", k_thres, 
    ", model was refit ", length(refits), 
    " times, at observations", refits)

plot_ks(ks, (L + 1):(N - M))

## ----lfosummary4sap, cache = FALSE--------------------------------------------
approx_elpd_4sap <- sum(approx_elpds_4sap, na.rm = TRUE)
rbind_print(
  "Approx LFO" = approx_elpd_4sap,
  "Exact LFO" = exact_elpd_4sap
)

## ----plot4sap, cache = FALSE--------------------------------------------------
dat_elpd_4sap <- data.frame(
  approx_elpd = approx_elpds_4sap,
  exact_elpd = exact_elpds_4sap
)

ggplot(dat_elpd_4sap, aes(x = approx_elpd, y = exact_elpd)) +
  geom_abline(color = "gray30") +
  geom_point(size = 2) +
  labs(x = "Approximate ELPDs", y = "Exact ELPDs")

## ----sessioninfo--------------------------------------------------------------
sessionInfo()

#' Extract pointwise log-likelihood from a Stan model
#'
#' Convenience function for extracting the pointwise log-likelihood
#' matrix or array from a `stanfit` object from the \pkg{rstan} package.
#' Note: recent versions of \pkg{rstan} now include a `loo()` method for
#' `stanfit` objects that handles this internally.
#'
#' @export
#' @param stanfit A `stanfit` object (\pkg{rstan} package).
#' @param parameter_name A character string naming the parameter (or generated
#'   quantity) in the Stan model corresponding to the log-likelihood.
#' @param merge_chains If `TRUE` (the default), all Markov chains are
#'   merged together (i.e., stacked) and a matrix is returned. If `FALSE`
#'   they are kept separate and an array is returned.
#' @return If `merge_chains=TRUE`, an \eqn{S} by \eqn{N} matrix of
#'   (post-warmup) extracted draws, where \eqn{S} is the size of the posterior
#'   sample and \eqn{N} is the number of data points. If
#'   `merge_chains=FALSE`, an \eqn{I} by \eqn{C} by \eqn{N} array, where
#'   \eqn{I \times C = S}{I * C = S}.
#'
#'
#' @details Stan does not automatically compute and store the log-likelihood. It
#'   is up to the user to incorporate it into the Stan program if it is to be
#'   extracted after fitting the model. In a Stan model, the pointwise log
#'   likelihood can be coded as a vector in the transformed parameters block
#'   (and then summed up in the model block) or it can be coded entirely in the
#'   generated quantities block. We recommend using the generated quantities
#'   block so that the computations are carried out only once per iteration
#'   rather than once per HMC leapfrog step.
#'
#'   For example, the following is the `generated quantities` block for
#'   computing and saving the log-likelihood for a linear regression model with
#'   `N` data points, outcome `y`, predictor matrix `X`,
#'   coefficients `beta`, and standard deviation `sigma`:
#'
#'  `vector[N] log_lik;`
#'
#'  `for (n in 1:N) log_lik[n] = normal_lpdf(y[n] | X[n, ] * beta, sigma);`
#'
#' @references
#' Stan Development Team (2017). The Stan C++ Library, Version 2.16.0.
#' <https://mc-stan.org/>
#'
#' Stan Development Team (2017). RStan: the R interface to Stan, Version 2.16.1.
#' <https://mc-stan.org/>
#'
extract_log_lik <-
  function(stanfit,
           parameter_name = "log_lik",
           merge_chains = TRUE) {
    if (!inherits(stanfit, "stanfit"))
      stop("Not a stanfit object.", call. = FALSE)
    if (stanfit@mode != 0)
      stop("Stan model does not contain posterior draws.", call. = FALSE)
    if (!requireNamespace("rstan", quietly = TRUE))
      stop("Please load the 'rstan' package.", call. = FALSE)

    if (merge_chains) {
      log_lik <- as.matrix(stanfit, pars = parameter_name)
    } else {
      log_lik <- as.array(stanfit, pars = parameter_name)
    }
    unname(log_lik)
  }

test_that("extract_log_lik throws appropriate errors", {
  x1 <- rnorm(100)
  expect_error(extract_log_lik(x1), regexp = "Not a stanfit object")
  x2 <- structure(x1, class = "stanfit")
  expect_error(extract_log_lik(x2)) # not an S4 object OR no applicable method (depending on R version)
})

library(loo)

context("extract_log_lik")
test_that("extract_log_lik throws appropriate errors", {
  x1 <- rnorm(100)
  expect_error(extract_log_lik(x1), regexp = "Not a stanfit object")
  x2 <- structure(x1, class = "stanfit")
  expect_error(extract_log_lik(x2)) # not an S4 object OR no applicable method (depending on R version)
})

.onAttach <- function(...) {
  ver <- utils::packageVersion("loo")
  packageStartupMessage("This is loo version ", ver)
  packageStartupMessage(
    "- Online documentation and vignettes at mc-stan.org/loo"
  )
  packageStartupMessage(
    "- As of v2.0.0 loo defaults to 1 core ",
    "but we recommend using as many as possible. ",
    "Use the 'cores' argument or set options(mc.cores = NUM_CORES) ",
    "for an entire session. "
  )
  if (os_is_windows()) {
    packageStartupMessage(
      "- Windows 10 users: loo may be very slow if 'mc.cores' ",
      "is set in your .Rprofile file (see https://github.com/stan-dev/loo/issues/94)."
    )
  }
}

#' LOO package glossary
#'
#' @name loo-glossary
#'
#' @template loo-and-psis-references
#' @template loo-uncertainty-reference
#' @template bayesvis-reference
#'
#' @description
#'   This page provides definitions of key terms. Also see the
#'   [FAQ page](https://mc-stan.org/loo/articles/online-only/faq.html) on
#'   the __loo__ website for answers to frequently asked questions.
#'
#'   Note: VGG2017 refers to Vehtari, Gelman, and Gabry (2017). See
#'   **References**, below.
#'
#' @section ELPD and `elpd_loo`:
#'
#' The ELPD is the theoretical expected log pointwise predictive density for a new
#' dataset (Eq 1 in VGG2017), which can be estimated, e.g., using
#' cross-validation. `elpd_loo` is the Bayesian LOO estimate of the
#' expected log pointwise predictive density (Eq 4 in VGG2017) and
#' is a sum of N individual pointwise log predictive densities. Probability
#' densities can be smaller or larger than 1, and thus log predictive densities
#' can be negative or positive. For simplicity the ELPD acronym is used also for
#' expected log pointwise predictive probabilities for discrete models.
#' Probabilities are always equal or less than 1, and thus log predictive
#' probabilities are 0 or negative.
#'
#' @section Standard error of `elpd_loo`:
#'
#' As `elpd_loo` is defined as the sum of N independent components (Eq 4 in
#' VGG2017), we can compute the standard error by using the standard deviation
#' of the N components and multiplying by `sqrt(N)` (Eq 23 in VGG2017).
#' This standard error is a coarse description of our uncertainty about the
#' predictive performance for unknown future data. When N is small or there is
#' severe model misspecification, the current SE estimate is overoptimistic and
#' the actual SE can even be twice as large. Even for moderate N, when the SE
#' estimate is an accurate estimate for the scale, it ignores the skewness. When
#' making model comparisons, the SE of the component-wise (pairwise) differences
#' should be used instead (see the `se_diff` section below and Eq 24 in
#' VGG2017). Sivula et al. (2022) discuss the conditions when the normal
#' approximation used for SE and `se_diff` is good.
#'
#' @section Monte Carlo SE of elpd_loo:
#'
#' The Monte Carlo standard error is the estimate for the computational accuracy
#' of MCMC and importance sampling used to compute `elpd_loo`. Usually this
#' is negligible compared to the standard describing the uncertainty due to
#' finite number of observations (Eq 23 in VGG2017).
#'
#' @section `p_loo` (effective number of parameters):
#'
#' `p_loo` is the difference between `elpd_loo` and the non-cross-validated
#' log posterior predictive density. It describes how much more difficult it
#' is to predict future data than the observed data. Asymptotically under
#' certain regularity conditions, `p_loo` can be interpreted as the
#' *effective number of parameters*. In well behaving cases `p_loo < N` and
#' `p_loo < p`, where `p` is the total number of parameters in the
#' model. `p_loo > N`  or `p_loo > p` indicates that the model has very
#' weak predictive capability and may indicate a severe model misspecification.
#' See below for more on interpreting `p_loo` when there are warnings
#' about high Pareto k diagnostic values.
#'
#' @section Pareto k estimates:
#'
#' The Pareto \eqn{k} estimate is a diagnostic for Pareto smoothed importance
#' sampling (PSIS), which is used to compute components of `elpd_loo`. In
#' importance-sampling LOO the full posterior distribution is used as the
#' proposal distribution. The Pareto k diagnostic estimates how far an
#' individual leave-one-out distribution is from the full distribution. If
#' leaving out an observation changes the posterior too much then importance
#' sampling is not able to give a reliable estimate. Pareto smoothing stabilizes
#' importance sampling and guarantees a finite variance estimate at the
#' cost of some bias.
#'
#' The diagnostic threshold for Pareto \eqn{k} depends on sample size
#' \eqn{S} (sample size dependent threshold was introduced by Vehtari
#' et al., 2024, and before that fixed thresholds of 0.5 and 0.7 were
#' recommended). For simplicity, `loo` package uses the nominal sample
#' size \eqn{S}  when computing the sample size specific
#' threshold. This provides an optimistic threshold if the effective
#' sample size is less than 2200, but even then if ESS/S > 1/2 the difference
#' is usually negligible. Thinning of MCMC draws can be used to improve
#' the ratio ESS/S.
#'
#' * If \eqn{k < \min(1 - 1 / \log_{10}(S), 0.7)}, where \eqn{S} is the
#'   sample size, the PSIS estimate and the corresponding Monte
#'   Carlo standard error estimate are reliable.
#'
#' * If \eqn{1 - 1 / \log_{10}(S) <= k < 0.7}, the PSIS estimate and the
#'   corresponding Monte Carlo standard error estimate are not
#'   reliable, but increasing the (effective) sample size \eqn{S} above
#'   2200 may help (this will increase the sample size specific
#'   threshold \eqn{(1 - 1 / \log_{10}(2200) > 0.7} and then the bias specific
#'   threshold 0.7 dominates).
#'
#' * If \eqn{0.7 <= k < 1}, the PSIS estimate and the corresponding Monte
#'   Carlo standard error have large bias and are not reliable. Increasing
#'   the sample size may reduce the variability in the \eqn{k} estimate, which
#'   may also result in a lower \eqn{k} estimate.
#'
#' * If \eqn{k \geq 1}{k >= 1}, the target distribution is estimated to
#'   have non-finite mean. The PSIS estimate and the corresponding Monte
#'   Carlo standard error are not well defined. Increasing the sample size
#'   may reduce the variability in \eqn{k} estimate, which may also result in
#'   a lower \eqn{k} estimate.
#'
#' Pareto \eqn{k} is also useful as a measure of influence of an
#' observation.  Highly influential observations have high \eqn{k}
#' values. Very high \eqn{k} values often indicate model
#' misspecification, outliers or mistakes in data processing. See
#' Section 6 of Gabry et al. (2019) for an example.
#'
#' \subsection{Interpreting `p_loo` when Pareto `k` is large}{
#' If \eqn{k > 0.7} then we can also look at
#' the `p_loo` estimate for some additional information about the problem:
#'
#' * If `p_loo << p` (the total number of parameters in the model),
#' then the model is likely to be misspecified. Posterior predictive checks
#' (PPCs) are then likely to also detect the problem. Try using an overdispersed
#' model, or add more structural information (nonlinearity, mixture model,
#' etc.).
#'
#' * If `p_loo < p` and the number of parameters `p` is relatively
#' large compared to the number of observations (e.g., `p>N/5`), it is
#' likely that the model is so flexible or the population prior so weak that it’s
#' difficult to predict the left out observation (even for the true model).
#' This happens, for example, in the simulated 8 schools (in VGG2017), random
#' effect models with a few observations per random effect, and Gaussian
#' processes and spatial models with short correlation lengths.
#'
#' * If `p_loo > p`, then the model is likely to be badly misspecified.
#' If the number of parameters `p<<N`, then PPCs are also likely to detect the
#' problem. See the case study at
#' <https://avehtari.github.io/modelselection/roaches.html> for an example.
#' If `p` is relatively large compared to the number of
#' observations, say `p>N/5` (more accurately we should count number of
#' observations influencing each parameter as in hierarchical models some groups
#' may have few observations and other groups many), it is possible that PPCs won't
#' detect the problem.
#' }
#'
#' @section elpd_diff:
#' `elpd_diff` is the difference in `elpd_loo` for two models. If more
#' than two models are compared, the difference is computed relative to the
#' model with highest `elpd_loo`.
#'
#' @section se_diff:
#'
#' The standard error of component-wise differences of elpd_loo (Eq 24 in
#' VGG2017) between two models. This SE is *smaller* than the SE for
#' individual models due to correlation (i.e., if some observations are easier
#' and some more difficult to predict for all models).
#'
#' @section `p_worse` (probability of worse predictive performance):
#'
#' `p_worse` is the estimated probability that a model has worse predictive
#' performance than the best-ranked model in the comparison, based on the normal
#' approximation to the uncertainty in `elpd_diff`. It is computed as
#'
#'     p_worse = pnorm(0, elpd_diff, se_diff).
#'
#' The best-ranked model (the first row in the `loo_compare()` output, where
#' `elpd_diff = 0`) always receives `NA`, since the comparison is defined
#' relative to that model.
#'
#' Because models are ordered by `elpd_loo` before computing `p_worse`, all
#' reported values are at least 0.5 by construction. A value close to 0.5
#' indicates that the models are nearly indistinguishable in predictive
#' performance and that the ranking could easily be reversed with different
#' data. A value close to 1 indicates that the lower-ranked model is almost
#' certainly worse. `p_worse` inherits all the limitations of `se_diff` and the
#' normal approximation on which it is based. In particular, when `se_diff` is
#' underestimated, `p_worse` will be estimated too close to 1, making a model
#' appear more clearly worse than the data actually support. Conversely, when
#' `elpd_diff` is biased due to an unreliable LOO approximation, `p_worse` can
#' point in the wrong direction entirely. When any of these conditions are
#' present, `diag_diff` or `diag_elpd` will be flagged in the `loo_compare()`
#' output. 
#' For further guidance, see the sections below and the case study on
#' [Uncertainty in Bayesian LOO-CV Model Comparison](
#' https://users.aalto.fi/~ave/casestudies/LOO_uncertainty/loo_uncertainty.html).
#'
#' @section `diag_diff` (pairwise comparison diagnostics):
#'
#' `diag_diff` is a diagnostic column in the `loo_compare()` output for each
#' model comparison against the current reference model. It flags conditions
#' under which the normal approximation behind `se_diff` and `p_worse` is likely
#' to be poorly calibrated. The column contains a short label when a condition
#' is detected, and is empty otherwise.
#'
#' The column `diag_diff` currently flags two problems:
#'
#' ### `N < 100`
#'
#' When the number of observations is small, we may assume `se_diff` to be
#' underestimated. As a rough heuristic one can multiply `se_diff` by 2 to
#' make a more conservative estimate.
#'
#' ### `|elpd_diff| < 4`
#'
#' When `|elpd_diff|` is below 4, the models have very similar predictive
#' performance. In this setting, Sivula et al. (2025) show that skewness in
#' the error distribution can make the normal approximation for `se_diff`
#' and `p_worse` miscalibrated, even for large N. In practice, this usually
#' supports treating the models as predictively similar.
#'
#' ### Relation between `N < 100` and `|elpd_diff| < 4`
#'
#' The conditions flagged by `diag_diff` are not independent: they tend to
#' co-occur, and when they do, some flags carry more information than others.
#' `loo_compare()` therefore follows a priority hierarchy and shows only the
#' most critical flag in the table output.
#'
#' The hierarchy is as follows:
#'
#' * **`N < 100` takes highest priority.** A small sample size undermines the
#' reliability of `se_diff` by underestimating uncertainty. Because of this,
#' even if `|elpd_diff| < 4` is also true for a comparison, the table will only
#' show `N < 100`. The small sample size renders the `|elpd_diff| < 4`
#' diagnostic less meaningful.
#'
#' * **`|elpd_diff| < 4` takes second priority.** When N >= 100 and the
#'   difference is small, the normal approximation is miscalibrated due to the
#'   skewness of the error distribution (Sivula et al., 2025). In this
#'   situation, `se_diff` exists and is not heavily biased in scale, but the
#'   shape of the approximation is wrong, making `p_worse` unreliable.
#'
#' For further guidance, see the case study on
#' [Uncertainty in Bayesian LOO-CV Model Comparison](
#' https://users.aalto.fi/~ave/casestudies/LOO_uncertainty/loo_uncertainty.html).
#' 
#' @section `diag_elpd`:
#'
#' `diag_elpd` is a diagnostic column in the `loo_compare()` output that flags
#' when the PSIS-LOO approximation for an individual model is unreliable. Unlike
#' `diag_diff`, which concerns the *comparison* between models, `diag_elpd`
#' concerns the quality of the `elpd_loo` estimate for each model individually.
#' It contains a short text label when a problem is detected, and is empty
#' otherwise.
#'
#' ### `K k_psis > t` (K observations with Pareto-k values > t)
#'
#' This label indicates that K observations for this model have Pareto-k values
#' above the PSIS reliability threshold `t` used by `loo` for that fit. The
#' threshold is sample-size dependent, and in many practical cases close to
#' 0.7. When this flag appears, the PSIS approximation can be unreliable for
#' those observations, and the resulting `elpd_loo` may be biased. Because
#' `elpd_diff` is a direct difference of two models' `elpd_loo` values, bias in
#' either model's estimate propagates directly into `elpd_diff` and `p_worse`.
#' This is qualitatively different from the calibration issues flagged by
#' `diag_diff`: here the estimate itself may be wrong, not just uncertain.
#'
#' See for further information on Pareto-k values the "Pareto k estimates"
#' section.
NULL

#' Moment matching for efficient approximate leave-one-out cross-validation (LOO)
#'
#' Moment matching algorithm for updating a loo object when Pareto k estimates
#' are large.
#'
#' @export loo_moment_match loo_moment_match.default
#' @param x A fitted model object.
#' @param loo A loo object to be modified.
#' @param post_draws A function the takes `x` as the first argument and returns
#'   a matrix of posterior draws of the model parameters.
#' @param log_lik_i A function that takes `x` and `i` and returns a matrix (one
#'   column per chain) or a vector (all chains stacked) of log-likelihood draws
#'   of the `i`th observation based on the model `x`. If the draws are obtained
#'   using MCMC, the matrix with MCMC chains separated is preferred.
#' @param unconstrain_pars A function that takes arguments `x`, and `pars` and
#'   returns posterior draws on the unconstrained space based on the posterior
#'   draws on the constrained space passed via `pars`.
#' @param log_prob_upars A function that takes arguments `x` and `upars` and
#'   returns a matrix of log-posterior density values of the unconstrained
#'   posterior draws passed via `upars`.
#' @param log_lik_i_upars A function that takes arguments `x`, `upars`, and `i`
#'   and returns a vector of log-likelihood draws of the `i`th observation based
#'   on the unconstrained posterior draws passed via `upars`.
#' @param max_iters Maximum number of moment matching iterations. Usually this
#'   does not need to be modified. If the maximum number of iterations is
#'   reached, there will be a warning, and increasing `max_iters` may improve
#'   accuracy.
#' @param k_threshold Threshold value for Pareto k values above which the moment
#'   matching algorithm is used. The default value is `min(1 - 1/log10(S), 0.7)`,
#'   where `S` is the sample size.
#' @param split Logical; Indicate whether to do the split transformation or not
#'   at the end of moment matching for each LOO fold.
#' @param cov Logical; Indicate whether to match the covariance matrix of the
#'   samples or not. If `FALSE`, only the mean and marginal variances are
#'   matched.
#' @template cores
#' @param ... Further arguments passed to the custom functions documented above.
#'
#' @return The `loo_moment_match()` methods return an updated `loo` object. The
#'   structure of the updated `loo` object is similar, but the method also
#'   stores the original Pareto k diagnostic values in the diagnostics field.
#'
#' @details The `loo_moment_match()` function is an S3 generic and we provide a
#'   default method that takes as arguments user-specified functions
#'   `post_draws`, `log_lik_i`, `unconstrain_pars`, `log_prob_upars`, and
#'   `log_lik_i_upars`. All of these functions should take `...`. as an argument
#'   in addition to those specified for each function.
#'
#' @seealso [loo()], [loo_moment_match_split()]
#' @template moment-matching-references
#'
#' @examples
#' # See the vignette for loo_moment_match()

#' @export
loo_moment_match <- function(x, ...) {
  UseMethod("loo_moment_match")
}


#' @describeIn loo_moment_match A default method that takes as arguments a
#'   user-specified model object `x`, a `loo` object and user-specified
#'   functions `post_draws`, `log_lik_i`, `unconstrain_pars`, `log_prob_upars`,
#'   and `log_lik_i_upars`.
#' @export
loo_moment_match.default <- function(x, loo, post_draws, log_lik_i,
                          unconstrain_pars, log_prob_upars,
                          log_lik_i_upars, max_iters = 30L,
                          k_threshold = NULL, split = TRUE,
                          cov = TRUE, cores = getOption("mc.cores", 1),
                          ...) {

  # input checks
  checkmate::assertClass(loo,classes = "loo")
  checkmate::assertFunction(post_draws)
  checkmate::assertFunction(log_lik_i)
  checkmate::assertFunction(unconstrain_pars)
  checkmate::assertFunction(log_prob_upars)
  checkmate::assertFunction(log_lik_i_upars)
  checkmate::assertNumber(max_iters)
  checkmate::assertNumber(k_threshold, null.ok=TRUE)
  checkmate::assertLogical(split)
  checkmate::assertLogical(cov)
  checkmate::assertNumber(cores)


  if ("psis_loo" %in% class(loo)) {
    is_method <- "psis"
  } else {
    stop("loo_moment_match currently supports only the \"psis\" importance sampling class.")
  }


  S <- dim(loo)[1]
  N <- dim(loo)[2]
  if (is.null(k_threshold)) {
    k_threshold <- ps_khat_threshold(S)
  }
  pars <- post_draws(x, ...)
  # transform the model parameters to unconstrained space
  upars <- unconstrain_pars(x, pars = pars, ...)
  # number of parameters in the **parameters** block only
  npars <- dim(upars)[2]
  # if more parameters than samples, do not do Cholesky transformation
  cov <- cov && S >= 10 * npars
  # compute log-probabilities of the original parameter values
  orig_log_prob <- log_prob_upars(x, upars = upars, ...)

  # loop over all observations whose Pareto k is high
  ks <- loo$diagnostics$pareto_k
  kfs <- rep(0,N)
  I <- which(ks > k_threshold)

  loo_moment_match_i_fun <- function(i) {
    loo_moment_match_i(i = i, x = x, log_lik_i = log_lik_i,
                       unconstrain_pars = unconstrain_pars,
                       log_prob_upars = log_prob_upars,
                       log_lik_i_upars = log_lik_i_upars,
                       max_iters = max_iters, k_threshold = k_threshold,
                       split = split, cov = cov, N = N, S = S, upars = upars,
                       orig_log_prob = orig_log_prob, k = ks[i],
                       is_method = is_method, npars = npars, ...)
  }

  if (cores == 1) {
    mm_list <- lapply(X = I, FUN = function(i) loo_moment_match_i_fun(i))
  }
  else {
    if (!os_is_windows()) {
      mm_list <- parallel::mclapply(X = I, mc.cores = cores,
                                    FUN = function(i) loo_moment_match_i_fun(i))
    }
    else {
      cl <- parallel::makePSOCKcluster(cores)
      on.exit(parallel::stopCluster(cl))
      mm_list <- parallel::parLapply(cl = cl, X = I,
                                    fun = function(i) loo_moment_match_i_fun(i))
    }
  }

  # update results
  for (ii in seq_along(I)) {
    i <- mm_list[[ii]]$i
    loo$pointwise[i, "elpd_loo"] <- mm_list[[ii]]$elpd_loo_i
    loo$pointwise[i, "p_loo"] <- mm_list[[ii]]$p_loo
    loo$pointwise[i, "mcse_elpd_loo"] <- mm_list[[ii]]$mcse_elpd_loo
    loo$pointwise[i, "looic"] <- mm_list[[ii]]$looic

    loo$diagnostics$pareto_k[i] <- mm_list[[ii]]$k
    loo$diagnostics$n_eff[i] <- mm_list[[ii]]$n_eff
    kfs[i] <- mm_list[[ii]]$kf

    if (!is.null(loo$psis_object)) {
      loo$psis_object$log_weights[, i] <- mm_list[[ii]]$lwi
    }
  }
  if (!is.null(loo$psis_object)) {
    attr(loo$psis_object, "norm_const_log") <- matrixStats::colLogSumExps(loo$psis_object$log_weights)
    loo$psis_object$diagnostics <- loo$diagnostics
  }

  # combined estimates
  cols_to_summarize <- !(colnames(loo$pointwise) %in% c("mcse_elpd_loo",
                                                        "influence_pareto_k"))
  loo$estimates <- table_of_estimates(loo$pointwise[, cols_to_summarize,
                                                    drop = FALSE])

  # these will be deprecated at some point
  loo$elpd_loo <- loo$estimates["elpd_loo","Estimate"]
  loo$p_loo <- loo$estimates["p_loo","Estimate"]
  loo$looic <- loo$estimates["looic","Estimate"]
  loo$se_elpd_loo <- loo$estimates["elpd_loo","SE"]
  loo$se_p_loo <- loo$estimates["p_loo","SE"]
  loo$se_looic <- loo$estimates["looic","SE"]

  # Warn if some Pareto ks are still high
  throw_pareto_warnings(loo$diagnostics$pareto_k, k_threshold)
  # if we don't split, accuracy may be compromised
  if (!split) {
    throw_large_kf_warning(kfs, k_threshold)
  }

  loo
}




# Internal functions ---------------


#' Do moment matching for a single observation.
#'
#' @noRd
#' @param i observation number.
#' @param x A fitted model object.
#' @param log_lik_i A function that takes `x` and `i` and returns a matrix (one
#'   column per chain) or a vector (all chains stacked) of log-likelihood draws
#'   of the `i`th observation based on the model `x`. If the draws are obtained
#'   using MCMC, the matrix with MCMC chains separated is preferred.
#' @param unconstrain_pars A function that takes arguments `x`, and `pars` and
#'   returns posterior draws on the unconstrained space based on the posterior
#'   draws on the constrained space passed via `pars`.
#' @param log_prob_upars A function that takes arguments `x` and `upars` and
#'   returns a matrix of log-posterior density values of the unconstrained
#'   posterior draws passed via `upars`.
#' @param log_lik_i_upars A function that takes arguments `x`, `upars`, and `i`
#'   and returns a vector of log-likelihood draws of the `i`th observation based
#'   on the unconstrained posterior draws passed via `upars`.
#' @param max_iters Maximum number of moment matching iterations. Usually this
#'   does not need to be modified. If the maximum number of iterations is
#'   reached, there will be a warning, and increasing `max_iters` may improve
#'   accuracy.
#' @param k_threshold Threshold value for Pareto k values above which the moment
#'   matching algorithm is used. The default value is 0.5.
#' @param split Logical; Indicate whether to do the split transformation or not
#'   at the end of moment matching for each LOO fold.
#' @param cov Logical; Indicate whether to match the covariance matrix of the
#'   samples or not. If `FALSE`, only the mean and marginal variances are
#'   matched.
#' @param N Number of observations.
#' @param S number of MCMC draws.
#' @param upars A matrix representing a sample of vector-valued parameters in
#'   the unconstrained space.
#' @param orig_log_prob log probability densities of the original draws from the
#'   model `x`.
#' @param k Pareto k value before moment matching
#' @template is_method
#' @param npars Number of parameters in the model
#' @param ... Further arguments passed to the custom functions documented above.
#' @return List with the updated elpd values and diagnostics
#'
loo_moment_match_i <- function(i,
                               x,
                               log_lik_i,
                               unconstrain_pars,
                               log_prob_upars,
                               log_lik_i_upars,
                               max_iters,
                               k_threshold,
                               split,
                               cov,
                               N,
                               S,
                               upars,
                               orig_log_prob,
                               k,
                               is_method,
                               npars,
                               ...) {
  # initialize values for this LOO-fold
  uparsi <- upars
  ki <- k
  kfi <- 0
  log_liki <- log_lik_i(x, i, ...)
  S_per_chain <- NROW(log_liki)
  N_chains <- NCOL(log_liki)
  dim(log_liki) <- c(S_per_chain, N_chains, 1)
  r_eff_i <- loo::relative_eff(exp(log_liki), cores = 1)
  dim(log_liki) <- NULL
  lpd <- matrixStats::logSumExp(log_liki) - log(length(log_liki))

  is_obj <- suppressWarnings(importance_sampling.default(-log_liki,
                                                         method = is_method,
                                                         r_eff = r_eff_i,
                                                         cores = 1))
  lwi <- as.vector(weights(is_obj))
  lwfi <- rep(-matrixStats::logSumExp(rep(0, S)),S)

  # initialize objects that keep track of the total transformation
  total_shift <- rep(0, npars)
  total_scaling <- rep(1, npars)
  total_mapping <- diag(npars)

  # try several transformations one by one
  # if one does not work, do not apply it and try another one
  # to accept the transformation, Pareto k needs to improve
  # when transformation succeeds, start again from the first one
  iterind <- 1
  while (iterind <= max_iters && ki > k_threshold) {

    if (iterind == max_iters) {
      throw_moment_match_max_iters_warning()
    }

    # 1. match means
    trans <- shift(x, uparsi, lwi)
    # gather updated quantities
    quantities_i <- try(update_quantities_i(x, trans$upars,  i = i,
                                        orig_log_prob = orig_log_prob,
                                        log_prob_upars = log_prob_upars,
                                        log_lik_i_upars = log_lik_i_upars,
                                        r_eff_i = r_eff_i,
                                        cores = 1,
                                        is_method = is_method,
                                        ...)
                        )
    if (inherits(quantities_i, "try-error")) {
      # Stan log prob caused an exception probably due to under- or
      # overflow of parameters to invalid values
      break
    }
    if (quantities_i$ki < ki) {
      uparsi <- trans$upars
      total_shift <- total_shift + trans$shift

      lwi <- quantities_i$lwi
      lwfi <- quantities_i$lwfi
      ki <- quantities_i$ki
      kfi <- quantities_i$kfi
      log_liki <- quantities_i$log_liki
      iterind <- iterind + 1
      next
    }

    # 2. match means and marginal variances
    trans <- shift_and_scale(x, uparsi, lwi)
    # gather updated quantities
    quantities_i <- try(update_quantities_i(x, trans$upars,  i = i,
                                        orig_log_prob = orig_log_prob,
                                        log_prob_upars = log_prob_upars,
                                        log_lik_i_upars = log_lik_i_upars,
                                        r_eff_i = r_eff_i,
                                        cores = 1,
                                        is_method = is_method,
                                        ...)
                        )
    if (inherits(quantities_i, "try-error")) {
      # Stan log prob caused an exception probably due to under- or
      # overflow of parameters to invalid values
      break
    }
    if (quantities_i$ki < ki) {
      uparsi <- trans$upars
      total_shift <- total_shift + trans$shift
      total_scaling <- total_scaling * trans$scaling

      lwi <- quantities_i$lwi
      lwfi <- quantities_i$lwfi
      ki <- quantities_i$ki
      kfi <- quantities_i$kfi
      log_liki <- quantities_i$log_liki
      iterind <- iterind + 1
      next
    }

    # 3. match means and covariances
    if (cov) {
      trans <- shift_and_cov(x, uparsi, lwi)
      # gather updated quantities
      quantities_i <- try(update_quantities_i(x, trans$upars,  i = i,
                                          orig_log_prob = orig_log_prob,
                                          log_prob_upars = log_prob_upars,
                                          log_lik_i_upars = log_lik_i_upars,
                                          r_eff_i = r_eff_i,
                                          cores = 1,
                                          is_method = is_method,
                                          ...)
                          )

      if (inherits(quantities_i, "try-error")) {
        # Stan log prob caused an exception probably due to under- or
        # overflow of parameters to invalid values
        break
      }
      if (quantities_i$ki < ki) {
        uparsi <- trans$upars
        total_shift <- total_shift + trans$shift
        total_mapping <- trans$mapping %*% total_mapping

        lwi <- quantities_i$lwi
        lwfi <- quantities_i$lwfi
        ki <- quantities_i$ki
        kfi <- quantities_i$kfi
        log_liki <- quantities_i$log_liki
        iterind <- iterind + 1
        next
      }
    }
    # none of the transformations improved khat
    # so there is no need to try further
    break
  }

  # transformations are now done
  # if we don't do split transform, or
  # if no transformations were successful
  # stop and collect values
  if (split && (iterind > 1)) {
    # compute split transformation
    split_obj <- loo_moment_match_split(
      x, upars, cov, total_shift, total_scaling, total_mapping, i,
      log_prob_upars = log_prob_upars, log_lik_i_upars = log_lik_i_upars,
      cores = 1, r_eff_i = r_eff_i, is_method = is_method, ...
    )
    log_liki <- split_obj$log_liki
    lwi <- split_obj$lwi
    lwfi <- split_obj$lwfi
    r_eff_i <- split_obj$r_eff_i
  }
  else {
    dim(log_liki) <- c(S_per_chain, N_chains, 1)
    r_eff_i <- loo::relative_eff(exp(log_liki), cores = 1)
    dim(log_liki) <- NULL
  }

  # pointwise estimates
  elpd_loo_i <- matrixStats::logSumExp(log_liki + lwi)
  mcse_elpd_loo <- mcse_elpd(
    ll = as.matrix(log_liki), lw = as.matrix(lwi),
    E_elpd = exp(elpd_loo_i), r_eff = r_eff_i
  )

  list(elpd_loo_i = elpd_loo_i,
       p_loo = lpd - elpd_loo_i,
       mcse_elpd_loo = mcse_elpd_loo,
       looic = -2 * elpd_loo_i,
       k = ki,
       kf = kfi,
       n_eff = min(1.0 / sum(exp(2 * lwi)),
                   1.0 / sum(exp(2 * lwfi))) * r_eff_i,
       lwi = lwi,
       i = i)
}




#' Update the importance weights, Pareto diagnostic and log-likelihood
#' for observation `i` based on model `x`.
#'
#' @noRd
#' @param x A fitted model object.
#' @param upars A matrix representing a sample of vector-valued parameters in
#' the unconstrained space.
#' @param i observation number.
#' @param orig_log_prob log probability densities of the original draws from
#' the model `x`.
#' @param log_prob_upars A function that takes arguments `x` and
#'   `upars` and returns a matrix of log-posterior density values of the
#'   unconstrained posterior draws passed via `upars`.
#' @param log_lik_i_upars A function that takes arguments `x`, `upars`,
#'   and `i` and returns a vector of log-likelihood draws of the `i`th
#'   observation based on the unconstrained posterior draws passed via
#'   `upars`.
#' @param r_eff_i MCMC effective sample size divided by the total sample size
#' for 1/exp(log_ratios) for observation i.
#' @template is_method
#' @return List with the updated importance weights, Pareto diagnostics and
#' log-likelihood values.
#'
update_quantities_i <- function(x, upars, i, orig_log_prob,
                                log_prob_upars, log_lik_i_upars,
                                r_eff_i, is_method, ...) {
  log_prob_new <- log_prob_upars(x, upars = upars, ...)
  log_liki_new <- log_lik_i_upars(x, upars = upars, i = i, ...)
  # compute new log importance weights

  # If log_liki_new and log_prob_new both have same element as Inf,
  # replace the log ratio with -Inf
  lr <- -log_liki_new + log_prob_new - orig_log_prob
  lr[is.na(lr)] <- -Inf
  is_obj_new <- suppressWarnings(importance_sampling.default(lr,
                                                             method = is_method,
                                                             r_eff = r_eff_i,
                                                             cores = 1))
  lwi_new <- as.vector(weights(is_obj_new))
  ki_new <- is_obj_new$diagnostics$pareto_k

  is_obj_f_new <- suppressWarnings(importance_sampling.default(log_prob_new - orig_log_prob,
                                                               method = is_method,
                                                               r_eff = r_eff_i,
                                                               cores = 1))
  lwfi_new <- as.vector(weights(is_obj_f_new))
  kfi_new <- is_obj_f_new$diagnostics$pareto_k

  # gather results
  list(
    lwi = lwi_new,
    lwfi = lwfi_new,
    ki = ki_new,
    kfi = kfi_new,
    log_liki = log_liki_new
  )
}



#' Shift a matrix of parameters to their weighted mean.
#' Also calls update_quantities_i which updates the importance weights based on
#' the supplied model object.
#'
#' @noRd
#' @param x A fitted model object.
#' @param upars A matrix representing a sample of vector-valued parameters in
#' the unconstrained space
#' @param lwi A vector representing the log-weight of each parameter
#' @return List with the shift that was performed, and the new parameter matrix.
#'
shift <- function(x, upars, lwi) {
  # compute moments using log weights
  mean_original <- colMeans(upars)
  mean_weighted <- colSums(exp(lwi) * upars)
  shift <- mean_weighted - mean_original
  # transform posterior draws
  upars_new <- sweep(upars, 2, shift, "+")
  list(
    upars = upars_new,
    shift = shift
  )
}




#' Shift a matrix of parameters to their weighted mean and scale the marginal
#' variances to match the weighted marginal variances. Also calls
#' update_quantities_i which updates the importance weights based on
#' the supplied model object.
#'
#' @noRd
#' @param x A fitted model object.
#' @param upars A matrix representing a sample of vector-valued parameters in
#' the unconstrained space
#' @param lwi A vector representing the log-weight of each parameter
#' @return List with the shift and scaling that were performed, and the new
#' parameter matrix.
#'
#'
shift_and_scale <- function(x, upars, lwi) {
  # compute moments using log weights
  S <- dim(upars)[1]
  mean_original <- colMeans(upars)
  mean_weighted <- colSums(exp(lwi) * upars)
  shift <- mean_weighted - mean_original
  mii <- exp(lwi)* upars^2
  mii <- colSums(mii) - mean_weighted^2
  mii <- mii*S/(S-1)
  scaling <- sqrt(mii / matrixStats::colVars(upars))
  # transform posterior draws
  upars_new <- sweep(upars, 2, mean_original, "-")
  upars_new <- sweep(upars_new, 2, scaling, "*")
  upars_new <- sweep(upars_new, 2, mean_weighted, "+")

  list(
    upars = upars_new,
    shift = shift,
    scaling = scaling
  )
}

#' Shift a matrix of parameters to their weighted mean and scale the covariance
#' to match the weighted covariance.
#' Also calls update_quantities_i which updates the importance weights based on
#' the supplied model object.
#'
#' @noRd
#' @param x A fitted model object.
#' @param upars A matrix representing a sample of vector-valued parameters in
#' the unconstrained space
#' @param lwi A vector representing the log-weight of each parameter
#' @return List with the shift and mapping that were performed, and the new
#' parameter matrix.
#'
shift_and_cov <- function(x, upars, lwi, ...) {
  # compute moments using log weights
  mean_original <- colMeans(upars)
  mean_weighted <- colSums(exp(lwi) * upars)
  shift <- mean_weighted - mean_original
  covv <- stats::cov(upars)
  wcovv <- stats::cov.wt(upars, wt = exp(lwi))$cov
  chol1 <- tryCatch(
    {
      chol(wcovv)
    },
    error = function(cond)
    {
      return(NULL)
    }
  )
  if (is.null(chol1)) {
    mapping <- diag(length(mean_original))
  }
  else {
    chol2 <- chol(covv)
    mapping <- t(chol1) %*% solve(t(chol2))
  }
  # transform posterior draws
  upars_new <- sweep(upars, 2, mean_original, "-")
  upars_new <- tcrossprod(upars_new, mapping)
  upars_new <- sweep(upars_new, 2, mean_weighted, "+")
  colnames(upars_new) <- colnames(upars)

  list(
    upars = upars_new,
    shift = shift,
    mapping = mapping
  )
}


#' Warning message if max_iters is reached
#' @noRd
throw_moment_match_max_iters_warning <- function() {
  warning(
    "The maximum number of moment matching iterations ('max_iters' argument)
    was reached.\n",
    "Increasing the value may improve accuracy.",
    call. = FALSE
  )
}

#' Warning message if not using split transformation and accuracy is
#' compromised
#' @noRd
throw_large_kf_warning <- function(kf, k_threshold) {
  if (any(kf > k_threshold)) {
    warning(
      "The accuracy of self-normalized importance sampling may be bad.\n",
      "Setting the argument 'split' to 'TRUE' will likely improve accuracy.",
      call. = FALSE
    )
  }
}

options(mc.cores = 1)

set.seed(123)
S <- 4000

# helper functions for sampling from the posterior distribution
rinvchisq <- function(n, df, scale = 1 / df, ...) {
  if ((length(scale) != 1) & (length(scale) != n)) {
    stop("scale should be a scalar or a vector of the same length as x")
  }
  if (df <= 0) {
    stop("df must be greater than zero")
  }
  if (any(scale <= 0)) {
    stop("scale must be greater than zero")
  }
  return((df * scale) / rchisq(n, df = df))
}

dinvchisq <- function(x, df, scale = 1 / df, log = FALSE, ...) {
  if (df <= 0) {
    stop("df must be greater than zero")
  }
  if (scale <= 0) {
    stop("scale must be greater than zero")
  }
  nu <- df / 2
  if (log) {
    return(ifelse(
      x > 0,
      nu *
        log(nu) -
        log(gamma(nu)) +
        nu * log(scale) -
        (nu + 1) * log(x) -
        (nu * scale / x),
      NA
    ))
  } else {
    return(ifelse(
      x > 0,
      (((nu)^(nu)) / gamma(nu)) *
        (scale^nu) *
        (x^(-(nu + 1))) *
        exp(-nu * scale / x),
      NA
    ))
  }
}


# generate toy data
# normally distributed data with known variance
data_sd <- 1.1
data_mean <- 1.3
n <- as.integer(30)
y <- rnorm(n = n, mean = data_mean, sd = data_sd)
y_tilde <- 11
y[1] <- y_tilde

ymean <- mean(y)
s2 <- sum((y - ymean)^2) / (n - 1)

# draws from the posterior distribution when including all observations
draws_full_posterior_sigma2 <- rinvchisq(S, n - 1, s2)
draws_full_posterior_mu <- rnorm(
  S,
  ymean,
  sqrt(draws_full_posterior_sigma2 / n)
)


# create a dummy model object
x <- list()
x$data <- list()
x$data$y <- y
x$data$n <- n
x$data$ymean <- ymean
x$data$s2 <- s2

x$draws <- data.frame(
  mu = draws_full_posterior_mu,
  sigma = sqrt(draws_full_posterior_sigma2)
)


# implement functions for moment matching loo

# extract original posterior draws
post_draws_test <- function(x, ...) {
  as.matrix(x$draws)
}

# extract original log lik draws
log_lik_i_test <- function(x, i, ...) {
  -0.5 *
    log(2 * pi) -
    log(x$draws$sigma) -
    1.0 / (2 * x$draws$sigma^2) * (x$data$y[i] - x$draws$mu)^2
}


loglik <- matrix(0, S, n)
for (j in seq(n)) {
  loglik[, j] <- log_lik_i_test(x, j)
}


# mu, log(sigma)
unconstrain_pars_test <- function(x, pars, ...) {
  upars <- as.matrix(pars)
  upars[, 2] <- log(upars[, 2])
  upars
}

log_prob_upars_test <- function(x, upars, ...) {
  dinvchisq(exp(upars[, 2])^2, x$data$n - 1, x$data$s2, log = TRUE) +
    dnorm(
      upars[, 1],
      x$data$ymean,
      exp(upars[, 2]) / sqrt(x$data$n),
      log = TRUE
    )
}

# compute log_lik_i values based on the unconstrained parameters
log_lik_i_upars_test <- function(x, upars, i, ...) {
  -0.5 *
    log(2 * pi) -
    upars[, 2] -
    1.0 / (2 * exp(upars[, 2])^2) * (x$data$y[i] - upars[, 1])^2
}


upars <- unconstrain_pars_test(x, x$draws)
lwi_1 <- -loglik[, 1]
lwi_1 <- lwi_1 - matrixStats::logSumExp(lwi_1)


test_that("log_prob_upars_test works", {
  upars <- unconstrain_pars_test(x, x$draws)
  xloo <- list()
  xloo$data <- list()
  xloo$data$y <- y[-1]
  xloo$data$n <- n - 1
  xloo$data$ymean <- mean(y[-1])
  xloo$data$s2 <- sum((y[-1] - mean(y[-1]))^2) / (n - 2)

  post1 <- log_prob_upars_test(x, upars)
  post1 <- post1 - matrixStats::logSumExp(post1)
  post2 <- log_prob_upars_test(xloo, upars) + loglik[, 1]
  post2 <- post2 - matrixStats::logSumExp(post2)
  expect_equal(post1, post2)
})


test_that("loo_moment_match.default warnings work", {
  # loo object
  loo_manual <- suppressWarnings(loo(loglik))
  loo_manual_tis <- suppressWarnings(loo(loglik, is_method = "tis"))

  expect_warning(
    loo_moment_match(
      x,
      loo_manual,
      post_draws_test,
      log_lik_i_test,
      unconstrain_pars_test,
      log_prob_upars_test,
      log_lik_i_upars_test,
      max_iters = 30L,
      k_thres = 0.5,
      split = FALSE,
      cov = TRUE,
      cores = 1
    ),
    "The accuracy of self-normalized importance sampling"
  )

  expect_warning(
    loo_moment_match(
      x,
      loo_manual,
      post_draws_test,
      log_lik_i_test,
      unconstrain_pars_test,
      log_prob_upars_test,
      log_lik_i_upars_test,
      max_iters = 30L,
      split = FALSE,
      cov = TRUE,
      cores = 1
    ),
    "The accuracy of self-normalized importance sampling"
  )

  expect_no_warning(loo_moment_match(
    x,
    loo_manual,
    post_draws_test,
    log_lik_i_test,
    unconstrain_pars_test,
    log_prob_upars_test,
    log_lik_i_upars_test,
    max_iters = 30L,
    k_thres = 100,
    split = TRUE,
    cov = TRUE,
    cores = 1
  ))

  expect_snapshot(loo_moment_match(
    x,
    loo_manual,
    post_draws_test,
    log_lik_i_test,
    unconstrain_pars_test,
    log_prob_upars_test,
    log_lik_i_upars_test,
    max_iters = 1,
    k_thres = 0.5,
    split = TRUE,
    cov = TRUE,
    cores = 1
  ))

  expect_error(
    loo_moment_match(
      x,
      loo_manual_tis,
      post_draws_test,
      log_lik_i_test,
      unconstrain_pars_test,
      log_prob_upars_test,
      log_lik_i_upars_test,
      max_iters = 30L,
      k_thres = 0.5,
      split = TRUE,
      cov = TRUE,
      cores = 1
    ),
    "loo_moment_match currently supports only"
  )
})


test_that("loo_moment_match.default works", {
  # allow -Inf
  lwi_x <- lwi_1
  lwi_x[which.min(lwi_1)] <- -Inf
  expect_no_error(suppressWarnings(importance_sampling.default(
    lwi_1,
    method = "psis",
    r_eff = 1,
    cores = 1
  )))

  # loo object
  loo_manual <- suppressWarnings(loo(loglik))

  loo_moment_match_object <- suppressWarnings(loo_moment_match(
    x,
    loo_manual,
    post_draws_test,
    log_lik_i_test,
    unconstrain_pars_test,
    log_prob_upars_test,
    log_lik_i_upars_test,
    max_iters = 30L,
    k_thres = 0.8,
    split = FALSE,
    cov = TRUE,
    cores = 1
  ))

  # diagnostic Pareto k decreases but influence pareto k stays the same
  expect_lt(
    loo_moment_match_object$diagnostics$pareto_k[1],
    loo_moment_match_object$pointwise[1, "influence_pareto_k"]
  )
  expect_equal(
    loo_moment_match_object$pointwise[, "influence_pareto_k"],
    loo_manual$pointwise[, "influence_pareto_k"]
  )
  expect_equal(
    loo_moment_match_object$pointwise[, "influence_pareto_k"],
    loo_manual$diagnostics$pareto_k
  )

  expect_snapshot_value(loo_moment_match_object, style = "serialize")

  loo_moment_match_object2 <- suppressWarnings(loo_moment_match(
    x,
    loo_manual,
    post_draws_test,
    log_lik_i_test,
    unconstrain_pars_test,
    log_prob_upars_test,
    log_lik_i_upars_test,
    max_iters = 30L,
    k_thres = 0.5,
    split = FALSE,
    cov = TRUE,
    cores = 1
  ))

  expect_snapshot_value(loo_moment_match_object2, style = "serialize")

  loo_moment_match_object3 <- suppressWarnings(loo_moment_match(
    x,
    loo_manual,
    post_draws_test,
    log_lik_i_test,
    unconstrain_pars_test,
    log_prob_upars_test,
    log_lik_i_upars_test,
    max_iters = 30L,
    k_thres = 0.5,
    split = TRUE,
    cov = TRUE,
    cores = 1
  ))

  expect_snapshot_value(loo_moment_match_object3, style = "serialize")

  loo_moment_match_object4 <- suppressWarnings(loo_moment_match(
    x,
    loo_manual,
    post_draws_test,
    log_lik_i_test,
    unconstrain_pars_test,
    log_prob_upars_test,
    log_lik_i_upars_test,
    max_iters = 30L,
    k_thres = 100,
    split = FALSE,
    cov = TRUE,
    cores = 1
  ))

  expect_equal(loo_manual, loo_moment_match_object4)

  loo_manual_with_psis <- suppressWarnings(loo(loglik, save_psis = TRUE))
  loo_moment_match_object5 <- suppressWarnings(loo_moment_match(
    x,
    loo_manual_with_psis,
    post_draws_test,
    log_lik_i_test,
    unconstrain_pars_test,
    log_prob_upars_test,
    log_lik_i_upars_test,
    max_iters = 30L,
    k_thres = 0.8,
    split = FALSE,
    cov = TRUE,
    cores = 1
  ))

  expect_equal(
    loo_moment_match_object5$diagnostics,
    loo_moment_match_object5$psis_object$diagnostics
  )
})

test_that("variance and covariance transformations work", {
  S <- 2000

  set.seed(8493874)
  draws_full_posterior_sigma2 <- rinvchisq(S, n - 1, s2)
  draws_full_posterior_mu <- rnorm(
    S,
    ymean,
    sqrt(draws_full_posterior_sigma2 / n)
  )

  x$draws <- data.frame(
    mu = draws_full_posterior_mu,
    sigma = sqrt(draws_full_posterior_sigma2)
  )
  loglik <- matrix(0, S, n)
  for (j in seq(n)) {
    loglik[, j] <- log_lik_i_test(x, j)
  }

  upars <- unconstrain_pars_test(x, x$draws)
  lwi_1 <- -loglik[, 1]
  lwi_1 <- lwi_1 - matrixStats::logSumExp(lwi_1)

  loo_manual <- suppressWarnings(loo(loglik))

  loo_moment_match_object <- suppressWarnings(loo_moment_match(
    x,
    loo_manual,
    post_draws_test,
    log_lik_i_test,
    unconstrain_pars_test,
    log_prob_upars_test,
    log_lik_i_upars_test,
    max_iters = 30L,
    k_thres = 0.0,
    split = FALSE,
    cov = TRUE,
    cores = 1
  ))

  expect_snapshot_value(loo_moment_match_object, style = "serialize")
})

test_that("loo_moment_match.default works with multiple cores", {
  # loo object
  loo_manual <- suppressWarnings(loo(loglik))

  loo_moment_match_manual3 <- suppressWarnings(loo_moment_match(
    x,
    loo_manual,
    post_draws_test,
    log_lik_i_test,
    unconstrain_pars_test,
    log_prob_upars_test,
    log_lik_i_upars_test,
    max_iters = 30L,
    k_thres = 0.5,
    split = FALSE,
    cov = TRUE,
    cores = 1
  ))

  loo_moment_match_manual4 <- suppressWarnings(loo_moment_match(
    x,
    loo_manual,
    post_draws_test,
    log_lik_i_test,
    unconstrain_pars_test,
    log_prob_upars_test,
    log_lik_i_upars_test,
    max_iters = 30L,
    k_thres = 0.5,
    split = FALSE,
    cov = TRUE,
    cores = 2
  ))

  expect_equal(
    loo_moment_match_manual3$diagnostics$pareto_k,
    loo_moment_match_manual4$diagnostics$pareto_k
  )
  expect_equal(
    loo_moment_match_manual3$diagnostics$n_eff,
    loo_moment_match_manual4$diagnostics$n_eff,
    tolerance = 5e-4
  )

  expect_equal(
    loo_moment_match_manual3$estimates,
    loo_moment_match_manual4$estimates
  )

  expect_equal(
    loo_moment_match_manual3$pointwise,
    loo_moment_match_manual4$pointwise,
    tolerance = 5e-4
  )
})


test_that("loo_moment_match_split works", {
  # skip on M1 Mac until we figure out why this test fails only on M1 Mac
  skip_if(Sys.info()[["sysname"]] == "Darwin" && R.version$arch == "aarch64")

  is_obj_1 <- suppressWarnings(importance_sampling.default(
    lwi_1,
    method = "psis",
    r_eff = 1,
    cores = 1
  ))
  lwi_1_ps <- as.vector(weights(is_obj_1))

  split <- loo_moment_match_split(
    x,
    upars,
    cov = FALSE,
    total_shift = c(0, 0),
    total_scaling = c(1, 1),
    total_mapping = diag(c(1, 1)),
    i = 1,
    log_prob_upars = log_prob_upars_test,
    log_lik_i_upars = log_lik_i_upars_test,
    cores = 1,
    r_eff_i = 1,
    is_method = "psis"
  )

  expect_named(split, c("lwi", "lwfi", "log_liki", "r_eff_i"))

  expect_equal(lwi_1_ps, split$lwi)

  split2 <- loo_moment_match_split(
    x,
    upars,
    cov = FALSE,
    total_shift = c(-0.1, -0.2),
    total_scaling = c(0.7, 0.7),
    total_mapping = matrix(c(1, 0.1, 0.1, 1), 2, 2),
    i = 1,
    log_prob_upars = log_prob_upars_test,
    log_lik_i_upars = log_lik_i_upars_test,
    cores = 1,
    r_eff_i = 1,
    is_method = "psis"
  )

  expect_snapshot_value(split2, style = "serialize")
})

test_that("passing arguments works", {
  log_lik_i_upars_test_additional_argument <- function(
    x,
    upars,
    i,
    passed_arg = FALSE,
    ...
  ) {
    if (!passed_arg) {
      warning("passed_arg was not passed here")
    }
    -0.5 *
      log(2 * pi) -
      upars[, 2] -
      1.0 / (2 * exp(upars[, 2])^2) * (x$data$y[i] - upars[, 1])^2
  }
  unconstrain_pars_test_additional_argument <- function(
    x,
    pars,
    passed_arg = FALSE,
    ...
  ) {
    if (!passed_arg) {
      warning("passed_arg was not passed here")
    }
    upars <- as.matrix(pars)
    upars[, 2] <- log(upars[, 2])
    upars
  }

  log_prob_upars_test_additional_argument <- function(
    x,
    upars,
    passed_arg = FALSE,
    ...
  ) {
    if (!passed_arg) {
      warning("passed_arg was not passed here")
    }
    dinvchisq(exp(upars[, 2])^2, x$data$n - 1, x$data$s2, log = TRUE) +
      dnorm(
        upars[, 1],
        x$data$ymean,
        exp(upars[, 2]) / sqrt(x$data$n),
        log = TRUE
      )
  }
  post_draws_test_additional_argument <- function(x, passed_arg = FALSE, ...) {
    if (!passed_arg) {
      warning("passed_arg was not passed here")
    }
    as.matrix(x$draws)
  }
  log_lik_i_test_additional_argument <- function(
    x,
    i,
    passed_arg = FALSE,
    ...
  ) {
    if (!passed_arg) {
      warning("passed_arg was not passed here")
    }
    -0.5 *
      log(2 * pi) -
      log(x$draws$sigma) -
      1.0 / (2 * x$draws$sigma^2) * (x$data$y[i] - x$draws$mu)^2
  }

  # loo object
  loo_manual <- suppressWarnings(loo(loglik))
  expect_silent(loo_moment_match(
    x,
    loo_manual,
    post_draws_test_additional_argument,
    log_lik_i_test_additional_argument,
    unconstrain_pars_test_additional_argument,
    log_prob_upars_test_additional_argument,
    log_lik_i_upars_test_additional_argument,
    max_iters = 30L,
    k_thres = 0.5,
    split = TRUE,
    cov = TRUE,
    cores = 1,
    passed_arg = TRUE
  ))
})

params <-
list(EVAL = TRUE)

## ---- SETTINGS-knitr, include=FALSE--------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----more-knitr-ops, include=FALSE---------------------------------------
knitr::opts_chunk$set(
  cache=TRUE,
  message=FALSE, 
  warning=FALSE
)

## ----lpdf, eval=FALSE----------------------------------------------------
#  /**
#   * Normal log-pdf for spatially lagged responses
#   *
#   * @param y Vector of response values.
#   * @param mu Mean parameter vector.
#   * @param sigma Positive scalar residual standard deviation.
#   * @param rho Positive scalar autoregressive parameter.
#   * @param W Spatial weight matrix.
#   *
#   * @return A scalar to be added to the log posterior.
#   */
#  real normal_lagsar_lpdf(vector y, vector mu, real sigma,
#                          real rho, matrix W) {
#    int N = rows(y);
#    real inv_sigma2 = 1 / square(sigma);
#    matrix[N, N] W_tilde = -rho * W;
#    vector[N] half_pred;
#  
#    for (n in 1:N) W_tilde[n,n] += 1;
#  
#    half_pred = W_tilde * (y - mdivide_left(W_tilde, mu));
#  
#    return 0.5 * log_determinant(crossprod(W_tilde) * inv_sigma2) -
#           0.5 * dot_self(half_pred) * inv_sigma2;
#  }

## ----setup, cache=FALSE--------------------------------------------------
library("loo")
library("brms")
library("bayesplot")
library("ggplot2")
color_scheme_set("brightblue")
theme_set(theme_default())


SEED <- 10001 
set.seed(SEED) # only sets seed for R (seed for Stan set later)

# loads COL.OLD data frame and COL.nb neighbor list
data(oldcol, package = "spdep") 

## ----data----------------------------------------------------------------
str(COL.OLD[, c("CRIME", "HOVAL", "INC")])

## ----fit, results="hide"-------------------------------------------------
fit <- brm(
  CRIME ~ INC + HOVAL, 
  data = COL.OLD,
  autocor = cor_lagsar(COL.nb),
  chains = 4,
  seed = SEED
)

## ----plot-lagsar, message=FALSE------------------------------------------
lagsar <- as.matrix(fit, pars = "lagsar")
estimates <- quantile(lagsar, probs = c(0.25, 0.5, 0.75))
mcmc_hist(lagsar) + 
  vline_at(estimates, linetype = 2, size = 1) +
  ggtitle("lagsar: posterior median and 50% central interval")

## ----approx--------------------------------------------------------------
posterior <- as.data.frame(fit)
y <- fit$data$CRIME
N <- length(y)
S <- nrow(posterior)
loglik <- yloo <- sdloo <- matrix(nrow = S, ncol = N)

for (s in 1:S) {
  p <- posterior[s, ]
  eta <- p$b_Intercept + p$b_INC * fit$data$INC + p$b_HOVAL * fit$data$HOVAL
  W_tilde <- diag(N) - p$lagsar * fit$autocor$W
  Cinv <- t(W_tilde) %*% W_tilde / p$sigma^2
  g <- Cinv %*% (y - solve(W_tilde, eta))
  cbar <- diag(Cinv)
  yloo[s, ] <- y - g / cbar
  sdloo[s, ] <- sqrt(1 / cbar)
  loglik[s, ] <- dnorm(y, yloo[s, ], sdloo[s, ], log = TRUE)
}

# use loo for psis smoothing
log_ratios <- -loglik
psis_result <- psis(log_ratios)

## ----plot, cache = FALSE-------------------------------------------------
plot(psis_result, label_points = TRUE)

## ----checklast, cache = FALSE--------------------------------------------
yloo_sub <- yloo[S, ]
sdloo_sub <- sdloo[S, ]
df <- data.frame(
  y = y, 
  yloo = yloo_sub,
  ymin = yloo_sub - sdloo_sub * 2,
  ymax = yloo_sub + sdloo_sub * 2
)
ggplot(data=df, aes(x = y, y = yloo, ymin = ymin, ymax = ymax)) +
  geom_errorbar(
    width = 1, 
    color = "skyblue3", 
    position = position_jitter(width = 0.25)
  ) +
  geom_abline(color = "gray30", size = 1.2) +
  geom_point()

## ----psisloo-------------------------------------------------------------
(psis_loo <- loo(loglik))

## ----fit_dummy, cache = TRUE---------------------------------------------
# see help("mi", "brms") for details on the mi() usage
fit_dummy <- brm(
  CRIME | mi() ~ INC + HOVAL, 
  data = COL.OLD,
  autocor = cor_lagsar(COL.nb), 
  chains = 0
)

## ----exact-loo-cv, results="hide", message=FALSE, warning=FALSE, cache = TRUE----
S <- 500
res <- vector("list", N)
loglik <- matrix(nrow = S, ncol = N)
for (i in seq_len(N)) {
  dat_mi <- COL.OLD
  dat_mi$CRIME[i] <- NA
  fit_i <- update(fit_dummy, newdata = dat_mi, 
                  # just for vignette
                  chains = 1, iter = S * 2)
  posterior <- as.data.frame(fit_i)
  yloo <- sdloo <- rep(NA, S)
  for (s in seq_len(S)) {
    p <- posterior[s, ]
    y_miss_i <- y
    y_miss_i[i] <- p$Ymi
    eta <- p$b_Intercept + p$b_INC * fit_i$data$INC + p$b_HOVAL * fit_i$data$HOVAL
    W_tilde <- diag(N) - p$lagsar * fit_i$autocor$W
    Cinv <- t(W_tilde) %*% W_tilde / p$sigma^2
    g <- Cinv %*% (y_miss_i - solve(W_tilde, eta))
    cbar <- diag(Cinv);
    yloo[s] <- y_miss_i[i] - g[i] / cbar[i]
    sdloo[s] <- sqrt(1 / cbar[i])
    loglik[s, i] <- dnorm(y[i], yloo[s], sdloo[s], log = TRUE)
  }
  ypred <- rnorm(S, yloo, sdloo)
  res[[i]] <- data.frame(y = c(posterior$Ymi, ypred))
  res[[i]]$type <- rep(c("pp", "loo"), each = S)
  res[[i]]$obs <- i
}
res <- do.call(rbind, res)

## ----yplots, cache = FALSE, fig.width=10, out.width="95%", fig.asp = 0.3----
res_sub <- res[res$obs %in% 1:4, ]
ggplot(res_sub, aes(y, fill = type)) +
  geom_density(alpha = 0.6) +
  facet_wrap("obs", scales = "fixed", ncol = 4)

## ----loo_exact, cache=FALSE----------------------------------------------
log_mean_exp <- function(x) {
  # more stable than log(mean(exp(x)))
  max_x <- max(x)
  max_x + log(sum(exp(x - max_x))) - log(length(x))
}
exact_elpds <- apply(loglik, 2, log_mean_exp)
exact_elpd <- sum(exact_elpds)
round(exact_elpd, 1)

## ----compare, fig.height=5-----------------------------------------------
df <- data.frame(
  approx_elpd = psis_loo$pointwise[, "elpd_loo"],
  exact_elpd = exact_elpds
)
ggplot(df, aes(x = approx_elpd, y = exact_elpd)) +
  geom_abline(color = "gray30") +
  geom_point(size = 2) +
  geom_point(data = df[4, ], size = 3, color = "red3") +
  xlab("Approximate elpds") +
  ylab("Exact elpds") +
  coord_fixed(xlim = c(-16, -3), ylim = c(-16, -3))

## ----pt4-----------------------------------------------------------------
without_pt_4 <- c(
  approx = sum(psis_loo$pointwise[-4, "elpd_loo"]),
  exact = sum(exact_elpds[-4])  
)
round(without_pt_4, 1)

## ----brms-stan-code, eval=FALSE------------------------------------------
#  // generated with brms 2.2.0
#  functions {
#  /**
#   * Normal log-pdf for spatially lagged responses
#   *
#   * @param y Vector of response values.
#   * @param mu Mean parameter vector.
#   * @param sigma Positive scalar residual standard deviation.
#   * @param rho Positive scalar autoregressive parameter.
#   * @param W Spatial weight matrix.
#   *
#   * @return A scalar to be added to the log posterior.
#   */
#    real normal_lagsar_lpdf(vector y, vector mu, real sigma,
#                            real rho, matrix W) {
#      int N = rows(y);
#      real inv_sigma2 = 1 / square(sigma);
#      matrix[N, N] W_tilde = -rho * W;
#      vector[N] half_pred;
#      for (n in 1:N) W_tilde[n, n] += 1;
#      half_pred = W_tilde * (y - mdivide_left(W_tilde, mu));
#      return 0.5 * log_determinant(crossprod(W_tilde) * inv_sigma2) -
#             0.5 * dot_self(half_pred) * inv_sigma2;
#    }
#  }
#  data {
#    int<lower=1> N;  // total number of observations
#    vector[N] Y;  // response variable
#    int<lower=0> Nmi;  // number of missings
#    int<lower=1> Jmi[Nmi];  // positions of missings
#    int<lower=1> K;  // number of population-level effects
#    matrix[N, K] X;  // population-level design matrix
#    matrix[N, N] W;  // spatial weight matrix
#    int prior_only;  // should the likelihood be ignored?
#  }
#  transformed data {
#    int Kc = K - 1;
#    matrix[N, K - 1] Xc;  // centered version of X
#    vector[K - 1] means_X;  // column means of X before centering
#    for (i in 2:K) {
#      means_X[i - 1] = mean(X[, i]);
#      Xc[, i - 1] = X[, i] - means_X[i - 1];
#    }
#  }
#  parameters {
#    vector[Nmi] Ymi;  // estimated missings
#    vector[Kc] b;  // population-level effects
#    real temp_Intercept;  // temporary intercept
#    real<lower=0> sigma;  // residual SD
#    real<lower=0,upper=1> lagsar;  // SAR parameter
#  }
#  transformed parameters {
#  }
#  model {
#    vector[N] Yl = Y;
#    vector[N] mu = Xc * b + temp_Intercept;
#    Yl[Jmi] = Ymi;
#    // priors including all constants
#    target += student_t_lpdf(temp_Intercept | 3, 34, 17);
#    target += student_t_lpdf(sigma | 3, 0, 17)
#      - 1 * student_t_lccdf(0 | 3, 0, 17);
#    // likelihood including all constants
#    if (!prior_only) {
#      target += normal_lagsar_lpdf(Yl | mu, sigma, lagsar, W);
#    }
#  }
#  generated quantities {
#    // actual population-level intercept
#    real b_Intercept = temp_Intercept - dot_product(means_X, b);
#  }

#' Efficient approximate LOO-CV using subsampling
#'
#' @description Efficiently approximates leave-one-out cross-validation (LOO-CV) via subsampling.
#' Less costly and more approximate computation is made for all LOO-fold,
#' and more costly and accurate computations are made only for m<N LOO-folds.
#'
#' @param x A function. The **Methods (by class)** section, below, has detailed
#'   descriptions of how to specify the inputs.
#'
#' @inheritParams loo
#' @param save_psis Should the `"psis"` object created internally by
#'   `loo_subsample()` be saved in the returned object? See [loo()] for details.
#' @template cores
#'
#' @details The `loo_subsample()` function is an S3 generic and a methods is
#'   currently provided for log-likelihood functions. The implementation works
#'   for both MCMC and for posterior approximations where it is possible to
#'   compute the log density for the approximation.
#'
#' @return `loo_subsample()` returns a named list with class `c("psis_loo_ss",
#'   "psis_loo", "loo")`. This has the same structure as objects returned by
#'   [loo()] but with the additional slot:
#'   * `loo_subsampling`: A list with two vectors, `log_p` and `log_g`, of the
#'   same length containing the posterior density and the approximation density
#'   for the individual draws.
#'
#' @seealso [loo()], [psis()], [loo_compare()]
#' @template loo-large-data-references
#'
#' @export loo_subsample loo_subsample.function
#'
loo_subsample <- function(x, ...) {
  UseMethod("loo_subsample")
}

#' @export
#' @templateVar fn loo_subsample
#' @template function
#' @param data,draws,... For `loo_subsample.function()`, these are the data,
#'   posterior draws, and other arguments to pass to the log-likelihood
#'   function. Note that for some `loo_approximation`s, the draws will be replaced
#'   by the posteriors summary statistics to compute loo approximations. See
#'   argument `loo_approximation` for details.
#' @param observations The subsample observations to use. The argument can take
#'   four (4) types of arguments:
#'   * `NULL` to use all observations. The algorithm then just uses
#'     standard `loo()` or `loo_approximate_posterior()`.
#'   * A single integer to specify the number of observations to be subsampled.
#'   * A vector of integers to provide the indices used to subset the data.
#'     _These observations need to be subsampled with the same scheme as given by
#'     the `estimator` argument_.
#'   * A `psis_loo_ss` object to use the same observations that were used in a
#'     previous call to `loo_subsample()`.
#'
#' @param log_p,log_g Should be supplied only if approximate posterior draws are
#'   used. The default (`NULL`) indicates draws are from "true" posterior (i.e.
#'   using MCMC). If not `NULL` then they should be specified as described in
#'   [loo_approximate_posterior()].
#'
#' @param loo_approximation What type of approximation of the loo_i's should be used?
#'   The default is `"plpd"` (the log predictive density using the posterior expectation).
#'   There are six different methods implemented to approximate loo_i's
#'   (see the references for more details):
#'   * `"plpd"`: uses the lpd based on point estimates (i.e., \eqn{p(y_i|\hat{\theta})}).
#'   * `"lpd"`: uses the lpds (i,e., \eqn{p(y_i|y)}).
#'   * `"tis"`: uses truncated importance sampling to approximate PSIS-LOO.
#'   * `"waic"`: uses waic (i.e., \eqn{p(y_i|y) - p_{waic}}).
#'   * `"waic_grad_marginal"`: uses waic approximation using first order delta
#'     method and posterior marginal variances to approximate \eqn{p_{waic}} (ie.
#'     \eqn{p(y_i|\hat{\theta})}-p_waic_grad_marginal). Requires gradient of
#'     likelihood function.
#'   * `"waic_grad"`: uses waic approximation using first order delta method and
#'     posterior covariance to approximate \eqn{p_{waic}} (ie.
#'     \eqn{p(y_i|\hat{\theta})}-p_waic_grad). Requires gradient of likelihood
#'     function.
#'   * `"waic_hess"`: uses waic approximation using second order delta method and
#'     posterior covariance to approximate \eqn{p_{waic}} (ie.
#'     \eqn{p(y_i|\hat{\theta})}-p_waic_grad). Requires gradient and Hessian of
#'     likelihood function.
#'
#'  As point estimates of \eqn{\hat{\theta}}, the posterior expectations
#'  of the parameters are used.
#'
#' @param loo_approximation_draws The number of posterior draws used when
#'   integrating over the posterior. This is used if `loo_approximation` is set
#'   to `"lpd"`, `"waic"`, or `"tis"`.
#'
#' @param estimator How should `elpd_loo`, `p_loo` and `looic` be estimated?
#'  The default is `"diff_srs"`.
#'  * `"diff_srs"`: uses the difference estimator with simple random sampling
#'    without replacement (srs). `p_loo` is estimated using standard srs.
#'    (Magnusson et al., 2020)
#'  * `"hh"`: uses the Hansen-Hurwitz estimator with sampling with replacement
#'    proportional to size, where `abs` of loo_approximation is used as size.
#'    (Magnusson et al., 2019)
#'  * `"srs"`: uses simple random sampling and ordinary estimation.
#'
#' @param llgrad The gradient of the log-likelihood. This
#'   is only used when `loo_approximation` is `"waic_grad"`,
#'   `"waic_grad_marginal"`, or `"waic_hess"`. The default is `NULL`.
#' @param llhess The Hessian of the log-likelihood. This is only used
#'        with `loo_approximation = "waic_hess"`. The default is `NULL`.
#'
loo_subsample.function <-
  function(x,
           ...,
           data = NULL,
           draws = NULL,
           observations = 400,
           log_p = NULL,
           log_g = NULL,
           r_eff = 1,
           save_psis = FALSE,
           cores = getOption("mc.cores", 1),
           loo_approximation = "plpd",
           loo_approximation_draws = NULL,
           estimator = "diff_srs",
           llgrad = NULL,
           llhess = NULL) {
    cores <- loo_cores(cores)
    # Asserting inputs
    .llfun <- validate_llfun(x)
    stopifnot(is.data.frame(data) || is.matrix(data), !is.null(draws))
    observations <- assert_observations(observations,
                                        N = dim(data)[1],
                                        estimator)
    checkmate::assert_numeric(log_p, len = length(log_g), null.ok = TRUE)
    checkmate::assert_null(dim(log_p))
    checkmate::assert_numeric(log_g, len = length(log_p), null.ok = TRUE)
    checkmate::assert_null(dim(log_g))

    if (is.null(log_p) && is.null(log_g)) {
        r_eff <- prepare_psis_r_eff(r_eff, len = dim(data)[1])
    }
    checkmate::assert_flag(save_psis)
    cores <- loo_cores(cores)

    checkmate::assert_choice(loo_approximation, choices = loo_approximation_choices(), null.ok = FALSE)
    checkmate::assert_int(loo_approximation_draws, lower = 1, upper = .ndraws(draws), null.ok = TRUE)
    checkmate::assert_choice(estimator, choices = estimator_choices())

    .llgrad <- .llhess <- NULL
    if (!is.null(llgrad)) .llgrad <- validate_llfun(llgrad)
    if (!is.null(llhess)) .llhess <- validate_llfun(llhess)

    # Fallbacks
    if (is.null(observations)) {
      if (is.null(log_p) && is.null(log_g)) {
        loo_obj <- loo.function(
          .llfun,
          ...,
          data = data,
          draws = draws,
          r_eff = r_eff,
          save_psis = save_psis,
          cores = cores
        )
      } else {
        loo_obj <- loo_approximate_posterior.function(
          .llfun,
          ...,
          log_p = log_p,
          log_g = log_g,
          data = data,
          draws = draws,
          save_psis = save_psis,
          cores = cores
        )
      }
      return(loo_obj)
    }

    # Compute loo approximation
    elpd_loo_approx <- elpd_loo_approximation(
      .llfun = .llfun,
      data = data,
      draws = draws,
      cores = cores,
      loo_approximation = loo_approximation,
      loo_approximation_draws = loo_approximation_draws,
      .llgrad = .llgrad,
      .llhess = .llhess
    )

    # Draw subsample of observations
    if (length(observations) == 1) {
      # Compute idxs
      idxs <- subsample_idxs(
        estimator = estimator,
        elpd_loo_approximation = elpd_loo_approx,
        observations = observations
      )
    } else {
      # Compute idxs
      idxs <- compute_idxs(observations)
    }
    data_subsample <- data[idxs$idx,, drop = FALSE]
    if (length(r_eff) > 1) {
      r_eff <- r_eff[idxs$idx]
    }

    # Compute elpd_loo
    if (!is.null(log_p) && !is.null(log_g)) {
      loo_obj <- loo_approximate_posterior.function(
        x = .llfun,
        data = data_subsample,
        draws = draws,
        log_p = log_p,
        log_g = log_g,
        save_psis = save_psis,
        cores = cores
      )
    } else {
      loo_obj <- loo.function(
        x = .llfun,
        data = data_subsample,
        draws = draws,
        r_eff = r_eff,
        save_psis = save_psis,
        cores = cores
      )
    }

    # Construct ss object and estimate
    loo_ss <- psis_loo_ss_object(x = loo_obj,
                                 idxs = idxs,
                                 elpd_loo_approx = elpd_loo_approx,
                                 loo_approximation = loo_approximation,
                                 loo_approximation_draws = loo_approximation_draws,
                                 estimator = estimator,
                                 .llfun = .llfun,
                                 .llgrad = .llgrad,
                                 .llhess = .llhess,
                                 data_dim = dim(data),
                                 ndraws = .ndraws(draws))
    loo_ss
  }


#' Update `psis_loo_ss` objects
#'
#' @details
#' If `observations` is updated then if a vector of indices or a `psis_loo_ss`
#' object is supplied the updated object will have exactly the observations
#' indicated by the vector or `psis_loo_ss` object. If a single integer is
#' supplied, new observations will be sampled to reach the supplied sample size.
#'
#' @export
#' @inheritParams loo_subsample.function
#' @param data,draws See [loo_subsample.function()].
#' @param object A `psis_loo_ss` object to update.
#' @param ... Currently not used.
#' @return A `psis_loo_ss` object.
#' @importFrom stats update
update.psis_loo_ss <- function(object, ...,
                               data = NULL,
                               draws = NULL,
                               observations = NULL,
                               r_eff = 1,
                               cores = getOption("mc.cores", 1),
                               loo_approximation = NULL,
                               loo_approximation_draws = NULL,
                               llgrad = NULL,
                               llhess = NULL) {
  # Fallback
  if (is.null(observations) &
     is.null(loo_approximation) &
     is.null(loo_approximation_draws) &
     is.null(llgrad) &
     is.null(llhess)) return(object)

  if (!is.null(data)) {
    stopifnot(is.data.frame(data) || is.matrix(data))
    checkmate::assert_true(all(dim(data) == object$loo_subsampling$data_dim))
  }
  if (!is.null(draws)) {
    # No current checks
  }
  cores <- loo_cores(cores)

  # Update elpd approximations
  if (!is.null(loo_approximation) | !is.null(loo_approximation_draws)) {
    stopifnot(is.data.frame(data) || is.matrix(data) & !is.null(draws))
    if (object$loo_subsampling$estimator %in% "hh_pps") {
      # HH estimation uses elpd_loo approx to sample,
      # so updating it will lead to incorrect results
      stop("Can not update loo_approximation when using PPS sampling.", call. = FALSE)
    }
    if (is.null(loo_approximation)) loo_approximation <- object$loo_subsampling$loo_approximation
    if (is.null(loo_approximation_draws)) loo_approximation_draws <- object$loo_subsampling$loo_approximation_draws
    if (is.null(llgrad)) .llgrad <- object$loo_subsampling$.llgrad else .llgrad <- validate_llfun(llgrad)
    if (is.null(llhess)) .llhess <- object$loo_subsampling$.llhess else .llhess <- validate_llfun(llhess)

    # Compute loo approximation
    elpd_loo_approx <-
      elpd_loo_approximation(.llfun = object$loo_subsampling$.llfun,
                             data = data, draws = draws,
                             cores = cores,
                             loo_approximation = loo_approximation,
                             loo_approximation_draws = loo_approximation_draws,
                             .llgrad = .llgrad, .llhess = .llhess)
    # Update object
    object$loo_subsampling$elpd_loo_approx <- elpd_loo_approx
    object$loo_subsampling$loo_approximation <- loo_approximation
    object$loo_subsampling["loo_approximation_draws"] <- list(loo_approximation_draws)
    object$loo_subsampling$.llgrad <- .llgrad
    object$loo_subsampling$.llhess <- .llhess
    object$pointwise[, "elpd_loo_approx"] <- object$loo_subsampling$elpd_loo_approx[object$pointwise[, "idx"]]
  }

  # Update observations
  if (!is.null(observations)) {
    observations <- assert_observations(observations,
                                        N = object$loo_subsampling$data_dim[1],
                                        object$loo_subsampling$estimator)
    if (length(observations) == 1) {
      checkmate::assert_int(observations, lower = nobs(object) + 1)
      stopifnot(is.data.frame(data) || is.matrix(data) & !is.null(draws))
    }

    # Compute subsample indices
    if (length(observations) > 1) {
      idxs <- compute_idxs(observations)
    } else {
      current_obs <- nobs(object)

      # If sampling with replacement
      if (object$loo_subsampling$estimator %in% c("hh_pps")) {
        idxs <- subsample_idxs(estimator = object$loo_subsampling$estimator,
                               elpd_loo_approximation = object$loo_subsampling$elpd_loo_approx,
                               observations = observations - current_obs)
      }
      # If sampling without replacement
      if (object$loo_subsampling$estimator %in% c("diff_srs", "srs")) {
        current_idxs <- obs_idx(object, rep = FALSE)
        new_idx <- (1:length(object$loo_subsampling$elpd_loo_approx))[-current_idxs]
        idxs <- subsample_idxs(estimator = object$loo_subsampling$estimator,
                               elpd_loo_approximation = object$loo_subsampling$elpd_loo_approx[-current_idxs],
                               observations = observations - current_obs)
        idxs$idx <- new_idx[idxs$idx]
      }
    }

    # Identify how to update object
    cidxs <- compare_idxs(idxs, object)

    # Compute new observations
    if (!is.null(cidxs$new)) {
      stopifnot(is.data.frame(data) || is.matrix(data) & !is.null(draws))
      data_new_subsample <- data[cidxs$new$idx,, drop = FALSE]
      if (length(r_eff) > 1) r_eff <- r_eff[cidxs$new$idx]

      if (!is.null(object$approximate_posterior$log_p) & !is.null(object$approximate_posterior$log_g)) {
        loo_obj <- loo_approximate_posterior.function(x = object$loo_subsampling$.llfun,
                                                  data = data_new_subsample,
                                                  draws = draws,
                                                  log_p = object$approximate_posterior$log_p,
                                                  log_g = object$approximate_posterior$log_g,
                                                  save_psis = !is.null(object$psis_object),
                                                  cores = cores)
      } else {
        loo_obj <- loo.function(x = object$loo_subsampling$.llfun,
                            data = data_new_subsample,
                            draws = draws,
                            r_eff = r_eff,
                            save_psis = !is.null(object$psis_object),
                            cores = cores)
      }
      # Add stuff to pointwise
      loo_obj$pointwise <-
        add_subsampling_vars_to_pointwise(loo_obj$pointwise,
                                          cidxs$new,
                                          object$loo_subsampling$elpd_loo_approx)
    } else {
      loo_obj <- NULL
    }

    if (length(observations) == 1) {
      # Add new samples pointwise and diagnostic
      object <- rbind_psis_loo_ss(object, x = loo_obj)

      # Update m_i for current pointwise (diagnostic stay the same)
      object$pointwise <- update_m_i_in_pointwise(object$pointwise, cidxs$add, type = "add")
    } else {
      # Add new samples pointwise and diagnostic
      object <- rbind_psis_loo_ss(object, loo_obj)

      # Replace m_i current pointwise and diagnostics
      object$pointwise <- update_m_i_in_pointwise(object$pointwise, cidxs$add, type = "replace")

      # Remove samples
      object <- remove_idx.psis_loo_ss(object, idxs = cidxs$remove)
      stopifnot(setequal(obs_idx(object), observations))

      # Order object as in observations
      object <- order.psis_loo_ss(object, observations)
    }
  }


  # Compute estimates
  if (object$loo_subsampling$estimator == "hh_pps") {
    object <- loo_subsample_estimation_hh(object)
  } else if (object$loo_subsampling$estimator == "diff_srs") {
    object <- loo_subsample_estimation_diff_srs(object)
  } else if (object$loo_subsampling$estimator == "srs") {
    object <- loo_subsample_estimation_srs(object)
  } else {
    stop("No correct estimator used.")
  }

  assert_psis_loo_ss(object)
  object
}

#' Get observation indices used in subsampling
#'
#' @param x A `psis_loo_ss` object.
#' @param rep If sampling with replacement is used, an observation can have
#'   multiple samples and these are then repeated in the returned object if
#'   `rep=TRUE` (e.g., a vector `c(1,1,2)` indicates that observation 1 has been
#'   subsampled two times). If `rep=FALSE` only the unique indices are returned.
#'
#' @return An integer vector.
#'
#' @export
obs_idx <- function(x, rep = TRUE) {
  checkmate::assert_class(x, "psis_loo_ss")
  if (rep) {
    idxs <- as.integer(rep(x$pointwise[,"idx"], x$pointwise[,"m_i"]))
  } else {
    idxs <- as.integer(x$pointwise[,"idx"])
  }
  idxs
}

#' The number of observations in a `psis_loo_ss` object.
#' @importFrom stats nobs
#' @param object a `psis_loo_ss` object.
#' @param ... Currently unused.
#' @export
nobs.psis_loo_ss <- function(object, ...) {
  as.integer(sum(object$pointwise[,"m_i"]))
}

# internal ----------------------------------------------------------------

#' The possible choices of loo_approximations implemented
#'
#' @details
#' The choice `psis` is returned if a `psis_loo` object
#' is converted to a `psis_loo_ss` object with `as.psis_loo_ss()`.
#' But `psis` cannot be chosen in the API of `loo_subsample()`.
#'
#' @noRd
#' @param api The choices available in the loo API or all possible choices.
#' @return A character vector of allowed choices.
loo_approximation_choices <- function(api = TRUE) {
  lac <- c("plpd", "lpd", "waic", "waic_grad_marginal", "waic_grad", "waic_hess", "tis", "sis", "none")
  if (!api) lac <- c(lac, "psis")
  lac
}

#' The estimators implemented
#'
#' @noRd
#' @return A character vector of allowed choices.
estimator_choices <- function() {
  c("hh_pps", "diff_srs", "srs")
}

## Approximate elpd -----

#' Utility function to apply user-specified log-likelihood to a single data point
#' @details
#' See `elpd_loo_approximation` and `compute_lpds` for usage examples
#' @noRd
#'
#' @return lpd value for a single data point i
lpd_i <- function(i, llfun, data, draws) {
  ll_i <- llfun(data_i = data[i,, drop=FALSE], draws = draws)
  ll_i <- as.vector(ll_i)
  lpd_i <- logMeanExp(ll_i)
  lpd_i
}


#' Utility function to compute lpd using user-defined likelihood function
#' using platform-dependent parallel backends when cores > 1
#'
#' @details
#' See `elpd_loo_approximation` for usage examples
#'
#' @noRd
#' @return a vector of computed log probability densities
compute_lpds <- function(N, data, draws, llfun, cores) {
  if (cores == 1) {
    lpds <- lapply(X = seq_len(N), FUN = lpd_i, llfun, data, draws)
  } else {
    if (.Platform$OS.type != "windows") {
      lpds <- mclapply(X = seq_len(N), mc.cores = cores, FUN = lpd_i, llfun, data, draws)
    } else {
      cl <- makePSOCKcluster(cores)
      on.exit(stopCluster(cl))
      lpds <- parLapply(cl, X = seq_len(N), fun = lpd_i, llfun, data, draws)
    }
  }

  unlist(lpds)
}

#' Compute approximation to loo_i:s
#'
#' @details
#' See [loo_subsample.function()] and the `loo_approximation` argument.
#' @noRd
#' @inheritParams loo_subsample.function
#'
#' @return a vector with approximations of elpd_{loo,i}s
elpd_loo_approximation <- function(.llfun, data, draws, cores, loo_approximation, loo_approximation_draws = NULL, .llgrad = NULL, .llhess = NULL) {
  checkmate::assert_function(.llfun, args = c("data_i", "draws"), ordered = TRUE)
  stopifnot(is.data.frame(data) || is.matrix(data), !is.null(draws))
  checkmate::assert_choice(loo_approximation, choices = loo_approximation_choices(), null.ok = FALSE)
  checkmate::assert_int(loo_approximation_draws, lower = 2, null.ok = TRUE)
  if (!is.null(.llgrad)) {
    checkmate::assert_function(.llgrad, args = c("data_i", "draws"), ordered = TRUE)
  }
  if (!is.null(.llhess)) {
    checkmate::assert_function(.llhess, args = c("data_i", "draws"), ordered = TRUE)
  }

  cores <- loo_cores(cores)
  N <- dim(data)[1]

  if (loo_approximation == "none") return(rep(1L,N))

  if (loo_approximation %in% c("tis", "sis")) {
    draws <- .thin_draws(draws, loo_approximation_draws)
    is_values <- suppressWarnings(loo.function(.llfun, data = data, draws = draws, is_method = loo_approximation))
    return(is_values$pointwise[, "elpd_loo"])
  }

  if (loo_approximation == "waic") {
    draws <- .thin_draws(draws, loo_approximation_draws)
    waic_full_obj <- waic.function(.llfun, data = data, draws = draws)
    return(waic_full_obj$pointwise[,"elpd_waic"])
  }

  # Compute the lpd or log p(y_i|y_{-i})
  if (loo_approximation == "lpd") {
    draws <- .thin_draws(draws, loo_approximation_draws)
    lpds <- compute_lpds(N, data, draws, .llfun, cores)
    return(lpds) # Use only the lpd
  }

  # Compute the point lpd or log p(y_i|\hat{\theta}) - also used in waic_delta approaches
  if (loo_approximation == "plpd" |
      loo_approximation == "waic_grad" |
      loo_approximation == "waic_grad_marginal" |
      loo_approximation == "waic_hess") {

    draws <- .thin_draws(draws, loo_approximation_draws)
    point_est <- .compute_point_estimate(draws)
    lpds <- compute_lpds(N, data, point_est, .llfun, cores)
    if (loo_approximation == "plpd") return(lpds) # Use only the lpd
  }

  if (loo_approximation == "waic_grad" |
      loo_approximation == "waic_grad_marginal" |
      loo_approximation == "waic_hess") {
    checkmate::assert_true(!is.null(.llgrad))

    point_est <- .compute_point_estimate(draws)
    # Compute the lpds
    lpds <- compute_lpds(N, data, point_est, .llfun, cores)

    if (loo_approximation == "waic_grad" |
        loo_approximation == "waic_hess") {
      cov_est <- stats::cov(draws)
    }

    if (loo_approximation == "waic_grad_marginal") {
      marg_vars <- apply(draws, MARGIN = 2, var)
    }

    p_eff_approx <- numeric(N)
    if (cores>1) warning("Multicore is not implemented for waic_delta",
                         call. = FALSE)

    if (loo_approximation == "waic_grad") {
      for(i in 1:nrow(data)) {
        grad_i <- t(.llgrad(data[i,,drop = FALSE], point_est))
        local_cov <- cov_est[rownames(grad_i), rownames(grad_i)]
        p_eff_approx[i] <-  t(grad_i) %*% local_cov %*% grad_i
      }
    } else if (loo_approximation == "waic_grad_marginal") {
      for(i in 1:nrow(data)) {
        grad_i <- t(.llgrad(data[i,,drop = FALSE], point_est))
        p_eff_approx[i] <- sum(grad_i * marg_vars[rownames(grad_i)] * grad_i)
      }
    } else if (loo_approximation == "waic_hess") {
      checkmate::assert_true(!is.null(.llhess))
      for(i in 1:nrow(data)) {
        grad_i <- t(.llgrad(data[i,,drop = FALSE], point_est))
        hess_i <- .llhess(data_i = data[i,,drop = FALSE], draws = point_est[,rownames(grad_i), drop = FALSE])[,,1]
        local_cov <- cov_est[rownames(grad_i), rownames(grad_i)]
        p_eff_approx[i] <- t(grad_i) %*% local_cov %*% grad_i +
          0.5 * sum(diag(local_cov %*% hess_i %*% local_cov %*% hess_i))
      }
    } else {
      stop(loo_approximation, " is not implemented!", call. = FALSE)
    }
    return(lpds - p_eff_approx)
  }

}


#' Compute a point estimate from a draws object
#'
#' @keywords internal
#' @export
#' @details This is a generic function to compute point estimates from draws
#'   objects. The function is internal and should only be used by developers to
#'   enable [loo_subsample()] for arbitrary draws objects.
#'
#' @param draws A draws object with draws from the posterior.
#' @return A 1 by P matrix with point estimates from a draws object.
.compute_point_estimate <- function(draws) {
  UseMethod(".compute_point_estimate")
}
#' @rdname dot-compute_point_estimate
#' @export
.compute_point_estimate.matrix <- function(draws) {
  t(as.matrix(colMeans(draws)))
}
#' @rdname dot-compute_point_estimate
#' @export
.compute_point_estimate.default <- function(draws) {
  stop(".compute_point_estimate() has not been implemented for objects of class '", class(draws), "'")
}

#' Thin a draws object
#'
#' @keywords internal
#' @export
#' @details This is a generic function to thin draws from arbitrary draws
#'   objects. The function is internal and should only be used by developers to
#'   enable [loo_subsample()] for arbitrary draws objects.
#'
#' @param draws A draws object with posterior draws.
#' @param loo_approximation_draws The number of posterior draws to return (ie after thinning).
#' @return A thinned draws object.
.thin_draws <- function(draws, loo_approximation_draws) {
  UseMethod(".thin_draws")
}
#' @rdname dot-thin_draws
#' @export
.thin_draws.matrix <- function(draws, loo_approximation_draws) {
  if (is.null(loo_approximation_draws)) return(draws)
  checkmate::assert_int(loo_approximation_draws, lower = 1, upper = .ndraws(draws), null.ok = TRUE)
  S <- .ndraws(draws)
  idx <- 1:loo_approximation_draws * S %/% loo_approximation_draws
  draws <- draws[idx, , drop = FALSE]
  draws
}
#' @rdname dot-thin_draws
#' @export
.thin_draws.numeric <- function(draws, loo_approximation_draws) {
  .thin_draws.matrix(as.matrix(draws), loo_approximation_draws)
}
#' @rdname dot-thin_draws
#' @export
.thin_draws.default <- function(draws, loo_approximation_draws) {
  stop(".thin_draws() has not been implemented for objects of class '", class(draws), "'")
}


#' The number of posterior draws in a draws object.
#'
#' @keywords internal
#' @export
#' @details This is a generic function to return the total number of draws from
#'   an arbitrary draws objects. The function is internal and should only be
#'   used by developers to enable [loo_subsample()] for arbitrary draws objects.
#'
#' @param x A draws object with posterior draws.
#' @return An integer with the number of draws.
.ndraws <- function(x) {
  UseMethod(".ndraws")
}
#' @rdname dot-ndraws
#' @export
.ndraws.matrix <- function(x) {
  nrow(x)
}
#' @rdname dot-ndraws
#' @export
.ndraws.default <- function(x) {
  stop(".ndraws() has not been implemented for objects of class '", class(x), "'")
}

## Subsampling -----

#' Subsampling strategy
#'
#' @noRd
#' @param estimator The estimator to use, see `estimator_choices()`.
#' @param elpd_loo_approximation A vector of loo approximations, see `elpd_loo_approximation()`.
#' @param observations The total number of subsample observations to sample.
#' @return A `subsample_idxs` data frame.
subsample_idxs <- function(estimator, elpd_loo_approximation, observations) {
  checkmate::assert_choice(estimator, choices = estimator_choices())
  checkmate::assert_numeric(elpd_loo_approximation)
  checkmate::assert_int(observations)

  if (estimator == "hh_pps") {
    pi_values <- pps_elpd_loo_approximation_to_pis(elpd_loo_approximation)
    idxs_df <- pps_sample(observations, pis = pi_values)
  }

  if (estimator == "diff_srs" | estimator == "srs") {
    if (observations > length(elpd_loo_approximation)) {
      stop("'observations' is larger than the total sample size in 'data'.", call. = FALSE)
    }
    idx <- 1:length(elpd_loo_approximation)
    idx_m <- idx[order(stats::runif(length(elpd_loo_approximation)))][1:observations]
    idx_m <- idx_m[order(idx_m)]
    idxs_df <- data.frame(idx=as.integer(idx_m), m_i=1L)
  }
  assert_subsample_idxs(x = idxs_df)
  idxs_df
}

#' Compute pis from approximation for use in pps sampling.
#' @noRd
#' @details pis are the sampling probabilities and sum to 1.
#' @inheritParams subsample_idxs
#' @return A vector of pis.
pps_elpd_loo_approximation_to_pis <- function(elpd_loo_approximation) {
  checkmate::assert_numeric(elpd_loo_approximation)
  pi_values <- abs(elpd_loo_approximation)
  pi_values <- pi_values/sum(pi_values) # \tilde{\pi}
  pi_values
}


#' Compute subsampling indices from an observation vector
#' @noRd
#' @param observation A vector of indices.
#' @return A `subsample_idxs` data frame.
compute_idxs <- function(observations) {
  checkmate::assert_integer(observations, lower = 1, min.len = 2, any.missing = FALSE)
  tab <- table(observations)
  idxs_df <- data.frame(idx = as.integer(names(tab)), m_i = as.integer(unname(tab)))
  assert_subsample_idxs(idxs_df)
  idxs_df
}


#' Compare the indices to prepare handling
#'
#' @details
#' The function compares the object and sampled indices into `new`
#' (observations not in `object`), `add` (observations in `object`), and
#' `remove` (observations in `object` but not in idxs).
#' @noRd
#' @param idxs A `subsample_idxs` data frame.
#' @param object A `psis_loo_ss` object.
#' @return A list of three `subsample_idxs` data frames. Elements without any
#'   observations return `NULL`.
compare_idxs  <- function(idxs, object) {
  assert_subsample_idxs(idxs)
  current_idx <- compute_idxs(obs_idx(object))
  result <- list()
  new_idx <- !(idxs$idx %in% current_idx$idx)
  remove_idx <- !(current_idx$idx %in% idxs$idx)

  result$new <- idxs[new_idx, ]
  if (nrow(result$new) == 0) {
    result["new"] <- NULL
  } else {
    assert_subsample_idxs(result$new)
  }

  result$add <- idxs[!new_idx, ]
  if (nrow(result$add) == 0) {
    result["add"] <- NULL
  } else {
    assert_subsample_idxs(result$add)
  }

  result$remove <- current_idx[remove_idx, ]
  if (nrow(result$remove) == 0) {
    result["remove"] <- NULL
  } else {
    assert_subsample_idxs(result$remove)
  }

  result
}


#' Draw a PPS sample with replacement and return a idx_df
#' @noRd
#' @details
#' We are sampling with replacement, hence we only want to compute elpd
#' for each observation once.
#' @param m The total sampling size.
#' @param pis The probability of selecting each observation.
#' @return a `subsample_idxs` data frame.
pps_sample <- function(m, pis) {
  checkmate::assert_int(m)
  checkmate::assert_numeric(pis, min.len = 2, lower = 0, upper = 1)
  idx <- sample(1:length(pis), size = m, replace = TRUE, prob = pis)
  idxs_df <- as.data.frame(table(idx), stringsAsFactors = FALSE)
  colnames(idxs_df) <- c("idx", "m_i")
  idxs_df$idx <- as.integer(idxs_df$idx)
  idxs_df$m_i <- as.integer(idxs_df$m_i)
  assert_subsample_idxs(idxs_df)
  idxs_df
}

## Constructor ---

#' Construct a `psis_loo_ss` object
#'
#' @noRd
#' @param x A `psis_loo` object.
#' @param idxs a `subsample_idxs` data frame.
#' @param elpd_loo_approximation A vector of loo approximations, see
#'   `elpd_loo_approximation()`.
#' @inheritParams loo_subsample
#' @param .llfun,.llgrad,.llhess  See llfun, llgrad and llhess in `loo_subsample()`.
#' @param data_dim Dimension of the data object.
#' @param ndraws Dimension of the draws object.
#' @return A `psis_loo_ss` object.
psis_loo_ss_object <- function(x,
                               idxs,
                               elpd_loo_approx,
                               loo_approximation, loo_approximation_draws,
                               estimator,
                               .llfun, .llgrad, .llhess,
                               data_dim, ndraws) {
  # Assertions
  checkmate::assert_class(x, "psis_loo")
  assert_subsample_idxs(idxs)
  checkmate::assert_numeric(elpd_loo_approx, any.missing = FALSE)
  checkmate::assert_choice(loo_approximation, loo_approximation_choices())
  checkmate::assert_int(loo_approximation_draws, null.ok = TRUE)
  checkmate::assert_choice(estimator, estimator_choices())
  checkmate::assert_function(.llfun, args = c("data_i", "draws"), ordered = TRUE)
  checkmate::assert_function(.llgrad, args = c("data_i", "draws"), ordered = TRUE, null.ok = TRUE)
  checkmate::assert_function(.llhess, args = c("data_i", "draws"), ordered = TRUE, null.ok = TRUE)
  checkmate::assert_integer(data_dim, len = 2, lower = 1, any.missing = FALSE)
  checkmate::assert_int(ndraws, lower = 1)

  # Construct object
  class(x) <- c("psis_loo_ss", class(x))
  x$pointwise <- add_subsampling_vars_to_pointwise(pointwise = x$pointwise, idxs, elpd_loo_approx)
  x$estimates <- cbind(x$estimates, matrix(0, nrow = nrow(x$estimates)))
  colnames(x$estimates)[ncol(x$estimates)] <- "subsampling SE"

  x$loo_subsampling <- list()
  x$loo_subsampling$elpd_loo_approx <- elpd_loo_approx
  x$loo_subsampling$loo_approximation <- loo_approximation
  x$loo_subsampling["loo_approximation_draws"] <- list(loo_approximation_draws)
  x$loo_subsampling$estimator <- estimator
  x$loo_subsampling$.llfun <- .llfun
  x$loo_subsampling[".llgrad"] <- list(.llgrad)
  x$loo_subsampling[".llhess"] <- list(.llhess)
  x$loo_subsampling$data_dim <- data_dim
  x$loo_subsampling$ndraws <- ndraws

  # Compute estimates
  if (estimator == "hh_pps") {
    x <- loo_subsample_estimation_hh(x)
  } else if (estimator == "diff_srs") {
    x <- loo_subsample_estimation_diff_srs(x)
  } else if (estimator == "srs") {
    x <- loo_subsample_estimation_srs(x)
  } else {
    stop("No correct estimator used.")
  }
  assert_psis_loo_ss(x)
  x
}

as.psis_loo_ss <- function(x) {
  UseMethod("as.psis_loo_ss")
}
#' @export
as.psis_loo_ss.psis_loo_ss <- function(x) {
  x
}
#' @export
as.psis_loo_ss.psis_loo <- function(x) {
  class(x) <- c("psis_loo_ss", class(x))
  x$estimates <- cbind(x$estimates, matrix(0, nrow = nrow(x$estimates)))
  colnames(x$estimates)[ncol(x$estimates)] <- "subsampling SE"
  x$pointwise <- cbind(x$pointwise,
                       matrix(1:nrow(x$pointwise), byrow = FALSE, ncol = 1),
                       matrix(rep(1,nrow(x$pointwise)), byrow = FALSE, ncol = 1),
                       x$pointwise[, "elpd_loo"])
  ncp <- ncol(x$pointwise)
  colnames(x$pointwise)[(ncp-2):ncp] <- c("idx", "m_i", "elpd_loo_approx")
  x$loo_subsampling <- list(elpd_loo_approx=x$pointwise[, "elpd_loo"],
                           loo_approximation = "psis",
                           loo_approximation_draws = NULL,
                           estimator = "diff_srs",
                           data_dim = c(nrow(x$pointwise), NA),
                           ndraws = NA)
  assert_psis_loo_ss(x)
  x
}

as.psis_loo <- function(x) {
  UseMethod("as.psis_loo")
}

#' @export
as.psis_loo.psis_loo <- function(x) {
  x
}
#' @export
as.psis_loo.psis_loo_ss <- function(x) {
  if (x$loo_subsampling$data_dim[1] == nrow(x$pointwise)) {
    x$estimates <- x$estimates[, 1:2]
    x$pointwise <- x$pointwise[, 1:5]
    x$loo_subsampling <- NULL
    loo_obj <- importance_sampling_loo_object(pointwise = x$pointwise[, 1:5],
                           diagnostics = x$diagnostics,
                           dims = attr(x, "dims"),
                           is_method = "psis",
                           is_object = x$psis_object)
    if (inherits(x, "psis_loo_ap")) {
      loo_obj$approximate_posterior <- list(log_p = x$approximate_posterior$log_p,
                                        log_g = x$approximate_posterior$log_g)
      class(loo_obj) <- c("psis_loo_ap", class(loo_obj))
      assert_psis_loo_ap(loo_obj)
    }
  } else {
    stop("A subsampling loo object can only be coerced to a loo object ",
         "if all observations in data have been subsampled.", call. = FALSE)
  }

  loo_obj
}

#' Add subsampling information to the pointwise element in a `psis_loo` object.
#' @noRd
#' @param pointwise The `pointwise` element in a `psis_loo` object.
#' @param idxs A `subsample_idxs` data frame.
#' @param elpd_loo_approximation A vector of loo approximations, see `elpd_loo_approximation()`.
#' @return A `pointwise` matrix with subsampling information.
add_subsampling_vars_to_pointwise <- function(pointwise, idxs, elpd_loo_approx) {
  checkmate::assert_matrix(pointwise,
                           any.missing = FALSE,
                           min.cols = 5)
  checkmate::assert_names(colnames(pointwise), identical.to = c("elpd_loo","mcse_elpd_loo","p_loo","looic", "influence_pareto_k"))
  assert_subsample_idxs(idxs)
  checkmate::assert_numeric(elpd_loo_approx)

  pw <- cbind(as.data.frame(pointwise), idxs)
  pw$elpd_loo_approx <- elpd_loo_approx[idxs$idx]
  pw <- as.matrix(pw)
  rownames(pw) <- NULL
  assert_subsampling_pointwise(pw)
  pw
}

#' Add `psis_loo` object to a `psis_loo_ss` object
#' @noRd
#' @param object A `psis_loo_ss` object.
#' @param x A `psis_loo` object.
#' @return An updated `psis_loo_ss` object.
rbind_psis_loo_ss <- function(object, x) {
  checkmate::assert_class(object, "psis_loo_ss")
  if (is.null(x)) return(object) # Fallback
  checkmate::assert_class(x, "psis_loo")
  assert_subsampling_pointwise(object$pointwise)
  assert_subsampling_pointwise(x$pointwise)
  checkmate::assert_disjunct(object$pointwise[, "idx"], x$pointwise[, "idx"])

  object$pointwise <- rbind(object$pointwise, x$pointwise)
  object$diagnostics$pareto_k <-
    c(object$diagnostics$pareto_k, x$diagnostics$pareto_k)
  object$diagnostics$n_eff <- c(object$diagnostics$n_eff, x$diagnostics$n_eff)
  object$diagnostics$r_eff <- c(object$diagnostics$r_eff, x$diagnostics$r_eff)
  attr(object, "dims")[2] <- nrow(object$pointwise)
  object
}

#' Remove observations in `idxs` from object
#' @noRd
#' @param object A `psis_loo_ss` object.
#' @param idxs A `subsample_idxs` data frame.
#' @return A `psis_loo_ss` object.
remove_idx.psis_loo_ss <- function(object, idxs) {
  checkmate::assert_class(object, "psis_loo_ss")
  if (is.null(idxs)) return(object) # Fallback
  assert_subsample_idxs(idxs)

  row_map <- data.frame(
    row_no = 1:nrow(object$pointwise),
    idx = object$pointwise[, "idx"]
  )
  row_map <- merge(row_map, idxs, by = "idx", all.y = TRUE)

  object$pointwise <- object$pointwise[-row_map$row_no,,drop = FALSE]
  object$diagnostics$pareto_k <- object$diagnostics$pareto_k[-row_map$row_no]
  object$diagnostics$n_eff <- object$diagnostics$n_eff[-row_map$row_no]
  object$diagnostics$r_eff <- object$diagnostics$r_eff[-row_map$row_no]
  attr(object, "dims")[2] <- nrow(object$pointwise)
  object
}

#' Order object by `observations`.
#' @noRd
#' @param x A `psis_loo_ss` object.
#' @param observations A vector with indices.
#' @return An ordered `psis_loo_ss` object.
order.psis_loo_ss <- function(x, observations) {
  checkmate::assert_class(x, "psis_loo_ss")
  checkmate::assert_integer(observations, len = nobs(x))
  if (identical(obs_idx(x), observations)) return(x) # Fallback
  checkmate::assert_set_equal(obs_idx(x), observations)

  row_map_x <- data.frame(row_no_x = 1:nrow(x$pointwise), idx = x$pointwise[, "idx"])
  row_map_obs <- data.frame(row_no_obs = 1:length(observations), idx = observations)
  row_map <- merge(row_map_obs, row_map_x, by = "idx", sort = FALSE)
  x$pointwise <- x$pointwise[row_map$row_no_x,,drop = FALSE]
  x$diagnostics$pareto_k <- x$diagnostics$pareto_k[row_map$row_no_x]
  x$diagnostics$n_eff <- x$diagnostics$n_eff[row_map$row_no_x]
  x$diagnostics$r_eff <- x$diagnostics$r_eff[row_map$row_no_x]
  x
}

#' Update m_i in a `pointwise` element.
#' @noRd
#' @param x A `psis_loo_ss` `pointwise` data frame.
#' @param idxs A `subsample_idxs` data frame.
#' @param type should the m_i:s in `idxs` `"replace"` the current m_i:s or
#'   `"add"` to them.
#' @return An ordered `psis_loo_ss` object.
update_m_i_in_pointwise <- function(pointwise, idxs, type = "replace") {
  assert_subsampling_pointwise(pointwise)
  if (is.null(idxs)) return(pointwise) # Fallback
  assert_subsample_idxs(idxs)
  checkmate::assert_choice(type, choices = c("replace", "add"))

  row_map <- data.frame(row_no = 1:nrow(pointwise), idx = pointwise[, "idx"])
  row_map <- merge(row_map, idxs, by = "idx", all.y = TRUE)

  if (type == "replace") {
    pointwise[row_map$row_no, "m_i"] <- row_map$m_i
  }
  if (type == "add") {
    pointwise[row_map$row_no, "m_i"] <- pointwise[row_map$row_no, "m_i"] + row_map$m_i
  }
  pointwise
}



## Estimation ---

#' Estimate the elpd using the Hansen-Hurwitz estimator (Magnusson et al., 2019)
#' @noRd
#' @param x A `psis_loo_ss` object.
#' @return A `psis_loo_ss` object.
loo_subsample_estimation_hh <- function(x) {
  checkmate::assert_class(x, "psis_loo_ss")
  N <- length(x$loo_subsampling$elpd_loo_approx)
  pis <- pps_elpd_loo_approximation_to_pis(x$loo_subsampling$elpd_loo_approx)
  pis_sample <- pis[x$pointwise[,"idx"]]

  hh_elpd_loo <- whhest(z = pis_sample, m_i = x$pointwise[, "m_i"], y = x$pointwise[, "elpd_loo"], N)
  srs_elpd_loo <- srs_est(y = x$pointwise[, "elpd_loo"], y_approx = pis_sample)
  x$estimates["elpd_loo", "Estimate"]  <- hh_elpd_loo$y_hat_ppz
  if (hh_elpd_loo$hat_v_y_ppz > 0) {
    x$estimates["elpd_loo", "SE"]  <- sqrt(hh_elpd_loo$hat_v_y_ppz)
  } else {
    warning("Negative estimate of SE, more subsampling obs. needed.", call. = FALSE)
    x$estimates["elpd_loo", "SE"]  <- NaN
  }
  x$estimates["elpd_loo", "subsampling SE"] <- sqrt(hh_elpd_loo$v_hat_y_ppz)

  hh_p_loo <- whhest(z = pis_sample, m_i = x$pointwise[,"m_i"], y = x$pointwise[,"p_loo"], N)
  x$estimates["p_loo", "Estimate"] <- hh_p_loo$y_hat_ppz
  if (hh_p_loo$hat_v_y_ppz > 0) {
    x$estimates["p_loo", "SE"]  <- sqrt(hh_p_loo$hat_v_y_ppz)
  } else {
    warning("Negative estimate of SE, more subsampling obs. needed.", call. = FALSE)
    x$estimates["elpd_loo", "SE"]  <- NaN
  }
  x$estimates["p_loo", "subsampling SE"] <- sqrt(hh_p_loo$v_hat_y_ppz)
  update_psis_loo_ss_estimates(x)
}

#' Update a `psis_loo_ss` object with generic estimates
#'
#' @noRd
#' @details
#' Updates a `psis_loo_ss` with generic estimates (looic)
#' and updates components in the object based on x$estimate.
#' @param x A `psis_loo_ss` object.
#' @return x A `psis_loo_ss` object.
update_psis_loo_ss_estimates <- function(x) {
  checkmate::assert_class(x, "psis_loo_ss")

  x$estimates["looic", "Estimate"] <- (-2) * x$estimates["elpd_loo", "Estimate"]
  x$estimates["looic", "SE"] <- 2 * x$estimates["elpd_loo", "SE"]
  x$estimates["looic", "subsampling SE"] <- 2 * x$estimates["elpd_loo", "subsampling SE"]

  x$elpd_loo <- x$estimates["elpd_loo", "Estimate"]
  x$p_loo <- x$estimates["p_loo", "Estimate"]
  x$looic <- x$estimates["looic", "Estimate"]
  x$se_elpd_loo <- x$estimates["elpd_loo", "SE"]
  x$se_p_loo <- x$estimates["p_loo", "SE"]
  x$se_looic <- x$estimates["looic", "SE"]

  x
}

#' Weighted Hansen-Hurwitz estimator (Magnusson et al., 2019)
#' @noRd
#' @param z Normalized probabilities for the observation.
#' @param m_i The number of times obs i was selected.
#' @param y The values observed.
#' @param N The total number of observations in finite population.
#' @return A list with estimates.
whhest <- function(z, m_i, y, N) {
  checkmate::assert_numeric(z, lower = 0, upper = 1)
  checkmate::assert_numeric(y, len = length(z))
  checkmate::assert_integerish(m_i, len = length(z))
  est_list <- list(m = sum(m_i))
  est_list$y_hat_ppz <- sum(m_i*(y/z))/est_list$m
  est_list$v_hat_y_ppz <- (sum(m_i*((y/z - est_list$y_hat_ppz)^2))/est_list$m)/(est_list$m-1)

  # See unbiadness proof in supplementary material to the article
  est_list$hat_v_y_ppz <-
    (sum(m_i*(y^2/z)) / est_list$m) +
    est_list$v_hat_y_ppz / N - est_list$y_hat_ppz^2 / N
  est_list
}


#' Estimate elpd using the difference estimator and SRS-WOR (Magnusson et al., 2020)
#' @noRd
#' @param x A `psis_loo_ss` object.
#' @return A `psis_loo_ss` object.
loo_subsample_estimation_diff_srs <- function(x) {
  checkmate::assert_class(x, "psis_loo_ss")

  elpd_loo_est <- srs_diff_est(y_approx = x$loo_subsampling$elpd_loo_approx, y = x$pointwise[, "elpd_loo"], y_idx = x$pointwise[, "idx"])
  x$estimates["elpd_loo", "Estimate"] <- elpd_loo_est$y_hat
  x$estimates["elpd_loo", "SE"] <- sqrt(elpd_loo_est$hat_v_y)
  x$estimates["elpd_loo", "subsampling SE"] <- sqrt(elpd_loo_est$v_y_hat)

  p_loo_est <- srs_est(y = x$pointwise[, "p_loo"], y_approx = x$loo_subsampling$elpd_loo_approx)
  x$estimates["p_loo", "Estimate"] <- p_loo_est$y_hat
  x$estimates["p_loo", "SE"] <- sqrt(p_loo_est$hat_v_y)
  x$estimates["p_loo", "subsampling SE"] <- sqrt(p_loo_est$v_y_hat)

  update_psis_loo_ss_estimates(x)
}

#' Difference estimation using SRS-WOR sampling (Magnusson et al., 2020)
#' @noRd
#' @param y_approx Approximated values of all observations.
#' @param y The values observed.
#' @param y_idx The index of `y` in `y_approx`.
#' @return A list with estimates.
srs_diff_est <- function(y_approx, y, y_idx) {
  checkmate::assert_numeric(y_approx)
  checkmate::assert_numeric(y, max.len = length(y_approx))
  checkmate::assert_integerish(y_idx, len = length(y))

  N <- length(y_approx)
  m <- length(y)
  y_approx_m <- y_approx[y_idx]

  e_i <- y - y_approx_m
  t_pi_tilde <- sum(y_approx)
  t_pi2_tilde <- sum(y_approx^2)
  t_e <- N * mean(e_i)
  t_hat_epsilon <- N * mean(y^2 - y_approx_m^2)

  est_list <- list(m = length(y), N = N)
  # eq (7)
  est_list$y_hat <- t_pi_tilde + t_e
  # eq (8)
  est_list$v_y_hat <- N^2 * (1 - m / N) * var(e_i) / m
  # eq (9) first row second `+` should be `-`
  # Supplementary material eq (6) has this correct
  # Here the variance is for sum, while in the paper the variance is for mean
  # which explains the proportional difference of 1/N
  est_list$hat_v_y <- (t_pi2_tilde + t_hat_epsilon) - # a (has been checked)
    (1/N) * (t_e^2 - est_list$v_y_hat + 2 * t_pi_tilde * est_list$y_hat - t_pi_tilde^2) # b
  est_list
}


#' Estimate elpd using the standard simple-re-sample without
#' resampling (SRS-WOR) estimator
#' @noRd
#' @param x A `psis_loo_ss` object.
#' @return A `psis_loo_ss` object.
loo_subsample_estimation_srs <- function(x) {
  checkmate::assert_class(x, "psis_loo_ss")

  elpd_loo_est <- srs_est(y = x$pointwise[, "elpd_loo"], y_approx = x$loo_subsampling$elpd_loo_approx)
  x$estimates["elpd_loo", "Estimate"] <- elpd_loo_est$y_hat
  x$estimates["elpd_loo", "SE"] <- sqrt(elpd_loo_est$hat_v_y)
  x$estimates["elpd_loo", "subsampling SE"] <- sqrt(elpd_loo_est$v_y_hat)

  p_loo_est <- srs_est(y = x$pointwise[, "p_loo"], y_approx = x$loo_subsampling$elpd_loo_approx)
  x$estimates["p_loo", "Estimate"] <- p_loo_est$y_hat
  x$estimates["p_loo", "SE"] <- sqrt(p_loo_est$hat_v_y)
  x$estimates["p_loo", "subsampling SE"] <- sqrt(p_loo_est$v_y_hat)

  update_psis_loo_ss_estimates(x)
}

#' Simple-re-sample without resampling (SRS-WOR) estimation
#' @noRd
#' @param y The values observed.
#' @param y_approx A vector of length N.
#' @return A list of estimates.
srs_est <- function(y, y_approx) {
  checkmate::assert_numeric(y)
  checkmate::assert_numeric(y_approx, min.len = length(y))
  N <- length(y_approx)
  m <- length(y)
  est_list <- list(m = m)
  est_list$y_hat <- N * mean(y)
  est_list$v_y_hat <- N^2 * (1-m/N) * var(y)/m
  est_list$hat_v_y <- N * var(y)

  est_list
}



## Specialized assertions of objects ---

#' Assert that the object has the expected properties
#' @noRd
#' @param x An object to assert.
#' @param N The total number of data points in data.
#' @param estimator The estimator used.
#' @return An asserted object of `x`.
assert_observations <- function(x, N, estimator) {
  checkmate::assert_int(N)
  checkmate::assert_choice(estimator, choices = estimator_choices())
  if (is.null(x)) return(x)
  if (checkmate::test_class(x, "psis_loo_ss")) {
    x <- obs_idx(x)
    checkmate::assert_integer(x, lower = 1, upper = N, any.missing = FALSE)
    return(x)
  }
  x <- as.integer(x)
  if (length(x) > 1) {
    checkmate::assert_integer(x, lower = 1, upper = N, any.missing = FALSE)
    if (estimator %in% "hh_pps") {
      message("Sampling proportional to elpd approximation and with replacement assumed.")
    }
    if (estimator %in% c("diff_srs", "srs")) {
      message("Simple random sampling with replacement assumed.")
    }
  } else {
    checkmate::assert_integer(x, lower = 1, any.missing = FALSE)
  }
  x
}

#' Assert that the object has the expected properties
#' @noRd
#' @inheritParams assert_observations
#' @return An asserted object of `x`.
assert_subsample_idxs <- function(x) {
  checkmate::assert_data_frame(x,
                               types = c("integer", "integer"),
                               any.missing = FALSE,
                               min.rows = 1,
                               col.names = "named")
  checkmate::assert_names(names(x), identical.to = c("idx", "m_i"))
  checkmate::assert_integer(x$idx, lower = 1, any.missing = FALSE, unique = TRUE)
  checkmate::assert_integer(x$m_i, lower = 1, any.missing = FALSE)
  x
}

#' Assert that the object has the expected properties
#' @noRd
#' @inheritParams assert_observations
#' @return An asserted object of `x`.
assert_psis_loo_ss <- function(x) {
  checkmate::assert_class(x, "psis_loo_ss")
  checkmate::assert_names(names(x), must.include = c("estimates", "pointwise", "diagnostics", "psis_object", "loo_subsampling"))
  checkmate::assert_names(rownames(x$estimates), must.include = c("elpd_loo", "p_loo", "looic"))
  checkmate::assert_names(colnames(x$estimates), must.include = c("Estimate", "SE", "subsampling SE"))
  assert_subsampling_pointwise(x$pointwise)
  checkmate::assert_names(names(x$loo_subsampling),
                          must.include = c("elpd_loo_approx",
                                           "loo_approximation", "loo_approximation_draws",
                                           "estimator",
                                           "data_dim", "ndraws"))
  checkmate::assert_numeric(x$loo_subsampling$elpd_loo_approx, any.missing = FALSE, len = x$loo_subsampling$data_dim[1])
  checkmate::assert_choice(x$loo_subsampling$loo_approximation, choices = loo_approximation_choices(api = FALSE))
  checkmate::assert_int(x$loo_subsampling$loo_approximation_draws, null.ok = TRUE)
  checkmate::assert_choice(x$loo_subsampling$estimator, choices = estimator_choices())
  checkmate::assert_integer(x$loo_subsampling$data_dim, any.missing = TRUE, len = 2)
  checkmate::assert_int(x$loo_subsampling$data_dim[1], na.ok = FALSE)
  checkmate::assert_integer(x$loo_subsampling$ndraws, len = 1, any.missing = TRUE)
  x
}

#' Assert that the object has the expected properties
#' @noRd
#' @inheritParams assert_observations
#' @return An asserted object of `x`.
assert_subsampling_pointwise <- function(x) {
  checkmate::assert_matrix(x,
                           any.missing = FALSE,
                           ncols = 8)
  checkmate::assert_names(colnames(x), identical.to = c("elpd_loo", "mcse_elpd_loo", "p_loo", "looic", "influence_pareto_k", "idx", "m_i", "elpd_loo_approx"))
  x
}

options(mc.cores = 1)

test_that("overall loo_subampling works as expected (compared with loo) for diff_est", {
  set.seed(123)
  N <- 1000
  K <- 10
  S <- 1000
  a0 <- 3
  b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y)
  b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y, K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    # each time called internally within loo the arguments will be equal to:
    # data_i: ith row of fdata (fake_data[i,, drop=FALSE])
    # draws: entire fake_posterior matrix
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(
    true_loo <- loo(
      llfun_test,
      draws = fake_posterior,
      data = fake_data,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(true_loo, "psis_loo")
  expect_silent(
    loo_ss <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 500,
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(loo_ss, "psis_loo_ss")

  # Check consistency
  expect_equal(
    loo_ss$pointwise[, "elpd_loo_approx"],
    loo_ss$loo_subsampling$elpd_loo_approx[loo_ss$pointwise[, "idx"]],
    ignore_attr = TRUE
  )

  # Expect values
  z <- 2
  expect_lte(
    loo_ss$estimates["elpd_loo", "Estimate"] -
      z * loo_ss$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["elpd_loo", "Estimate"] +
      z * loo_ss$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_lte(
    loo_ss$estimates["p_loo", "Estimate"] -
      z * loo_ss$estimates["p_loo", "subsampling SE"],
    true_loo$estimates["p_loo", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["p_loo", "Estimate"] +
      z * loo_ss$estimates["p_loo", "subsampling SE"],
    true_loo$estimates["p_loo", "Estimate"]
  )
  expect_lte(
    loo_ss$estimates["looic", "Estimate"] -
      z * loo_ss$estimates["looic", "subsampling SE"],
    true_loo$estimates["looic", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["looic", "Estimate"] +
      z * loo_ss$estimates["looic", "subsampling SE"],
    true_loo$estimates["looic", "Estimate"]
  )

  expect_failure(expect_equal(
    true_loo$estimates["elpd_loo", "Estimate"],
    loo_ss$estimates["elpd_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    true_loo$estimates["p_loo", "Estimate"],
    loo_ss$estimates["p_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    true_loo$estimates["looic", "Estimate"],
    loo_ss$estimates["looic", "Estimate"],
    tolerance = 0.00000001
  ))

  # Test that observations works as expected
  expect_message(
    loo_ss2 <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = obs_idx(loo_ss),
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_equal(loo_ss2$estimates, loo_ss$estimates, tolerance = 0.00000001)
  expect_silent(
    loo_ss2 <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = loo_ss,
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_equal(loo_ss2$estimates, loo_ss$estimates, tolerance = 0.00000001)

  # Test lpd
  expect_silent(
    loo_ss_lpd <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 500,
      loo_approximation = "lpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(loo_ss_lpd, "psis_loo_ss")
  z <- 2
  expect_lte(
    loo_ss_lpd$estimates["elpd_loo", "Estimate"] -
      z * loo_ss_lpd$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_gte(
    loo_ss_lpd$estimates["elpd_loo", "Estimate"] +
      z * loo_ss_lpd$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_lte(
    loo_ss_lpd$estimates["p_loo", "Estimate"] -
      z * loo_ss_lpd$estimates["p_loo", "subsampling SE"],
    true_loo$estimates["p_loo", "Estimate"]
  )
  expect_gte(
    loo_ss_lpd$estimates["p_loo", "Estimate"] +
      z * loo_ss_lpd$estimates["p_loo", "subsampling SE"],
    true_loo$estimates["p_loo", "Estimate"]
  )
  expect_lte(
    loo_ss_lpd$estimates["looic", "Estimate"] -
      z * loo_ss_lpd$estimates["looic", "subsampling SE"],
    true_loo$estimates["looic", "Estimate"]
  )
  expect_gte(
    loo_ss_lpd$estimates["looic", "Estimate"] +
      z * loo_ss_lpd$estimates["looic", "subsampling SE"],
    true_loo$estimates["looic", "Estimate"]
  )

  expect_failure(expect_equal(
    true_loo$estimates["elpd_loo", "Estimate"],
    loo_ss_lpd$estimates["elpd_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    true_loo$estimates["p_loo", "Estimate"],
    loo_ss_lpd$estimates["p_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    true_loo$estimates["looic", "Estimate"],
    loo_ss_lpd$estimates["looic", "Estimate"],
    tolerance = 0.00000001
  ))

  expect_silent(
    loo_ss_lpd10 <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 500,
      loo_approximation = "lpd",
      loo_approximation_draws = 10,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(loo_ss_lpd10, "psis_loo_ss")

  z <- 2
  expect_lte(
    loo_ss_lpd10$estimates["elpd_loo", "Estimate"] -
      z * loo_ss_lpd10$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_gte(
    loo_ss_lpd10$estimates["elpd_loo", "Estimate"] +
      z * loo_ss_lpd10$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_lte(
    loo_ss_lpd10$estimates["p_loo", "Estimate"] -
      z * loo_ss_lpd10$estimates["p_loo", "subsampling SE"],
    true_loo$estimates["p_loo", "Estimate"]
  )
  expect_gte(
    loo_ss_lpd10$estimates["p_loo", "Estimate"] +
      z * loo_ss_lpd10$estimates["p_loo", "subsampling SE"],
    true_loo$estimates["p_loo", "Estimate"]
  )
  expect_lte(
    loo_ss_lpd10$estimates["looic", "Estimate"] -
      z * loo_ss_lpd10$estimates["looic", "subsampling SE"],
    true_loo$estimates["looic", "Estimate"]
  )
  expect_gte(
    loo_ss_lpd10$estimates["looic", "Estimate"] +
      z * loo_ss_lpd10$estimates["looic", "subsampling SE"],
    true_loo$estimates["looic", "Estimate"]
  )

  expect_failure(expect_equal(
    true_loo$estimates["elpd_loo", "Estimate"],
    loo_ss_lpd10$estimates["elpd_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    true_loo$estimates["p_loo", "Estimate"],
    loo_ss_lpd10$estimates["p_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    true_loo$estimates["looic", "Estimate"],
    loo_ss_lpd10$estimates["looic", "Estimate"],
    tolerance = 0.00000001
  ))

  # Test conversion of objects
  expect_silent(true_loo_2 <- loo:::as.psis_loo.psis_loo(true_loo))
  expect_silent(true_loo_ss <- loo:::as.psis_loo_ss.psis_loo(true_loo))
  expect_s3_class(true_loo_ss, "psis_loo_ss")
  expect_silent(true_loo_conv <- loo:::as.psis_loo.psis_loo_ss(true_loo_ss))
  expect_failure(expect_s3_class(true_loo_conv, "psis_loo_ss"))
  expect_equal(true_loo_conv, true_loo)
  expect_error(loo:::as.psis_loo.psis_loo_ss(loo_ss))
})

test_that("loo with subsampling of all observations works as ordinary loo.", {
  set.seed(123)
  N <- 1000
  K <- 10
  S <- 1000
  a0 <- 3
  b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y)
  b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y, K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(
    true_loo <- loo(
      llfun_test,
      draws = fake_posterior,
      data = fake_data,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(true_loo, "psis_loo")
  expect_silent(
    loo_ss <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 1000,
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(loo_ss, "psis_loo_ss")
  expect_error(
    loo_ss <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 1001,
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )

  expect_equal(
    true_loo$estimates["elpd_loo", "Estimate"],
    loo_ss$estimates["elpd_loo", "Estimate"],
    tolerance = 0.00000001
  )
  expect_equal(
    true_loo$estimates["p_loo", "Estimate"],
    loo_ss$estimates["p_loo", "Estimate"],
    tolerance = 0.00000001
  )
  expect_equal(
    true_loo$estimates["looic", "Estimate"],
    loo_ss$estimates["looic", "Estimate"],
    tolerance = 0.00000001
  )

  expect_equal(dim(true_loo), dim(loo_ss))
  expect_equal(true_loo$diagnostics, loo_ss$diagnostics)
  expect_equal(max(loo_ss$pointwise[, "m_i"]), 1)
})

test_that("overall loo_subsample works with diff_srs as expected (compared with loo)", {
  set.seed(123)
  N <- 1000
  K <- 10
  S <- 1000
  a0 <- 3
  b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y)
  b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y, K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(
    true_loo <- loo(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_silent(
    loo_ss <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 200,
      loo_approximation = "plpd",
      estimator = "diff_srs",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_equal(
    true_loo$estimates[1, 1],
    loo_ss$estimates[1, 1],
    tolerance = 0.1
  )
})

test_that("Test the srs estimator with 'none' approximation", {
  set.seed(123)
  N <- 1000
  K <- 10
  S <- 1000
  a0 <- 3
  b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y)
  b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y, K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(
    true_loo <- loo(
      llfun_test,
      draws = fake_posterior,
      data = fake_data,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(true_loo, "psis_loo")
  expect_silent(
    loo_ss <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 200,
      loo_approximation = "none",
      estimator = "srs",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(loo_ss, "psis_loo_ss")
  expect_error(
    loo_ss <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 1100,
      loo_approximation = "none",
      estimator = "srs",
      r_eff = rep(1, nrow(fake_data))
    )
  )

  expect_equal(length(obs_idx(loo_ss)), nobs(loo_ss))

  # Check consistency
  expect_equal(
    loo_ss$pointwise[, "elpd_loo_approx"],
    loo_ss$loo_subsampling$elpd_loo_approx[loo_ss$pointwise[, "idx"]],
    ignore_attr = TRUE
  )

  # Expect values
  z <- 2
  expect_lte(
    loo_ss$estimates["elpd_loo", "Estimate"] -
      z * loo_ss$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["elpd_loo", "Estimate"] +
      z * loo_ss$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_lte(
    loo_ss$estimates["p_loo", "Estimate"] -
      z * loo_ss$estimates["p_loo", "subsampling SE"],
    true_loo$estimates["p_loo", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["p_loo", "Estimate"] +
      z * loo_ss$estimates["p_loo", "subsampling SE"],
    true_loo$estimates["p_loo", "Estimate"]
  )
  expect_lte(
    loo_ss$estimates["looic", "Estimate"] -
      z * loo_ss$estimates["looic", "subsampling SE"],
    true_loo$estimates["looic", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["looic", "Estimate"] +
      z * loo_ss$estimates["looic", "subsampling SE"],
    true_loo$estimates["looic", "Estimate"]
  )

  expect_failure(expect_equal(
    true_loo$estimates["elpd_loo", "Estimate"],
    loo_ss$estimates["elpd_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    true_loo$estimates["p_loo", "Estimate"],
    loo_ss$estimates["p_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    true_loo$estimates["looic", "Estimate"],
    loo_ss$estimates["looic", "Estimate"],
    tolerance = 0.00000001
  ))
})

test_that("Test the Hansen-Hurwitz estimator", {
  set.seed(123)
  N <- 1000
  K <- 10
  S <- 1000
  a0 <- 3
  b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y)
  b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y, K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(
    true_loo <- loo(
      llfun_test,
      draws = fake_posterior,
      data = fake_data,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(true_loo, "psis_loo")
  expect_silent(
    loo_ss <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 300,
      loo_approximation = "plpd",
      estimator = "hh_pps",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(loo_ss, "psis_loo_ss")
  expect_silent(
    loo_ss_max <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 1100,
      loo_approximation = "plpd",
      estimator = "hh_pps",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(loo_ss_max, "psis_loo_ss")
  expect_silent(
    loo_ss_max2 <- update(
      loo_ss,
      draws = fake_posterior,
      data = fake_data,
      observations = 1100,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_equal(nobs(loo_ss_max2), 1100)
  expect_gt(max(loo_ss_max2$pointwise[, "m_i"]), 1)
  expect_error(
    loo_ss_max2 <- update(
      loo_ss_max2,
      draws = fake_posterior,
      data = fake_data,
      observations = 300,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_silent(
    loo_ss_max3 <- update(
      loo_ss,
      draws = fake_posterior,
      data = fake_data,
      observations = 1500,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_silent(
    loo_ss2 <- update(
      loo_ss,
      draws = fake_posterior,
      data = fake_data,
      observations = loo_ss,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_error(
    loo_ss2 <- update(
      loo_ss,
      draws = fake_posterior,
      data = fake_data,
      observations = loo_ss,
      loo_approximation = "lpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_equal(loo_ss$estimates, loo_ss2$estimates)
  expect_equal(length(obs_idx(loo_ss_max)), length(obs_idx(loo_ss_max2)))
  expect_equal(length(obs_idx(loo_ss_max)), nobs(loo_ss_max))

  # Check consistency
  expect_equal(
    loo_ss$pointwise[, "elpd_loo_approx"],
    loo_ss$loo_subsampling$elpd_loo_approx[loo_ss$pointwise[, "idx"]],
    ignore_attr = TRUE
  )
  # Check consistency
  expect_equal(
    loo_ss_max$pointwise[, "elpd_loo_approx"],
    loo_ss_max$loo_subsampling$elpd_loo_approx[loo_ss_max$pointwise[, "idx"]],
    ignore_attr = TRUE
  )

  # Expect values
  z <- 2
  expect_lte(
    loo_ss$estimates["elpd_loo", "Estimate"] -
      z * loo_ss$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["elpd_loo", "Estimate"] +
      z * loo_ss$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_lte(
    loo_ss$estimates["p_loo", "Estimate"] -
      z * loo_ss$estimates["p_loo", "subsampling SE"],
    true_loo$estimates["p_loo", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["p_loo", "Estimate"] +
      z * loo_ss$estimates["p_loo", "subsampling SE"],
    true_loo$estimates["p_loo", "Estimate"]
  )
  expect_lte(
    loo_ss$estimates["looic", "Estimate"] -
      z * loo_ss$estimates["looic", "subsampling SE"],
    true_loo$estimates["looic", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["looic", "Estimate"] +
      z * loo_ss$estimates["looic", "subsampling SE"],
    true_loo$estimates["looic", "Estimate"]
  )

  expect_failure(expect_equal(
    true_loo$estimates["elpd_loo", "Estimate"],
    loo_ss$estimates["elpd_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    true_loo$estimates["p_loo", "Estimate"],
    loo_ss$estimates["p_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    true_loo$estimates["looic", "Estimate"],
    loo_ss$estimates["looic", "Estimate"],
    tolerance = 0.00000001
  ))

  expect_lte(
    loo_ss_max$estimates["elpd_loo", "Estimate"] -
      z * loo_ss_max$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_gte(
    loo_ss_max$estimates["elpd_loo", "Estimate"] +
      z * loo_ss_max$estimates["elpd_loo", "subsampling SE"],
    true_loo$estimates["elpd_loo", "Estimate"]
  )
})


test_that("update.psis_loo_ss works as expected (compared with loo)", {
  set.seed(123)
  N <- 1000
  K <- 10
  S <- 1000
  a0 <- 3
  b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y)
  b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y, K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(
    true_loo <- loo(
      llfun_test,
      draws = fake_posterior,
      data = fake_data,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(true_loo, "psis_loo")
  expect_silent(
    loo_ss <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 500,
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_s3_class(loo_ss, "psis_loo_ss")

  # Check error when draws and data dimensions differ
  expect_error(
    loo_ss2 <- update(
      object = loo_ss,
      draws = cbind(fake_posterior, 1),
      data = fake_data,
      observations = 600,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_error(
    loo_ss2 <- update(
      object = loo_ss,
      draws = fake_posterior,
      data = fake_data[-1, ],
      observations = 600,
      r_eff = rep(1, nrow(fake_data))
    )
  )

  # Add tests for adding observations
  expect_silent(
    loo_ss2 <- update(
      object = loo_ss,
      draws = fake_posterior,
      data = fake_data,
      observations = 600,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_equal(dim(loo_ss2)[2] - dim(loo_ss)[2], expected = 100)
  expect_equal(dim(loo_ss2)[2], expected = dim(loo_ss2$pointwise)[1])
  expect_length(loo_ss2$diagnostics$pareto_k, 600)
  expect_length(loo_ss2$diagnostics$n_eff, 600)
  for (i in 1:nrow(loo_ss2$estimates)) {
    expect_lt(
      loo_ss2$estimates[i, "subsampling SE"],
      loo_ss$estimates[i, "subsampling SE"]
    )
  }

  expect_silent(
    loo_ss2b <- update(
      object = loo_ss,
      draws = fake_posterior,
      data = fake_data
    )
  )
  expect_equal(loo_ss2b$estimates, loo_ss$estimates)
  expect_equal(loo_ss2b$pointwise, loo_ss$pointwise)
  expect_equal(loo_ss2b$diagnostics$pareto_k, loo_ss$diagnostics$pareto_k)
  expect_equal(loo_ss2b$diagnostics$n_eff, loo_ss$diagnostics$n_eff)

  expect_silent(
    loo_ss3 <- update(
      object = loo_ss2,
      draws = fake_posterior,
      data = fake_data,
      observations = loo_ss
    )
  )
  expect_equal(loo_ss3$estimates, loo_ss$estimates)
  expect_equal(loo_ss3$pointwise, loo_ss$pointwise)
  expect_equal(loo_ss3$diagnostics$pareto_k, loo_ss$diagnostics$pareto_k)
  expect_equal(loo_ss3$diagnostics$n_eff, loo_ss$diagnostics$n_eff)

  expect_silent(
    loo_ss4 <- update(
      object = loo_ss,
      draws = fake_posterior,
      data = fake_data,
      observations = 1000,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_equal(loo_ss4$estimates[, 1], true_loo$estimates[, 1])
  expect_equal(
    loo_ss4$estimates[, 2],
    true_loo$estimates[, 2],
    tolerance = 0.001
  )

  expect_silent(
    loo_ss5 <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 1000,
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )

  ss4_order <- order(loo_ss4$pointwise[, "idx"])
  expect_equal(
    loo_ss4$pointwise[ss4_order, c(1, 3, 4)],
    loo_ss5$pointwise[, c(1, 3, 4)]
  )
  expect_equal(
    loo_ss4$diagnostics$pareto_k[ss4_order],
    loo_ss5$diagnostics$pareto_k
  )
  expect_equal(loo_ss4$diagnostics$n_eff[ss4_order], loo_ss5$diagnostics$n_eff)
  expect_equal(
    loo_ss4$pointwise[ss4_order, c(1, 3, 4)],
    true_loo$pointwise[, c(1, 3, 4)]
  )
  expect_equal(
    loo_ss4$diagnostics$pareto_k[ss4_order],
    true_loo$diagnostics$pareto_k
  )
  expect_equal(loo_ss4$diagnostics$n_eff[ss4_order], true_loo$diagnostics$n_eff)

  expect_error(
    loo_ss_min <- update(
      object = loo_ss,
      draws = fake_posterior,
      data = fake_data,
      observations = 50,
      r_eff = rep(1, nrow(fake_data))
    )
  )

  expect_silent(true_loo_ss <- loo:::as.psis_loo_ss.psis_loo(true_loo))
  expect_silent(
    loo_ss_subset0 <- update(
      true_loo_ss,
      observations = loo_ss,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_true(identical(obs_idx(loo_ss_subset0), obs_idx(loo_ss)))
  expect_silent(
    loo_ss_subset1 <- update(
      object = loo_ss,
      observations = loo_ss,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_message(
    loo_ss_subset2 <- update(
      object = loo_ss,
      observations = obs_idx(loo_ss)[1:10],
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_equal(nobs(loo_ss_subset2), 10)

  expect_silent(true_loo_ss <- loo:::as.psis_loo_ss.psis_loo(true_loo))
  set.seed(4711)
  expect_silent(
    loo_ss2 <- update(
      object = loo_ss,
      draws = fake_posterior,
      data = fake_data,
      observations = 600,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_silent(
    loo_ss2_subset0 <- update(
      object = true_loo_ss,
      observations = loo_ss2,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_true(setequal(obs_idx(loo_ss2), obs_idx(loo_ss2_subset0)))
  expect_true(identical(obs_idx(loo_ss2), obs_idx(loo_ss2_subset0)))
  expect_true(identical(loo_ss2$diagnostic, loo_ss2_subset0$diagnostic))

  # Add tests for changing approx variable
  expect_silent(
    loo_ss_lpd <- update(
      object = loo_ss,
      draws = fake_posterior,
      data = fake_data,
      loo_approximation = "lpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_failure(expect_equal(
    loo_ss_lpd$loo_subsampling$elpd_loo_approx,
    loo_ss$loo_subsampling$elpd_loo_approx
  ))
  expect_equal(dim(loo_ss_lpd)[2], dim(loo_ss)[2])
  expect_equal(dim(loo_ss_lpd)[2], dim(loo_ss_lpd$pointwise)[1])
  expect_length(loo_ss_lpd$diagnostics$pareto_k, 500)
  expect_length(loo_ss_lpd$diagnostics$n_eff, 500)
  expect_failure(expect_equal(
    loo_ss_lpd$estimates[1, "subsampling SE"],
    loo_ss$estimates[1, "subsampling SE"]
  ))
  expect_failure(expect_equal(
    loo_ss_lpd$estimates[3, "subsampling SE"],
    loo_ss$estimates[3, "subsampling SE"]
  ))
})

test_that("loo_compare_subsample", {
  skip_on_cran() # to get under cran check time limit

  set.seed(123)
  N <- 1000
  x1 <- rnorm(N)
  x2 <- rnorm(N)
  x3 <- rnorm(N)
  sigma <- 2
  y <- rnorm(N, 1 + 2 * x1 - 2 * x2 - 1 * x3, sd = sigma)
  X <- cbind("x0" = rep(1, N), x1, x2, x3)

  # Generate samples from posterior
  samples_blin <- function(X, y, sigma, draws = 1000) {
    XtX <- t(X) %*% X
    b_hat <- solve(XtX) %*% (t(X) %*% y)
    Lambda_n = XtX + diag(ncol(X))
    mu_n <- solve(Lambda_n) %*%
      (XtX %*% b_hat + diag(ncol(X)) %*% rep(0, ncol(X)))
    L <- t(chol(sigma^2 * solve(Lambda_n)))
    draws_mat <- matrix(0, ncol = ncol(X), nrow = draws)
    for (i in 1:draws) {
      z <- rnorm(length(mu_n))
      draws_mat[i, ] <- L %*% z + mu_n
    }
    draws_mat
  }

  fake_posterior1 <- samples_blin(X[, 1:2], y, sigma, draws = 1000)
  fake_posterior2 <- samples_blin(X[, 1:3], y, sigma, draws = 1000)
  fake_posterior3 <- samples_blin(X, y, sigma, draws = 1000)

  fake_data1 <- data.frame(y, X[, 1:2])
  fake_data2 <- data.frame(y, X[, 1:3])
  fake_data3 <- data.frame(y, X)

  llfun_test <- function(data_i, draws) {
    dnorm(
      x = data_i$y,
      mean = draws %*% t(data_i[, -1, drop = FALSE]),
      sd = sigma,
      log = TRUE
    )
  }

  expect_silent(
    l1 <- loo(
      llfun_test,
      data = fake_data1,
      draws = fake_posterior1,
      r_eff = rep(1, N)
    )
  )
  expect_silent(
    l2 <- loo(
      llfun_test,
      data = fake_data2,
      draws = fake_posterior2,
      r_eff = rep(1, N)
    )
  )
  expect_silent(
    l3 <- loo(
      llfun_test,
      data = fake_data3,
      draws = fake_posterior3,
      r_eff = rep(1, N)
    )
  )

  expect_silent(
    lss1 <- loo_subsample(
      llfun_test,
      data = fake_data1,
      draws = fake_posterior1,
      observations = 100,
      r_eff = rep(1, N)
    )
  )
  expect_silent(
    lss2 <- loo_subsample(
      llfun_test,
      data = fake_data2,
      draws = fake_posterior2,
      observations = 100,
      r_eff = rep(1, N)
    )
  )
  expect_silent(
    lss3 <- loo_subsample(
      llfun_test,
      data = fake_data3,
      draws = fake_posterior3,
      observations = 100,
      r_eff = rep(1, N)
    )
  )
  expect_silent(
    lss2o1 <- loo_subsample(
      llfun_test,
      data = fake_data2,
      draws = fake_posterior2,
      observations = lss1,
      r_eff = rep(1, N)
    )
  )
  expect_silent(
    lss3o1 <- loo_subsample(
      llfun_test,
      data = fake_data3,
      draws = fake_posterior3,
      observations = lss1,
      r_eff = rep(1, N)
    )
  )
  expect_silent(
    lss2hh <- loo_subsample(
      llfun_test,
      data = fake_data2,
      draws = fake_posterior2,
      observations = 100,
      estimator = "hh_pps",
      r_eff = rep(1, N)
    )
  )

  expect_snapshot(
    lcss <- loo:::loo_compare.psis_loo_ss_list(x = list(lss1, lss2, lss3))
  )
  expect_warning(
    lcss2 <- loo:::loo_compare.psis_loo_ss_list(x = list(lss1, lss2, lss3o1))
  )
  expect_silent(
    lcsso <- loo:::loo_compare.psis_loo_ss_list(x = list(lss1, lss2o1, lss3o1))
  )
  expect_warning(
    lcssohh <- loo:::loo_compare.psis_loo_ss_list(
      x = list(lss1, lss2hh, lss3o1)
    )
  )
  expect_message(
    lcssf1 <- loo:::loo_compare.psis_loo_ss_list(
      x = list(loo:::as.psis_loo_ss.psis_loo(l1), lss2o1, lss3o1)
    )
  )
  expect_message(
    lcssf2 <- loo:::loo_compare.psis_loo_ss_list(
      x = list(
        loo:::as.psis_loo_ss.psis_loo(l1),
        lss2o1,
        loo:::as.psis_loo_ss.psis_loo(l3)
      )
    )
  )

  expect_equal(lcss[, 1], lcsso[, 1], tolerance = 1)
  expect_equal(lcss2[, 1], lcsso[, 1], tolerance = 1)
  expect_equal(lcssohh[, 1], lcsso[, 1], tolerance = 1)
  expect_equal(lcssf1[, 1], lcsso[, 1], tolerance = 1)
  expect_equal(lcssf2[, 1], lcsso[, 1], tolerance = 1)

  expect_gt(lcss[, 2][2], lcsso[, 2][2])
  expect_gt(lcss[, 2][3], lcsso[, 2][3])
  expect_gt(lcss2[, 2][2], lcsso[, 2][2])
  expect_equal(lcss2[, 2][3], lcsso[, 2][3])
  expect_gt(lcssohh[, 2][2], lcsso[, 2][2])
  expect_equal(lcssohh[, 2][3], lcsso[, 2][3])

  expect_silent(
    lcss2m <- loo:::loo_compare.psis_loo_ss_list(x = list(lss2o1, lss3o1))
  )
  expect_equal(unname(lcss2m[,]), unname(lcsso[1:2, ]))

  expect_snapshot(lcssapi <- loo_compare(lss1, lss2, lss3))
  expect_equal(lcssapi, lcss)
  expect_warning(lcssohhapi <- loo_compare(lss1, lss2hh, lss3o1))
  expect_equal(lcssohhapi, lcssohh)
  expect_silent(lcss2mapi <- loo_compare(lss2o1, lss3o1))
  expect_equal(lcss2mapi, lcss2m)
})

test_that("Test 'tis' and 'sis'", {
  skip_on_cran()

  set.seed(123)
  N <- 1000
  K <- 10
  S <- 1000
  a0 <- 3
  b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y)
  b <- b0 + N * K - sum(y)
  fake_posterior <- draws <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y, K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(
    loo_ss_full <-
      loo_subsample(
        x = llfun_test,
        draws = fake_posterior,
        data = fake_data,
        observations = 1000,
        loo_approximation = "plpd",
        r_eff = rep(1, nrow(fake_data))
      )
  )
  expect_silent(
    loo_ss_plpd <-
      loo_subsample(
        x = llfun_test,
        draws = fake_posterior,
        data = fake_data,
        observations = 100,
        loo_approximation = "plpd",
        r_eff = rep(1, nrow(fake_data))
      )
  )
  expect_silent(
    loo_ss_tis_S1000 <-
      loo_subsample(
        x = llfun_test,
        draws = fake_posterior,
        data = fake_data,
        observations = 100,
        loo_approximation = "tis",
        r_eff = rep(1, nrow(fake_data))
      )
  )
  expect_silent(
    loo_ss_tis_S100 <-
      loo_subsample(
        x = llfun_test,
        draws = fake_posterior,
        data = fake_data,
        observations = 100,
        loo_approximation = "tis",
        loo_approximation_draws = 100,
        r_eff = rep(1, nrow(fake_data))
      )
  )
  expect_silent(
    loo_ss_tis_S10 <-
      loo_subsample(
        x = llfun_test,
        draws = fake_posterior,
        data = fake_data,
        observations = 100,
        loo_approximation = "tis",
        loo_approximation_draws = 10,
        r_eff = rep(1, nrow(fake_data))
      )
  )
  expect_silent(
    loo_ss_sis_S1000 <-
      loo_subsample(
        x = llfun_test,
        draws = fake_posterior,
        data = fake_data,
        observations = 100,
        loo_approximation = "sis",
        r_eff = rep(1, nrow(fake_data))
      )
  )
  expect_silent(
    loo_ss_sis_S100 <-
      loo_subsample(
        x = llfun_test,
        draws = fake_posterior,
        data = fake_data,
        observations = 100,
        loo_approximation = "sis",
        loo_approximation_draws = 100,
        r_eff = rep(1, nrow(fake_data))
      )
  )
  expect_silent(
    loo_ss_sis_S10 <-
      loo_subsample(
        x = llfun_test,
        draws = fake_posterior,
        data = fake_data,
        observations = 100,
        loo_approximation = "sis",
        loo_approximation_draws = 10,
        r_eff = rep(1, nrow(fake_data))
      )
  )

  SEs <- 4
  expect_gt(
    loo_ss_tis_S1000$estimates["elpd_loo", "Estimate"] +
      SEs * loo_ss_tis_S1000$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_lt(
    loo_ss_tis_S1000$estimates["elpd_loo", "Estimate"] -
      SEs * loo_ss_tis_S1000$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_gt(
    loo_ss_tis_S100$estimates["elpd_loo", "Estimate"] +
      SEs * loo_ss_tis_S100$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_lt(
    loo_ss_tis_S100$estimates["elpd_loo", "Estimate"] -
      SEs * loo_ss_tis_S100$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_gt(
    loo_ss_tis_S10$estimates["elpd_loo", "Estimate"] +
      SEs * loo_ss_tis_S10$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_lt(
    loo_ss_tis_S10$estimates["elpd_loo", "Estimate"] -
      SEs * loo_ss_tis_S10$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )

  expect_gt(
    loo_ss_sis_S1000$estimates["elpd_loo", "Estimate"] +
      SEs * loo_ss_sis_S1000$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_lt(
    loo_ss_sis_S1000$estimates["elpd_loo", "Estimate"] -
      SEs * loo_ss_sis_S1000$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_gt(
    loo_ss_sis_S100$estimates["elpd_loo", "Estimate"] +
      SEs * loo_ss_sis_S100$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_lt(
    loo_ss_sis_S100$estimates["elpd_loo", "Estimate"] -
      SEs * loo_ss_sis_S100$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_gt(
    loo_ss_sis_S10$estimates["elpd_loo", "Estimate"] +
      SEs * loo_ss_sis_S10$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_lt(
    loo_ss_sis_S10$estimates["elpd_loo", "Estimate"] -
      SEs * loo_ss_sis_S10$estimates["elpd_loo", "subsampling SE"],
    loo_ss_full$estimates["elpd_loo", "Estimate"]
  )
})

options(mc.cores = 1)

generate_test_elpd_dataset <- function() {
  N <- 10
  K <- 10
  S <- 1000
  a0 <- 3
  b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y)
  b <- b0 + N * K - sum(y)
  fake_posterior <- draws <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y, K)
  rm(N, K, S, a0, b0, p, y, a, b)

  list(fake_posterior = fake_posterior, fake_data = fake_data)
}

test_elpd_loo_approximation <- function(cores) {
  set.seed(123)
  test_data <- generate_test_elpd_dataset()
  fake_posterior <- test_data$fake_posterior
  fake_data <- test_data$fake_data

  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  # Compute plpd approximation
  expect_silent(
    pi_vals <- loo:::elpd_loo_approximation(
      .llfun = llfun_test,
      data = fake_data,
      draws = fake_posterior,
      loo_approximation = "plpd",
      cores = cores
    )
  )
  # Compute it manually
  point <- mean(fake_posterior)
  llik <- dbinom(fake_data$y, size = fake_data$K, prob = point, log = TRUE)
  abs_lliks <- abs(llik)
  man_elpd_loo_approximation <- abs_lliks / sum(abs_lliks)
  expect_equal(
    abs(pi_vals) / sum(abs(pi_vals)),
    man_elpd_loo_approximation,
    tolerance = 0.00001
  )

  # Compute lpd approximation
  expect_silent(
    pi_vals <- loo:::elpd_loo_approximation(
      .llfun = llfun_test,
      data = fake_data,
      draws = fake_posterior,
      loo_approximation = "lpd",
      cores = cores
    )
  )
  # Compute it manually
  llik <- numeric(10)
  for (i in seq_along(fake_data$y)) {
    llik[i] <- loo:::logMeanExp(dbinom(
      fake_data$y[i],
      size = fake_data$K,
      prob = fake_posterior,
      log = TRUE
    ))
  }
  abs_lliks <- abs(llik)
  man_approx_loo_variable <- abs_lliks / sum(abs_lliks)
  expect_equal(
    abs(pi_vals) / sum(abs(pi_vals)),
    man_approx_loo_variable,
    tolerance = 0.00001
  )

  # Compute waic approximation
  expect_silent(
    pi_vals_waic <- loo:::elpd_loo_approximation(
      .llfun = llfun_test,
      data = fake_data,
      draws = fake_posterior,
      loo_approximation = "waic",
      cores = cores
    )
  )
  expect_true(all(pi_vals > pi_vals_waic))
  expect_true(sum(pi_vals) - sum(pi_vals_waic) < 1)

  # Compute tis approximation
  expect_silent(
    pi_vals_tis <- loo:::elpd_loo_approximation(
      .llfun = llfun_test,
      data = fake_data,
      draws = fake_posterior,
      loo_approximation = "tis",
      loo_approximation_draws = 100,
      cores = cores
    )
  )
  expect_true(all(pi_vals > pi_vals_tis))
  expect_true(sum(pi_vals) - sum(pi_vals_tis) < 1)
}

test_that("elpd_loo_approximation works as expected", {
  test_elpd_loo_approximation(1)
})

test_that("elpd_loo_approximation with multiple cores", {
  test_elpd_loo_approximation(2)
})

test_that("Test loo_approximation_draws", {
  set.seed(123)
  N <- 1000
  K <- 10
  S <- 1000
  a0 <- 3
  b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y)
  b <- b0 + N * K - sum(y)
  fake_posterior <- draws <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y, K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(
    res1 <- loo:::elpd_loo_approximation(
      .llfun = llfun_test,
      data = fake_data,
      draws = fake_posterior,
      loo_approximation = "waic",
      loo_approximation_draws = NULL,
      cores = 1
    )
  )
  expect_silent(
    res2 <- loo:::elpd_loo_approximation(
      .llfun = llfun_test,
      data = fake_data,
      draws = fake_posterior,
      loo_approximation = "waic",
      loo_approximation_draws = 10,
      cores = 1
    )
  )
  expect_silent(
    res3 <- loo:::elpd_loo_approximation(
      .llfun = llfun_test,
      data = fake_data,
      draws = fake_posterior[1:10 * 100, ],
      loo_approximation = "waic",
      loo_approximation_draws = NULL,
      cores = 1
    )
  )
  expect_silent(
    res4 <- loo:::elpd_loo_approximation(
      .llfun = llfun_test,
      data = fake_data,
      draws = fake_posterior[1:10 * 100, , drop = FALSE],
      loo_approximation = "waic",
      loo_approximation_draws = NULL,
      cores = 1
    )
  )
  expect_failure(expect_equal(res1, res3))
  expect_equal(res2, res3)

  expect_silent(
    loo_ss1 <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 100,
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_silent(
    loo_ss2 <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 100,
      loo_approximation = "plpd",
      loo_approximation_draws = 10,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_silent(
    loo_ss3 <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 100,
      loo_approximation = "plpd",
      loo_approximation_draws = 31,
      r_eff = rep(1, nrow(fake_data))
    )
  )
  expect_error(
    loo_ss4 <- loo_subsample(
      x = llfun_test,
      draws = fake_posterior,
      data = fake_data,
      observations = 100,
      loo_approximation = "plpd",
      loo_approximation_draws = 3100,
      r_eff = rep(1, nrow(fake_data))
    )
  )

  expect_equal(
    names(loo_ss1$loo_subsampling),
    c(
      "elpd_loo_approx",
      "loo_approximation",
      "loo_approximation_draws",
      "estimator",
      ".llfun",
      ".llgrad",
      ".llhess",
      "data_dim",
      "ndraws"
    )
  )
  expect_null(loo_ss1$loo_subsampling$loo_approximation_draws)
  expect_equal(loo_ss2$loo_subsampling$loo_approximation_draws, 10L)
  expect_equal(loo_ss3$loo_subsampling$loo_approximation_draws, 31L)
})


test_that("waic using delta method and gradient", {
  if (FALSE) {
    # Code to generate testdata - saved and loaded to avoid dependency of mvtnorm
    set.seed(123)
    N <- 400
    beta <- c(1, 2)
    X_full <- matrix(rep(1, N), ncol = 1)
    X_full <- cbind(X_full, runif(N))
    S <- 1000
    y_full <- rnorm(n = N, mean = X_full %*% beta, sd = 1)
    X <- X_full
    y <- y_full
    Lambda_0 <- diag(length(beta))
    mu_0 <- c(0, 0)
    b_hat <- solve(t(X) %*% X) %*% t(X) %*% y
    mu_n <- solve(t(X) %*% X) %*% (t(X) %*% X %*% b_hat + Lambda_0 %*% mu_0)
    Lambda_n <- t(X) %*% X + Lambda_0
    # Uncomment row below when running. Commented out to remove CHECK warnings
    # fake_posterior <- mvtnorm::rmvnorm(n = S, mean = mu_n, sigma = solve(Lambda_n))
    colnames(fake_posterior) <- c("a", "b")
    fake_data <- data.frame(y, X)
    save(
      fake_posterior,
      fake_data,
      file = test_path("data-for-tests/normal_reg_waic_test_example.rda")
    )
  } else {
    load(file = test_path("data-for-tests/normal_reg_waic_test_example.rda"))
  }

  .llfun <- function(data_i, draws) {
    # data_i: ith row of fdata (fake_data[i,, drop=FALSE])
    # draws: entire fake_posterior matrix
    dnorm(
      data_i$y,
      mean = draws[, c("a", "b")] %*% t(as.matrix(data_i[, c("X1", "X2")])),
      sd = 1,
      log = TRUE
    )
  }

  .llgrad <- function(data_i, draws) {
    x_i <- data_i[, "X2"]
    gr <- cbind(
      data_i$y - draws[, "a"] - draws[, "b"] * x_i,
      (data_i$y - draws[, "a"] - draws[, "b"] * x_i) * x_i
    )
    colnames(gr) <- c("a", "b")
    gr
  }

  fake_posterior <- cbind(fake_posterior, runif(nrow(fake_posterior)))

  expect_silent(
    approx_loo_waic <- loo:::elpd_loo_approximation(
      .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      loo_approximation = "waic"
    )
  )
  expect_silent(
    approx_loo_waic_delta <- loo:::elpd_loo_approximation(
      .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      loo_approximation = "waic_grad",
      .llgrad = .llgrad
    )
  )
  expect_silent(
    approx_loo_waic_delta_diag <- loo:::elpd_loo_approximation(
      .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      loo_approximation = "waic_grad_marginal",
      .llgrad = .llgrad
    )
  )

  # Test that the approaches should not deviate too much
  diff_waic_delta <- mean(approx_loo_waic - approx_loo_waic_delta)
  diff_waic_delta_diag <- mean(approx_loo_waic - approx_loo_waic_delta_diag)
  expect_equal(approx_loo_waic, approx_loo_waic_delta_diag, tolerance = 0.1)
  expect_equal(approx_loo_waic, approx_loo_waic_delta, tolerance = 0.01)

  # Test usage in subsampling_loo
  expect_silent(
    loo_ss_waic <- loo_subsample(
      x = .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      r_eff = rep(1, nrow(fake_data)),
      loo_approximation = "waic",
      observations = 50,
      llgrad = .llgrad
    )
  )
  expect_silent(
    loo_ss_waic_delta <- loo_subsample(
      x = .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      r_eff = rep(1, nrow(fake_data)),
      loo_approximation = "waic_grad",
      observations = 50,
      llgrad = .llgrad
    )
  )
  expect_silent(
    loo_ss_waic_delta_marginal <- loo_subsample(
      x = .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      r_eff = rep(1, nrow(fake_data)),
      loo_approximation = "waic_grad_marginal",
      observations = 50,
      llgrad = .llgrad
    )
  )
  expect_silent(
    loo_ss_plpd <- loo_subsample(
      x = .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      r_eff = rep(1, nrow(fake_data)),
      loo_approximation = "plpd",
      observations = 50,
      llgrad = .llgrad
    )
  )
  expect_error(
    loo_ss_waic_delta <- loo_subsample(
      x = .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      r_eff = rep(1, nrow(fake_data)),
      loo_approximation = "waic_grad",
      observations = 50
    )
  )
})

test_that("waic using delta 2nd order method", {
  if (FALSE) {
    # Code to generate testdata - saved and loaded to avoid dependency of MCMCPack
    set.seed(123)
    N <- 100
    beta <- c(1, 2)
    X_full <- matrix(rep(1, N), ncol = 1)
    X_full <- cbind(X_full, runif(N))
    S <- 1000
    y_full <- rnorm(n = N, mean = X_full %*% beta, sd = 0.5)
    X <- X_full
    y <- y_full
    # Uncomment row below when running. Commented out to remove CHECK warnings
    # fake_posterior <- MCMCpack::MCMCregress(y~x, data = data.frame(y = y,x=X[,2]), thin = 10, mcmc = 10000) # Because Im lazy
    fake_posterior <- as.matrix(fake_posterior)
    fake_posterior[, "sigma2"] <- sqrt(fake_posterior[, "sigma2"])
    colnames(fake_posterior) <- c("a", "b", "sigma")
    fake_data <- data.frame(y, X)
    save(
      fake_posterior,
      fake_data,
      file = test_path("data-for-tests/normal_reg_waic_test_example2.rda"),
      compression_level = 9
    )
  } else {
    load(file = test_path("data-for-tests/normal_reg_waic_test_example2.rda"))
  }

  .llfun <- function(data_i, draws) {
    # data_i: ith row of fdata (data_i <- fake_data[i,, drop=FALSE])
    # draws: entire fake_posterior matrix
    dnorm(
      data_i$y,
      mean = draws[, c("a", "b")] %*% t(as.matrix(data_i[, c("X1", "X2")])),
      sd = draws[, c("sigma")],
      log = TRUE
    )
  }

  .llgrad <- function(data_i, draws) {
    sigma <- draws[, "sigma"]
    sigma2 <- sigma^2
    b <- draws[, "b"]
    a <- draws[, "a"]
    x_i <- unlist(data_i[, c("X1", "X2")])
    e <- (data_i$y - draws[, "a"] * x_i[1] - draws[, "b"] * x_i[2])

    gr <- cbind(
      e * x_i[1] / sigma2,
      e * x_i[2] / sigma2,
      -1 / sigma + e^2 / (sigma2 * sigma)
    )
    colnames(gr) <- c("a", "b", "sigma")
    gr
  }

  .llhess <- function(data_i, draws) {
    hess_array <- array(
      0,
      dim = c(ncol(draws), ncol(draws), nrow(draws)),
      dimnames = list(colnames(draws), colnames(draws), NULL)
    )
    sigma <- draws[, "sigma"]
    sigma2 <- sigma^2
    sigma3 <- sigma2 * sigma
    b <- draws[, "b"]
    a <- draws[, "a"]
    x_i <- unlist(data_i[, c("X1", "X2")])
    e <- (data_i$y - draws[, "a"] * x_i[1] - draws[, "b"] * x_i[2])

    hess_array[1, 1, ] <- -x_i[1]^2 / sigma2
    hess_array[1, 2, ] <- hess_array[2, 1, ] <- -x_i[1] * x_i[2] / sigma2
    hess_array[2, 2, ] <- -x_i[2]^2 / sigma2
    hess_array[3, 1, ] <- hess_array[1, 3, ] <- -2 * x_i[1] * e / sigma3
    hess_array[3, 2, ] <- hess_array[2, 3, ] <- -2 * x_i[2] * e / sigma3
    hess_array[3, 3, ] <- 1 / sigma2 - 3 * e^2 / (sigma2^2)
    hess_array
  }

  #data <- fake_data
  fake_posterior <- cbind(fake_posterior, runif(nrow(fake_posterior)))
  #draws <- fake_posterior <- cbind(fake_posterior, runif(nrow(fake_posterior)))

  expect_silent(
    approx_loo_waic <- loo:::elpd_loo_approximation(
      .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      loo_approximation = "waic"
    )
  )
  expect_silent(
    approx_loo_waic_delta <- loo:::elpd_loo_approximation(
      .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      loo_approximation = "waic_grad",
      .llgrad = .llgrad
    )
  )
  expect_silent(
    approx_loo_waic_delta2 <- loo:::elpd_loo_approximation(
      .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      loo_approximation = "waic_hess",
      .llgrad = .llgrad,
      .llhess = .llhess
    )
  )

  # Test that the approaches should not deviate too much
  expect_equal(approx_loo_waic, approx_loo_waic_delta2, tolerance = 0.01)
  expect_equal(approx_loo_waic, approx_loo_waic_delta, tolerance = 0.01)

  expect_silent(
    test_loo_ss_waic <- loo_subsample(
      x = .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      r_eff = rep(1, nrow(fake_data)),
      loo_approximation = "waic",
      observations = 50,
      llgrad = .llgrad
    )
  )
  expect_error(
    test_loo_ss_delta2 <- loo_subsample(
      x = .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      r_eff = rep(1, nrow(fake_data)),
      loo_approximation = "waic_hess",
      observations = 50,
      llgrad = .llgrad
    )
  )
  expect_silent(
    test_loo_ss_delta2 <- loo_subsample(
      x = .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      r_eff = rep(1, nrow(fake_data)),
      loo_approximation = "waic_hess",
      observations = 50,
      llgrad = .llgrad,
      llhess = .llhess
    )
  )
  expect_silent(
    test_loo_ss_delta <- loo_subsample(
      x = .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      r_eff = rep(1, nrow(fake_data)),
      loo_approximation = "waic_grad",
      observations = 50,
      llgrad = .llgrad
    )
  )
  expect_silent(
    test_loo_ss_point <- loo_subsample(
      x = .llfun,
      data = fake_data,
      draws = fake_posterior,
      cores = 1,
      r_eff = rep(1, nrow(fake_data)),
      loo_approximation = "plpd",
      observations = 50,
      llgrad = .llgrad
    )
  )
})


test_that("whhest works as expected", {
  N <- 100
  m <- 10
  z <- rep(1 / N, m)
  y <- 1:10
  m_i <- rep(1, m)
  expect_silent(whe <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  expect_equal(whe$y_hat_ppz, 550)
  man_var <- (sum((whe$y_hat_ppz - y / z)^2) / (m - 1)) / m
  expect_equal(whe$v_hat_y_ppz, man_var)
  z <- 1:10 / (sum(1:10) * 10)
  expect_silent(whe <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  expect_equal(whe$y_hat_ppz, 550)
  expect_equal(whe$v_hat_y_ppz, 0)

  # School book example
  # https://newonlinecourses.science.psu.edu/stat506/node/15/
  z <- c(650 / 15650, 2840 / 15650, 3200 / 15650)
  y <- c(420, 1785, 2198)
  m_i <- c(1, 1, 1)
  N <- 10
  expect_silent(whe <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  expect_equal(round(whe$y_hat_ppz, 2), 10232.75, tolerance = 0)
  expect_equal(whe$v_hat_y_ppz, 73125.74, tolerance = 0.01)
  # Double check that it is rounding error
  man_var_round <- (sum((round(y / z, 2) - 10232.75)^2)) * (1 / 2) * (1 / 3)
  expect_equal(man_var_round, 73125.74, tolerance = 0.001)
  man_var_exact <- (sum((y / z - 10232.75)^2)) * (1 / 2) * (1 / 3)
  expect_equal(whe$v_hat_y_ppz, man_var_exact, tolerance = 0.001)

  # Add test for variance estimation
  N <- 100
  m <- 10
  y <- rep(1:10, 1)
  true_var <- var(rep(y, 10)) * (99)
  z <- rep(1 / N, m)
  m_i <- rep(100000, m)
  expect_silent(whe <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  expect_equal(true_var, whe$hat_v_y_ppz, tolerance = 0.01)

  # Add tests for m_i
  N <- 100
  y <- rep(1:10, 2)
  m <- length(y)
  z <- rep(1 / N, m)
  m_i <- rep(1, m)
  expect_silent(whe1 <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  y <- rep(1:10)
  m <- length(y)
  z <- rep(1 / N, m)
  m_i <- rep(2, m)
  expect_silent(whe2 <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  expect_equal(whe1$y_hat_ppz, whe2$y_hat_ppz)
  expect_equal(whe1$v_hat_y_ppz, whe2$v_hat_y_ppz)
  expect_equal(whe1$hat_v_y_ppz, whe1$hat_v_y_ppz)
})


test_that("srs_diff_est works as expected", {
  set.seed(1234)
  N <- 1000
  y_true <- 1:N
  sigma_hat_true <- sqrt(N * sum((y_true - mean(y_true))^2) / length(y_true))
  y_approx <- rnorm(N, y_true, 0.1)
  m <- 100
  sigma_hat <- y_hat <- se_y_hat <- numeric(10000)
  for (i in 1:10000) {
    y_idx <- sample(1:N, size = m)
    y <- y_true[y_idx]
    res <- loo:::srs_diff_est(y_approx, y, y_idx)
    y_hat[i] <- res$y_hat
    se_y_hat[i] <- sqrt(res$v_y_hat)
    sigma_hat[i] <- sqrt(res$hat_v_y)
  }
  expect_equal(mean(y_hat), sum(y_true), tolerance = 0.1)

  in_ki <- y_hat + 2 * se_y_hat > sum(y_true) &
    y_hat - 2 * se_y_hat < sum(y_true)
  expect_equal(mean(in_ki), 0.95, tolerance = 0.01)

  # Should be  unbiased
  expect_equal(mean(sigma_hat), sigma_hat_true, tolerance = 0.1)

  m <- N
  y_idx <- sample(1:N, size = m)
  y <- y_true[y_idx]
  res <- loo:::srs_diff_est(y_approx, y, y_idx)
  expect_equal(res$y_hat, 500500, tolerance = 0.0001)
  expect_equal(res$v_y_hat, 0, tolerance = 0.0001)
  expect_equal(sqrt(res$hat_v_y), sigma_hat_true, tolerance = 0.1)
})

test_that("srs_est works as expected", {
  set.seed(1234)
  # Cochran 1976 example Table 2.2

  y <- c(
    rep(42, 23),
    rep(41, 4),
    36,
    32,
    29,
    27,
    27,
    23,
    19,
    16,
    16,
    15,
    15,
    14,
    11,
    10,
    9,
    7,
    6,
    6,
    6,
    5,
    5,
    4,
    3
  )
  expect_equal(sum(y), 1471)
  approx_loo <- rep(0L, 676)
  expect_equal(sum(y^2), 54497)
  res <- loo:::srs_est(y = y, approx_loo)
  expect_equal(res$y_hat, 19888, tolerance = 0.0001)
  expect_equal(res$v_y_hat, 676^2 * 229 * (1 - 0.074) / 50, tolerance = 0.0001)
  expect_equal(res$hat_v_y, 676 * var(y), tolerance = 0.0001)

  # Simulation example
  set.seed(1234)
  N <- 1000
  y_true <- 1:N
  sigma_hat_true <- sqrt(N * sum((y_true - mean(y_true))^2) / length(y_true))

  m <- 100
  y_hat <- se_y_hat <- sigma_hat <- numeric(10000)
  for (i in 1:10000) {
    y_idx <- sample(1:N, size = m)
    y <- y_true[y_idx]
    res <- loo:::srs_est(y = y, y_approx = y_true)
    y_hat[i] <- res$y_hat
    se_y_hat[i] <- sqrt(res$v_y_hat)
    sigma_hat[i] <- sqrt(res$hat_v_y)
  }
  expect_equal(mean(y_hat), sum(y_true), tolerance = 0.1)

  in_ki <- y_hat + 2 * se_y_hat > sum(y_true) &
    y_hat - 2 * se_y_hat < sum(y_true)
  expect_equal(mean(in_ki), 0.95, tolerance = 0.01)

  # Should be  unbiased
  expect_equal(mean(sigma_hat), sigma_hat_true, tolerance = 0.1)

  m <- N
  y_idx <- sample(1:N, size = m)
  y <- y_true[y_idx]
  res <- loo:::srs_est(y, y_true)
  expect_equal(res$y_hat, 500500, tolerance = 0.0001)
  expect_equal(res$v_y_hat, 0, tolerance = 0.0001)
})

options(mc.cores = 1)

test_that("Test loo_subsampling and loo_approx with radon data", {
  skip_on_cran() # avoid going over time limit for tests

  load(test_path("data-for-tests/test_radon_laplace_loo.rda"))
  # Rename to spot variable leaking errors
  llfun_test <- llfun
  log_p_test <- log_p
  log_g_test <- log_q
  draws_test <- draws
  data_test <- data
  rm(llfun, log_p, log_q, draws, data)

  set.seed(134)
  expect_silent(
    full_loo <- loo(
      llfun_test,
      draws = draws_test,
      data = data_test,
      r_eff = rep(1, nrow(data_test))
    )
  )
  expect_s3_class(full_loo, "psis_loo")

  set.seed(134)
  expect_silent(
    loo_ss <- loo_subsample(
      x = llfun_test,
      draws = draws_test,
      data = data_test,
      observations = 200,
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(data_test))
    )
  )
  expect_s3_class(loo_ss, "psis_loo_ss")

  set.seed(134)
  expect_silent(
    loo_ap_ss <- loo_subsample(
      x = llfun_test,
      draws = draws_test,
      data = data_test,
      log_p = log_p_test,
      log_g = log_g_test,
      observations = 200,
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(data_test))
    )
  )
  expect_s3_class(loo_ap_ss, "psis_loo_ss")
  expect_s3_class(loo_ap_ss, "psis_loo_ap")

  expect_silent(
    loo_ap_ss_full <- loo_subsample(
      x = llfun_test,
      log_p = log_p_test,
      log_g = log_g_test,
      draws = draws_test,
      data = data_test,
      observations = NULL,
      loo_approximation = "plpd",
      r_eff = rep(1, nrow(data_test))
    )
  )
  expect_failure(expect_s3_class(loo_ap_ss_full, "psis_loo_ss"))
  expect_s3_class(loo_ap_ss_full, "psis_loo_ap")

  # Expect similar results
  z <- 2
  expect_lte(
    loo_ss$estimates["elpd_loo", "Estimate"] -
      z * loo_ss$estimates["elpd_loo", "subsampling SE"],
    full_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["elpd_loo", "Estimate"] +
      z * loo_ss$estimates["elpd_loo", "subsampling SE"],
    full_loo$estimates["elpd_loo", "Estimate"]
  )
  expect_lte(
    loo_ss$estimates["p_loo", "Estimate"] -
      z * loo_ss$estimates["p_loo", "subsampling SE"],
    full_loo$estimates["p_loo", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["p_loo", "Estimate"] +
      z * loo_ss$estimates["p_loo", "subsampling SE"],
    full_loo$estimates["p_loo", "Estimate"]
  )
  expect_lte(
    loo_ss$estimates["looic", "Estimate"] -
      z * loo_ss$estimates["looic", "subsampling SE"],
    full_loo$estimates["looic", "Estimate"]
  )
  expect_gte(
    loo_ss$estimates["looic", "Estimate"] +
      z * loo_ss$estimates["looic", "subsampling SE"],
    full_loo$estimates["looic", "Estimate"]
  )

  expect_failure(expect_equal(
    full_loo$estimates["elpd_loo", "Estimate"],
    loo_ss$estimates["elpd_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    full_loo$estimates["p_loo", "Estimate"],
    loo_ss$estimates["p_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    full_loo$estimates["looic", "Estimate"],
    loo_ss$estimates["looic", "Estimate"],
    tolerance = 0.00000001
  ))

  z <- 2
  expect_lte(
    loo_ap_ss$estimates["elpd_loo", "Estimate"] -
      z * loo_ap_ss$estimates["elpd_loo", "subsampling SE"],
    loo_ap_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_gte(
    loo_ap_ss$estimates["elpd_loo", "Estimate"] +
      z * loo_ap_ss$estimates["elpd_loo", "subsampling SE"],
    loo_ap_ss_full$estimates["elpd_loo", "Estimate"]
  )
  expect_lte(
    loo_ap_ss$estimates["p_loo", "Estimate"] -
      z * loo_ap_ss$estimates["p_loo", "subsampling SE"],
    loo_ap_ss_full$estimates["p_loo", "Estimate"]
  )
  expect_gte(
    loo_ap_ss$estimates["p_loo", "Estimate"] +
      z * loo_ap_ss$estimates["p_loo", "subsampling SE"],
    loo_ap_ss_full$estimates["p_loo", "Estimate"]
  )
  expect_lte(
    loo_ap_ss$estimates["looic", "Estimate"] -
      z * loo_ap_ss$estimates["looic", "subsampling SE"],
    loo_ap_ss_full$estimates["looic", "Estimate"]
  )
  expect_gte(
    loo_ap_ss$estimates["looic", "Estimate"] +
      z * loo_ap_ss$estimates["looic", "subsampling SE"],
    loo_ap_ss_full$estimates["looic", "Estimate"]
  )

  expect_failure(expect_equal(
    loo_ap_ss_full$estimates["elpd_loo", "Estimate"],
    loo_ap_ss$estimates["elpd_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    loo_ap_ss_full$estimates["p_loo", "Estimate"],
    loo_ap_ss$estimates["p_loo", "Estimate"],
    tolerance = 0.00000001
  ))
  expect_failure(expect_equal(
    loo_ap_ss_full$estimates["looic", "Estimate"],
    loo_ap_ss$estimates["looic", "Estimate"],
    tolerance = 0.00000001
  ))

  # Correct printout
  expect_failure(expect_output(
    print(full_loo),
    "Posterior approximation correction used\\."
  ))
  expect_failure(expect_output(
    print(full_loo),
    "subsampled log-likelihood\nvalues"
  ))

  expect_failure(expect_output(
    print(loo_ss),
    "Posterior approximation correction used\\."
  ))
  expect_output(print(loo_ss), "subsampled log-likelihood\nvalues")

  expect_output(print(loo_ap_ss), "Posterior approximation correction used\\.")
  expect_output(print(loo_ap_ss), "subsampled log-likelihood\nvalues")

  expect_output(
    print(loo_ap_ss_full),
    "Posterior approximation correction used\\."
  )
  expect_failure(expect_output(
    print(loo_ap_ss_full),
    "subsampled log-likelihood\nvalues"
  ))

  # Test conversion of objects
  expect_silent(loo_ap_full <- loo:::as.psis_loo.psis_loo(loo_ap_ss_full))
  expect_s3_class(loo_ap_full, "psis_loo_ap")
  expect_silent(loo_ap_full_ss <- loo:::as.psis_loo_ss.psis_loo(loo_ap_full))
  expect_s3_class(loo_ap_full_ss, "psis_loo_ss")
  expect_s3_class(loo_ap_full_ss, "psis_loo_ap")
  expect_silent(loo_ap_full2 <- loo:::as.psis_loo.psis_loo_ss(loo_ap_full_ss))
  expect_s3_class(loo_ap_full2, "psis_loo_ap")
  expect_failure(expect_s3_class(loo_ap_full2, "psis_loo_ss"))
  expect_equal(loo_ap_full2, loo_ap_full)

  # Test update
  set.seed(4712)
  expect_silent(
    loo_ss2 <- update(
      loo_ss,
      draws = draws_test,
      data = data_test,
      observations = 1000,
      r_eff = rep(1, nrow(data_test))
    )
  )
  expect_gt(dim(loo_ss2)[2], dim(loo_ss)[2])
  expect_gt(dim(loo_ss2$pointwise)[1], dim(loo_ss$pointwise)[1])
  expect_equal(nobs(loo_ss), 200)
  expect_equal(nobs(loo_ss2), 1000)
  for (i in 1:nrow(loo_ss2$estimates)) {
    expect_lt(
      loo_ss2$estimates[i, "subsampling SE"],
      loo_ss$estimates[i, "subsampling SE"]
    )
  }

  set.seed(4712)
  expect_silent(
    loo_ap_ss2 <- update(
      object = loo_ap_ss,
      draws = draws_test,
      data = data_test,
      observations = 2000
    )
  )
  expect_gt(dim(loo_ap_ss2)[2], dim(loo_ap_ss)[2])
  expect_gt(dim(loo_ap_ss2$pointwise)[1], dim(loo_ap_ss$pointwise)[1])
  expect_equal(nobs(loo_ap_ss), 200)
  expect_equal(nobs(loo_ap_ss2), 2000)
  for (i in 1:nrow(loo_ap_ss2$estimates)) {
    expect_lt(
      loo_ap_ss2$estimates[i, "subsampling SE"],
      loo_ap_ss$estimates[i, "subsampling SE"]
    )
  }

  expect_equal(round(full_loo$estimates), round(loo_ap_ss_full$estimates))
  expect_failure(expect_equal(full_loo$estimates, loo_ap_ss_full$estimates))
  expect_equal(dim(full_loo), dim(loo_ap_ss_full))
  expect_s3_class(loo_ap_ss_full, "psis_loo_ap")
})


test_that("Test the vignette", {
  skip_on_cran()

  # NOTE
  # If any of these test fails, the vignette probably needs to be updated

  if (FALSE) {
    # Generate vignette test case
    library("rstan")
    stan_code <- "
    data {
      int<lower=0> N;             // number of data points
      int<lower=0> P;             // number of predictors (including intercept)
      matrix[N,P] X;              // predictors (including 1s for intercept)
      int<lower=0,upper=1> y[N];  // binary outcome
    }
    parameters {
      vector[P] beta;
    }
    model {
      beta ~ normal(0, 1);
      y ~ bernoulli_logit(X * beta);
    }
    "
    # logistic <- function(x) {1  / (1 + exp(-x))}
    # logit <- function(x) {log(x) - log(1-x)}
    llfun_logistic <- function(data_i, draws) {
      x_i <- as.matrix(data_i[,
        which(grepl(colnames(data_i), pattern = "X")),
        drop = FALSE
      ])
      y_pred <- draws %*% t(x_i)
      dbinom(x = data_i$y, size = 1, prob = 1 / (1 + exp(-y_pred)), log = TRUE)
    }

    # Prepare data
    url <- "http://stat.columbia.edu/~gelman/arm/examples/arsenic/wells.dat"
    wells <- read.table(url)
    wells$dist100 <- with(wells, dist / 100)
    X <- model.matrix(~ dist100 + arsenic, wells)
    standata <- list(y = wells$switch, X = X, N = nrow(X), P = ncol(X))

    # Fit model
    set.seed(4711)
    fit_1 <- stan(model_code = stan_code, data = standata, seed = 4711)
    print(fit_1, pars = "beta")

    parameter_draws <- extract(fit_1)$beta
    stan_df <- as.data.frame(standata)
    loo_i(1, llfun_logistic, data = stan_df, draws = parameter_draws)

    sm <- stan_model(model_code = stan_code)
    set.seed(4711)
    fit_laplace <- optimizing(sm, data = standata, draws = 2000, seed = 42)
    parameter_draws_laplace <- fit_laplace$theta_tilde
    log_p <- fit_laplace$log_p # The log density of the posterior
    log_g <- fit_laplace$log_g # The log density of the approximation

    # For comparisons
    standata$X[, "arsenic"] <- log(standata$X[, "arsenic"])
    stan_df2 <- as.data.frame(standata)
    set.seed(4711)
    fit_2 <- stan(fit = fit_1, data = standata, seed = 4711)
    parameter_draws_2 <- extract(fit_2)$beta

    save(
      llfun_logistic,
      stan_df,
      stan_df2,
      parameter_draws,
      parameter_draws_laplace,
      parameter_draws_2,
      log_p,
      log_g,
      file = test_path("data-for-tests/loo_subsample_vignette.rda"),
      compression_level = 9
    )
  } else {
    load(test_path("data-for-tests/loo_subsample_vignette.rda"))
  }

  set.seed(4711)
  expect_no_warning(
    looss_1 <- loo_subsample(
      llfun_logistic,
      draws = parameter_draws,
      data = stan_df,
      observations = 100
    )
  )
  expect_snapshot(print(looss_1))
  expect_s3_class(looss_1, c("psis_loo_ss", "psis_loo", "loo"))

  set.seed(4711)
  expect_no_warning(
    looss_1b <- update(
      looss_1,
      draws = parameter_draws,
      data = stan_df,
      observations = 200
    )
  )
  expect_snapshot(print(looss_1b))
  expect_s3_class(looss_1b, c("psis_loo_ss", "psis_loo", "loo"))

  set.seed(4711)
  expect_no_warning(
    looss_2 <- loo_subsample(
      x = llfun_logistic,
      draws = parameter_draws,
      data = stan_df,
      observations = 100,
      estimator = "hh_pps",
      loo_approximation = "lpd",
      loo_approximation_draws = 100
    )
  )
  expect_output(
    print(looss_2),
    "Computed from 4000 by 100 subsampled log-likelihood"
  )
  expect_output(print(looss_2), "values from 3020 total observations.")
  expect_output(
    print(looss_2),
    "MCSE and ESS estimates assume independent draws"
  )
  # Currently failing
  # expect_output(print(looss_2), "elpd_loo  -1968.9 15.4            0.5")
  # expect_output(print(looss_2), "p_loo         3.5  0.2            0.5")
  expect_s3_class(looss_2, c("psis_loo_ss", "psis_loo", "loo"))

  set.seed(4711)
  expect_no_warning(
    aploo_1 <- loo_approximate_posterior(
      llfun_logistic,
      draws = parameter_draws_laplace,
      data = stan_df,
      log_p = log_p,
      log_g = log_g
    )
  )
  expect_snapshot(print(aploo_1))
  expect_equal(length(pareto_k_ids(aploo_1, threshold = 0.5)), 31)
  expect_s3_class(aploo_1, c("psis_loo_ap", "psis_loo", "loo"))

  set.seed(4711)
  expect_no_warning(
    looapss_1 <- loo_subsample(
      llfun_logistic,
      draws = parameter_draws_laplace,
      data = stan_df,
      log_p = log_p,
      log_g = log_g,
      observations = 100
    )
  )
  expect_snapshot(print(looapss_1))
  expect_equal(length(pareto_k_ids(looapss_1, threshold = 0.5)), 3)

  # Loo compare
  set.seed(4711)
  expect_no_warning(
    looss_1 <- loo_subsample(
      llfun_logistic,
      draws = parameter_draws,
      data = stan_df,
      observations = 100
    )
  )
  set.seed(4712)
  expect_no_warning(
    looss_2 <- loo_subsample(
      x = llfun_logistic,
      draws = parameter_draws_2,
      data = stan_df2,
      observations = 100
    )
  )
  expect_snapshot(print(looss_2))

  expect_warning(
    comp <- loo_compare(looss_1, looss_2),
    "Different subsamples in 'model2' and 'model1'. Naive diff SE is used."
  )
  expect_snapshot(print(comp))

  set.seed(4712)
  expect_no_warning(
    looss_2_m <- loo_subsample(
      x = llfun_logistic,
      draws = parameter_draws_2,
      data = stan_df2,
      observations = looss_1
    )
  )
  expect_message(
    looss_2_m <- suppressWarnings(loo_subsample(
      x = llfun_logistic,
      draws = parameter_draws_2,
      data = stan_df2,
      observations = obs_idx(looss_1)
    )),
    "Simple random sampling with replacement assumed."
  )

  expect_silent(comp <- loo_compare(looss_1, looss_2_m))
  expect_snapshot(print(comp))

  set.seed(4712)
  expect_no_warning(
    looss_1 <- update(
      looss_1,
      draws = parameter_draws,
      data = stan_df,
      observations = 200
    )
  )
  expect_no_warning(
    looss_2_m <- update(
      looss_2_m,
      draws = parameter_draws_2,
      data = stan_df2,
      observations = looss_1
    )
  )
  expect_silent(comp2 <- loo_compare(looss_1, looss_2_m))
  expect_snapshot(print(comp2))

  expect_no_warning(
    looss_2_full <- loo(
      x = llfun_logistic,
      draws = parameter_draws_2,
      data = stan_df2
    )
  )
  expect_message(
    comp3 <- loo_compare(x = list(looss_1, looss_2_full)),
    "Estimated elpd_diff using observations included in loo calculations for all models."
  )
  expect_snapshot(print(comp3))
})

library(loo)
options(mc.cores = 1)

context("loo_subsampling")

test_that("overall loo_subampling works as expected (compared with loo) for diff_est", {
  set.seed(123)
  N <- 1000; K <- 10; S <- 1000; a0 <- 3; b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y); b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y,K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    # each time called internally within loo the arguments will be equal to:
    # data_i: ith row of fdata (fake_data[i,, drop=FALSE])
    # draws: entire fake_posterior matrix
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(true_loo <- loo(llfun_test, draws = fake_posterior, data = fake_data, r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(true_loo, "psis_loo")
  expect_silent(loo_ss <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 500, loo_approximation = "plpd", r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(loo_ss, "psis_loo_ss")

  # Check consistency
  expect_equivalent(loo_ss$pointwise[, "elpd_loo_approx"],
                    loo_ss$loo_subsampling$elpd_loo_approx[loo_ss$pointwise[, "idx"]])

  # Expect values
  z <- 2
  expect_lte(loo_ss$estimates["elpd_loo", "Estimate"] - z * loo_ss$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_gte(loo_ss$estimates["elpd_loo", "Estimate"] + z * loo_ss$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_lte(loo_ss$estimates["p_loo", "Estimate"] - z * loo_ss$estimates["p_loo", "subsampling SE"], true_loo$estimates["p_loo", "Estimate"])
  expect_gte(loo_ss$estimates["p_loo", "Estimate"] + z * loo_ss$estimates["p_loo", "subsampling SE"], true_loo$estimates["p_loo", "Estimate"])
  expect_lte(loo_ss$estimates["looic", "Estimate"] - z * loo_ss$estimates["looic", "subsampling SE"], true_loo$estimates["looic", "Estimate"])
  expect_gte(loo_ss$estimates["looic", "Estimate"] + z * loo_ss$estimates["looic", "subsampling SE"], true_loo$estimates["looic", "Estimate"])

  expect_failure(expect_equal(true_loo$estimates["elpd_loo", "Estimate"], loo_ss$estimates["elpd_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(true_loo$estimates["p_loo", "Estimate"], loo_ss$estimates["p_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(true_loo$estimates["looic", "Estimate"], loo_ss$estimates["looic", "Estimate"], tol = 0.00000001))

  # Test that observations works as expected
  expect_message(loo_ss2 <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = obs_idx(loo_ss), loo_approximation = "plpd", r_eff = rep(1, nrow(fake_data))))
  expect_equal(loo_ss2$estimates, loo_ss$estimates, tol = 0.00000001)
  expect_silent(loo_ss2 <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = loo_ss, loo_approximation = "plpd", r_eff = rep(1, nrow(fake_data))))
  expect_equal(loo_ss2$estimates, loo_ss$estimates, tol = 0.00000001)

  # Test lpd
  expect_silent(loo_ss_lpd <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 500, loo_approximation = "lpd", r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(loo_ss_lpd, "psis_loo_ss")
  z <- 2
  expect_lte(loo_ss_lpd$estimates["elpd_loo", "Estimate"] - z * loo_ss_lpd$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_gte(loo_ss_lpd$estimates["elpd_loo", "Estimate"] + z * loo_ss_lpd$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_lte(loo_ss_lpd$estimates["p_loo", "Estimate"] - z * loo_ss_lpd$estimates["p_loo", "subsampling SE"], true_loo$estimates["p_loo", "Estimate"])
  expect_gte(loo_ss_lpd$estimates["p_loo", "Estimate"] + z * loo_ss_lpd$estimates["p_loo", "subsampling SE"], true_loo$estimates["p_loo", "Estimate"])
  expect_lte(loo_ss_lpd$estimates["looic", "Estimate"] - z * loo_ss_lpd$estimates["looic", "subsampling SE"], true_loo$estimates["looic", "Estimate"])
  expect_gte(loo_ss_lpd$estimates["looic", "Estimate"] + z * loo_ss_lpd$estimates["looic", "subsampling SE"], true_loo$estimates["looic", "Estimate"])

  expect_failure(expect_equal(true_loo$estimates["elpd_loo", "Estimate"], loo_ss_lpd$estimates["elpd_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(true_loo$estimates["p_loo", "Estimate"], loo_ss_lpd$estimates["p_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(true_loo$estimates["looic", "Estimate"], loo_ss_lpd$estimates["looic", "Estimate"], tol = 0.00000001))

  expect_silent(loo_ss_lpd10 <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 500, loo_approximation = "lpd", loo_approximation_draws = 10, r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(loo_ss_lpd10, "psis_loo_ss")

  z <- 2
  expect_lte(loo_ss_lpd10$estimates["elpd_loo", "Estimate"] - z * loo_ss_lpd10$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_gte(loo_ss_lpd10$estimates["elpd_loo", "Estimate"] + z * loo_ss_lpd10$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_lte(loo_ss_lpd10$estimates["p_loo", "Estimate"] - z * loo_ss_lpd10$estimates["p_loo", "subsampling SE"], true_loo$estimates["p_loo", "Estimate"])
  expect_gte(loo_ss_lpd10$estimates["p_loo", "Estimate"] + z * loo_ss_lpd10$estimates["p_loo", "subsampling SE"], true_loo$estimates["p_loo", "Estimate"])
  expect_lte(loo_ss_lpd10$estimates["looic", "Estimate"] - z * loo_ss_lpd10$estimates["looic", "subsampling SE"], true_loo$estimates["looic", "Estimate"])
  expect_gte(loo_ss_lpd10$estimates["looic", "Estimate"] + z * loo_ss_lpd10$estimates["looic", "subsampling SE"], true_loo$estimates["looic", "Estimate"])

  expect_failure(expect_equal(true_loo$estimates["elpd_loo", "Estimate"], loo_ss_lpd10$estimates["elpd_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(true_loo$estimates["p_loo", "Estimate"], loo_ss_lpd10$estimates["p_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(true_loo$estimates["looic", "Estimate"], loo_ss_lpd10$estimates["looic", "Estimate"], tol = 0.00000001))

  # Test conversion of objects
  expect_silent(true_loo_2 <- loo:::as.psis_loo.psis_loo(true_loo))
  expect_silent(true_loo_ss <- loo:::as.psis_loo_ss.psis_loo(true_loo))
  expect_s3_class(true_loo_ss, "psis_loo_ss")
  expect_silent(true_loo_conv <- loo:::as.psis_loo.psis_loo_ss(true_loo_ss))
  expect_failure(expect_s3_class(true_loo_conv, "psis_loo_ss"))
  expect_equal(true_loo_conv, true_loo)
  expect_error(loo:::as.psis_loo.psis_loo_ss(loo_ss))

})

test_that("loo with subsampling of all observations works as ordinary loo.", {
  set.seed(123)
  N <- 1000; K <- 10; S <- 1000; a0 <- 3; b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y); b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y,K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(true_loo <- loo(llfun_test, draws = fake_posterior, data = fake_data, r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(true_loo, "psis_loo")
  expect_silent(loo_ss <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 1000, loo_approximation = "plpd", r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(loo_ss, "psis_loo_ss")
  expect_error(loo_ss <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 1001, loo_approximation = "plpd", r_eff = rep(1, nrow(fake_data))))

  expect_equal(true_loo$estimates["elpd_loo", "Estimate"], loo_ss$estimates["elpd_loo", "Estimate"], tol = 0.00000001)
  expect_equal(true_loo$estimates["p_loo", "Estimate"], loo_ss$estimates["p_loo", "Estimate"], tol = 0.00000001)
  expect_equal(true_loo$estimates["looic", "Estimate"], loo_ss$estimates["looic", "Estimate"], tol = 0.00000001)

  expect_equal(dim(true_loo),dim(loo_ss))
  expect_equal(true_loo$diagnostics, loo_ss$diagnostics)
  expect_equal(max(loo_ss$pointwise[, "m_i"]), 1)

})

test_that("overall loo_subsample works with diff_srs as expected (compared with loo)", {
  set.seed(123)
  N <- 1000; K <- 10; S <- 1000; a0 <- 3; b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y); b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y,K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(true_loo <- loo(x = llfun_test, draws = fake_posterior, data = fake_data, r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 200, loo_approximation = "plpd", estimator = "diff_srs", r_eff  = rep(1, nrow(fake_data))))
  expect_equal(true_loo$estimates[1,1], loo_ss$estimates[1,1], tol = 0.1)

})

test_that("Test the srs estimator with 'none' approximation", {
  set.seed(123)
  N <- 1000; K <- 10; S <- 1000; a0 <- 3; b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y); b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y,K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(true_loo <- loo(llfun_test, draws = fake_posterior, data = fake_data, r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(true_loo, "psis_loo")
  expect_silent(loo_ss <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 200, loo_approximation = "none", estimator = "srs", r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(loo_ss, "psis_loo_ss")
  expect_error(loo_ss <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 1100, loo_approximation = "none", estimator = "srs", r_eff = rep(1, nrow(fake_data))))

  expect_equal(length(obs_idx(loo_ss)), nobs(loo_ss))


  # Check consistency
  expect_equivalent(loo_ss$pointwise[, "elpd_loo_approx"],
                    loo_ss$loo_subsampling$elpd_loo_approx[loo_ss$pointwise[, "idx"]])

  # Expect values
  z <- 2
  expect_lte(loo_ss$estimates["elpd_loo", "Estimate"] - z * loo_ss$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_gte(loo_ss$estimates["elpd_loo", "Estimate"] + z * loo_ss$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_lte(loo_ss$estimates["p_loo", "Estimate"] - z * loo_ss$estimates["p_loo", "subsampling SE"], true_loo$estimates["p_loo", "Estimate"])
  expect_gte(loo_ss$estimates["p_loo", "Estimate"] + z * loo_ss$estimates["p_loo", "subsampling SE"], true_loo$estimates["p_loo", "Estimate"])
  expect_lte(loo_ss$estimates["looic", "Estimate"] - z * loo_ss$estimates["looic", "subsampling SE"], true_loo$estimates["looic", "Estimate"])
  expect_gte(loo_ss$estimates["looic", "Estimate"] + z * loo_ss$estimates["looic", "subsampling SE"], true_loo$estimates["looic", "Estimate"])

  expect_failure(expect_equal(true_loo$estimates["elpd_loo", "Estimate"], loo_ss$estimates["elpd_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(true_loo$estimates["p_loo", "Estimate"], loo_ss$estimates["p_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(true_loo$estimates["looic", "Estimate"], loo_ss$estimates["looic", "Estimate"], tol = 0.00000001))

})

test_that("Test the Hansen-Hurwitz estimator", {
  set.seed(123)
  N <- 1000; K <- 10; S <- 1000; a0 <- 3; b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y); b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y,K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(true_loo <- loo(llfun_test, draws = fake_posterior, data = fake_data, r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(true_loo, "psis_loo")
  expect_silent(loo_ss <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 300, loo_approximation = "plpd", estimator = "hh_pps", r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(loo_ss, "psis_loo_ss")
  expect_silent(loo_ss_max <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 1100, loo_approximation = "plpd", estimator = "hh_pps", r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(loo_ss_max, "psis_loo_ss")
  expect_silent(loo_ss_max2 <- update(loo_ss, draws = fake_posterior, data = fake_data, observations = 1100, r_eff = rep(1, nrow(fake_data))))
  expect_equal(nobs(loo_ss_max2), 1100)
  expect_gt(max(loo_ss_max2$pointwise[, "m_i"]), 1)
  expect_error(loo_ss_max2 <- update(loo_ss_max2, draws = fake_posterior, data = fake_data, observations = 300, r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss_max3 <- update(loo_ss, draws = fake_posterior, data = fake_data, observations = 1500, r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss2 <- update(loo_ss, draws = fake_posterior, data = fake_data, observations = loo_ss, r_eff = rep(1, nrow(fake_data))))
  expect_error(loo_ss2 <- update(loo_ss, draws = fake_posterior, data = fake_data, observations = loo_ss, loo_approximation = "lpd", r_eff = rep(1, nrow(fake_data))))
  expect_equal(loo_ss$estimates, loo_ss2$estimates)
  expect_equal(length(obs_idx(loo_ss_max)), length(obs_idx(loo_ss_max2)))
  expect_equal(length(obs_idx(loo_ss_max)), nobs(loo_ss_max))


  # Check consistency
  expect_equivalent(loo_ss$pointwise[, "elpd_loo_approx"],
                    loo_ss$loo_subsampling$elpd_loo_approx[loo_ss$pointwise[, "idx"]])
  # Check consistency
  expect_equivalent(loo_ss_max$pointwise[, "elpd_loo_approx"],
                    loo_ss_max$loo_subsampling$elpd_loo_approx[loo_ss_max$pointwise[, "idx"]])

  # Expect values
  z <- 2
  expect_lte(loo_ss$estimates["elpd_loo", "Estimate"] - z * loo_ss$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_gte(loo_ss$estimates["elpd_loo", "Estimate"] + z * loo_ss$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_lte(loo_ss$estimates["p_loo", "Estimate"] - z * loo_ss$estimates["p_loo", "subsampling SE"], true_loo$estimates["p_loo", "Estimate"])
  expect_gte(loo_ss$estimates["p_loo", "Estimate"] + z * loo_ss$estimates["p_loo", "subsampling SE"], true_loo$estimates["p_loo", "Estimate"])
  expect_lte(loo_ss$estimates["looic", "Estimate"] - z * loo_ss$estimates["looic", "subsampling SE"], true_loo$estimates["looic", "Estimate"])
  expect_gte(loo_ss$estimates["looic", "Estimate"] + z * loo_ss$estimates["looic", "subsampling SE"], true_loo$estimates["looic", "Estimate"])

  expect_failure(expect_equal(true_loo$estimates["elpd_loo", "Estimate"], loo_ss$estimates["elpd_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(true_loo$estimates["p_loo", "Estimate"], loo_ss$estimates["p_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(true_loo$estimates["looic", "Estimate"], loo_ss$estimates["looic", "Estimate"], tol = 0.00000001))

  expect_lte(loo_ss_max$estimates["elpd_loo", "Estimate"] - z * loo_ss_max$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])
  expect_gte(loo_ss_max$estimates["elpd_loo", "Estimate"] + z * loo_ss_max$estimates["elpd_loo", "subsampling SE"], true_loo$estimates["elpd_loo", "Estimate"])

})


test_that("update.psis_loo_ss works as expected (compared with loo)", {


  set.seed(123)
  N <- 1000; K <- 10; S <- 1000; a0 <- 3; b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y); b <- b0 + N * K - sum(y)
  fake_posterior <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y,K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(true_loo <- loo(llfun_test, draws = fake_posterior, data = fake_data, r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(true_loo, "psis_loo")
  expect_silent(loo_ss <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 500, loo_approximation = "plpd", r_eff = rep(1, nrow(fake_data))))
  expect_s3_class(loo_ss, "psis_loo_ss")

  # Check error when draws and data dimensions differ
  expect_error(loo_ss2 <- update(object = loo_ss, draws = cbind(fake_posterior, 1), data = fake_data, observations = 600, r_eff = rep(1, nrow(fake_data))))
  expect_error(loo_ss2 <- update(object = loo_ss, draws = fake_posterior, data = fake_data[-1,], observations = 600, r_eff = rep(1, nrow(fake_data))))

  # Add tests for adding observations
  expect_silent(loo_ss2 <- update(object = loo_ss, draws = fake_posterior, data = fake_data, observations = 600, r_eff = rep(1, nrow(fake_data))))
  expect_equal(dim(loo_ss2)[2] - dim(loo_ss)[2], expected = 100)
  expect_equal(dim(loo_ss2)[2], expected = dim(loo_ss2$pointwise)[1])
  expect_length(loo_ss2$diagnostics$pareto_k, 600)
  expect_length(loo_ss2$diagnostics$n_eff, 600)
  for(i in 1:nrow(loo_ss2$estimates)) {
    expect_lt(loo_ss2$estimates[i, "subsampling SE"],
              loo_ss$estimates[i, "subsampling SE"])
  }

  expect_silent(loo_ss2b <- update(object = loo_ss, draws = fake_posterior, data = fake_data))
  expect_equal(loo_ss2b$estimates, loo_ss$estimates)
  expect_equal(loo_ss2b$pointwise, loo_ss$pointwise)
  expect_equal(loo_ss2b$diagnostics$pareto_k, loo_ss$diagnostics$pareto_k)
  expect_equal(loo_ss2b$diagnostics$n_eff, loo_ss$diagnostics$n_eff)

  expect_silent(loo_ss3 <- update(object = loo_ss2, draws = fake_posterior, data = fake_data, observations = loo_ss))
  expect_equal(loo_ss3$estimates, loo_ss$estimates)
  expect_equal(loo_ss3$pointwise, loo_ss$pointwise)
  expect_equal(loo_ss3$diagnostics$pareto_k, loo_ss$diagnostics$pareto_k)
  expect_equal(loo_ss3$diagnostics$n_eff, loo_ss$diagnostics$n_eff)

  expect_silent(loo_ss4 <- update(object = loo_ss, draws = fake_posterior, data = fake_data, observations = 1000, r_eff = rep(1, nrow(fake_data))))
  expect_equal(loo_ss4$estimates[,1], true_loo$estimates[,1])
  expect_equal(loo_ss4$estimates[,2], true_loo$estimates[,2], tol = 0.001)

  expect_silent(loo_ss5 <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 1000, loo_approximation = "plpd", r_eff = rep(1, nrow(fake_data))))

  ss4_order <- order(loo_ss4$pointwise[, "idx"])
  expect_equal(loo_ss4$pointwise[ss4_order,c(1,3,4)], loo_ss5$pointwise[,c(1,3,4)])
  expect_equal(loo_ss4$diagnostics$pareto_k[ss4_order], loo_ss5$diagnostics$pareto_k)
  expect_equal(loo_ss4$diagnostics$n_eff[ss4_order], loo_ss5$diagnostics$n_eff)
  expect_equal(loo_ss4$pointwise[ss4_order,c(1,3,4)], true_loo$pointwise[,c(1,3,4)])
  expect_equal(loo_ss4$diagnostics$pareto_k[ss4_order], true_loo$diagnostics$pareto_k)
  expect_equal(loo_ss4$diagnostics$n_eff[ss4_order], true_loo$diagnostics$n_eff)

  expect_error(loo_ss_min <- update(object = loo_ss, draws = fake_posterior, data = fake_data, observations = 50, r_eff = rep(1, nrow(fake_data))))

  expect_silent(true_loo_ss <- loo:::as.psis_loo_ss.psis_loo(true_loo))
  expect_silent(loo_ss_subset0 <- update(true_loo_ss, observations = loo_ss, r_eff = rep(1, nrow(fake_data))))
  expect_true(identical(obs_idx(loo_ss_subset0), obs_idx(loo_ss)))
  expect_silent(loo_ss_subset1 <- update(object = loo_ss, observations = loo_ss, r_eff = rep(1, nrow(fake_data))))
  expect_message(loo_ss_subset2 <- update(object = loo_ss, observations = obs_idx(loo_ss)[1:10], r_eff = rep(1, nrow(fake_data))))
  expect_equal(nobs(loo_ss_subset2), 10)


  expect_silent(true_loo_ss <- loo:::as.psis_loo_ss.psis_loo(true_loo))
  set.seed(4711)
  expect_silent(loo_ss2 <- update(object = loo_ss, draws = fake_posterior, data = fake_data, observations = 600, r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss2_subset0 <- update(object = true_loo_ss, observations = loo_ss2, r_eff = rep(1, nrow(fake_data))))
  expect_true(setequal(obs_idx(loo_ss2), obs_idx(loo_ss2_subset0)))
  expect_true(identical(obs_idx(loo_ss2), obs_idx(loo_ss2_subset0)))
  expect_true(identical(loo_ss2$diagnostic, loo_ss2_subset0$diagnostic))

  # Add tests for changing approx variable
  expect_silent(loo_ss_lpd <- update(object = loo_ss, draws = fake_posterior, data = fake_data, loo_approximation = "lpd", r_eff = rep(1, nrow(fake_data))))
  expect_failure(expect_equal(loo_ss_lpd$loo_subsampling$elpd_loo_approx, loo_ss$loo_subsampling$elpd_loo_approx))
  expect_equal(dim(loo_ss_lpd)[2], dim(loo_ss)[2])
  expect_equal(dim(loo_ss_lpd)[2], dim(loo_ss_lpd$pointwise)[1])
  expect_length(loo_ss_lpd$diagnostics$pareto_k, 500)
  expect_length(loo_ss_lpd$diagnostics$n_eff, 500)
  expect_failure(expect_equal(loo_ss_lpd$estimates[1, "subsampling SE"], loo_ss$estimates[1, "subsampling SE"]))
  expect_failure(expect_equal(loo_ss_lpd$estimates[3, "subsampling SE"], loo_ss$estimates[3, "subsampling SE"]))

})







context("loo_subsampling_approximations")

geterate_test_elpd_dataset <- function() {
  N <- 10; K <- 10; S <- 1000; a0 <- 3; b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y); b <- b0 + N * K - sum(y)
  fake_posterior <- draws <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y,K)
  rm(N, K, S, a0, b0, p, y, a, b)

  list(fake_posterior = fake_posterior, fake_data = fake_data)
}

test_elpd_loo_approximation <- function(cores) {
  set.seed(123)
  test_data <- geterate_test_elpd_dataset()
  fake_posterior <- test_data$fake_posterior
  fake_data <- test_data$fake_data

  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  # Compute plpd approximation
  expect_silent(pi_vals <- loo:::elpd_loo_approximation(.llfun = llfun_test, data = fake_data, draws = fake_posterior, loo_approximation = "plpd", cores = cores))
  # Compute it manually
  point <- mean(fake_posterior)
  llik <- dbinom(fake_data$y, size = fake_data$K, prob = point, log = TRUE)
  abs_lliks <- abs(llik)
  man_elpd_loo_approximation <- abs_lliks/sum(abs_lliks)
  expect_equal(abs(pi_vals)/sum(abs(pi_vals)), man_elpd_loo_approximation, tol = 0.00001)

  # Compute lpd approximation
  expect_silent(pi_vals <- loo:::elpd_loo_approximation(.llfun = llfun_test, data = fake_data, draws = fake_posterior, loo_approximation = "lpd", cores = cores))
  # Compute it manually
  llik <- numeric(10)
  for(i in seq_along(fake_data$y)){
    llik[i] <- loo:::logMeanExp(dbinom(fake_data$y[i], size = fake_data$K, prob = fake_posterior, log = TRUE))
  }
  abs_lliks <- abs(llik)
  man_approx_loo_variable <- abs_lliks/sum(abs_lliks)
  expect_equal(abs(pi_vals)/sum(abs(pi_vals)), man_approx_loo_variable, tol = 0.00001)

  # Compute waic approximation
  expect_silent(pi_vals_waic <- loo:::elpd_loo_approximation(.llfun = llfun_test, data = fake_data, draws = fake_posterior, loo_approximation = "waic", cores = cores))
  expect_true(all(pi_vals > pi_vals_waic))
  expect_true(sum(pi_vals) - sum(pi_vals_waic) < 1)

  # Compute tis approximation
  expect_silent(pi_vals_tis <- loo:::elpd_loo_approximation(.llfun = llfun_test,
                                                            data = fake_data,
                                                            draws = fake_posterior,
                                                            loo_approximation = "tis",
                                                            loo_approximation_draws = 100,
                                                            cores = cores))
  expect_true(all(pi_vals > pi_vals_tis))
  expect_true(sum(pi_vals) - sum(pi_vals_tis) < 1)
}

test_that("elpd_loo_approximation works as expected", {
  test_elpd_loo_approximation(1)
})

test_that("elpd_loo_approximation with multiple cores", {
  test_elpd_loo_approximation(2)
})

test_that("Test loo_approximation_draws", {


  set.seed(123)
  N <- 1000; K <- 10; S <- 1000; a0 <- 3; b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y); b <- b0 + N * K - sum(y)
  fake_posterior <- draws <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y,K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(res1 <- loo:::elpd_loo_approximation(.llfun = llfun_test, data = fake_data, draws = fake_posterior, loo_approximation = "waic", loo_approximation_draws = NULL, cores = 1))
  expect_silent(res2 <- loo:::elpd_loo_approximation(.llfun = llfun_test, data = fake_data, draws = fake_posterior, loo_approximation = "waic", loo_approximation_draws = 10, cores = 1))
  expect_silent(res3 <- loo:::elpd_loo_approximation(.llfun = llfun_test, data = fake_data, draws = fake_posterior[1:10*100,], loo_approximation = "waic", loo_approximation_draws = NULL, cores = 1))
  expect_silent(res4 <- loo:::elpd_loo_approximation(.llfun = llfun_test, data = fake_data, draws = fake_posterior[1:10*100,, drop = FALSE], loo_approximation = "waic", loo_approximation_draws = NULL, cores = 1))
  expect_failure(expect_equal(res1, res3))
  expect_equal(res2, res3)

  expect_silent(loo_ss1 <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 100, loo_approximation = "plpd", r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss2 <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 100, loo_approximation = "plpd", loo_approximation_draws = 10, r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss3 <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 100, loo_approximation = "plpd", loo_approximation_draws = 31, r_eff = rep(1, nrow(fake_data))))
  expect_error(loo_ss4 <- loo_subsample(x = llfun_test, draws = fake_posterior, data = fake_data, observations = 100, loo_approximation = "plpd", loo_approximation_draws = 3100, r_eff = rep(1, nrow(fake_data))))

  expect_equal(names(loo_ss1$loo_subsampling), c("elpd_loo_approx", "loo_approximation", "loo_approximation_draws", "estimator", ".llfun", ".llgrad", ".llhess", "data_dim", "ndraws"))
  expect_null(loo_ss1$loo_subsampling$loo_approximation_draws)
  expect_equal(loo_ss2$loo_subsampling$loo_approximation_draws, 10L)
  expect_equal(loo_ss3$loo_subsampling$loo_approximation_draws, 31L)

})



test_that("waic using delta method and gradient", {


  if (FALSE){
    # Code to generate testdata - saved and loaded to avoid dependency of mvtnorm
    set.seed(123)
    N <- 400; beta <- c(1,2); X_full <- matrix(rep(1,N), ncol = 1); X_full <- cbind(X_full, runif(N)); S <- 1000
    y_full <- rnorm(n = N, mean = X_full%*%beta, sd = 1)
    X <- X_full; y <- y_full
    Lambda_0 <- diag(length(beta)); mu_0 <- c(0,0)
    b_hat <- solve(t(X)%*%X)%*%t(X)%*%y
    mu_n <- solve(t(X)%*%X)%*%(t(X)%*%X%*%b_hat + Lambda_0%*%mu_0)
    Lambda_n <- t(X)%*%X + Lambda_0
    # Uncomment row below when running. Commented out to remove CHECK warnings
    # fake_posterior <- mvtnorm::rmvnorm(n = S, mean = mu_n, sigma = solve(Lambda_n))
    colnames(fake_posterior) <- c("a", "b")
    fake_data <- data.frame(y, X)
    save(fake_posterior, fake_data, file = test_path("data-for-tests/normal_reg_waic_test_example.rda"))
  } else {
    load(file = test_path("data-for-tests/normal_reg_waic_test_example.rda"))
  }

  .llfun <- function(data_i, draws) {
    # data_i: ith row of fdata (fake_data[i,, drop=FALSE])
    # draws: entire fake_posterior matrix
    dnorm(data_i$y, mean = draws[, c("a", "b")] %*% t(as.matrix(data_i[, c("X1", "X2")])), sd = 1, log = TRUE)
  }

  .llgrad <- function(data_i, draws) {
    x_i <- data_i[, "X2"]
    gr <- cbind(data_i$y - draws[,"a"] - draws[,"b"]*x_i,
                (data_i$y - draws[,"a"] - draws[,"b"]*x_i) * x_i)
    colnames(gr) <- c("a", "b")
    gr
  }

  fake_posterior <- cbind(fake_posterior, runif(nrow(fake_posterior)))

  expect_silent(approx_loo_waic <- loo:::elpd_loo_approximation(.llfun, data = fake_data, draws = fake_posterior, cores = 1, loo_approximation = "waic"))
  expect_silent(approx_loo_waic_delta <- loo:::elpd_loo_approximation(.llfun, data = fake_data, draws = fake_posterior, cores = 1, loo_approximation = "waic_grad", .llgrad = .llgrad))
  expect_silent(approx_loo_waic_delta_diag <- loo:::elpd_loo_approximation(.llfun, data = fake_data, draws = fake_posterior, cores = 1, loo_approximation = "waic_grad_marginal", .llgrad = .llgrad))

  # Test that the approaches should not deviate too much
  diff_waic_delta <- mean(approx_loo_waic - approx_loo_waic_delta)
  diff_waic_delta_diag <- mean(approx_loo_waic - approx_loo_waic_delta_diag)
  expect_equal(approx_loo_waic,approx_loo_waic_delta_diag, tol = 0.1)
  expect_equal(approx_loo_waic,approx_loo_waic_delta, tol = 0.01)

  # Test usage in subsampling_loo
  expect_silent(loo_ss_waic <- loo_subsample(x = .llfun, data = fake_data, draws = fake_posterior, cores = 1, r_eff = rep(1, nrow(fake_data)), loo_approximation = "waic", observations = 50, llgrad = .llgrad))
  expect_silent(loo_ss_waic_delta <- loo_subsample(x = .llfun, data = fake_data, draws = fake_posterior, cores = 1, r_eff = rep(1, nrow(fake_data)), loo_approximation = "waic_grad", observations = 50, llgrad = .llgrad))
  expect_silent(loo_ss_waic_delta_marginal <- loo_subsample(x = .llfun, data = fake_data, draws = fake_posterior, cores = 1, r_eff = rep(1, nrow(fake_data)), loo_approximation = "waic_grad_marginal", observations = 50, llgrad = .llgrad))
  expect_silent(loo_ss_plpd <- loo_subsample(x = .llfun, data = fake_data, draws = fake_posterior, cores = 1, r_eff = rep(1, nrow(fake_data)), loo_approximation = "plpd", observations = 50, llgrad = .llgrad))
  expect_error(loo_ss_waic_delta <- loo_subsample(x = .llfun, data = fake_data, draws = fake_posterior, cores = 1, r_eff = rep(1, nrow(fake_data)), loo_approximation = "waic_grad", observations = 50))
})

test_that("waic using delta 2nd order method", {


  if (FALSE){
    # Code to generate testdata - saved and loaded to avoid dependency of MCMCPack
    set.seed(123)
    N <- 100; beta <- c(1,2); X_full <- matrix(rep(1,N), ncol = 1); X_full <- cbind(X_full, runif(N)); S <- 1000
    y_full <- rnorm(n = N, mean = X_full%*%beta, sd = 0.5)
    X <- X_full; y <- y_full
    # Uncomment row below when running. Commented out to remove CHECK warnings
    # fake_posterior <- MCMCpack::MCMCregress(y~x, data = data.frame(y = y,x=X[,2]), thin = 10, mcmc = 10000) # Because Im lazy
    fake_posterior <- as.matrix(fake_posterior)
    fake_posterior[,"sigma2"] <- sqrt(fake_posterior[,"sigma2"])
    colnames(fake_posterior) <- c("a", "b", "sigma")
    fake_data <- data.frame(y, X)
    save(fake_posterior, fake_data, file = test_path("data-for-tests/normal_reg_waic_test_example2.rda"), compression_level = 9)
  } else {
    load(file = test_path("data-for-tests/normal_reg_waic_test_example2.rda"))
  }

  .llfun <- function(data_i, draws) {
    # data_i: ith row of fdata (data_i <- fake_data[i,, drop=FALSE])
    # draws: entire fake_posterior matrix
    dnorm(data_i$y, mean = draws[, c("a", "b")] %*% t(as.matrix(data_i[, c("X1", "X2")])), sd = draws[, c("sigma")], log = TRUE)
  }

  .llgrad <- function(data_i, draws) {
    sigma <- draws[,"sigma"]
    sigma2 <- sigma^2
    b <- draws[,"b"]
    a <- draws[,"a"]
    x_i <- unlist(data_i[, c("X1", "X2")])
    e <- (data_i$y - draws[,"a"] * x_i[1] - draws[,"b"] * x_i[2])

    gr <- cbind(e * x_i[1] / sigma2,
                e * x_i[2] / sigma2,
                - 1 / sigma + e^2 / (sigma2 * sigma))
    colnames(gr) <- c("a", "b", "sigma")
    gr
  }

  .llhess <- function(data_i, draws) {
    hess_array <- array(0, dim = c(ncol(draws), ncol(draws), nrow(draws)), dimnames = list(colnames(draws),colnames(draws),NULL))
    sigma <- draws[,"sigma"]
    sigma2 <- sigma^2
    sigma3 <- sigma2*sigma
    b <- draws[,"b"]
    a <- draws[,"a"]
    x_i <- unlist(data_i[, c("X1", "X2")])
    e <- (data_i$y - draws[,"a"] * x_i[1] - draws[,"b"] * x_i[2])

    hess_array[1,1,] <- - x_i[1]^2 / sigma2
    hess_array[1,2,] <- hess_array[2,1,] <- - x_i[1] * x_i[2] / sigma2
    hess_array[2,2,] <- - x_i[2]^2 / sigma2
    hess_array[3,1,] <- hess_array[1,3,] <- -2 * x_i[1] * e / sigma3
    hess_array[3,2,] <- hess_array[2,3,] <- -2 * x_i[2] * e / sigma3
    hess_array[3,3,] <- 1 / sigma2 - 3 * e^2 / (sigma2^2)
    hess_array
  }

  #data <- fake_data
  fake_posterior <- cbind(fake_posterior, runif(nrow(fake_posterior)))
  #draws <- fake_posterior <- cbind(fake_posterior, runif(nrow(fake_posterior)))

  expect_silent(approx_loo_waic <- loo:::elpd_loo_approximation(.llfun, data = fake_data, draws = fake_posterior, cores = 1, loo_approximation = "waic"))
  expect_silent(approx_loo_waic_delta <- loo:::elpd_loo_approximation(.llfun, data = fake_data, draws = fake_posterior, cores = 1, loo_approximation = "waic_grad", .llgrad = .llgrad))
  expect_silent(approx_loo_waic_delta2 <- loo:::elpd_loo_approximation(.llfun, data = fake_data, draws = fake_posterior, cores = 1, loo_approximation = "waic_hess", .llgrad = .llgrad, .llhess = .llhess))

  # Test that the approaches should not deviate too much
  expect_equal(approx_loo_waic,approx_loo_waic_delta2, tol = 0.01)
  expect_equal(approx_loo_waic,approx_loo_waic_delta, tol = 0.01)

  expect_silent(test_loo_ss_waic <- loo_subsample(x = .llfun, data = fake_data, draws = fake_posterior, cores = 1, r_eff = rep(1, nrow(fake_data)), loo_approximation = "waic", observations = 50, llgrad = .llgrad))
  expect_error(test_loo_ss_delta2 <- loo_subsample(x = .llfun, data = fake_data, draws = fake_posterior, cores = 1, r_eff = rep(1, nrow(fake_data)), loo_approximation = "waic_hess", observations = 50, llgrad = .llgrad))
  expect_silent(test_loo_ss_delta2 <- loo_subsample(x = .llfun, data = fake_data, draws = fake_posterior, cores = 1, r_eff = rep(1, nrow(fake_data)), loo_approximation = "waic_hess", observations = 50, llgrad = .llgrad, llhess = .llhess))
  expect_silent(test_loo_ss_delta <- loo_subsample(x = .llfun, data = fake_data, draws = fake_posterior, cores = 1, r_eff = rep(1, nrow(fake_data)), loo_approximation = "waic_grad", observations = 50, llgrad = .llgrad))
  expect_silent(test_loo_ss_point <- loo_subsample(x = .llfun, data = fake_data, draws = fake_posterior, cores = 1, r_eff = rep(1, nrow(fake_data)), loo_approximation = "plpd", observations = 50, llgrad = .llgrad))
})







context("loo_subsampling_estimation")

test_that("whhest works as expected", {


  N <- 100
  m <- 10
  z <- rep(1/N, m)
  y <- 1:10
  m_i <- rep(1,m)
  expect_silent(whe <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  expect_equal(whe$y_hat_ppz, 550)
  man_var <- (sum((whe$y_hat_ppz - y/z)^2)/(m-1))/m
  expect_equal(whe$v_hat_y_ppz, man_var)
  z <- 1:10/(sum(1:10)*10)
  expect_silent(whe <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  expect_equal(whe$y_hat_ppz, 550)
  expect_equal(whe$v_hat_y_ppz, 0)

  # School book example
  # https://newonlinecourses.science.psu.edu/stat506/node/15/
  z <- c(650/15650, 2840/15650, 3200/15650)
  y <- c(420, 1785, 2198)
  m_i <- c(1,1,1)
  N <- 10
  expect_silent(whe <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  expect_equal(round(whe$y_hat_ppz, 2), 10232.75, tol = 0)
  expect_equal(whe$v_hat_y_ppz, 73125.74, tol = 0.01)
  # Double check that it is rounding error
  man_var_round <- (sum((round(y/z,2) - 10232.75)^2)) * (1/2) * (1/3)
  expect_equal(man_var_round, 73125.74, tol = 0.001)
  man_var_exact <- (sum((y/z - 10232.75)^2)) * (1/2) * (1/3)
  expect_equal(whe$v_hat_y_ppz, man_var_exact, tol = 0.001)

  # Add test for variance estimation
  N <- 100
  m <- 10
  y <- rep(1:10, 1)
  true_var <- var(rep(y, 10)) * (99)
  z <- rep(1/N, m)
  m_i <- rep(100000, m)
  expect_silent(whe <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  expect_equal(true_var, whe$hat_v_y_ppz, tol = 0.01)

  # Add tests for m_i
  N <- 100
  y <- rep(1:10, 2)
  m <- length(y)
  z <- rep(1/N, m)
  m_i <- rep(1,m)
  expect_silent(whe1 <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  y <- rep(1:10)
  m <- length(y)
  z <- rep(1/N, m)
  m_i <- rep(2,m)
  expect_silent(whe2 <- loo:::whhest(z = z, m_i = m_i, y = y, N = N))
  expect_equal(whe1$y_hat_ppz, whe2$y_hat_ppz)
  expect_equal(whe1$v_hat_y_ppz, whe2$v_hat_y_ppz)
  expect_equal(whe1$hat_v_y_ppz, whe1$hat_v_y_ppz)

})


test_that("srs_diff_est works as expected", {


  set.seed(1234)
  N <- 1000
  y_true <- 1:N
  sigma_hat_true <- sqrt(N * sum((y_true - mean(y_true))^2) / length(y_true))
  y_approx <- rnorm(N, y_true, 0.1)
  m <- 100
  sigma_hat <- y_hat <- se_y_hat <- numeric(10000)
  for(i in 1:10000){
    y_idx <- sample(1:N, size = m)
    y <- y_true[y_idx]
    res <- loo:::srs_diff_est(y_approx, y, y_idx)
    y_hat[i] <- res$y_hat
    se_y_hat[i] <- sqrt(res$v_y_hat)
    sigma_hat[i] <- sqrt(res$hat_v_y)
  }
  expect_equal(mean(y_hat), sum(y_true), tol = 0.1)

  in_ki <- y_hat + 2 * se_y_hat > sum(y_true) & y_hat - 2*se_y_hat < sum(y_true)
  expect_equal(mean(in_ki), 0.95, tol = 0.01)

  # Should be  unbiased
  expect_equal(mean(sigma_hat), sigma_hat_true, tol = 0.1)

  m <- N
  y_idx <- sample(1:N, size = m)
  y <- y_true[y_idx]
  res <- loo:::srs_diff_est(y_approx, y, y_idx)
  expect_equal(res$y_hat, 500500, tol = 0.0001)
  expect_equal(res$v_y_hat, 0, tol = 0.0001)
  expect_equal(sqrt(res$hat_v_y), sigma_hat_true, tol = 0.1)

})

test_that("srs_est works as expected", {


  set.seed(1234)
  # Cochran 1976 example Table 2.2

  y <- c(rep(42,23),rep(41,4), 36, 32, 29, 27, 27, 23, 19, 16, 16, 15, 15, 14, 11, 10, 9, 7, 6, 6, 6, 5, 5, 4, 3)
  expect_equal(sum(y), 1471)
  approx_loo <- rep(0L, 676)
  expect_equal(sum(y^2), 54497)
  res <- loo:::srs_est(y = y, approx_loo)
  expect_equal(res$y_hat, 19888, tol = 0.0001)
  expect_equal(res$v_y_hat, 676^2*229*(1-0.074)/50, tol = 0.0001)
  expect_equal(res$hat_v_y, 676 * var(y), tol = 0.0001)

  # Simulation example
  set.seed(1234)
  N <- 1000
  y_true <- 1:N
  sigma_hat_true <- sqrt(N * sum((y_true - mean(y_true))^2) / length(y_true))

  m <- 100
  y_hat <- se_y_hat <- sigma_hat <- numeric(10000)
  for(i in 1:10000){
    y_idx <- sample(1:N, size = m)
    y <- y_true[y_idx]
    res <- loo:::srs_est(y = y, y_approx = y_true)
    y_hat[i] <- res$y_hat
    se_y_hat[i] <- sqrt(res$v_y_hat)
    sigma_hat[i] <- sqrt(res$hat_v_y)
  }
  expect_equal(mean(y_hat), sum(y_true), tol = 0.1)

  in_ki <- y_hat + 2*se_y_hat > sum(y_true) & y_hat - 2*se_y_hat < sum(y_true)
  expect_equal(mean(in_ki), 0.95, tol = 0.01)

  # Should be  unbiased
  expect_equal(mean(sigma_hat), sigma_hat_true, tol = 0.1)

  m <- N
  y_idx <- sample(1:N, size = m)
  y <- y_true[y_idx]
  res <- loo:::srs_est(y, y_true)
  expect_equal(res$y_hat, 500500, tol = 0.0001)
  expect_equal(res$v_y_hat, 0, tol = 0.0001)

})




context("loo_subsampling cases")

test_that("Test loo_subsampling and loo_approx with radon data", {
  skip_on_cran() # avoid going over time limit for tests

  load(test_path("data-for-tests/test_radon_laplace_loo.rda"))
  # Rename to spot variable leaking errors
  llfun_test <- llfun
  log_p_test <- log_p
  log_g_test <- log_q
  draws_test <- draws
  data_test <- data
  rm(llfun, log_p,log_q, draws, data)

  set.seed(134)
  expect_silent(full_loo <- loo(llfun_test, draws = draws_test, data = data_test, r_eff = rep(1, nrow(data_test))))
  expect_s3_class(full_loo, "psis_loo")

  set.seed(134)
  expect_silent(loo_ss <- loo_subsample(x = llfun_test, draws = draws_test, data = data_test, observations = 200, loo_approximation = "plpd", r_eff = rep(1, nrow(data_test))))
  expect_s3_class(loo_ss, "psis_loo_ss")

  set.seed(134)
  expect_silent(loo_ap_ss <- loo_subsample(x = llfun_test, draws = draws_test, data = data_test, log_p = log_p_test, log_g = log_g_test, observations = 200, loo_approximation = "plpd", r_eff = rep(1, nrow(data_test))))
  expect_s3_class(loo_ap_ss, "psis_loo_ss")
  expect_s3_class(loo_ap_ss, "psis_loo_ap")

  expect_silent(loo_ap_ss_full <- loo_subsample(x = llfun_test, log_p = log_p_test, log_g = log_g_test, draws = draws_test, data = data_test, observations = NULL, loo_approximation = "plpd", r_eff = rep(1, nrow(data_test))))
  expect_failure(expect_s3_class(loo_ap_ss_full, "psis_loo_ss"))
  expect_s3_class(loo_ap_ss_full, "psis_loo_ap")

  # Expect similar results
  z <- 2
  expect_lte(loo_ss$estimates["elpd_loo", "Estimate"] - z * loo_ss$estimates["elpd_loo", "subsampling SE"], full_loo$estimates["elpd_loo", "Estimate"])
  expect_gte(loo_ss$estimates["elpd_loo", "Estimate"] + z * loo_ss$estimates["elpd_loo", "subsampling SE"], full_loo$estimates["elpd_loo", "Estimate"])
  expect_lte(loo_ss$estimates["p_loo", "Estimate"] - z * loo_ss$estimates["p_loo", "subsampling SE"], full_loo$estimates["p_loo", "Estimate"])
  expect_gte(loo_ss$estimates["p_loo", "Estimate"] + z * loo_ss$estimates["p_loo", "subsampling SE"], full_loo$estimates["p_loo", "Estimate"])
  expect_lte(loo_ss$estimates["looic", "Estimate"] - z * loo_ss$estimates["looic", "subsampling SE"], full_loo$estimates["looic", "Estimate"])
  expect_gte(loo_ss$estimates["looic", "Estimate"] + z * loo_ss$estimates["looic", "subsampling SE"], full_loo$estimates["looic", "Estimate"])

  expect_failure(expect_equal(full_loo$estimates["elpd_loo", "Estimate"], loo_ss$estimates["elpd_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(full_loo$estimates["p_loo", "Estimate"], loo_ss$estimates["p_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(full_loo$estimates["looic", "Estimate"], loo_ss$estimates["looic", "Estimate"], tol = 0.00000001))

  z <- 2
  expect_lte(loo_ap_ss$estimates["elpd_loo", "Estimate"] - z * loo_ap_ss$estimates["elpd_loo", "subsampling SE"], loo_ap_ss_full$estimates["elpd_loo", "Estimate"])
  expect_gte(loo_ap_ss$estimates["elpd_loo", "Estimate"] + z * loo_ap_ss$estimates["elpd_loo", "subsampling SE"], loo_ap_ss_full$estimates["elpd_loo", "Estimate"])
  expect_lte(loo_ap_ss$estimates["p_loo", "Estimate"] - z * loo_ap_ss$estimates["p_loo", "subsampling SE"], loo_ap_ss_full$estimates["p_loo", "Estimate"])
  expect_gte(loo_ap_ss$estimates["p_loo", "Estimate"] + z * loo_ap_ss$estimates["p_loo", "subsampling SE"], loo_ap_ss_full$estimates["p_loo", "Estimate"])
  expect_lte(loo_ap_ss$estimates["looic", "Estimate"] - z * loo_ap_ss$estimates["looic", "subsampling SE"], loo_ap_ss_full$estimates["looic", "Estimate"])
  expect_gte(loo_ap_ss$estimates["looic", "Estimate"] + z * loo_ap_ss$estimates["looic", "subsampling SE"], loo_ap_ss_full$estimates["looic", "Estimate"])

  expect_failure(expect_equal(loo_ap_ss_full$estimates["elpd_loo", "Estimate"], loo_ap_ss$estimates["elpd_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(loo_ap_ss_full$estimates["p_loo", "Estimate"], loo_ap_ss$estimates["p_loo", "Estimate"], tol = 0.00000001))
  expect_failure(expect_equal(loo_ap_ss_full$estimates["looic", "Estimate"], loo_ap_ss$estimates["looic", "Estimate"], tol = 0.00000001))

  # Correct printout
  expect_failure(expect_output(print(full_loo), "Posterior approximation correction used\\."))
  expect_failure(expect_output(print(full_loo), "subsampled log-likelihood\nvalues"))

  expect_failure(expect_output(print(loo_ss), "Posterior approximation correction used\\."))
  expect_output(print(loo_ss), "subsampled log-likelihood\nvalues")

  expect_output(print(loo_ap_ss), "Posterior approximation correction used\\.")
  expect_output(print(loo_ap_ss), "subsampled log-likelihood\nvalues")

  expect_output(print(loo_ap_ss_full), "Posterior approximation correction used\\.")
  expect_failure(expect_output(print(loo_ap_ss_full), "subsampled log-likelihood\nvalues"))

  # Test conversion of objects
  expect_silent(loo_ap_full <- loo:::as.psis_loo.psis_loo(loo_ap_ss_full))
  expect_s3_class(loo_ap_full, "psis_loo_ap")
  expect_silent(loo_ap_full_ss <- loo:::as.psis_loo_ss.psis_loo(loo_ap_full))
  expect_s3_class(loo_ap_full_ss, "psis_loo_ss")
  expect_s3_class(loo_ap_full_ss, "psis_loo_ap")
  expect_silent(loo_ap_full2 <- loo:::as.psis_loo.psis_loo_ss(loo_ap_full_ss))
  expect_s3_class(loo_ap_full2, "psis_loo_ap")
  expect_failure(expect_s3_class(loo_ap_full2, "psis_loo_ss"))
  expect_equal(loo_ap_full2,loo_ap_full)

  # Test update
  set.seed(4712)
  expect_silent(loo_ss2 <- update(loo_ss, draws = draws_test, data = data_test, observations = 1000, r_eff = rep(1, nrow(data_test))))
  expect_gt(dim(loo_ss2)[2], dim(loo_ss)[2])
  expect_gt(dim(loo_ss2$pointwise)[1], dim(loo_ss$pointwise)[1])
  expect_equal(nobs(loo_ss), 200)
  expect_equal(nobs(loo_ss2), 1000)
  for(i in 1:nrow(loo_ss2$estimates)) {
    expect_lt(loo_ss2$estimates[i, "subsampling SE"],
              loo_ss$estimates[i, "subsampling SE"])
  }

  set.seed(4712)
  expect_silent(loo_ap_ss2 <- update(object = loo_ap_ss, draws = draws_test, data = data_test, observations = 2000))
  expect_gt(dim(loo_ap_ss2)[2], dim(loo_ap_ss)[2])
  expect_gt(dim(loo_ap_ss2$pointwise)[1], dim(loo_ap_ss$pointwise)[1])
  expect_equal(nobs(loo_ap_ss), 200)
  expect_equal(nobs(loo_ap_ss2), 2000)
  for(i in 1:nrow(loo_ap_ss2$estimates)) {
    expect_lt(loo_ap_ss2$estimates[i, "subsampling SE"],
              loo_ap_ss$estimates[i, "subsampling SE"])
  }

  expect_equal(round(full_loo$estimates), round(loo_ap_ss_full$estimates))
  expect_failure(expect_equal(full_loo$estimates, loo_ap_ss_full$estimates))
  expect_equal(dim(full_loo), dim(loo_ap_ss_full))
  expect_s3_class(loo_ap_ss_full, "psis_loo_ap")

})


test_that("Test the vignette", {
  skip_on_cran()


  # NOTE
  # If any of these test fails, the vignette probably needs to be updated

  if (FALSE) {
    # Generate vignette test case
    library("rstan")
    stan_code <- "
    data {
      int<lower=0> N;             // number of data points
      int<lower=0> P;             // number of predictors (including intercept)
      matrix[N,P] X;              // predictors (including 1s for intercept)
      int<lower=0,upper=1> y[N];  // binary outcome
    }
    parameters {
      vector[P] beta;
    }
    model {
      beta ~ normal(0, 1);
      y ~ bernoulli_logit(X * beta);
    }
    "
    # logistic <- function(x) {1  / (1 + exp(-x))}
    # logit <- function(x) {log(x) - log(1-x)}
    llfun_logistic <- function(data_i, draws) {
      x_i <- as.matrix(data_i[, which(grepl(colnames(data_i), pattern = "X")), drop=FALSE])
      y_pred <- draws %*% t(x_i)
      dbinom(x = data_i$y, size = 1, prob = 1  / (1 + exp(-y_pred)), log = TRUE)
    }

    # Prepare data
    url <- "http://stat.columbia.edu/~gelman/arm/examples/arsenic/wells.dat"
    wells <- read.table(url)
    wells$dist100 <- with(wells, dist / 100)
    X <- model.matrix(~ dist100 + arsenic, wells)
    standata <- list(y = wells$switch, X = X, N = nrow(X), P = ncol(X))

    # Fit model
    set.seed(4711)
    fit_1 <- stan(model_code = stan_code, data = standata, seed = 4711)
    print(fit_1, pars = "beta")

    parameter_draws <- extract(fit_1)$beta
    stan_df <- as.data.frame(standata)
    loo_i(1, llfun_logistic, data = stan_df, draws = parameter_draws)

    sm <- stan_model(model_code = stan_code)
    set.seed(4711)
    fit_laplace <- optimizing(sm, data = standata, draws = 2000, seed = 42)
    parameter_draws_laplace <- fit_laplace$theta_tilde
    log_p <- fit_laplace$log_p # The log density of the posterior
    log_g <- fit_laplace$log_g # The log density of the approximation

    # For comparisons
    standata$X[, "arsenic"] <- log(standata$X[, "arsenic"])
    stan_df2 <- as.data.frame(standata)
    set.seed(4711)
    fit_2 <- stan(fit = fit_1, data = standata, seed = 4711)
    parameter_draws_2 <- extract(fit_2)$beta

    save(llfun_logistic,
         stan_df, stan_df2,
         parameter_draws, parameter_draws_laplace, parameter_draws_2,
         log_p, log_g,
         file = test_path("data-for-tests/loo_subsample_vignette.rda"), compression_level = 9)

  } else {
    load(test_path("data-for-tests/loo_subsample_vignette.rda"))
  }

  set.seed(4711)
  expect_no_warning(looss_1 <- loo_subsample(llfun_logistic, draws = parameter_draws, data = stan_df, observations = 100))
  expect_output(print(looss_1), "Computed from 4000 by 100 subsampled log-likelihood")
  expect_output(print(looss_1), "values from 3020 total observations.")
  expect_output(print(looss_1), "MCSE and ESS estimates assume independent draws")
  expect_output(print(looss_1), "elpd_loo  -1968.5 15.6            0.3")
  expect_output(print(looss_1), "p_loo         3.1  0.1            0.4")
  expect_s3_class(looss_1, c("psis_loo_ss", "psis_loo", "loo"))

  set.seed(4711)
  expect_no_warning(looss_1b <- update(looss_1, draws = parameter_draws, data = stan_df, observations = 200))
  expect_output(print(looss_1b), "Computed from 4000 by 200 subsampled log-likelihood")
  expect_output(print(looss_1b), "values from 3020 total observations.")
  expect_output(print(looss_1b), "MCSE and ESS estimates assume independent draws")
  expect_output(print(looss_1b), "elpd_loo  -1968.3 15.6            0.2")
  expect_output(print(looss_1b), "p_loo         3.2  0.1            0.4")
  expect_s3_class(looss_1b, c("psis_loo_ss", "psis_loo", "loo"))

  set.seed(4711)
  expect_no_warning(looss_2 <- loo_subsample(x = llfun_logistic, draws = parameter_draws, data = stan_df, observations = 100, estimator = "hh_pps", loo_approximation = "lpd", loo_approximation_draws = 100))
  expect_output(print(looss_2), "Computed from 4000 by 100 subsampled log-likelihood")
  expect_output(print(looss_2), "values from 3020 total observations.")
  expect_output(print(looss_2), "MCSE and ESS estimates assume independent draws")
  # Currently failing
  # expect_output(print(looss_2), "elpd_loo  -1968.9 15.4            0.5")
  # expect_output(print(looss_2), "p_loo         3.5  0.2            0.5")
  expect_s3_class(looss_2, c("psis_loo_ss", "psis_loo", "loo"))

  set.seed(4711)
  expect_no_warning(aploo_1 <- loo_approximate_posterior(llfun_logistic, draws = parameter_draws_laplace, data = stan_df, log_p = log_p, log_g = log_g))
  expect_output(print(aploo_1), "Computed from 2000 by 3020 log-likelihood matrix")
  expect_output(print(aploo_1), "MCSE and ESS estimates assume independent draws")
  expect_output(print(aploo_1), "elpd_loo  -1968.4 15.6")
  expect_output(print(aploo_1), "p_loo         3.2  0.2")
  expect_output(print(aploo_1), "Posterior approximation correction used.")
  expect_output(print(aploo_1), "All Pareto k estimates are good")
  expect_equal(length(pareto_k_ids(aploo_1,threshold=0.5)), 31)
  expect_s3_class(aploo_1, c("psis_loo_ap", "psis_loo", "loo"))

  set.seed(4711)
  expect_no_warning(looapss_1 <- loo_subsample(llfun_logistic, draws = parameter_draws_laplace, data = stan_df, log_p = log_p, log_g = log_g, observations = 100))
  expect_output(print(looapss_1), "Computed from 2000 by 100 subsampled log-likelihood")
  expect_output(print(looapss_1), "MCSE and ESS estimates assume independent draws")
  expect_output(print(looapss_1), "values from 3020 total observations.")
  expect_output(print(looapss_1), "elpd_loo  -1968.2 15.6            0.4")
  expect_output(print(looapss_1), "p_loo         2.9  0.1            0.5")
  expect_output(print(looapss_1), "All Pareto k estimates are good")
  expect_equal(length(pareto_k_ids(looapss_1,threshold=0.5)), 3)

  # Loo compare
  set.seed(4711)
  expect_no_warning(looss_1 <- loo_subsample(llfun_logistic, draws = parameter_draws, data = stan_df, observations = 100))
  set.seed(4712)
  expect_no_warning(looss_2 <- loo_subsample(x = llfun_logistic, draws = parameter_draws_2, data = stan_df2, observations = 100))
  expect_output(print(looss_2), "Computed from 4000 by 100 subsampled log-likelihood")
  expect_output(print(looss_2), "MCSE and ESS estimates assume independent draws")
  expect_output(print(looss_2), "values from 3020 total observations.")
  expect_output(print(looss_2), "elpd_loo  -1952.0 16.2            0.2")
  expect_output(print(looss_2), "p_loo         2.6  0.1            0.3")

  expect_warning(comp <- loo_compare(looss_1, looss_2), "Different subsamples in 'model2' and 'model1'. Naive diff SE is used.")
  expect_output(print(comp), "model1 16.5      22.5     0.4")

  set.seed(4712)
  expect_no_warning(looss_2_m <- loo_subsample(x = llfun_logistic, draws = parameter_draws_2, data = stan_df2, observations = looss_1))
  expect_message(looss_2_m <- suppressWarnings(loo_subsample(x = llfun_logistic, draws = parameter_draws_2, data = stan_df2, observations = obs_idx(looss_1))),
                 "Simple random sampling with replacement assumed.")

  expect_silent(comp <- loo_compare(looss_1, looss_2_m))
  expect_output(print(comp), "model1 16.1       4.4     0.1")

  set.seed(4712)
  expect_no_warning(looss_1 <- update(looss_1, draws = parameter_draws, data = stan_df, observations = 200))
  expect_no_warning(looss_2_m <- update(looss_2_m, draws = parameter_draws_2, data = stan_df2, observations = looss_1))
  expect_silent(comp2 <- loo_compare(looss_1, looss_2_m))
  expect_output(print(comp2), "model1 16.3       4.4     0.1")

  expect_no_warning(looss_2_full <- loo(x = llfun_logistic, draws = parameter_draws_2, data = stan_df2))
  expect_message(comp3 <- loo_compare(x = list(looss_1, looss_2_full)),
                 "Estimated elpd_diff using observations included in loo calculations for all models.")
  expect_output(print(comp3), "model1 16.5       4.4     0.3")

})


context("loo_compare_subsample")

test_that("loo_compare_subsample", {
  skip_on_cran() # to get under cran check time limit

  set.seed(123)
  N <- 1000
  x1 <- rnorm(N)
  x2 <- rnorm(N)
  x3 <- rnorm(N)
  sigma <- 2
  y <- rnorm(N, 1 + 2*x1 - 2*x2 - 1*x3, sd = sigma)
  X <- cbind("x0" = rep(1, N), x1, x2, x3)

  # Generate samples from posterior
  samples_blin <- function(X, y, sigma, draws = 1000){
    XtX <- t(X)%*%X
    b_hat <- solve(XtX) %*% (t(X) %*% y)
    Lambda_n = XtX + diag(ncol(X))
    mu_n <- solve(Lambda_n) %*% (XtX %*% b_hat + diag(ncol(X)) %*% rep(0,ncol(X)))
    L <- t(chol(sigma^2 * solve(Lambda_n)))
    draws_mat <- matrix(0, ncol=ncol(X), nrow = draws)
    for(i in 1:draws){
      z <- rnorm(length(mu_n))
      draws_mat[i,] <- L %*% z + mu_n
    }
    draws_mat
  }

  fake_posterior1 <- samples_blin(X[,1:2], y, sigma, draws = 1000)
  fake_posterior2 <- samples_blin(X[,1:3], y, sigma, draws = 1000)
  fake_posterior3 <- samples_blin(X, y, sigma, draws = 1000)

  fake_data1 <- data.frame(y, X[,1:2])
  fake_data2 <- data.frame(y, X[,1:3])
  fake_data3 <- data.frame(y, X)

  llfun_test <- function(data_i, draws) {
    dnorm(x = data_i$y, mean = draws %*% t(data_i[,-1, drop=FALSE]), sd = sigma, log = TRUE)
  }

  expect_silent(l1 <- loo(llfun_test, data = fake_data1, draws = fake_posterior1, r_eff = rep(1, N)))
  expect_silent(l2 <- loo(llfun_test, data = fake_data2, draws = fake_posterior2, r_eff = rep(1, N)))
  expect_silent(l3 <- loo(llfun_test, data = fake_data3, draws = fake_posterior3, r_eff = rep(1, N)))

  expect_silent(lss1 <- loo_subsample(llfun_test, data = fake_data1, draws = fake_posterior1, observations = 100, r_eff = rep(1, N)))
  expect_silent(lss2 <- loo_subsample(llfun_test, data = fake_data2, draws = fake_posterior2, observations = 100, r_eff = rep(1, N)))
  expect_silent(lss3 <- loo_subsample(llfun_test, data = fake_data3, draws = fake_posterior3, observations = 100, r_eff = rep(1, N)))
  expect_silent(lss2o1 <- loo_subsample(llfun_test, data = fake_data2, draws = fake_posterior2, observations = lss1, r_eff = rep(1, N)))
  expect_silent(lss3o1 <- loo_subsample(llfun_test, data = fake_data3, draws = fake_posterior3, observations = lss1, r_eff = rep(1, N)))
  expect_silent(lss2hh <- loo_subsample(llfun_test, data = fake_data2, draws = fake_posterior2, observations = 100, estimator = "hh_pps", r_eff = rep(1, N)))

  expect_warning(lcss <- loo:::loo_compare.psis_loo_ss_list(x = list(lss1, lss2, lss3)))
  expect_warning(lcss2 <- loo:::loo_compare.psis_loo_ss_list(x = list(lss1, lss2, lss3o1)))
  expect_silent(lcsso <- loo:::loo_compare.psis_loo_ss_list(x = list(lss1, lss2o1, lss3o1)))
  expect_warning(lcssohh <- loo:::loo_compare.psis_loo_ss_list(x = list(lss1, lss2hh, lss3o1)))
  expect_message(lcssf1 <- loo:::loo_compare.psis_loo_ss_list(x = list(loo:::as.psis_loo_ss.psis_loo(l1), lss2o1, lss3o1)))
  expect_message(lcssf2 <- loo:::loo_compare.psis_loo_ss_list(x = list(loo:::as.psis_loo_ss.psis_loo(l1), lss2o1, loo:::as.psis_loo_ss.psis_loo(l3))))

  expect_equal(lcss[,1], lcsso[,1], tolerance = 1)
  expect_equal(lcss2[,1], lcsso[,1], tolerance = 1)
  expect_equal(lcssohh[,1], lcsso[,1], tolerance = 1)
  expect_equal(lcssf1[,1], lcsso[,1], tolerance = 1)
  expect_equal(lcssf2[,1], lcsso[,1], tolerance = 1)

  expect_gt(lcss[,2][2], lcsso[,2][2])
  expect_gt(lcss[,2][3], lcsso[,2][3])
  expect_gt(lcss2[,2][2], lcsso[,2][2])
  expect_equal(lcss2[,2][3], lcsso[,2][3])
  expect_gt(lcssohh[,2][2], lcsso[,2][2])
  expect_equal(lcssohh[,2][3], lcsso[,2][3])

  expect_silent(lcss2m <- loo:::loo_compare.psis_loo_ss_list(x = list(lss2o1, lss3o1)))
  expect_equal(unname(lcss2m[,]), unname(lcsso[1:2,]))

  expect_warning(lcssapi <- loo_compare(lss1, lss2, lss3))
  expect_equal(lcssapi, lcss)
  expect_warning(lcssohhapi <- loo_compare(lss1, lss2hh, lss3o1))
  expect_equal(lcssohhapi, lcssohh)
  expect_silent(lcss2mapi <- loo_compare(lss2o1, lss3o1))
  expect_equal(lcss2mapi, lcss2m)

})

context("subsample with tis, sis")

test_that("Test 'tis' and 'sis'", {
  skip_on_cran()

  set.seed(123)
  N <- 1000; K <- 10; S <- 1000; a0 <- 3; b0 <- 2
  p <- 0.7
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y); b <- b0 + N * K - sum(y)
  fake_posterior <- draws <- as.matrix(rbeta(S, a, b))
  fake_data <- data.frame(y,K)
  rm(N, K, S, a0, b0, p, y, a, b)
  llfun_test <- function(data_i, draws) {
    dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
  }

  expect_silent(loo_ss_full <-
                  loo_subsample(x = llfun_test,
                                draws = fake_posterior,
                                data = fake_data,
                                observations = 1000,
                                loo_approximation = "plpd",
                                r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss_plpd <-
                  loo_subsample(x = llfun_test,
                                draws = fake_posterior,
                                data = fake_data,
                                observations = 100,
                                loo_approximation = "plpd",
                                r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss_tis_S1000 <-
                  loo_subsample(x = llfun_test,
                                draws = fake_posterior,
                                data = fake_data,
                                observations = 100,
                                loo_approximation = "tis",
                                r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss_tis_S100 <-
                  loo_subsample(x = llfun_test,
                                draws = fake_posterior,
                                data = fake_data,
                                observations = 100,
                                loo_approximation = "tis",
                                loo_approximation_draws = 100,
                                r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss_tis_S10 <-
                  loo_subsample(x = llfun_test,
                                draws = fake_posterior,
                                data = fake_data,
                                observations = 100,
                                loo_approximation = "tis",
                                loo_approximation_draws = 10,
                                r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss_sis_S1000 <-
                  loo_subsample(x = llfun_test,
                                draws = fake_posterior,
                                data = fake_data,
                                observations = 100,
                                loo_approximation = "sis",
                                r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss_sis_S100 <-
                  loo_subsample(x = llfun_test,
                                draws = fake_posterior,
                                data = fake_data,
                                observations = 100,
                                loo_approximation = "sis",
                                loo_approximation_draws = 100,
                                r_eff = rep(1, nrow(fake_data))))
  expect_silent(loo_ss_sis_S10 <-
                  loo_subsample(x = llfun_test,
                                draws = fake_posterior,
                                data = fake_data,
                                observations = 100,
                                loo_approximation = "sis",
                                loo_approximation_draws = 10,
                                r_eff = rep(1, nrow(fake_data))))


  SEs <- 4
  expect_gt(loo_ss_tis_S1000$estimates["elpd_loo", "Estimate"] + SEs*loo_ss_tis_S1000$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
  expect_lt(loo_ss_tis_S1000$estimates["elpd_loo", "Estimate"] - SEs*loo_ss_tis_S1000$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
  expect_gt(loo_ss_tis_S100$estimates["elpd_loo", "Estimate"] + SEs*loo_ss_tis_S100$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
  expect_lt(loo_ss_tis_S100$estimates["elpd_loo", "Estimate"] - SEs*loo_ss_tis_S100$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
  expect_gt(loo_ss_tis_S10$estimates["elpd_loo", "Estimate"] + SEs*loo_ss_tis_S10$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
  expect_lt(loo_ss_tis_S10$estimates["elpd_loo", "Estimate"] - SEs*loo_ss_tis_S10$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])

  expect_gt(loo_ss_sis_S1000$estimates["elpd_loo", "Estimate"] + SEs*loo_ss_sis_S1000$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
  expect_lt(loo_ss_sis_S1000$estimates["elpd_loo", "Estimate"] - SEs*loo_ss_sis_S1000$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
  expect_gt(loo_ss_sis_S100$estimates["elpd_loo", "Estimate"] + SEs*loo_ss_sis_S100$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
  expect_lt(loo_ss_sis_S100$estimates["elpd_loo", "Estimate"] - SEs*loo_ss_sis_S100$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
  expect_gt(loo_ss_sis_S10$estimates["elpd_loo", "Estimate"] + SEs*loo_ss_sis_S10$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
  expect_lt(loo_ss_sis_S10$estimates["elpd_loo", "Estimate"] - SEs*loo_ss_sis_S10$estimates["elpd_loo", "subsampling SE"], loo_ss_full$estimates["elpd_loo", "Estimate"])
})

#' Pareto smoothed importance sampling (PSIS)
#'
#' Implementation of Pareto smoothed importance sampling (PSIS), a method for
#' stabilizing importance ratios. The version of PSIS implemented here
#' corresponds to the algorithm presented in Vehtari, Simpson, Gelman, Yao,
#' and Gabry (2024).
#' For PSIS diagnostics see the [pareto-k-diagnostic] page.
#'
#' @export
#' @param log_ratios An array, matrix, or vector of importance ratios on the log
#'   scale (for PSIS-LOO these are *negative* log-likelihood values). See the
#'   **Methods (by class)** section below for a detailed description of how
#'   to specify the inputs for each method.
#' @param ... Arguments passed on to the various methods.
#' @template cores
#' @param r_eff Vector of relative effective sample size estimates containing
#'   one element per observation. The values provided should be the relative
#'   effective sample sizes of `1/exp(log_ratios)` (i.e., `1/ratios`).
#'   This is related to the relative efficiency of estimating the normalizing
#'   term in self-normalizing importance sampling. If `r_eff` is not
#'   provided then the reported PSIS effective sample sizes and Monte Carlo
#'   error estimates can be over-optimistic. If the posterior draws are (near)
#'   independent then `r_eff=1` can be used. `r_eff` has to be a scalar (same
#'   value is used for all observations) or a vector with length equal to the
#'   number of observations. The default value is 1. See the [relative_eff()]
#'   helper function for computing `r_eff`.
#'
#' @return The `psis()` methods return an object of class `"psis"`,
#'   which is a named list with the following components:
#'
#' \describe{
#'   \item{`log_weights`}{
#'     Vector or matrix of smoothed (and truncated) but *unnormalized* log
#'     weights. To get normalized weights use the
#'     [`weights()`][weights.importance_sampling] method provided for objects of
#'     class `"psis"`.
#'   }
#'  \item{`diagnostics`}{
#'    A named list containing two vectors:
#'    * `pareto_k`: Estimates of the shape parameter \eqn{k} of the
#'      generalized Pareto distribution. See the [pareto-k-diagnostic]
#'      page for details.
#'    * `n_eff`: PSIS effective sample size estimates.
#'  }
#' }
#'
#' Objects of class `"psis"` also have the following [attributes][attributes()]:
#' \describe{
#'   \item{`norm_const_log`}{
#'     Vector of precomputed values of `colLogSumExps(log_weights)` that are
#'     used internally by the `weights` method to normalize the log weights.
#'   }
#'   \item{`tail_len`}{
#'     Vector of tail lengths used for fitting the generalized Pareto distribution.
#'   }
#'   \item{`r_eff`}{
#'     If specified, the user's `r_eff` argument.
#'   }
#'   \item{`dims`}{
#'     Integer vector of length 2 containing `S` (posterior sample size)
#'     and `N` (number of observations).
#'   }
#'   \item{`method`}{
#'     Method used for importance sampling, here `psis`.
#'   }
#' }
#'
#' @seealso
#' * [loo()] for approximate LOO-CV using PSIS.
#' * [pareto-k-diagnostic] for PSIS diagnostics.
#' * The __loo__ package [vignettes](https://mc-stan.org/loo/articles/index.html)
#'   for demonstrations.
#' * The [FAQ page](https://mc-stan.org/loo/articles/online-only/faq.html) on
#'   the __loo__ website for answers to frequently asked questions.
#'
#' @template loo-and-psis-references
#'
#' @examples
#' log_ratios <- -1 * example_loglik_array()
#' r_eff <- relative_eff(exp(-log_ratios))
#' psis_result <- psis(log_ratios, r_eff = r_eff)
#' str(psis_result)
#' plot(psis_result)
#'
#' # extract smoothed weights
#' lw <- weights(psis_result) # default args are log=TRUE, normalize=TRUE
#' ulw <- weights(psis_result, normalize=FALSE) # unnormalized log-weights
#'
#' w <- weights(psis_result, log=FALSE) # normalized weights (not log-weights)
#' uw <- weights(psis_result, log=FALSE, normalize = FALSE) # unnormalized weights
#'
#'
#'
psis <- function(log_ratios, ...) UseMethod("psis")

#' @export
#' @templateVar fn psis
#' @template array
#'
psis.array <-
  function(log_ratios, ..., r_eff = 1, cores = getOption("mc.cores", 1)) {
    importance_sampling.array(
      log_ratios = log_ratios,
      ...,
      r_eff = r_eff,
      cores = cores,
      method = "psis"
    )
  }


#' @export
#' @templateVar fn psis
#' @template matrix
#'
psis.matrix <-
  function(log_ratios, ..., r_eff = 1, cores = getOption("mc.cores", 1)) {
    importance_sampling.matrix(
      log_ratios,
      ...,
      r_eff = r_eff,
      cores = cores,
      method = "psis"
    )
  }

#' @export
#' @templateVar fn psis
#' @template vector
#'
psis.default <-
  function(log_ratios, ..., r_eff = 1) {
    importance_sampling.default(
      log_ratios = log_ratios,
      ...,
      r_eff = r_eff,
      method = "psis"
    )
  }


#' @rdname psis
#' @export
#' @param x For `is.psis()`, an object to check.
is.psis <- function(x) {
  inherits(x, "psis") && is.list(x)
}


# internal ----------------------------------------------------------------

#' @noRd
#' @seealso importance_sampling_object
psis_object <-
  function(unnormalized_log_weights, pareto_k, tail_len, r_eff) {
    importance_sampling_object(
      unnormalized_log_weights = unnormalized_log_weights,
      pareto_k = pareto_k,
      tail_len = tail_len,
      r_eff = r_eff,
      method = "psis"
    )
  }


#' @noRd
#' @seealso do_importance_sampling
do_psis <- function(log_ratios, r_eff, cores, method) {
  do_importance_sampling(
    log_ratios = log_ratios,
    r_eff = r_eff,
    cores = cores,
    method = "psis"
  )
}

#' Extract named components from each list in the list of lists obtained by
#' parallelizing `do_psis_i()`
#'
#' @noRd
#' @param x List of lists.
#' @param item String naming the component or attribute to pull out of each list
#'   (or list-like object).
#' @param fun,fun.val passed to `vapply()`'s `FUN` and `FUN.VALUE` arguments.
#' @return Numeric vector or matrix.
#'
psis_apply <- function(x, item, fun = c("[[", "attr"), fun_val = numeric(1)) {
  if (!is.list(x)) {
    stop("Internal error ('x' must be a list for psis_apply)")
  }
  vapply(x, FUN = match.arg(fun), FUN.VALUE = fun_val, item)
}

#' PSIS on a single vector
#'
#' @noRd
#' @param log_ratios_i A vector of log importance ratios (for `loo()`, negative
#'   log likelihoods).
#' @param tail_len_i An integer tail length.
#' @param ... Not used. Included to conform to API for differen IS methods.
#'
#' @details
#' * If there are enough tail samples then the tail is smoothed with PSIS
#' * The log weights (or log ratios if no smoothing) larger than the largest raw
#'   ratio are set to the largest raw ratio
#'
#' @return A named list containing:
#' * `lw`: vector of unnormalized log weights
#' * `pareto_k`: scalar Pareto k estimate.
#'
do_psis_i <- function(log_ratios_i, tail_len_i, ...) {
  S <- length(log_ratios_i)
  # shift log ratios for safer exponentation
  lw_i <- log_ratios_i - max(log_ratios_i)
  khat <- Inf

  if (enough_tail_samples(tail_len_i)) {
    ord <- sort.int(lw_i, index.return = TRUE)
    tail_ids <- seq(S - tail_len_i + 1, S)
    lw_tail <- ord$x[tail_ids]
    if (abs(max(lw_tail) - min(lw_tail)) < .Machine$double.eps / 100) {
      warning(
        "Can't fit generalized Pareto distribution ",
        "because all tail values are the same.",
        call. = FALSE
      )
    } else {
      cutoff <- ord$x[min(tail_ids) - 1] # largest value smaller than tail values
      smoothed <- psis_smooth_tail(lw_tail, cutoff)
      khat <- smoothed$k
      lw_i[ord$ix[tail_ids]] <- smoothed$tail
    }
  }

  # truncate at max of raw wts (i.e., 0 since max has been subtracted)
  lw_i[lw_i > 0] <- 0
  # shift log weights back so that the smallest log weights remain unchanged
  lw_i <- lw_i + max(log_ratios_i)

  list(log_weights = lw_i, pareto_k = khat)
}

#' PSIS tail smoothing for a single vector
#'
#' @noRd
#' @param x Vector of tail elements already sorted in ascending order.
#' @return A named list containing:
#' * `tail`: vector same size as `x` containing the logs of the
#'   order statistics of the generalized pareto distribution.
#' * `k`: scalar shape parameter estimate.
#'
psis_smooth_tail <- function(x, cutoff) {
  len <- length(x)
  exp_cutoff <- exp(cutoff)

  # save time not sorting since x already sorted
  fit <- posterior::gpdfit(exp(x) - exp_cutoff, sort_x = FALSE)
  k <- fit$k
  sigma <- fit$sigma
  if (is.finite(k)) {
    p <- (seq_len(len) - 0.5) / len
    qq <- posterior::qgeneralized_pareto(p, 0, sigma, k) + exp_cutoff
    tail <- log(qq)
  } else {
    tail <- x
  }
  list(tail = tail, k = k)
}


#' Calculate tail lengths to use for fitting the GPD
#'
#' The number of weights (i.e., tail length) used to fit the generalized Pareto
#' distribution is now decreasing with the number of posterior draws S, and is
#' also adjusted based on the relative MCMC neff for `exp(log_lik)`. This will
#' answer the questions about the asymptotic properties, works better for thick
#' tailed proposal distributions, and is adjusted based on dependent Markov chain
#' samples. Specifically, the tail length is now `3*sqrt(S)/r_eff` but capped at
#' 20% of the total number of weights.
#'
#' @noRd
#' @param r_eff A N-vector or scalar of relative MCMC effective sample sizes of
#'   `exp(log-lik matrix)`. The default value is 1.
#' @param S The (integer) size of posterior sample.
#' @return An N-vector of tail lengths.
#'
n_pareto <- function(r_eff, S) {
  if (isTRUE(is.null(r_eff) || all(is.na(r_eff)))) {
    r_eff <- 1
  }
  ceiling(pmin(0.2 * S, 3 * sqrt(S / r_eff)))
}

#' Check for enough tail samples to fit GPD
#'
#' @noRd
#' @param tail_len Integer tail length.
#' @param min_len The minimum allowed tail length.
#' @return `TRUE` or `FALSE`
#'
enough_tail_samples <- function(tail_len, min_len = 5) {
  tail_len >= min_len
}


#' Throw warnings about Pareto k estimates
#'
#' @noRd
#' @param k A vector of Pareto k estimates.
#' @param k_threshold The value at which to warn about high Pareto k estimates.
#' @return Nothing, just possibly throws warnings.
#'
throw_pareto_warnings <- function(k, k_threshold) {
  if (isTRUE(any(k > k_threshold))) {
    .warn("Some Pareto k diagnostic values are too high. ", .k_help())
  }
}

#' Warn if not enough tail samples to fit GPD
#'
#' @noRd
#' @param tail_lengths Vector of tail lengths.
#' @return `tail_lengths`, invisibly.
#'
throw_tail_length_warnings <- function(tail_lengths) {
  tail_len_bad <- !sapply(tail_lengths, enough_tail_samples)
  if (any(tail_len_bad)) {
    if (length(tail_lengths) == 1) {
      warning(
        "Not enough tail samples to fit the generalized Pareto distribution.",
        call. = FALSE,
        immediate. = TRUE
      )
    } else {
      bad <- which(tail_len_bad)
      Nbad <- length(bad)
      warning(
        "Not enough tail samples to fit the generalized Pareto distribution ",
        "in some or all columns of matrix of log importance ratios. ",
        "Skipping the following columns: ",
        paste(if (Nbad <= 10) bad else bad[1:10], collapse = ", "),
        if (Nbad > 10) {
          paste0(", ... [", Nbad - 10, " more not printed].\n")
        } else {
          "\n"
        },
        call. = FALSE,
        immediate. = TRUE
      )
    }
  }
  invisible(tail_lengths)
}

#' Prepare `r_eff` to pass to `psis()` and throw warnings/errors if necessary
#'
#' @noRd
#' @param r_eff User's `r_eff` argument.
#' @param len The length `r_eff` should have if not `NULL` or `NA`.
#' @return
#' * If `r_eff` has length `len` then `r_eff` is returned.
#' * If `r_eff` is `NULL` then `rep(1, len)` is returned.
#' * If `r_eff` is `NA` then `rep(1, len)` is returned.
#' * If `r_eff` is a scalar then `rep(r_eff, len)` is returned.
#' * If `r_eff` is not a scalar but the length is not `len` then an error is thrown.
#' * If `r_eff` has length `len` but has `NA`s then `NA`s are filled in with `1`s.
#'
prepare_psis_r_eff <- function(r_eff, len) {
  if (isTRUE(is.null(r_eff) || all(is.na(r_eff)))) {
    r_eff <- rep(1, len)
  } else if (length(r_eff) == 1) {
    r_eff <- rep(r_eff, len)
  } else if (length(r_eff) != len) {
    stop(
      "'r_eff' must have one value or one value per observation.",
      call. = FALSE
    )
  } else if (anyNA(r_eff)) {
    message("Replacing NAs in `r_eff` with 1s")
    r_eff[is.na(r_eff)] <- 1
  }
  r_eff
}

#' Check if `psis()` was called from one of the loo methods
#'
#' @noRd
#' @return `TRUE` if the `loo()` array, matrix, or function method is found in
#'   the active call list, `FALSE` otherwise.
#'
called_from_loo <- function() {
  calls <- sys.calls()
  txt <- unlist(lapply(calls, deparse))
  patts <- "loo.array\\(|loo.matrix\\(|loo.function\\("
  check <- sapply(txt, function(x) grepl(patts, x))
  isTRUE(any(check))
}

#' Warning message about missing `r_eff` argument
#' @noRd
throw_psis_r_eff_warning <- function() {
  warning(
    "Relative effective sample sizes ('r_eff' argument) not specified. ",
    "PSIS ESS (n_eff) will not be adjusted based on MCMC ESS (n_eff).",
    call. = FALSE
  )
}

options(mc.cores = 1)
options(loo.cores = NULL)
set.seed(123)

LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()
LLvec <- LLmat[, 1]
chain_id <- rep(1:2, each = dim(LLarr)[1])
r_eff_arr <- relative_eff(exp(LLarr))
r_eff_vec <- relative_eff(exp(LLvec), chain_id = chain_id)
psis1 <- psis(log_ratios = -LLarr, r_eff = r_eff_arr)

test_that("psis results haven't changed", {
  expect_snapshot_value(psis1, style = "serialize")
})

test_that("psis returns object with correct structure", {
  expect_true(is.psis(psis1))
  expect_false(is.loo(psis1))
  expect_false(is.psis_loo(psis1))

  expect_named(psis1, c("log_weights", "diagnostics"))
  expect_named(psis1$diagnostics, c("pareto_k", "n_eff", "r_eff"))
  expect_equal(dim(psis1), dim(LLmat))
  expect_length(psis1$diagnostics$pareto_k, dim(psis1)[2])
  expect_length(psis1$diagnostics$n_eff, dim(psis1)[2])
})


test_that("psis methods give same results", {
  psis2 <- suppressWarnings(psis(-LLmat, r_eff = r_eff_arr))
  expect_identical(psis1, psis2)

  psisvec <- suppressWarnings(psis(-LLvec, r_eff = r_eff_vec))
  psismat <- suppressWarnings(psis(-LLmat[, 1], r_eff = r_eff_vec))
  expect_identical(psisvec, psismat)
})

test_that("psis throws correct errors and warnings", {
  # r_eff default no warnings
  expect_no_warning(psis(-LLarr))
  expect_no_warning(psis(-LLmat))
  expect_no_warning(psis(-LLmat[, 1]))

  # r_eff=NULL no warnings
  expect_silent(psis(-LLarr, r_eff = NULL))
  expect_silent(psis(-LLmat, r_eff = NULL))
  expect_silent(psis(-LLmat[, 1], r_eff = NULL))

  # r_eff=NA disables warnings
  expect_silent(psis(-LLarr, r_eff = NA))
  expect_silent(psis(-LLmat, r_eff = NA))
  expect_silent(psis(-LLmat[, 1], r_eff = NA))

  # r_eff default and r_eff=NA give same answer
  expect_equal(
    suppressWarnings(psis(-LLarr)),
    psis(-LLarr, r_eff = NA)
  )

  # r_eff=NULL and r_eff=NA give same answer
  expect_equal(
    suppressWarnings(psis(-LLarr, r_eff = NULL)),
    psis(-LLarr, r_eff = NA)
  )

  # r_eff scalar is fine
  expect_silent(psis(-LLarr, r_eff = r_eff_arr[1]))

  # r_eff non-scalar wrong length is error
  expect_error(psis(-LLarr, r_eff = r_eff_arr[-1]), "one value per observation")

  # r_eff has some NA values which are replaced with 1
  r_eff_arr[2] <- NA
  expect_snapshot(psis(-LLarr, r_eff = r_eff_arr))

  # tail length warnings
  expect_snapshot(psis(-LLarr[1:5, , ]))

  # no NAs or non-finite values allowed
  LLmat[1, 1] <- NA
  expect_error(psis(-LLmat), "NAs not allowed in input")

  LLmat[1, 1] <- 1
  LLmat[10, 2] <- -Inf
  expect_error(psis(-LLmat), "All input values must be finite or -Inf")
  # log ratio of -Inf is allowed
  LLmat[10, 2] <- Inf
  expect_no_error(psis(-LLmat))

  # no lists allowed
  expect_error(
    expect_warning(psis(as.list(-LLvec))),
    "List not allowed as input"
  )

  # if array, must be 3-D array
  dim(LLarr) <- c(2, 250, 2, 32)
  expect_error(
    psis(-LLarr),
    "length(dim(log_ratios)) == 3 is not TRUE",
    fixed = TRUE
  )
})

test_that("throw_tail_length_warnings gives correct output", {
  expect_silent(throw_tail_length_warnings(10))
  expect_equal(throw_tail_length_warnings(10), 10)
  expect_warning(throw_tail_length_warnings(1), "Not enough tail samples")
  expect_warning(
    throw_tail_length_warnings(c(1, 10, 2)),
    "Skipping the following columns: 1, 3"
  )
  expect_warning(throw_tail_length_warnings(rep(1, 21)), "11 more not printed")
})


test_that("weights method returns correct output", {
  # default arguments
  expect_identical(weights(psis1), weights(psis1, normalize = TRUE, log = TRUE))

  # unnormalized log-weights same as in psis object
  expect_equal(psis1$log_weights, weights(psis1, normalize = FALSE))

  # normalized weights sum to 1
  expect_equal(
    colSums(weights(psis1, normalize = TRUE, log = FALSE)),
    rep(1, ncol(psis1$log_weights))
  )
})


test_that("psis_n_eff methods works properly", {
  w <- weights(psis1, normalize = TRUE, log = FALSE)
  expect_equal(psis_n_eff.default(w[, 1], r_eff = 1), 1 / sum(w[, 1]^2))
  expect_equal(psis_n_eff.default(w[, 1], r_eff = 2), 2 / sum(w[, 1]^2))
  expect_equal(
    psis_n_eff.default(w[, 1], r_eff = 2),
    psis_n_eff.matrix(w, r_eff = rep(2, ncol(w)))[1]
  )
  expect_no_warning(psis_n_eff.default(w[, 1]))
  expect_no_warning(psis_n_eff.matrix(w))
})


test_that("do_psis_i throws warning if all tail values the same", {
  xx <- c(1, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4)
  expect_warning(
    val <- do_psis_i(xx, tail_len_i = 6),
    "all tail values are the same"
  )
  expect_equal(val$pareto_k, Inf)
})

test_that("psis_smooth_tail returns original tail values if k is infinite", {
  xx <- log(c(2, 2, 2, 2, 3, 4, 5, 6))
  val <- suppressWarnings(psis_smooth_tail(xx, 0))
  expect_equal(val$tail, xx)
  expect_equal(val$k, Inf)
})

options(mc.cores = 1)
options(loo.cores = NULL)
set.seed(123)

LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()
LLvec <- LLmat[, 1]
chain_id <- rep(1:2, each = dim(LLarr)[1])
r_eff_arr <- relative_eff(exp(LLarr))
r_eff_vec <- relative_eff(exp(LLvec), chain_id = chain_id)
psis1 <- psis(log_ratios = -LLarr, r_eff = r_eff_arr)

test_that("psis results haven't changed", {
  expect_snapshot_value(psis1, style = "serialize")
})

test_that("psis returns object with correct structure", {
  expect_true(is.psis(psis1))
  expect_false(is.loo(psis1))
  expect_false(is.psis_loo(psis1))

  expect_named(psis1, c("log_weights", "diagnostics"))
  expect_named(psis1$diagnostics, c("pareto_k", "n_eff", "r_eff"))
  expect_equal(dim(psis1), dim(LLmat))
  expect_length(psis1$diagnostics$pareto_k, dim(psis1)[2])
  expect_length(psis1$diagnostics$n_eff, dim(psis1)[2])
})


test_that("psis methods give same results", {
  psis2 <- suppressWarnings(psis(-LLmat, r_eff = r_eff_arr))
  expect_identical(psis1, psis2)

  psisvec <- suppressWarnings(psis(-LLvec, r_eff = r_eff_vec))
  psismat <- suppressWarnings(psis(-LLmat[, 1], r_eff = r_eff_vec))
  expect_identical(psisvec, psismat)
})

test_that("psis throws correct errors and warnings", {
  # r_eff default no warnings
  expect_no_warning(psis(-LLarr))
  expect_no_warning(psis(-LLmat))
  expect_no_warning(psis(-LLmat[, 1]))

  # r_eff=NULL no warnings
  expect_silent(psis(-LLarr, r_eff = NULL))
  expect_silent(psis(-LLmat, r_eff = NULL))
  expect_silent(psis(-LLmat[, 1], r_eff = NULL))

  # r_eff=NA disables warnings
  expect_silent(psis(-LLarr, r_eff = NA))
  expect_silent(psis(-LLmat, r_eff = NA))
  expect_silent(psis(-LLmat[, 1], r_eff = NA))

  # r_eff default and r_eff=NA give same answer
  expect_equal(
    suppressWarnings(psis(-LLarr)),
    psis(-LLarr, r_eff = NA)
  )

  # r_eff=NULL and r_eff=NA give same answer
  expect_equal(
    suppressWarnings(psis(-LLarr, r_eff = NULL)),
    psis(-LLarr, r_eff = NA)
  )

  # r_eff scalar is fine
  expect_silent(psis(-LLarr, r_eff = r_eff_arr[1]))

  # r_eff non-scalar wrong length is error
  expect_error(psis(-LLarr, r_eff = r_eff_arr[-1]), "one value per observation")

  # r_eff has some NA values which are replaced with 1
  r_eff_arr[2] <- NA
  expect_snapshot(psis(-LLarr, r_eff = r_eff_arr))

  # tail length warnings
  expect_snapshot(psis(-LLarr[1:5, , ]))

  # no NAs or non-finite values allowed
  LLmat[1, 1] <- NA
  expect_error(psis(-LLmat), "NAs not allowed in input")

  LLmat[1, 1] <- 1
  LLmat[10, 2] <- -Inf
  expect_error(psis(-LLmat), "All input values must be finite or -Inf")
  # log ratio of -Inf is allowed
  LLmat[10, 2] <- Inf
  expect_no_error(psis(-LLmat))

  # no lists allowed
  expect_error(
    expect_warning(psis(as.list(-LLvec))),
    "List not allowed as input"
  )

  # if array, must be 3-D array
  dim(LLarr) <- c(2, 250, 2, 32)
  expect_error(
    psis(-LLarr),
    "length(dim(log_ratios)) == 3 is not TRUE",
    fixed = TRUE
  )
})

test_that("throw_tail_length_warnings gives correct output", {
  expect_silent(throw_tail_length_warnings(10))
  expect_equal(throw_tail_length_warnings(10), 10)
  expect_warning(throw_tail_length_warnings(1), "Not enough tail samples")
  expect_warning(
    throw_tail_length_warnings(c(1, 10, 2)),
    "Skipping the following columns: 1, 3"
  )
  expect_warning(throw_tail_length_warnings(rep(1, 21)), "11 more not printed")
})


test_that("weights method returns correct output", {
  # default arguments
  expect_identical(weights(psis1), weights(psis1, normalize = TRUE, log = TRUE))

  # unnormalized log-weights same as in psis object
  expect_equal(psis1$log_weights, weights(psis1, normalize = FALSE))

  # normalized weights sum to 1
  expect_equal(
    colSums(weights(psis1, normalize = TRUE, log = FALSE)),
    rep(1, ncol(psis1$log_weights))
  )
})


test_that("psis_n_eff methods works properly", {
  w <- weights(psis1, normalize = TRUE, log = FALSE)
  expect_equal(psis_n_eff.default(w[, 1], r_eff = 1), 1 / sum(w[, 1]^2))
  expect_equal(psis_n_eff.default(w[, 1], r_eff = 2), 2 / sum(w[, 1]^2))
  expect_equal(
    psis_n_eff.default(w[, 1], r_eff = 2),
    psis_n_eff.matrix(w, r_eff = rep(2, ncol(w)))[1]
  )
  expect_no_warning(psis_n_eff.default(w[, 1]))
  expect_no_warning(psis_n_eff.matrix(w))
})


test_that("do_psis_i throws warning if all tail values the same", {
  xx <- c(1, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4)
  expect_warning(
    val <- do_psis_i(xx, tail_len_i = 6),
    "all tail values are the same"
  )
  expect_equal(val$pareto_k, Inf)
})

test_that("psis_smooth_tail returns original tail values if k is infinite", {
  # skip on M1 Mac until we figure out why this test fails only on M1 Mac
  skip_if(Sys.info()[["sysname"]] == "Darwin" && R.version$arch == "aarch64")

  xx <- c(1, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4)
  val <- suppressWarnings(psis_smooth_tail(xx, 3))
  expect_equal(val$tail, xx)
  expect_equal(val$k, Inf)
})

#' Estimate leave-one-out predictive performance..
#'
#' The `loo_predictive_metric()` function computes estimates of leave-one-out
#' predictive metrics given a set of predictions and observations. Currently
#' supported metrics are mean absolute error, mean squared error and root mean
#' squared error for continuous predictions and accuracy and balanced accuracy
#' for binary classification. Predictions are passed on to the [E_loo()]
#' function, so this function assumes that the PSIS approximation is working
#' well.
#'
#' @param x A numeric matrix of predictions.
#' @param y A numeric vector of observations. Length should be equal to the
#'     number of rows in `x`.
#' @param log_lik A matrix of pointwise log-likelihoods. Should be of same
#'     dimension as `x`.
#' @param metric The type of predictive metric to be used. Currently
#'     supported options are `"mae"`, `"rmse"` and `"mse"` for regression and
#'     for binary classification `"acc"` and `"balanced_acc"`.
#'     \describe{
#'       \item{`"mae"`}{
#'          Mean absolute error.
#'       }
#'       \item{`"mse"`}{
#'          Mean squared error.
#'       }
#'       \item{`"rmse"`}{
#'          Root mean squared error, given by as the square root of `MSE`.
#'       }
#'       \item{`"acc"`}{
#'          The proportion of predictions indicating the correct outcome.
#'       }
#'       \item{`"balanced_acc"`}{
#'          Balanced accuracy is given by the average of true positive and true
#'          negative rates.
#'       }
#'     }
#' @param r_eff A Vector of relative effective sample size estimates containing
#'     one element per observation. See [psis()] for more details.
#' @param cores The number of cores to use for parallelization of `[psis()]`.
#'   See [psis()] for details.
#' @param ... Additional arguments passed on to [E_loo()]
#'
#' @return A list with the following components:
#' \describe{
#'   \item{`estimate`}{
#'   Estimate of the given metric.
#'   }
#'  \item{`se`}{
#'   Standard error of the estimate.
#'   }
#'  }
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("rstanarm", quietly = TRUE)) {
#' # Use rstanarm package to quickly fit a model and get both a log-likelihood
#' # matrix and draws from the posterior predictive distribution
#' library("rstanarm")
#'
#' # data from help("lm")
#' ctl <- c(4.17,5.58,5.18,6.11,4.50,4.61,5.17,4.53,5.33,5.14)
#' trt <- c(4.81,4.17,4.41,3.59,5.87,3.83,6.03,4.89,4.32,4.69)
#' d <- data.frame(
#'   weight = c(ctl, trt),
#'   group = gl(2, 10, 20, labels = c("Ctl","Trt"))
#' )
#' fit <- stan_glm(weight ~ group, data = d, refresh = 0)
#' ll <- log_lik(fit)
#' r_eff <- relative_eff(exp(-ll), chain_id = rep(1:4, each = 1000))
#'
#' mu_pred <- posterior_epred(fit)
#' # Leave-one-out mean absolute error of predictions
#' mae <- loo_predictive_metric(x = mu_pred, y = d$weight, log_lik = ll,
#'                             pred_error = 'mae', r_eff = r_eff)
#' # Leave-one-out 90%-quantile of mean absolute error
#' mae_90q <- loo_predictive_metric(x = mu_pred, y = d$weight, log_lik = ll,
#'                                 pred_error = 'mae', r_eff = r_eff,
#'                                 type = 'quantile', probs = 0.9)
#' }
#' }
loo_predictive_metric <- function(x, ...) {
  UseMethod("loo_predictive_metric")
}

#' @rdname loo_predictive_metric
#' @export
loo_predictive_metric.matrix <-
  function(x,
           y,
           log_lik,
           ...,
           metric = c("mae", "rmse", "mse", "acc", "balanced_acc"),
           r_eff = 1,
           cores = getOption("mc.cores", 1)) {
    stopifnot(
      is.numeric(x),
      is.numeric(y),
      identical(ncol(x), length(y)),
      identical(dim(x), dim(log_lik))
    )
    metric <- match.arg(metric)
    psis_object <- psis(-log_lik, r_eff = r_eff, cores = cores)
    pred_loo <- E_loo(x,
                      psis_object = psis_object,
                      log_ratios = -log_lik,
                      ...)$value

    predictive_metric_fun <- .loo_predictive_metric_fun(metric)

    predictive_metric_fun(y, pred_loo)
  }


# ----------------------------- Internals -----------------------------

#' Select predictive metric function based on user's `metric` argument
#'
#' @noRd
#' @param metric The metric used.
#' @return The function used to compute predictive error or accuracy specified
#' by the argument `metric`.
.loo_predictive_metric_fun <- function(metric) {
  switch(
    metric,
    'mae' = .mae,
    'rmse' = .rmse,
    'mse' = .mse,
    'acc' = .accuracy,
    'balanced_acc' = .balanced_accuracy
  )
}

#' Mean absolute error
#'
#' @noRd
#' @param y A vector of observed values
#' @param yhat A vector of predictions
.mae <-function(y, yhat) {
  stopifnot(length(y) == length(yhat))
  n <- length(y)
  e <- abs(y - yhat)
  list(estimate = mean(e), se = sd(e) / sqrt(n))
}

#' Mean squared error
#'
#' @noRd
#' @param y A vector of observed values
#' @param yhat A vector of predictions
.mse <-function(y, yhat) {
  stopifnot(length(y) == length(yhat))
  n <- length(y)
  e <- (y - yhat)^2
  list(estimate = mean(e), se = sd(e) / sqrt(n))
}

#' Root mean squared error
#'
#' @noRd
#' @param y A vector of observed values
#' @param yhat A vector of predictions
.rmse <-function(y, yhat) {
  est <- .mse(y, yhat)
  mean_mse <- est$estimate
  var_mse <- est$se^2
  var_rmse <- var_mse / mean_mse / 4 # Comes from the first order Taylor approx.
  return(list(estimate = sqrt(mean_mse), se = sqrt(var_rmse)))
}

#' Classification accuracy
#'
#' @noRd
#' @param y A vector of observed values
#' @param yhat A vector of predictions
.accuracy <- function(y, yhat) {
  stopifnot(length(y) == length(yhat),
            all(y <= 1 & y >= 0),
            all(yhat <= 1 & yhat >= 0))
  n <- length(y)
  yhat <- as.integer(yhat > 0.5)
  acc <- as.integer(yhat == y)
  est <- mean(acc)
  list(estimate = est, se = sqrt(est * (1-est) / n) )
}

#' Balanced classification accuracy
#'
#' @noRd
#' @param y A vector of observed values
#' @param yhat A vector of predictions
.balanced_accuracy <- function(y, yhat) {
  stopifnot(length(y) == length(yhat),
            all(y <= 1 & y >= 0),
            all(yhat <= 1 & yhat >= 0))
  n <- length(y)
  yhat <- as.integer(yhat > 0.5)
  mask <- y == 0

  tn <- mean(yhat[mask] == y[mask]) # True negatives
  tp <- mean(yhat[!mask] == y[!mask]) # True positives

  bls_acc <- (tp + tn) / 2
  # This approximation has quite large bias for small samples
  bls_acc_var <- (tp * (1 - tp) + tn * (1 - tn)) / 4
  list(estimate = bls_acc, se = sqrt(bls_acc_var / n))
}

LL <- example_loglik_matrix()
chain_id <- rep(1:2, each = dim(LL)[1] / 2)
r_eff <- relative_eff(exp(LL), chain_id)
# cores = 1 because forked processes can segfault on CRAN's macOS machines if
# R is linked to Apple's Accelerate/vecLib BLAS, which is not fork-safe
psis_obj <- psis(-LL, r_eff = r_eff, cores = 1)

set.seed(123)
x <- matrix(rnorm(length(LL)), nrow = nrow(LL), ncol = ncol(LL))
x_prob <- 1 / (1 + exp(-x))
y <- rnorm(ncol(LL))
y_binary <- rbinom(ncol(LL), 1, 0.5)

mae_mean <- loo_predictive_metric(x, y, LL, metric = 'mae', r_eff = r_eff)
mae_quant <- loo_predictive_metric(x, y, LL, metric = 'mae', r_eff = r_eff,
                                  type = 'quantile', probs = 0.9)

rmse_mean <- loo_predictive_metric(x, y, LL, metric = 'rmse', r_eff = r_eff)
rmse_quant <- loo_predictive_metric(x, y, LL, metric = 'rmse', r_eff = r_eff,
                                   type = 'quantile', probs = 0.9)

mse_mean <- loo_predictive_metric(x, y, LL, metric = 'mse', r_eff = r_eff)
mse_quant <- loo_predictive_metric(x, y, LL, metric = 'mse', r_eff = r_eff,
                                  type = 'quantile', probs = 0.9)

acc_mean <- loo_predictive_metric(x_prob, y_binary, LL, metric = 'acc', r_eff = r_eff)
acc_quant <- loo_predictive_metric(x_prob, y_binary, LL, metric = 'acc', r_eff = r_eff,
                                  type = 'quantile', probs = 0.9)

bacc_mean <- loo_predictive_metric(x_prob, y_binary, LL, metric = 'balanced_acc', r_eff = r_eff)
bacc_quant <- loo_predictive_metric(x_prob, y_binary, LL, metric = 'balanced_acc', r_eff = r_eff,
                                  type = 'quantile', probs = 0.9)

test_that('loo_predictive_metric stops with incorrect inputs', {
  expect_error(loo_predictive_metric(as.character(x), y, LL, r_eff = r_eff),
               'no applicable method',
               fixed = TRUE)

  expect_error(loo_predictive_metric(x, as.character(y), LL, r_eff = r_eff),
               'is.numeric(y) is not TRUE',
               fixed = TRUE)

  x_invalid <- matrix(rnorm(9), nrow = 3)
  expect_error(loo_predictive_metric(x_invalid, y, LL, r_eff = r_eff),
               'identical(ncol(x), length(y)) is not TRUE',
               fixed = TRUE)

  x_invalid <- matrix(rnorm(64), nrow = 2)
  expect_error(loo_predictive_metric(x_invalid, y, LL, r_eff = r_eff),
               'identical(dim(x), dim(log_lik)) is not TRUE',
               fixed = TRUE)
})


test_that('loo_predictive_metric return types are correct', {
  # MAE
  expect_type(mae_mean, 'list')
  expect_type(mae_quant, 'list')
  expect_named(mae_mean, c('estimate', 'se'))
  expect_named(mae_quant, c('estimate', 'se'))
  # RMSE
  expect_type(rmse_mean, 'list')
  expect_type(rmse_quant, 'list')
  expect_named(rmse_mean, c('estimate', 'se'))
  expect_named(rmse_quant, c('estimate', 'se'))
  # MSE
  expect_type(mse_mean, 'list')
  expect_type(mse_quant, 'list')
  expect_named(mse_mean, c('estimate', 'se'))
  expect_named(mse_quant, c('estimate', 'se'))
  # Accuracy
  expect_type(acc_mean, 'list')
  expect_type(acc_quant, 'list')
  expect_named(acc_mean, c('estimate', 'se'))
  expect_named(acc_quant, c('estimate', 'se'))
  # Balanced accuracy
  expect_type(bacc_mean, 'list')
  expect_type(bacc_quant, 'list')
  expect_named(bacc_mean, c('estimate', 'se'))
  expect_named(bacc_quant, c('estimate', 'se'))
})

test_that('loo_predictive_metric is equal to snapshot', {
  expect_snapshot_value(mae_mean, style = "serialize")
  expect_snapshot_value(mae_quant, style = "serialize")
  expect_snapshot_value(rmse_mean, style = "serialize")
  expect_snapshot_value(rmse_quant, style = "serialize")
  expect_snapshot_value(mse_mean, style = "serialize")
  expect_snapshot_value(mse_quant, style = "serialize")
  expect_snapshot_value(acc_mean, style = "serialize")
  expect_snapshot_value(acc_quant, style = "serialize")
  expect_snapshot_value(bacc_mean, style = "serialize")
  expect_snapshot_value(bacc_quant, style = "serialize")
})

test_that('MAE computation is correct', {
  expect_equal(
    .mae(rep(0.5, 5), rep(1, 5))$estimate,
    0.5)
  expect_equal(
    .mae(rep(0.5, 5), rep(1, 5))$se,
    0.0)
  expect_error(
    .mae(rep(0.5, 5), rep(1, 3)),
    'length(y) == length(yhat) is not TRUE',
    fixed = TRUE)
})

test_that('MSE computation is correct', {
  expect_equal(
    .mse(rep(0.5, 5), rep(1, 5))$estimate,
    0.25)
  expect_equal(
    .mse(rep(0.5, 5), rep(1, 5))$se,
    0.0)
  expect_error(
    .mse(rep(0.5, 5), rep(1, 3)),
    'length(y) == length(yhat) is not TRUE',
    fixed = TRUE)
})

test_that('RMSE computation is correct', {
  expect_equal(
    .rmse(rep(0.5, 5), rep(1, 5))$estimate,
    sqrt(0.25))
  expect_equal(
    .mse(rep(0.5, 5), rep(1, 5))$se,
    0.0)
  expect_error(
    .mse(rep(0.5, 5), rep(1, 3)),
    'length(y) == length(yhat) is not TRUE',
    fixed = TRUE)
})

test_that('Accuracy computation is correct', {
  expect_equal(
    .accuracy(c(0, 0, 0, 1, 1, 1), c(0.2, 0.2, 0.2, 0.7, 0.7, 0.7))$estimate,
    1.0
  )
  expect_error(
    .accuracy(c(1, 0), c(0.5)),
    'length(y) == length(yhat) is not TRUE',
    fixed = TRUE)
  expect_error(
    .accuracy(c(2, 1), c(0.5, 0.5)),
    'all(y <= 1 & y >= 0) is not TRUE',
    fixed = TRUE
  )
  expect_error(
    .accuracy(c(1, 0), c(1.1, 0.5)),
    'all(yhat <= 1 & yhat >= 0) is not TRUE',
    fixed = TRUE
  )
})

test_that('Balanced accuracy computation is correct', {
  expect_equal(
    .balanced_accuracy(c(0, 0, 1, 1, 1, 1), c(0.9, 0.9, 0.9, 0.9, 0.9, 0.9))$estimate,
    0.5
  )
  expect_error(
    .balanced_accuracy(c(1, 0), c(0.5)),
    'length(y) == length(yhat) is not TRUE',
    fixed = TRUE)
  expect_error(
    .balanced_accuracy(c(2, 1), c(0.5, 0.5)),
    'all(y <= 1 & y >= 0) is not TRUE',
    fixed = TRUE
  )
  expect_error(
    .balanced_accuracy(c(1, 0), c(1.1, 0.5)),
    'all(yhat <= 1 & yhat >= 0) is not TRUE',
    fixed = TRUE
  )
})

LL <- example_loglik_matrix()
chain_id <- rep(1:2, each = dim(LL)[1] / 2)
r_eff <- relative_eff(exp(LL), chain_id)
psis_obj <- psis(-LL, r_eff = r_eff, cores = 2)

set.seed(123)
x <- matrix(rnorm(length(LL)), nrow = nrow(LL), ncol = ncol(LL))
x_prob <- 1 / (1 + exp(-x))
y <- rnorm(ncol(LL))
y_binary <- rbinom(ncol(LL), 1, 0.5)

mae_mean <- loo_predictive_metric(x, y, LL, metric = 'mae', r_eff = r_eff)
mae_quant <- loo_predictive_metric(x, y, LL, metric = 'mae', r_eff = r_eff,
                                  type = 'quantile', probs = 0.9)

rmse_mean <- loo_predictive_metric(x, y, LL, metric = 'rmse', r_eff = r_eff)
rmse_quant <- loo_predictive_metric(x, y, LL, metric = 'rmse', r_eff = r_eff,
                                   type = 'quantile', probs = 0.9)

mse_mean <- loo_predictive_metric(x, y, LL, metric = 'mse', r_eff = r_eff)
mse_quant <- loo_predictive_metric(x, y, LL, metric = 'mse', r_eff = r_eff,
                                  type = 'quantile', probs = 0.9)

acc_mean <- loo_predictive_metric(x_prob, y_binary, LL, metric = 'acc', r_eff = r_eff)
acc_quant <- loo_predictive_metric(x_prob, y_binary, LL, metric = 'acc', r_eff = r_eff,
                                  type = 'quantile', probs = 0.9)

bacc_mean <- loo_predictive_metric(x_prob, y_binary, LL, metric = 'balanced_acc', r_eff = r_eff)
bacc_quant <- loo_predictive_metric(x_prob, y_binary, LL, metric = 'balanced_acc', r_eff = r_eff,
                                  type = 'quantile', probs = 0.9)

test_that('loo_predictive_metric stops with incorrect inputs', {
  expect_error(loo_predictive_metric(as.character(x), y, LL, r_eff = r_eff),
               'no applicable method',
               fixed = TRUE)

  expect_error(loo_predictive_metric(x, as.character(y), LL, r_eff = r_eff),
               'is.numeric(y) is not TRUE',
               fixed = TRUE)

  x_invalid <- matrix(rnorm(9), nrow = 3)
  expect_error(loo_predictive_metric(x_invalid, y, LL, r_eff = r_eff),
               'identical(ncol(x), length(y)) is not TRUE',
               fixed = TRUE)

  x_invalid <- matrix(rnorm(64), nrow = 2)
  expect_error(loo_predictive_metric(x_invalid, y, LL, r_eff = r_eff),
               'identical(dim(x), dim(log_lik)) is not TRUE',
               fixed = TRUE)
})


test_that('loo_predictive_metric return types are correct', {
  # MAE
  expect_type(mae_mean, 'list')
  expect_type(mae_quant, 'list')
  expect_named(mae_mean, c('estimate', 'se'))
  expect_named(mae_quant, c('estimate', 'se'))
  # RMSE
  expect_type(rmse_mean, 'list')
  expect_type(rmse_quant, 'list')
  expect_named(rmse_mean, c('estimate', 'se'))
  expect_named(rmse_quant, c('estimate', 'se'))
  # MSE
  expect_type(mse_mean, 'list')
  expect_type(mse_quant, 'list')
  expect_named(mse_mean, c('estimate', 'se'))
  expect_named(mse_quant, c('estimate', 'se'))
  # Accuracy
  expect_type(acc_mean, 'list')
  expect_type(acc_quant, 'list')
  expect_named(acc_mean, c('estimate', 'se'))
  expect_named(acc_quant, c('estimate', 'se'))
  # Balanced accuracy
  expect_type(bacc_mean, 'list')
  expect_type(bacc_quant, 'list')
  expect_named(bacc_mean, c('estimate', 'se'))
  expect_named(bacc_quant, c('estimate', 'se'))
})

test_that('loo_predictive_metric is equal to snapshot', {
  expect_snapshot_value(mae_mean, style = "serialize")
  expect_snapshot_value(mae_quant, style = "serialize")
  expect_snapshot_value(rmse_mean, style = "serialize")
  expect_snapshot_value(rmse_quant, style = "serialize")
  expect_snapshot_value(mse_mean, style = "serialize")
  expect_snapshot_value(mse_quant, style = "serialize")
  expect_snapshot_value(acc_mean, style = "serialize")
  expect_snapshot_value(acc_quant, style = "serialize")
  expect_snapshot_value(bacc_mean, style = "serialize")
  expect_snapshot_value(bacc_quant, style = "serialize")
})

test_that('MAE computation is correct', {
  expect_equal(
    .mae(rep(0.5, 5), rep(1, 5))$estimate,
    0.5)
  expect_equal(
    .mae(rep(0.5, 5), rep(1, 5))$se,
    0.0)
  expect_error(
    .mae(rep(0.5, 5), rep(1, 3)),
    'length(y) == length(yhat) is not TRUE',
    fixed = TRUE)
})

test_that('MSE computation is correct', {
  expect_equal(
    .mse(rep(0.5, 5), rep(1, 5))$estimate,
    0.25)
  expect_equal(
    .mse(rep(0.5, 5), rep(1, 5))$se,
    0.0)
  expect_error(
    .mse(rep(0.5, 5), rep(1, 3)),
    'length(y) == length(yhat) is not TRUE',
    fixed = TRUE)
})

test_that('RMSE computation is correct', {
  expect_equal(
    .rmse(rep(0.5, 5), rep(1, 5))$estimate,
    sqrt(0.25))
  expect_equal(
    .mse(rep(0.5, 5), rep(1, 5))$se,
    0.0)
  expect_error(
    .mse(rep(0.5, 5), rep(1, 3)),
    'length(y) == length(yhat) is not TRUE',
    fixed = TRUE)
})

test_that('Accuracy computation is correct', {
  expect_equal(
    .accuracy(c(0, 0, 0, 1, 1, 1), c(0.2, 0.2, 0.2, 0.7, 0.7, 0.7))$estimate,
    1.0
  )
  expect_error(
    .accuracy(c(1, 0), c(0.5)),
    'length(y) == length(yhat) is not TRUE',
    fixed = TRUE)
  expect_error(
    .accuracy(c(2, 1), c(0.5, 0.5)),
    'all(y <= 1 & y >= 0) is not TRUE',
    fixed = TRUE
  )
  expect_error(
    .accuracy(c(1, 0), c(1.1, 0.5)),
    'all(yhat <= 1 & yhat >= 0) is not TRUE',
    fixed = TRUE
  )
})

test_that('Balanced accuracy computation is correct', {
  expect_equal(
    .balanced_accuracy(c(0, 0, 1, 1, 1, 1), c(0.9, 0.9, 0.9, 0.9, 0.9, 0.9))$estimate,
    0.5
  )
  expect_error(
    .balanced_accuracy(c(1, 0), c(0.5)),
    'length(y) == length(yhat) is not TRUE',
    fixed = TRUE)
  expect_error(
    .balanced_accuracy(c(2, 1), c(0.5, 0.5)),
    'all(y <= 1 & y >= 0) is not TRUE',
    fixed = TRUE
  )
  expect_error(
    .balanced_accuracy(c(1, 0), c(1.1, 0.5)),
    'all(yhat <= 1 & yhat >= 0) is not TRUE',
    fixed = TRUE
  )
})

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----stancode-----------------------------------------------------------------
# Note: some syntax used in this Stan program requires RStan >= 2.26 (or CmdStanR)
# To use an older version of RStan change the line declaring `y` to: int y[N];
stancode <- "
data {
  int<lower=1> K;
  int<lower=1> N;
  matrix[N,K] x;
  array[N] int y;
  vector[N] offset;

  real beta_prior_scale;
  real alpha_prior_scale;
}
parameters {
  vector[K] beta;
  real intercept;
}
model {
  y ~ poisson(exp(x * beta + intercept + offset));
  beta ~ normal(0,beta_prior_scale);
  intercept ~ normal(0,alpha_prior_scale);
}
generated quantities {
  vector[N] log_lik;
  for (n in 1:N)
    log_lik[n] = poisson_lpmf(y[n] | exp(x[n] * beta + intercept + offset[n]));
}
"

## ----setup, message=FALSE-----------------------------------------------------
library("rstan")
library("loo")
seed <- 9547
set.seed(seed)

## ----modelfit, message=FALSE--------------------------------------------------
# Prepare data
data(roaches, package = "rstanarm")
roaches$roach1 <- sqrt(roaches$roach1)
y <- roaches$y
x <- roaches[,c("roach1", "treatment", "senior")]
offset <- log(roaches[,"exposure2"])
n <- dim(x)[1]
k <- dim(x)[2]

standata <- list(N = n, K = k, x = as.matrix(x), y = y, offset = offset, beta_prior_scale = 2.5, alpha_prior_scale = 5.0)

# Compile
stanmodel <- stan_model(model_code = stancode)

# Fit model
fit <- sampling(stanmodel, data = standata, seed = seed, refresh = 0)
print(fit, pars = "beta")

## ----loo1---------------------------------------------------------------------
loo1 <- loo(fit)
loo1

## ----loo_moment_match---------------------------------------------------------
# available in rstan >= 2.21
loo2 <- loo(fit, moment_match = TRUE)
loo2

## ----stanfitfuns--------------------------------------------------------------
# create a named list of draws for use with rstan methods
.rstan_relist <- function(x, skeleton) {
  out <- utils::relist(x, skeleton)
  for (i in seq_along(skeleton)) {
    dim(out[[i]]) <- dim(skeleton[[i]])
  }
  out
}

# rstan helper function to get dims of parameters right
.create_skeleton <- function(pars, dims) {
  out <- lapply(seq_along(pars), function(i) {
    len_dims <- length(dims[[i]])
    if (len_dims < 1) return(0)
    return(array(0, dim = dims[[i]]))
  })
  names(out) <- pars
  out
}

# extract original posterior draws
post_draws_stanfit <- function(x, ...) {
  as.matrix(x)
}

# compute a matrix of log-likelihood values for the ith observation
# matrix contains information about the number of MCMC chains
log_lik_i_stanfit <- function(x, i, parameter_name = "log_lik", ...) {
  loo::extract_log_lik(x, parameter_name, merge_chains = FALSE)[, , i]
}

# transform parameters to the unconstraint space
unconstrain_pars_stanfit <- function(x, pars, ...) {
  skeleton <- .create_skeleton(x@sim$pars_oi, x@par_dims[x@sim$pars_oi])
  upars <- apply(pars, 1, FUN = function(theta) {
    rstan::unconstrain_pars(x, .rstan_relist(theta, skeleton))
  })
  # for one parameter models
  if (is.null(dim(upars))) {
    dim(upars) <- c(1, length(upars))
  }
  t(upars)
}

# compute log_prob for each posterior draws on the unconstrained space
log_prob_upars_stanfit <- function(x, upars, ...) {
  apply(upars, 1, rstan::log_prob, object = x,
        adjust_transform = TRUE, gradient = FALSE)
}

# compute log_lik values based on the unconstrained parameters
log_lik_i_upars_stanfit <- function(x, upars, i, parameter_name = "log_lik",
                                  ...) {
  S <- nrow(upars)
  out <- numeric(S)
  for (s in seq_len(S)) {
    out[s] <- rstan::constrain_pars(x, upars = upars[s, ])[[parameter_name]][i]
  }
  out
}

## ----loo_moment_match.default, message=FALSE----------------------------------
loo3 <- loo::loo_moment_match.default(
  x = fit,
  loo = loo1,
  post_draws = post_draws_stanfit,
  log_lik_i = log_lik_i_stanfit,
  unconstrain_pars = unconstrain_pars_stanfit,
  log_prob_upars = log_prob_upars_stanfit,
  log_lik_i_upars = log_lik_i_upars_stanfit
)
loo3

N <- 50; K <- 10; S <- 100; a0 <- 3; b0 <- 2
p <- rbeta(1, a0, b0)
y <- rbinom(N, size = K, prob = p)
a <- a0 + sum(y); b <- b0 + N * K - sum(y)
draws <- as.matrix(rbeta(S, a, b))
data <- data.frame(y,K)
llfun <- function(data_i, draws) {
  dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
}
llmat_from_fn <- sapply(1:N, function(i) llfun(data[i,, drop=FALSE], draws))

#' Efficient approximate leave-one-out cross-validation (LOO) for posterior
#' approximations
#'
#' @param x A log-likelihood array, matrix, or function.
#'   The **Methods (by class)** section, below, has detailed descriptions of how
#'   to specify the inputs for each method.
#' @param save_psis Should the `"psis"` object created internally by
#'   `loo_approximate_posterior()` be saved in the returned object? See
#'   [loo()] for details.
#' @template cores
#' @inheritParams psis_approximate_posterior
#'
#' @details The `loo_approximate_posterior()` function is an S3 generic and
#'   methods are provided for 3-D pointwise log-likelihood arrays, pointwise
#'   log-likelihood matrices, and log-likelihood functions. The implementation
#'   works for posterior approximations where it is possible to compute the log
#'   density for the posterior approximation.
#'
#' @return The `loo_approximate_posterior()` methods return a named list with
#'   class `c("psis_loo_ap", "psis_loo", "loo")`. It has the same structure
#'   as the objects returned by [loo()] but with the additional slot:
#' \describe{
#'  \item{`posterior_approximation`}{
#'   A list with two vectors, `log_p` and `log_g` of the same length
#'   containing the posterior density and the approximation density
#'   for the individual draws.
#'  }
#' }
#'
#' @seealso [loo()], [psis()], [loo_compare()]
#' @template loo-large-data-references
#'
#' @export loo_approximate_posterior
#' @export loo_approximate_posterior.array loo_approximate_posterior.matrix loo_approximate_posterior.function
#'
loo_approximate_posterior <- function(x, log_p, log_g, ...) {
  UseMethod("loo_approximate_posterior")
}

#' @export
#' @templateVar fn loo_approximate_posterior
#' @template array
loo_approximate_posterior.array <-
  function(x,
           log_p,
           log_g,
           ...,
           save_psis = FALSE,
           cores = getOption("mc.cores", 1)) {
    checkmate::assert_flag(save_psis)
    checkmate::assert_int(cores)
    checkmate::assert_matrix(log_p, mode = "numeric", nrows = dim(x)[1], ncols = dim(x)[2])
    checkmate::assert_matrix(log_g, mode = "numeric", nrows = nrow(log_p), ncols = ncol(log_p))

    ll <- llarray_to_matrix(x)
    log_p <- as.vector(log_p)
    log_g <- as.vector(log_g)
    loo_approximate_posterior.matrix(
      ll,
      log_p = log_p,
      log_g = log_g,
      ...,
      save_psis = save_psis,
      cores = cores
    )
  }

#' @export
#' @templateVar fn loo_approximate_posterior
#' @template matrix
loo_approximate_posterior.matrix <-
  function(x,
           log_p,
           log_g,
           ...,
           save_psis = FALSE,
           cores = getOption("mc.cores", 1)) {
    checkmate::assert_flag(save_psis)
    checkmate::assert_int(cores)
    checkmate::assert_numeric(log_p, len = nrow(x))
    checkmate::assert_null(dim(log_p))
    checkmate::assert_numeric(log_g, len = length(log_p))
    checkmate::assert_null(dim(log_g))

    ap_psis <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = x,
        ...,
        cores = cores,
        save_psis = save_psis
      )
    ap_psis$approximate_posterior <- list(log_p = log_p, log_g = log_g)
    class(ap_psis) <- c("psis_loo_ap", class(ap_psis))
    assert_psis_loo_ap(ap_psis)
    ap_psis
  }

#' @export
#' @templateVar fn loo_approximate_posterior
#' @template function
#' @param data,draws,... For the `loo_approximate_posterior.function()` method,
#'   these are the data, posterior draws, and other arguments to pass to the
#'   log-likelihood function. See the **Methods (by class)** section below for
#'   details on how to specify these arguments.
#'
loo_approximate_posterior.function <-
  function(x,
           ...,
           data = NULL,
           draws = NULL,
           log_p = NULL,
           log_g = NULL,
           save_psis = FALSE,
           cores = getOption("mc.cores", 1)) {

    checkmate::assert_numeric(log_p, len = length(log_g))
    checkmate::assert_numeric(log_g, len = length(log_p))
    cores <- loo_cores(cores)
    stopifnot(is.data.frame(data) || is.matrix(data), !is.null(draws))
    .llfun <- validate_llfun(x)
    N <- dim(data)[1]

    psis_list <- parallel_psis_list(N = N,
                                    .loo_i = .loo_ap_i,
                                    .llfun = .llfun,
                                    data = data,
                                    draws = draws,
                                    r_eff = 1, # r_eff is ignored
                                    save_psis = save_psis,
                                    log_p = log_p,
                                    log_g = log_g,
                                    cores = cores,
                                    ...)

    pointwise <- lapply(psis_list, "[[", "pointwise")
    if (save_psis) {
      psis_object_list <- lapply(psis_list, "[[", "psis_object")
      psis_out <- list2importance_sampling(psis_object_list)
      diagnostics <- psis_out$diagnostics
    } else {
      diagnostics_list <- lapply(psis_list, "[[", "diagnostics")
      diagnostics <- list(
        pareto_k = psis_apply(diagnostics_list, "pareto_k"),
        n_eff = psis_apply(diagnostics_list, "n_eff")
      )
    }

    ap_psis <- importance_sampling_loo_object(
      pointwise = do.call(rbind, pointwise),
      diagnostics = diagnostics,
      dims = c(attr(psis_list[[1]], "S"), N),
      is_method = "psis",
      is_object = if (save_psis) psis_out else NULL
    )

    ap_psis$approximate_posterior <- list(log_p = log_p, log_g = log_g)
    class(ap_psis) <- c("psis_loo_ap", class(ap_psis))
    assert_psis_loo_ap(ap_psis)
    ap_psis
  }

# Function that is passed to the FUN argument of lapply, mclapply, or parLapply
# for the loo_approximate_posterior.function method. The arguments and return
# value are the same as the ones documented for the user-facing loo_i function.
.loo_ap_i <-
  function(i,
           llfun,
           ...,
           data,
           draws,
           log_p,
           log_g,
           r_eff = 1,
           save_psis = FALSE,
           is_method) {

    if (is_method != "psis") stop(is_method, " not implemented for aploo.")
    d_i <- data[i, , drop = FALSE]
    ll_i <- llfun(data_i = d_i, draws = draws, ...)
    if (!is.matrix(ll_i)) {
      ll_i <- as.matrix(ll_i)
    }
    psis_out <- ap_psis(log_ratios = -ll_i, log_p = log_p, log_g = log_g, cores = 1)

    structure(
      list(
        pointwise = pointwise_loo_calcs(ll_i, psis_out),
        diagnostics = psis_out$diagnostics,
        psis_object = if (save_psis) psis_out else NULL
      ),
      S = dim(psis_out)[1],
      N = 1
    )
  }


assert_psis_loo_ap <- function(x) {
  checkmate::assert_class(x, "psis_loo_ap")
  checkmate::assert_names(names(x), must.include = c("estimates", "pointwise", "diagnostics", "psis_object", "approximate_posterior"))
  checkmate::assert_names(names(x$approximate_posterior), must.include = c("log_p", "log_g"))
  checkmate::assert_numeric(x$approximate_posterior$log_p, len = length(x$approximate_posterior$log_g), any.missing = FALSE)
  checkmate::assert_numeric(x$approximate_posterior$log_g, len = length(x$approximate_posterior$log_p), any.missing = FALSE)
}

# Create test data
# Checked by Mans M and Paul B 24th of June 2019
set.seed(123)
N <- 50
K <- 10
S <- 1000
a0 <- 1
b0 <- 1
p <- 0.5
y <- rbinom(N, size = K, prob = p)
fake_data <- data.frame(y, K)

# The log posterior
log_post <- function(p, y, a0, b0, K) {
  log_lik <- sum(dbinom(x = y, size = K, prob = p, log = TRUE)) # the log likelihood
  log_post <- log_lik + dbeta(x = p, shape1 = a0, shape2 = b0, log = TRUE) # the log prior
  log_post
}
it <- optim(
  par = 0.5,
  fn = log_post,
  control = list(fnscale = -1),
  hessian = TRUE,
  y = y,
  a0 = a0,
  b0 = b0,
  K = K,
  lower = 0.01,
  upper = 0.99,
  method = "Brent"
)
lap_params <- c(mu = it$par, sd = sqrt(solve(-it$hessian)))

a <- a0 + sum(y)
b <- b0 + N * K - sum(y)
fake_true_posterior <- as.matrix(rbeta(S, a, b))
fake_laplace_posterior <- as.matrix(rnorm(
  n = S,
  mean = lap_params["mu"],
  sd = lap_params["sd"]
))
# mean(fake_laplace_posterior); sd(fake_laplace_posterior)

p_draws <- as.vector(fake_laplace_posterior)
log_p <- numeric(S)
for (s in 1:S) {
  log_p[s] <- log_post(p_draws[s], y = y, a0 = a0, b0 = b0, K = K)
}
log_g <- as.vector(dnorm(
  as.vector(fake_laplace_posterior),
  mean = lap_params["mu"],
  sd = lap_params["sd"],
  log = TRUE
))

llfun <- function(data_i, draws) {
  dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
}

ll <- matrix(0, nrow = S, ncol = N)
for (j in 1:N) {
  ll[, j] <- llfun(
    data_i = fake_data[j, , drop = FALSE],
    draws = fake_laplace_posterior
  )
}


test_that("loo_approximate_posterior.array works as loo_approximate_posterior.matrix", {
  # Create array with two "chains"
  log_p_mat <- matrix(log_p, nrow = (S / 2), ncol = 2)
  log_g_mat <- matrix(log_g, nrow = (S / 2), ncol = 2)
  ll_array <- array(0, dim = c((S / 2), 2, ncol(ll)))
  ll_array[, 1, ] <- ll[1:(S / 2), ]
  ll_array[, 2, ] <- ll[(S / 2 + 1):S, ]

  # Assert that they are ok
  expect_equal(ll_array[1:2, 1, 1:2], ll[1:2, 1:2], ignore_attr = TRUE)
  expect_equal(
    ll_array[1:2, 2, 1:2],
    ll[(S / 2 + 1):((S / 2) + 2), 1:2],
    ignore_attr = TRUE
  )

  # Compute aploo
  expect_silent(
    aploo1 <- loo_approximate_posterior.matrix(
      x = ll,
      log_p = log_p,
      log_g = log_g
    )
  )
  expect_silent(
    aploo2 <- loo_approximate_posterior.array(
      x = ll_array,
      log_p = log_p_mat,
      log_g = log_g_mat
    )
  )
  expect_silent(aploo1b <- loo.matrix(x = ll, r_eff = rep(1, N)))

  # Check equivalence
  expect_equal(aploo1$estimates, aploo2$estimates)
  expect_equal(class(aploo1), class(aploo2))
  expect_failure(expect_equal(aploo1b$estimates, aploo2$estimates))
  expect_failure(expect_equal(class(aploo1), class(aploo1b)))

  # Should fail with matrix
  expect_error(
    aploo2 <- loo_approximate_posterior.matrix(
      x = ll,
      log_p = as.matrix(log_p),
      log_g = log_g
    )
  )
  expect_error(
    aploo2 <- loo_approximate_posterior.matrix(
      x = ll,
      log_p = as.matrix(log_p),
      log_g = as.matrix(log_g)
    )
  )

  # Expect log_p and log_g be stored in the approximate_posterior in the same way
  expect_length(aploo1$approximate_posterior$log_p, nrow(ll))
  expect_length(aploo1$approximate_posterior$log_g, nrow(ll))
  expect_equal(
    aploo1$approximate_posterior$log_p,
    aploo2$approximate_posterior$log_p
  )
  expect_equal(
    aploo1$approximate_posterior$log_g,
    aploo2$approximate_posterior$log_g
  )
})

test_that("loo_approximate_posterior.function works as loo_approximate_posterior.matrix", {
  # Compute aploo
  expect_silent(
    aploo1 <- loo_approximate_posterior.matrix(
      x = ll,
      log_p = log_p,
      log_g = log_g
    )
  )
  expect_silent(aploo1b <- loo.matrix(x = ll, r_eff = rep(1, N)))
  expect_silent(
    aploo2 <- loo_approximate_posterior.function(
      x = llfun,
      log_p = log_p,
      log_g = log_g,
      data = fake_data,
      draws = fake_laplace_posterior
    )
  )

  # Check equivalence
  expect_equal(aploo1$estimates, aploo2$estimates)
  expect_equal(class(aploo1), class(aploo2))
  expect_failure(expect_equal(aploo1b$estimates, aploo2$estimates))

  # Check equivalence
  # Expect log_p and log_g be stored in the approximate_posterior in the same way
  expect_length(
    aploo2$approximate_posterior$log_p,
    nrow(fake_laplace_posterior)
  )
  expect_length(
    aploo2$approximate_posterior$log_g,
    nrow(fake_laplace_posterior)
  )
  expect_equal(
    aploo1$approximate_posterior$log_p,
    aploo2$approximate_posterior$log_p
  )
  expect_equal(
    aploo1$approximate_posterior$log_g,
    aploo2$approximate_posterior$log_g
  )
})

suppressPackageStartupMessages(library(loo))

context("loo_approximate_posterior")

# Create test data
# Checked by Mans M and Paul B 24th of June 2019
set.seed(123)
N <- 50
K <- 10
S <- 1000
a0 <- 1
b0 <- 1
p <- 0.5
y <- rbinom(N, size = K, prob = p)
fake_data <- data.frame(y,K)

# The log posterior
log_post <- function(p, y, a0, b0, K) {
  log_lik <- sum(dbinom(x = y, size = K, prob = p, log = TRUE))  # the log likelihood
  log_post <- log_lik + dbeta(x = p, shape1 = a0, shape2 = b0, log = TRUE)  # the log prior
  log_post
}
it <- optim(par = 0.5, fn = log_post, control = list(fnscale = -1),
            hessian = TRUE,  y = y, a0 = a0, b0 = b0, K = K,
            lower = 0.01, upper = 0.99, method = "Brent")
lap_params <- c(mu = it$par, sd = sqrt(solve(-it$hessian)))

a <- a0 + sum(y)
b <- b0 + N * K - sum(y)
fake_true_posterior <- as.matrix(rbeta(S, a, b))
fake_laplace_posterior <- as.matrix(rnorm(n = S, mean = lap_params["mu"], sd = lap_params["sd"]))
# mean(fake_laplace_posterior); sd(fake_laplace_posterior)

p_draws <- as.vector(fake_laplace_posterior)
log_p <- numeric(S)
for(s in 1:S){
  log_p[s] <- log_post(p_draws[s], y = y, a0 = a0, b0 = b0, K = K)
}
log_g <- as.vector(dnorm(as.vector(fake_laplace_posterior), mean = lap_params["mu"], sd = lap_params["sd"], log = TRUE))

llfun <- function(data_i, draws) {
  dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
}

ll <- matrix(0, nrow = S, ncol = N)
for(j in 1:N){
  ll[, j] <- llfun(data_i = fake_data[j, , drop=FALSE], draws = fake_laplace_posterior)
}


test_that("loo_approximate_posterior.array works as loo_approximate_posterior.matrix", {

  # Create array with two "chains"
  log_p_mat <- matrix(log_p, nrow = (S/2), ncol = 2)
  log_g_mat <- matrix(log_g, nrow = (S/2), ncol = 2)
  ll_array <- array(0, dim = c((S/2), 2 , ncol(ll)))
  ll_array[,1,] <- ll[1:(S/2),]
  ll_array[,2,] <- ll[(S/2 + 1):S,]

  # Assert that they are ok
  expect_equivalent(ll_array[1:2,1,1:2], ll[1:2,1:2])
  expect_equivalent(ll_array[1:2,2,1:2], ll[(S/2+1):((S/2)+2),1:2])

  # Compute aploo
  expect_silent(aploo1 <- loo_approximate_posterior.matrix(x = ll, log_p = log_p, log_g = log_g))
  expect_silent(aploo2 <- loo_approximate_posterior.array(x = ll_array, log_p = log_p_mat, log_g = log_g_mat))
  expect_silent(aploo1b <- loo.matrix(x = ll, r_eff = rep(1, N)))

  # Check equivalence
  expect_equal(aploo1$estimates, aploo2$estimates)
  expect_equal(class(aploo1), class(aploo2))
  expect_failure(expect_equal(aploo1b$estimates, aploo2$estimates))
  expect_failure(expect_equal(class(aploo1), class(aploo1b)))

  # Should fail with matrix
  expect_error(aploo2 <- loo_approximate_posterior.matrix(x = ll, log_p = as.matrix(log_p), log_g = log_g))
  expect_error(aploo2 <- loo_approximate_posterior.matrix(x = ll, log_p = as.matrix(log_p), log_g = as.matrix(log_g)))

  # Expect log_p and log_g be stored in the approximate_posterior in the same way
  expect_length(aploo1$approximate_posterior$log_p, nrow(ll))
  expect_length(aploo1$approximate_posterior$log_g, nrow(ll))
  expect_equal(aploo1$approximate_posterior$log_p, aploo2$approximate_posterior$log_p)
  expect_equal(aploo1$approximate_posterior$log_g, aploo2$approximate_posterior$log_g)

})

test_that("loo_approximate_posterior.function works as loo_approximate_posterior.matrix", {


  # Compute aploo
  expect_silent(aploo1 <- loo_approximate_posterior.matrix(x = ll, log_p = log_p, log_g = log_g))
  expect_silent(aploo1b <- loo.matrix(x = ll, r_eff = rep(1, N)))
  expect_silent(aploo2 <- loo_approximate_posterior.function(x = llfun, log_p = log_p, log_g = log_g, data = fake_data, draws = fake_laplace_posterior))

  # Check equivalence
  expect_equal(aploo1$estimates, aploo2$estimates)
  expect_equal(class(aploo1), class(aploo2))
  expect_failure(expect_equal(aploo1b$estimates, aploo2$estimates))

  # Check equivalence
  # Expect log_p and log_g be stored in the approximate_posterior in the same way
  expect_length(aploo2$approximate_posterior$log_p, nrow(fake_laplace_posterior))
  expect_length(aploo2$approximate_posterior$log_g, nrow(fake_laplace_posterior))
  expect_equal(aploo1$approximate_posterior$log_p, aploo2$approximate_posterior$log_p)
  expect_equal(aploo1$approximate_posterior$log_g, aploo2$approximate_posterior$log_g)

})

#' Split moment matching for efficient approximate leave-one-out cross-validation (LOO)
#'
#' A function that computes the split moment matching importance sampling loo.
#' Takes in the moment matching total transformation, transforms only half
#' of the draws, and computes a single elpd using multiple importance sampling.
#'
#' @param x A fitted model object.
#' @param upars A matrix containing the model parameters in unconstrained space
#'   where they can have any real value.
#' @param cov Logical; Indicate whether to match the covariance matrix of the
#'   samples or not. If `FALSE`, only the mean and marginal variances are
#'   matched.
#' @param total_shift A vector representing the total shift made by the moment
#'   matching algorithm.
#' @param total_scaling A vector representing the total scaling of marginal
#'   variance made by the moment matching algorithm.
#' @param total_mapping A vector representing the total covariance
#'   transformation made by the moment matching algorithm.
#' @param i Observation index.
#' @param log_prob_upars A function that takes arguments `x` and `upars` and
#'   returns a matrix of log-posterior density values of the unconstrained
#'   posterior draws passed via `upars`.
#' @param log_lik_i_upars A function that takes arguments `x`, `upars`, and `i`
#'   and returns a vector of log-likelihood draws of the `i`th observation based
#'   on the unconstrained posterior draws passed via `upars`.
#' @param r_eff_i MCMC relative effective sample size of the `i`'th log
#'   likelihood draws.
#' @template cores
#' @template is_method
#' @param ... Further arguments passed to the custom functions documented above.
#'
#' @return A list containing the updated log-importance weights and
#' log-likelihood values. Also returns the updated MCMC effective sample size
#' and the integrand-specific log-importance weights.
#'
#'
#' @seealso [loo()], [loo_moment_match()]
#' @template moment-matching-references
#'
#'
loo_moment_match_split <- function(x, upars, cov, total_shift, total_scaling,
                     total_mapping, i, log_prob_upars,
                     log_lik_i_upars, r_eff_i, cores,
                     is_method, ...) {
  S <- dim(upars)[1]
  S_half <- as.integer(0.5 * S)
  mean_original <- colMeans(upars)

  # accumulated affine transformation
  upars_trans <- sweep(upars, 2, mean_original, "-")
  upars_trans <- sweep(upars_trans, 2, total_scaling, "*")
  if (cov) {
    upars_trans <- tcrossprod(upars_trans, total_mapping)
  }
  upars_trans <- sweep(upars_trans, 2, total_shift + mean_original, "+")
  attributes(upars_trans) <- attributes(upars)

  # inverse accumulated affine transformation
  upars_trans_inv <- sweep(upars, 2, mean_original, "-")
  if (cov) {
    upars_trans_inv <- tcrossprod(upars_trans_inv, solve(total_mapping))
  }
  upars_trans_inv <- sweep(upars_trans_inv, 2, total_scaling, "/")
  upars_trans_inv <- sweep(upars_trans_inv, 2, mean_original - total_shift, "+")
  attributes(upars_trans_inv) <- attributes(upars)

  # first half of upars_trans_half are T(theta)
  # second half are theta
  upars_trans_half <- upars
  take <- seq_len(S_half)
  upars_trans_half[take, ] <- upars_trans[take, , drop = FALSE]

  # first half of upars_half_inv are theta
  # second half are T^-1 (theta)
  upars_trans_half_inv <- upars
  take <- seq_len(S)[-seq_len(S_half)]
  upars_trans_half_inv[take, ] <- upars_trans_inv[take, , drop = FALSE]

  # compute log likelihoods and log probabilities
  log_prob_half_trans <- log_prob_upars(x, upars = upars_trans_half, ...)
  log_prob_half_trans_inv <- log_prob_upars(x, upars = upars_trans_half_inv, ...)
  log_liki_half <- log_lik_i_upars(x, upars = upars_trans_half, i = i, ...)

  # compute weights
  log_prob_half_trans_inv <- (log_prob_half_trans_inv -
                              log(prod(total_scaling)) -
                              log(det(total_mapping)))
  stable_S <- log_prob_half_trans > log_prob_half_trans_inv

  lwi_half <- -log_liki_half + log_prob_half_trans
  lwi_half[stable_S] <- lwi_half[stable_S] -
    (log_prob_half_trans[stable_S] +
      log1p(exp(log_prob_half_trans_inv[stable_S] -
                  log_prob_half_trans[stable_S])))

  lwi_half[!stable_S] <- lwi_half[!stable_S] -
    (log_prob_half_trans_inv[!stable_S] +
        log1p(exp(log_prob_half_trans[!stable_S] -
                   log_prob_half_trans_inv[!stable_S])))

  # lwi_half may have NaNs if computation involves -Inf + Inf
  # replace NaN log ratios with -Inf
  lr <- lwi_half
  lr[is.na(lr)] <- -Inf
  is_obj_half <- suppressWarnings(importance_sampling.default(lr,
                                                              method = is_method,
                                                              r_eff = r_eff_i,
                                                              cores = cores))
  lwi_half <- as.vector(weights(is_obj_half))

  # lwi_half may have NaNs if computation involves -Inf + Inf
  # replace NaN log ratios with -Inf
  lr <- lwi_half + log_liki_half
  lr[is.na(lr)] <- -Inf
  is_obj_f_half <- suppressWarnings(importance_sampling.default(lr,
                                                                method = is_method,
                                                                r_eff = r_eff_i,
                                                                cores = cores))
  lwfi_half <- as.vector(weights(is_obj_f_half))

  # relative_eff recomputation
  # currently ignores chain information
  # since we have two proposal distributions
  # compute S_eff separately from both and take the smaller
  take <- seq_len(S)[-seq_len(S_half)]
  log_liki_half_1 <- log_liki_half[take, drop = FALSE]
  dim(log_liki_half_1) <- c(length(take), 1, 1)
  take <- seq_len(S)[-seq_len(S_half)]
  log_liki_half_2 <- log_liki_half[take, drop = FALSE]
  dim(log_liki_half_2) <- c(length(take), 1, 1)
  r_eff_i1 <- loo::relative_eff(exp(log_liki_half_1), cores = cores)
  r_eff_i2 <- loo::relative_eff(exp(log_liki_half_2), cores = cores)
  r_eff_i <- min(r_eff_i1,r_eff_i2)

  list(
    lwi = lwi_half,
    lwfi = lwfi_half,
    log_liki = log_liki_half,
    r_eff_i = r_eff_i
  )
}

#' Generic function for K-fold cross-validation for developers
#'
#' @description For developers of Bayesian modeling packages, **loo** includes
#'   a generic function `kfold()` so that methods may be defined for K-fold
#'   CV without name conflicts between packages. See, for example, the
#'   `kfold()` methods in the **rstanarm** and **brms** packages.
#'
#'   The **Value** section below describes the objects that `kfold()`
#'   methods should return in order to be compatible with
#'   [loo_compare()] and the **loo** package print methods.
#'
#'
#' @name kfold-generic
#' @param x A fitted model object.
#' @param ... Arguments to pass to specific methods.
#'
#' @return For developers defining a `kfold()` method for a class
#'   `"foo"`, the `kfold.foo()` function should return a list with class
#'   `c("kfold", "loo")` with at least the following named elements:
#'    * `"estimates"`: A `1x2` matrix containing the ELPD estimate and its
#'      standard error. The matrix must have row name "`elpd_kfold`" and column
#'      names `"Estimate"` and `"SE"`.
#'    * `"pointwise"`: A `Nx1` matrix with column name `"elpd_kfold"` containing
#'      the pointwise contributions for each data point.
#'
#'   It is important for the object to have at least these classes and
#'   components so that it is compatible with other functions like
#'   [loo_compare()] and `print()` methods.
#'
NULL

#' @rdname kfold-generic
#' @export
kfold <- function(x, ...) {
  UseMethod("kfold")
}

#' @rdname kfold-generic
#' @export
is.kfold <- function(x) {
  inherits(x, "kfold") && is.loo(x)
}

#' @export
dim.kfold <- function(x) {
  attr(x, "dims")
}

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----eval=FALSE---------------------------------------------------------------
#  library("rstan")
#  
#  # Prepare data
#  url <- "http://stat.columbia.edu/~gelman/arm/examples/arsenic/wells.dat"
#  wells <- read.table(url)
#  wells$dist100 <- with(wells, dist / 100)
#  X <- model.matrix(~ dist100 + arsenic, wells)
#  standata <- list(y = wells$switch, X = X, N = nrow(X), P = ncol(X))
#  
#  # Fit model
#  fit_1 <- stan("logistic.stan", data = standata)
#  print(fit_1, pars = "beta")

## ----eval=FALSE---------------------------------------------------------------
#  library("loo")
#  
#  # Extract pointwise log-likelihood
#  # using merge_chains=FALSE returns an array, which is easier to
#  # use with relative_eff()
#  log_lik_1 <- extract_log_lik(fit_1, merge_chains = FALSE)
#  
#  # as of loo v2.0.0 we can optionally provide relative effective sample sizes
#  # when calling loo, which allows for better estimates of the PSIS effective
#  # sample sizes and Monte Carlo error
#  r_eff <- relative_eff(exp(log_lik_1), cores = 2)
#  
#  # preferably use more than 2 cores (as many cores as possible)
#  # will use value of 'mc.cores' option if cores is not specified
#  loo_1 <- loo(log_lik_1, r_eff = r_eff, cores = 2)
#  print(loo_1)

## ----eval=FALSE---------------------------------------------------------------
#  standata$X[, "arsenic"] <- log(standata$X[, "arsenic"])
#  fit_2 <- stan(fit = fit_1, data = standata)
#  
#  log_lik_2 <- extract_log_lik(fit_2, merge_chains = FALSE)
#  r_eff_2 <- relative_eff(exp(log_lik_2))
#  loo_2 <- loo(log_lik_2, r_eff = r_eff_2, cores = 2)
#  print(loo_2)

## ----eval=FALSE---------------------------------------------------------------
#  # Compare
#  comp <- loo_compare(loo_1, loo_2)

## ----eval=FALSE---------------------------------------------------------------
#  print(comp) # can set simplify=FALSE for more detailed print output

#' Print methods
#'
#' @export
#' @param x An object returned by [loo()], [psis()], or [waic()].
#' @param digits An integer passed to [base::round()].
#' @param plot_k Logical. If `TRUE` the estimates of the Pareto shape
#'   parameter \eqn{k} are plotted. Ignored if `x` was generated by
#'   [waic()]. To just plot \eqn{k} without printing use the
#'   [plot()][pareto-k-diagnostic] method for 'loo' objects.
#' @param ... Arguments passed to [plot.psis_loo()] if `plot_k` is
#'   `TRUE`.
#'
#' @return `x`, invisibly.
#'
#' @seealso [pareto-k-diagnostic]
#'
print.loo <- function(x, digits = 1, ...) {
  cat("\n")
  print_dims(x)
  if (!("estimates" %in% names(x))) {
    x <- convert_old_object(x)
  }
  cat("\n")
  print(.fr(as.data.frame(x$estimates), digits), quote = FALSE)
  return(invisible(x))
}

#' @export
#' @rdname print.loo
print.waic <- function(x, digits = 1, ...) {
  print.loo(x, digits = digits, ...)
  throw_pwaic_warnings(x$pointwise[, "p_waic"], digits = digits, warn = FALSE)
  invisible(x)
}

#' @export
#' @rdname print.loo
print.psis_loo <- function(x, digits = 1, plot_k = FALSE, ...) {
  print.loo(x, digits = digits, ...)
  cat("------\n")
  print_mcse_summary(x, digits = digits)
  S <- dim(x)[1]
  k_threshold <- ps_khat_threshold(S)
  if (length(pareto_k_ids(x, threshold = k_threshold))) {
    cat("\n")
  }
  print(pareto_k_table(x), digits = digits)
  cat(.k_help())
  if (plot_k) {
    graphics::plot(x, ...)
  }
  invisible(x)
}

#' @export
#' @rdname print.loo
print.importance_sampling_loo <- function(x, digits = 1, plot_k = FALSE, ...) {
  print.loo(x, digits = digits, ...)
  cat("------\n")
  invisible(x)
}

#' @export
#' @rdname print.loo
print.psis_loo_ap <- function(x, digits = 1, plot_k = FALSE, ...) {
  print.loo(x, digits = digits, ...)
  cat("------\n")
  cat("Posterior approximation correction used.\n")
  attr(x, 'r_eff') <- 1
  print_mcse_summary(x, digits = digits)
  S <- dim(x)[1]
  k_threshold <- ps_khat_threshold(S)
  if (length(pareto_k_ids(x, threshold = k_threshold))) {
    cat("\n")
  }
  print(pareto_k_table(x), digits = digits)
  cat(.k_help())
  if (plot_k) {
    graphics::plot(x, ...)
  }
  invisible(x)
}


#' @export
#' @rdname print.loo
print.psis <- function(x, digits = 1, plot_k = FALSE, ...) {
  print_dims(x)
  print_reff_summary(x, digits)
  print(pareto_k_table(x), digits = digits)
  cat(.k_help())
  if (plot_k) {
    graphics::plot(x, ...)
  }
  invisible(x)
}

#' @export
#' @rdname print.loo
print.importance_sampling <- function(x, digits = 1, plot_k = FALSE, ...) {
  print_dims(x)
  if (plot_k) {
    graphics::plot(x, ...)
  }
  invisible(x)
}

#' @export
#' @rdname print.loo
print.kfold <- function(x, digits = 1, plot_k = FALSE, ...) {
  print.loo(x, digits = digits, ...)

  if ("diagnostics" %in% names(x)) {
    cat("------\n")
    S <- dim(x)[1]
    k_threshold <- ps_khat_threshold(S)
    if (length(pareto_k_ids(x, threshold = k_threshold))) {
      cat("\n")
    }
    print(pareto_k_table(x), digits = digits)
    cat(.k_help())
  
    if (plot_k) {
      graphics::plot(x, ...)
    }
  }
  return(invisible(x))
  }

# internal ----------------------------------------------------------------

#' Print dimensions of log-likelihood or log-weights matrix
#'
#' @export
#' @keywords internal
#'
#' @param x The object returned by [psis()], [loo()], or [waic()].
#' @param ... Ignored.
print_dims <- function(x, ...) UseMethod("print_dims")

#' @rdname print_dims
#' @export
print_dims.importance_sampling <- function(x, ...) {
  cat(
    "Computed from",
    paste(dim(x), collapse = " by "),
    "log-weights matrix.\n"
  )
}

#' @rdname print_dims
#' @export
print_dims.psis_loo <- function(x, ...) {
  cat(
    "Computed from",
    paste(dim(x), collapse = " by "),
    "log-likelihood matrix.\n"
  )
}

#' @rdname print_dims
#' @export
print_dims.importance_sampling_loo <- function(x, ...) {
  cat(
    "Computed from",
    paste(dim(x), collapse = " by "),
    "log-likelihood matrix using", class(x)[1], ".\n"
  )
}

#' @rdname print_dims
#' @export
print_dims.waic <- function(x, ...) {
  cat(
    "Computed from",
    paste(dim(x), collapse = " by "),
    "log-likelihood matrix.\n"
  )
}

#' @rdname print_dims
#' @export
print_dims.kfold <- function(x, ...) {
  K <- attr(x, "K", exact = TRUE)
  if (!is.null(K)) {
    cat("Based on", paste0(K, "-fold"), "cross-validation.\n")
  }
}

#' @rdname print_dims
#' @export
print_dims.psis_loo_ss <- function(x, ...) {
  cat(
    "Computed from",
    paste(c(dim(x)[1], nobs(x)) , collapse = " by "),
    "subsampled log-likelihood\nvalues from",
    length(x$loo_subsampling$elpd_loo_approx),
    "total observations.\n"
  )
}

print_reff_summary <- function(x, digits) {
  r_eff <- x$diagnostics$r_eff
  if (is.null(r_eff)) {
    if (!is.null(x$psis_object)) {
      r_eff <- attr(x$psis_object,'r_eff')
    } else {
      r_eff <- attr(x,'r_eff')
    }
  }
  if (!is.null(r_eff)) {
    if (all(r_eff==1)) {
      cat(
        "MCSE and ESS estimates assume independent draws (r_eff=1).\n"
      )
    } else {
      cat(paste0(
        "MCSE and ESS estimates assume MCMC draws (r_eff in [",
        .fr(min(r_eff), digits),
        ", ",
        .fr(max(r_eff), digits),
        "]).\n"
      ))
    }
  }
}

print_mcse_summary <- function(x, digits) {
  mcse_val <- mcse_loo(x)
  cat(
    "MCSE of elpd_loo is",
    paste0(.fr(mcse_val, digits), ".\n")
  )
  print_reff_summary(x, digits)
}

# print and warning helpers
.fr <- function(x, digits) format(round(x, digits), nsmall = digits)
.warn <- function(..., call. = FALSE) warning(..., call. = call.)
.k_help <- function() "See help('pareto-k-diagnostic') for details.\n"

# compatibility with old loo objects
convert_old_object <- function(x, digits = 1, ...) {
  z <- x[-grep("pointwise|pareto_k|n_eff", names(x))]
  uz <- unlist(z)
  nms <- names(uz)
  ses <- grepl("se", nms)
  list(estimates = data.frame(Estimate = uz[!ses], SE = uz[ses]))
}

#' Truncated importance sampling (TIS)
#'
#' Implementation of truncated (self-normalized) importance sampling (TIS),
#' truncated at S^(1/2) as recommended by Ionides (2008).
#'
#' @param log_ratios An array, matrix, or vector of importance ratios on the log
#'   scale (for Importance sampling LOO, these are *negative* log-likelihood
#'   values). See the **Methods (by class)** section below for a detailed
#'   description of how to specify the inputs for each method.
#' @template cores
#' @param ... Arguments passed on to the various methods.
#' @param r_eff Vector of relative effective sample size estimates containing
#'   one element per observation. The values provided should be the relative
#'   effective sample sizes of `1/exp(log_ratios)` (i.e., `1/ratios`).
#'   This is related to the relative efficiency of estimating the normalizing
#'   term in self-normalizing importance sampling. See the [relative_eff()]
#'   helper function for computing `r_eff`. If using `psis` with
#'   draws of the `log_ratios` not obtained from MCMC then the warning
#'   message thrown when not specifying `r_eff` can be disabled by
#'   setting `r_eff` to `NA`.
#'
#' @return The `tis()` methods return an object of class `"tis"`,
#'   which is a named list with the following components:
#'
#' \describe{
#'   \item{`log_weights`}{
#'     Vector or matrix of smoothed (and truncated) but *unnormalized* log
#'     weights. To get normalized weights use the
#'     [`weights()`][weights.importance_sampling] method provided for objects of
#'     class `tis`.
#'   }
#'  \item{`diagnostics`}{
#'    A named list containing one vector:
#'    * `pareto_k`: Not used in `tis`, all set to 0.
#'    * `n_eff`: Effective sample size estimates.
#'  }
#' }
#'
#' Objects of class `"tis"` also have the following [attributes][attributes()]:
#' \describe{
#'   \item{`norm_const_log`}{
#'     Vector of precomputed values of `colLogSumExps(log_weights)` that are
#'     used internally by the [weights()]method to normalize the log weights.
#'   }
#'   \item{`r_eff`}{
#'     If specified, the user's `r_eff` argument.
#'   }
#'   \item{`tail_len`}{
#'     Not used for `tis`.
#'   }
#'   \item{`dims`}{
#'     Integer vector of length 2 containing `S` (posterior sample size)
#'     and `N` (number of observations).
#'   }
#'   \item{`method`}{
#'     Method used for importance sampling, here `tis`.
#'   }
#' }
#'
#' @seealso
#' * [psis()] for approximate LOO-CV using PSIS.
#' * [loo()] for approximate LOO-CV.
#' * [pareto-k-diagnostic] for PSIS diagnostics.
#'
#' @references
#' Ionides, Edward L. (2008). Truncated importance sampling.
#' *Journal of Computational and Graphical Statistics* 17(2): 295--311.
#'
#' @examples
#' log_ratios <- -1 * example_loglik_array()
#' r_eff <- relative_eff(exp(-log_ratios))
#' tis_result <- tis(log_ratios, r_eff = r_eff)
#' str(tis_result)
#'
#' # extract smoothed weights
#' lw <- weights(tis_result) # default args are log=TRUE, normalize=TRUE
#' ulw <- weights(tis_result, normalize=FALSE) # unnormalized log-weights
#'
#' w <- weights(tis_result, log=FALSE) # normalized weights (not log-weights)
#' uw <- weights(tis_result, log=FALSE, normalize = FALSE) # unnormalized weights
#'
#' @export
tis <- function(log_ratios, ...) UseMethod("tis")

#' @export
#' @templateVar fn tis
#' @template array
#'
tis.array <-
  function(log_ratios, ...,
           r_eff = NULL,
           cores = getOption("mc.cores", 1)) {
  importance_sampling.array(log_ratios = log_ratios, ...,
                            r_eff = r_eff,
                            cores = cores,
                            method = "tis")
  }

#' @export
#' @templateVar fn tis
#' @template matrix
#'
tis.matrix <-
  function(log_ratios,
           ...,
           r_eff = NULL,
           cores = getOption("mc.cores", 1)) {
    importance_sampling.matrix(log_ratios,
                               ...,
                               r_eff = r_eff,
                               cores = cores,
                               method = "tis")
  }

#' @export
#' @templateVar fn tis
#' @template vector
#'
tis.default <-
  function(log_ratios, ..., r_eff = NULL) {
    importance_sampling.default(log_ratios = log_ratios, ...,
                                r_eff = r_eff, method = "tis")
  }


#' @rdname psis
#' @export
is.tis <- function(x) {
  inherits(x, "tis") && is.list(x)
}


# internal ----------------------------------------------------------------

#' Truncated Importance Sampling on a single vector
#'
#' @noRd
#' @param log_ratios_i A vector of log importance ratios (for `loo()`, negative
#'   log likelihoods).
#' @param ... Not used. Included to conform to PSIS API.
#'
#' @details Implementation of Truncated importance sampling (TIS), a method for
#' stabilizing importance ratios. The version of TIS implemented here
#' corresponds to the algorithm presented in Ionides (2008) with truncation at
#' sqrt(S).
#'
#' @return A named list containing:
#' * `lw`: vector of unnormalized log weights
#' * `pareto_k`: scalar Pareto k estimate. For 'tis', this defaults  to 0.
#'
do_tis_i <- function(log_ratios_i, ...) {
  S <- length(log_ratios_i)
  log_Z <- logSumExp(log_ratios_i) - log(S) # Normalization term, c-hat in Ionides (2008) appendix
  log_cutpoint <- log_Z + 0.5 * log(S)
  lw_i <- pmin(log_ratios_i, log_cutpoint)
  list(log_weights = lw_i, pareto_k = 0)
}

#' Generic (expected) log-predictive density
#'
#' The `elpd()` methods for arrays and matrices can compute the expected log
#' pointwise predictive density for a new dataset or the log pointwise
#' predictive density of the observed data (an overestimate of the elpd).
#'
#' @export
#' @param x A log-likelihood array or matrix. The **Methods (by class)**
#'   section, below, has detailed descriptions of how to specify the inputs for
#'   each method.
#' @param ... Currently ignored.
#'
#' @details The `elpd()` function is an S3 generic and methods are provided for
#'   3-D pointwise log-likelihood arrays and matrices.
#'
#' @seealso The vignette *Holdout validation and K-fold cross-validation of Stan
#'   programs with the loo package* for demonstrations of using the `elpd()`
#'   methods.
#'
#' @examples
#' # Calculate the lpd of the observed data
#' LLarr <- example_loglik_array()
#' elpd(LLarr)
#'
elpd <- function(x, ...) {
  UseMethod("elpd")
}

#' @export
#' @templateVar fn elpd
#' @template array
#'
elpd.array <- function(x, ...) {
  ll <- llarray_to_matrix(x)
  elpd.matrix(ll)
}

#' @export
#' @templateVar fn elpd
#' @template matrix
#'
elpd.matrix <- function(x, ...) {
  pointwise <- pointwise_elpd_calcs(x)
  elpd_object(pointwise, dim(x))
}



# internal ----------------------------------------------------------------
pointwise_elpd_calcs <- function(ll){
  elpd <- colLogSumExps(ll) - log(nrow(ll))
  ic <- -2 * elpd
  cbind(elpd, ic)
}

elpd_object <- function(pointwise, dims) {
  if (!is.matrix(pointwise)) stop("Internal error ('pointwise' must be a matrix)")

  cols_to_summarize <- colnames(pointwise)
  estimates <- table_of_estimates(pointwise[, cols_to_summarize, drop=FALSE])
  out <- nlist(estimates, pointwise)
  structure(
    out,
    dims = dims,
    class = c("elpd_generic", "loo")
  )
}
#' @export
print_dims.elpd_generic <- function(x, ...) {
  cat(
    "Computed from",
    paste(dim(x), collapse = " by "),
    "log-likelihood matrix using the generic elpd function\n"
  )
}

#' Model comparison (deprecated, old version)
#'
#' **This function is deprecated**. Please use the new [loo_compare()] function
#' instead.
#'
#' @export
#' @param ... At least two objects returned by [loo()] (or [waic()]).
#' @param x A list of at least two objects returned by [loo()] (or
#'   [waic()]). This argument can be used as an alternative to
#'   specifying the objects in `...`.
#'
#' @return A vector or matrix with class `'compare.loo'` that has its own
#'   print method. If exactly two objects are provided in `...` or
#'   `x`, then the difference in expected predictive accuracy and the
#'   standard error of the difference are returned. If more than two objects are
#'   provided then a matrix of summary information is returned (see **Details**).
#'
#' @details
#'   When comparing two fitted models, we can estimate the difference in their
#'   expected predictive accuracy by the difference in `elpd_loo` or
#'   `elpd_waic` (or multiplied by -2, if desired, to be on the
#'   deviance scale).
#'
#'   *When that difference, `elpd_diff`, is positive then the expected
#'   predictive accuracy for the second model is higher. A negative
#'   `elpd_diff` favors the first model.*
#'
#'   When using `compare()` with more than two models, the values in the
#'   `elpd_diff` and `se_diff` columns of the returned matrix are
#'   computed by making pairwise comparisons between each model and the model
#'   with the best ELPD (i.e., the model in the first row).
#'   Although the `elpd_diff` column is equal to the difference in
#'   `elpd_loo`, do not expect the `se_diff` column to be equal to the
#'   the difference in `se_elpd_loo`.
#'
#'   To compute the standard error of the difference in ELPD we use a
#'   paired estimate to take advantage of the fact that the same set of _N_
#'   data points was used to fit both models. These calculations should be most
#'   useful when _N_ is large, because then non-normality of the
#'   distribution is not such an issue when estimating the uncertainty in these
#'   sums. These standard errors, for all their flaws, should give a better
#'   sense of uncertainty than what is obtained using the current standard
#'   approach of comparing differences of deviances to a Chi-squared
#'   distribution, a practice derived for Gaussian linear models or
#'   asymptotically, and which only applies to nested models in any case.
#'
#' @template loo-and-psis-references
#'
#' @examples
#' \dontrun{
#' loo1 <- loo(log_lik1)
#' loo2 <- loo(log_lik2)
#' print(compare(loo1, loo2), digits = 3)
#' print(compare(x = list(loo1, loo2)))
#'
#' waic1 <- waic(log_lik1)
#' waic2 <- waic(log_lik2)
#' compare(waic1, waic2)
#' }
#'
compare <- function(..., x = list()) {
  .Deprecated("loo_compare")
  dots <- list(...)
  if (length(dots)) {
    if (length(x)) {
      stop("If 'x' is specified then '...' should not be specified.",
           call. = FALSE)
    }
    nms <- as.character(match.call(expand.dots = TRUE))[-1L]
  } else {
    if (!is.list(x) || !length(x)) {
      stop("'x' must be a list.", call. = FALSE)
    }
    dots <- x
    nms <- names(dots)
    if (!length(nms)) {
      nms <- paste0("model", seq_along(dots))
    }
  }

  if (!all(sapply(dots, is.loo))) {
    stop("All inputs should have class 'loo'.")
  }
  if (length(dots) <= 1L) {
    stop("'compare' requires at least two models.")
  } else if (length(dots) == 2L) {
    loo1 <- dots[[1]]
    loo2 <- dots[[2]]
    comp <- compare_two_models(loo1, loo2)
    class(comp) <- c(class(comp), "old_compare.loo")
    return(comp)
  } else {
    Ns <- sapply(dots, function(x) nrow(x$pointwise))
    if (!all(Ns == Ns[1L])) {
      stop("Not all models have the same number of data points.", call. = FALSE)
    }

    x <- sapply(dots, function(x) {
      est <- x$estimates
      setNames(c(est), nm = c(rownames(est), paste0("se_", rownames(est))) )
    })
    colnames(x) <- nms
    rnms <- rownames(x)
    comp <- x
    ord <- order(x[grep("^elpd", rnms), ], decreasing = TRUE)
    comp <- t(comp)[ord, ]
    patts <- c("elpd", "p_", "^waic$|^looic$", "^se_waic$|^se_looic$")
    col_ord <- unlist(sapply(patts, function(p) grep(p, colnames(comp))),
                      use.names = FALSE)
    comp <- comp[, col_ord]

    # compute elpd_diff and se_elpd_diff relative to best model
    rnms <- rownames(comp)
    diffs <- mapply(elpd_diffs, dots[ord[1]], dots[ord])
    elpd_diff <- apply(diffs, 2, sum)
    se_diff <- apply(diffs, 2, se_elpd_diff)
    comp <- cbind(elpd_diff = elpd_diff, se_diff = se_diff, comp)
    rownames(comp) <- rnms
    class(comp) <- c("compare.loo", class(comp), "old_compare.loo")
    comp
  }
}



# internal ----------------------------------------------------------------
compare_two_models <- function(loo_a, loo_b, return = c("elpd_diff", "se"), check_dims = TRUE) {
  if (check_dims) {
    if (dim(loo_a$pointwise)[1] != dim(loo_b$pointwise)[1]) {
      stop(paste("Models don't have the same number of data points.",
                 "\nFound N_1 =", dim(loo_a$pointwise)[1], "and N_2 =", dim(loo_b$pointwise)[1]), call. = FALSE)
    }
  }

  diffs <- elpd_diffs(loo_a, loo_b)
  comp <- c(elpd_diff = sum(diffs), se = se_elpd_diff(diffs))
  structure(comp, class = "compare.loo")
}

set.seed(123)

LLarr <- example_loglik_array()
LLarr2 <- array(rnorm(prod(dim(LLarr)), c(LLarr), 0.5), dim = dim(LLarr))
LLarr3 <- array(rnorm(prod(dim(LLarr)), c(LLarr), 1), dim = dim(LLarr))
w1 <- suppressWarnings(waic(LLarr))
w2 <- suppressWarnings(waic(LLarr2))

test_that("loo_compare throws appropriate errors", {
  w3 <- suppressWarnings(waic(LLarr[,, -1]))
  w4 <- suppressWarnings(waic(LLarr[,, -(1:2)]))

  expect_error(loo_compare(2, 3), "must be a list if not a 'loo' object")
  expect_error(
    loo_compare(w1, w2, x = list(w1, w2)),
    "If 'x' is a list then '...' should not be specified"
  )
  expect_error(loo_compare(w1, list(1, 2, 3)), "class 'loo'")
  expect_error(loo_compare(w1), "requires at least two models")
  expect_error(loo_compare(x = list(w1)), "requires at least two models")
  expect_error(
    loo_compare(w1, w3),
    "All models must have the same number of observations, but models have inconsistent observation counts: 'model1' (32), 'model2' (31)",
    fixed = TRUE
  )
  expect_error(
    loo_compare(w1, w2, w3),
    "All models must have the same number of observations, but models have inconsistent observation counts: 'model1' (32), 'model2' (32), 'model3' (31)",
    fixed = TRUE
  )
  expect_error(
    loo_compare(x = list("Model A" = w1, "Model B" = w2, "Model C" = w3)),
    "All models must have the same number of observations, but models have inconsistent observation counts: 'Model A' (32), 'Model B' (32), 'Model C' (31)",
    fixed = TRUE
  )
})

test_that("loo_compare throws appropriate warnings", {
  w3 <- w1
  w4 <- w2
  class(w3) <- class(w4) <- c("kfold", "loo")
  attr(w3, "K") <- 2
  attr(w4, "K") <- 3
  expect_warning(
    loo_compare(w3, w4),
    "Not all kfold objects have the same K value"
  )

  class(w4) <- c("psis_loo", "loo")
  attr(w4, "K") <- NULL
  expect_warning(loo_compare(w3, w4), "Comparing LOO-CV to K-fold-CV")

  w3 <- w1
  w4 <- w2
  attr(w3, "yhash") <- "a"
  attr(w4, "yhash") <- "b"
  expect_warning(loo_compare(w3, w4), "Not all models have the same y variable")

  set.seed(123)
  w_list <- lapply(1:25, function(x) {
    suppressWarnings(waic(LLarr + rnorm(1, 0, 0.1)))
  })
  expect_warning(
    loo_compare(w_list),
    "Difference in performance potentially due to chance. See McLatchie and Vehtari (2023) for details.",
    fixed = TRUE
  )

  w_list_short <- lapply(1:4, function(x) {
    suppressWarnings(waic(LLarr + rnorm(1, 0, 0.1)))
  })
  expect_no_warning(loo_compare(w_list_short))
})


comp_colnames <- c(
  "model",
  "elpd_diff",
  "se_diff",
  "p_worse",
  "diag_diff",
  "diag_elpd",
  "elpd_waic",
  "se_elpd_waic",
  "p_waic",
  "se_p_waic",
  "waic",
  "se_waic"
)

test_that("loo_compare returns expected results (2 models)", {
  comp1 <- loo_compare(w1, w1)
  expect_s3_class(comp1, "compare.loo")
  expect_s3_class(comp1, "data.frame")
  expect_equal(colnames(comp1), comp_colnames)
  expect_equal(comp1$model, c("model1", "model2"))
  expect_equal(comp1$elpd_diff, c(0, 0), ignore_attr = TRUE)
  expect_equal(comp1$se_diff, c(0, 0), ignore_attr = TRUE)
  expect_equal(comp1$p_worse, c(NA_real_, NA_real_), ignore_attr = TRUE)
  expect_snapshot_value(comp1, style = "serialize")
  expect_snapshot(print(comp1))

  comp2 <- loo_compare(w1, w2)
  expect_s3_class(comp2, "compare.loo")
  expect_equal(colnames(comp2), comp_colnames)
  expect_equal(comp2$p_worse, c(NA, 1))
  expect_equal(comp2$diag_diff, c("", "N < 100"))
  expect_equal(comp2$diag_elpd, c("", ""))
  expect_snapshot_value(comp2, style = "serialize")
  expect_snapshot(print(comp2))
  expect_snapshot(print(comp2, p_worse = FALSE))

  # specifying objects via ... and via arg x gives equal results
  expect_equal(comp2, loo_compare(x = list(w1, w2)))

  # custom naming works
  comp3 <- loo_compare(x = list("A" = w2, "B" = w1))
  expect_equal(comp3$model, c("B", "A"))
})


test_that("loo_compare returns expected result (3 models)", {
  w3 <- suppressWarnings(waic(LLarr3))
  comp1 <- loo_compare(w1, w2, w3)

  expect_equal(colnames(comp1), comp_colnames)
  expect_equal(comp1$model, c("model1", "model2", "model3"))
  expect_equal(comp1$p_worse, c(NA, 1, 1))
  expect_equal(comp1$diag_diff, c("", "N < 100", "N < 100"))
  expect_s3_class(comp1, "compare.loo")
  expect_s3_class(comp1, "data.frame")
  expect_snapshot_value(comp1, style = "serialize")
  expect_snapshot(print(comp1))

  # specifying objects via '...' gives equivalent results (equal
  # except rownames) to using 'x' argument
  expect_equal(comp1, loo_compare(x = list(w1, w2, w3)), ignore_attr = TRUE)
})

# Tests for deprecated compare() ------------------------------------------

test_that("compare throws deprecation warnings", {
  expect_warning(loo::compare(w1, w2), "Deprecated")
  expect_warning(loo::compare(w1, w1, w2), "Deprecated")
})

test_that("compare returns expected result (2 models)", {
  expect_warning(comp1 <- loo::compare(w1, w1), "Deprecated")
  expect_equal(comp1[1:2], c(elpd_diff = 0, se = 0))

  expect_warning(comp2 <- loo::compare(w1, w2), "Deprecated")
  expect_equal(round(comp2[1:2], 3), c(elpd_diff = -4.057, se = 0.088))
  expect_s3_class(comp2, "old_compare.loo")

  # specifying objects via ... and via arg x gives equal results
  expect_warning(comp_via_list <- loo::compare(x = list(w1, w2)), "Deprecated")
  expect_equal(comp2, comp_via_list)
})

test_that("compare returns expected result (3 models)", {
  w3 <- suppressWarnings(waic(LLarr3))
  expect_warning(comp1 <- loo::compare(w1, w2, w3), "Deprecated")

  expect_equal(
    colnames(comp1),
    c(
      "elpd_diff",
      "se_diff",
      "elpd_waic",
      "se_elpd_waic",
      "p_waic",
      "se_p_waic",
      "waic",
      "se_waic"
    )
  )
  expect_equal(rownames(comp1), c("w1", "w2", "w3"))
  expect_equal(comp1[1, 1], 0)
  expect_s3_class(comp1, "compare.loo")
  expect_s3_class(comp1, "matrix")
  expect_snapshot_value(comp1, style = "serialize")

  # specifying objects via '...' gives equivalent results (equal
  # except rownames) to using 'x' argument
  expect_warning(
    comp_via_list <- loo::compare(x = list(w1, w2, w3)),
    "Deprecated"
  )
  expect_equal(comp1, comp_via_list, ignore_attr = TRUE)
})

test_that("compare throws appropriate errors", {
  expect_error(
    suppressWarnings(loo::compare(w1, w2, x = list(w1, w2))),
    "should not be specified"
  )
  expect_error(suppressWarnings(loo::compare(x = 2)), "must be a list")
  expect_error(
    suppressWarnings(loo::compare(x = list(2))),
    "should have class 'loo'"
  )
  expect_error(
    suppressWarnings(loo::compare(x = list(w1))),
    "requires at least two models"
  )

  w3 <- suppressWarnings(waic(LLarr2[,, -1]))
  expect_error(
    suppressWarnings(loo::compare(x = list(w1, w3))),
    "same number of data points"
  )
  expect_error(
    suppressWarnings(loo::compare(x = list(w1, w2, w3))),
    "same number of data points"
  )
})

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----stancode-----------------------------------------------------------------
# Note: some syntax used in this Stan program requires RStan >= 2.26 (or CmdStanR)
# To use an older version of RStan change the line declaring `y` to: int y[N];
stancode <- "
data {
  int<lower=1> K;
  int<lower=1> N;
  matrix[N,K] x;
  array[N] int y;
  vector[N] offset;

  real beta_prior_scale;
  real alpha_prior_scale;
}
parameters {
  vector[K] beta;
  real intercept;
}
model {
  y ~ poisson(exp(x * beta + intercept + offset));
  beta ~ normal(0,beta_prior_scale);
  intercept ~ normal(0,alpha_prior_scale);
}
generated quantities {
  vector[N] log_lik;
  for (n in 1:N)
    log_lik[n] = poisson_lpmf(y[n] | exp(x[n] * beta + intercept + offset[n]));
}
"

## ----setup, message=FALSE-----------------------------------------------------
library("rstan")
library("loo")
seed <- 9547
set.seed(seed)

## ----modelfit-holdout, message=FALSE------------------------------------------
# Prepare data
data(roaches, package = "rstanarm")
roaches$roach1 <- sqrt(roaches$roach1)
roaches$offset <- log(roaches[,"exposure2"])
# 20% of the data goes to the test set:
roaches$test <- 0
roaches$test[sample(.2 * seq_len(nrow(roaches)))] <- 1
# data to "train" the model
data_train <- list(y = roaches$y[roaches$test == 0],
                   x = as.matrix(roaches[roaches$test == 0,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$test == 0,]),
                   K = 3,
                   offset = roaches$offset[roaches$test == 0],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )
# data to "test" the model
data_test <- list(y = roaches$y[roaches$test == 1],
                   x = as.matrix(roaches[roaches$test == 1,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$test == 1,]),
                   K = 3,
                   offset = roaches$offset[roaches$test == 1],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )

## ----fit-train----------------------------------------------------------------
# Compile
stanmodel <- stan_model(model_code = stancode)
# Fit model
fit <- sampling(stanmodel, data = data_train, seed = seed, refresh = 0)

## ----gen-test-----------------------------------------------------------------
gen_test <- gqs(stanmodel, draws = as.matrix(fit), data= data_test)
log_pd <- extract_log_lik(gen_test)

## ----elpd-holdout-------------------------------------------------------------
(elpd_holdout <- elpd(log_pd))

## ----prepare-folds, message=FALSE---------------------------------------------
# Prepare data
roaches$fold <- kfold_split_random(K = 10, N = nrow(roaches))

## -----------------------------------------------------------------------------
# Prepare a matrix with the number of post-warmup iterations by number of observations:
log_pd_kfold <- matrix(nrow = 4000, ncol = nrow(roaches))
# Loop over the folds
for(k in 1:10){
  data_train <- list(y = roaches$y[roaches$fold != k],
                   x = as.matrix(roaches[roaches$fold != k,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$fold != k,]),
                   K = 3,
                   offset = roaches$offset[roaches$fold != k],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )
  data_test <- list(y = roaches$y[roaches$fold == k],
                   x = as.matrix(roaches[roaches$fold == k,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$fold == k,]),
                   K = 3,
                   offset = roaches$offset[roaches$fold == k],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )
  fit <- sampling(stanmodel, data = data_train, seed = seed, refresh = 0)
  gen_test <- gqs(stanmodel, draws = as.matrix(fit), data= data_test)
  log_pd_kfold[, roaches$fold == k] <- extract_log_lik(gen_test)
}

## ----elpd-kfold---------------------------------------------------------------
(elpd_kfold <- elpd(log_pd_kfold))

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----more-knitr-ops, include=FALSE--------------------------------------------
knitr::opts_chunk$set(
  cache=TRUE,
  message=FALSE, 
  warning=FALSE
)

## ----lpdf, eval=FALSE---------------------------------------------------------
#  /**
#   * Normal log-pdf for spatially lagged responses
#   *
#   * @param y Vector of response values.
#   * @param mu Mean parameter vector.
#   * @param sigma Positive scalar residual standard deviation.
#   * @param rho Positive scalar autoregressive parameter.
#   * @param W Spatial weight matrix.
#   *
#   * @return A scalar to be added to the log posterior.
#   */
#  real normal_lagsar_lpdf(vector y, vector mu, real sigma,
#                          real rho, matrix W) {
#    int N = rows(y);
#    real inv_sigma2 = 1 / square(sigma);
#    matrix[N, N] W_tilde = -rho * W;
#    vector[N] half_pred;
#  
#    for (n in 1:N) W_tilde[n,n] += 1;
#  
#    half_pred = W_tilde * (y - mdivide_left(W_tilde, mu));
#  
#    return 0.5 * log_determinant(crossprod(W_tilde) * inv_sigma2) -
#           0.5 * dot_self(half_pred) * inv_sigma2;
#  }

## ----setup, cache=FALSE-------------------------------------------------------
library("loo")
library("brms")
library("bayesplot")
library("ggplot2")
color_scheme_set("brightblue")
theme_set(theme_default())


SEED <- 10001 
set.seed(SEED) # only sets seed for R (seed for Stan set later)

# loads COL.OLD data frame and COL.nb neighbor list
data(oldcol, package = "spdep") 

## ----data---------------------------------------------------------------------
str(COL.OLD[, c("CRIME", "HOVAL", "INC")])

## ----fit, results="hide"------------------------------------------------------
fit <- brm(
  CRIME ~ INC + HOVAL + sar(COL.nb, type = "lag"), 
  data = COL.OLD,
  data2 = list(COL.nb = COL.nb),
  chains = 4,
  seed = SEED
)

## ----plot-lagsar, message=FALSE-----------------------------------------------
lagsar <- as.matrix(fit, pars = "lagsar")
estimates <- quantile(lagsar, probs = c(0.25, 0.5, 0.75))
mcmc_hist(lagsar) + 
  vline_at(estimates, linetype = 2, size = 1) +
  ggtitle("lagsar: posterior median and 50% central interval")

## ----approx-------------------------------------------------------------------
posterior <- as.data.frame(fit)
y <- fit$data$CRIME
N <- length(y)
S <- nrow(posterior)
loglik <- yloo <- sdloo <- matrix(nrow = S, ncol = N)

for (s in 1:S) {
  p <- posterior[s, ]
  eta <- p$b_Intercept + p$b_INC * fit$data$INC + p$b_HOVAL * fit$data$HOVAL
  W_tilde <- diag(N) - p$lagsar * spdep::nb2mat(COL.nb)
  Cinv <- t(W_tilde) %*% W_tilde / p$sigma^2
  g <- Cinv %*% (y - solve(W_tilde, eta))
  cbar <- diag(Cinv)
  yloo[s, ] <- y - g / cbar
  sdloo[s, ] <- sqrt(1 / cbar)
  loglik[s, ] <- dnorm(y, yloo[s, ], sdloo[s, ], log = TRUE)
}

# use loo for psis smoothing
log_ratios <- -loglik
psis_result <- psis(log_ratios)

## ----plot, cache = FALSE------------------------------------------------------
plot(psis_result, label_points = TRUE)

## ----checklast, cache = FALSE-------------------------------------------------
yloo_sub <- yloo[S, ]
sdloo_sub <- sdloo[S, ]
df <- data.frame(
  y = y, 
  yloo = yloo_sub,
  ymin = yloo_sub - sdloo_sub * 2,
  ymax = yloo_sub + sdloo_sub * 2
)
ggplot(data=df, aes(x = y, y = yloo, ymin = ymin, ymax = ymax)) +
  geom_errorbar(
    width = 1, 
    color = "skyblue3", 
    position = position_jitter(width = 0.25)
  ) +
  geom_abline(color = "gray30", size = 1.2) +
  geom_point()

## ----psisloo------------------------------------------------------------------
(psis_loo <- loo(loglik))

## ----fit_dummy, cache = TRUE--------------------------------------------------
# see help("mi", "brms") for details on the mi() usage
fit_dummy <- brm(
  CRIME | mi() ~ INC + HOVAL + sar(COL.nb, type = "lag"), 
  data = COL.OLD,
  data2 = list(COL.nb = COL.nb),
  chains = 0
)

## ----exact-loo-cv, results="hide", message=FALSE, warning=FALSE, cache = TRUE----
S <- 500
res <- vector("list", N)
loglik <- matrix(nrow = S, ncol = N)
for (i in seq_len(N)) {
  dat_mi <- COL.OLD
  dat_mi$CRIME[i] <- NA
  fit_i <- update(fit_dummy, newdata = dat_mi, 
                  # just for vignette
                  chains = 1, iter = S * 2)
  posterior <- as.data.frame(fit_i)
  yloo <- sdloo <- rep(NA, S)
  for (s in seq_len(S)) {
    p <- posterior[s, ]
    y_miss_i <- y
    y_miss_i[i] <- p$Ymi
    eta <- p$b_Intercept + p$b_INC * fit_i$data$INC + p$b_HOVAL * fit_i$data$HOVAL
    W_tilde <- diag(N) - p$lagsar * spdep::nb2mat(COL.nb)
    Cinv <- t(W_tilde) %*% W_tilde / p$sigma^2
    g <- Cinv %*% (y_miss_i - solve(W_tilde, eta))
    cbar <- diag(Cinv);
    yloo[s] <- y_miss_i[i] - g[i] / cbar[i]
    sdloo[s] <- sqrt(1 / cbar[i])
    loglik[s, i] <- dnorm(y[i], yloo[s], sdloo[s], log = TRUE)
  }
  ypred <- rnorm(S, yloo, sdloo)
  res[[i]] <- data.frame(y = c(posterior$Ymi, ypred))
  res[[i]]$type <- rep(c("pp", "loo"), each = S)
  res[[i]]$obs <- i
}
res <- do.call(rbind, res)

## ----yplots, cache = FALSE, fig.width=10, out.width="95%", fig.asp = 0.3------
res_sub <- res[res$obs %in% 1:4, ]
ggplot(res_sub, aes(y, fill = type)) +
  geom_density(alpha = 0.6) +
  facet_wrap("obs", scales = "fixed", ncol = 4)

## ----loo_exact, cache=FALSE---------------------------------------------------
log_mean_exp <- function(x) {
  # more stable than log(mean(exp(x)))
  max_x <- max(x)
  max_x + log(sum(exp(x - max_x))) - log(length(x))
}
exact_elpds <- apply(loglik, 2, log_mean_exp)
exact_elpd <- sum(exact_elpds)
round(exact_elpd, 1)

## ----compare, fig.height=5----------------------------------------------------
df <- data.frame(
  approx_elpd = psis_loo$pointwise[, "elpd_loo"],
  exact_elpd = exact_elpds
)
ggplot(df, aes(x = approx_elpd, y = exact_elpd)) +
  geom_abline(color = "gray30") +
  geom_point(size = 2) +
  geom_point(data = df[4, ], size = 3, color = "red3") +
  xlab("Approximate elpds") +
  ylab("Exact elpds") +
  coord_fixed(xlim = c(-16, -3), ylim = c(-16, -3))

## ----pt4----------------------------------------------------------------------
without_pt_4 <- c(
  approx = sum(psis_loo$pointwise[-4, "elpd_loo"]),
  exact = sum(exact_elpds[-4])  
)
round(without_pt_4, 1)

## ----brms-stan-code, eval=FALSE-----------------------------------------------
#  // generated with brms 2.2.0
#  functions {
#  /**
#   * Normal log-pdf for spatially lagged responses
#   *
#   * @param y Vector of response values.
#   * @param mu Mean parameter vector.
#   * @param sigma Positive scalar residual standard deviation.
#   * @param rho Positive scalar autoregressive parameter.
#   * @param W Spatial weight matrix.
#   *
#   * @return A scalar to be added to the log posterior.
#   */
#    real normal_lagsar_lpdf(vector y, vector mu, real sigma,
#                            real rho, matrix W) {
#      int N = rows(y);
#      real inv_sigma2 = 1 / square(sigma);
#      matrix[N, N] W_tilde = -rho * W;
#      vector[N] half_pred;
#      for (n in 1:N) W_tilde[n, n] += 1;
#      half_pred = W_tilde * (y - mdivide_left(W_tilde, mu));
#      return 0.5 * log_determinant(crossprod(W_tilde) * inv_sigma2) -
#             0.5 * dot_self(half_pred) * inv_sigma2;
#    }
#  }
#  data {
#    int<lower=1> N;  // total number of observations
#    vector[N] Y;  // response variable
#    int<lower=0> Nmi;  // number of missings
#    int<lower=1> Jmi[Nmi];  // positions of missings
#    int<lower=1> K;  // number of population-level effects
#    matrix[N, K] X;  // population-level design matrix
#    matrix[N, N] W;  // spatial weight matrix
#    int prior_only;  // should the likelihood be ignored?
#  }
#  transformed data {
#    int Kc = K - 1;
#    matrix[N, K - 1] Xc;  // centered version of X
#    vector[K - 1] means_X;  // column means of X before centering
#    for (i in 2:K) {
#      means_X[i - 1] = mean(X[, i]);
#      Xc[, i - 1] = X[, i] - means_X[i - 1];
#    }
#  }
#  parameters {
#    vector[Nmi] Ymi;  // estimated missings
#    vector[Kc] b;  // population-level effects
#    real temp_Intercept;  // temporary intercept
#    real<lower=0> sigma;  // residual SD
#    real<lower=0,upper=1> lagsar;  // SAR parameter
#  }
#  transformed parameters {
#  }
#  model {
#    vector[N] Yl = Y;
#    vector[N] mu = Xc * b + temp_Intercept;
#    Yl[Jmi] = Ymi;
#    // priors including all constants
#    target += student_t_lpdf(temp_Intercept | 3, 34, 17);
#    target += student_t_lpdf(sigma | 3, 0, 17)
#      - 1 * student_t_lccdf(0 | 3, 0, 17);
#    // likelihood including all constants
#    if (!prior_only) {
#      target += normal_lagsar_lpdf(Yl | mu, sigma, lagsar, W);
#    }
#  }
#  generated quantities {
#    // actual population-level intercept
#    real b_Intercept = temp_Intercept - dot_product(means_X, b);
#  }

#' Diagnostics for Pareto smoothed importance sampling (PSIS)
#'
#' Print a diagnostic table summarizing the estimated Pareto shape parameters
#' and PSIS effective sample sizes, find the indexes of observations for which
#' the estimated Pareto shape parameter \eqn{k} is larger than some
#' `threshold` value, or plot observation indexes vs. diagnostic estimates.
#' The **Details** section below provides a brief overview of the
#' diagnostics, but we recommend consulting Vehtari, Gelman, and Gabry (2017)
#' and Vehtari, Simpson, Gelman, Yao, and Gabry (2024) for full details.
#'
#' @name pareto-k-diagnostic
#' @param x An object created by [loo()] or [psis()].
#' @param threshold For `pareto_k_ids()`, `threshold` is the minimum \eqn{k}
#'   value to flag (default is a sample size `S` dependent threshold
#'   `1 - 1 / log10(S)`). For `mcse_loo()`, if any \eqn{k} estimates are
#'   greater than `threshold` the MCSE estimate is returned as `NA`
#'   See **Details** for the motivation behind these defaults.
#'
#' @details
#'
#' The reliability and approximate convergence rate of the PSIS-based
#' estimates can be assessed using the estimates for the shape
#' parameter \eqn{k} of the generalized Pareto distribution. The
#' diagnostic threshold for Pareto \eqn{k} depends on sample size
#' \eqn{S} (sample size dependent threshold was introduced by Vehtari
#' et al. (2024), and before that fixed thresholds of 0.5 and 0.7 were
#' recommended). For simplicity, `loo` package uses the nominal sample
#' size \eqn{S} when computing the sample size specific
#' threshold. This provides an optimistic threshold if the effective
#' sample size is less than 2200, but if MCMC-ESS > S/2 the difference
#' is usually negligible. Thinning of MCMC draws can be used to
#' improve the ratio ESS/S.
#'
#' * If \eqn{k < min(1 - 1 / log10(S), 0.7)}, where \eqn{S} is the
#'   sample size, the PSIS estimate and the corresponding Monte Carlo
#'   standard error estimate are reliable.
#'
#' * If \eqn{1 - 1 / log10(S) <= k < 0.7}, the PSIS estimate and the
#'   corresponding Monte Carlo standard error estimate are not
#'   reliable, but increasing the (effective) sample size \eqn{S} above
#'   2200 may help (this will increase the sample size specific
#'   threshold \eqn{(1-1/log10(2200)>0.7} and then the bias specific
#'   threshold 0.7 dominates).
#'
#' * If \eqn{0.7 <= k < 1}, the PSIS estimate and the corresponding Monte
#'   Carlo standard error have large bias and are not reliable. Increasing
#'   the sample size may reduce the variability in \eqn{k} estimate, which
#'   may result in lower \eqn{k} estimate, too.
#'
#' * If \eqn{k \geq 1}{k >= 1}, the target distribution is estimated to
#'   have a non-finite mean. The PSIS estimate and the corresponding Monte
#'   Carlo standard error are not well defined. Increasing the sample size
#'   may reduce the variability in the \eqn{k} estimate, which
#'   may also result in a lower \eqn{k} estimate.
#'
#' \subsection{What if the estimated tail shape parameter \eqn{k}
#' exceeds the diagnostic threshold?}{ Importance sampling is likely to
#' work less well if the marginal posterior \eqn{p(\theta^s | y)} and
#' LOO posterior \eqn{p(\theta^s | y_{-i})} are very different, which
#' is more likely to happen with a non-robust model and highly
#' influential observations.  If the estimated tail shape parameter
#' \eqn{k} exceeds the diagnostic threshold, the user should be
#' warned. (Note: If \eqn{k} is greater than the diagnostic threshold
#' then WAIC is also likely to fail, but WAIC lacks as accurate
#' diagnostic.)  When using PSIS in the context of approximate LOO-CV,
#' we recommend one of the following actions:
#'
#' * With some additional computations, it is possible to transform
#'   the MCMC draws from the posterior distribution to obtain more
#'   reliable importance sampling estimates. This results in a smaller
#'   shape parameter \eqn{k}.  See [loo_moment_match()] and the
#'   vignette *Avoiding model refits in leave-one-out cross-validation
#'   with moment matching* for an example of this.
#'
#' * Sampling from a leave-one-out mixture distribution (see the
#'   vignette *Mixture IS leave-one-out cross-validation for
#'   high-dimensional Bayesian models*), directly from \eqn{p(\theta^s
#'   | y_{-i})} for the problematic observations \eqn{i}, or using
#'   \eqn{K}-fold cross-validation (see the vignette *Holdout
#'   validation and K-fold cross-validation of Stan programs with the
#'   loo package*) will generally be more stable.
#'
#' * Using a model that is more robust to anomalous observations will
#'   generally make approximate LOO-CV more stable.
#'
#' }
#'
#' \subsection{Observation influence statistics}{ The estimated shape parameter
#' \eqn{k} for each observation can be used as a measure of the observation's
#' influence on posterior distribution of the model. These can be obtained with
#' `pareto_k_influence_values()`.
#' }
#'
#' \subsection{Effective sample size and error estimates}{ In the case that we
#' obtain the samples from the proposal distribution via MCMC the **loo**
#' package also computes estimates for the Monte Carlo error and the effective
#' sample size for importance sampling, which are more accurate for PSIS than
#' for IS and TIS (see Vehtari et al (2024) for details). However, the PSIS
#' effective sample size estimate will be
#' **over-optimistic when the estimate of \eqn{k} is greater than**
#' \eqn{min(1-1/log10(S), 0.7)}, where \eqn{S} is the sample size.
#' }
#'
#' @seealso
#'  * [psis()] for the implementation of the PSIS algorithm.
#'  * The [FAQ page](https://mc-stan.org/loo/articles/online-only/faq.html) on
#'    the __loo__ website for answers to frequently asked questions.
#'
#' @template loo-and-psis-references
#'
NULL

#' @rdname pareto-k-diagnostic
#' @export
#' @return `pareto_k_table()` returns an object of class
#'   `"pareto_k_table"`, which is a matrix with columns `"Count"`,
#'   `"Proportion"`, and `"Min. n_eff"`, and has its own print method.
#'
pareto_k_table <- function(x) {
  k <- pareto_k_values(x)
  n_eff <- try(psis_n_eff_values(x), silent = TRUE)
  if (inherits(n_eff, "try-error")) {
    n_eff <- rep(NA, length(k))
  }

  S <- dim(x)[1]
  k_threshold <- ps_khat_threshold(S)
  kcut <- k_cut(k, k_threshold)
  n_eff[k>k_threshold] <- NA
  min_n_eff <- min_n_eff_by_k(n_eff, kcut)
  count <- table(kcut)
  out <- cbind(
    Count = count,
    Proportion = prop.table(count),
    "Min. n_eff" = min_n_eff
  )
  attr(out, "k_threshold") <- k_threshold
  structure(out, class = c("pareto_k_table", class(out)))
}

#' @export
print.pareto_k_table <- function(x, digits = 1, ...) {
  count <- x[, "Count"]
  k_threshold <- attr(x, "k_threshold")

  if (sum(count[2:3]) == 0) {
    cat(paste0("\nAll Pareto k estimates are good (k < ",
               round(k_threshold,2), ").\n"))
  } else {
    tab <- cbind(
      " " = rep("", 3),
      " " = c("(good)", "(bad)", "(very bad)"),
      "Count" = .fr(count, 0),
      "Pct.   " = paste0(.fr(100 * x[, "Proportion"], digits), "%"),
      # Print ESS as n_eff terms has been deprecated
      "Min. ESS" = round(x[, "Min. n_eff"])
    )
    tab2 <- rbind(tab)
    cat("Pareto k diagnostic values:\n")
    rownames(tab2) <- format(rownames(tab2), justify = "right")
    print(tab2, quote = FALSE)

    invisible(x)
  }
}

#' @rdname pareto-k-diagnostic
#' @export
#' @return `pareto_k_ids()` returns an integer vector indicating which
#' observations have Pareto \eqn{k} estimates above `threshold`.
#'
pareto_k_ids <- function(x, threshold = NULL) {
  if (is.null(threshold)) {
    S <- dim(x)[1]
    threshold <- ps_khat_threshold(S)
  }
  k <- pareto_k_values(x)
  which(k > threshold)
}

#' @rdname pareto-k-diagnostic
#' @export
#' @return `pareto_k_values()` returns a vector of the estimated Pareto
#'   \eqn{k} parameters. These represent the reliability of sampling.
pareto_k_values <- function(x) {
  k <- x$diagnostics[["pareto_k"]]
  if (is.null(k)) {
    # for compatibility with objects from loo < 2.0.0
    k <- x[["pareto_k"]]
  }
  if (is.null(k)) {
    stop("No Pareto k estimates found.", call. = FALSE)
  }
  return(k)
}

#' @rdname pareto-k-diagnostic
#' @export
#' @return `pareto_k_influence_values()` returns a vector of the estimated Pareto
#'   \eqn{k} parameters. These represent influence of the observations on the
#'   model posterior distribution.
pareto_k_influence_values <- function(x) {
  if ("influence_pareto_k" %in% colnames(x$pointwise)) {
    k <- x$pointwise[,"influence_pareto_k"]
  }
  else {
    stop("No Pareto k influence estimates found.", call. = FALSE)
  }
  return(k)
}

#' @rdname pareto-k-diagnostic
#' @export
#' @return `psis_n_eff_values()` returns a vector of the estimated PSIS
#'   effective sample sizes.
psis_n_eff_values <- function(x) {
  n_eff <- x$diagnostics[["n_eff"]]
  if (is.null(n_eff)) {
    # Print ESS as n_eff terms has been deprecated
    stop("No PSIS ESS estimates found.", call. = FALSE)
  }
  return(n_eff)
}

#' @rdname pareto-k-diagnostic
#' @export
#' @return `mcse_loo()` returns the Monte Carlo standard error (MCSE)
#'   estimate for PSIS-LOO. MCSE will be NA if any Pareto \eqn{k} values are
#'   above `threshold`.
#'
mcse_loo <- function(x, threshold = NULL) {
  stopifnot(is.psis_loo(x))
  S <- dim(x)[1]
  if (is.null(threshold)) {
    k_threshold <- ps_khat_threshold(S)
  } else {
    k_threshold <- threshold
  }
  if (any(pareto_k_values(x) > k_threshold, na.rm = TRUE)) {
    return(NA)
  }
  mc_var <- x$pointwise[, "mcse_elpd_loo"]^2
  sqrt(sum(mc_var))
}

#' @rdname pareto-k-diagnostic
#' @aliases plot.loo
#' @export
#' @param label_points,... For the `plot()` method, if `label_points` is
#'   `TRUE` the observation numbers corresponding to any values of \eqn{k}
#'   greater than the diagnostic threshold will be displayed in the plot.
#'   Any arguments specified in `...` will be passed to [graphics::text()]
#'   and can be used to control the appearance of the labels.
#' @param diagnostic For the `plot` method, which diagnostic should be
#'   plotted? The options are `"k"` for Pareto \eqn{k} estimates (the
#'   default), or `"ESS"` or `"n_eff"` for PSIS effective sample size estimates.
#' @param main For the `plot()` method, a title for the plot.
#'
#' @return The `plot()` method is called for its side effect and does not
#'   return anything. If `x` is the result of a call to [loo()]
#'   or [psis()] then `plot(x, diagnostic)` produces a plot of
#'   the estimates of the Pareto shape parameters (`diagnostic = "k"`) or
#'   estimates of the PSIS effective sample sizes (`diagnostic = "ESS"`).
#'
plot.psis_loo <- function(x,
                          diagnostic = c("k", "ESS", "n_eff"),
                          ...,
                          label_points = FALSE,
                          main = "PSIS diagnostic plot") {
  diagnostic <- match.arg(diagnostic)
  k <- pareto_k_values(x)
  k[is.na(k)] <- 0  # FIXME when reloo is changed to make NA k values -Inf
  k_inf <- !is.finite(k)
  if (any(k_inf)) {
    warning(signif(100 * mean(k_inf), 2),
            "% of Pareto k estimates are Inf/NA/NaN and not plotted.")
  }

  if (diagnostic == "ESS" || diagnostic == "n_eff") {
    n_eff <- psis_n_eff_values(x)
  } else {
    n_eff <- NULL
  }
  S <- dim(x)[1]
  k_threshold <- ps_khat_threshold(S)

  plot_diagnostic(
    k = k,
    n_eff = n_eff,
    threshold = k_threshold,
    ...,
    label_points = label_points,
    main = main
  )
}

#' @export
#' @noRd
#' @rdname pareto-k-diagnostic
plot.loo <- plot.psis_loo

#' @export
#' @rdname pareto-k-diagnostic
plot.psis <- function(x, diagnostic = c("k", "ESS", "n_eff"), ...,
                      label_points = FALSE,
                      main = "PSIS diagnostic plot") {
  plot.psis_loo(x, diagnostic = diagnostic, ...,
                label_points = label_points, main = main)
}



# internal ----------------------------------------------------------------

plot_diagnostic <-
  function(k,
           n_eff = NULL,
           threshold = 0.7,
           ...,
           label_points = FALSE,
           main = "PSIS diagnostic plot") {
    use_n_eff <- !is.null(n_eff)
    graphics::plot(
      x = if (use_n_eff) n_eff else k,
      xlab = "Data point",
      # Print ESS as n_eff terms has been deprecated
      ylab = if (use_n_eff) "PSIS ESS" else "Pareto shape k",
      type = "n",
      bty = "l",
      yaxt = "n",
      main = main
    )
    graphics::axis(side = 2, las = 1)

    in_range <- function(x, lb_ub) {
      x >= lb_ub[1L] & x <= lb_ub[2L]
    }

    if (!use_n_eff) {
      krange <- range(k, na.rm = TRUE)
      breaks <- c(0, threshold, 1)
      hex_clrs <- c("#C79999", "#7C0000")
      ltys <- c(3, 2, 1)
      for (j in seq_along(breaks)) {
        val <- breaks[j]
        if (in_range(val, krange))
          graphics::abline(
            h = val,
            col = ifelse(val == 0, "darkgray", hex_clrs[j - 1]),
            lty = ltys[j],
            lwd = 1
          )
      }
    }

    breaks <- c(-Inf, threshold, 1)
    hex_clrs <- c("#6497b1", "#005b96", "#03396c")
    clrs <- ifelse(
      in_range(k, breaks[1:2]),
      hex_clrs[1],
      ifelse(in_range(k, breaks[2:3]), hex_clrs[2], hex_clrs[3])
    )
    if (all(k < threshold) || !label_points) {
      graphics::points(x = if (use_n_eff) n_eff else k,
                       col = clrs, pch = 3, cex = .6)
      return(invisible())
    } else {
      graphics::points(x = which(k < threshold), 
                       y = if (use_n_eff) n_eff[k < threshold] else k[k < threshold],
                       col = clrs[k < threshold], pch = 3, cex = .6)
      sel <- !in_range(k, breaks[1:2])
      dots <- list(...)
      txt_args <- c(
        list(
          x = seq_along(k)[sel],
          y = if (use_n_eff) n_eff[sel] else k[sel],
          labels = seq_along(k)[sel]
        ),
        if (length(dots)) dots
      )
      if (!("adj" %in% names(txt_args))) txt_args$adj <- 2 / 3
      if (!("cex" %in% names(txt_args))) txt_args$cex <- 0.75
      if (!("col" %in% names(txt_args))) txt_args$col <- clrs[sel]

      do.call(graphics::text, txt_args)
    }
  }


#' Convert numeric Pareto k values to a factor variable.
#'
#' @noRd
#' @param k Vector of Pareto k estimates.
#' @return A factor variable (the same length as k) with 3 levels.
#'
k_cut <- function(k, threshold) {
  cut(
    k,
    breaks = c(-Inf, threshold, 1, Inf),
    labels = c(paste0("(-Inf, ", round(threshold,2), "]"),
               paste0("(", round(threshold,2), ", 1]"),
               "(1, Inf)")
  )
}

#' Calculate the minimum PSIS n_eff within groups defined by Pareto k values
#'
#' @noRd
#' @param n_eff Vector of PSIS n_eff estimates.
#' @param kcut Factor returned by the k_cut() function.
#' @return Vector of length `nlevels(kcut)` containing the minimum n_eff within
#'   each k group. If there are no k values in a group the corresponding element
#'   of the returned vector is NA.
min_n_eff_by_k <- function(n_eff, kcut) {
  n_eff_split <- split(n_eff, f = kcut)
  n_eff_split <- sapply(n_eff_split, function(x) {
    # some k groups might be empty.
    # split gives numeric(0) but replace with NA
    if (!length(x)) NA else x
  })
  sapply(n_eff_split, min)
}

#' Pareto-smoothing k-hat threshold
#'
#' Given sample size S computes khat threshold for reliable Pareto
#' smoothed estimate (to have small probability of large error). See
#' section 3.2.4, equation (13). Sample sizes 100, 320, 1000, 2200,
#' 10000 correspond to thresholds 0.5, 0.6, 0.67, 0.7, 0.75. Although
#' with bigger sample size S we can achieve estimates with small
#' probability of large error, it is difficult to get accurate MCSE
#' estimates as the bias starts to dominate when k > 0.7 (see Section 3.2.3).
#' Thus the sample size dependent k-hat threshold is capped at 0.7.
#' @param S sample size
#' @param ... unused
#' @return threshold
#' @noRd
ps_khat_threshold <- function(S, ...) {
  min(1 - 1 / log10(S), 0.7)
}

load(test_path("data-for-tests/test_data_psis_approximate_posterior.rda"))

test_that("Laplace approximation, independent posterior", {
  log_p <- test_data_psis_approximate_posterior$laplace_independent$log_p
  log_g <- test_data_psis_approximate_posterior$laplace_independent$log_q
  ll <- test_data_psis_approximate_posterior$laplace_independent$log_liks
  expect_silent(
    psis_lap <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_lap, "psis")
  expect_lt(pareto_k_values(psis_lap), 0.7)

  expect_silent(
    psis_lap_ll <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = ll,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_lap_ll, "loo")
  expect_true(all(pareto_k_values(psis_lap_ll) < 0.7))
})


test_that("Laplace approximation, correlated posterior", {
  log_p <- test_data_psis_approximate_posterior$laplace_correlated$log_p
  log_g <- test_data_psis_approximate_posterior$laplace_correlated$log_q
  ll <- test_data_psis_approximate_posterior$laplace_correlated$log_liks
  expect_silent(
    psis_lap <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_lap, "psis")
  expect_lt(pareto_k_values(psis_lap), 0.7)

  expect_silent(
    psis_lap_ll <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = ll,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_lap_ll, "loo")
  expect_true(all(pareto_k_values(psis_lap_ll) < 0.7))
})

test_that("Laplace approximation, normal model", {
  log_p <- test_data_psis_approximate_posterior$laplace_normal$log_p
  log_g <- test_data_psis_approximate_posterior$laplace_normal$log_q
  ll <- test_data_psis_approximate_posterior$laplace_normal$log_liks
  expect_no_warning(
    psis_lap <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_lap, "psis")
  expect_gt(pareto_k_values(psis_lap), 0.5)

  expect_warning(
    psis_lap_ll <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = ll,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_lap_ll, "loo")
  expect_true(all(pareto_k_values(psis_lap_ll) > 0.5))
})


test_that("ADVI fullrank approximation, independent posterior", {
  log_p <- test_data_psis_approximate_posterior$fullrank_independent$log_p
  log_g <- test_data_psis_approximate_posterior$fullrank_independent$log_q
  ll <- test_data_psis_approximate_posterior$fullrank_independent$log_liks
  expect_silent(
    psis_advi <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi, "psis")
  expect_lt(pareto_k_values(psis_advi), 0.7)

  expect_silent(
    psis_advi_ll <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = ll,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) < 0.7))
})


test_that("ADVI fullrank approximation, correlated posterior", {
  log_p <- test_data_psis_approximate_posterior$fullrank_correlated$log_p
  log_g <- test_data_psis_approximate_posterior$fullrank_correlated$log_q
  ll <- test_data_psis_approximate_posterior$fullrank_correlated$log_liks
  expect_silent(
    psis_advi <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi, "psis")
  expect_lt(pareto_k_values(psis_advi), 0.7)

  expect_silent(
    psis_advi_ll <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = ll,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) < 0.7))
})

test_that("ADVI fullrank approximation, correlated posterior", {
  log_p <- test_data_psis_approximate_posterior$fullrank_normal$log_p
  log_g <- test_data_psis_approximate_posterior$fullrank_normal$log_q
  ll <- test_data_psis_approximate_posterior$fullrank_normal$log_liks
  expect_warning(
    psis_advi <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi, "psis")
  expect_gt(pareto_k_values(psis_advi), 0.7)

  expect_warning(
    psis_advi_ll <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = ll,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) > 0.7))
})


test_that("ADVI meanfield approximation, independent posterior", {
  log_p <- test_data_psis_approximate_posterior$meanfield_independent$log_p
  log_g <- test_data_psis_approximate_posterior$meanfield_independent$log_q
  ll <- test_data_psis_approximate_posterior$meanfield_independent$log_liks
  expect_silent(
    psis_advi <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi, "psis")
  expect_lt(pareto_k_values(psis_advi), 0.7)

  expect_silent(
    psis_advi_ll <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = ll,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) < 0.7))
})


test_that("ADVI meanfield approximation, correlated posterior", {
  log_p <- test_data_psis_approximate_posterior$meanfield_correlated$log_p
  log_g <- test_data_psis_approximate_posterior$meanfield_correlated$log_q
  ll <- test_data_psis_approximate_posterior$meanfield_correlated$log_liks
  expect_warning(
    psis_advi <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi, "psis")
  expect_gt(pareto_k_values(psis_advi), 0.7)

  expect_warning(
    psis_advi_ll <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = ll,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) > 0.5))
  expect_true(any(pareto_k_values(psis_advi_ll) > 0.7))
})

test_that("ADVI meanfield approximation, normal model", {
  log_p <- test_data_psis_approximate_posterior$meanfield_normal$log_p
  log_g <- test_data_psis_approximate_posterior$meanfield_normal$log_q
  ll <- test_data_psis_approximate_posterior$meanfield_normal$log_liks
  expect_warning(
    psis_advi <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi, "psis")
  expect_gt(pareto_k_values(psis_advi), 0.7)

  expect_warning(
    psis_advi_ll <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = ll,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) > 0.7))
})


test_that("ADVI meanfield approximation, normal model", {
  log_p <- test_data_psis_approximate_posterior$meanfield_normal$log_p
  log_g <- test_data_psis_approximate_posterior$meanfield_normal$log_q
  ll <- test_data_psis_approximate_posterior$meanfield_normal$log_liks
  expect_warning(
    psis_advi <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi, "psis")
  expect_gt(pareto_k_values(psis_advi), 0.7)

  expect_warning(
    psis_advi_ll <-
      psis_approximate_posterior(
        log_p = log_p,
        log_g = log_g,
        log_liks = ll,
        cores = 1,
        save_psis = FALSE
      )
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) > 0.7))
})


test_that("Deprecation of log_q argument", {
  log_p <- test_data_psis_approximate_posterior$laplace_independent$log_p
  log_g <- test_data_psis_approximate_posterior$laplace_independent$log_q
  ll <- test_data_psis_approximate_posterior$laplace_independent$log_liks
  expect_warning(
    psis_lap <-
      loo:::psis_approximate_posterior(
        log_p = log_p,
        log_q = log_g,
        cores = 1,
        save_psis = FALSE
      ),
    regexp = "argument log_q has been changed to log_g"
  )
  expect_s3_class(psis_lap, "psis")
  expect_lt(pareto_k_values(psis_lap), 0.7)
})

set.seed(1414)

LLarr <- example_loglik_array()
waic1 <- suppressWarnings(waic(LLarr))
loo1 <- suppressWarnings(loo(LLarr))
psis1 <- suppressWarnings(psis(-LLarr))
r_eff_arr <- relative_eff(exp(LLarr))
loo1_r_eff <- suppressWarnings(loo(LLarr, r_eff = r_eff_arr))


# plotting ----------------------------------------------------------------
test_that("plot methods don't error", {
  expect_silent(plot(loo1, label_points = FALSE))
  expect_silent(plot(psis1, label_points = TRUE))
  expect_silent(plot(psis1, diagnostic = "n_eff", label_points = FALSE))

  loo1$diagnostics$pareto_k[1] <- 10
  expect_silent(plot(loo1, label_points = TRUE))

  expect_output(print(loo1, plot_k = TRUE))
  expect_output(print(psis1, plot_k = TRUE))
})

test_that("plot methods throw appropriate errors/warnings", {
  expect_error(plot(waic1), regexp = "No Pareto k estimates found")

  loo1$diagnostics$pareto_k[1:5] <- Inf
  psis1$diagnostics$pareto_k[1:5] <- Inf
  expect_warning(
    plot(loo1),
    regexp = "estimates are Inf/NA/NaN and not plotted."
  )
  expect_warning(
    plot(psis1),
    regexp = "estimates are Inf/NA/NaN and not plotted."
  )
})


# printing ----------------------------------------------------------------
test_that("print.waic output is ok", {
  expect_snapshot(print(waic1))
})

test_that("print.psis_loo and print.psis output ok", {
  expect_snapshot(print(psis1))
  expect_snapshot(print(loo1))
  expect_snapshot(print(loo1_r_eff))

  loo1$diagnostics$pareto_k <- psis1$diagnostics$pareto_k <- runif(32, 0, .49)
  expect_output(print(loo1), regexp = "Pareto k estimates are good")
  expect_output(print(psis1), regexp = "Pareto k estimates are good")

  loo1$diagnostics$pareto_k[1] <- psis1$diagnostics$pareto_k[1] <- 0.71
  expect_output(print(loo1), regexp = "Pareto k diagnostic")
  loo1$diagnostics$pareto_k[1] <- psis1$diagnostics$pareto_k[1] <- 1.1
  expect_output(print(loo1), regexp = "Pareto k diagnostic")
})


# pareto_k_[ids,values,table] ---------------------------------------------
test_that("pareto_k_values works for psis_loo and psis objects, errors for waic", {
  kpsis <- pareto_k_values(psis1)
  kloo <- pareto_k_values(loo1)
  expect_identical(kpsis, kloo)
  expect_identical(kpsis, psis1$diagnostics$pareto_k)

  expect_error(pareto_k_values(waic1), "No Pareto k estimates found")
})

test_that("pareto_k_influence_values works for psis_loo objects, errors for psis waic", {
  kloo <- pareto_k_influence_values(loo1)
  kloo2 <- pareto_k_values(loo1)
  expect_identical(kloo, kloo2)

  expect_error(
    pareto_k_influence_values(psis1),
    "No Pareto k influence estimates found"
  )
  expect_error(
    pareto_k_influence_values(waic1),
    "No Pareto k influence estimates found"
  )
})

test_that("pareto_k_ids identifies correct observations", {
  for (j in 1:5) {
    loo1$diagnostics$pareto_k <- psis1$diagnostics$pareto_k <- runif(
      32,
      .25,
      1.25
    )
    expect_identical(
      pareto_k_ids(loo1, threshold = 0.5),
      pareto_k_ids(psis1, threshold = 0.5)
    )
    expect_identical(
      pareto_k_ids(loo1, threshold = 0.5),
      which(pareto_k_values(loo1) > 0.5)
    )
    expect_identical(
      pareto_k_ids(psis1, threshold = 0.7),
      which(pareto_k_values(psis1) > 0.7)
    )
  }
})

test_that("pareto_k_table gives correct output", {
  threshold <- ps_khat_threshold(dim(psis1)[1])
  psis1$diagnostics$pareto_k[1:10] <- runif(10, 0, threshold)
  psis1$diagnostics$pareto_k[11:20] <- runif(10, threshold + 0.01, 0.99)
  psis1$diagnostics$pareto_k[21:32] <- runif(12, 1, 10)
  k <- pareto_k_values(psis1)
  tab <- pareto_k_table(psis1)

  expect_output(print(tab), "Pareto k diagnostic values")
  expect_identical(colnames(tab), c("Count", "Proportion", "Min. n_eff"))
  expect_equal(sum(tab[, "Count"]), length(k))
  expect_equal(sum(tab[, "Proportion"]), 1)

  expect_equal(sum(k <= threshold), tab[1, 1])
  expect_equal(sum(k > threshold & k <= 1), tab[2, 1])
  expect_equal(sum(k > 1), tab[3, 1])

  # if n_eff is NULL
  psis1$diagnostics$n_eff <- NULL
  tab2 <- pareto_k_table(psis1)
  expect_output(print(tab2), "<NA>")
  expect_equal(unname(tab2[, "Min. n_eff"]), rep(NA_real_, 3))

  psis1$diagnostics$pareto_k[1:32] <- 0.4
  expect_output(
    print(pareto_k_table(psis1)),
    paste0("All Pareto k estimates are good (k < ", round(threshold, 2), ")"),
    fixed = TRUE
  )
})


# psis_neff and mcse_loo --------------------------------------------------
test_that("psis_n_eff_values extractor works", {
  n_eff_psis <- psis1$diagnostics$n_eff
  expect_type(n_eff_psis, "double")
  expect_identical(psis_n_eff_values(psis1), n_eff_psis)
  expect_identical(psis_n_eff_values(psis1), psis_n_eff_values(loo1))

  psis1$diagnostics$n_eff <- NULL
  expect_error(psis_n_eff_values(psis1), "No PSIS ESS estimates found")
})

test_that("mcse_loo extractor gives correct value", {
  mcse <- mcse_loo(loo1)
  expect_type(mcse, "double")
  expect_snapshot_value(mcse, style = "serialize")
})

test_that("mcse_loo returns NA when it should", {
  loo1$diagnostics$pareto_k[1] <- 1.5
  mcse <- mcse_loo(loo1)
  expect_equal(mcse, NA)
})

test_that("mcse_loo errors if not psis_loo object", {
  expect_error(mcse_loo(psis1), "psis_loo")
})

# print.loo kfold objects --------------------------------------------------

test_that("print.loo supports kfold with pareto-k diagnostics - calibrated", {
  kfold1 <- readRDS("data-for-tests/kfold-calibrated.Rds")

  expect_snapshot(print(kfold1))
})

test_that("print.loo supports kfold with pareto-k diagnostics - miscalibrated", {
  kfold1 <- readRDS("data-for-tests/kfold-miscalibrated.Rds")
  
  expect_snapshot(print(kfold1))
})

test_that("print.loo supports kfold without pareto-k diagnostics", {
  kfold1 <- readRDS("data-for-tests/kfold-miscalibrated.Rds")
  kfold1$diagnostics <- NULL

  expect_snapshot(print(kfold1))
})

library(loo)

context("psis_approximate_posterior")

load(test_path("data-for-tests/test_data_psis_approximate_posterior.rda"))

test_that("Laplace approximation, independent posterior", {
  log_p <- test_data_psis_approximate_posterior$laplace_independent$log_p
  log_g <- test_data_psis_approximate_posterior$laplace_independent$log_q
  ll <- test_data_psis_approximate_posterior$laplace_independent$log_liks
  expect_silent(
    psis_lap <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_lap, "psis")
  expect_lt(pareto_k_values(psis_lap), 0.7)

  expect_silent(
    psis_lap_ll <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, log_liks = ll , cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_lap_ll, "loo")
  expect_true(all(pareto_k_values(psis_lap_ll) < 0.7))
})


test_that("Laplace approximation, correlated posterior", {
  log_p <- test_data_psis_approximate_posterior$laplace_correlated$log_p
  log_g <- test_data_psis_approximate_posterior$laplace_correlated$log_q
  ll <- test_data_psis_approximate_posterior$laplace_correlated$log_liks
  expect_silent(
    psis_lap <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_lap, "psis")
  expect_lt(pareto_k_values(psis_lap), 0.7)

  expect_silent(
    psis_lap_ll <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, log_liks = ll , cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_lap_ll, "loo")
  expect_true(all(pareto_k_values(psis_lap_ll) < 0.7))
})

test_that("Laplace approximation, normal model", {
  log_p <- test_data_psis_approximate_posterior$laplace_normal$log_p
  log_g <- test_data_psis_approximate_posterior$laplace_normal$log_q
  ll <- test_data_psis_approximate_posterior$laplace_normal$log_liks
  expect_no_warning(
    psis_lap <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_lap, "psis")
  expect_gt(pareto_k_values(psis_lap), 0.5)

  expect_warning(
    psis_lap_ll <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, log_liks = ll , cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_lap_ll, "loo")
  expect_true(all(pareto_k_values(psis_lap_ll) > 0.5))
})



test_that("ADVI fullrank approximation, independent posterior", {
  log_p <- test_data_psis_approximate_posterior$fullrank_independent$log_p
  log_g <- test_data_psis_approximate_posterior$fullrank_independent$log_q
  ll <- test_data_psis_approximate_posterior$fullrank_independent$log_liks
  expect_silent(
    psis_advi <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi, "psis")
  expect_lt(pareto_k_values(psis_advi), 0.7)

  expect_silent(
    psis_advi_ll <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, log_liks = ll , cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) < 0.7))
})


test_that("ADVI fullrank approximation, correlated posterior", {
  log_p <- test_data_psis_approximate_posterior$fullrank_correlated$log_p
  log_g <- test_data_psis_approximate_posterior$fullrank_correlated$log_q
  ll <- test_data_psis_approximate_posterior$fullrank_correlated$log_liks
  expect_silent(
    psis_advi <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi, "psis")
  expect_lt(pareto_k_values(psis_advi), 0.7)

  expect_silent(
    psis_advi_ll <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, log_liks = ll , cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) < 0.7))
})

test_that("ADVI fullrank approximation, correlated posterior", {
  log_p <- test_data_psis_approximate_posterior$fullrank_normal$log_p
  log_g <- test_data_psis_approximate_posterior$fullrank_normal$log_q
  ll <- test_data_psis_approximate_posterior$fullrank_normal$log_liks
  expect_warning(
    psis_advi <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi, "psis")
  expect_gt(pareto_k_values(psis_advi), 0.7)

  expect_warning(
    psis_advi_ll <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, log_liks = ll , cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) > 0.7))
})


test_that("ADVI meanfield approximation, independent posterior", {
  log_p <- test_data_psis_approximate_posterior$meanfield_independent$log_p
  log_g <- test_data_psis_approximate_posterior$meanfield_independent$log_q
  ll <- test_data_psis_approximate_posterior$meanfield_independent$log_liks
  expect_silent(
    psis_advi <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi, "psis")
  expect_lt(pareto_k_values(psis_advi), 0.7)

  expect_silent(
    psis_advi_ll <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, log_liks = ll , cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) < 0.7))
})


test_that("ADVI meanfield approximation, correlated posterior", {
  log_p <- test_data_psis_approximate_posterior$meanfield_correlated$log_p
  log_g <- test_data_psis_approximate_posterior$meanfield_correlated$log_q
  ll <- test_data_psis_approximate_posterior$meanfield_correlated$log_liks
  expect_warning(
    psis_advi <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi, "psis")
  expect_gt(pareto_k_values(psis_advi), 0.7)

  expect_warning(
    psis_advi_ll <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, log_liks = ll , cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) > 0.5))
  expect_true(any(pareto_k_values(psis_advi_ll) > 0.7))
})

test_that("ADVI meanfield approximation, normal model", {
  log_p <- test_data_psis_approximate_posterior$meanfield_normal$log_p
  log_g <- test_data_psis_approximate_posterior$meanfield_normal$log_q
  ll <- test_data_psis_approximate_posterior$meanfield_normal$log_liks
  expect_warning(
    psis_advi <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi, "psis")
  expect_gt(pareto_k_values(psis_advi), 0.7)

  expect_warning(
    psis_advi_ll <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, log_liks = ll , cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) > 0.7))
})


test_that("ADVI meanfield approximation, normal model", {

  log_p <- test_data_psis_approximate_posterior$meanfield_normal$log_p
  log_g <- test_data_psis_approximate_posterior$meanfield_normal$log_q
  ll <- test_data_psis_approximate_posterior$meanfield_normal$log_liks
  expect_warning(
    psis_advi <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi, "psis")
  expect_gt(pareto_k_values(psis_advi), 0.7)

  expect_warning(
    psis_advi_ll <-
      psis_approximate_posterior(log_p = log_p, log_g = log_g, log_liks = ll , cores = 1, save_psis = FALSE)
  )
  expect_s3_class(psis_advi_ll, "loo")
  expect_true(all(pareto_k_values(psis_advi_ll) > 0.7))
})


test_that("Deprecation of log_q argument", {

  log_p <- test_data_psis_approximate_posterior$laplace_independent$log_p
  log_g <- test_data_psis_approximate_posterior$laplace_independent$log_q
  ll <- test_data_psis_approximate_posterior$laplace_independent$log_liks
  expect_warning(
    psis_lap <-
      loo:::psis_approximate_posterior(log_p = log_p, log_q = log_g, cores = 1, save_psis = FALSE)
    , regexp = "argument log_q has been changed to log_g"
  )
  expect_s3_class(psis_lap, "psis")
  expect_lt(pareto_k_values(psis_lap), 0.7)
})

library(loo)
set.seed(1414)

context("print, plot, diagnostics")

LLarr <- example_loglik_array()
waic1 <- suppressWarnings(waic(LLarr))
loo1 <- suppressWarnings(loo(LLarr))
psis1 <- suppressWarnings(psis(-LLarr))
r_eff_arr <- relative_eff(exp(LLarr))
loo1_r_eff <- suppressWarnings(loo(LLarr, r_eff = r_eff_arr))


# plotting ----------------------------------------------------------------
test_that("plot methods don't error", {
  expect_silent(plot(loo1, label_points = FALSE))
  expect_silent(plot(psis1, label_points = TRUE))
  expect_silent(plot(psis1, diagnostic = "n_eff", label_points = FALSE))

  loo1$diagnostics$pareto_k[1] <- 10
  expect_silent(plot(loo1, label_points = TRUE))

  expect_output(print(loo1, plot_k = TRUE))
  expect_output(print(psis1, plot_k = TRUE))
})

test_that("plot methods throw appropriate errors/warnings", {
  expect_error(plot(waic1), regexp = "No Pareto k estimates found")

  loo1$diagnostics$pareto_k[1:5] <- Inf
  psis1$diagnostics$pareto_k[1:5] <- Inf
  expect_warning(plot(loo1), regexp = "estimates are Inf/NA/NaN and not plotted.")
  expect_warning(plot(psis1), regexp = "estimates are Inf/NA/NaN and not plotted.")
})



# printing ----------------------------------------------------------------
lldim_msg <- paste0("Computed from ", prod(dim(LLarr)[1:2]) , " by ",
                    dim(LLarr)[3], " log-likelihood matrix")
lwdim_msg <- paste0("Computed from ", prod(dim(LLarr)[1:2]) , " by ",
                    dim(LLarr)[3], " log-weights matrix")

test_that("print.waic output is ok",{
  expect_output(print(waic1), lldim_msg)
  expect_output(print(waic1),
    "p_waic estimates greater than 0.4. We recommend trying loo instead."
  )
})

test_that("print.psis_loo and print.psis output ok",{
  expect_output(print(psis1), lwdim_msg)
  expect_output(print(psis1), "Pareto k estimates are good")
  expect_output(print(loo1), lldim_msg)
  expect_output(print(loo1), "MCSE and ESS estimates assume independent draws")
  expect_output(print(loo1_r_eff), "MCSE and ESS estimates assume MCMC draws")
  expect_output(print(loo1), "Pareto k estimates are good")

  loo1$diagnostics$pareto_k <- psis1$diagnostics$pareto_k <- runif(32, 0, .49)
  expect_output(print(loo1), regexp = "Pareto k estimates are good")
  expect_output(print(psis1), regexp = "Pareto k estimates are good")

  loo1$diagnostics$pareto_k[1] <- psis1$diagnostics$pareto_k[1] <- 0.71
  expect_output(print(loo1), regexp = "Pareto k diagnostic")
  loo1$diagnostics$pareto_k[1] <- psis1$diagnostics$pareto_k[1] <- 1.1
  expect_output(print(loo1), regexp = "Pareto k diagnostic")
})


# pareto_k_[ids,values,table] ---------------------------------------------
test_that("pareto_k_values works for psis_loo and psis objects, errors for waic", {
  kpsis <- pareto_k_values(psis1)
  kloo <- pareto_k_values(loo1)
  expect_identical(kpsis, kloo)
  expect_identical(kpsis, psis1$diagnostics$pareto_k)

  expect_error(pareto_k_values(waic1), "No Pareto k estimates found")
})

test_that("pareto_k_influence_values works for psis_loo objects, errors for psis waic", {
  kloo <- pareto_k_influence_values(loo1)
  kloo2 <- pareto_k_values(loo1)
  expect_identical(kloo, kloo2)

  expect_error(pareto_k_influence_values(psis1), "No Pareto k influence estimates found")
  expect_error(pareto_k_influence_values(waic1), "No Pareto k influence estimates found")
})

test_that("pareto_k_ids identifies correct observations", {
  for (j in 1:5) {
    loo1$diagnostics$pareto_k <- psis1$diagnostics$pareto_k <- runif(32, .25, 1.25)
    expect_identical(
      pareto_k_ids(loo1, threshold = 0.5),
      pareto_k_ids(psis1, threshold = 0.5)
    )
    expect_identical(
      pareto_k_ids(loo1, threshold = 0.5),
      which(pareto_k_values(loo1) > 0.5)
    )
    expect_identical(
      pareto_k_ids(psis1, threshold = 0.7),
      which(pareto_k_values(psis1) > 0.7)
    )
  }
})

test_that("pareto_k_table gives correct output", {
  threshold <- ps_khat_threshold(dim(psis1)[1])
  psis1$diagnostics$pareto_k[1:10] <- runif(10, 0, threshold)
  psis1$diagnostics$pareto_k[11:20] <- runif(10, threshold+0.01, 0.99)
  psis1$diagnostics$pareto_k[21:32] <- runif(12, 1, 10)
  k <- pareto_k_values(psis1)
  tab <- pareto_k_table(psis1)

  expect_output(print(tab), "Pareto k diagnostic values")
  expect_identical(colnames(tab), c("Count", "Proportion", "Min. n_eff"))
  expect_equal(sum(tab[, "Count"]), length(k))
  expect_equal(sum(tab[, "Proportion"]), 1)

  expect_equal(sum(k <= threshold), tab[1,1])
  expect_equal(sum(k > threshold & k <= 1), tab[2,1])
  expect_equal(sum(k > 1), tab[3,1])

  # if n_eff is NULL
  psis1$diagnostics$n_eff <- NULL
  tab2 <- pareto_k_table(psis1)
  expect_output(print(tab2), "<NA>")
  expect_equal(unname(tab2[, "Min. n_eff"]), rep(NA_real_, 3))

  psis1$diagnostics$pareto_k[1:32] <- 0.4
  expect_output(print(pareto_k_table(psis1)),
                paste0("All Pareto k estimates are good (k < ", round(threshold,2), ")"),
                fixed = TRUE)
})



# psis_neff and mcse_loo --------------------------------------------------
test_that("psis_n_eff_values extractor works", {
  n_eff_psis <- psis1$diagnostics$n_eff
  expect_type(n_eff_psis, "double")
  expect_identical(psis_n_eff_values(psis1), n_eff_psis)
  expect_identical(psis_n_eff_values(psis1), psis_n_eff_values(loo1))

  psis1$diagnostics$n_eff <- NULL
  expect_error(psis_n_eff_values(psis1), "No PSIS ESS estimates found")
})

test_that("mcse_loo extractor gives correct value", {
  mcse <- mcse_loo(loo1)
  expect_type(mcse, "double")
  expect_equal_to_reference(mcse, "reference-results/mcse_loo.rds")
})

test_that("mcse_loo returns NA when it should", {
  loo1$diagnostics$pareto_k[1] <- 1.5
  mcse <- mcse_loo(loo1)
  expect_equal(mcse, NA)
})

test_that("mcse_loo errors if not psis_loo object", {
  expect_error(mcse_loo(psis1), "psis_loo")
})

#' A parent class for different importance sampling methods.
#'
#' @inheritParams psis
#' @param method The importance sampling method to use. The following methods
#'   are implemented:
#' * [`"psis"`][psis]: Pareto-Smoothed Importance Sampling (PSIS). Default method.
#' * [`"tis"`][tis]: Truncated Importance Sampling (TIS) with truncation at
#'   `sqrt(S)`, where `S` is the number of posterior draws.
#' * [`"sis"`][sis]: Standard Importance Sampling (SIS).
#'
importance_sampling <- function(log_ratios, method, ...) {
  UseMethod("importance_sampling")
}


#' @rdname importance_sampling
#' @inheritParams psis
#' @export
importance_sampling.array <-
  function(log_ratios, method,
           ...,
           r_eff = 1,
           cores = getOption("mc.cores", 1)) {
    cores <- loo_cores(cores)
    stopifnot(length(dim(log_ratios)) == 3)
    assert_importance_sampling_method_is_implemented(method)
    log_ratios <- validate_ll(log_ratios)
    log_ratios <- llarray_to_matrix(log_ratios)
    r_eff <- prepare_psis_r_eff(r_eff, len = ncol(log_ratios))
    do_importance_sampling(log_ratios, r_eff = r_eff, cores = cores, method = method)
  }

#' @rdname importance_sampling
#' @inheritParams psis
#' @export
importance_sampling.matrix <-
  function(log_ratios, method,
           ...,
           r_eff = 1,
           cores = getOption("mc.cores", 1)) {
    cores <- loo_cores(cores)
    assert_importance_sampling_method_is_implemented(method)
    log_ratios <- validate_ll(log_ratios)
    r_eff <- prepare_psis_r_eff(r_eff, len = ncol(log_ratios))
    do_importance_sampling(log_ratios, r_eff = r_eff, cores = cores, method = method)
  }

#' @rdname importance_sampling
#' @inheritParams psis
#' @export
importance_sampling.default <-
  function(log_ratios, method, ..., r_eff = 1) {
    stopifnot(is.null(dim(log_ratios)) || length(dim(log_ratios)) == 1)
    assert_importance_sampling_method_is_implemented(method)
    dim(log_ratios) <- c(length(log_ratios), 1)
    r_eff <- prepare_psis_r_eff(r_eff, len = 1)
    importance_sampling.matrix(log_ratios, r_eff = r_eff, cores = 1, method = method)
  }


#' @export
dim.importance_sampling <- function(x) {
  attr(x, "dims")
}


#' Extract importance sampling weights
#'
#' @export
#' @export weights.importance_sampling
#' @method weights importance_sampling
#' @param object An object returned by [psis()], [tis()], or [sis()].
#' @param log Should the weights be returned on the log scale? Defaults to
#'   `TRUE`.
#' @param normalize Should the weights be normalized? Defaults to `TRUE`.
#' @param ... Ignored.
#'
#' @return The `weights()` method returns an object with the same dimensions as
#'   the `log_weights` component of `object`. The `normalize` and `log`
#'   arguments control whether the returned weights are normalized and whether
#'   or not to return them on the log scale.
#'
#' @examples
#' # See the examples at help("psis")
#'
weights.importance_sampling <-
  function(object,
           ...,
           log = TRUE,
           normalize = TRUE) {
    out <- object[["log_weights"]] # smoothed but unnormalized log weights
    if (normalize) {
      out <- sweep(out,
                   MARGIN = 2,
                   STATS = attr(object, "norm_const_log"), # colLogSumExp(log_weights)
                   check.margin = FALSE)
    }
    if (!log) {
      out <- exp(out)
    }

    return(out)
  }

# internal ----------------------------------------------------------------

#' Validate selected importance sampling method
#' @noRd
#' @keywords internal
#' @description
#' Currently implemented importance sampling methods
assert_importance_sampling_method_is_implemented <- function(x){
  if (!x %in% implemented_is_methods()) {
    stop("Importance sampling method '",
         x,
         "' is not implemented. Implemented methods: '",
         paste0(implemented_is_methods(), collapse = "', '"),
         "'")
  }
}
implemented_is_methods <- function() c("psis", "tis", "sis")


#' Structure the object returned by the importance_sampling methods
#'
#' @noRd
#' @param unnormalized_log_weights Smoothed and possibly truncated log weights,
#'   but unnormalized.
#' @param pareto_k Vector of GPD k estimates.
#' @param tail_len Vector of tail lengths used to fit GPD.
#' @param r_eff Vector of relative MCMC ESS (n_eff) for `exp(log lik)`
#' @template is_method
#' @return A list of class `"psis"` with structure described in the main doc at
#'   the top of this file.
#'
importance_sampling_object <-
  function(unnormalized_log_weights,
           pareto_k,
           tail_len,
           r_eff,
           method) {
    stopifnot(is.matrix(unnormalized_log_weights))
    methods <- unique(method)
    stopifnot(all(methods %in% implemented_is_methods()))
    if (length(methods) == 1) {
      method <- methods
      classes <- c(tolower(method), "importance_sampling", "list")
    } else {
      classes <- c("importance_sampling", "list")
    }

    norm_const_log <- matrixStats::colLogSumExps(unnormalized_log_weights)
    out <- structure(
      list(
        log_weights = unnormalized_log_weights,
        diagnostics = list(pareto_k = pareto_k, n_eff = NULL, r_eff = r_eff)
      ),
      # attributes
      norm_const_log = norm_const_log,
      tail_len = tail_len,
      r_eff = r_eff,
      dims = dim(unnormalized_log_weights),
      method = method,
      class = classes
    )

    # need normalized weights (not on log scale) for psis_n_eff
    w <- weights(out, normalize = TRUE, log = FALSE)
    out$diagnostics[["n_eff"]] <- psis_n_eff(w, r_eff)
    return(out)
  }

#' Do importance sampling given matrix of log weights
#'
#' @noRd
#' @param lr Matrix of log ratios (`-loglik`)
#' @param r_eff Vector of relative effective sample sizes
#' @param cores User's integer `cores` argument
#' @return A list with class `"psis"` and structure described in the main doc at
#'   the top of this file.
#'
do_importance_sampling <- function(log_ratios, r_eff, cores, method) {
  stopifnot(cores == as.integer(cores))
  assert_importance_sampling_method_is_implemented(method)
  N <- ncol(log_ratios)
  S <- nrow(log_ratios)
  k_threshold <- ps_khat_threshold(S)
  tail_len <- n_pareto(r_eff, S)

  if (method == "psis") {
    is_fun <- do_psis_i
    throw_tail_length_warnings(tail_len)
  } else if (method == "tis") {
    is_fun <- do_tis_i
  } else if (method == "sis") {
    is_fun <- do_sis_i
  } else {
    stop("Incorrect IS method.")
  }

  if (cores == 1) {
    lw_list <- lapply(seq_len(N), function(i)
      is_fun(log_ratios_i = log_ratios[, i], tail_len_i = tail_len[i]))
  } else {
    if (!os_is_windows()) {
      lw_list <- parallel::mclapply(
        X = seq_len(N),
        mc.cores = cores,
        FUN = function(i)
          is_fun(log_ratios_i = log_ratios[, i], tail_len_i = tail_len[i])
      )
    } else {
      cl <- parallel::makePSOCKcluster(cores)
      on.exit(parallel::stopCluster(cl))
      lw_list <-
        parallel::parLapply(
          cl = cl,
          X = seq_len(N),
          fun = function(i)
            is_fun(log_ratios_i = log_ratios[, i], tail_len_i = tail_len[i])
        )
    }
  }

  log_weights <- psis_apply(lw_list, "log_weights", fun_val = numeric(S))
  pareto_k <- psis_apply(lw_list, "pareto_k")
  throw_pareto_warnings(pareto_k, k_threshold)

  importance_sampling_object(
    unnormalized_log_weights = log_weights,
    pareto_k = pareto_k,
    tail_len = tail_len,
    r_eff = r_eff,
    method = rep(method, length(pareto_k)) # Conform to other attr that exist per obs.
  )
}

#' Datasets for loo examples and vignettes
#'
#' Small datasets for use in **loo** examples and vignettes. The `Kline`
#' and `milk` datasets are also included in the **rethinking** package
#' (McElreath, 2016a), but we include them here as **rethinking** is not
#' on CRAN.
#'
#' @name loo-datasets
#' @aliases Kline milk voice voice_loo
#'
#' @details
#' Currently the data sets included are:
#' * `Kline`:
#'   Small dataset from Kline and Boyd (2010) on tool complexity and demography
#'   in Oceanic islands societies. This data is discussed in detail in
#'   McElreath (2016a,2016b). [(Link to variable descriptions)](https://www.rdocumentation.org/packages/rethinking/versions/1.59/topics/Kline)
#' * `milk`:
#'   Small dataset from Hinde and Milligan (2011) on primate milk
#'   composition.This data is discussed in detail in McElreath (2016a,2016b).
#'   [(Link to variable descriptions)](https://www.rdocumentation.org/packages/rethinking/versions/1.59/topics/milk)
#' * `voice`:
#'    Voice rehabilitation data from Tsanas et al. (2014).
#' @references
#' Hinde and Milligan. 2011. *Evolutionary Anthropology* 20:9-23.
#'
#' Kline, M.A. and R. Boyd. 2010. *Proc R Soc B* 277:2559-2564.
#'
#' McElreath, R. (2016a). rethinking: Statistical Rethinking book package.
#' R package version 1.59.
#'
#' McElreath, R. (2016b). *Statistical rethinking: A Bayesian course with
#' examples in R and Stan*. Chapman & Hall/CRC.
#'
#' A. Tsanas, M.A. Little, C. Fox, L.O. Ramig: Objective automatic assessment of
#' rehabilitative speech treatment in Parkinson's disease, IEEE
#' Transactions on Neural Systems and Rehabilitation Engineering, Vol. 22, pp.
#' 181-190, January 2014
#'
#' @examples
#' str(Kline)
#' str(milk)
#'
NULL

#' Diagnostics for Laplace and ADVI approximations and Laplace-loo and ADVI-loo
#'
#' @param log_p The log-posterior (target) evaluated at S samples from the
#'   proposal distribution (g). A vector of length S.
#' @param log_g The log-density (proposal) evaluated at S samples from the
#'   proposal distribution (g). A vector of length S.
#' @param log_q Deprecated argument name (the same as log_g).
#' @param log_liks A log-likelihood matrix of size S * N, where N is the number
#'   of observations and S is the number of samples from q. See
#'   [loo.matrix()] for details. Default is `NULL`. Then only the
#'   posterior is evaluated using the k_hat diagnostic.
#' @inheritParams  loo
#'
#' @return
#' If log likelihoods are supplied, the function returns a `"loo"` object,
#' otherwise the function returns a `"psis"` object.
#'
#' @seealso [loo()] and [psis()]
#'
#' @template loo-and-psis-references
#'
#' @keywords internal
#'
psis_approximate_posterior <- function(log_p = NULL, log_g = NULL, log_liks = NULL,
                                       cores, save_psis, ..., log_q = NULL) {
  if (!is.null(log_q)) {
    .Deprecated(msg = "psis_approximate_posterior() argument log_q has been changed to log_g")
    log_g <- log_q
  }
  checkmate::assert_numeric(log_p, any.missing = FALSE, len = length(log_g), null.ok = FALSE)
  checkmate::assert_numeric(log_g, any.missing = FALSE, len = length(log_p), null.ok = FALSE)
  checkmate::assert_matrix(log_liks, null.ok = TRUE, nrows = length(log_p))
  checkmate::assert_integerish(cores)
  checkmate::assert_flag(save_psis)

  if (is.null(log_liks)) {
    approx_correction <- log_p - log_g
    # Handle underflow/overflow
    approx_correction <- approx_correction - max(approx_correction)
    log_ratios <- matrix(approx_correction, ncol = 1)
  } else {
    log_ratios <- correct_log_ratios(log_ratios = -log_liks, log_p = log_p, log_g = log_g)
  }
  psis_out <- psis.matrix(log_ratios, cores = cores, r_eff = rep(1, ncol(log_ratios)))

  if (is.null(log_liks)) {
    return(psis_out)
  }

  pointwise <- pointwise_loo_calcs(log_liks, psis_out)
  importance_sampling_loo_object(
    pointwise = pointwise,
    diagnostics = psis_out$diagnostics,
    dims = dim(psis_out),
    is_method = "psis",
    is_object = if (save_psis) psis_out else NULL
  )
}



#' Correct log ratios for posterior approximations
#'
#' @inheritParams psis_approximate_posterior
#' @inheritParams ap_psis
#' @noRd
#' @keywords internal
correct_log_ratios <- function(log_ratios, log_p, log_g) {
  approx_correction <- log_p - log_g
  log_ratios <- log_ratios + approx_correction
  # Handle underflow/overflow
  log_ratio_max <- apply(log_ratios, 2, max)
  log_ratios <- sweep(log_ratios, MARGIN = 2, STATS = log_ratio_max)

  log_ratios
}


#' Pareto smoothed importance sampling (PSIS)
#' using approximate posteriors
#' @inheritParams psis_approximate_posterior
#' @param log_ratios The log-likelihood ratios (ie -log_liks)
#' @param ... Currently not in use.
ap_psis <- function(log_ratios, log_p, log_g, ...) {
  UseMethod("ap_psis")
}

#' @export
#' @templateVar fn ap_psis
#' @template array
#'
ap_psis.array <-
  function(log_ratios, log_p, log_g, ...,
           cores = getOption("mc.cores", 1)) {
    cores <- loo_cores(cores)
    stopifnot(length(dim(log_ratios)) == 3)
    log_ratios <- validate_ll(log_ratios)
    log_ratios <- llarray_to_matrix(log_ratios)

    r_eff <- prepare_psis_r_eff(r_eff, len = ncol(log_ratios))
    ap_psis.matrix(log_ratios = log_ratios,
                   log_p = log_p,
                   log_g = log_g,
                   cores = 1)
  }

#' @export
#' @templateVar fn ap_psis
#' @template matrix
#'
ap_psis.matrix <- function(log_ratios, log_p, log_g,
           ...,
           cores = getOption("mc.cores", 1)) {
    checkmate::assert_numeric(log_p, len = nrow(log_ratios))
    checkmate::assert_numeric(log_g, len = nrow(log_ratios))
    cores <- loo_cores(cores)
    log_ratios <- validate_ll(log_ratios)

    log_ratios <- correct_log_ratios(log_ratios, log_p = log_p, log_g = log_g)

    do_psis(log_ratios, r_eff = rep(1, ncol(log_ratios)), cores = cores)
  }

#' @export
#' @templateVar fn ap_psis
#' @template vector
#'
ap_psis.default <- function(log_ratios, log_p, log_g, ...) {
    stopifnot(is.null(dim(log_ratios)) || length(dim(log_ratios)) == 1)
    dim(log_ratios) <- c(length(log_ratios), 1)
    warning("llfun values do not return a matrix, coerce to matrix")
    ap_psis.matrix(as.matrix(log_ratios), log_p, log_g, cores = 1)
  }

#' Compute weighted expectations
#'
#' The `E_loo()` function computes weighted expectations (means, variances,
#' quantiles) using the importance weights obtained from the [PSIS][psis()]
#' smoothing procedure. The expectations estimated by the `E_loo()` function
#' assume that the PSIS approximation is working well.
#' **A small [Pareto k][pareto-k-diagnostic] estimate is necessary,
#' but not sufficient, for `E_loo()` to give reliable estimates**. If the
#' `log_ratios` argument is provided, `E_loo()` also computes a function
#' specific Pareto k diagnostic, which must also be small for a reliable
#' estimate. See more details below.
#'
#' @export
#' @param x A numeric vector or matrix.
#' @param psis_object An object returned by [psis()].
#' @param log_ratios Optionally, a vector or matrix (the same dimensions as `x`)
#'   of raw (not smoothed) log ratios. If working with log-likelihood values,
#'   the log ratios are the **negative** of those values. If `log_ratios` is
#'   specified we are able to compute more accurate [Pareto k][pareto-k-diagnostic]
#'   diagnostics specific to `E_loo()`.
#' @param type The type of expectation to compute. The options are
#'   `"mean"`, `"variance"`, `"sd"`, and `"quantile"`.
#' @param probs For computing quantiles, a vector of probabilities.
#' @param ... Arguments passed to individual methods.
#'
#' @return A named list with the following components:
#' \describe{
#'   \item{`value`}{
#'   The result of the computation.
#'
#'   For the matrix method, `value` is a vector with `ncol(x)`
#'   elements, with one exception: when `type="quantile"` and
#'   multiple values are specified in `probs` the `value` component of
#'   the returned object is a `length(probs)` by `ncol(x)` matrix.
#'
#'   For the default/vector method the `value` component is scalar, with
#'   one exception: when `type="quantile"` and multiple values
#'   are specified in `probs` the `value` component is a vector with
#'   `length(probs)` elements.
#'   }
#'  \item{`pareto_k`}{
#'   Function-specific diagnostic.
#'
#'   For the matrix method it will be a vector of length `ncol(x)`
#'   containing estimates of the shape parameter \eqn{k} of the
#'   generalized Pareto distribution. For the default/vector method,
#'   the estimate is a scalar. If `log_ratios` is not specified when
#'   calling `E_loo()`, the smoothed log-weights are used to estimate
#'   Pareto-k's, which may produce optimistic estimates.
#'
#'   For `type="mean"`, `type="var"`, and `type="sd"`, the returned Pareto-k is
#'   usually the maximum of the Pareto-k's for the left and right tail of \eqn{hr}
#'   and the right tail of \eqn{r}, where \eqn{r} is the importance ratio and
#'   \eqn{h=x} for `type="mean"` and \eqn{h=x^2} for `type="var"` and `type="sd"`.
#'   If \eqn{h} is binary, constant, or not finite, or if `type="quantile"`, the
#'   returned Pareto-k is the Pareto-k for the right tail of \eqn{r}.
#'  }
#' }
#'
#'
#' @examples
#' \donttest{
#' if (requireNamespace("rstanarm", quietly = TRUE)) {
#' # Use rstanarm package to quickly fit a model and get both a log-likelihood
#' # matrix and draws from the posterior predictive distribution
#' library("rstanarm")
#'
#' # data from help("lm")
#' ctl <- c(4.17,5.58,5.18,6.11,4.50,4.61,5.17,4.53,5.33,5.14)
#' trt <- c(4.81,4.17,4.41,3.59,5.87,3.83,6.03,4.89,4.32,4.69)
#' d <- data.frame(
#'   weight = c(ctl, trt),
#'   group = gl(2, 10, 20, labels = c("Ctl","Trt"))
#' )
#' fit <- stan_glm(weight ~ group, data = d, refresh = 0)
#' yrep <- posterior_predict(fit)
#' dim(yrep)
#'
#' log_ratios <- -1 * log_lik(fit)
#' dim(log_ratios)
#'
#' r_eff <- relative_eff(exp(-log_ratios), chain_id = rep(1:4, each = 1000))
#' psis_object <- psis(log_ratios, r_eff = r_eff, cores = 2)
#'
#' E_loo(yrep, psis_object, type = "mean")
#' E_loo(yrep, psis_object, type = "var")
#' E_loo(yrep, psis_object, type = "sd")
#' E_loo(yrep, psis_object, type = "quantile", probs = 0.5) # median
#' E_loo(yrep, psis_object, type = "quantile", probs = c(0.1, 0.9))
#'
#' # We can get more accurate Pareto k diagnostic if we also provide
#' # the log_ratios argument
#' E_loo(yrep, psis_object, type = "mean", log_ratios = log_ratios)
#' }
#' }
#'
E_loo <- function(x, psis_object, ...) {
  UseMethod("E_loo")
}

#' @rdname E_loo
#' @export
E_loo.default <-
  function(x,
           psis_object,
           ...,
           type = c("mean", "variance", "sd", "quantile"),
           probs = NULL,
           log_ratios = NULL) {
    stopifnot(
      is.numeric(x),
      is.psis(psis_object),
      length(x) == dim(psis_object)[1],
      is.null(log_ratios) || (length(x) == length(log_ratios))
    )
    type <- match.arg(type)
    E_fun <- .E_fun(type)
    w <- as.vector(weights(psis_object, log = FALSE))
    x <- as.vector(x)
    out <- E_fun(x, w, probs)

    if (is.null(log_ratios)) {
      # Use of smoothed ratios gives slightly optimistic
      # Pareto-k's, but these are still better than nothing
      log_ratios <- weights(psis_object, log = TRUE)
    }
    h <- switch(
      type,
      "mean" = x,
      "variance" = x^2,
      "sd" = x^2,
      "quantile" = NULL
    )
    khat <- E_loo_khat.default(h, psis_object, log_ratios)
    list(value = out, pareto_k = khat)
  }

#' @rdname E_loo
#' @export
E_loo.matrix <-
  function(x,
           psis_object,
           ...,
           type = c("mean", "variance", "sd", "quantile"),
           probs = NULL,
           log_ratios = NULL) {
    stopifnot(
      is.numeric(x),
      is.psis(psis_object),
      identical(dim(x), dim(psis_object)),
      is.null(log_ratios) || identical(dim(x), dim(log_ratios))
    )
    type <- match.arg(type)
    E_fun <- .E_fun(type)
    fun_val <- numeric(1)
    if (type == "quantile") {
      stopifnot(
        is.numeric(probs),
        length(probs) >= 1,
        all(probs > 0 & probs < 1)
      )
      fun_val <- numeric(length(probs))
    }
    w <- weights(psis_object, log = FALSE)

    out <- vapply(seq_len(ncol(x)), function(i) {
      E_fun(x[, i], w[, i], probs = probs)
    }, FUN.VALUE = fun_val)

    if (is.null(log_ratios)) {
      # Use of smoothed ratios gives slightly optimistic
      # Pareto-k's, but these are still better than nothing
      log_ratios <- weights(psis_object, log = TRUE)
    }
    h <- switch(
      type,
      "mean" = x,
      "variance" = x^2,
      "sd" = x^2,
      "quantile" = NULL
    )
    khat <- E_loo_khat.matrix(h, psis_object, log_ratios)
    list(value = out, pareto_k = khat)
  }



#' Select the function to use based on user's 'type' argument
#'
#' @noRd
#' @param type User's `type` argument.
#' @return The function for computing the weighted expectation specified by
#'   `type`.
#'
.E_fun <- function(type = c("mean", "variance", "sd", "quantile")) {
  switch(
    type,
    "mean" = .wmean,
    "variance" = .wvar,
    "sd" = .wsd,
    "quantile" = .wquant
  )
}

#' loo-weighted mean, variance, and quantiles
#'
#' @noRd
#' @param x,w Vectors of the same length. This should be checked inside
#'   `E_loo()` before calling these functions.
#' @param probs Vector of probabilities.
#' @param ... ignored. Having ... allows `probs` to be passed to `.wmean()` and
#'   `.wvar()` in `E_loo()` without resulting in an error.
#'
.wmean <- function(x, w, ...) {
  sum(w * x)
}
.wvar <- function(x, w, ...) {
  # The denominator (1- sum(w^2)) is equal to (ESS-1)/ESS, where effective
  # sample size ESS is estimated with the generic target quantity invariant
  # estimate 1/sum(w^2), see e.g. "Monte Carlo theory, methods and examples"
  # by Owen (2013).
  (sum(.wmean(x^2, w)) - sum(.wmean(x, w)^2)) / (1 - sum(w^2))
}
.wsd <- function(x, w, ...) {
  sqrt(.wvar(x, w))
}
.wquant <- function(x, w, probs, ...) {
  if (all(w == w[1])) {
    return(quantile(x, probs = probs, names = FALSE))
  }

  ord <- order(x)
  x <- x[ord]
  w <- w[ord]
  ww <- cumsum(w)
  ww <- ww / ww[length(ww)]

  qq <- numeric(length(probs))
  for (j in seq_along(probs)) {
    ids <- which(ww >= probs[j])
    wi <- min(ids)
    if (wi == 1) {
      qq[j] <- x[1]
    } else {
      w1 <- ww[wi - 1]
      x1 <- x[wi - 1]
      qq[j] <- x1 + (x[wi] - x1) * (probs[j] - w1) / (ww[wi] - w1)
    }
  }
  return(qq)
}

#' Compute function-specific k-hat diagnostics
#'
#' @noRd
#' @param log_ratios Vector or matrix of raw (not smoothed) log ratios with the
#'   same dimensions as `x`. If working with log-likelihood values, the log
#'   ratios are the negative of those values.
#' @return Vector (of length `NCOL(x)`) of k-hat estimates.
#'
E_loo_khat <- function(x, psis_object, log_ratios, ...) {
  UseMethod("E_loo_khat")
}
#' @export
E_loo_khat.default <- function(x, psis_object, log_ratios, ...) {
  .E_loo_khat_i(x, log_ratios, attr(psis_object, "tail_len"))
}
#' @export
E_loo_khat.matrix <- function(x, psis_object, log_ratios, ...) {
  tail_lengths <- attr(psis_object, "tail_len")
  if (is.null(x)) {
    sapply(seq_len(ncol(log_ratios)), function(i) {
      .E_loo_khat_i(x, log_ratios[, i], tail_lengths[i])
    })
  } else {
    sapply(seq_len(ncol(log_ratios)), function(i) {
      .E_loo_khat_i(x[, i], log_ratios[, i], tail_lengths[i])
    })
  }
}

#' Compute function-specific khat estimates
#'
#' @noRd
#' @param x_i Vector of values of function h(theta)
#' @param log_ratios_i S-vector of log_ratios, log(r(theta)), for a single
#'   observation.
#' @param tail_len_i Integer tail length used for fitting GPD.
#' @return Scalar h-specific k-hat estimate.
#'
.E_loo_khat_i <- function(x_i, log_ratios_i, tail_len_i) {
  h_theta <- x_i
  r_theta <- exp(log_ratios_i - max(log_ratios_i))
  khat_r <- posterior::pareto_khat(r_theta, tail = "right", ndraws_tail = tail_len_i)
  if (is.list(khat_r)) { # retain compatiblity with older posterior that returned a list
    khat_r <- khat_r$khat
  }
  if (is.null(x_i) || is_constant(x_i) || length(unique(x_i))==2 ||
        anyNA(x_i) || any(is.infinite(x_i))) {
    khat_r
  } else {
    khat_hr <- posterior::pareto_khat(h_theta * r_theta, tail = "both", ndraws_tail = tail_len_i)
    if (is.list(khat_hr)) { # retain compatiblity with older posterior that returned a list
      khat_hr <- khat_hr$khat
    }
    if (is.na(khat_hr) && is.na(khat_r)) {
      k <- NA
    } else {
      k <- max(khat_hr, khat_r, na.rm=TRUE)
    }
    k
  }
}

LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()
LLvec <- LLmat[, 1]
chain_id <- rep(1:2, each = dim(LLarr)[1])
r_eff_mat <- relative_eff(exp(LLmat), chain_id)
r_eff_vec <- relative_eff(exp(LLvec), chain_id = chain_id)
# cores = 1 because forked processes can segfault on CRAN's macOS machines if
# R is linked to Apple's Accelerate/vecLib BLAS, which is not fork-safe
psis_mat <- psis(-LLmat, r_eff = r_eff_mat, cores = 1)
psis_vec <- psis(-LLvec, r_eff = r_eff_vec)

set.seed(123)
x <- matrix(rnorm(length(LLmat)), nrow = nrow(LLmat), ncol = ncol(LLmat))
log_rats <- -LLmat

# matrix method
E_test_mean <- E_loo(x, psis_mat, type = "mean", log_ratios = log_rats)
E_test_var <- E_loo(x, psis_mat, type = "var", log_ratios = log_rats)
E_test_sd <- E_loo(x, psis_mat, type = "sd", log_ratios = log_rats)
E_test_quant <- E_loo(
  x,
  psis_mat,
  type = "quantile",
  probs = 0.5,
  log_ratios = log_rats
)
E_test_quant2 <- E_loo(
  x,
  psis_mat,
  type = "quantile",
  probs = c(0.1, 0.9),
  log_ratios = log_rats
)

# vector method
E_test_mean_vec <- E_loo(
  x[, 1],
  psis_vec,
  type = "mean",
  log_ratios = log_rats[, 1]
)
E_test_var_vec <- E_loo(
  x[, 1],
  psis_vec,
  type = "var",
  log_ratios = log_rats[, 1]
)
E_test_sd_vec <- E_loo(
  x[, 1],
  psis_vec,
  type = "sd",
  log_ratios = log_rats[, 1]
)
E_test_quant_vec <- E_loo(
  x[, 1],
  psis_vec,
  type = "quant",
  probs = 0.5,
  log_ratios = log_rats[, 1]
)
E_test_quant_vec2 <- E_loo(
  x[, 1],
  psis_vec,
  type = "quant",
  probs = c(0.1, 0.5, 0.9),
  log_ratios = log_rats[, 1]
)

# E_loo_khat
khat <- loo:::E_loo_khat.matrix(x, psis_mat, log_rats)

test_that("E_loo return types correct for matrix method", {
  expect_type(E_test_mean, "list")
  expect_named(E_test_mean, c("value", "pareto_k"))
  expect_length(E_test_mean, 2)
  expect_length(E_test_mean$value, ncol(x))
  expect_length(E_test_mean$pareto_k, ncol(x))

  expect_type(E_test_var, "list")
  expect_named(E_test_var, c("value", "pareto_k"))
  expect_length(E_test_var, 2)
  expect_length(E_test_var$value, ncol(x))
  expect_length(E_test_var$pareto_k, ncol(x))

  expect_type(E_test_sd, "list")
  expect_named(E_test_sd, c("value", "pareto_k"))
  expect_length(E_test_sd, 2)
  expect_length(E_test_sd$value, ncol(x))
  expect_length(E_test_sd$pareto_k, ncol(x))

  expect_type(E_test_quant, "list")
  expect_named(E_test_quant, c("value", "pareto_k"))
  expect_length(E_test_quant, 2)
  expect_length(E_test_quant$value, ncol(x))
  expect_length(E_test_quant$pareto_k, ncol(x))

  expect_type(E_test_quant2, "list")
  expect_named(E_test_quant2, c("value", "pareto_k"))
  expect_length(E_test_quant2, 2)
  expect_equal(dim(E_test_quant2$value), c(2, ncol(x)))
  expect_length(E_test_quant2$pareto_k, ncol(x))
})

test_that("E_loo return types correct for default/vector method", {
  expect_type(E_test_mean_vec, "list")
  expect_named(E_test_mean_vec, c("value", "pareto_k"))
  expect_length(E_test_mean_vec, 2)
  expect_length(E_test_mean_vec$value, 1)
  expect_length(E_test_mean_vec$pareto_k, 1)

  expect_type(E_test_var_vec, "list")
  expect_named(E_test_var_vec, c("value", "pareto_k"))
  expect_length(E_test_var_vec, 2)
  expect_length(E_test_var_vec$value, 1)
  expect_length(E_test_var_vec$pareto_k, 1)

  expect_type(E_test_sd_vec, "list")
  expect_named(E_test_sd_vec, c("value", "pareto_k"))
  expect_length(E_test_sd_vec, 2)
  expect_length(E_test_sd_vec$value, 1)
  expect_length(E_test_sd_vec$pareto_k, 1)

  expect_type(E_test_quant_vec, "list")
  expect_named(E_test_quant_vec, c("value", "pareto_k"))
  expect_length(E_test_quant_vec, 2)
  expect_length(E_test_quant_vec$value, 1)
  expect_length(E_test_quant_vec$pareto_k, 1)

  expect_type(E_test_quant_vec2, "list")
  expect_named(E_test_quant_vec2, c("value", "pareto_k"))
  expect_length(E_test_quant_vec2, 2)
  expect_length(E_test_quant_vec2$value, 3)
  expect_length(E_test_quant_vec2$pareto_k, 1)
})

test_that("E_loo.default equal to snapshots", {
  expect_snapshot_value(E_test_mean_vec, style = "serialize")
  expect_snapshot_value(E_test_var_vec, style = "serialize")
  expect_snapshot_value(E_test_sd_vec, style = "serialize")
  expect_snapshot_value(E_test_quant_vec, style = "serialize")
  expect_snapshot_value(E_test_quant_vec2, style = "serialize")
})

test_that("E_loo.matrix equal to snapshots", {
  expect_snapshot_value(E_test_mean, style = "serialize")
  expect_snapshot_value(E_test_var, style = "serialize")
  expect_snapshot_value(E_test_sd, style = "serialize")
  expect_snapshot_value(E_test_quant, style = "serialize")
  expect_snapshot_value(E_test_quant2, style = "serialize")
})

test_that("E_loo throws correct errors and warnings", {
  # warnings
  expect_no_warning(E_loo.matrix(x, psis_mat))
  # no warnings if x is constant, binary, NA, NaN, Inf
  expect_no_warning(E_loo.matrix(x * 0, psis_mat))
  expect_no_warning(E_loo.matrix(0 + (x > 0), psis_mat))
  expect_no_warning(E_loo.matrix(x + NA, psis_mat))
  expect_no_warning(E_loo.matrix(x * NaN, psis_mat))
  expect_no_warning(E_loo.matrix(x * Inf, psis_mat))
  expect_no_warning(E_test <- E_loo.default(x[, 1], psis_vec))
  expect_length(E_test$pareto_k, 1)

  # errors
  expect_error(E_loo(x, 1), "is.psis")
  expect_error(
    E_loo(x, psis_mat, type = "quantile", probs = 2),
    "all(probs > 0 & probs < 1) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    E_loo(rep("a", nrow(x)), psis_vec),
    "is.numeric(x) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    E_loo(1:10, psis_vec),
    "length(x) == dim(psis_object)[1] is not TRUE",
    fixed = TRUE
  )
  expect_error(
    E_loo(cbind(1:10, 1:10), psis_mat),
    "identical(dim(x), dim(psis_object)) is not TRUE",
    fixed = TRUE
  )
})


test_that("weighted quantiles work", {
  .wquant_rapprox <- function(x, w, probs) {
    stopifnot(all(probs > 0 & probs < 1))
    ord <- order(x)
    d <- x[ord]
    ww <- w[ord]
    p <- cumsum(ww) / sum(ww)
    stats::approx(p, d, probs, rule = 2)$y
  }
  .wquant_sim <- function(x, w, probs, n_sims) {
    xx <- sample(x, size = n_sims, replace = TRUE, prob = w / sum(w))
    quantile(xx, probs, names = FALSE)
  }

  set.seed(123)
  pr <- seq(0.025, 0.975, 0.025)

  x1 <- rnorm(100)
  w1 <- rlnorm(100)
  expect_equal(
    .wquant(x1, w1, pr),
    .wquant_rapprox(x1, w1, pr)
  )

  x1 <- rnorm(1e4)
  w1 <- rlnorm(1e4)
  # expect_equal(
  #   .wquant(x1, w1, pr),
  #   .wquant_sim(x1, w1, pr, n_sim = 5e6),
  #   tol = 0.005
  # )

  expect_equal(
    .wquant(x1, rep(1, length(x1)), pr),
    quantile(x1, probs = pr, names = FALSE)
  )
})

test_that("weighted variance works", {
  x <- rnorm(100)
  w <- rep(0.01, 100)
  expect_equal(.wvar(x, w), var(x))
  expect_equal(.wsd(x, w), sqrt(.wvar(x, w)))

  w <- c(rep(0.1, 10), rep(0, 90))
  expect_equal(.wvar(x, w), var(x[w > 0]))
})

LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()
LLvec <- LLmat[, 1]
chain_id <- rep(1:2, each = dim(LLarr)[1])
r_eff_mat <- relative_eff(exp(LLmat), chain_id)
r_eff_vec <- relative_eff(exp(LLvec), chain_id = chain_id)
psis_mat <- psis(-LLmat, r_eff = r_eff_mat, cores = 2)
psis_vec <- psis(-LLvec, r_eff = r_eff_vec)

set.seed(123)
x <- matrix(rnorm(length(LLmat)), nrow = nrow(LLmat), ncol = ncol(LLmat))
log_rats <- -LLmat

# matrix method
E_test_mean <- E_loo(x, psis_mat, type = "mean", log_ratios = log_rats)
E_test_var <- E_loo(x, psis_mat, type = "var", log_ratios = log_rats)
E_test_sd <- E_loo(x, psis_mat, type = "sd", log_ratios = log_rats)
E_test_quant <- E_loo(
  x,
  psis_mat,
  type = "quantile",
  probs = 0.5,
  log_ratios = log_rats
)
E_test_quant2 <- E_loo(
  x,
  psis_mat,
  type = "quantile",
  probs = c(0.1, 0.9),
  log_ratios = log_rats
)

# vector method
E_test_mean_vec <- E_loo(
  x[, 1],
  psis_vec,
  type = "mean",
  log_ratios = log_rats[, 1]
)
E_test_var_vec <- E_loo(
  x[, 1],
  psis_vec,
  type = "var",
  log_ratios = log_rats[, 1]
)
E_test_sd_vec <- E_loo(
  x[, 1],
  psis_vec,
  type = "sd",
  log_ratios = log_rats[, 1]
)
E_test_quant_vec <- E_loo(
  x[, 1],
  psis_vec,
  type = "quant",
  probs = 0.5,
  log_ratios = log_rats[, 1]
)
E_test_quant_vec2 <- E_loo(
  x[, 1],
  psis_vec,
  type = "quant",
  probs = c(0.1, 0.5, 0.9),
  log_ratios = log_rats[, 1]
)

# E_loo_khat
khat <- loo:::E_loo_khat.matrix(x, psis_mat, log_rats)

test_that("E_loo return types correct for matrix method", {
  expect_type(E_test_mean, "list")
  expect_named(E_test_mean, c("value", "pareto_k"))
  expect_length(E_test_mean, 2)
  expect_length(E_test_mean$value, ncol(x))
  expect_length(E_test_mean$pareto_k, ncol(x))

  expect_type(E_test_var, "list")
  expect_named(E_test_var, c("value", "pareto_k"))
  expect_length(E_test_var, 2)
  expect_length(E_test_var$value, ncol(x))
  expect_length(E_test_var$pareto_k, ncol(x))

  expect_type(E_test_sd, "list")
  expect_named(E_test_sd, c("value", "pareto_k"))
  expect_length(E_test_sd, 2)
  expect_length(E_test_sd$value, ncol(x))
  expect_length(E_test_sd$pareto_k, ncol(x))

  expect_type(E_test_quant, "list")
  expect_named(E_test_quant, c("value", "pareto_k"))
  expect_length(E_test_quant, 2)
  expect_length(E_test_quant$value, ncol(x))
  expect_length(E_test_quant$pareto_k, ncol(x))

  expect_type(E_test_quant2, "list")
  expect_named(E_test_quant2, c("value", "pareto_k"))
  expect_length(E_test_quant2, 2)
  expect_equal(dim(E_test_quant2$value), c(2, ncol(x)))
  expect_length(E_test_quant2$pareto_k, ncol(x))
})

test_that("E_loo return types correct for default/vector method", {
  expect_type(E_test_mean_vec, "list")
  expect_named(E_test_mean_vec, c("value", "pareto_k"))
  expect_length(E_test_mean_vec, 2)
  expect_length(E_test_mean_vec$value, 1)
  expect_length(E_test_mean_vec$pareto_k, 1)

  expect_type(E_test_var_vec, "list")
  expect_named(E_test_var_vec, c("value", "pareto_k"))
  expect_length(E_test_var_vec, 2)
  expect_length(E_test_var_vec$value, 1)
  expect_length(E_test_var_vec$pareto_k, 1)

  expect_type(E_test_sd_vec, "list")
  expect_named(E_test_sd_vec, c("value", "pareto_k"))
  expect_length(E_test_sd_vec, 2)
  expect_length(E_test_sd_vec$value, 1)
  expect_length(E_test_sd_vec$pareto_k, 1)

  expect_type(E_test_quant_vec, "list")
  expect_named(E_test_quant_vec, c("value", "pareto_k"))
  expect_length(E_test_quant_vec, 2)
  expect_length(E_test_quant_vec$value, 1)
  expect_length(E_test_quant_vec$pareto_k, 1)

  expect_type(E_test_quant_vec2, "list")
  expect_named(E_test_quant_vec2, c("value", "pareto_k"))
  expect_length(E_test_quant_vec2, 2)
  expect_length(E_test_quant_vec2$value, 3)
  expect_length(E_test_quant_vec2$pareto_k, 1)
})

test_that("E_loo.default equal to snapshots", {
  expect_snapshot_value(E_test_mean_vec, style = "serialize")
  expect_snapshot_value(E_test_var_vec, style = "serialize")
  expect_snapshot_value(E_test_sd_vec, style = "serialize")
  expect_snapshot_value(E_test_quant_vec, style = "serialize")
  expect_snapshot_value(E_test_quant_vec2, style = "serialize")
})

test_that("E_loo.matrix equal to snapshots", {
  expect_snapshot_value(E_test_mean, style = "serialize")
  expect_snapshot_value(E_test_var, style = "serialize")
  expect_snapshot_value(E_test_sd, style = "serialize")
  expect_snapshot_value(E_test_quant, style = "serialize")
  expect_snapshot_value(E_test_quant2, style = "serialize")
})

test_that("E_loo throws correct errors and warnings", {
  # warnings
  expect_no_warning(E_loo.matrix(x, psis_mat))
  # no warnings if x is constant, binary, NA, NaN, Inf
  expect_no_warning(E_loo.matrix(x * 0, psis_mat))
  expect_no_warning(E_loo.matrix(0 + (x > 0), psis_mat))
  expect_no_warning(E_loo.matrix(x + NA, psis_mat))
  expect_no_warning(E_loo.matrix(x * NaN, psis_mat))
  expect_no_warning(E_loo.matrix(x * Inf, psis_mat))
  expect_no_warning(E_test <- E_loo.default(x[, 1], psis_vec))
  expect_length(E_test$pareto_k, 1)

  # errors
  expect_error(E_loo(x, 1), "is.psis")
  expect_error(
    E_loo(x, psis_mat, type = "quantile", probs = 2),
    "all(probs > 0 & probs < 1) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    E_loo(rep("a", nrow(x)), psis_vec),
    "is.numeric(x) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    E_loo(1:10, psis_vec),
    "length(x) == dim(psis_object)[1] is not TRUE",
    fixed = TRUE
  )
  expect_error(
    E_loo(cbind(1:10, 1:10), psis_mat),
    "identical(dim(x), dim(psis_object)) is not TRUE",
    fixed = TRUE
  )
})


test_that("weighted quantiles work", {
  .wquant_rapprox <- function(x, w, probs) {
    stopifnot(all(probs > 0 & probs < 1))
    ord <- order(x)
    d <- x[ord]
    ww <- w[ord]
    p <- cumsum(ww) / sum(ww)
    stats::approx(p, d, probs, rule = 2)$y
  }
  .wquant_sim <- function(x, w, probs, n_sims) {
    xx <- sample(x, size = n_sims, replace = TRUE, prob = w / sum(w))
    quantile(xx, probs, names = FALSE)
  }

  set.seed(123)
  pr <- seq(0.025, 0.975, 0.025)

  x1 <- rnorm(100)
  w1 <- rlnorm(100)
  expect_equal(
    .wquant(x1, w1, pr),
    .wquant_rapprox(x1, w1, pr)
  )

  x1 <- rnorm(1e4)
  w1 <- rlnorm(1e4)
  # expect_equal(
  #   .wquant(x1, w1, pr),
  #   .wquant_sim(x1, w1, pr, n_sim = 5e6),
  #   tol = 0.005
  # )

  expect_equal(
    .wquant(x1, rep(1, length(x1)), pr),
    quantile(x1, probs = pr, names = FALSE)
  )
})

test_that("weighted variance works", {
  x <- rnorm(100)
  w <- rep(0.01, 100)
  expect_equal(.wvar(x, w), var(x))
  expect_equal(.wsd(x, w), sqrt(.wvar(x, w)))

  w <- c(rep(0.1, 10), rep(0, 90))
  expect_equal(.wvar(x, w), var(x[w > 0]))
})

#' Detect if OS is Windows
#' @noRd
os_is_windows <- function() {
  checkmate::test_os("windows")
}

#' More stable version of `log(mean(exp(x)))`
#'
#' @noRd
#' @param x A numeric vector.
#' @return A scalar equal to `log(mean(exp(x)))`.
#'
logMeanExp <- function(x) {
  logS <- log(length(x))
  matrixStats::logSumExp(x) - logS
}

#' More stable version of `log(colMeans(exp(x)))`
#'
#' @noRd
#' @param x A matrix.
#' @return A vector where each element is `logMeanExp()` of a column of `x`.
#'
colLogMeanExps <- function(x) {
  logS <- log(nrow(x))
  matrixStats::colLogSumExps(x) - logS
}

#' Compute point estimates and standard errors from pointwise vectors
#'
#' @noRd
#' @param x A matrix.
#' @return An `ncol(x)` by 2 matrix with columns `"Estimate"` and `"SE"`
#'   and rownames equal to `colnames(x)`.
#'
table_of_estimates <- function(x) {
  out <- cbind(
    Estimate = matrixStats::colSums2(x),
    SE = sqrt(nrow(x) * matrixStats::colVars(x))
  )
  rownames(out) <- colnames(x)
  return(out)
}


# validating and reshaping arrays/matrices  -------------------------------

#' Check for `NA` and non-finite values in log-lik (or log-ratios)
#' array/matrix/vector
#'
#' @noRd
#' @param x Array/matrix/vector of log-likelihood or log-ratio values.
#' @return `x`, invisibly, if no error is thrown.
#'
validate_ll <- function(x) {
  if (is.list(x)) {
    stop("List not allowed as input.")
  } else if (anyNA(x)) {
    stop("NAs not allowed in input.")
  } else if (any(x == Inf)) {
    stop("All input values must be finite or -Inf.")
  }
  invisible(x)
}

#' Convert iter by chain by obs array to (iter * chain) by obs matrix
#'
#' @noRd
#' @param x Array to convert.
#' @return An (iter * chain) by obs matrix.
#'
llarray_to_matrix <- function(x) {
  stopifnot(is.array(x), length(dim(x)) == 3)
  xdim <- dim(x)
  dim(x) <- c(prod(xdim[1:2]), xdim[3])
  unname(x)
}

#' Convert (iter * chain) by obs matrix to iter by chain by obs array
#'
#' @noRd
#' @param x matrix to convert.
#' @param chain_id vector of chain ids.
#' @return iter by chain by obs array
#'
llmatrix_to_array <- function(x, chain_id) {
  stopifnot(is.matrix(x), all(chain_id == as.integer(chain_id)))

  lldim <- dim(x)
  n_chain <- length(unique(chain_id))
  chain_id <- as.integer(chain_id)
  chain_counts <- as.numeric(table(chain_id))

  if (length(chain_id) != lldim[1]) {
    stop("Number of rows in matrix not equal to length(chain_id).",
         call. = FALSE)
  } else if (any(chain_counts != chain_counts[1])) {
    stop("Not all chains have same number of iterations.",
         call. = FALSE)
  } else if (max(chain_id) != n_chain) {
    stop("max(chain_id) not equal to the number of chains.",
         call. = FALSE)
  }

  n_iter <- lldim[1] / n_chain
  n_obs <- lldim[2]
  a <- array(data = NA, dim = c(n_iter, n_chain, n_obs))
  for (c in seq_len(n_chain)) {
    a[, c, ] <- x[chain_id == c, , drop = FALSE]
  }
  return(a)
}


#' Validate that log-lik function exists and has correct arg names
#'
#' @noRd
#' @param x A function with arguments `data_i` and `draws`.
#' @return Either returns `x` or throws an error.
#'
validate_llfun <- function(x) {
  f <- match.fun(x)
  must_have <- c("data_i", "draws")
  arg_names <- names(formals(f))
  if (!all(must_have %in% arg_names)) {
    stop(
      "Log-likelihood function must have at least the arguments ",
      "'data_i' and 'draws'",
      call. = FALSE
    )
  }
  return(f)
}


#' Named lists
#'
#' Create a named list using specified names or, if names are omitted, using the
#' names of the objects in the list. The code `list(a = a, b = b)` becomes
#' `nlist(a,b)` and `list(a = a, b = 2)` becomes `nlist(a, b = 2)`, etc.
#'
#' @export
#' @keywords internal
#' @param ... Objects to include in the list.
#' @return A named list.
#' @examples
#'
#' # All variables already defined
#' a <- rnorm(100)
#' b <- mat.or.vec(10, 3)
#' nlist(a,b)
#'
#' # Define some variables in the call and take the rest from the environment
#' nlist(a, b, veggies = c("lettuce", "spinach"), fruits = c("banana", "papaya"))
#'
nlist <- function(...) {
  m <- match.call()
  out <- list(...)
  no_names <- is.null(names(out))
  has_name <- if (no_names) FALSE else nzchar(names(out))
  if (all(has_name))
    return(out)
  nms <- as.character(m)[-1L]
  if (no_names) {
    names(out) <- nms
  } else {
    names(out)[!has_name] <- nms[!has_name]
  }

  return(out)
}


# Check how many cores to use and throw deprecation warning if loo.cores is used
loo_cores <- function(cores) {
  loo_cores_op <- getOption("loo.cores", NA)
  if (!is.na(loo_cores_op) && (loo_cores_op != cores)) {
    cores <- loo_cores_op
    warning("'loo.cores' is deprecated, please use 'mc.cores' or pass 'cores' explicitly.",
            call. = FALSE)
  }
  return(cores)
}


# nocov start
# release reminders (for devtools)
release_questions <- function() {
  c(
    "Have you updated references?",
    "Have you updated inst/CITATION?",
    "Have you updated the vignettes?"
  )
}
# nocov end

is_constant <- function(x, tol = .Machine$double.eps) {
  abs(max(x) - min(x)) < tol
}

LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()

test_that("example_loglik_array and example_loglik_matrix dimensions ok", {
  dim_arr <- dim(LLarr)
  dim_mat <- dim(LLmat)
  expect_equal(dim_mat[1], dim_arr[1] * dim_arr[2])
  expect_equal(dim_mat[2], dim_arr[3])
})

test_that("example_loglik_array and example_loglik_matrix contain same values", {
  expect_equal(LLmat[1:500, ], LLarr[, 1, ])
  expect_equal(LLmat[501:1000, ], LLarr[, 2, ])
})

test_that("reshaping functions result in correct dimensions", {
  LLmat2 <- llarray_to_matrix(LLarr)
  expect_identical(LLmat2, LLmat)

  LLarr2 <- llmatrix_to_array(LLmat2, chain_id = rep(1:2, each = 500))
  expect_identical(LLarr2, LLarr)
})

test_that("reshaping functions throw correct errors", {
  expect_error(
    llmatrix_to_array(LLmat, chain_id = rep(1:2, times = c(400, 600))),
    regexp = "Not all chains have same number of iterations",
    fixed = TRUE
  )
  expect_error(
    llmatrix_to_array(LLmat, chain_id = rep(1:2, each = 400)),
    regexp = "Number of rows in matrix not equal to length(chain_id)",
    fixed = TRUE
  )
  expect_error(
    llmatrix_to_array(LLmat, chain_id = rep(2:3, each = 500)),
    regexp = "max(chain_id) not equal to the number of chains",
    fixed = TRUE
  )
  expect_error(
    llmatrix_to_array(LLmat, chain_id = rnorm(1000)),
    regexp = "all(chain_id == as.integer(chain_id)) is not TRUE",
    fixed = TRUE
  )
})

test_that("colLogMeanExps(x) = log(colMeans(exp(x))) ", {
  expect_equal(colLogMeanExps(LLmat), log(colMeans(exp(LLmat))))
})

test_that("validating log-lik objects and functions works", {
  f_ok <- function(data_i, draws) return(NULL)
  f_bad1 <- function(data_i) return(NULL)
  f_bad2 <- function(data, draws) return(NULL)
  expect_equal(validate_llfun(f_ok), f_ok)

  bad_msg <- "Log-likelihood function must have at least the arguments 'data_i' and 'draws'"
  expect_error(validate_llfun(f_bad1), bad_msg)
  expect_error(validate_llfun(f_bad2), bad_msg)
})

test_that("nlist works", {
  a <- 1
  b <- 2
  c <- 3
  nlist_val <- list(nlist(a, b, c), nlist(a, b, c = "tornado"))
  nlist_ans <- list(
    list(a = 1, b = 2, c = 3),
    list(a = 1, b = 2, c = "tornado")
  )
  expect_equal(nlist_val, nlist_ans)
  expect_equal(nlist(a = 1, b = 2, c = 3), list(a = 1, b = 2, c = 3))
})

test_that("loo_cores works", {
  expect_equal(loo_cores(10), 10)
  options(mc.cores = 2)
  expect_equal(loo_cores(getOption("mc.cores", 1)), 2)
  options(mc.cores = 1)

  options(loo.cores = 2)
  expect_warning(expect_equal(loo_cores(10), 2), "deprecated")
  options(loo.cores = NULL)
})

#' Efficient approximate leave-one-out cross-validation (LOO)
#'
#' The `loo()` methods for arrays, matrices, and functions compute PSIS-LOO
#' CV, efficient approximate leave-one-out (LOO) cross-validation for Bayesian
#' models using Pareto smoothed importance sampling ([PSIS][psis()]). This is
#' an implementation of the methods described in Vehtari, Gelman, and Gabry
#' (2017) and Vehtari, Simpson, Gelman, Yao, and Gabry (2024).
#'
#' @export loo loo.array loo.matrix loo.function
#' @param x A log-likelihood array, matrix, or function. The **Methods (by class)**
#'   section, below, has detailed descriptions of how to specify the inputs for
#'   each method.
#' @param r_eff Vector of relative effective sample size estimates for the
#'   likelihood (`exp(log_lik)`) of each observation. This is related to
#'   the relative efficiency of estimating the normalizing term in
#'   self-normalized importance sampling when using posterior draws obtained
#'   with MCMC. If MCMC draws are used and `r_eff` is not provided then
#'   the reported PSIS effective sample sizes and Monte Carlo error estimates
#'   can be over-optimistic. If the posterior draws are (near) independent then
#'   `r_eff=1` can be used. `r_eff` has to be a scalar (same value is used
#'    for all observations) or a vector with length equal to the number of
#'    observations. The default value is 1. See the [relative_eff()] helper
#'    functions for help computing `r_eff`.
#' @param save_psis Should the `psis` object created internally by `loo()` be
#'   saved in the returned object? The `loo()` function calls [psis()]
#'   internally but by default discards the (potentially large) `psis` object
#'   after using it to compute the LOO-CV summaries. Setting `save_psis=TRUE`
#'   will add a `psis_object` component to the list returned by `loo`.
#'   This is useful if you plan to use the [E_loo()] function to compute
#'   weighted expectations after running `loo`. Several functions in the
#'   \pkg{bayesplot} package also accept `psis` objects.
#' @template cores
#' @template is_method
#'
#' @details The `loo()` function is an S3 generic and methods are provided for
#'   3-D pointwise log-likelihood arrays, pointwise log-likelihood matrices, and
#'   log-likelihood functions. The array and matrix methods are the most
#'   convenient, but for models fit to very large datasets the `loo.function()`
#'   method is more memory efficient and may be preferable.
#'
#' @section Defining `loo()` methods in a package: Package developers can define
#'   `loo()` methods for fitted models objects. See the example `loo.stanfit()`
#'   method in the **Examples** section below for an example of defining a
#'   method that calls `loo.array()`. The `loo.stanreg()` method in the
#'   **rstanarm** package is an example of defining a method that calls
#'   `loo.function()`.
#'
#' @return The `loo()` methods return a named list with class
#'   `c("psis_loo", "loo")` and components:
#' \describe{
#'  \item{`estimates`}{
#'   A matrix with two columns (`Estimate`, `SE`) and three rows (`elpd_loo`,
#'   `p_loo`, `looic`). This contains point estimates and standard errors of the
#'   expected log pointwise predictive density ([`elpd_loo`][loo-glossary]), the
#'   effective number of parameters ([`p_loo`][loo-glossary]) and the LOO
#'   information criterion `looic` (which is just `-2 * elpd_loo`, i.e.,
#'   converted to deviance scale).
#'  }
#'
#'  \item{`pointwise`}{
#'   A matrix with five columns (and number of rows equal to the number of
#'   observations) containing the pointwise contributions of the measures
#'   (`elpd_loo`, `mcse_elpd_loo`, `p_loo`, `looic`, `influence_pareto_k`).
#'   in addition to the three measures in `estimates`, we also report
#'   pointwise values of the Monte Carlo standard error of [`elpd_loo`][loo-glossary]
#'   ([`mcse_elpd_loo`][loo-glossary]), and statistics describing the influence of
#'   each observation on the posterior distribution (`influence_pareto_k`).
#'   These are the estimates of the shape parameter \eqn{k} of the
#'   generalized Pareto fit to the importance ratios for each leave-one-out
#'   distribution (see the [pareto-k-diagnostic] page for details).
#'  }
#'
#'  \item{`diagnostics`}{
#'  A named list containing two vectors:
#'    * `pareto_k`: Importance sampling reliability diagnostics. By default,
#'      these are equal to the `influence_pareto_k` in `pointwise`.
#'      Some algorithms can improve importance sampling reliability and
#'      modify these diagnostics. See the [pareto-k-diagnostic] page for details.
#'    * `n_eff`: PSIS effective sample size estimates.
#'  }
#'
#'  \item{`psis_object`}{
#'  This component will be `NULL` unless the `save_psis` argument is set to
#'  `TRUE` when calling `loo()`. In that case `psis_object` will be the object
#'  of class `"psis"` that is created when the `loo()` function calls [psis()]
#'  internally to do the PSIS procedure.
#'  }
#' }
#'
#' @seealso
#'  * The __loo__ package [vignettes](https://mc-stan.org/loo/articles/index.html)
#'    for demonstrations.
#'  * The [FAQ page](https://mc-stan.org/loo/articles/online-only/faq.html) on
#'    the __loo__ website for answers to frequently asked questions.
#'  * [psis()] for the underlying Pareto Smoothed Importance Sampling (PSIS)
#'    procedure used in the LOO-CV approximation.
#'  * [pareto-k-diagnostic] for convenience functions for looking at diagnostics.
#'  * [loo_compare()] for model comparison.
#'
#' @template loo-and-psis-references
#'
#' @examples
#' ### Array and matrix methods (using example objects included with loo package)
#' # Array method
#' LLarr <- example_loglik_array()
#' rel_n_eff <- relative_eff(exp(LLarr))
#' loo(LLarr, r_eff = rel_n_eff, cores = 2)
#'
#' # Matrix method
#' LLmat <- example_loglik_matrix()
#' rel_n_eff <- relative_eff(exp(LLmat), chain_id = rep(1:2, each = 500))
#' loo(LLmat, r_eff = rel_n_eff, cores = 2)
#'
#'
#' ### Using log-likelihood function instead of array or matrix
#' set.seed(124)
#'
#' # Simulate data and draw from posterior
#' N <- 50; K <- 10; S <- 100; a0 <- 3; b0 <- 2
#' p <- rbeta(1, a0, b0)
#' y <- rbinom(N, size = K, prob = p)
#' a <- a0 + sum(y); b <- b0 + N * K - sum(y)
#' fake_posterior <- as.matrix(rbeta(S, a, b))
#' dim(fake_posterior) # S x 1
#' fake_data <- data.frame(y,K)
#' dim(fake_data) # N x 2
#'
#' llfun <- function(data_i, draws) {
#'   # each time called internally within loo the arguments will be equal to:
#'   # data_i: ith row of fake_data (fake_data[i,, drop=FALSE])
#'   # draws: entire fake_posterior matrix
#'   dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
#' }
#'
#' # Use the loo_i function to check that llfun works on a single observation
#' # before running on all obs. For example, using the 3rd obs in the data:
#' loo_3 <- loo_i(i = 3, llfun = llfun, data = fake_data, draws = fake_posterior)
#' print(loo_3$pointwise[, "elpd_loo"])
#'
#' # Use loo.function method (default r_eff=1 is used as this posterior not obtained via MCMC)
#' loo_with_fn <- loo(llfun, draws = fake_posterior, data = fake_data)
#'
#' # If we look at the elpd_loo contribution from the 3rd obs it should be the
#' # same as what we got above with the loo_i function and i=3:
#' print(loo_with_fn$pointwise[3, "elpd_loo"])
#' print(loo_3$pointwise[, "elpd_loo"])
#'
#' # Check that the loo.matrix method gives same answer as loo.function method
#' log_lik_matrix <- sapply(1:N, function(i) {
#'   llfun(data_i = fake_data[i,, drop=FALSE], draws = fake_posterior)
#' })
#' loo_with_mat <- loo(log_lik_matrix)
#' all.equal(loo_with_mat$estimates, loo_with_fn$estimates) # should be TRUE!
#'
#'
#' \dontrun{
#' ### For package developers: defining loo methods
#'
#' # An example of a possible loo method for 'stanfit' objects (rstan package).
#' # A similar method is included in the rstan package.
#' # In order for users to be able to call loo(stanfit) instead of
#' # loo.stanfit(stanfit) the NAMESPACE needs to be handled appropriately
#' # (roxygen2 and devtools packages are good for that).
#' #
#' loo.stanfit <-
#'  function(x,
#'          pars = "log_lik",
#'          ...,
#'          save_psis = FALSE,
#'          cores = getOption("mc.cores", 1)) {
#'   stopifnot(length(pars) == 1L)
#'   LLarray <- loo::extract_log_lik(stanfit = x,
#'                                   parameter_name = pars,
#'                                   merge_chains = FALSE)
#'   r_eff <- loo::relative_eff(x = exp(LLarray), cores = cores)
#'   loo::loo.array(LLarray,
#'                  r_eff = r_eff,
#'                  cores = cores,
#'                  save_psis = save_psis)
#' }
#' }
#'
#'

loo <- function(x, ...) {
  UseMethod("loo")
}

#' @export
#' @templateVar fn loo
#' @template array
#'
loo.array <-
  function(x,
           ...,
           r_eff = 1,
           save_psis = FALSE,
           cores = getOption("mc.cores", 1),
           is_method = c("psis", "tis", "sis")) {
    is_method <- match.arg(is_method)
    psis_out <- importance_sampling.array(log_ratios = -x, r_eff = r_eff, cores = cores, method = is_method)
    ll <- llarray_to_matrix(x)
    pointwise <- pointwise_loo_calcs(ll, psis_out)
    importance_sampling_loo_object(
      pointwise = pointwise,
      diagnostics = psis_out$diagnostics,
      dims = dim(psis_out),
      is_method = is_method,
      is_object = if (save_psis) psis_out else NULL
    )
  }

#' @export
#' @templateVar fn loo
#' @template matrix
#'
loo.matrix <-
  function(x,
           ...,
           r_eff = 1,
           save_psis = FALSE,
           cores = getOption("mc.cores", 1),
           is_method = c("psis", "tis", "sis")) {
    is_method <- match.arg(is_method)
    psis_out <-
      importance_sampling.matrix(
        log_ratios = -x,
        r_eff = r_eff,
        cores = cores,
        method = is_method
      )
    pointwise <- pointwise_loo_calcs(x, psis_out)
    importance_sampling_loo_object(
      pointwise = pointwise,
      diagnostics = psis_out$diagnostics,
      dims = dim(psis_out),
      is_method = is_method,
      is_object = if (save_psis) psis_out else NULL
    )
  }

#' @export
#' @templateVar fn loo
#' @template function
#' @param data,draws,... For the `loo.function()` method and the `loo_i()`
#'   function, these are the data, posterior draws, and other arguments to pass
#'   to the log-likelihood function. See the **Methods (by class)** section
#'   below for details on how to specify these arguments.
#'
loo.function <-
  function(x,
           ...,
           data = NULL,
           draws = NULL,
           r_eff = 1,
           save_psis = FALSE,
           cores = getOption("mc.cores", 1),
           is_method = c("psis", "tis", "sis")) {
    is_method <- match.arg(is_method)
    cores <- loo_cores(cores)
    stopifnot(is.data.frame(data) || is.matrix(data), !is.null(draws))
    assert_importance_sampling_method_is_implemented(is_method)
    .llfun <- validate_llfun(x)
    N <- dim(data)[1]

    r_eff <- prepare_psis_r_eff(r_eff, len = N)

    psis_list <-
      parallel_importance_sampling_list(
        N = N,
        .loo_i = .loo_i,
        .llfun = .llfun,
        data = data,
        draws = draws,
        r_eff = r_eff,
        save_psis = save_psis,
        cores = cores,
        method = is_method,
        ...
      )

    pointwise <- lapply(psis_list, "[[", "pointwise")
    if (save_psis) {
      psis_object_list <- lapply(psis_list, "[[", "psis_object")
      psis_out <- list2importance_sampling(psis_object_list)
      diagnostics <- psis_out$diagnostics
    } else {
      diagnostics_list <- lapply(psis_list, "[[", "diagnostics")
      diagnostics <- list(
        pareto_k = psis_apply(diagnostics_list, "pareto_k"),
        n_eff = psis_apply(diagnostics_list, "n_eff"),
        r_eff = psis_apply(diagnostics_list, "r_eff")
      )
    }

    importance_sampling_loo_object(
      pointwise = do.call(rbind, pointwise),
      diagnostics = diagnostics,
      dims = c(attr(psis_list[[1]], "S"), N),
      is_method = is_method,
      is_object = if (save_psis) psis_out else NULL
    )
  }


#' @description The `loo_i()` function enables testing log-likelihood
#'   functions for use with the `loo.function()` method.
#'
#' @rdname loo
#' @export
#'
#' @param i For `loo_i()`, an integer in `1:N`.
#' @param llfun For `loo_i()`, the same as `x` for the
#'   `loo.function()` method. A log-likelihood function as described in the
#'   **Methods (by class)** section.
#'
#' @return The `loo_i()` function returns a named list with components
#'   `pointwise` and `diagnostics`. These components have the same
#'   structure as the `pointwise` and `diagnostics` components of the
#'   object returned by `loo()` except they contain results for only a single
#'   observation.
#'
loo_i <-
  function(i,
           llfun,
           ...,
           data = NULL,
           draws = NULL,
           r_eff = 1,
           is_method = "psis"
           ) {
    stopifnot(
      i == as.integer(i),
      is.function(llfun) || is.character(llfun),
      is.data.frame(data) || is.matrix(data),
      i <= dim(data)[1],
      !is.null(draws),
      is_method %in% implemented_is_methods()
    )
    .loo_i(
      i = as.integer(i),
      llfun = match.fun(llfun),
      data = data,
      draws = draws,
      r_eff = r_eff[i],
      save_psis = FALSE,
      is_method = is_method,
      ...
    )
  }


# Function that is passed to the FUN argument of lapply, mclapply, or parLapply
# for the loo.function method. The arguments and return value are the same as
# the ones documented above for the user-facing loo_i function.
.loo_i <-
  function(i,
           llfun,
           ...,
           data,
           draws,
           r_eff = 1,
           save_psis = FALSE,
           is_method) {

    if (!is.null(r_eff)) {
      r_eff <- r_eff[i]
    }
    d_i <- data[i, , drop = FALSE]
    ll_i <- llfun(data_i = d_i, draws = draws, ...)
    if (!is.matrix(ll_i)) {
      ll_i <- as.matrix(ll_i)
    }
    psis_out <-
      importance_sampling.matrix(
        log_ratios = -ll_i,
        r_eff = r_eff,
        cores = 1,
        method = is_method
      )
    structure(
      list(
        pointwise = pointwise_loo_calcs(ll_i, psis_out),
        diagnostics = psis_out$diagnostics,
        psis_object = if (save_psis) psis_out else NULL
      ),
      S = dim(psis_out)[1],
      N = 1
    )
  }


#' @export
dim.loo <- function(x) {
  attr(x, "dims")
}

#' @rdname loo
#' @export
is.loo <- function(x) {
  inherits(x, "loo")
}

#' @export
dim.psis_loo <- function(x) {
  attr(x, "dims")
}

#' @rdname loo
#' @export
is.psis_loo <- function(x) {
  inherits(x, "psis_loo") && is.loo(x)
}


# internal ----------------------------------------------------------------

#' Compute pointwise elpd_loo, p_loo, looic from log lik matrix and
#' psis log weights
#'
#' @noRd
#' @param ll Log-likelihood matrix.
#' @param psis_object The object returned by `psis()`.
#' @return Named list with pointwise elpd_loo, mcse_elpd_loo, p_loo, looic,
#' and influence_pareto_k.
#'
pointwise_loo_calcs <- function(ll, psis_object) {
  if (!is.matrix(ll)) {
    ll <- as.matrix(ll)
  }
  lw <- weights(psis_object, normalize = TRUE, log = TRUE)
  elpd_loo <- matrixStats::colLogSumExps(ll + lw)
  lpd <- matrixStats::colLogSumExps(ll) - log(nrow(ll)) # colLogMeanExps
  p_loo <- lpd - elpd_loo
  mcse_elpd_loo <- mcse_elpd(ll, lw, E_elpd = elpd_loo, r_eff = relative_eff(psis_object))
  looic <- -2 * elpd_loo
  influence_pareto_k <- psis_object$diagnostics$pareto_k
  cbind(elpd_loo, mcse_elpd_loo, p_loo, looic, influence_pareto_k)
}

#' Structure the object returned by the loo methods
#'
#' @noRd
#' @param pointwise Matrix containing columns elpd_loo, mcse_elpd_loo, p_loo,
#'   looic, influence_pareto_k.
#' @param diagnostics Named list containing vector `pareto_k` and vector `n_eff`.
#' @param dims Log likelihood matrix dimensions (attribute of `"psis"` object).
#' @template is_method
#' @param is_object An object of class `"psis"/"tis"/"sis"`, as returned by the `psis()`/`tis()`/`sis()` function.
#' @return A `'importance_sampling_loo'` object as described in the Value section of the [loo()]
#'   function documentation.
#'
importance_sampling_loo_object <- function(pointwise, diagnostics, dims,
                                           is_method, is_object = NULL) {
  if (!is.matrix(pointwise)) stop("Internal error ('pointwise' must be a matrix)")
  if (!is.list(diagnostics)) stop("Internal error ('diagnostics' must be a list)")
  assert_importance_sampling_method_is_implemented(is_method)

  cols_to_summarize <- !(colnames(pointwise) %in% c("mcse_elpd_loo", "influence_pareto_k"))
  estimates <- table_of_estimates(pointwise[, cols_to_summarize, drop=FALSE])

  out <- nlist(estimates, pointwise, diagnostics)
  if (is.null(is_object)) {
    out[paste0(is_method, "_object")] <- list(NULL)
  } else {
    out[[paste0(is_method, "_object")]] <- is_object
  }

  # maintain backwards compatibility
  old_nms <- c("elpd_loo", "p_loo", "looic", "se_elpd_loo", "se_p_loo", "se_looic")
  out <- c(out, setNames(as.list(estimates), old_nms))

  structure(
    out,
    dims = dims,
    class = c(paste0(is_method, "_loo"), "importance_sampling_loo", "loo")
  )
}


#' Compute Monte Carlo standard error for ELPD
#'
#' @noRd
#' @param ll Log-likelihood matrix.
#' @param E_elpd elpd_loo column of pointwise matrix.
#' @param psis_object Object returned by [psis()].
#' @param n_samples Deprecated
#' @return Vector of standard error estimates.
#'
mcse_elpd <- function(ll, lw, E_elpd, r_eff, n_samples = NULL) {
  lik <- exp(ll)
  w2 <- exp(lw)^2
  E_epd <- exp(E_elpd)
  if (length(r_eff) == 1 && !is.null(ncol(ll))) {
    r_eff <- rep(r_eff, ncol(ll))
  }
  var_elpd <-
    vapply(
      seq_len(ncol(w2)),
      FUN.VALUE = numeric(1),
      FUN = function(i) {
        # Variance in linear scale
        # Equation (6) in Vehtari et al. (2024)
        var_epd_i <- sum(w2[, i] * (lik[, i] - E_epd[i]) ^ 2) / r_eff[i]
        # Compute variance in log scale by match the variance of a
        # log-normal approximation
        # https://en.wikipedia.org/wiki/Log-normal_distribution#Arithmetic_moments
        log(1 + var_epd_i / E_epd[i]^2)
      }
    )
  sqrt(var_elpd)
}


#' Warning message if r_eff not specified
#' @noRd
throw_loo_r_eff_warning <- function() {
  warning(
    "Relative effective sample sizes ('r_eff' argument) not specified.\n",
    "For models fit with MCMC, the reported PSIS ESS and \n",
    "MCSE estimates can be over-optimistic.",
    call. = FALSE
  )
}

#' Combine many psis objects into a single psis object
#'
#' @noRd
#' @param objects List of `"psis"` objects, each for a single observation.
#' @return A single `"psis"` object.
#'
list2importance_sampling <- function(objects) {
  log_weights <- sapply(objects, "[[", "log_weights")
  diagnostics <- lapply(objects, "[[", "diagnostics")

  method <- psis_apply(objects, "method", fun = "attr", fun_val = character(1))
  methods <- unique(method)
  if (length(methods) == 1) {
    method <- methods
    classes <- c(methods, "importance_sampling", "list")
  } else {
    classes <- c("importance_sampling", "list")
  }

  structure(
    list(
      log_weights = log_weights,
      diagnostics = list(
        pareto_k = psis_apply(diagnostics, item = "pareto_k"),
        n_eff = psis_apply(diagnostics, item = "n_eff"),
        r_eff = psis_apply(diagnostics, item = "r_eff")
      )
    ),
    norm_const_log = psis_apply(objects, "norm_const_log", fun = "attr"),
    tail_len = psis_apply(objects, "tail_len", fun = "attr"),
    r_eff = psis_apply(objects, "r_eff", fun = "attr"),
    dims = dim(log_weights),
    method = method,
    class = classes
  )
}

#' Extractor methods
#'
#' These are only defined in order to deprecate with a warning (rather than
#' remove and break backwards compatibility) the old way of accessing the point
#' estimates in a `"psis_loo"` or `"psis"` object. The new way as of
#' v2.0.0 is to get them from the `"estimates"` component of the object.
#'
#' @name old-extractors
#' @keywords internal
#' @param x,i,exact,name See \link{Extract}.
#'
NULL

#' @rdname old-extractors
#' @keywords internal
#' @export
`[.loo` <- function(x, i) {
  flags <- c("elpd_loo", "se_elpd_loo", "p_loo", "se_p_loo", "looic", "se_looic",
            "elpd_waic", "se_elpd_waic", "p_waic", "se_p_waic", "waic", "se_waic")

  if (is.character(i)) {
    needs_warning <- which(flags == i)
    if (length(needs_warning)) {
      warning(
        "Accessing ", flags[needs_warning], " using '[' is deprecated ",
        "and will be removed in a future release. ",
        "Please extract the ", flags[needs_warning],
        " estimate from the 'estimates' component instead.",
        call. = FALSE
      )
    }
  }
  NextMethod()
}

#' @rdname old-extractors
#' @keywords internal
#' @export
`[[.loo` <- function(x, i, exact=TRUE) {
  flags <- c("elpd_loo", "se_elpd_loo", "p_loo", "se_p_loo", "looic", "se_looic",
             "elpd_waic", "se_elpd_waic", "p_waic", "se_p_waic", "waic", "se_waic")

  if (is.character(i)) {
    needs_warning <- which(flags == i)
    if (length(needs_warning)) {
      warning(
        "Accessing ", flags[needs_warning], " using '[[' is deprecated ",
        "and will be removed in a future release. ",
        "Please extract the ", flags[needs_warning],
        " estimate from the 'estimates' component instead.",
        call. = FALSE
      )
    }
  }
  NextMethod()
}

#' @rdname old-extractors
#' @keywords internal
#' @export
#'
`$.loo` <- function(x, name) {
  flags <- c("elpd_loo", "se_elpd_loo", "p_loo", "se_p_loo", "looic", "se_looic",
             "elpd_waic", "se_elpd_waic", "p_waic", "se_p_waic", "waic", "se_waic")
  needs_warning <- which(flags == name)
  if (length(needs_warning)) {
    warning(
      "Accessing ", flags[needs_warning], " using '$' is deprecated ",
      "and will be removed in a future release. ",
      "Please extract the ", flags[needs_warning],
      " estimate from the 'estimates' component instead.",
      call. = FALSE
    )
  }
  NextMethod()
}


#' Parallel psis list computations
#'
#' @details Refactored function to handle parallel computations
#' for psis_list
#'
#' @keywords internal
#' @inheritParams loo.function
#' @param .loo_i The function used to compute individual loo contributions.
#' @param .llfun See `llfun` in [loo.function()].
#' @param N The total number of observations (i.e. `nrow(data)`).
#' @param method See `is_method` for [loo()]
#'
parallel_psis_list <- function(N, .loo_i, .llfun,
                               data, draws, r_eff,
                               save_psis, cores,
                               ...){
  parallel_importance_sampling_list(N, .loo_i, .llfun,
                                    data, draws, r_eff,
                                    save_psis, cores,
                                    method = "psis", ...)
}

#' @rdname parallel_psis_list
parallel_importance_sampling_list <- function(N, .loo_i, .llfun,
                                              data, draws, r_eff,
                                              save_psis, cores,
                                              method, ...){
  if (cores == 1) {
    psis_list <-
      lapply(
        X = seq_len(N),
        FUN = .loo_i,
        llfun = .llfun,
        data = data,
        draws = draws,
        r_eff = r_eff,
        save_psis = save_psis,
        is_method = method,
        ...
      )
  } else {
    if (!os_is_windows()) {
      # On Mac or Linux use mclapply() for multiple cores
      psis_list <-
        parallel::mclapply(
          mc.cores = cores,
          X = seq_len(N),
          FUN = .loo_i,
          llfun = .llfun,
          data = data,
          draws = draws,
          r_eff = r_eff,
          save_psis = save_psis,
          is_method = method,
          ...
        )
    } else {
      # On Windows use makePSOCKcluster() and parLapply() for multiple cores
      cl <- parallel::makePSOCKcluster(cores)
      on.exit(parallel::stopCluster(cl))
      psis_list <-
        parallel::parLapply(
          cl = cl,
          X = seq_len(N),
          fun = .loo_i,
          llfun = .llfun,
          data = data,
          draws = draws,
          r_eff = r_eff,
          save_psis = save_psis,
          is_method = method,
          ...
        )
    }
  }
}

loo1 <- suppressWarnings(loo(example_loglik_matrix()))

test_that("pointwise throws the right errors", {
  expect_error(
    pointwise(loo1, "xxx"),
    "'xxx' not found",
    fixed = TRUE
  )
  expect_error(
    pointwise(loo1, c("elpd_loo", "p_loo")),
    "length(estimate) == 1 is not TRUE",
    fixed = TRUE
  )
  expect_error(
    pointwise(loo1, 1),
    "is.character(estimate) is not TRUE",
    fixed = TRUE
  )
  loo1$pointwise <- NULL
  expect_error(
    pointwise(loo1, "xxx"),
    "No pointwise estimates found"
  )
})

test_that("pointwise returns correct estimate", {
  expect_equal(pointwise(loo1, "elpd_loo"), loo1$pointwise[, "elpd_loo"])
  expect_equal(
    pointwise(loo1, "mcse_elpd_loo"),
    loo1$pointwise[, "mcse_elpd_loo"]
  )
  expect_equal(pointwise(loo1, "p_loo"), loo1$pointwise[, "p_loo"])
  expect_equal(pointwise(loo1, "looic"), loo1$pointwise[, "looic"])
  expect_equal(
    pointwise(loo1, "influence_pareto_k"),
    loo1$pointwise[, "influence_pareto_k"]
  )
})

options(mc.cores = 1)
options(loo.cores = NULL)
set.seed(123)

LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()
LLvec <- LLmat[, 1]
chain_id <- rep(1:2, each = dim(LLarr)[1])
r_eff_arr <- relative_eff(exp(LLarr))
r_eff_vec <- relative_eff(exp(LLvec), chain_id = chain_id)
psis1 <- psis(log_ratios = -LLarr, r_eff = r_eff_arr)
tis1 <- tis(log_ratios = -LLarr, r_eff = r_eff_arr)
is1 <- sis(log_ratios = -LLarr, r_eff = r_eff_arr)


test_that("tis and is runs", {
  LLvec[1] <- -10
  expect_silent(tis1 <- tis(log_ratios = -LLvec, r_eff = r_eff_vec))
  expect_silent(is1 <- sis(log_ratios = -LLvec, r_eff = r_eff_vec))
  expect_failure(expect_equal(tis1$log_weights, is1$log_weights))
  expect_failure(expect_equal(tis1$log_weights, psis1$log_weights))
})

test_that("tis() and sis() returns object with correct structure for tis/sis", {
  expect_false(is.psis(tis1))
  expect_false(is.psis(is1))
  expect_true(is.tis(tis1))
  expect_false(is.tis(is1))
  expect_false(is.sis(tis1))
  expect_true(is.sis(is1))

  expect_false(is.loo(tis1))
  expect_false(is.loo(is1))

  expect_false(is.psis_loo(tis1))
  expect_false(is.psis_loo(is1))

  expect_named(tis1, c("log_weights", "diagnostics"))
  expect_named(is1, c("log_weights", "diagnostics"))

  expect_named(tis1$diagnostics, c("pareto_k", "n_eff", "r_eff"))
  expect_named(is1$diagnostics, c("pareto_k", "n_eff", "r_eff"))

  expect_equal(dim(tis1), dim(LLmat))
  expect_equal(dim(is1), dim(LLmat))

  expect_length(tis1$diagnostics$pareto_k, dim(psis1)[2])
  expect_length(is1$diagnostics$pareto_k, dim(psis1)[2])

  expect_length(tis1$diagnostics$n_eff, dim(psis1)[2])
  expect_length(is1$diagnostics$n_eff, dim(psis1)[2])

  expect_equal(attr(psis1, "method")[1], "psis")
  expect_equal(attr(tis1, "method")[1], "tis")
  expect_equal(attr(is1, "method")[1], "sis")
})


test_that("psis methods give same results", {
  tis2 <- suppressWarnings(tis(-LLmat, r_eff = r_eff_arr))
  expect_identical(tis1, tis2)

  tisvec <- suppressWarnings(tis(-LLvec, r_eff = r_eff_vec))
  tismat <- suppressWarnings(tis(-LLmat[, 1], r_eff = r_eff_vec))
  expect_identical(tisvec, tismat)

  is2 <- suppressWarnings(sis(-LLmat, r_eff = r_eff_arr))
  expect_identical(is1, is2)

  isvec <- suppressWarnings(sis(-LLvec, r_eff = r_eff_vec))
  ismat <- suppressWarnings(sis(-LLmat[, 1], r_eff = r_eff_vec))
  expect_identical(isvec, ismat)
})


test_that("tis throws correct errors and warnings", {
  # r_eff default no warnings
  expect_silent(tis(-LLarr))
  expect_silent(tis(-LLmat))
  expect_silent(tis(-LLmat[, 1]))

  # r_eff=NULL no warnings
  expect_silent(tis(-LLarr, r_eff = NULL))
  expect_silent(tis(-LLmat, r_eff = NULL))
  expect_silent(tis(-LLmat[, 1], r_eff = NULL))

  # r_eff=NA no warnings
  expect_silent(tis(-LLarr, r_eff = NA))
  expect_silent(tis(-LLmat, r_eff = NA))
  expect_silent(tis(-LLmat[, 1], r_eff = NA))

  # r_eff default and r_eff=NA give same answer
  expect_equal(
    suppressWarnings(tis(-LLarr)),
    tis(-LLarr, r_eff = NA)
  )

  # r_eff=NULL and r_eff=NA give same answer
  expect_equal(
    suppressWarnings(tis(-LLarr, r_eff = NULL)),
    tis(-LLarr, r_eff = NA)
  )

  # r_eff scalar is fine
  expect_silent(tis(-LLarr, r_eff = r_eff_arr[1]))

  # r_eff wrong length is error
  expect_error(tis(-LLarr, r_eff = r_eff_arr[-1]), "one value per observation")

  # r_eff has some NA values which are replaced with 1
  r_eff_arr[2] <- NA
  expect_snapshot(psis(-LLarr, r_eff = r_eff_arr))

  # no NAs or non-finite values allowed
  LLmat[1, 1] <- NA
  expect_error(tis(-LLmat), "NAs not allowed in input")

  LLmat[1, 1] <- 1
  LLmat[10, 2] <- -Inf
  expect_error(tis(-LLmat), "All input values must be finite or -Inf")
  LLmat[10, 2] <- Inf
  expect_no_error(tis(-LLmat))

  # no lists allowed
  expect_error(expect_warning(
    tis(as.list(-LLvec)),
    "List not allowed as input"
  ))

  # if array, must be 3-D array
  dim(LLarr) <- c(2, 250, 2, 32)
  expect_error(
    tis(-LLarr),
    "length(dim(log_ratios)) == 3 is not TRUE",
    fixed = TRUE
  )
})


test_that("explict test of values for 'sis' and 'tis'", {
  lw <- 1:16
  expect_silent(tis_true <- tis(log_ratios = lw, r_eff = NA))
  expect_equal(
    as.vector(weights(tis_true, log = TRUE, normalize = FALSE)),
    c(
      -14.0723,
      -13.0723,
      -12.0723,
      -11.0723,
      -10.0723,
      -9.0723,
      -8.0723,
      -7.0723,
      -6.0723,
      -5.0723,
      -4.0723,
      -3.0723,
      -2.0723,
      -1.0723,
      -0.0723,
      0.
    ) +
      15.07238,
    tolerance = 0.001
  )
  expect_silent(is_true <- sis(log_ratios = lw, r_eff = NA))
  expect_equal(
    as.vector(weights(is_true, log = TRUE, normalize = FALSE)),
    lw,
    tolerance = 0.00001
  )

  lw <- c(
    0.7609420,
    1.3894140,
    0.4158346,
    2.5307927,
    4.3379119,
    2.4159240,
    2.2462172,
    0.8057697,
    0.9333107,
    1.5599302
  )

  expect_silent(tis_true <- tis(log_ratios = lw, r_eff = NA))
  expect_equal(
    as.vector(weights(tis_true, log = TRUE, normalize = FALSE)),
    c(
      -2.931,
      -2.303,
      -3.276,
      -1.161,
      0,
      -1.276,
      -1.446,
      -2.886,
      -2.759,
      -2.132
    ) +
      3.692668,
    tolerance = 0.001
  )
  expect_silent(is_true <- sis(log_ratios = lw, r_eff = NA))
  expect_equal(
    as.vector(weights(is_true, log = TRUE, normalize = FALSE)),
    lw,
    tolerance = 0.00001
  )
})


test_that("tis_loo and sis_loo are returned", {
  LLmat <- example_loglik_matrix()
  loo_psis <- suppressWarnings(loo(LLmat, r_eff = NA, is_method = "psis"))
  loo_tis <- suppressWarnings(loo(LLmat, r_eff = NA, is_method = "tis"))
  loo_sis <- suppressWarnings(loo(LLmat, r_eff = NA, is_method = "sis"))

  expect_s3_class(loo_tis, "tis_loo")
  expect_s3_class(loo_sis, "sis_loo")
  expect_s3_class(loo_tis, "importance_sampling_loo")
  expect_s3_class(loo_sis, "importance_sampling_loo")

  expect_snapshot(print(loo_tis))
  expect_snapshot(print(loo_sis))
})

options(mc.cores = 1)
set.seed(123)

LLarr <- example_loglik_array()
r_eff_arr <- relative_eff(exp(LLarr))
loo1 <- suppressWarnings(loo(LLarr, r_eff = r_eff_arr))
waic1 <- suppressWarnings(waic(LLarr))
test_that("extracting estimates by name is deprecated for loo objects", {
  # $ method
  expect_snapshot(loo1$elpd_loo)
  expect_equal(
    suppressWarnings(loo1$elpd_loo),
    loo1$estimates["elpd_loo", "Estimate"]
  )
  expect_snapshot(loo1$se_elpd_loo)
  expect_equal(
    suppressWarnings(loo1$se_elpd_loo),
    loo1$estimates["elpd_loo", "SE"]
  )
  expect_snapshot(loo1$p_loo)
  expect_equal(
    suppressWarnings(loo1$p_loo),
    loo1$estimates["p_loo", "Estimate"]
  )
  expect_snapshot(loo1$se_p_loo)
  expect_equal(
    suppressWarnings(loo1$se_p_loo),
    loo1$estimates["p_loo", "SE"]
  )
  expect_snapshot(loo1$looic)
  expect_equal(
    suppressWarnings(loo1$looic),
    loo1$estimates["looic", "Estimate"]
  )
  expect_snapshot(loo1$se_looic)
  expect_equal(
    suppressWarnings(loo1$se_looic),
    loo1$estimates["looic", "SE"]
  )

  # [ method
  expect_snapshot(loo1["elpd_loo"])
  expect_equal(
    suppressWarnings(loo1["elpd_loo"][[1]]),
    loo1$estimates["elpd_loo", "Estimate"]
  )
  expect_snapshot(loo1["se_elpd_loo"])
  expect_equal(
    suppressWarnings(loo1["se_elpd_loo"][[1]]),
    loo1$estimates["elpd_loo", "SE"]
  )
  expect_snapshot(loo1["p_loo"])
  expect_equal(
    suppressWarnings(loo1["p_loo"][[1]]),
    loo1$estimates["p_loo", "Estimate"]
  )
  expect_snapshot(loo1["se_p_loo"])
  expect_equal(
    suppressWarnings(loo1["se_p_loo"][[1]]),
    loo1$estimates["p_loo", "SE"]
  )
  expect_snapshot(loo1["looic"])
  expect_equal(
    suppressWarnings(loo1["looic"][[1]]),
    loo1$estimates["looic", "Estimate"]
  )
  expect_snapshot(loo1["se_looic"])
  expect_equal(
    suppressWarnings(loo1["se_looic"][[1]]),
    loo1$estimates["looic", "SE"]
  )

  # [[ method
  expect_snapshot(loo1[["elpd_loo"]])
  expect_equal(
    suppressWarnings(loo1[["elpd_loo"]]),
    loo1$estimates["elpd_loo", "Estimate"]
  )
  expect_snapshot(loo1[["se_elpd_loo"]])
  expect_equal(
    suppressWarnings(loo1[["se_elpd_loo"]]),
    loo1$estimates["elpd_loo", "SE"]
  )
  expect_snapshot(loo1[["p_loo"]])
  expect_equal(
    suppressWarnings(loo1[["p_loo"]]),
    loo1$estimates["p_loo", "Estimate"]
  )
  expect_snapshot(loo1[["se_p_loo"]])
  expect_equal(
    suppressWarnings(loo1[["se_p_loo"]]),
    loo1$estimates["p_loo", "SE"]
  )
  expect_snapshot(loo1[["looic"]])
  expect_equal(
    suppressWarnings(loo1[["looic"]]),
    loo1$estimates["looic", "Estimate"]
  )
  expect_snapshot(loo1[["se_looic"]])
  expect_equal(
    suppressWarnings(loo1[["se_looic"]]),
    loo1$estimates["looic", "SE"]
  )
})

test_that("extracting estimates by name is deprecated for waic objects", {
  expect_snapshot(waic1$elpd_waic)
  expect_equal(
    suppressWarnings(waic1$elpd_waic),
    waic1$estimates["elpd_waic", "Estimate"]
  )
  expect_snapshot(waic1$se_elpd_waic)
  expect_equal(
    suppressWarnings(waic1$se_elpd_waic),
    waic1$estimates["elpd_waic", "SE"]
  )
  expect_snapshot(waic1$p_waic)
  expect_equal(
    suppressWarnings(waic1$p_waic),
    waic1$estimates["p_waic", "Estimate"]
  )
  expect_snapshot(waic1$se_p_waic)
  expect_equal(
    suppressWarnings(waic1$se_p_waic),
    waic1$estimates["p_waic", "SE"]
  )
  expect_snapshot(waic1$waic)
  expect_equal(
    suppressWarnings(waic1$waic),
    waic1$estimates["waic", "Estimate"]
  )
  expect_snapshot(waic1$se_waic)
  expect_equal(
    suppressWarnings(waic1$se_waic),
    waic1$estimates["waic", "SE"]
  )

  # [ method
  expect_snapshot(waic1["elpd_waic"])
  expect_equal(
    suppressWarnings(waic1["elpd_waic"][[1]]),
    waic1$estimates["elpd_waic", "Estimate"]
  )
  expect_snapshot(waic1["se_elpd_waic"])
  expect_equal(
    suppressWarnings(waic1["se_elpd_waic"][[1]]),
    waic1$estimates["elpd_waic", "SE"]
  )
  expect_snapshot(waic1["p_waic"])
  expect_equal(
    suppressWarnings(waic1["p_waic"][[1]]),
    waic1$estimates["p_waic", "Estimate"]
  )
  expect_snapshot(waic1["se_p_waic"])
  expect_equal(
    suppressWarnings(waic1["se_p_waic"][[1]]),
    waic1$estimates["p_waic", "SE"]
  )
  expect_snapshot(waic1["waic"])
  expect_equal(
    suppressWarnings(waic1["waic"][[1]]),
    waic1$estimates["waic", "Estimate"]
  )
  expect_snapshot(waic1["se_waic"])
  expect_equal(
    suppressWarnings(waic1["se_waic"][[1]]),
    waic1$estimates["waic", "SE"]
  )

  # [[ method
  expect_snapshot(waic1[["elpd_waic"]])
  expect_equal(
    suppressWarnings(waic1[["elpd_waic"]]),
    waic1$estimates["elpd_waic", "Estimate"]
  )
  expect_snapshot(waic1[["se_elpd_waic"]])
  expect_equal(
    suppressWarnings(waic1[["se_elpd_waic"]]),
    waic1$estimates["elpd_waic", "SE"]
  )
  expect_snapshot(waic1[["p_waic"]])
  expect_equal(
    suppressWarnings(waic1[["p_waic"]]),
    waic1$estimates["p_waic", "Estimate"]
  )
  expect_snapshot(waic1[["se_p_waic"]])
  expect_equal(
    suppressWarnings(waic1[["se_p_waic"]]),
    waic1$estimates["p_waic", "SE"]
  )
  expect_snapshot(waic1[["waic"]])
  expect_equal(
    suppressWarnings(waic1[["waic"]]),
    waic1$estimates["waic", "Estimate"]
  )
  expect_snapshot(waic1[["se_waic"]])
  expect_equal(
    suppressWarnings(waic1[["se_waic"]]),
    waic1$estimates["waic", "SE"]
  )
})

options(mc.cores = 1)
set.seed(123)

LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()
LLvec <- LLmat[, 1]
chain_id <- rep(1:2, each = nrow(LLarr))
r_eff_arr <- relative_eff(exp(LLarr))
r_eff_mat <- relative_eff(exp(LLmat), chain_id = chain_id)

loo1 <- suppressWarnings(loo(LLarr, r_eff = r_eff_arr))
waic1 <- suppressWarnings(waic(LLarr))
elpd1 <- suppressWarnings(elpd(LLarr))

test_that("using loo.cores is deprecated", {
  options(mc.cores = NULL)
  options(loo.cores = 1)
  expect_warning(loo(LLarr, r_eff = r_eff_arr, cores = 2), "loo.cores")
  options(loo.cores = NULL)
  options(mc.cores = 1)
})

test_that("loo, waic and elpd results haven't changed", {
  expect_snapshot_value(loo1, style = "serialize")
  expect_snapshot_value(waic1, style = "serialize")
  expect_snapshot_value(elpd1, style = "serialize")
})

test_that("loo with cores=1 and cores=2 gives same results", {
  loo2 <- suppressWarnings(loo(LLarr, r_eff = r_eff_arr, cores = 2))
  expect_equal(loo1$estimates, loo2$estimates)
})

test_that("waic returns object with correct structure", {
  expect_true(is.waic(waic1))
  expect_true(is.loo(waic1))
  expect_false(is.psis_loo(waic1))
  expect_named(
    waic1,
    c(
      "estimates",
      "pointwise",

      # deprecated but still there
      "elpd_waic",
      "p_waic",
      "waic",
      "se_elpd_waic",
      "se_p_waic",
      "se_waic"
    )
  )
  est_names <- dimnames(waic1$estimates)
  expect_equal(est_names[[1]], c("elpd_waic", "p_waic", "waic"))
  expect_equal(est_names[[2]], c("Estimate", "SE"))
  expect_equal(colnames(waic1$pointwise), est_names[[1]])
  expect_equal(dim(waic1), dim(LLmat))
})

test_that("loo returns object with correct structure", {
  expect_false(is.waic(loo1))
  expect_true(is.loo(loo1))
  expect_true(is.psis_loo(loo1))
  expect_named(
    loo1,
    c(
      "estimates",
      "pointwise",
      "diagnostics",
      "psis_object",

      # deprecated but still there
      "elpd_loo",
      "p_loo",
      "looic",
      "se_elpd_loo",
      "se_p_loo",
      "se_looic"
    )
  )
  expect_named(loo1$diagnostics, c("pareto_k", "n_eff", "r_eff"))
  expect_equal(dimnames(loo1$estimates)[[1]], c("elpd_loo", "p_loo", "looic"))
  expect_equal(dimnames(loo1$estimates)[[2]], c("Estimate", "SE"))
  expect_equal(
    colnames(loo1$pointwise),
    c("elpd_loo", "mcse_elpd_loo", "p_loo", "looic", "influence_pareto_k")
  )
  expect_equal(dim(loo1), dim(LLmat))
})


test_that("elpd returns object with correct structure", {
  expect_true(is.loo(elpd1))
  expect_named(
    elpd1,
    c(
      "estimates",
      "pointwise"
    )
  )
  est_names <- dimnames(elpd1$estimates)
  expect_equal(est_names[[1]], c("elpd", "ic"))
  expect_equal(est_names[[2]], c("Estimate", "SE"))
  expect_equal(colnames(elpd1$pointwise), est_names[[1]])
  expect_equal(dim(elpd1), dim(LLmat))
})


test_that("two pareto k values are equal", {
  expect_identical(
    loo1$pointwise[, "influence_pareto_k"],
    loo1$diagnostics$pareto_k
  )
})

test_that("loo.array and loo.matrix give same result", {
  l2 <- suppressWarnings(loo(LLmat, r_eff = r_eff_mat))
  expect_identical(loo1$estimates, l2$estimates)
  expect_identical(loo1$diagnostics, l2$diagnostics)

  # the mcse_elpd_loo columns won't be identical because we use sampling
  expect_identical(loo1$pointwise[, -2], l2$pointwise[, -2])
  expect_equal(loo1$pointwise[, 2], l2$pointwise[, 2], tolerance = 0.005)
})

test_that("loo.array runs with multiple cores", {
  loo_with_arr1 <- loo(LLarr, cores = 1, r_eff = NA)
  loo_with_arr2 <- loo(LLarr, cores = 2, r_eff = NA)
  expect_identical(loo_with_arr1$estimates, loo_with_arr2$estimates)
})

test_that("waic.array and waic.matrix give same result", {
  waic2 <- suppressWarnings(waic(LLmat))
  expect_identical(waic1, waic2)
})

test_that("elpd.array and elpd.matrix give same result", {
  elpd2 <- suppressWarnings(elpd(LLmat))
  expect_identical(elpd1, elpd2)
})

test_that("loo, waic, and elpd error with vector input", {
  expect_error(loo(LLvec), regexp = "no applicable method")
  expect_error(waic(LLvec), regexp = "no applicable method")
  expect_error(elpd(LLvec), regexp = "no applicable method")
})


# testing function methods ------------------------------------------------
source(test_path("data-for-tests/function_method_stuff.R"))

waic_with_fn <- waic(llfun, data = data, draws = draws)
waic_with_mat <- waic(llmat_from_fn)

loo_with_fn <- loo(
  llfun,
  data = data,
  draws = draws,
  r_eff = rep(1, nrow(data))
)
loo_with_mat <- loo(
  llmat_from_fn,
  r_eff = rep(1, ncol(llmat_from_fn)),
  save_psis = TRUE
)

test_that("loo.cores deprecation warning works with function method", {
  options(loo.cores = 1)
  expect_warning(
    loo(
      llfun,
      cores = 2,
      data = data,
      draws = draws,
      r_eff = rep(1, nrow(data))
    ),
    "loo.cores"
  )
  options(loo.cores = NULL)
})

test_that("loo_i results match loo results for ith data point", {
  expect_no_warning(
    loo_i_val <- loo_i(i = 2, llfun = llfun, data = data, draws = draws),
  )
  expect_equal(
    loo_i_val$pointwise[, "elpd_loo"],
    loo_with_fn$pointwise[2, "elpd_loo"]
  )
  expect_equal(
    loo_i_val$pointwise[, "p_loo"],
    loo_with_fn$pointwise[2, "p_loo"]
  )
  expect_equal(
    loo_i_val$diagnostics$pareto_k,
    loo_with_fn$diagnostics$pareto_k[2]
  )
  expect_equal(loo_i_val$diagnostics$n_eff, loo_with_fn$diagnostics$n_eff[2])
})

test_that("function and matrix methods return same result", {
  expect_equal(waic_with_mat, waic_with_fn)
  expect_identical(loo_with_mat$estimates, loo_with_fn$estimates)
  expect_identical(loo_with_mat$diagnostics, loo_with_fn$diagnostics)
  expect_identical(dim(loo_with_mat), dim(loo_with_fn))
})

test_that("loo.function runs with multiple cores", {
  loo_with_fn1 <- loo(
    llfun,
    data = data,
    draws = draws,
    r_eff = rep(1, nrow(data)),
    cores = 1
  )
  loo_with_fn2 <- loo(
    llfun,
    data = data,
    draws = draws,
    r_eff = rep(1, nrow(data)),
    cores = 2
  )
  expect_identical(loo_with_fn2$estimates, loo_with_fn1$estimates)
})

test_that("save_psis option to loo.function makes correct psis object", {
  loo_with_fn2 <- loo.function(
    llfun,
    data = data,
    draws = draws,
    r_eff = rep(1, nrow(data)),
    save_psis = TRUE
  )
  expect_identical(loo_with_fn2$psis_object, loo_with_mat$psis_object)
})

test_that("loo doesn't throw r_eff warnings", {
  expect_no_warning(loo(-LLarr))
  expect_no_warning(loo(-LLmat))
  expect_no_warning(loo(llfun, data = data, draws = draws))
})

library(loo)

context("pointwise convenience function")

loo1 <- suppressWarnings(loo(example_loglik_matrix()))

test_that("pointwise throws the right errors", {
  expect_error(
    pointwise(loo1, "xxx"),
    "'xxx' not found",
    fixed = TRUE
  )
  expect_error(
    pointwise(loo1, c("elpd_loo", "p_loo")),
    "length(estimate) == 1 is not TRUE",
    fixed = TRUE
  )
  expect_error(
    pointwise(loo1, 1),
    "is.character(estimate) is not TRUE",
    fixed = TRUE
  )
  loo1$pointwise <- NULL
  expect_error(
    pointwise(loo1, "xxx"),
    "No pointwise estimates found"
  )
})

test_that("pointwise returns correct estimate", {
  expect_equal(pointwise(loo1, "elpd_loo"), loo1$pointwise[, "elpd_loo"])
  expect_equal(pointwise(loo1, "mcse_elpd_loo"), loo1$pointwise[, "mcse_elpd_loo"])
  expect_equal(pointwise(loo1, "p_loo"), loo1$pointwise[, "p_loo"])
  expect_equal(pointwise(loo1, "looic"), loo1$pointwise[, "looic"])
  expect_equal(pointwise(loo1, "influence_pareto_k"), loo1$pointwise[, "influence_pareto_k"])
})

library(loo)
options(mc.cores=1)
options(loo.cores=NULL)
set.seed(123)

context("tis and is")

LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()
LLvec <- LLmat[, 1]
chain_id <- rep(1:2, each = dim(LLarr)[1])
r_eff_arr <- relative_eff(exp(LLarr))
r_eff_vec <- relative_eff(exp(LLvec), chain_id = chain_id)
psis1 <- psis(log_ratios = -LLarr, r_eff = r_eff_arr)
tis1 <- tis(log_ratios = -LLarr, r_eff = r_eff_arr)
is1 <- sis(log_ratios = -LLarr, r_eff = r_eff_arr)


test_that("tis and is runs", {
  LLvec[1] <- -10
  expect_silent(tis1 <- tis(log_ratios = -LLvec, r_eff = r_eff_vec))
  expect_silent(is1 <- sis(log_ratios = -LLvec, r_eff = r_eff_vec))
  expect_failure(expect_equal(tis1$log_weights, is1$log_weights))
  expect_failure(expect_equal(tis1$log_weights, psis1$log_weights))
})

test_that("tis() and sis() returns object with correct structure for tis/sis", {

  expect_false(is.psis(tis1))
  expect_false(is.psis(is1))
  expect_true(is.tis(tis1))
  expect_false(is.tis(is1))
  expect_false(is.sis(tis1))
  expect_true(is.sis(is1))

  expect_false(is.loo(tis1))
  expect_false(is.loo(is1))

  expect_false(is.psis_loo(tis1))
  expect_false(is.psis_loo(is1))

  expect_named(tis1, c("log_weights", "diagnostics"))
  expect_named(is1, c("log_weights", "diagnostics"))

  expect_named(tis1$diagnostics, c("pareto_k", "n_eff", "r_eff"))
  expect_named(is1$diagnostics, c("pareto_k", "n_eff", "r_eff"))

  expect_equal(dim(tis1), dim(LLmat))
  expect_equal(dim(is1), dim(LLmat))

  expect_length(tis1$diagnostics$pareto_k, dim(psis1)[2])
  expect_length(is1$diagnostics$pareto_k, dim(psis1)[2])

  expect_length(tis1$diagnostics$n_eff, dim(psis1)[2])
  expect_length(is1$diagnostics$n_eff, dim(psis1)[2])

  expect_equal(attr(psis1, "method")[1], "psis")
  expect_equal(attr(tis1, "method")[1], "tis")
  expect_equal(attr(is1, "method")[1], "sis")
})


test_that("psis methods give same results", {
  tis2 <- suppressWarnings(tis(-LLmat, r_eff = r_eff_arr))
  expect_identical(tis1, tis2)

  tisvec <- suppressWarnings(tis(-LLvec, r_eff = r_eff_vec))
  tismat <- suppressWarnings(tis(-LLmat[, 1], r_eff = r_eff_vec))
  expect_identical(tisvec, tismat)

  is2 <- suppressWarnings(sis(-LLmat, r_eff = r_eff_arr))
  expect_identical(is1, is2)

  isvec <- suppressWarnings(sis(-LLvec, r_eff = r_eff_vec))
  ismat <- suppressWarnings(sis(-LLmat[, 1], r_eff = r_eff_vec))
  expect_identical(isvec, ismat)
})



test_that("tis throws correct errors and warnings", {
  # r_eff default no warnings
  expect_silent(tis(-LLarr))
  expect_silent(tis(-LLmat))
  expect_silent(tis(-LLmat[, 1]))

  # r_eff=NULL no warnings
  expect_silent(tis(-LLarr, r_eff = NULL))
  expect_silent(tis(-LLmat, r_eff = NULL))
  expect_silent(tis(-LLmat[,1], r_eff = NULL))

  # r_eff=NA no warnings
  expect_silent(tis(-LLarr, r_eff = NA))
  expect_silent(tis(-LLmat, r_eff = NA))
  expect_silent(tis(-LLmat[,1], r_eff = NA))

  # r_eff default and r_eff=NA give same answer
  expect_equal(
    suppressWarnings(tis(-LLarr)),
                     tis(-LLarr, r_eff = NA)
  )

  # r_eff=NULL and r_eff=NA give same answer
  expect_equal(
    suppressWarnings(tis(-LLarr, r_eff = NULL)),
                     tis(-LLarr, r_eff = NA)
  )

  # r_eff scalar is fine
  expect_silent(tis(-LLarr, r_eff = r_eff_arr[1]))

  # r_eff wrong length is error
  expect_error(tis(-LLarr, r_eff = r_eff_arr[-1]), "one value per observation")

  # r_eff has some NA values causes error
  r_eff_arr[2] <- NA
  expect_error(tis(-LLarr, r_eff = r_eff_arr), "mix NA and not NA values")

  # no NAs or non-finite values allowed
  LLmat[1,1] <- NA
  expect_error(tis(-LLmat), "NAs not allowed in input")

  LLmat[1,1] <- 1
  LLmat[10, 2] <- -Inf
  expect_error(tis(-LLmat), "All input values must be finite or -Inf")
  LLmat[10, 2] <- Inf
  expect_no_error(tis(-LLmat))

  # no lists allowed
  expect_error(expect_warning(tis(as.list(-LLvec)), "List not allowed as input"))

  # if array, must be 3-D array
  dim(LLarr) <- c(2, 250, 2, 32)
  expect_error(
    tis(-LLarr),
    "length(dim(log_ratios)) == 3 is not TRUE",
    fixed = TRUE
  )
})


test_that("explict test of values for 'sis' and 'tis'", {
  lw <- 1:16
  expect_silent(tis_true <- tis(log_ratios = lw, r_eff = NA))
  expect_equal(as.vector(weights(tis_true, log = TRUE, normalize = FALSE)),
               c(-14.0723, -13.0723, -12.0723, -11.0723, -10.0723, -9.0723, -8.0723, -7.0723, -6.0723, -5.0723, -4.0723, -3.0723, -2.0723, -1.0723, -0.0723, 0.) + 15.07238, tol = 0.001)
  expect_silent(is_true <- sis(log_ratios = lw, r_eff = NA))
  expect_equal(as.vector(weights(is_true, log = TRUE, normalize = FALSE)),
               lw, tol = 0.00001)

  lw <- c(0.7609420, 1.3894140, 0.4158346, 2.5307927, 4.3379119, 2.4159240, 2.2462172, 0.8057697, 0.9333107, 1.5599302)

  expect_silent(tis_true <- tis(log_ratios = lw, r_eff = NA))
  expect_equal(as.vector(weights(tis_true, log = TRUE, normalize = FALSE)),
               c(-2.931, -2.303, -3.276, -1.161,  0, -1.276, -1.446, -2.886, -2.759, -2.132) + 3.692668,
               tol = 0.001)
  expect_silent(is_true <- sis(log_ratios = lw, r_eff = NA))
  expect_equal(as.vector(weights(is_true, log = TRUE, normalize = FALSE)),
               lw, tol = 0.00001)
})


test_that("tis_loo and sis_loo are returned", {
  LLmat <- example_loglik_matrix()
  loo_psis <- suppressWarnings(loo(LLmat, r_eff = NA, is_method = "psis"))
  loo_tis <- suppressWarnings(loo(LLmat, r_eff = NA, is_method = "tis"))
  loo_sis <- suppressWarnings(loo(LLmat, r_eff = NA, is_method = "sis"))

  expect_s3_class(loo_tis, "tis_loo")
  expect_s3_class(loo_sis, "sis_loo")
  expect_s3_class(loo_tis, "importance_sampling_loo")
  expect_s3_class(loo_sis, "importance_sampling_loo")

  expect_output(print(loo_tis), regexp = "tis_loo")
  expect_output(print(loo_sis), regexp = "sis_loo")
})

library(loo)
options(mc.cores = 1)
set.seed(123)

context("Depracted extractors")

LLarr <- example_loglik_array()
r_eff_arr <- relative_eff(exp(LLarr))
loo1 <- suppressWarnings(loo(LLarr, r_eff = r_eff_arr))
waic1 <- suppressWarnings(waic(LLarr))

expect_warning_fixed <- function(object, regexp = NULL) {
  expect_warning(object, regexp = regexp, fixed = TRUE)
}

test_that("extracting estimates by name is deprecated for loo objects", {
  # $ method
  expect_equal(
    expect_warning_fixed(loo1$elpd_loo, "elpd_loo using '$' is deprecated"),
    loo1$estimates["elpd_loo", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(loo1$se_elpd_loo, "se_elpd_loo using '$' is deprecated"),
    loo1$estimates["elpd_loo", "SE"]
  )
  expect_equal(
    expect_warning_fixed(loo1$p_loo, "p_loo using '$' is deprecated"),
    loo1$estimates["p_loo", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(loo1$se_p_loo, "se_p_loo using '$' is deprecated"),
    loo1$estimates["p_loo", "SE"]
  )
  expect_equal(
    expect_warning_fixed(loo1$looic, "looic using '$' is deprecated"),
    loo1$estimates["looic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(loo1$se_looic, "se_looic using '$' is deprecated"),
    loo1$estimates["looic", "SE"]
  )

  # [ method
  expect_equal(
    expect_warning_fixed(loo1["elpd_loo"], "elpd_loo using '[' is deprecated")[[1]],
    loo1$estimates["elpd_loo", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(loo1["se_elpd_loo"], "se_elpd_loo using '[' is deprecated")[[1]],
    loo1$estimates["elpd_loo", "SE"]
  )
  expect_equal(
    expect_warning_fixed(loo1["p_loo"], "p_loo using '[' is deprecated")[[1]],
    loo1$estimates["p_loo", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(loo1["se_p_loo"], "se_p_loo using '[' is deprecated")[[1]],
    loo1$estimates["p_loo", "SE"]
  )
  expect_equal(
    expect_warning_fixed(loo1["looic"], "looic using '[' is deprecated")[[1]],
    loo1$estimates["looic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(loo1["se_looic"], "se_looic using '[' is deprecated")[[1]],
    loo1$estimates["looic", "SE"]
  )


  # [[ method
  expect_equal(
    expect_warning_fixed(loo1[["elpd_loo"]], "elpd_loo using '[[' is deprecated"),
    loo1$estimates["elpd_loo", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(loo1[["se_elpd_loo"]], "se_elpd_loo using '[[' is deprecated"),
    loo1$estimates["elpd_loo", "SE"]
  )
  expect_equal(
    expect_warning_fixed(loo1[["p_loo"]], "p_loo using '[[' is deprecated"),
    loo1$estimates["p_loo", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(loo1[["se_p_loo"]], "se_p_loo using '[[' is deprecated"),
    loo1$estimates["p_loo", "SE"]
  )
  expect_equal(
    expect_warning_fixed(loo1[["looic"]], "looic using '[[' is deprecated"),
    loo1$estimates["looic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(loo1[["se_looic"]], "se_looic using '[[' is deprecated"),
    loo1$estimates["looic", "SE"]
  )
})

test_that("extracting estimates by name is deprecated for waic objects", {
  expect_equal(
    expect_warning_fixed(waic1$elpd_waic, "elpd_waic using '$' is deprecated"),
    waic1$estimates["elpd_waic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(waic1$se_elpd_waic, "se_elpd_waic using '$' is deprecated"),
    waic1$estimates["elpd_waic", "SE"]
  )
  expect_equal(
    expect_warning_fixed(waic1$p_waic, "p_waic using '$' is deprecated"),
    waic1$estimates["p_waic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(waic1$se_p_waic, "se_p_waic using '$' is deprecated"),
    waic1$estimates["p_waic", "SE"]
  )
  expect_equal(
    expect_warning_fixed(waic1$waic, "waic using '$' is deprecated"),
    waic1$estimates["waic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(waic1$se_waic, "se_waic using '$' is deprecated"),
    waic1$estimates["waic", "SE"]
  )


  expect_equal(
    expect_warning_fixed(waic1["elpd_waic"], "elpd_waic using '[' is deprecated")[[1]],
    waic1$estimates["elpd_waic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(waic1["se_elpd_waic"], "se_elpd_waic using '[' is deprecated")[[1]],
    waic1$estimates["elpd_waic", "SE"]
  )
  expect_equal(
    expect_warning_fixed(waic1["p_waic"], "p_waic using '[' is deprecated")[[1]],
    waic1$estimates["p_waic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(waic1["se_p_waic"], "se_p_waic using '[' is deprecated")[[1]],
    waic1$estimates["p_waic", "SE"]
  )
  expect_equal(
    expect_warning_fixed(waic1["waic"], "waic using '[' is deprecated")[[1]],
    waic1$estimates["waic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(waic1["se_waic"], "se_waic using '[' is deprecated")[[1]],
    waic1$estimates["waic", "SE"]
  )


  expect_equal(
    expect_warning_fixed(waic1[["elpd_waic"]], "elpd_waic using '[[' is deprecated"),
    waic1$estimates["elpd_waic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(waic1[["se_elpd_waic"]], "se_elpd_waic using '[[' is deprecated"),
    waic1$estimates["elpd_waic", "SE"]
  )
  expect_equal(
    expect_warning_fixed(waic1[["p_waic"]], "p_waic using '[[' is deprecated"),
    waic1$estimates["p_waic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(waic1[["se_p_waic"]], "se_p_waic using '[[' is deprecated"),
    waic1$estimates["p_waic", "SE"]
  )
  expect_equal(
    expect_warning_fixed(waic1[["waic"]], "waic using '[[' is deprecated"),
    waic1$estimates["waic", "Estimate"]
  )
  expect_equal(
    expect_warning_fixed(waic1[["se_waic"]], "se_waic using '[[' is deprecated"),
    waic1$estimates["waic", "SE"]
  )
})

#' Objects to use in examples and tests
#'
#' Example pointwise log-likelihood objects to use in demonstrations and tests.
#' See the **Value** and **Examples** sections below.
#'
#' @export
#' @return
#' `example_loglik_array()` returns a 500 (draws) x 2 (chains) x 32
#' (observations) pointwise log-likelihood array.
#'
#' `example_loglik_matrix()` returns the same pointwise log-likelihood values
#' as `example_loglik_array()` but reshaped into a 1000 (draws*chains) x 32
#' (observations) matrix.
#'
#' @examples
#' LLarr <- example_loglik_array()
#' (dim_arr <- dim(LLarr))
#' LLmat <- example_loglik_matrix()
#' (dim_mat <- dim(LLmat))
#'
#' all.equal(dim_mat[1], dim_arr[1] * dim_arr[2])
#' all.equal(dim_mat[2], dim_arr[3])
#'
#' all.equal(LLarr[, 1, ], LLmat[1:500, ])
#' all.equal(LLarr[, 2, ], LLmat[501:1000, ])
#'
example_loglik_array <- function() {
  # .example_loglik_array exists in R/sysdata.R
  return(.example_loglik_array)
}

#' @rdname example_loglik_array
#' @export
example_loglik_matrix <- function() {
  ll <- example_loglik_array()
  return(llarray_to_matrix(ll))
}

options(mc.cores = 1)
set.seed(123)

LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()

test_that("relative_eff results haven't changed", {
  expect_snapshot_value(relative_eff(exp(LLarr)), style = "serialize")
})

test_that("relative_eff is equal to ESS / S", {
  dims <- dim(LLarr)
  ess <- r_eff <- rep(NA, dims[3])
  for (j in 1:dims[3]) {
    r_eff[j] <- relative_eff(LLarr[,, 1, drop = FALSE])
    ess[j] <- posterior::ess_mean(LLarr[,, 1])
  }
  S <- prod(dim(LLarr)[1:2])
  expect_equal(r_eff, ess / S)
})

test_that("relative_eff array and matrix methods return identical output", {
  r_eff_arr <- relative_eff(exp(LLarr))
  r_eff_mat <- relative_eff(exp(LLmat), chain_id = rep(1:2, each = nrow(LLarr)))
  expect_identical(r_eff_arr, r_eff_mat)
})

test_that("relative_eff matrix and function methods return identical output", {
  source(test_path("data-for-tests/function_method_stuff.R"))
  chain <- rep(1, nrow(draws))
  r_eff_mat <- relative_eff(llmat_from_fn, chain_id = chain)
  r_eff_fn <- relative_eff(
    llfun,
    chain_id = chain,
    data = data,
    draws = draws,
    cores = 1
  )
  expect_identical(r_eff_mat, r_eff_fn)
})

test_that("relative_eff with multiple cores runs", {
  skip_on_cran()
  source(test_path("data-for-tests/function_method_stuff.R"))
  dim(llmat_from_fn) <- c(nrow(llmat_from_fn), 1, ncol(llmat_from_fn))
  r_eff_arr <- relative_eff(llmat_from_fn, cores = 2)
  r_eff_fn <-
    relative_eff(
      llfun,
      chain_id = rep(1, nrow(draws)),
      data = data,
      draws = draws,
      cores = 2
    )
  expect_identical(r_eff_arr, r_eff_fn)
})

library(loo)
options(mc.cores = 1)
set.seed(123)

context("relative_eff methods")

LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()

test_that("relative_eff results haven't changed", {
  expect_equal_to_reference(relative_eff(exp(LLarr)), "reference-results/relative_eff.rds")
})

test_that("relative_eff is equal to ESS / S", {
  dims <- dim(LLarr)
  ess <- r_eff <- rep(NA, dims[3])
  for (j in 1:dims[3]) {
    r_eff[j] <- relative_eff(LLarr[,,1, drop=FALSE])
    ess[j] <- ess_rfun(LLarr[,,1])
  }
  S <- prod(dim(LLarr)[1:2])
  expect_equal(r_eff, ess / S)
})

test_that("relative_eff array and matrix methods return identical output", {
  r_eff_arr <- relative_eff(exp(LLarr))
  r_eff_mat <- relative_eff(exp(LLmat), chain_id = rep(1:2, each = nrow(LLarr)))
  expect_identical(r_eff_arr, r_eff_mat)
})

test_that("relative_eff matrix and function methods return identical output", {
  source(test_path("data-for-tests/function_method_stuff.R"))
  chain <- rep(1, nrow(draws))
  r_eff_mat <- relative_eff(llmat_from_fn, chain_id = chain)
  r_eff_fn <- relative_eff(llfun, chain_id = chain, data = data, draws = draws, cores = 1)
  expect_identical(r_eff_mat, r_eff_fn)
})

test_that("relative_eff with multiple cores runs", {
  skip_on_cran()
  source(test_path("data-for-tests/function_method_stuff.R"))
  dim(llmat_from_fn) <- c(nrow(llmat_from_fn), 1, ncol(llmat_from_fn))
  r_eff_arr <- relative_eff(llmat_from_fn, cores = 2)
  r_eff_fn <-
    relative_eff(
      llfun,
      chain_id = rep(1, nrow(draws)),
      data = data,
      draws = draws,
      cores = 2
    )
  expect_identical(r_eff_arr, r_eff_fn)
})

#' Efficient LOO-CV and WAIC for Bayesian models
#'
#' @importFrom stats sd var quantile setNames weights rnorm qnorm
#' @importFrom matrixStats logSumExp colLogSumExps colSums2 colVars colMaxs
#'
#' @description
#' \if{html}{
#'   \figure{stanlogo.png}{options: width="50" alt="mc-stan.org"}
#' }
#' *Stan Development Team*
#'
#' This package implements the methods described in Vehtari, Gelman, and
#' Gabry (2017), Vehtari, Simpson, Gelman, Yao, and Gabry (2024), and
#' Yao et al. (2018). To get started see the **loo** package
#' [vignettes](https://mc-stan.org/loo/articles/index.html), the
#' [loo()] function for efficient approximate leave-one-out
#' cross-validation (LOO-CV), the [psis()] function for the Pareto
#' smoothed importance sampling (PSIS) algorithm, or
#' [loo_model_weights()] for an implementation of Bayesian stacking of
#' predictive distributions from multiple models.
#'
#'
#' @details Leave-one-out cross-validation (LOO-CV) and the widely applicable
#'   information criterion (WAIC) are methods for estimating pointwise
#'   out-of-sample prediction accuracy from a fitted Bayesian model using the
#'   log-likelihood evaluated at the posterior simulations of the parameter
#'   values. LOO-CV and WAIC have various advantages over simpler estimates of
#'   predictive error such as AIC and DIC but are less used in practice because
#'   they involve additional computational steps. This package implements the
#'   fast and stable computations for approximate LOO-CV laid out in Vehtari,
#'   Gelman, and Gabry (2017). From existing posterior simulation draws, we
#'   compute LOO-CV using Pareto smoothed importance sampling (PSIS; Vehtari,
#'   Simpson, Gelman, Yao, and Gabry, 2024), a new procedure for stabilizing
#'   and diagnosing importance weights. As a byproduct of our calculations,
#'   we also obtain approximate standard errors for estimated predictive
#'   errors and for comparing of predictive errors between two models.
#'
#'   We recommend PSIS-LOO-CV instead of WAIC, because PSIS provides useful
#'   diagnostics and effective sample size and Monte Carlo standard error
#'   estimates.
#'
#'
#' @template loo-and-psis-references
#' @template stacking-references
#' @template loo-large-data-references
#'
#' @references
#' Epifani, I., MacEachern, S. N., and Peruggia, M. (2008). Case-deletion
#' importance sampling estimators: Central limit theorems and related results.
#' *Electronic Journal of Statistics* **2**, 774-806.
#'
#' Gelfand, A. E. (1996). Model determination using sampling-based methods. In
#' *Markov Chain Monte Carlo in Practice*, ed. W. R. Gilks, S. Richardson,
#' D. J. Spiegelhalter, 145-162. London: Chapman and Hall.
#'
#' Gelfand, A. E., Dey, D. K., and Chang, H. (1992). Model determination using
#' predictive distributions with implementation via sampling-based methods. In
#' *Bayesian Statistics 4*, ed. J. M. Bernardo, J. O. Berger, A. P. Dawid,
#' and A. F. M. Smith, 147-167. Oxford University Press.
#'
#' Gelman, A., Hwang, J., and Vehtari, A. (2014). Understanding predictive
#' information criteria for Bayesian models. *Statistics and Computing*
#' **24**, 997-1016.
#'
#' Ionides, E. L. (2008). Truncated importance sampling. *Journal of
#' Computational and Graphical Statistics* **17**, 295-311.
#'
#' Koopman, S. J., Shephard, N., and Creal, D. (2009). Testing the assumptions
#' behind importance sampling. *Journal of Econometrics* **149**, 2-11.
#'
#' Peruggia, M. (1997). On the variability of case-deletion importance sampling
#' weights in the Bayesian linear model. *Journal of the American
#' Statistical Association* **92**, 199-207.
#'
#' Stan Development Team (2017). The Stan C++ Library, Version 2.17.0.
#' <https://mc-stan.org>.
#'
#' Stan Development Team (2018). RStan: the R interface to Stan, Version 2.17.3.
#' <https://mc-stan.org>.
#'
#' Watanabe, S. (2010). Asymptotic equivalence of Bayes cross validation and
#' widely applicable information criterion in singular learning theory.
#' *Journal of Machine Learning Research* **11**, 3571-3594.
#'
#' Zhang, J., and Stephens, M. A. (2009). A new and efficient estimation method
#' for the generalized Pareto distribution. *Technometrics* **51**,
#' 316-325.
#'
"_PACKAGE"

#' Convenience function for computing relative efficiencies
#'
#' `relative_eff()` computes the the MCMC effective sample size divided by
#' the total sample size.
#'
#' @export
#' @param x A vector, matrix, 3-D array, or function. See the **Methods (by
#'   class)** section below for details on specifying `x`, but where
#'   "log-likelihood" is mentioned replace it with one of the following
#'   depending on the use case:
#'   * For use with the [loo()] function, the values in `x` (or generated by
#'     `x`, if a function) should be **likelihood** values
#'     (i.e., `exp(log_lik)`), not on the log scale.
#'   * For generic use with [psis()], the values in `x` should be the reciprocal
#'     of the importance ratios (i.e., `exp(-log_ratios)`).
#' @param chain_id A vector of length `NROW(x)` containing MCMC chain
#'   indexes for each each row of `x` (if a matrix) or each value in
#'   `x` (if a vector). No `chain_id` is needed if `x` is a 3-D
#'   array. If there are `C` chains then valid chain indexes are values
#'   in `1:C`.
#' @param cores The number of cores to use for parallelization.
#' @return A vector of relative effective sample sizes.
#'
#' @examples
#' LLarr <- example_loglik_array()
#' LLmat <- example_loglik_matrix()
#' dim(LLarr)
#' dim(LLmat)
#'
#' rel_n_eff_1 <- relative_eff(exp(LLarr))
#' rel_n_eff_2 <- relative_eff(exp(LLmat), chain_id = rep(1:2, each = 500))
#' all.equal(rel_n_eff_1, rel_n_eff_2)
#'
relative_eff <- function(x, ...) {
  UseMethod("relative_eff")
}

#' @export
#' @templateVar fn relative_eff
#' @template vector
#'
relative_eff.default <- function(x, chain_id, ...) {
  dim(x) <- c(length(x), 1)
  class(x) <- "matrix"
  relative_eff.matrix(x, chain_id)
}

#' @export
#' @templateVar fn relative_eff
#' @template matrix
#'
relative_eff.matrix <- function(x, chain_id, ..., cores = getOption("mc.cores", 1)) {
  x <- llmatrix_to_array(x, chain_id)
  relative_eff.array(x, cores = cores)
}

#' @export
#' @templateVar fn relative_eff
#' @template array
#'
relative_eff.array <- function(x, ..., cores = getOption("mc.cores", 1)) {
  stopifnot(length(dim(x)) == 3)
  S <- prod(dim(x)[1:2]) # posterior sample size = iter * chains

  if (cores == 1) {
    n_eff_vec <- apply(x, 3, posterior::ess_mean)
  } else {
    if (!os_is_windows()) {
      n_eff_list <-
        parallel::mclapply(
          mc.cores = cores,
          X = seq_len(dim(x)[3]),
          FUN = function(i) posterior::ess_mean(x[, , i, drop = TRUE])
        )
    } else {
      cl <- parallel::makePSOCKcluster(cores)
      on.exit(parallel::stopCluster(cl))
      n_eff_list <-
        parallel::parLapply(
          cl = cl,
          X = seq_len(dim(x)[3]),
          fun = function(i) posterior::ess_mean(x[, , i, drop = TRUE])
        )
    }
    n_eff_vec <- unlist(n_eff_list, use.names = FALSE)
  }

  return(n_eff_vec / S)
}

#' @export
#' @templateVar fn relative_eff
#' @template function
#' @param data,draws,... Same as for the [loo()] function method.
#'
relative_eff.function <-
  function(x,
           chain_id,
           ...,
           cores = getOption("mc.cores", 1),
           data = NULL,
           draws = NULL) {

    f_i <- validate_llfun(x) # not really an llfun, should return exp(ll) or exp(-ll)
    N <- dim(data)[1]

    if (cores == 1) {
      n_eff_list <-
        lapply(
          X = seq_len(N),
          FUN = function(i) {
            val_i <- f_i(data_i = data[i, , drop = FALSE], draws = draws, ...)
            relative_eff.default(as.vector(val_i), chain_id = chain_id, cores = 1)
          }
        )
    } else {
      if (!os_is_windows()) {
        n_eff_list <-
          parallel::mclapply(
            X = seq_len(N),
            FUN = function(i) {
              val_i <- f_i(data_i = data[i, , drop = FALSE], draws = draws, ...)
              relative_eff.default(as.vector(val_i), chain_id = chain_id, cores = 1)
            },
            mc.cores = cores
          )
      } else {
        cl <- parallel::makePSOCKcluster(cores)
        parallel::clusterExport(cl=cl, varlist=c("draws", "chain_id", "data"), envir=environment())
        on.exit(parallel::stopCluster(cl))
        n_eff_list <-
          parallel::parLapply(
            cl = cl,
            X = seq_len(N),
            fun = function(i) {
              val_i <- f_i(data_i = data[i, , drop = FALSE], draws = draws, ...)
              relative_eff.default(as.vector(val_i), chain_id = chain_id, cores = 1)
            }
          )
      }
    }

    n_eff_vec <- unlist(n_eff_list, use.names = FALSE)
    return(n_eff_vec)
  }

#' @export
#' @describeIn relative_eff
#'   If `x` is an object of class `"psis"`, `relative_eff()` simply returns
#'   the `r_eff` attribute of `x`.
relative_eff.importance_sampling <- function(x, ...) {
  attr(x, "r_eff")
}


# internal ----------------------------------------------------------------


#' Effective sample size for PSIS
#'
#' @noRd
#' @param w A vector or matrix (one column per observation) of normalized Pareto
#'   smoothed weights (not log weights).
#' @param r_eff Relative effective sample size of `exp(log_lik)` or
#'   `exp(-log_ratios)`. `r_eff` should be a scalar if `w` is a
#'   vector and a vector of length `ncol(w)` if `w` is a matrix.
#' @return A scalar if `w` is a vector. A vector of length `ncol(w)`
#'   if `w` is matrix.
#'
psis_n_eff <- function(w, ...) {
  UseMethod("psis_n_eff")
}

#' @export
psis_n_eff.default <- function(w, r_eff = NULL, ...) {
  ss <- sum(w^2)
  if (is.null(r_eff)) {
    return(1 / ss)
  }
  stopifnot(length(r_eff) == 1)
  1 / ss * r_eff
}

#' @export
psis_n_eff.matrix <- function(w, r_eff = NULL, ...) {
  ss <- colSums(w^2)
  if (is.null(r_eff)) {
    return(1 / ss)
  }
  if (length(r_eff) != length(ss) && length(r_eff) != 1) {
    stop("r_eff must have length 1 or ncol(w).", call. = FALSE)
  }
  1 / ss * r_eff
}

#' Compare `psis_loo_ss` objects
#' @noRd
#' @param x A list with `psis_loo` objects.
#' @param ... Currently ignored.
#' @return A `compare.loo_ss` object.
#' @author Mans Magnusson
#' @export
loo_compare.psis_loo_ss_list <- function(x, ...) {

  checkmate::assert_list(x, any.missing = FALSE, min.len = 1)
  for(i in seq_along(x)){
    if (!inherits(x[[i]], "psis_loo_ss")) x[[i]] <- as.psis_loo_ss.psis_loo(x[[i]])
  }

  loo_compare_checks.psis_loo_ss_list(x)

  comp <- loo_compare_matrix.psis_loo_ss_list(x)
  ord <- loo_compare_order(x)
  names(x) <- rownames(comp)[ord]

  rnms <- rownames(comp)
  elpd_diff_mat <- matrix(0, nrow = nrow(comp), ncol = 3,
                          dimnames = list(rnms, c("elpd_diff", "se_diff", "subsampling_se_diff")))
  for(i in 2:length(ord)){
    elpd_diff_mat[i,] <- loo_compare_ss(ref_loo = x[ord[1]], compare_loo = x[ord[i]])
  }
  comp <- cbind(elpd_diff_mat, comp)
  rownames(comp) <- rnms

  class(comp) <- c("compare.loo_ss", "compare.loo", class(comp))
  return(comp)
}

#' Compare a reference loo object with a comparison loo object
#' @noRd
#' @param ref_loo A named list with a `psis_loo_ss` object.
#' @param compare_loo A named list with a  `psis_loo_ss` object.
#' @return A 1 by 3 elpd_diff estimation.
loo_compare_ss <- function(ref_loo, compare_loo){
  checkmate::assert_list(ref_loo, names = "named")
  checkmate::assert_list(compare_loo, names = "named")
  checkmate::assert_class(ref_loo[[1]], "psis_loo_ss")
  checkmate::assert_class(compare_loo[[1]], "psis_loo_ss")


  ref_idx <- obs_idx(ref_loo[[1]])
  compare_idx <- obs_idx(compare_loo[[1]])
  intersect_idx <- base::intersect(ref_idx, compare_idx)
  ref_subset_of_compare <- base::setequal(intersect_idx, ref_idx)
  compare_subset_of_ref <- base::setequal(intersect_idx, compare_idx)

  # Using HH estimation
  if (ref_loo[[1]]$loo_subsampling$estimator == "hh_pps" | compare_loo[[1]]$loo_subsampling$estimator == "hh_pps"){
    warning("Hansen-Hurwitz estimator used. Naive diff SE is used.", call. = FALSE)
    return(loo_compare_ss_naive(ref_loo, compare_loo))
  }

  # Same observations in both
  if (compare_subset_of_ref & ref_subset_of_compare){
    return(loo_compare_ss_diff(ref_loo, compare_loo))
  }

  # Use subset
  if (compare_subset_of_ref | ref_subset_of_compare){
    if (compare_subset_of_ref) ref_loo[[1]] <- update(object = ref_loo[[1]], observations = compare_loo[[1]])
    if (ref_subset_of_compare) compare_loo[[1]] <- update(compare_loo[[1]], observations = ref_loo[[1]])
    message("Estimated elpd_diff using observations included in loo calculations for all models.")
    return(loo_compare_ss_diff(ref_loo, compare_loo))
  }

  # If different samples
  if (!compare_subset_of_ref & !ref_subset_of_compare){
    warning("Different subsamples in '", names(ref_loo), "' and '", names(compare_loo),
            "'. Naive diff SE is used.", call. = FALSE)
    return(loo_compare_ss_naive(ref_loo, compare_loo))
  }
}

#' Compute a naive diff SE
#' @noRd
#' @inheritParams loo_compare_ss
#' @return a 1 by 3 elpd_diff estimation
loo_compare_ss_naive <- function(ref_loo, compare_loo){
  checkmate::assert_list(ref_loo, names = "named")
  checkmate::assert_list(compare_loo, names = "named")
  checkmate::assert_class(ref_loo[[1]], "psis_loo_ss")
  checkmate::assert_class(compare_loo[[1]], "psis_loo_ss")

  elpd_loo_diff <- ref_loo[[1]]$estimates["elpd_loo","Estimate"] - compare_loo[[1]]$estimates["elpd_loo","Estimate"]
  elpd_loo_diff_se <- sqrt(
    (ref_loo[[1]]$estimates["elpd_loo","SE"])^2 +
    (compare_loo[[1]]$estimates["elpd_loo","SE"])^2)
  elpd_loo_diff_subsampling_se <- sqrt(
      (ref_loo[[1]]$estimates["elpd_loo","subsampling SE"])^2 +
      (compare_loo[[1]]$estimates["elpd_loo","subsampling SE"])^2)

  c(elpd_loo_diff, elpd_loo_diff_se, elpd_loo_diff_subsampling_se)
}

#' Compare a effective diff SE
#' @noRd
#' @inheritParams loo_compare_ss
#' @return a 1 by 3 elpd_diff estimation
loo_compare_ss_diff <- function(ref_loo, compare_loo){
  checkmate::assert_list(ref_loo, names = "named")
  checkmate::assert_list(compare_loo, names = "named")
  checkmate::assert_class(ref_loo[[1]], "psis_loo_ss")
  checkmate::assert_class(compare_loo[[1]], "psis_loo_ss")
  checkmate::assert_true(identical(obs_idx(ref_loo[[1]]), obs_idx(compare_loo[[1]])))

  # Assert not none as loo approximation
  checkmate::assert_true(ref_loo[[1]]$loo_subsampling$loo_approximation != "none")
  checkmate::assert_true(compare_loo[[1]]$loo_subsampling$loo_approximation != "none")

  diff_approx <- ref_loo[[1]]$loo_subsampling$elpd_loo_approx - compare_loo[[1]]$loo_subsampling$elpd_loo_approx
  diff_sample <- ref_loo[[1]]$pointwise[,"elpd_loo"] - compare_loo[[1]]$pointwise[,"elpd_loo"]
  est <- srs_diff_est(diff_approx, y = diff_sample, y_idx = ref_loo[[1]]$pointwise[,"idx"])

  elpd_loo_diff <- est$y_hat
  elpd_loo_diff_se <- sqrt(est$hat_v_y)
  elpd_loo_diff_subsampling_se <- sqrt(est$v_y_hat)

  c(elpd_loo_diff, elpd_loo_diff_se, elpd_loo_diff_subsampling_se)
}


#' Check list of `psis_loo` objects
#' @details Similar to `loo_compare_checks()` but checks dim size rather than
#' pointwise dim since different pointwise sizes of `psis_loo_ss` will work.
#' Can probably be removed by refactoring `loo_compare_checks()`.
#' @noRd
#' @inheritParams loo_compare_ss
#' @return A 1 by 3 elpd_diff estimation.
loo_compare_checks.psis_loo_ss_list <- function(loos) {
  ## errors
  if (length(loos) <= 1L) {
    stop("'loo_compare' requires at least two models.", call.=FALSE)
  }
  if (!all(sapply(loos, is.loo))) {
    stop("All inputs should have class 'loo'.", call.=FALSE)
  }

  Ns <- sapply(loos, function(x) x$loo_subsampling$data_dim[1])
  if (!all(Ns == Ns[1L])) {
    stop("Not all models have the same number of data points.", call.=FALSE)
  }

  ## warnings

  yhash <- lapply(loos, attr, which = "yhash")
  yhash_ok <- sapply(yhash, function(x) { # ok only if all yhash are same (all NULL is ok)
    isTRUE(all.equal(x, yhash[[1]]))
  })
  if (!all(yhash_ok)) {
    warning("Not all models have the same y variable. ('yhash' attributes do not match)",
            call. = FALSE)
  }

  if (all(sapply(loos, is.kfold))) {
    Ks <- unlist(lapply(loos, attr, which = "K"))
    if (!all(Ks == Ks[1])) {
      warning("Not all kfold objects have the same K value. ",
              "For a more accurate comparison use the same number of folds. ",
              call. = FALSE)
    }
  } else if (any(sapply(loos, is.kfold)) && any(sapply(loos, is.psis_loo))) {
    warning("Comparing LOO-CV to K-fold-CV. ",
            "For a more accurate comparison use the same number of folds ",
            "or loo for all models compared.",
            call. = FALSE)
  }
}

#' @rdname loo_compare
#' @export
print.compare.loo_ss <- function(x, ..., digits = 1) {
  xcopy <- x
  if (NCOL(xcopy) >= 2) {
    xcopy <- xcopy[, c("elpd_diff", "se_diff", "subsampling_se_diff")]
  }
  print(.fr(xcopy, digits), quote = FALSE)
  invisible(x)
}


#' Compute comparison matrix for `psis_loo_ss` objects
#' @noRd
#' @keywords internal
#' @param loos List of `psis_loo_ss` objects.
#' @return A `compare.loo_ss` matrix.
loo_compare_matrix.psis_loo_ss_list <- function(loos){
  tmp <- sapply(loos, function(x) {
    est <- x$estimates
    setNames(c(est), nm = c(rownames(est),
                            paste0("se_", rownames(est)),
                            paste0("subsampling_se_", rownames(est))))
  })
  colnames(tmp) <- find_model_names(loos)
  rnms <- rownames(tmp)
  comp <- tmp
  ord <- loo_compare_order(loos)
  comp <- t(comp)[ord, ]
  patts <- c("elpd", "p_", "^waic$|^looic$", "se_waic$|se_looic$")
  col_ord <- unlist(sapply(patts, function(p) grep(p, colnames(comp))),
                    use.names = FALSE)
  comp <- comp[, col_ord]
  comp
}

#' Model averaging/weighting via stacking or pseudo-BMA weighting
#'
#' Model averaging via stacking of predictive distributions, pseudo-BMA
#' weighting or pseudo-BMA+ weighting with the Bayesian bootstrap. See Yao et
#' al. (2018), Vehtari, Gelman, and Gabry (2017), and Vehtari, Simpson,
#' Gelman, Yao, and Gabry (2024) for background.
#'
#' @export
#' @param x A list of `"psis_loo"` objects (objects returned by [loo()]) or
#'   pointwise log-likelihood matrices, one for each model. If the list
#'   elements are named the names will be used to label the models in the
#'   results. Each matrix/object should have dimensions \eqn{S} by \eqn{N},
#'   where \eqn{S} is the size of the posterior sample (with all chains merged)
#'   and \eqn{N} is the number of data points. If `x` is a list of
#'   log-likelihood matrices then [loo()] is called internally on each matrix.
#'   Currently the `loo_model_weights()` function is not implemented to be used
#'   with results from K-fold CV, but you can still obtain weights using K-fold
#'   CV results by calling the `stacking_weights()` or `pseudobma_weights()`
#'   function directly.
#' @param method Either `"stacking"` (the default) or `"pseudobma"`, indicating which method
#'   to use for obtaining the weights. `"stacking"` refers to stacking of
#'   predictive distributions and  `"pseudobma"` refers to pseudo-BMA+ weighting
#'   (or plain pseudo-BMA weighting if argument `BB` is `FALSE`).
#' @param BB Logical used when `"method"`=`"pseudobma"`. If
#'   `TRUE` (the default), the Bayesian bootstrap will be used to adjust
#'   the pseudo-BMA weighting, which is called pseudo-BMA+ weighting. It helps
#'   regularize the weight away from 0 and 1, so as to reduce the variance.
#' @param BB_n For pseudo-BMA+ weighting only, the number of samples to use for
#'   the Bayesian bootstrap. The default is `BB_n=1000`.
#' @param alpha Positive scalar shape parameter in the Dirichlet distribution
#'   used for the Bayesian bootstrap. The default is `alpha=1`, which
#'   corresponds to a uniform distribution on the simplex space.
#' @param optim_method If `method="stacking"`, a string passed to the `method`
#'   argument of [stats::constrOptim()] to specify the optimization algorithm.
#'   The default is `optim_method="BFGS"`, but other options are available (see
#'   [stats::optim()]).
#' @param optim_control If `method="stacking"`, a list of control parameters for
#'   optimization passed to the `control` argument of [stats::constrOptim()].
#' @param r_eff_list Optionally, a list of relative effective sample size
#'   estimates for the likelihood `(exp(log_lik))` of each observation in
#'   each model. See [psis()] and  [relative_eff()] helper
#'   function for computing `r_eff`. If `x` is a list of `"psis_loo"`
#'   objects then `r_eff_list` is ignored.
#' @template cores
#' @param ... Unused, except for the generic to pass arguments to individual
#'   methods.
#'
#' @return A numeric vector containing one weight for each model.
#'
#' @details
#' `loo_model_weights()` is a wrapper around the `stacking_weights()` and
#' `pseudobma_weights()` functions that implements stacking, pseudo-BMA, and
#' pseudo-BMA+ weighting for combining multiple predictive distributions. We can
#' use approximate or exact leave-one-out cross-validation (LOO-CV) or K-fold CV
#' to estimate the expected log predictive density (ELPD).
#'
#' The stacking method (`method="stacking"`), which is the default for
#' `loo_model_weights()`, combines all models by maximizing the leave-one-out
#' predictive density of the combination distribution. That is, it finds the
#' optimal linear combining weights for maximizing the leave-one-out log score.
#'
#' The pseudo-BMA method (`method="pseudobma"`) finds the relative weights
#' proportional to the ELPD of each model. However, when
#' `method="pseudobma"`, the default is to also use the Bayesian bootstrap
#' (`BB=TRUE`), which corresponds to the pseudo-BMA+ method. The Bayesian
#' bootstrap  takes into account the uncertainty of finite data points and
#' regularizes the weights away from the extremes of 0 and 1.
#'
#' In general, we recommend stacking for averaging predictive distributions,
#' while pseudo-BMA+ can serve as a computationally easier alternative.
#'
#' @seealso
#' * The __loo__ package [vignettes](https://mc-stan.org/loo/articles/), particularly
#'   [Bayesian Stacking and Pseudo-BMA weights using the __loo__ package](https://mc-stan.org/loo/articles/loo2-weights.html).
#' * [loo()] for details on leave-one-out ELPD estimation.
#' * [constrOptim()] for the choice of optimization methods and control-parameters.
#' * [relative_eff()] for computing `r_eff`.
#'
#' @template loo-and-psis-references
#' @template stacking-references
#'
#' @examples
#' \dontrun{
#' ### Demonstrating usage after fitting models with RStan
#' library(rstan)
#'
#' # generate fake data from N(0,1).
#' N <- 100
#' y <- rnorm(N, 0, 1)
#'
#' # Suppose we have three models: N(-1, sigma), N(0.5, sigma) and N(0.6,sigma).
#' stan_code <- "
#'   data {
#'     int N;
#'     vector[N] y;
#'     real mu_fixed;
#'   }
#'   parameters {
#'     real<lower=0> sigma;
#'   }
#'   model {
#'     sigma ~ exponential(1);
#'     y ~ normal(mu_fixed, sigma);
#'   }
#'   generated quantities {
#'     vector[N] log_lik;
#'     for (n in 1:N) log_lik[n] = normal_lpdf(y[n]| mu_fixed, sigma);
#'   }"
#'
#' mod <- stan_model(model_code = stan_code)
#' fit1 <- sampling(mod, data=list(N=N, y=y, mu_fixed=-1))
#' fit2 <- sampling(mod, data=list(N=N, y=y, mu_fixed=0.5))
#' fit3 <- sampling(mod, data=list(N=N, y=y, mu_fixed=0.6))
#' model_list <- list(fit1, fit2, fit3)
#' log_lik_list <- lapply(model_list, extract_log_lik)
#'
#' # optional but recommended
#' r_eff_list <- lapply(model_list, function(x) {
#'   ll_array <- extract_log_lik(x, merge_chains = FALSE)
#'   relative_eff(exp(ll_array))
#' })
#'
#' # stacking method:
#' wts1 <- loo_model_weights(
#'   log_lik_list,
#'   method = "stacking",
#'   r_eff_list = r_eff_list,
#'   optim_control = list(reltol=1e-10)
#' )
#' print(wts1)
#'
#' # can also pass a list of psis_loo objects to avoid recomputing loo
#' loo_list <- lapply(1:length(log_lik_list), function(j) {
#'   loo(log_lik_list[[j]], r_eff = r_eff_list[[j]])
#' })
#'
#' wts2 <- loo_model_weights(
#'   loo_list,
#'   method = "stacking",
#'   optim_control = list(reltol=1e-10)
#' )
#' all.equal(wts1, wts2)
#'
#' # can provide names to be used in the results
#' loo_model_weights(setNames(loo_list, c("A", "B", "C")))
#'
#'
#' # pseudo-BMA+ method:
#' set.seed(1414)
#' loo_model_weights(loo_list, method = "pseudobma")
#'
#' # pseudo-BMA method (set BB = FALSE):
#' loo_model_weights(loo_list, method = "pseudobma", BB = FALSE)
#'
#' # calling stacking_weights or pseudobma_weights directly
#' lpd1 <- loo(log_lik_list[[1]], r_eff = r_eff_list[[1]])$pointwise[,1]
#' lpd2 <- loo(log_lik_list[[2]], r_eff = r_eff_list[[2]])$pointwise[,1]
#' lpd3 <- loo(log_lik_list[[3]], r_eff = r_eff_list[[3]])$pointwise[,1]
#' stacking_weights(cbind(lpd1, lpd2, lpd3))
#' pseudobma_weights(cbind(lpd1, lpd2, lpd3))
#' pseudobma_weights(cbind(lpd1, lpd2, lpd3), BB = FALSE)
#' }
#'
loo_model_weights <- function(x, ...) {
  UseMethod("loo_model_weights")
}

#' @rdname loo_model_weights
#' @export
#' @export loo_model_weights.default
loo_model_weights.default <-
  function(x,
           ...,
           method = c("stacking", "pseudobma"),
           optim_method = "BFGS",
           optim_control = list(),
           BB = TRUE,
           BB_n = 1000,
           alpha = 1,
           r_eff_list = NULL,
           cores = getOption("mc.cores", 1)) {

    cores <- loo_cores(cores)
    method <- match.arg(method)
    K <- length(x) # number of models

    if (is.matrix(x[[1]])) {
      N <- ncol(x[[1]]) # number of data points
      validate_log_lik_list(x)
      validate_r_eff_list(r_eff_list, K, N)
      lpd_point <- matrix(NA, N, K)
      elpd_loo <- rep(NA, K)
      for (k in 1:K) {
        r_eff_k <- r_eff_list[[k]] # possibly NULL
        log_likelihood <- x[[k]]
        loo_object <- loo(log_likelihood, r_eff = r_eff_k, cores = cores)
        lpd_point[, k] <- loo_object$pointwise[, "elpd_loo"]    #calculate log(p_k (y_i | y_-i))
        elpd_loo[k] <- loo_object$estimates["elpd_loo", "Estimate"]
      }
    } else if (is.psis_loo(x[[1]])) {
      validate_psis_loo_list(x)
      lpd_point <- do.call(cbind, lapply(x, function(obj) obj$pointwise[, "elpd_loo"]))
      elpd_loo <- sapply(x, function(obj) obj$estimates["elpd_loo", "Estimate"])
    } else {
      stop("'x' must be a list of matrices or a list of 'psis_loo' objects.")
    }

    ## 1) stacking on log score
    if (method =="stacking") {
      wts <- stacking_weights(
        lpd_point = lpd_point,
        optim_method = optim_method,
        optim_control = optim_control
      )

    } else {
      # method =="pseudobma"
      wts <- pseudobma_weights(
        lpd_point = lpd_point,
        BB = BB,
        BB_n = BB_n,
        alpha = alpha
      )
    }

    if (is.matrix(x[[1]])) {
      if (!is.null(names(x)) && all(nzchar(names(x)))) {
        wts <- setNames(wts, names(x))
      }
    } else { # list of loo objects
      wts <- setNames(wts, find_model_names(x))
    }
    wts
  }


#' @rdname loo_model_weights
#' @export
#' @param lpd_point If calling `stacking_weights()` or `pseudobma_weights()`
#'   directly, a matrix of pointwise leave-one-out (or K-fold) log likelihoods
#'   evaluated for different models. It should be a \eqn{N} by \eqn{K}  matrix
#'   where \eqn{N} is sample size and \eqn{K} is the number of models. Each
#'   column corresponds to one model. These values can be calculated
#'   approximately using [loo()] or by running exact leave-one-out or K-fold
#'   cross-validation.
#'
#' @importFrom stats constrOptim
#'
stacking_weights <-
  function(lpd_point,
           optim_method = "BFGS",
           optim_control = list()) {

    stopifnot(is.matrix(lpd_point))
    N <- nrow(lpd_point)
    K <- ncol(lpd_point)
    if (K < 2) {
      stop("At least two models are required for stacking weights.")
    }

    negative_log_score_loo <- function(w) {
      # objective function: log score
      stopifnot(length(w) == K - 1)
      w_full <- c(w, 1 - sum(w))
      # avoid over- and underflows using log weights and rowLogSumExps
      sum <- sum(matrixStats::rowLogSumExps(sweep(lpd_point[1:N,], 2, log(w_full), '+')))
      return(-as.numeric(sum))
    }

    gradient <- function(w) {
      # gradient of the objective function
      stopifnot(length(w) == K - 1)
      w_full <- c(w, 1 - sum(w))
      grad <- rep(0, K - 1)
      # avoid over- and underflows using log weights, rowLogSumExps,
      # and by subtracting the row maximum of lpd_point
      mlpd <- matrixStats::rowMaxs(lpd_point)
      for (k in 1:(K - 1)) {
        grad[k] <- sum((exp(lpd_point[, k] - mlpd) - exp(lpd_point[, K] - mlpd)) / exp(matrixStats::rowLogSumExps(sweep(lpd_point, 2, log(w_full), '+')) - mlpd))
      }
      return(-grad)
    }

    ui <- rbind(rep(-1, K - 1), diag(K - 1))  # K-1 simplex constraint matrix
    ci <- c(-1, rep(0, K - 1))
    w <- constrOptim(
      theta = rep(1 / K, K - 1),
      f = negative_log_score_loo,
      grad = gradient,
      ui = ui,
      ci = ci,
      method = optim_method,
      control = optim_control
    )$par

    wts <- structure(
      c(w, 1 - sum(w)),
      names = paste0("model", 1:K),
      class = c("stacking_weights")
    )

    return(wts)
  }


#' @rdname loo_model_weights
#' @export
#'
pseudobma_weights <-
  function(lpd_point,
           BB = TRUE,
           BB_n = 1000,
           alpha = 1) {
    stopifnot(is.matrix(lpd_point))
    N <- nrow(lpd_point)
    K <- ncol(lpd_point)
    if (K < 2) {
      stop("At least two models are required for pseudo-BMA weights.")
    }

    if (!BB) {
      elpd <- colSums2(lpd_point)
      uwts <- exp(elpd - max(elpd))
      wts <- structure(
        uwts / sum(uwts),
        names = paste0("model", 1:K),
        class = "pseudobma_weights"
      )
      return(wts)
    }

    temp <- matrix(NA, BB_n, K)
    BB_weighting <- dirichlet_rng(BB_n, rep(alpha, N))
    for (bb in 1:BB_n) {
      z_bb <- BB_weighting[bb, ] %*% lpd_point * N
      uwts <- exp(z_bb - max(z_bb))
      temp[bb, ] <- uwts / sum(uwts)
    }
    wts <- structure(
      colMeans(temp),
      names = paste0("model", 1:K),
      class = "pseudobma_bb_weights"
    )
    return(wts)
  }


#' Generate dirichlet simulations, rewritten version
#' @importFrom stats rgamma
#' @noRd
dirichlet_rng <- function(n, alpha) {
  K <- length(alpha)
  gamma_sim <- matrix(rgamma(K * n, alpha), ncol = K, byrow = TRUE)
  gamma_sim / rowSums(gamma_sim)
}

#' @export
print.stacking_weights <- function(x, digits = 3, ...) {
  cat("Method: stacking\n------\n")
  print_weight_vector(x, digits = digits)
}

#' @export
print.pseudobma_weights <- function(x, digits = 3, ...) {
  cat("Method: pseudo-BMA\n------\n")
  print_weight_vector(x, digits = digits)
}

#' @export
print.pseudobma_bb_weights <- function(x, digits = 3, ...) {
  cat("Method: pseudo-BMA+ with Bayesian bootstrap\n------\n")
  print_weight_vector(x, digits = digits)
}

print_weight_vector <- function(x, digits) {
  z <- cbind(x)
  colnames(z) <- "weight"
  print(.fr(z, digits = digits), quote = FALSE)
  invisible(x)
}

#' Validate r_eff_list argument if provided
#'
#' @noRd
#' @param r_eff_list User's `r_eff_list` argument
#' @param K Required length of `r_eff_list` (number of models).
#' @param N Required length of each element of `r_eff_list` (number of data points).
#' @return Either throws an error or returns `TRUE` invisibly.
#'
validate_r_eff_list <- function(r_eff_list, K, N) {
  if (is.null(r_eff_list)) return(invisible(TRUE))

  if (length(r_eff_list) != K) {
    stop("If r_eff_list is specified then it must contain ",
         "one component for each model being compared.",
         call. = FALSE)
  }
  if (any(sapply(r_eff_list, length) != N)) {
    stop("Each component of r_eff list must have the same length ",
         "as the number of columns in the log-likelihood matrix.",
         call. = FALSE)
  }
  invisible(TRUE)
}


#' Validate log-likelihood list argument
#'
#' Checks that log-likelihood list has at least 2 elements and that each element
#' has the same dimensions.
#'
#' @noRd
#' @param log_lik_list User's list of log-likelihood matrices (the `x` argument
#'   to loo_model_weights).
#' @return Either throws an error or returns `TRUE` invisibly.
#'
validate_log_lik_list <- function(log_lik_list) {
  stopifnot(is.list(log_lik_list))
  if (length(log_lik_list) < 2) {
    stop("At least two models are required.", call. = FALSE)
  }
  if (length(unique(sapply(log_lik_list, ncol))) != 1 |
     length(unique(sapply(log_lik_list, nrow))) != 1) {
    stop("Each log-likelihood matrix must have the same dimensions.", call. = FALSE)
  }
  invisible(TRUE)
}

validate_psis_loo_list <- function(psis_loo_list) {
  stopifnot(is.list(psis_loo_list))
  if (length(psis_loo_list) < 2) {
    stop("At least two models are required.", call. = FALSE)
  }
  if (!all(sapply(psis_loo_list, is.psis_loo))) {
    stop("List elements must all be 'psis_loo' objects or log-likelihood matrices.")
  }

  dims <- sapply(psis_loo_list, dim)
  if (length(unique(dims[1, ])) != 1 |
      length(unique(dims[2, ])) != 1) {
    stop("Each object in the list must have the same dimensions.", call. = FALSE)
  }
  invisible(TRUE)
}

# generate fake data
set.seed(123)
y <- rnorm(50, 0, 1)
sd_sim1 <- abs(rnorm(500, 1.5, 0.1))
sd_sim2 <- abs(rnorm(500, 1.2, 0.1))
sd_sim3 <- abs(rnorm(500, 1, 0.05))
log_lik1 <- log_lik2 <- log_lik3 <- matrix(NA, 500, 50)
for (s in 1:500) {
  log_lik1[s, ] <- dnorm(y, -1, sd_sim1[s], log = T)
  log_lik2[s, ] <- dnorm(y, 0.7, sd_sim2[s], log = T)
  log_lik3[s, ] <- dnorm(y, 1, sd_sim3[s], log = T)
}

ll_list <- list(log_lik1, log_lik2, log_lik3)
r_eff_list <- list(rep(0.9, 50), rep(0.9, 50), rep(0.9, 50))

loo_list <- lapply(1:length(ll_list), function(j) {
  loo(ll_list[[j]], r_eff = r_eff_list[[j]])
})

tol <- 0.01 # absolute tolerance of weights

test_that("loo_model_weights throws correct errors and warnings", {
  expect_error(
    loo_model_weights(log_lik1),
    "list of matrices or a list of 'psis_loo' objects"
  )
  expect_error(loo_model_weights(list(log_lik1)), "At least two models")
  expect_error(loo_model_weights(list(loo_list[[1]])), "At least two models")
  expect_error(
    loo_model_weights(list(log_lik1), method = "pseudobma"),
    "At least two models"
  )

  expect_error(
    loo_model_weights(list(log_lik1, log_lik2[-1, ])),
    "same dimensions"
  )
  expect_error(
    loo_model_weights(list(log_lik1, log_lik2, log_lik3[, -1])),
    "same dimensions"
  )

  loo_list2 <- loo_list
  attr(loo_list2[[3]], "dims") <- c(10, 10)
  expect_error(loo_model_weights(loo_list2), "same dimensions")

  expect_error(
    loo_model_weights(ll_list, r_eff_list = r_eff_list[-1]),
    "one component for each model"
  )

  r_eff_list[[3]] <- rep(0.9, 51)
  expect_error(
    loo_model_weights(ll_list, r_eff_list = r_eff_list),
    "same length as the number of columns"
  )

  expect_error(
    loo_model_weights(list(loo_list[[1]], 2)),
    "List elements must all be 'psis_loo' objects or log-likelihood matrices",
    fixed = TRUE
  )

  expect_no_warning(loo_model_weights(ll_list))
})


test_that("loo_model_weights (stacking and pseudo-BMA) gives expected result", {
  w1 <- loo_model_weights(ll_list, method = "stacking", r_eff_list = r_eff_list)
  expect_type(w1, "double")
  expect_s3_class(w1, "stacking_weights")
  expect_length(w1, 3)
  expect_named(w1, paste0("model", c(1:3)))
  expect_snapshot_value(as.numeric(w1), style = "serialize")
  expect_snapshot(print(w1))

  w1_b <- loo_model_weights(loo_list)
  expect_identical(w1, w1_b)

  w2 <- loo_model_weights(
    ll_list,
    r_eff_list = r_eff_list,
    method = "pseudobma",
    BB = TRUE
  )
  expect_type(w2, "double")
  expect_s3_class(w2, "pseudobma_bb_weights")
  expect_length(w2, 3)
  expect_named(w2, paste0("model", c(1:3)))
  expect_snapshot_value(as.numeric(w2), style = "serialize")
  expect_snapshot(print(w2))

  w3 <- loo_model_weights(
    ll_list,
    r_eff_list = r_eff_list,
    method = "pseudobma",
    BB = FALSE
  )
  expect_type(w3, "double")
  expect_length(w3, 3)
  expect_named(w3, paste0("model", c(1:3)))
  expect_equal(
    as.numeric(w3),
    c(5.365279e-05, 9.999436e-01, 2.707028e-06),
    tolerance = tol
  )
  expect_snapshot(print(w3))

  w3_b <- loo_model_weights(loo_list, method = "pseudobma", BB = FALSE)
  expect_identical(w3, w3_b)
})

test_that("stacking_weights and pseudobma_weights throw correct errors", {
  xx <- cbind(rnorm(10))
  expect_error(stacking_weights(xx), "two models are required")
  expect_error(pseudobma_weights(xx), "two models are required")
})

test_that("loo_model_weights uses correct names for list of loo objects", {
  loo1 <- loo_list[[1]]
  loo2 <- loo_list[[2]]
  loo3 <- loo_list[[3]]

  expect_named(
    loo_model_weights(list(loo1, loo2, loo3)),
    c("model1", "model2", "model3")
  )
  expect_named(
    loo_model_weights(list("a" = loo1, loo2, "c" = loo3)),
    c("a", "model2", "c")
  )
  expect_named(
    loo_model_weights(list(`a` = loo1, `b` = loo2, `c` = loo3)),
    c("a", "b", "c")
  )
})

library(loo)

context("loo_model_weights")

# generate fake data
set.seed(123)
y<-rnorm(50,0,1)
sd_sim1<- abs(rnorm(500,1.5, 0.1))
sd_sim2<- abs(rnorm(500,1.2, 0.1))
sd_sim3<- abs(rnorm(500,1, 0.05))
log_lik1 <- log_lik2 <- log_lik3 <- matrix(NA, 500, 50)
for(s in 1:500) {
  log_lik1[s,] <- dnorm(y,-1,sd_sim1[s], log=T)
  log_lik2[s,] <-  dnorm(y,0.7,sd_sim2[s], log=T)
  log_lik3[s,] <-  dnorm(y,1,sd_sim3[s], log=T)
}

ll_list <- list(log_lik1, log_lik2,log_lik3)
r_eff_list <- list(rep(0.9,50), rep(0.9,50), rep(0.9,50))

loo_list <- lapply(1:length(ll_list), function(j) {
  loo(ll_list[[j]], r_eff = r_eff_list[[j]])
})

tol <- 0.01 # absoulte tolerance of weights

test_that("loo_model_weights throws correct errors and warnings", {
  expect_error(loo_model_weights(log_lik1), "list of matrices or a list of 'psis_loo' objects")
  expect_error(loo_model_weights(list(log_lik1)), "At least two models")
  expect_error(loo_model_weights(list(loo_list[[1]])), "At least two models")
  expect_error(loo_model_weights(list(log_lik1), method = "pseudobma"), "At least two models")

  expect_error(loo_model_weights(list(log_lik1, log_lik2[-1, ])), "same dimensions")
  expect_error(loo_model_weights(list(log_lik1, log_lik2, log_lik3[, -1])), "same dimensions")

  loo_list2 <- loo_list
  attr(loo_list2[[3]], "dims") <- c(10, 10)
  expect_error(loo_model_weights(loo_list2), "same dimensions")

  expect_error(loo_model_weights(ll_list, r_eff_list = r_eff_list[-1]),
               "one component for each model")

  r_eff_list[[3]] <- rep(0.9, 51)
  expect_error(loo_model_weights(ll_list, r_eff_list = r_eff_list),
               "same length as the number of columns")

  expect_error(loo_model_weights(list(loo_list[[1]], 2)),
               "List elements must all be 'psis_loo' objects or log-likelihood matrices",
               fixed = TRUE)

  expect_no_warning(loo_model_weights(ll_list))
})


test_that("loo_model_weights (stacking and pseudo-BMA) gives expected result", {
  w1 <- loo_model_weights(ll_list, method = "stacking", r_eff_list = r_eff_list)
  expect_type(w1,"double")
  expect_s3_class(w1, "stacking_weights")
  expect_length(w1, 3)
  expect_named(w1, paste0("model"  ,c(1:3)))
  expect_equal_to_reference(as.numeric(w1), "reference-results/model_weights_stacking.rds",
                            tolerance  = tol, scale=1)
  expect_output(print(w1), "Method: stacking")

  w1_b <- loo_model_weights(loo_list)
  expect_identical(w1, w1_b)

  w2 <- loo_model_weights(ll_list, r_eff_list=r_eff_list,
                      method = "pseudobma", BB = TRUE)
  expect_type(w2, "double")
  expect_s3_class(w2, "pseudobma_bb_weights")
  expect_length(w2, 3)
  expect_named(w2, paste0("model", c(1:3)))
  expect_equal_to_reference(as.numeric(w2), "reference-results/model_weights_pseudobma.rds",
                            tolerance  = tol, scale=1)
  expect_output(print(w2), "Method: pseudo-BMA+")

  w3 <- loo_model_weights(ll_list, r_eff_list=r_eff_list,
                      method = "pseudobma", BB = FALSE)
  expect_type(w3,"double")
  expect_length(w3, 3)
  expect_named(w3, paste0("model"  ,c(1:3)))
  expect_equal(as.numeric(w3), c(5.365279e-05, 9.999436e-01, 2.707028e-06),
               tolerance  = tol, scale = 1)
  expect_output(print(w3), "Method: pseudo-BMA")

  w3_b <- loo_model_weights(loo_list, method = "pseudobma", BB = FALSE)
  expect_identical(w3, w3_b)
})

test_that("stacking_weights and pseudobma_weights throw correct errors", {
  xx <- cbind(rnorm(10))
  expect_error(stacking_weights(xx), "two models are required")
  expect_error(pseudobma_weights(xx), "two models are required")
})

test_that("loo_model_weights uses correct names for list of loo objects", {
  loo1 <- loo_list[[1]]
  loo2 <- loo_list[[2]]
  loo3 <- loo_list[[3]]

  expect_named(
    loo_model_weights(list(loo1, loo2, loo3)),
    c("model1", "model2", "model3")
  )
  expect_named(
    loo_model_weights(list("a" = loo1, loo2, "c" = loo3)),
    c("a", "model2", "c")
  )
  expect_named(
    loo_model_weights(list(`a` = loo1, `b` = loo2, `c` = loo3)),
    c("a", "b", "c")
  )
})

#' Standard importance sampling (SIS)
#'
#' Implementation of standard importance sampling (SIS).
#'
#' @param log_ratios An array, matrix, or vector of importance ratios on the log
#'   scale (for Importance sampling LOO, these are *negative* log-likelihood
#'   values). See the **Methods (by class)** section below for a detailed
#'   description of how to specify the inputs for each method.
#' @template cores
#' @param ... Arguments passed on to the various methods.
#' @param r_eff Vector of relative effective sample size estimates containing
#'   one element per observation. The values provided should be the relative
#'   effective sample sizes of `1/exp(log_ratios)` (i.e., `1/ratios`).
#'   This is related to the relative efficiency of estimating the normalizing
#'   term in self-normalizing importance sampling. See the [relative_eff()]
#'   helper function for computing `r_eff`. If using `psis` with
#'   draws of the `log_ratios` not obtained from MCMC then the warning
#'   message thrown when not specifying `r_eff` can be disabled by
#'   setting `r_eff` to `NA`.
#'
#' @return The `sis()` methods return an object of class `"sis"`,
#'   which is a named list with the following components:
#'
#' \describe{
#'   \item{`log_weights`}{
#'     Vector or matrix of smoothed but *unnormalized* log
#'     weights. To get normalized weights use the
#'     [`weights()`][weights.importance_sampling] method provided for objects of
#'     class `sis`.
#'   }
#'  \item{`diagnostics`}{
#'    A named list containing one vector:
#'    * `pareto_k`: Not used in `sis`, all set to 0.
#'    * `n_eff`: effective sample size estimates.
#'  }
#' }
#'
#' Objects of class `"sis"` also have the following [attributes][attributes()]:
#' \describe{
#'   \item{`norm_const_log`}{
#'     Vector of precomputed values of `colLogSumExps(log_weights)` that are
#'     used internally by the `weights` method to normalize the log weights.
#'   }
#'   \item{`r_eff`}{
#'     If specified, the user's `r_eff` argument.
#'   }
#'   \item{`tail_len`}{
#'     Not used for `sis`.
#'   }
#'   \item{`dims`}{
#'     Integer vector of length 2 containing `S` (posterior sample size)
#'     and `N` (number of observations).
#'   }
#'   \item{`method`}{
#'     Method used for importance sampling, here `sis`.
#'   }
#' }
#'
#' @seealso
#' * [psis()] for approximate LOO-CV using PSIS.
#' * [loo()] for approximate LOO-CV.
#' * [pareto-k-diagnostic] for PSIS diagnostics.
#'
#' @template loo-and-psis-references
#'
#' @examples
#' log_ratios <- -1 * example_loglik_array()
#' r_eff <- relative_eff(exp(-log_ratios))
#' sis_result <- sis(log_ratios, r_eff = r_eff)
#' str(sis_result)
#'
#' # extract smoothed weights
#' lw <- weights(sis_result) # default args are log=TRUE, normalize=TRUE
#' ulw <- weights(sis_result, normalize=FALSE) # unnormalized log-weights
#'
#' w <- weights(sis_result, log=FALSE) # normalized weights (not log-weights)
#' uw <- weights(sis_result, log=FALSE, normalize = FALSE) # unnormalized weights
#'
#' @export
sis <- function(log_ratios, ...) UseMethod("sis")

#' @export
#' @templateVar fn sis
#' @template array
#'
sis.array <-
  function(log_ratios, ...,
           r_eff = NULL,
           cores = getOption("mc.cores", 1)) {
  importance_sampling.array(log_ratios = log_ratios, ...,
                            r_eff = r_eff,
                            cores = cores,
                            method = "sis")
  }

#' @export
#' @templateVar fn sis
#' @template matrix
#'
sis.matrix <-
  function(log_ratios,
           ...,
           r_eff = NULL,
           cores = getOption("mc.cores", 1)) {
    importance_sampling.matrix(log_ratios,
                               ...,
                               r_eff = r_eff,
                               cores = cores,
                               method = "sis")
  }

#' @export
#' @templateVar fn sis
#' @template vector
#'
sis.default <-
  function(log_ratios, ..., r_eff = NULL) {
    importance_sampling.default(log_ratios = log_ratios, ...,
                                r_eff = r_eff,
                                method = "sis")
  }

#' @rdname psis
#' @export
is.sis <- function(x) {
  inherits(x, "sis") && is.list(x)
}


# internal ----------------------------------------------------------------


#' Standard IS on a single vector
#'
#' @noRd
#' @param log_ratios_i A vector of log importance ratios (for `loo()`, negative
#'   log likelihoods).
#' @param ... Not used. Included to conform to PSIS API.
#'
#' @details Implementation standard importance sampling.
#' @return A named list containing:
#' * `lw`: vector of unnormalized log weights
#' * `pareto_k`: scalar Pareto k estimate. For IS, this defaults to 0.
do_sis_i <- function(log_ratios_i, ...) {
  S <- length(log_ratios_i)
  list(log_weights = log_ratios_i, pareto_k = 0)
}

#' Widely applicable information criterion (WAIC)
#'
#' The `waic()` methods can be used to compute WAIC from the pointwise
#' log-likelihood. However, we recommend LOO-CV using PSIS (as implemented by
#' the [loo()] function) because PSIS provides useful diagnostics as well as
#' effective sample size and Monte Carlo estimates.
#'
#' @export waic waic.array waic.matrix waic.function
#' @inheritParams loo
#'
#' @return A named list (of class `c("waic", "loo")`) with components:
#'
#' \describe{
#'  \item{`estimates`}{
#'  A matrix with two columns (`"Estimate"`, `"SE"`) and three
#'  rows (`"elpd_waic"`, `"p_waic"`, `"waic"`). This contains
#'  point estimates and standard errors of the expected log pointwise predictive
#'  density (`elpd_waic`), the effective number of parameters
#'  (`p_waic`) and the information criterion `waic` (which is just
#'  `-2 * elpd_waic`, i.e., converted to deviance scale).
#'  }
#'  \item{`pointwise`}{
#'  A matrix with three columns (and number of rows equal to the number of
#'  observations) containing the pointwise contributions of each of the above
#'  measures (`elpd_waic`, `p_waic`, `waic`).
#'  }
#' }
#'
#' @seealso
#' * The __loo__ package [vignettes](https://mc-stan.org/loo/articles/) and
#'   Vehtari, Gelman, and Gabry (2017) and Vehtari, Simpson, Gelman, Yao,
#'   and Gabry (2024) for more details on why we prefer `loo()` to `waic()`.
#' * [loo_compare()] for comparing models on approximate LOO-CV or WAIC.
#'
#' @references
#' Watanabe, S. (2010). Asymptotic equivalence of Bayes cross validation and
#' widely applicable information criterion in singular learning theory.
#' *Journal of Machine Learning Research* **11**, 3571-3594.
#'
#' @template loo-and-psis-references
#'
#' @examples
#' ### Array and matrix methods
#' LLarr <- example_loglik_array()
#' dim(LLarr)
#'
#' LLmat <- example_loglik_matrix()
#' dim(LLmat)
#'
#' waic_arr <- waic(LLarr)
#' waic_mat <- waic(LLmat)
#' identical(waic_arr, waic_mat)
#'
#'
#' \dontrun{
#' log_lik1 <- extract_log_lik(stanfit1)
#' log_lik2 <- extract_log_lik(stanfit2)
#' (waic1 <- waic(log_lik1))
#' (waic2 <- waic(log_lik2))
#' print(compare(waic1, waic2), digits = 2)
#' }
#'
waic <- function(x, ...) {
  UseMethod("waic")
}

#' @export
#' @templateVar fn waic
#' @template array
#'
waic.array <- function(x, ...) {
  waic.matrix(llarray_to_matrix(x), ...)
}

#' @export
#' @templateVar fn waic
#' @template matrix
#'
waic.matrix <- function(x, ...) {
  ll <- validate_ll(x)
  lldim <- dim(ll)
  lpd <- matrixStats::colLogSumExps(ll) - log(nrow(ll)) # colLogMeanExps
  p_waic <- matrixStats::colVars(ll)
  elpd_waic <- lpd - p_waic
  waic <- -2 * elpd_waic
  pointwise <- cbind(elpd_waic, p_waic, waic)

  throw_pwaic_warnings(pointwise[, "p_waic"], digits = 1)
  return(waic_object(pointwise, dims = lldim))
}


#' @export
#' @templateVar fn waic
#' @template function
#' @param draws,data,... For the function method only. See the
#' **Methods (by class)** section below for details on these arguments.
#'
waic.function <-
  function(x,
           ...,
           data = NULL,
           draws = NULL) {
    stopifnot(is.data.frame(data) || is.matrix(data), !is.null(draws))

    .llfun <- validate_llfun(x)
    N <- dim(data)[1]
    S <- length(as.vector(.llfun(data_i = data[1,, drop=FALSE], draws = draws, ...)))
    waic_list <- lapply(seq_len(N), FUN = function(i) {
      ll_i <- .llfun(data_i = data[i,, drop=FALSE], draws = draws, ...)
      ll_i <- as.vector(ll_i)
      lpd_i <- logMeanExp(ll_i)
      p_waic_i <- var(ll_i)
      elpd_waic_i <- lpd_i - p_waic_i
      c(elpd_waic = elpd_waic_i, p_waic = p_waic_i)
    })
    pointwise <- do.call(rbind, waic_list)
    pointwise <- cbind(pointwise, waic = -2 * pointwise[, "elpd_waic"])

    throw_pwaic_warnings(pointwise[, "p_waic"], digits = 1)
    waic_object(pointwise, dims = c(S, N))
  }


#' @export
dim.waic <- function(x) {
  attr(x, "dims")
}

#' @rdname waic
#' @export
is.waic <- function(x) {
  inherits(x, "waic") && is.loo(x)
}


# internal ----------------------------------------------------------------

# structure the object returned by the waic methods
waic_object <- function(pointwise, dims) {
  estimates <- table_of_estimates(pointwise)
  out <- nlist(estimates, pointwise)
  # maintain backwards compatibility
  old_nms <- c("elpd_waic", "p_waic", "waic", "se_elpd_waic", "se_p_waic", "se_waic")
  out <- c(out, setNames(as.list(estimates), old_nms))
  structure(
    out,
    dims = dims,
    class = c("waic", "loo")
  )
}

# waic warnings
# @param p 'p_waic' estimates
throw_pwaic_warnings <- function(p, digits = 1, warn = TRUE) {
  badp <- p > 0.4
  if (any(badp)) {
    count <- sum(badp)
    prop <- count / length(badp)
    msg <- paste0("\n", count, " (", .fr(100 * prop, digits),
                  "%) p_waic estimates greater than 0.4. ",
                  "We recommend trying loo instead.")
    if (warn) .warn(msg) else cat(msg, "\n")
  }
  invisible(NULL)
}

#' Pareto smoothed importance sampling (deprecated, old version)
#'
#' As of version `2.0.0` this function is **deprecated**. Please use the
#' [psis()] function for the new PSIS algorithm.
#'
#' @export
#' @param lw A matrix or vector of log weights. For computing LOO, `lw =
#'   -log_lik`, the *negative* of an \eqn{S} (simulations) by \eqn{N} (data
#'   points) pointwise log-likelihood matrix.
#' @param wcp The proportion of importance weights to use for the generalized
#'   Pareto fit. The `100*wcp`\% largest weights are used as the sample
#'   from which to estimate the parameters of the generalized Pareto
#'   distribution.
#' @param wtrunc For truncating very large weights to \eqn{S}^`wtrunc`. Set
#'   to zero for no truncation.
#' @param cores The number of cores to use for parallelization. This defaults to
#'   the option `mc.cores` which can be set for an entire R session by
#'   `options(mc.cores = NUMBER)`, the old option `loo.cores` is now
#'   deprecated but will be given precedence over `mc.cores` until it is
#'   removed. **As of version 2.0.0, the default is now 1 core if
#'   `mc.cores` is not set, but we recommend using as many (or close to as
#'   many) cores as possible.**
#' @param llfun,llargs See [loo.function()].
#' @param ... Ignored when `psislw()` is called directly. The `...` is
#'   only used internally when `psislw()` is called by the [loo()]
#'   function.
#'
#' @return A named list with components `lw_smooth` (modified log weights) and
#'   `pareto_k` (estimated generalized Pareto shape parameter(s) k).
#'
#' @seealso [pareto-k-diagnostic] for PSIS diagnostics.
#'
#' @template loo-and-psis-references
#'
#' @importFrom parallel mclapply makePSOCKcluster stopCluster parLapply
#'
psislw <- function(lw, wcp = 0.2, wtrunc = 3/4,
                   cores = getOption("mc.cores", 1),
                   llfun = NULL, llargs = NULL,
                   ...) {
  .Deprecated("psis")

  cores <- loo_cores(cores)

  .psis <- function(lw_i) {
    x <- lw_i - max(lw_i)
    cutoff <- lw_cutpoint(x, wcp, MIN_CUTOFF)
    above_cut <- x > cutoff
    x_body <- x[!above_cut]
    x_tail <- x[above_cut]

    tail_len <- length(x_tail)
    if (tail_len < MIN_TAIL_LENGTH || all(x_tail == x_tail[1])) {
      if (all(x_tail == x_tail[1]))
        warning(
          "All tail values are the same. ",
          "Weights are truncated but not smoothed.",
          call. = FALSE
        )
      else if (tail_len < MIN_TAIL_LENGTH)
        warning(
          "Too few tail samples to fit generalized Pareto distribution.\n",
          "Weights are truncated but not smoothed.",
          call. = FALSE
        )

      x_new <- x
      k <- Inf
    } else {
      # store order of tail samples, fit gPd to the right tail samples, compute
      # order statistics for the fit, remap back to the original order, join
      # body and gPd smoothed tail
      tail_ord <- order(x_tail)
      exp_cutoff <- exp(cutoff)
      fit <- posterior::gpdfit(exp(x_tail) - exp_cutoff, wip=FALSE, min_grid_pts = 80)
      k <- fit$k
      sigma <- fit$sigma
      prb <- (seq_len(tail_len) - 0.5) / tail_len
      qq <- posterior::qgeneralized_pareto(prb, 0, sigma, k) + exp_cutoff
      smoothed_tail <- rep.int(0, tail_len)
      smoothed_tail[tail_ord] <- log(qq)
      x_new <- x
      x_new[!above_cut] <- x_body
      x_new[above_cut] <- smoothed_tail
    }
    # truncate (if wtrunc > 0) and renormalize,
    # return log weights and pareto k
    lw_new <- lw_normalize(lw_truncate(x_new, wtrunc))
    nlist(lw_new, k)
  }

  .psis_loop <- function(i) {
    if (LL_FUN) {
      ll_i <- llfun(i = i,
                    data = llargs$data[i,, drop=FALSE],
                    draws = llargs$draws)
      lw_i <- -1 * ll_i
    } else {
      lw_i <- lw[, i]
      ll_i <- -1 * lw_i
    }
    psis <- .psis(lw_i)
    if (FROM_LOO)
      nlist(lse = logSumExp(ll_i + psis$lw_new), k = psis$k)
    else
      psis
  }

  # minimal cutoff value. there must be at least 5 log-weights larger than this
  # in order to fit the gPd to the tail
  MIN_CUTOFF <- -700
  MIN_TAIL_LENGTH <- 5
  dots <- list(...)
  FROM_LOO <- if ("COMPUTE_LOOS" %in% names(dots))
    dots$COMPUTE_LOOS else FALSE

  if (!missing(lw)) {
    if (!is.matrix(lw))
      lw <- as.matrix(lw)
    N <- ncol(lw)
    LL_FUN <- FALSE
  } else {
    if (is.null(llfun) || is.null(llargs))
      stop("Either 'lw' or 'llfun' and 'llargs' must be specified.")
    N <- llargs$N
    LL_FUN <- TRUE
  }
  if (cores == 1) {
    # don't call functions from parallel package if cores=1
    out <- lapply(X = 1:N, FUN = .psis_loop)
  } else {
    # parallelize
    if (.Platform$OS.type != "windows") {
      out <- mclapply(X = 1:N, FUN = .psis_loop, mc.cores = cores)
    } else {
      # nocov start
      cl <- makePSOCKcluster(cores)
      on.exit(stopCluster(cl))
      out <- parLapply(cl, X = 1:N, fun = .psis_loop)
      # nocov end
    }
  }

  pareto_k <- vapply(out, "[[", 2L, FUN.VALUE = numeric(1))
  psislw_warnings(pareto_k)

  if (FROM_LOO) {
    loos <- vapply(out, "[[", 1L, FUN.VALUE = numeric(1))
    nlist(loos, pareto_k)
  } else {
    funval <- if (LL_FUN) llargs$S else nrow(lw)
    lw_smooth <- vapply(out, "[[", 1L, FUN.VALUE = numeric(funval))
    out <- nlist(lw_smooth, pareto_k)
    class(out) <- c("psis", "list")
    return(out)
  }
}


# internal ----------------------------------------------------------------
lw_cutpoint <- function(y, wcp, min_cut) {
  if (min_cut < log(.Machine$double.xmin))
    min_cut <- -700

  cp <- quantile(y, 1 - wcp, names = FALSE)
  max(cp, min_cut)
}

lw_truncate <- function(y, wtrunc) {
  if (wtrunc == 0)
    return(y)

  logS <- log(length(y))
  lwtrunc <- wtrunc * logS - logS + logSumExp(y)
  y[y > lwtrunc] <- lwtrunc
  y
}

lw_normalize <- function(y) {
  y - logSumExp(y)
}


# warnings about pareto k values ------------------------------------------
psislw_warnings <- function(k) {
  if (any(k > 0.7)) {
    .warn(
      "Some Pareto k diagnostic values are too high. ",
      .k_help()
    )
  } else if (any(k > 0.5)) {
    .warn(
      "Some Pareto k diagnostic values are slightly high. ",
      .k_help()
    )
  }
}

SW <- suppressWarnings

set.seed(123)
x <- matrix(rnorm(5000), 100, 50)

expect_deprecated <- function(object) {
  testthat::expect_warning(object, "deprecated", ignore.case = TRUE)
}

test_that("psislw throws deprecation warning", {
  expect_deprecated(psislw(x[, 1]))
})


test_that("psislw handles special cases, throws appropriate errors/warnings", {
  expect_snapshot(psis <- psislw(x[, 1], wcp = 0.01))
  expect_true(is.infinite(psis$pareto_k))

  expect_error(
    expect_deprecated(psislw(wcp = 0.2)),
    regexp = "'lw' or 'llfun' and 'llargs' must be specified"
  )
})

test_that("psislw returns expected results", {
  psis <- SW(psislw(x[, 1]))
  lw <- psis$lw_smooth
  expect_equal(length(psis), 2L)
  expect_equal(nrow(lw), nrow(x))
  expect_equal(lw[1], -5.6655489517740527106)
  expect_equal(lw[50], -5.188442371693668953)
  expect_equal(range(lw), c(-7.4142421808626526314, -2.6902215137943321643))
  expect_equal(psis$pareto_k, 0.17364505906017813075)
})

test_that("psislw function and matrix methods return same result", {
  set.seed(024)

  # fake data and posterior draws
  N <- 50
  K <- 10
  S <- 100
  a0 <- 3
  b0 <- 2
  p <- rbeta(1, a0, b0)
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y)
  b <- b0 + N * K - sum(y)
  draws <- rbeta(S, a, b)
  data <- data.frame(y, K)
  llfun <- function(i, data, draws) {
    dbinom(data$y, size = data$K, prob = draws, log = TRUE)
  }
  psislw_with_fn <- SW(psislw(llfun = llfun, llargs = nlist(data, draws, N, S)))

  # Check that we get same answer if using log-likelihood matrix
  ll <- sapply(1:N, function(i) llfun(i, data[i, , drop = FALSE], draws))
  psislw_with_mat <- SW(psislw(-ll))
  expect_equal(psislw_with_fn, psislw_with_mat)
})

test_that("psislw_warnings helper works properly", {
  k <- c(0, 0.1, 0.55, 0.75)
  expect_silent(psislw_warnings(k[1:2]))
  expect_warning(
    psislw_warnings(k[1:3]),
    "Some Pareto k diagnostic values are slightly high"
  )
  expect_warning(
    psislw_warnings(k),
    "Some Pareto k diagnostic values are too high"
  )
})

library(loo)
SW <- suppressWarnings

context("psislw")

set.seed(123)
x <- matrix(rnorm(5000), 100, 50)

expect_deprecated <- function(object) {
  testthat::expect_warning(object, "deprecated", ignore.case = TRUE)
}

test_that("psislw throws deprecation warning", {
  expect_deprecated(psislw(x[, 1]))
})


test_that("psislw handles special cases, throws appropriate errors/warnings", {
  expect_warning(
    psis <- psislw(x[, 1], wcp = 0.01),
    regexp = "All tail values are the same. Weights are truncated but not smoothed"
  )
  expect_true(is.infinite(psis$pareto_k))

  expect_warning(
    psislw(x[, 1], wcp = 0.01),
    regexp = "Some Pareto k diagnostic values are too high. See help('pareto-k-diagnostic') for details",
    fixed = TRUE
  )

  expect_error(
    expect_deprecated(psislw(wcp = 0.2)),
    regexp = "'lw' or 'llfun' and 'llargs' must be specified"
  )
})

test_that("psislw returns expected results", {
  psis <- SW(psislw(x[, 1]))
  lw <- psis$lw_smooth
  expect_equal(length(psis), 2L)
  expect_equal(nrow(lw), nrow(x))
  expect_equal(lw[1], -5.6655489517740527106)
  expect_equal(lw[50], -5.188442371693668953)
  expect_equal(range(lw), c(-7.4142421808626526314, -2.6902215137943321643))
  expect_equal(psis$pareto_k, 0.17364505906017813075)
})

test_that("psislw function and matrix methods return same result", {
  set.seed(024)

  # fake data and posterior draws
  N <- 50; K <- 10; S <- 100; a0 <- 3; b0 <- 2
  p <- rbeta(1, a0, b0)
  y <- rbinom(N, size = K, prob = p)
  a <- a0 + sum(y); b <- b0 + N * K - sum(y)
  draws <- rbeta(S, a, b)
  data <- data.frame(y,K)
  llfun <- function(i, data, draws) {
    dbinom(data$y, size = data$K, prob = draws, log = TRUE)
  }
  psislw_with_fn <- SW(psislw(llfun = llfun, llargs = nlist(data, draws, N, S)))

  # Check that we get same answer if using log-likelihood matrix
  ll <- sapply(1:N, function(i) llfun(i, data[i,, drop=FALSE], draws))
  psislw_with_mat <- SW(psislw(-ll))
  expect_equal(psislw_with_fn, psislw_with_mat)
})

test_that("psislw_warnings helper works properly", {
  k <- c(0, 0.1, 0.55, 0.75)
  expect_silent(psislw_warnings(k[1:2]))
  expect_warning(psislw_warnings(k[1:3]),
                 "Some Pareto k diagnostic values are slightly high")
  expect_warning(psislw_warnings(k),
                 "Some Pareto k diagnostic values are too high")
})

#' Estimate parameters of the Generalized Pareto distribution
#'
#' Given a sample \eqn{x}, Estimate the parameters \eqn{k} and \eqn{\sigma} of
#' the generalized Pareto distribution (GPD), assuming the location parameter is
#' 0. By default the fit uses a prior for \eqn{k}, which will stabilize
#' estimates for very small sample sizes (and low effective sample sizes in the
#' case of MCMC samples). The weakly informative prior is a Gaussian prior
#' centered at 0.5.
#'
#' @export
#' @param x A numeric vector. The sample from which to estimate the parameters.
#' @param wip Logical indicating whether to adjust \eqn{k} based on a weakly
#'   informative Gaussian prior centered on 0.5. Defaults to `TRUE`.
#' @param min_grid_pts The minimum number of grid points used in the fitting
#'   algorithm. The actual number used is `min_grid_pts + floor(sqrt(length(x)))`.
#' @param sort_x If `TRUE` (the default), the first step in the fitting
#'   algorithm is to sort the elements of `x`. If `x` is already
#'   sorted in ascending order then `sort_x` can be set to `FALSE` to
#'   skip the initial sorting step.
#' @return A named list with components `k` and `sigma`.
#'
#' @details Here the parameter \eqn{k} is the negative of \eqn{k} in Zhang &
#'   Stephens (2009).
#'
#' @seealso [psis()], [pareto-k-diagnostic]
#'
#' @references
#' Zhang, J., and Stephens, M. A. (2009). A new and efficient estimation method
#' for the generalized Pareto distribution. *Technometrics* **51**, 316-325.
#'
gpdfit <- function(x, wip = TRUE, min_grid_pts = 30, sort_x = TRUE) {
  posterior::gpdfit(
    x = x,
    wip = wip,
    min_grid_pts = min_grid_pts,
    sort_x = sort_x
  )
}

test_that("gpdfit returns correct result", {
  set.seed(123)
  x <- rexp(100)
  gpdfit_val_old <- unlist(gpdfit(x, wip = FALSE, min_grid_pts = 80))
  expect_snapshot_value(gpdfit_val_old, style = "serialize")

  gpdfit_val_wip <- unlist(gpdfit(x, wip = TRUE, min_grid_pts = 80))
  expect_snapshot_value(gpdfit_val_wip, style = "serialize")

  gpdfit_val_wip_default_grid <- unlist(gpdfit(x, wip = TRUE))
  expect_snapshot_value(gpdfit_val_wip_default_grid, style = "serialize")
})

test_that("gpdfit returns correct result", {
  set.seed(123)
  x <- rexp(100)
  gpdfit_val_old <- unlist(gpdfit(x, wip = FALSE, min_grid_pts = 80))
  expect_snapshot_value(gpdfit_val_old, style = "serialize")

  gpdfit_val_wip <- unlist(gpdfit(x, wip = TRUE, min_grid_pts = 80))
  expect_snapshot_value(gpdfit_val_wip, style = "serialize")

  gpdfit_val_wip_default_grid <- unlist(gpdfit(x, wip = TRUE))
  expect_snapshot_value(gpdfit_val_wip_default_grid, style = "serialize")
})

test_that("qgpd returns the correct result ", {
  probs <- seq(from = 0, to = 1, by = 0.25)
  q1 <- qgpd(probs, k = 1, sigma = 1)
  expect_equal(q1, c(0, 1 / 3, 1, 3, Inf))

  q2 <- qgpd(probs, k = 1, sigma = 0)
  expect_true(all(is.nan(q2)))
})

params <-
list(EVAL = TRUE)

## ----SETTINGS-knitr, include=FALSE--------------------------------------------
stopifnot(require(knitr))
opts_chunk$set(
  comment=NA,
  eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
  dev = "png",
  dpi = 150,
  fig.asp = 0.618,
  fig.width = 5,
  out.width = "60%",
  fig.align = "center"
)

## ----setup, message=FALSE-----------------------------------------------------
library("rstan")
library("loo")
set.seed(4711)

## ----llfun_logistic-----------------------------------------------------------
# we'll add an argument log to toggle whether this is a log-likelihood or 
# likelihood function. this will be useful later in the vignette.
llfun_logistic <- function(data_i, draws, log = TRUE) {
  x_i <- as.matrix(data_i[, which(grepl(colnames(data_i), pattern = "X")), drop=FALSE])
  logit_pred <- draws %*% t(x_i)
  dbinom(x = data_i$y, size = 1, prob = 1/(1 + exp(-logit_pred)), log = log)
}

## ----eval=FALSE---------------------------------------------------------------
#  # Prepare data
#  url <- "http://stat.columbia.edu/~gelman/arm/examples/arsenic/wells.dat"
#  wells <- read.table(url)
#  wells$dist100 <- with(wells, dist / 100)
#  X <- model.matrix(~ dist100 + arsenic, wells)
#  standata <- list(y = wells$switch, X = X, N = nrow(X), P = ncol(X))
#  
#  # Compile
#  stan_mod <- stan_model("logistic.stan")
#  
#  # Fit model
#  fit_1 <- sampling(stan_mod, data = standata, seed = 4711)
#  print(fit_1, pars = "beta")

## ----eval=FALSE---------------------------------------------------------------
#  # used for draws argument to loo_i
#  parameter_draws_1 <- extract(fit_1)$beta
#  
#  # used for data argument to loo_i
#  stan_df_1 <- as.data.frame(standata)
#  
#  # compute relative efficiency (this is slow and optional but is recommended to allow
#  # for adjusting PSIS effective sample size based on MCMC effective sample size)
#  r_eff <- relative_eff(llfun_logistic,
#                        log = FALSE, # relative_eff wants likelihood not log-likelihood values
#                        chain_id = rep(1:4, each = 1000),
#                        data = stan_df_1,
#                        draws = parameter_draws_1,
#                        cores = 2)
#  
#  loo_i(i = 1, llfun_logistic, r_eff = r_eff, data = stan_df_1, draws = parameter_draws_1)

## ----eval=FALSE---------------------------------------------------------------
#  set.seed(4711)
#  loo_ss_1 <-
#    loo_subsample(
#      llfun_logistic,
#      observations = 100, # take a subsample of size 100
#      cores = 2,
#      # these next objects were computed above
#      r_eff = r_eff,
#      draws = parameter_draws_1,
#      data = stan_df_1
#    )
#  print(loo_ss_1)

## ----eval=FALSE---------------------------------------------------------------
#  set.seed(4711)
#  loo_ss_1b <-
#    update(
#      loo_ss_1,
#      observations = 200, # subsample 200 instead of 100
#      r_eff = r_eff,
#      draws = parameter_draws_1,
#      data = stan_df_1
#    )
#  print(loo_ss_1b)

## ----eval=FALSE---------------------------------------------------------------
#  set.seed(4711)
#  loo_ss_1c <-
#    loo_subsample(
#      x = llfun_logistic,
#      r_eff = r_eff,
#      draws = parameter_draws_1,
#      data = stan_df_1,
#      observations = 100,
#      estimator = "hh_pps", # use Hansen-Hurwitz
#      loo_approximation = "lpd", # use lpd instead of plpd
#      loo_approximation_draws = 100,
#      cores = 2
#    )
#  print(loo_ss_1c)

## ----eval=FALSE---------------------------------------------------------------
#  fit_laplace <- optimizing(stan_mod, data = standata, draws = 2000,
#                            importance_resampling = TRUE)
#  parameter_draws_laplace <- fit_laplace$theta_tilde # draws from approximate posterior
#  log_p <- fit_laplace$log_p # log density of the posterior
#  log_g <- fit_laplace$log_g # log density of the approximation

## ----eval=FALSE---------------------------------------------------------------
#  set.seed(4711)
#  loo_ap_1 <-
#    loo_approximate_posterior(
#      x = llfun_logistic,
#      draws = parameter_draws_laplace,
#      data = stan_df_1,
#      log_p = log_p,
#      log_g = log_g,
#      cores = 2
#    )
#  print(loo_ap_1)

## ----eval=FALSE---------------------------------------------------------------
#  set.seed(4711)
#  loo_ap_ss_1 <-
#    loo_subsample(
#      x = llfun_logistic,
#      draws = parameter_draws_laplace,
#      data = stan_df_1,
#      log_p = log_p,
#      log_g = log_g,
#      observations = 100,
#      cores = 2
#    )
#  print(loo_ap_ss_1)

## ----eval=FALSE---------------------------------------------------------------
#  standata$X[, "arsenic"] <- log(standata$X[, "arsenic"])
#  fit_2 <- sampling(stan_mod, data = standata)
#  parameter_draws_2 <- extract(fit_2)$beta
#  stan_df_2 <- as.data.frame(standata)
#  
#  # recompute subsampling loo for first model for demonstration purposes
#  
#  # compute relative efficiency (this is slow and optional but is recommended to allow
#  # for adjusting PSIS effective sample size based on MCMC effective sample size)
#  r_eff_1 <- relative_eff(
#    llfun_logistic,
#    log = FALSE, # relative_eff wants likelihood not log-likelihood values
#    chain_id = rep(1:4, each = 1000),
#    data = stan_df_1,
#    draws = parameter_draws_1,
#    cores = 2
#  )
#  
#  set.seed(4711)
#  loo_ss_1 <- loo_subsample(
#    x = llfun_logistic,
#    r_eff = r_eff_1,
#    draws = parameter_draws_1,
#    data = stan_df_1,
#    observations = 200,
#    cores = 2
#  )
#  
#  # compute subsampling loo for a second model (with log-arsenic)
#  
#  r_eff_2 <- relative_eff(
#    llfun_logistic,
#    log = FALSE, # relative_eff wants likelihood not log-likelihood values
#    chain_id = rep(1:4, each = 1000),
#    data = stan_df_2,
#    draws = parameter_draws_2,
#    cores = 2
#  )
#  loo_ss_2 <- loo_subsample(
#    x = llfun_logistic,
#    r_eff = r_eff_2,
#    draws = parameter_draws_2,
#    data = stan_df_2,
#    observations = 200,
#    cores = 2
#  )
#  
#  print(loo_ss_2)

## ----eval=FALSE---------------------------------------------------------------
#  # Compare
#  comp <- loo_compare(loo_ss_1, loo_ss_2)
#  print(comp)

## ----eval=FALSE---------------------------------------------------------------
#  loo_ss_2 <-
#    loo_subsample(
#      x = llfun_logistic,
#      r_eff = r_eff_2,
#      draws = parameter_draws_2,
#      data = stan_df_2,
#      observations = loo_ss_1,
#      cores = 2
#    )

## ----eval=FALSE---------------------------------------------------------------
#  idx <- obs_idx(loo_ss_1)
#  loo_ss_2 <- loo_subsample(
#    x = llfun_logistic,
#    r_eff = r_eff_2,
#    draws = parameter_draws_2,
#    data = stan_df_2,
#    observations = idx,
#    cores = 2
#  )

## ----eval=FALSE---------------------------------------------------------------
#  comp <- loo_compare(loo_ss_1, loo_ss_2)
#  print(comp)

## ----eval=FALSE---------------------------------------------------------------
#  # use loo() instead of loo_subsample() to compute full PSIS-LOO for model 2
#  loo_full_2 <- loo(
#    x = llfun_logistic,
#    r_eff = r_eff_2,
#    draws = parameter_draws_2,
#    data = stan_df_2,
#    cores = 2
#  )
#  loo_compare(loo_ss_1, loo_full_2)

#' Convenience function for extracting pointwise estimates
#'
#' @export
#' @param x A `loo` object, for example one returned by [loo()],
#'   [loo_subsample()], [loo_approximate_posterior()], [loo_moment_match()], etc.
#' @param estimate Which pointwise estimate to return. By default all are
#'   returned. The objects returned by the different functions ([loo()],
#'   [loo_subsample()], etc.) have slightly different estimates available.
#'   Typically at a minimum the estimates `elpd_loo`, `looic`, `mcse_elpd_loo`,
#'   `p_loo`, and `influence_pareto_k` will be available, but there may be
#'   others.
#' @param ... Currently ignored.
#' @return A vector of length equal to the number of observations.
#'
#' @examples
#' x <- loo(example_loglik_array())
#' pointwise(x, "elpd_loo")
#'
pointwise <- function(x, estimate, ...) {
  UseMethod("pointwise")
}

#' @rdname pointwise
#' @export
pointwise.loo <- function(x, estimate, ...) {
  stopifnot(is.character(estimate), length(estimate) == 1)
  pw <- x$pointwise
  if (is.null(pw)) {
    stop("No pointwise estimates found.", call. = FALSE)
  }
  estimates <- colnames(pw)
  if (!(estimate %in% estimates)) {
    stop(
      "'", estimate, "' not found.",
      " Available estimates are: \n",
      paste(shQuote(estimates), collapse=", ")
    )
  }
  pw[, estimate]
}

#' Continuously ranked probability score
#'
#' The `crps()` and `scrps()` functions and their `loo_*()` counterparts can be
#' used to compute the continuously ranked probability score (CRPS) and scaled
#' CRPS (SCRPS) (as defined by Bolin and Wallin, 2023). CRPS is a proper scoring rule, and
#' strictly proper when the first moment of the predictive distribution is
#' finite. Both can be expressed in terms of samples form the predictive
#' distribution. See, for example, a paper by Gneiting and Raftery (2007)
#' for a comprehensive discussion on CRPS.
#'
#' To compute (S)CRPS, the user needs to provide two sets of draws, `x` and
#' `x2`, from the predictive distribution. This is due to the fact that formulas
#' used to compute CRPS involve an expectation of the absolute difference of `x`
#' and `x2`, both having the same distribution. See the `permutations` argument,
#' as well as Gneiting and Raftery (2007) for details.
#'
#' @export
#' @param x A `S` by `N` matrix (draws by observations), or a vector of length
#'   `S` when only single observation is provided in `y`.
#' @param x2 Independent draws from the same distribution as draws in `x`.
#'   Should be of the identical dimension.
#' @param y A vector of observations or a single value.
#' @param permutations An integer, with default value of 1,  specifying how many
#'   times the expected value of  |X - X'| (`|x - x2|`) is computed. The row
#'   order of `x2` is shuffled as elements `x` and `x2` are typically drawn
#'   given the same values of parameters. This happens, e.g., when one calls
#'   `posterior_predict()` twice for a fitted \pkg{rstanarm} or \pkg{brms}
#'   model. Generating more permutations is expected to decrease the variance of
#'   the computed expected value.
#' @param ... Passed on to [E_loo()] in the `loo_*()` version of these
#'   functions.
#'
#' @return A list containing two elements: `estimates` and `pointwise`.
#'   The former reports estimator and standard error and latter the pointwise
#'   values. Following Bolin & Wallin (2023), a larger value is better.
#'
#' @examples
#' \dontrun{
#' # An example using rstanarm
#' library(rstanarm)
#' data("kidiq")
#' fit <- stan_glm(kid_score ~ mom_hs + mom_iq, data = kidiq)
#' ypred1 <- posterior_predict(fit)
#' ypred2 <- posterior_predict(fit)
#' crps(ypred1, ypred2, y = fit$y)
#' loo_crps(ypred1, ypred2, y = fit$y, log_lik = log_lik(fit))
#' }
#'
#' @references
#' Bolin, D., & Wallin, J. (2023). Local scale invariance and robustness of
#' proper scoring rules. Statistical Science, 38(1):140-159.
#'
#' Gneiting, T., & Raftery, A. E. (2007). Strictly Proper Scoring Rules,
#' Prediction, and Estimation. Journal of the American Statistical Association,
#' 102(477), 359–378.
crps <- function(x, ...) {
  UseMethod("crps")
}


#' @rdname crps
#' @export
scrps <- function(x, ...) {
  UseMethod("scrps")
}

#' @rdname crps
#' @export
loo_crps <- function(x, ...) {
  UseMethod("loo_crps")
}

#' @rdname crps
#' @export
loo_scrps <- function(x, ...) {
  UseMethod("loo_scrps")
}


#' @rdname crps
#' @export
crps.matrix <- function(x, x2, y, ..., permutations = 1) {
  validate_crps_input(x, x2, y)
  repeats <- replicate(permutations, EXX_compute(x, x2), simplify = F)
  EXX <- Reduce(`+`, repeats) / permutations
  EXy <- colMeans(abs(sweep(x, 2, y)))
  crps_output(.crps_fun(EXX, EXy))
}


#' Method for a single data point
#' @rdname crps
#' @export
crps.numeric <- function(x, x2, y, ..., permutations = 1) {
  stopifnot(length(x) == length(x2),
            length(y) == 1)
  crps.matrix(as.matrix(x), as.matrix(x2), y, permutations)
}


#' @rdname crps
#' @export
#' @param log_lik A log-likelihood matrix the same size as `x`.
#' @param r_eff An optional vector of relative effective sample size estimates
#'   containing one element per observation. See [psis()] for details.
#' @param cores The number of cores to use for parallelization of `[psis()]`.
#'   See [psis()] for details.
loo_crps.matrix <-
  function(x,
           x2,
           y,
           log_lik,
           ...,
           permutations = 1,
           r_eff = 1,
           cores = getOption("mc.cores", 1)) {
  validate_crps_input(x, x2, y, log_lik)
  repeats <- replicate(permutations,
                       EXX_loo_compute(x, x2, log_lik, r_eff = r_eff, ...),
                       simplify = F)
  EXX <- Reduce(`+`, repeats) / permutations
  psis_obj <- psis(-log_lik, r_eff = r_eff, cores = cores)
  EXy <- E_loo(abs(sweep(x, 2, y)), psis_obj, log_ratios = -log_lik, ...)$value
  crps_output(.crps_fun(EXX, EXy))
}


#' @rdname crps
#' @export
scrps.matrix <- function(x, x2, y, ..., permutations = 1) {
  validate_crps_input(x, x2, y)
  repeats <- replicate(permutations, EXX_compute(x, x2), simplify = F)
  EXX <- Reduce(`+`, repeats) / permutations
  EXy <- colMeans(abs(sweep(x, 2, y)))
  crps_output(.crps_fun(EXX, EXy, scale = TRUE))
}

#' @rdname crps
#' @export
scrps.numeric <- function(x, x2, y, ..., permutations = 1) {
  stopifnot(length(x) == length(x2),
            length(y) == 1)
  scrps.matrix(as.matrix(x), as.matrix(x2), y, permutations)
}


#' @rdname crps
#' @export
loo_scrps.matrix <-
  function(
    x,
    x2,
    y,
    log_lik,
    ...,
    permutations = 1,
    r_eff = 1,
    cores = getOption("mc.cores", 1)) {
  validate_crps_input(x, x2, y, log_lik)
  repeats <- replicate(permutations,
                       EXX_loo_compute(x, x2, log_lik, r_eff = r_eff, ...),
                       simplify = F)
  EXX <- Reduce(`+`, repeats) / permutations
  psis_obj <- psis(-log_lik, r_eff = r_eff, cores = cores)
  EXy <- E_loo(abs(sweep(x, 2, y)), psis_obj, log_ratios = -log_lik, ...)$value
  crps_output(.crps_fun(EXX, EXy, scale = TRUE))
}

# ------------ Internals ----------------


EXX_compute <- function(x, x2) {
  S <- nrow(x)
  colMeans(abs(x - x2[sample(1:S),]))
}


EXX_loo_compute <- function(x, x2, log_lik, r_eff = 1, ...) {
  S <- nrow(x)
  shuffle <- sample (1:S)
  x2 <- x2[shuffle,]
  log_lik2 <- log_lik[shuffle,]
  psis_obj_joint <- psis(-log_lik - log_lik2 , r_eff = r_eff)
  E_loo(abs(x - x2), psis_obj_joint, log_ratios = -log_lik - log_lik2, ...)$value
}


#' Function to compute crps and scrps
#' @noRd
.crps_fun <- function(EXX, EXy, scale = FALSE) {
  if (scale) return(-EXy/EXX - 0.5 * log(EXX))
  0.5 * EXX - EXy
}

#' Compute output data for crps functions
#' @noRd
crps_output <- function(crps_pw) {
  n <- length(crps_pw)
  out <- list()
  out$estimates <- c(mean(crps_pw), sd(crps_pw) / sqrt(n))
  names(out$estimates) <- c('Estimate', 'SE')
  out$pointwise <- crps_pw
  out
}

#' Validate input of CRPS functions
#'
#' Check that predictive draws and observed data are of compatible shape
#' @noRd
validate_crps_input <- function(x, x2, y, log_lik = NULL) {
  stopifnot(is.numeric(x),
            is.numeric(x2),
            is.numeric(y),
            identical(dim(x), dim(x2)),
            ncol(x) == length(y),
            ifelse(is.null(log_lik), TRUE, identical(dim(log_lik), dim(x)))
            )
}

set.seed(123456789)
n <- 10
S <- 100
y <- rnorm(n)
x1 <- matrix(rnorm(n * S), nrow = S)
x2 <- matrix(rnorm(n * S), nrow = S)
ll <- matrix(rnorm(n * S) * 0.1 - 1, nrow = S)

with_seed <- function(seed, code) {
  code <- substitute(code)
  orig.seed <- .Random.seed
  on.exit(.Random.seed <<- orig.seed)
  set.seed(seed)
  eval.parent(code)
}

test_that("crps computation is correct", {
  expect_equal(.crps_fun(2.0, 1.0), 0.0)
  expect_equal(.crps_fun(1.0, 2.0), -1.5)
  expect_equal(.crps_fun(pi, pi^2), 0.5 * pi - pi^2)

  expect_equal(.crps_fun(1.0, 0.0, scale = TRUE), 0.0)
  expect_equal(.crps_fun(1.0, 2.0, scale = TRUE), -2.0)
  expect_equal(.crps_fun(pi, pi^2, scale = TRUE), -pi^2/pi - 0.5 * log(pi))
})

test_that("crps matches snapshots", {
  expect_snapshot_value(with_seed(1, crps(x1, x2, y)), style = "serialize")
  expect_snapshot_value(with_seed(1, scrps(x1, x2, y)), style = "serialize")
  expect_snapshot_value(with_seed(1, loo_crps(x1, x2, y, ll)), style = "serialize")
  expect_snapshot_value(with_seed(1, loo_scrps(x1, x2, y, ll)), style = "serialize")
})

test_that("input validation throws correct errors", {
  expect_error(validate_crps_input(as.character(x1), x2, y),
               "is.numeric(x) is not TRUE",
               fixed = TRUE)
  expect_error(validate_crps_input(x1, as.character(x2), y),
               "is.numeric(x2) is not TRUE",
               fixed = TRUE)
  expect_error(validate_crps_input(x1, x2, c('a', 'b')),
               "is.numeric(y) is not TRUE",
               fixed = TRUE)
  expect_error(validate_crps_input(x1, t(x2), y),
               "identical(dim(x), dim(x2)) is not TRUE",
               fixed = TRUE)
  expect_error(validate_crps_input(x1, x2, c(1, 2)),
               "ncol(x) == length(y) is not TRUE",
               fixed = TRUE)
  expect_error(validate_crps_input(x1, x2, y, t(ll)),
               "ifelse(is.null(log_lik), TRUE, identical(dim(log_lik), dim(x))) is not TRUE",
               fixed = TRUE)
})

test_that("methods for single data point don't error", {
  expect_silent(crps(x1[,1], x2[,1], y[1]))
  expect_silent(scrps(x1[,1], x2[,1], y[1]))
})

set.seed(123456789)
n <- 10
S <- 100
y <- rnorm(n)
x1 <- matrix(rnorm(n * S), nrow = S)
x2 <- matrix(rnorm(n * S), nrow = S)
ll <- matrix(rnorm(n * S) * 0.1 - 1, nrow = S)

with_seed <- function(seed, code) {
  code <- substitute(code)
  orig.seed <- .Random.seed
  on.exit(.Random.seed <<- orig.seed)
  set.seed(seed)
  eval.parent(code)
}

test_that("crps computation is correct", {
  expect_equal(.crps_fun(2.0, 1.0), 0.0)
  expect_equal(.crps_fun(1.0, 2.0), -1.5)
  expect_equal(.crps_fun(pi, pi^2), 0.5 * pi - pi^2)

  expect_equal(.crps_fun(1.0, 0.0, scale = TRUE), 0.0)
  expect_equal(.crps_fun(1.0, 2.0, scale = TRUE), -2.0)
  expect_equal(.crps_fun(pi, pi^2, scale = TRUE), -pi^2/pi - 0.5 * log(pi))
})

test_that("crps matches references", {
  expect_equal_to_reference(with_seed(1, crps(x1, x2, y)), 'reference-results/crps.rds')
  expect_equal_to_reference(with_seed(1, scrps(x1, x2, y)), 'reference-results/scrps.rds')
  expect_equal_to_reference(with_seed(1, loo_crps(x1, x2, y, ll)), 'reference-results/loo_crps.rds')
  expect_equal_to_reference(with_seed(1, loo_scrps(x1, x2, y, ll)), 'reference-results/loo_scrps.rds')
})

test_that("input validation throws correct errors", {
  expect_error(validate_crps_input(as.character(x1), x2, y),
               "is.numeric(x) is not TRUE",
               fixed = TRUE)
  expect_error(validate_crps_input(x1, as.character(x2), y),
               "is.numeric(x2) is not TRUE",
               fixed = TRUE)
  expect_error(validate_crps_input(x1, x2, c('a', 'b')),
               "is.numeric(y) is not TRUE",
               fixed = TRUE)
  expect_error(validate_crps_input(x1, t(x2), y),
               "identical(dim(x), dim(x2)) is not TRUE",
               fixed = TRUE)
  expect_error(validate_crps_input(x1, x2, c(1, 2)),
               "ncol(x) == length(y) is not TRUE",
               fixed = TRUE)
  expect_error(validate_crps_input(x1, x2, y, t(ll)),
               "ifelse(is.null(log_lik), TRUE, identical(dim(log_lik), dim(x))) is not TRUE",
               fixed = TRUE)
})

test_that("methods for single data point don't error", {
  expect_silent(crps(x1[,1], x2[,1], y[1]))
  expect_silent(scrps(x1[,1], x2[,1], y[1]))
})

#' Helper functions for K-fold cross-validation
#'
#' @description These functions can be used to generate indexes for use with
#'   K-fold cross-validation. See the **Details** section for explanations.
#'
#' @name kfold-helpers
#' @param K The number of folds to use.
#' @param N The number of observations in the data.
#' @param x A discrete variable of length `N` with at least `K` levels (unique
#'   values). Will be coerced to a [factor][factor()].
#'
#' @return An integer vector of length `N` where each element is an index in `1:K`.
#'
#' @details
#' `kfold_split_random()` splits the data into `K` groups
#' of equal size (or roughly equal size).
#'
#' For a categorical variable `x` `kfold_split_stratified()`
#' splits the observations into `K` groups ensuring that relative
#' category frequencies are approximately preserved.
#'
#' For a grouping variable `x`, `kfold_split_grouped()` places
#' all observations in `x` from the same group/level together in
#' the same fold. The selection of which groups/levels go into which
#' fold (relevant when when there are more groups than folds) is
#' randomized.
#'
#' @examples
#' ids <- kfold_split_random(K = 5, N = 20)
#' print(ids)
#' table(ids)
#'
#'
#' x <- sample(c(0, 1), size = 200, replace = TRUE, prob = c(0.05, 0.95))
#' table(x)
#' ids <- kfold_split_stratified(K = 5, x = x)
#' print(ids)
#' table(ids, x)
#'
#' grp <- gl(n = 50, k = 15, labels = state.name)
#' length(grp)
#' head(table(grp))
#'
#' ids_10 <- kfold_split_grouped(K = 10, x = grp)
#' (tab_10 <- table(grp, ids_10))
#' colSums(tab_10)
#'
#' ids_9 <- kfold_split_grouped(K = 9, x = grp)
#' (tab_9 <- table(grp, ids_9))
#' colSums(tab_9)
#'
NULL

#' @rdname kfold-helpers
#' @export
kfold_split_random <- function(K = 10, N = NULL) {
  stopifnot(
    !is.null(N),
    K == as.integer(K),
    N == as.integer(N),
    length(K) == 1,
    length(N) == 1,
    K > 1,
    K <= N
  )
  perm <- sample.int(N)
  idx <- ceiling(seq(from = 1, to = N, length.out = K + 1))
  bins <- .bincode(perm, breaks = idx, right = FALSE, include.lowest = TRUE)
  return(bins)
}

#' @rdname kfold-helpers
#' @export
kfold_split_stratified <- function(K = 10, x = NULL) {
  stopifnot(
    !is.null(x),
    K == as.integer(K),
    length(K) == 1,
    K > 1,
    K <= length(x)
  )
  x <- as.integer(as.factor(x))
  Nlev <- length(unique(x))
  N <- length(x)
  xids <- numeric()
  for (l in 1:Nlev) {
    idx <- which(x == l)
    if (length(idx) > 1) {
      xids <- c(xids, sample(idx))
    } else {
      xids <- c(xids, idx)
    }
  }
  bins <- rep(NA, N)
  bins[xids] <- rep(1:K, ceiling(N/K))[1:N]
  return(bins)
}

#' @rdname kfold-helpers
#' @export
kfold_split_grouped <- function(K = 10, x = NULL) {
  stopifnot(
    !is.null(x),
    K == as.integer(K),
    length(K) == 1,
    K > 1,
    K <= length(x)
  )

  Nlev <- length(unique(x))
  if (Nlev < K) {
    stop("'K' must not be bigger than the number of levels/groups in 'x'.")
  }
  x <- as.integer(as.factor(x))
  if (Nlev == K) {
    return(x)
  }

  # Otherwise we have Nlev > K
  S1 <- ceiling(Nlev / K)  # number of levels in largest groups of levels
  N_S2 <- S1 * K - Nlev    # number of groups of levels of size S1 - 1
  N_S1 <- K - N_S2         # number of groups of levels of size S1

  perm <- sample.int(Nlev) # permute group levels
  brks <- seq(from = S1 + 0.5, by = S1, length.out = N_S1)
  if (N_S2 > 0) {
    brks2 <- seq(from = brks[N_S1] + S1 - 1, by = S1 - 1, length.out = N_S2 - 1)
    brks <- c(brks, brks2)
  }
  grps <- findInterval(perm, vec = brks) + 1  # +1 so min is 1 not 0

  bins <- rep(NA, length(x))
  for (j in perm) {
    bins[x == j] <- grps[j]
  }
  return(bins)
}

set.seed(14014)

test_that("kfold_split_random works", {
  fold_rand <- kfold_split_random(10, 100)
  expect_length(fold_rand, 100)
  expect_equal(sort(unique(fold_rand)), 1:10)
  expect_equal(sum(fold_rand == 2), sum(fold_rand == 9))
})

test_that("kfold_split_stratified works", {
  y <- rep(c(0, 1), times = c(10, 190))
  fold_strat <- kfold_split_stratified(5, y)
  expect_true(all(table(fold_strat) == 40))

  y <- rep(c(1, 2, 3), times = c(15, 33, 42))
  fold_strat <- kfold_split_stratified(7, y)
  expect_equal(range(table(fold_strat)), c(12, 13))

  y <- mtcars$cyl
  fold_strat <- kfold_split_stratified(10, y)
  expect_equal(range(table(fold_strat)), c(3, 4))

  # test when a group has 1 observation
  # https://github.com/stan-dev/loo/issues/277
  y <- rep(c(1, 2, 3), times = c(20, 40, 1))
  expect_silent(fold_strat <- kfold_split_stratified(5, y)) # used to be a warning before fixing issue #277
  tab <- table(fold_strat, y)
  expect_equal(tab[1, ], c("1" = 4, "2" = 8, "3" = 1))
  for (i in 2:nrow(tab)) {
    expect_equal(tab[i, ], c("1" = 4, "2" = 8, "3" = 0))
  }
})

test_that("kfold_split_grouped works", {
  grp <- gl(n = 50, k = 15, labels = state.name)
  fold_group <- kfold_split_grouped(x = grp)
  expect_true(all(table(fold_group) == 75))
  expect_equal(sum(table(fold_group)), length(grp))

  fold_group <- kfold_split_grouped(K = 9, x = grp)
  expect_false(all(table(fold_group) == 75))
  expect_equal(sum(table(fold_group)), length(grp))

  grp <- gl(n = 50, k = 4, labels = state.name)
  grp[grp == "Montana"] <- "Utah"
  fold_group <- kfold_split_grouped(K = 10, x = grp)
  expect_equal(sum(table(fold_group)), length(grp) - 4)

  grp <- rep(c("A", "B"), each = 20)
  fold_group <- kfold_split_grouped(K = 2, x = grp)
  expect_equal(fold_group, as.integer(as.factor(grp)))
})

test_that("kfold helpers throw correct errors", {
  expect_error(kfold_split_random(10), "!is.null(N) is not TRUE", fixed = TRUE)
  expect_error(
    kfold_split_random(10.5, 100),
    "K == as.integer(K) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_random(10, 100.5),
    "N == as.integer(N) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_random(K = c(1, 1), N = 100),
    "length(K) == 1 is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_random(N = c(100, 100)),
    "length(N) == 1 is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_random(K = 5, N = 4),
    "K <= N is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_random(K = 1, N = 4),
    "K > 1 is not TRUE",
    fixed = TRUE
  )

  y <- sample(c(0, 1), size = 200, replace = TRUE, prob = c(0.05, 0.95))
  expect_error(
    kfold_split_stratified(10),
    "!is.null(x) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_stratified(10.5, y),
    "K == as.integer(K) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_stratified(K = c(1, 1), y),
    "length(K) == 1 is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_stratified(K = 201, y),
    "K <= length(x) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_stratified(K = 1, y),
    "K > 1 is not TRUE",
    fixed = TRUE
  )

  grp <- gl(n = 50, k = 15)
  expect_error(kfold_split_grouped(10), "!is.null(x) is not TRUE", fixed = TRUE)
  expect_error(
    kfold_split_grouped(3, c(1, 1, 1)),
    "'K' must not be bigger than the number of levels/groups in 'x'",
    fixed = TRUE
  )
  expect_error(
    kfold_split_grouped(10.5, grp),
    "K == as.integer(K) is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_grouped(K = c(1, 1), grp),
    "length(K) == 1 is not TRUE",
    fixed = TRUE
  )
  expect_error(
    kfold_split_grouped(K = 1, grp),
    "K > 1 is not TRUE",
    fixed = TRUE
  )
})


test_that("print_dims.kfold works", {
  xx <- structure(list(), K = 17, class = c("kfold", "loo"))
  expect_snapshot(print_dims(xx))

  attr(xx, "K") <- NULL
  expect_silent(print_dims(xx))
})

library(loo)
set.seed(14014)

context("kfold helper functions")

test_that("kfold_split_random works", {
  fold_rand <- kfold_split_random(10, 100)
  expect_length(fold_rand, 100)
  expect_equal(sort(unique(fold_rand)), 1:10)
  expect_equal(sum(fold_rand == 2), sum(fold_rand == 9))
})

test_that("kfold_split_stratified works", {
  y <- rep(c(0, 1), times = c(10, 190))
  fold_strat <- kfold_split_stratified(5, y)
  expect_true(all(table(fold_strat) == 40))

  y <- rep(c(1, 2, 3), times = c(15, 33, 42))
  fold_strat <- kfold_split_stratified(7, y)
  expect_equal(range(table(fold_strat)), c(12, 13))

  y <- mtcars$cyl
  fold_strat <- kfold_split_stratified(10, y)
  expect_equal(range(table(fold_strat)), c(3, 4))
})

test_that("kfold_split_grouped works", {
  grp <- gl(n = 50, k = 15, labels = state.name)
  fold_group <- kfold_split_grouped(x = grp)
  expect_true(all(table(fold_group) == 75))
  expect_equal(sum(table(fold_group)), length(grp))

  fold_group <- kfold_split_grouped(K = 9, x = grp)
  expect_false(all(table(fold_group) == 75))
  expect_equal(sum(table(fold_group)), length(grp))

  grp <- gl(n = 50, k = 4, labels = state.name)
  grp[grp == "Montana"] <- "Utah"
  fold_group <- kfold_split_grouped(K = 10, x = grp)
  expect_equal(sum(table(fold_group)), length(grp) - 4)

  grp <- rep(c("A","B"), each = 20)
  fold_group <- kfold_split_grouped(K = 2, x = grp)
  expect_equal(fold_group, as.integer(as.factor(grp)))
})

test_that("kfold helpers throw correct errors", {
  expect_error(kfold_split_random(10), "!is.null(N) is not TRUE", fixed = TRUE)
  expect_error(kfold_split_random(10.5, 100), "K == as.integer(K) is not TRUE", fixed = TRUE)
  expect_error(kfold_split_random(10, 100.5), "N == as.integer(N) is not TRUE", fixed = TRUE)
  expect_error(kfold_split_random(K = c(1,1), N = 100), "length(K) == 1 is not TRUE", fixed = TRUE)
  expect_error(kfold_split_random(N = c(100, 100)), "length(N) == 1 is not TRUE", fixed = TRUE)
  expect_error(kfold_split_random(K = 5, N = 4), "K <= N is not TRUE", fixed = TRUE)
  expect_error(kfold_split_random(K = 1, N = 4), "K > 1 is not TRUE", fixed = TRUE)

  y <- sample(c(0, 1), size = 200, replace = TRUE, prob = c(0.05, 0.95))
  expect_error(kfold_split_stratified(10), "!is.null(x) is not TRUE", fixed = TRUE)
  expect_error(kfold_split_stratified(10.5, y), "K == as.integer(K) is not TRUE", fixed = TRUE)
  expect_error(kfold_split_stratified(K = c(1,1), y), "length(K) == 1 is not TRUE", fixed = TRUE)
  expect_error(kfold_split_stratified(K = 201, y), "K <= length(x) is not TRUE", fixed = TRUE)
  expect_error(kfold_split_stratified(K = 1, y), "K > 1 is not TRUE", fixed = TRUE)

  grp <- gl(n = 50, k = 15)
  expect_error(kfold_split_grouped(10), "!is.null(x) is not TRUE", fixed = TRUE)
  expect_error(kfold_split_grouped(3, c(1,1,1)), "'K' must not be bigger than the number of levels/groups in 'x'", fixed = TRUE)
  expect_error(kfold_split_grouped(10.5, grp), "K == as.integer(K) is not TRUE", fixed = TRUE)
  expect_error(kfold_split_grouped(K = c(1,1), grp), "length(K) == 1 is not TRUE", fixed = TRUE)
  expect_error(kfold_split_grouped(K = 1, grp), "K > 1 is not TRUE", fixed = TRUE)
})


test_that("print_dims.kfold works", {
  xx <- structure(list(), K = 17, class = c("kfold", "loo"))
  expect_output(print_dims(xx), "Based on 17-fold cross-validation")

  attr(xx, "K") <- NULL
  expect_silent(print_dims(xx))
})

#' Model comparison
#'
#' @description Compare fitted models based on [ELPD][loo-glossary].
#'
#' @export
#' @param x An object of class `"loo"` or a list of such objects. If a list is
#'   used then the list names will be used as the model names in the output. See
#'   **Examples**.
#' @param ... Additional objects of class `"loo"`, if not passed in as a single
#'   list.
#'
#' @return A data frame with class `"compare.loo"` that has its own
#'   print method. See the **Details** and **Examples** sections.
#'
#' @details
#'   When comparing two fitted models, we can estimate the difference in their
#'   expected predictive accuracy by the difference in
#'   [`elpd_loo`][loo-glossary] or `elpd_waic` (or multiplied by \eqn{-2}, if
#'   desired, to be on the deviance scale).
#'
#' ## `elpd_diff` and `se_diff`
#'   When using `loo_compare()`, the returned data frame will have one row per
#'   model and several columns of estimates. The values of
#'   [`elpd_diff`][loo-glossary] and [`se_diff`][loo-glossary] are computed by
#'   making pairwise comparisons between each model and the model with the
#'   largest ELPD (the model listed first). Therefore, the first `elpd_diff`
#'   value will always be `0` (i.e., the difference between the preferred model
#'   and itself) and the rest of the values will be negative.
#'
#'   To compute the standard error of the difference in [ELPD][loo-glossary] ---
#'   which should not be expected to equal the difference of the standard errors
#'   --- we use a paired estimate to take advantage of the fact that the same
#'   set of \eqn{N} data points was used to fit both models. These calculations
#'   should be most useful when \eqn{N} is large, because then non-normality of
#'   the distribution is not such an issue when estimating the uncertainty in
#'   these sums. These standard errors, for all their flaws, should give a
#'   better sense of uncertainty than what is obtained using the current
#'   standard approach of comparing differences of deviances to a Chi-squared
#'   distribution, a practice derived for Gaussian linear models or
#'   asymptotically, and which only applies to nested models in any case.
#'
#' ## `p_worse`, `diag_diff`, and `diag_elpd`
#'   The values in the `p_worse` column show the probability of each model
#'   having worse ELPD than the best model. These probabilities are computed
#'   with a normal approximation using the values from `elpd_diff` and
#'   `se_diff`. Sivula et al. (2025) present the conditions when the normal
#'   approximation used for SE and `se_diff` is good, and the column
#'   `diag_diff` contains possible diagnostic messages:
#'
#'   * `N < 100` (small data)
#'   * `|elpd_diff| < 4` (models make similar predictions)
#'
#'   If either of these diagnostic messages is shown, the error distribution is
#'   skewed or thick tailed and the normal approximation based on `elpd_diff`
#'   and `se_diff` is not well calibrated. In that case, the probabilities
#'   `p_worse` are likely to be too large. However, `elpd_diff` and `se_diff`
#'   will still be indicative of the differences and uncertainties (for example,
#'   if `|elpd_diff|` is many times larger than `se_diff` the difference is quite
#'   certain). In addition, if the model is not well specificed and there are
#'   outliers, the error distribution can also be skewed or thick tailed and the
#'   normal approximation is not well calibrated. Possible model misspecification
#'   and outliers can be diagnosed with usual predictive checking methods.
#'
#'   The column `diag_elpd` shows the PSIS-LOO Pareto k diagnostic for the
#'   pointwise ELPD computations for each model. If `K k_psis > 0.7` is shown,
#'   where `K` is the number of high Pareto k values in the PSIS
#'   computation, then there may be significant bias in `elpd_diff` favoring
#'   models with a large number of high Pareto k values.
#'
#' ## Warnings for many model comparisons
#'   If more than \eqn{11} models are compared, we internally recompute the model
#'   differences using the median model by ELPD as the baseline model. We then
#'   estimate whether the differences in predictive performance are potentially
#'   due to chance as described by McLatchie and Vehtari (2023). This will flag
#'   a warning if it is deemed that there is a risk of over-fitting due to the
#'   selection process. In that case users are recommended to avoid model
#'   selection based on LOO-CV, and instead to favor model averaging/stacking or
#'   projection predictive inference.
#'
#' @seealso
#' * The [FAQ page](https://mc-stan.org/loo/articles/online-only/faq.html) on
#'   the __loo__ website for answers to frequently asked questions.
#' @template loo-and-compare-references
#'
#' @examples
#' # very artificial example, just for demonstration!
#' LL <- example_loglik_array()
#' loo1 <- loo(LL)     # should be worst model when compared
#' loo2 <- loo(LL + 1) # should be second best model when compared
#' loo3 <- loo(LL + 2) # should be best model when compared
#'
#' comp <- loo_compare(loo1, loo2, loo3)
#' print(comp, digits = 2)
#'
#' # can use a list of objects with custom names
#' # the names will be used in the output
#' loo_compare(list("apple" = loo1, "banana" = loo2, "cherry" = loo3))
#'
#' \dontrun{
#' # works for waic (and kfold) too
#' loo_compare(waic(LL), waic(LL - 10))
#' }
#'
loo_compare <- function(x, ...) {
  UseMethod("loo_compare")
}

#' @rdname loo_compare
#' @export
loo_compare.default <- function(x, ...) {
  if (is.loo(x)) {
    dots <- list(...)
    loos <- c(list(x), dots)
  } else {
    if (!is.list(x) || !length(x)) {
      stop("'x' must be a list if not a 'loo' object.")
    }
    if (length(list(...))) {
      stop("If 'x' is a list then '...' should not be specified.")
    }
    loos <- x
  }

  # if subsampling is used
  if (any(sapply(loos, inherits, "psis_loo_ss"))) {
    return(loo_compare.psis_loo_ss_list(loos))
  }

  # run pre-comparison checks
  loo_compare_checks(loos)

  # compute elpd_diff and se_elpd_diff relative to best model
  comp <- loo_compare_matrix(loos)
  ord <- loo_compare_order(loos)
  rnms <- rownames(comp)
  diffs <- mapply(FUN = elpd_diffs, loos[ord[1]], loos[ord])
  colnames(diffs) <- rnms
  elpd_diff <- apply(diffs, 2, sum)
  se_diff <- apply(diffs, 2, se_elpd_diff)

  # compute probabilities that a model has worse elpd than the best model
  # using a normal approximation (Sivula et al., 2025)
  p_worse <- stats::pnorm(0, elpd_diff, se_diff)
  p_worse[elpd_diff == 0] <- NA

  comp <- cbind(
    data.frame(
      model = rnms,
      elpd_diff = elpd_diff,
      se_diff = se_diff,
      p_worse = p_worse,
      diag_diff = diag_diff(nrow(diffs), elpd_diff),
      diag_elpd = diag_elpd(loos[ord])
    ),
    as.data.frame(comp)
  )
  rownames(comp) <- NULL

  # run order statistics-based checks for many model comparisons
  loo_order_stat_check(loos, ord)

  class(comp) <- c("compare.loo", class(comp))
  comp
}

#' @rdname loo_compare
#' @export
#' @param digits For the print method only, the number of digits to use when
#'   printing.
#' @param p_worse For the print method only, should we include the normal
#'   approximation based probability of each model having worse performance than
#'   the best model? The default is `TRUE`.
print.compare.loo <- function(x, ..., digits = 1, p_worse = TRUE) {
  if (inherits(x, "old_compare.loo")) {
    return(unclass(x))
  }
  if (!inherits(x, "data.frame")) {
    class(x) <- c(class(x), "data.frame")
  }
  if (!all(c("model", "elpd_diff", "se_diff") %in% colnames(x))) {
    print(as.data.frame(x))
    return(x)
  }
  x2 <- cbind(
    model = x$model,
    .fr(x[, c("elpd_diff", "se_diff")], digits)
  )
  if (p_worse && "p_worse" %in% colnames(x)) {
    x2 <- cbind(
      x2,
      p_worse = .fr(x[, "p_worse"], digits = 2),
      diag_diff = x[, "diag_diff"],
      diag_elpd = x[, "diag_elpd"]
    )
  }
  print(x2, quote = FALSE, row.names = FALSE)

  # show glossary for diagnostic flags
  has_diag <- any(nzchar(x[["diag_diff"]], keepNA = FALSE), na.rm = TRUE) ||
              any(nzchar(x[["diag_elpd"]], keepNA = FALSE), na.rm = TRUE)
  if (has_diag && p_worse) {
    message(
      "\nDiagnostic flags present.\n",
      "See ?`loo-glossary` (sections `diag_diff` and `diag_elpd`)\n",
      "or https://mc-stan.org/loo/reference/loo-glossary.html."
    )
  }
  invisible(x)
}


# internal ----------------------------------------------------------------

#' Compute pointwise elpd differences
#' @noRd
#' @param loo_a,loo_b Two `"loo"` objects.
elpd_diffs <- function(loo_a, loo_b) {
  pt_a <- loo_a$pointwise
  pt_b <- loo_b$pointwise
  elpd <- grep("^elpd", colnames(pt_a))
  pt_b[, elpd] - pt_a[, elpd]
}

#' Compute standard error of the elpd difference
#' @noRd
#' @param diffs Vector of pointwise elpd differences
se_elpd_diff <- function(diffs) {
  N <- length(diffs)
  # As `elpd_diff` is defined as the sum of N independent components,
  # we can compute the standard error by using the standard deviation
  # of the N components and multiplying by `sqrt(N)`.
  sqrt(N) * sd(diffs)
}

#' Perform checks on `"loo"` objects before comparison
#' @noRd
#' @param loos List of `"loo"` objects.
#' @return Nothing, just possibly throws errors/warnings.
loo_compare_checks <- function(loos) {
  ## errors
  if (length(loos) <= 1L) {
    stop("'loo_compare' requires at least two models.", call.=FALSE)
  }
  if (!all(sapply(loos, is.loo))) {
    stop("All inputs should have class 'loo'.", call.=FALSE)
  }

  Ns <- vapply(loos, function(x) nrow(x$pointwise), integer(1))
  if (any(Ns != Ns[1L])) {
    stop(
      paste0(
        "All models must have the same number of observations, but models have inconsistent observation counts: ",
        paste(paste0("'", find_model_names(loos), "' (", Ns, ")"), collapse = ", ")
      ),
      call. = FALSE
    )
  }

  ## warnings

  yhash <- lapply(loos, attr, which = "yhash")
  yhash_ok <- sapply(yhash, function(x) { # ok only if all yhash are same (all NULL is ok)
    isTRUE(all.equal(x, yhash[[1]]))
  })
  if (!all(yhash_ok)) {
    warning("Not all models have the same y variable. ('yhash' attributes do not match)",
            call. = FALSE)
  }

  if (all(sapply(loos, is.kfold))) {
    Ks <- unlist(lapply(loos, attr, which = "K"))
    if (!all(Ks == Ks[1])) {
      warning("Not all kfold objects have the same K value. ",
              "For a more accurate comparison use the same number of folds. ",
              call. = FALSE)
    }
  } else if (any(sapply(loos, is.kfold)) && any(sapply(loos, is.psis_loo))) {
    warning("Comparing LOO-CV to K-fold-CV. ",
            "For a more accurate comparison use the same number of folds ",
            "or loo for all models compared.",
            call. = FALSE)
  }
}


#' Find the model names associated with `"loo"` objects
#'
#' @export
#' @param x List of `"loo"` objects.
#' @return Character vector of model names the same length as `x.`
#'
find_model_names <- function(x) {
  stopifnot(is.list(x))
  out_names <- character(length(x))

  names1 <- names(x)
  names2 <- lapply(x, "attr", "model_name", exact = TRUE)
  names3 <- lapply(x, "[[", "model_name")
  names4 <- paste0("model", seq_along(x))

  for (j in seq_along(x)) {
    if (isTRUE(nzchar(names1[j]))) {
      out_names[j] <- names1[j]
    } else if (length(names2[[j]])) {
      out_names[j] <- names2[[j]]
    } else if (length(names3[[j]])) {
      out_names[j] <- names3[[j]]
    } else {
      out_names[j] <- names4[j]
    }
  }
  out_names
}


#' Compute the loo_compare matrix
#' @noRd
#' @param loos List of `"loo"` objects.
loo_compare_matrix <- function(loos){
  tmp <- sapply(loos, function(x) {
    est <- x$estimates
    setNames(c(est), nm = c(rownames(est), paste0("se_", rownames(est))))
  })
  colnames(tmp) <- find_model_names(loos)
  rnms <- rownames(tmp)
  comp <- tmp
  ord <- loo_compare_order(loos)
  comp <- t(comp)[ord, ]
  patts <- c("elpd", "p_", "^waic$|^looic$", "^se_waic$|^se_looic$")
  col_ord <- unlist(sapply(patts, function(p) grep(p, colnames(comp))),
                    use.names = FALSE)
  comp <- comp[, col_ord]
  comp
}

#' Computes the order of loos for comparison
#' @noRd
#' @param loos List of `"loo"` objects.
loo_compare_order <- function(loos){
  tmp <- sapply(loos, function(x) {
    est <- x$estimates
    setNames(c(est), nm = c(rownames(est), paste0("se_", rownames(est))))
  })
  colnames(tmp) <- find_model_names(loos)
  rnms <- rownames(tmp)
  ord <- order(tmp[grep("^elpd", rnms), ], decreasing = TRUE)
  ord
}

#' Perform checks on `"loo"` objects __after__ comparison
#' @noRd
#' @param loos List of `"loo"` objects.
#' @param ord List of `"loo"` object orderings.
#' @return Nothing, just possibly throws errors/warnings.
loo_order_stat_check <- function(loos, ord) {

  ## breaks

  if (length(loos) <= 11L) {
    # procedure cannot be diagnosed for fewer than ten candidate models
    # (total models = worst model + ten candidates)
    # break from function
    return(NULL)
  }

  ## warnings

  # compute the elpd differences from the median model
  baseline_idx <- middle_idx(ord)
  diffs <- mapply(FUN = elpd_diffs, loos[ord[baseline_idx]], loos[ord])
  elpd_diff <- apply(diffs, 2, sum)

  # estimate the standard deviation of the upper-half-normal
  diff_median <- stats::median(elpd_diff)
  elpd_diff_trunc <- elpd_diff[elpd_diff >= diff_median]
  n_models <- sum(!is.na(elpd_diff_trunc))
  candidate_sd <- sqrt(1 / n_models * sum(elpd_diff_trunc^2, na.rm = TRUE))

  # estimate expected best diff under null hypothesis
  K <- length(loos) - 1
  order_stat <- order_stat_heuristic(K, candidate_sd)

  if (max(elpd_diff) <= order_stat) {
    # flag warning if we suspect no model is theoretically better than the baseline
    warning("Difference in performance potentially due to chance. ",
            "See McLatchie and Vehtari (2023) for details.",
            call. = FALSE)
  }
}

#' Returns the middle index of a vector
#' @noRd
#' @param vec A vector.
#' @return Integer index value.
middle_idx <- function(vec) floor(length(vec) / 2)

#' Computes maximum order statistic from K Gaussians
#' @noRd
#' @param K Number of Gaussians.
#' @param c Scaling of the order statistic.
#' @return Numeric expected maximum from K samples from a Gaussian with mean
#' zero and scale `"c"`
order_stat_heuristic <- function(K, c) {
  qnorm(p = 1 - 1 / (K * 2), mean = 0, sd = c)
}

#' Count number of high Pareto k values in PSIS-LOO and create diagnostic message
#' @noRd
#' @param loos Ordered list of loo objects.
#' @return Character vector of diagnostic messages.
diag_elpd <- function(loos) {
  sapply(loos, function(loo) {
    k <- loo$diagnostics[["pareto_k"]]
    if (is.null(k)) {
      out <- ""
    } else {
      S <- dim(loo)[1]
      khat_threshold <- ps_khat_threshold(S)
      K <- sum(k > khat_threshold)
      out <- ifelse(K == 0, "", paste0(K, " k_psis > ", round(khat_threshold, 2)))
    }
    out
  })
}

#' Create diagnostic for elpd differences
#' @noRd
#' @param N Number of data points.
#' @param elpd_diff Vector of elpd differences.
#' @return Character vector of diagnostic messages.
diag_diff <- function(N, elpd_diff) {
  if (N < 100) {
    diag_diff <- rep("N < 100", length(elpd_diff))
    diag_diff[elpd_diff == 0] <- ""
  } else {
    diag_diff <- rep("", length(elpd_diff))
    diag_diff[elpd_diff > -4 & elpd_diff != 0] <- "|elpd_diff| < 4"
  }
  diag_diff
}

# loo_moment_match.default warnings work

    Code
      loo_moment_match(x, loo_manual, post_draws_test, log_lik_i_test,
        unconstrain_pars_test, log_prob_upars_test, log_lik_i_upars_test, max_iters = 1,
        k_thres = 0.5, split = TRUE, cov = TRUE, cores = 1)
    Condition
      Warning:
      The maximum number of moment matching iterations ('max_iters' argument)
          was reached.
      Increasing the value may improve accuracy.
      Warning:
      Some Pareto k diagnostic values are too high. See help('pareto-k-diagnostic') for details.
    Output
      
      Computed from 4000 by 30 log-likelihood matrix.
      
               Estimate   SE
      elpd_loo    -74.0 18.8
      p_loo        11.7 11.1
      looic       148.0 37.6
      ------
      MCSE of elpd_loo is NA.
      MCSE and ESS estimates assume independent draws (r_eff=1).
      
      Pareto k diagnostic values:
                               Count Pct.    Min. ESS
      (-Inf, 0.7]   (good)     29    96.7%   3758    
         (0.7, 1]   (bad)       1     3.3%   <NA>    
         (1, Inf)   (very bad)  0     0.0%   <NA>    
      See help('pareto-k-diagnostic') for details.

# loo_moment_match.default works

    WAoAAAACAAQFAAACAwAAAAMTAAAACgAAAg4AAAAGwFJQKLlLGUdAJeeelRE5/EBiUCi5SxlH
    QDIH7c01qv5AJJ/C+xNRK0BCB+3NNar+AAAEAgAAAAEABAAJAAAAA2RpbQAAAA0AAAACAAAA
    AwAAAAIAAAQCAAAAAQAEAAkAAAAIZGltbmFtZXMAAAATAAAAAgAAABAAAAADAAQACQAAAAhl
    bHBkX2xvbwAEAAkAAAAFcF9sb28ABAAJAAAABWxvb2ljAAAAEAAAAAIABAAJAAAACEVzdGlt
    YXRlAAQACQAAAAJTRQAAAP4AAAIOAAAAlsAz2SCKL/OPv/uoQkpWQUi//t6gM8RRaL/7ME6e
    7mn4v/slBOxuIMC//9jJ8p3XcL/7L6Fev0U4wAA2/FFJKpi//ScFgWbouL/8Pocg7R74v/0p
    qq1wTGC/+x8Kh2qjyL/7JFoKTeIwv/soK3EUBUC//KBEQBPCGMAAKtiBefwUv/s4omTXOVjA
    A0qKEEuTHL/7hlp0sBHYv/xVS8d60Ki//yJy272SKL/7oVoD6+JIv/7iRaQt0cC//VaEo30D
    QL/85OxxwadowAHqd1LUWNy/+9Uw/+PdqL/7IYmDCMJQv/+TLUcS2KC//UsYUnCS0D/ADGFe
    xkQGP2Gfcpi2J8Y/Y+tB1l78/T9hmK3v3QgHP2GaKR7p1QA/ZQkD2NQUNz9hp3Fk+oG7P2Wd
    0KgLpYI/Yjujzb+UVT9hx0KNqoPZP2J1vTdzw+s/YaMJLQv8lD9hpM5D/++/P2GZy1TTq7Y/
    YfB+8HlL/j9lqO+CzesuP2GpxXMbu9A/cGE1wkkFgT9hvBsU728qP2HQSdm5/Yo/Y/zERK8U
    MT9hni0KCSNZP2O2kFJN6R4/Ylo+R+/P/j9iFB/WgCCUP2s9ryP5flA/YdKJGMivnD9hmtGw
    ba2EP2SA235CBJ8/YotQDpXP+EAkqqbkVm+LP5IXf6UkPQA/lrxpW1EOAD+SDmssqA4AP5IS
    8YLzxAA/mQEBdV99AD+SOdWCpaAAP5otz1Wr6wA/k0U02BITAD+SYx8kcE4AP5PhNpO18QA/
    ki0P3eyAAD+SMjlDn5oAP5IR1Ro8TgA/krLjpKmXAD+aTlBMoagAP5JAN/f7OgA/qwiqhKr0
    AD+SbOl4OJEAP5J03K3aTgA/lsJL+aFkAD+SFPC7p4gAP5Y04S7O5wA/k4CJvuH6AD+S+DEh
    3hkAP6Omxr93f4A/kp2i9JwgAD+SFOjEYZoAP5fSMVoWZAA/lAsiPEe5AEBD2SCKL/OPQAuo
    QkpWQUhADt6gM8RRaEALME6e7mn4QAslBOxuIMBAD9jJ8p3XcEALL6Fev0U4QBA2/FFJKphA
    DScFgWbouEAMPocg7R74QA0pqq1wTGBACx8Kh2qjyEALJFoKTeIwQAsoK3EUBUBADKBEQBPC
    GEAQKtiBefwUQAs4omTXOVhAE0qKEEuTHEALhlp0sBHYQAxVS8d60KhADyJy272SKEALoVoD
    6+JIQA7iRaQt0cBADVaEo30DQEAM5OxxwadoQBHqd1LUWNxAC9Uw/+PdqEALIYmDCMJQQA+T
    LUcS2KBADUsYUnCS0D/2XcffnmqVv5oq/ZSIVXI/rc6o9F9nBr9RT6Q+S52Qv3AJF5hupNw/
    teKv62UxfL+O7vQo7zgIP8HB+bh4VvU/p+nRG8MPFz+j+bKjZAlIv6dNik1Iv/W/pGhy2+ci
    f7+iYyxrlH4Ov4shzHMAyAw/r8r08fJQBz+9B+WoAPW6v6EcmkpM9v8/sa5n3E19AL+gl0c6
    bbNbP5Z/I4ovlu4/xB68LmfXyr+KgBHuAEaAP7sysVWUlHQ/tAdcziRrAj+tf7e/7a72P6bM
    edRbec6/pOE4IxDMN79Twd0WCzSwP8WPgi1BERa/kp3azvQypgAABAIAAAH/AAAADQAAAAIA
    AAAeAAAABQAABAIAAAL/AAAAEwAAAAIAAAD+AAAAEAAAAAUABAAJAAAACGVscGRfbG9vAAQA
    CQAAAA1tY3NlX2VscGRfbG9vAAQACQAAAAVwX2xvbwAEAAkAAAAFbG9vaWMABAAJAAAAEmlu
    Zmx1ZW5jZV9wYXJldG9fawAAAP4AAAITAAAAAwAAAA4AAAAeP+fdHDQ/QUa/mir9lIhVcj+t
    zqj0X2cGv1FPpD5LnZC/cAkXmG6k3D+14q/rZTF8v47u9CjvOAg/wcH5uHhW9T+n6dEbww8X
    P6P5sqNkCUi/p02KTUi/9b+kaHLb5yJ/v6JjLGuUfg6/iyHMcwDIDD+vyvTx8lAHP70H5agA
    9bq/oRyaSkz2/z+xrmfcTX0Av6CXRzpts1s/ln8jii+W7j/EHrwuZ9fKv4qAEe4ARoA/uzKx
    VZSUdD+0B1zOJGsCP61/t7/trvY/psx51Ft5zr+k4TgjEMw3v1PB3RYLNLA/xY+CLUERFr+S
    ndrO9DKmAAAADgAAAB5ARQcvr5nxRUCurmCYTUMgQK6GZIiNpldArq7KVLrefECurq6LsT47
    QK5xjnyab6BArq3SHuGPKUCuZrTzmGSJQK6kDS9Qm5FArqvKHptSZ0CuoDIHBkg3QK6uB1vl
    FFtArq3uNy14MkCurq95Y/PBQK6pD98CEa1ArmWF3fh0P0CuraLKskjiQK1cwbtJXeNArqxg
    HKDKmkCuqx85Cm9rQK6FgdGWcINArq598JLLRECuil5ipQqpQK6iHFktv3tArqazEBI7tkCt
    65ndUr6PQK6q4cDxAOlArq6k0f++pUCufAQQaRkWQK6e1Q/LzlsAAAAOAAAAHj/wAAAAAAAA
    P/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAA
    AAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAA
    AAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/w
    AAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAA
    P/AAAAAAAAA/8AAAAAAAAAAABAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAADAAQACQAAAAhw
    YXJldG9fawAEAAkAAAAFbl9lZmYABAAJAAAABXJfZWZmAAAA/gAAAP4AAAAOAAAAAcBSUCi5
    SxlHAAAADgAAAAFAJeeelRE5/AAAAA4AAAABQGJQKLlLGUcAAAAOAAAAAUAyB+3NNar+AAAA
    DgAAAAFAJJ/C+xNRKwAAAA4AAAABQEIH7c01qv4AAAQCAAAD/wAAABAAAAAKAAQACQAAAAll
    c3RpbWF0ZXMABAAJAAAACXBvaW50d2lzZQAEAAkAAAALZGlhZ25vc3RpY3MABAAJAAAAC3Bz
    aXNfb2JqZWN0AAQACQAAAAhlbHBkX2xvbwAEAAkAAAAFcF9sb28ABAAJAAAABWxvb2ljAAQA
    CQAAAAtzZV9lbHBkX2xvbwAEAAkAAAAIc2VfcF9sb28ABAAJAAAACHNlX2xvb2ljAAAEAgAA
    AAEABAAJAAAABGRpbXMAAAANAAAAAgAAD6AAAAAeAAAEAgAAAAEABAAJAAAABWNsYXNzAAAA
    EAAAAAMABAAJAAAACHBzaXNfbG9vAAQACQAAABdpbXBvcnRhbmNlX3NhbXBsaW5nX2xvbwAE
    AAkAAAADbG9vAAAA/g==

---

    WAoAAAACAAQFAAACAwAAAAMTAAAACgAAAg4AAAAGwFJ5ESZa2ndAJy7h/Y9DekBieREmWtp3
    QDKrWN/9QglAJecFy/N4C0BCq1jf/UIJAAAEAgAAAAEABAAJAAAAA2RpbQAAAA0AAAACAAAA
    AwAAAAIAAAQCAAAAAQAEAAkAAAAIZGltbmFtZXMAAAATAAAAAgAAABAAAAADAAQACQAAAAhl
    bHBkX2xvbwAEAAkAAAAFcF9sb28ABAAJAAAABWxvb2ljAAAAEAAAAAIABAAJAAAACEVzdGlt
    YXRlAAQACQAAAAJTRQAAAP4AAAIOAAAAlsA0fMI+bvhOv/uoQkpWQUi//t6gM8RRaL/7ME6e
    7mn4v/slBOxuIMC//9jJ8p3XcL/7L6Fev0U4wAA2/FFJKpi//ScFgWbouL/8Pocg7R74v/0p
    qq1wTGC/+x8Kh2qjyL/7JFoKTeIwv/soK3EUBUC//KBEQBPCGMAAKtiBefwUv/s4omTXOVjA
    A0qKEEuTHL/7hlp0sBHYv/xVS8d60Ki//yJy272SKL/7oVoD6+JIv/7iRaQt0cC//VaEo30D
    QL/85OxxwadowAHqd1LUWNy/+9Uw/+PdqL/7IYmDCMJQv/+TLUcS2KC//UsYUnCS0D+TxTjn
    PtpbP2Gfcpi2J8Y/Y+tB1l78/T9hmK3v3QgHP2GaKR7p1QA/ZQkD2NQUNz9hp3Fk+oG7P2Wd
    0KgLpYI/Yjujzb+UVT9hx0KNqoPZP2J1vTdzw+s/YaMJLQv8lD9hpM5D/++/P2GZy1TTq7Y/
    YfB+8HlL/j9lqO+CzesuP2GpxXMbu9A/cGE1wkkFgT9hvBsU728qP2HQSdm5/Yo/Y/zERK8U
    MT9hni0KCSNZP2O2kFJN6R4/Ylo+R+/P/j9iFB/WgCCUP2s9ryP5flA/YdKJGMivnD9hmtGw
    ba2EP2SA235CBJ8/YotQDpXP+EAl8epM1HkJP5IXf6UkPQA/lrxpW1EOAD+SDmssqA4AP5IS
    8YLzxAA/mQEBdV99AD+SOdWCpaAAP5otz1Wr6wA/k0U02BITAD+SYx8kcE4AP5PhNpO18QA/
    ki0P3eyAAD+SMjlDn5oAP5IR1Ro8TgA/krLjpKmXAD+aTlBMoagAP5JAN/f7OgA/qwiqhKr0
    AD+SbOl4OJEAP5J03K3aTgA/lsJL+aFkAD+SFPC7p4gAP5Y04S7O5wA/k4CJvuH6AD+S+DEh
    3hkAP6Omxr93f4A/kp2i9JwgAD+SFOjEYZoAP5fSMVoWZAA/lAsiPEe5AEBEfMI+bvhOQAuo
    QkpWQUhADt6gM8RRaEALME6e7mn4QAslBOxuIMBAD9jJ8p3XcEALL6Fev0U4QBA2/FFJKphA
    DScFgWbouEAMPocg7R74QA0pqq1wTGBACx8Kh2qjyEALJFoKTeIwQAsoK3EUBUBADKBEQBPC
    GEAQKtiBefwUQAs4omTXOVhAE0qKEEuTHEALhlp0sBHYQAxVS8d60KhADyJy272SKEALoVoD
    6+JIQA7iRaQt0cBADVaEo30DQEAM5OxxwadoQBHqd1LUWNxAC9Uw/+PdqEALIYmDCMJQQA+T
    LUcS2KBADUsYUnCS0D/2XcffnmqVv5oq/ZSIVXI/rc6o9F9nBr9RT6Q+S52Qv3AJF5hupNw/
    teKv62UxfL+O7vQo7zgIP8HB+bh4VvU/p+nRG8MPFz+j+bKjZAlIv6dNik1Iv/W/pGhy2+ci
    f7+iYyxrlH4Ov4shzHMAyAw/r8r08fJQBz+9B+WoAPW6v6EcmkpM9v8/sa5n3E19AL+gl0c6
    bbNbP5Z/I4ovlu4/xB68LmfXyr+KgBHuAEaAP7sysVWUlHQ/tAdcziRrAj+tf7e/7a72P6bM
    edRbec6/pOE4IxDMN79Twd0WCzSwP8WPgi1BERa/kp3azvQypgAABAIAAAH/AAAADQAAAAIA
    AAAeAAAABQAABAIAAAL/AAAAEwAAAAIAAAD+AAAAEAAAAAUABAAJAAAACGVscGRfbG9vAAQA
    CQAAAA1tY3NlX2VscGRfbG9vAAQACQAAAAVwX2xvbwAEAAkAAAAFbG9vaWMABAAJAAAAEmlu
    Zmx1ZW5jZV9wYXJldG9fawAAAP4AAAITAAAAAwAAAA4AAAAeP7Z3M/Z+3oW/mir9lIhVcj+t
    zqj0X2cGv1FPpD5LnZC/cAkXmG6k3D+14q/rZTF8v47u9CjvOAg/wcH5uHhW9T+n6dEbww8X
    P6P5sqNkCUi/p02KTUi/9b+kaHLb5yJ/v6JjLGuUfg6/iyHMcwDIDD+vyvTx8lAHP70H5agA
    9bq/oRyaSkz2/z+xrmfcTX0Av6CXRzpts1s/ln8jii+W7j/EHrwuZ9fKv4qAEe4ARoA/uzKx
    VZSUdD+0B1zOJGsCP61/t7/trvY/psx51Ft5zr+k4TgjEMw3v1PB3RYLNLA/xY+CLUERFr+S
    ndrO9DKmAAAADgAAAB5AFp345qujjUCurmCYTUMgQK6GZIiNpldArq7KVLrefECurq6LsT47
    QK5xjnyab6BArq3SHuGPKUCuZrTzmGSJQK6kDS9Qm5FArqvKHptSZ0CuoDIHBkg3QK6uB1vl
    FFtArq3uNy14MkCurq95Y/PBQK6pD98CEa1ArmWF3fh0P0CuraLKskjiQK1cwbtJXeNArqxg
    HKDKmkCuqx85Cm9rQK6FgdGWcINArq598JLLRECuil5ipQqpQK6iHFktv3tArqazEBI7tkCt
    65ndUr6PQK6q4cDxAOlArq6k0f++pUCufAQQaRkWQK6e1Q/LzlsAAAAOAAAAHj/wAAAAAAAA
    P/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAA
    AAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAA
    AAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/w
    AAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAA
    P/AAAAAAAAA/8AAAAAAAAAAABAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAADAAQACQAAAAhw
    YXJldG9fawAEAAkAAAAFbl9lZmYABAAJAAAABXJfZWZmAAAA/gAAAP4AAAAOAAAAAcBSeREm
    Wtp3AAAADgAAAAFAJy7h/Y9DegAAAA4AAAABQGJ5ESZa2ncAAAAOAAAAAUAyq1jf/UIJAAAA
    DgAAAAFAJecFy/N4CwAAAA4AAAABQEKrWN/9QgkAAAQCAAAD/wAAABAAAAAKAAQACQAAAAll
    c3RpbWF0ZXMABAAJAAAACXBvaW50d2lzZQAEAAkAAAALZGlhZ25vc3RpY3MABAAJAAAAC3Bz
    aXNfb2JqZWN0AAQACQAAAAhlbHBkX2xvbwAEAAkAAAAFcF9sb28ABAAJAAAABWxvb2ljAAQA
    CQAAAAtzZV9lbHBkX2xvbwAEAAkAAAAIc2VfcF9sb28ABAAJAAAACHNlX2xvb2ljAAAEAgAA
    AAEABAAJAAAABGRpbXMAAAANAAAAAgAAD6AAAAAeAAAEAgAAAAEABAAJAAAABWNsYXNzAAAA
    EAAAAAMABAAJAAAACHBzaXNfbG9vAAQACQAAABdpbXBvcnRhbmNlX3NhbXBsaW5nX2xvbwAE
    AAkAAAADbG9vAAAA/g==

---

    WAoAAAACAAQFAAACAwAAAAMTAAAACgAAAg4AAAAGwFKt/dMmDJVAKNZHY+jUZEBirf3TJgyV
    QDN+yjVXeO1AJ45qhpBmBkBDfso1V3jtAAAEAgAAAAEABAAJAAAAA2RpbQAAAA0AAAACAAAA
    AwAAAAIAAAQCAAAAAQAEAAkAAAAIZGltbmFtZXMAAAATAAAAAgAAABAAAAADAAQACQAAAAhl
    bHBkX2xvbwAEAAkAAAAFcF9sb28ABAAJAAAABWxvb2ljAAAAEAAAAAIABAAJAAAACEVzdGlt
    YXRlAAQACQAAAAJTRQAAAP4AAAIOAAAAlsA1UHTxm8DDv/uoQkpWQUi//t6gM8RRaL/7ME6e
    7mn4v/slBOxuIMC//9jJ8p3XcL/7L6Fev0U4wAA2/FFJKpi//ScFgWbouL/8Pocg7R74v/0p
    qq1wTGC/+x8Kh2qjyL/7JFoKTeIwv/soK3EUBUC//KBEQBPCGMAAKtiBefwUv/s4omTXOVjA
    A0qKEEuTHL/7hlp0sBHYv/xVS8d60Ki//yJy272SKL/7oVoD6+JIv/7iRaQt0cC//VaEo30D
    QL/85OxxwadowAHqd1LUWNy/+9Uw/+PdqL/7IYmDCMJQv/+TLUcS2KC//UsYUnCS0D+cqILP
    qzBKP2Gfcpi2J8Y/Y+tB1l78/T9hmK3v3QgHP2GaKR7p1QA/ZQkD2NQUNz9hp3Fk+oG7P2Wd
    0KgLpYI/Yjujzb+UVT9hx0KNqoPZP2J1vTdzw+s/YaMJLQv8lD9hpM5D/++/P2GZy1TTq7Y/
    YfB+8HlL/j9lqO+CzesuP2GpxXMbu9A/cGE1wkkFgT9hvBsU728qP2HQSdm5/Yo/Y/zERK8U
    MT9hni0KCSNZP2O2kFJN6R4/Ylo+R+/P/j9iFB/WgCCUP2s9ryP5flA/YdKJGMivnD9hmtGw
    ba2EP2SA235CBJ8/YotQDpXP+EAnmU+zLgnzP5IXf6UkPQA/lrxpW1EOAD+SDmssqA4AP5IS
    8YLzxAA/mQEBdV99AD+SOdWCpaAAP5otz1Wr6wA/k0U02BITAD+SYx8kcE4AP5PhNpO18QA/
    ki0P3eyAAD+SMjlDn5oAP5IR1Ro8TgA/krLjpKmXAD+aTlBMoagAP5JAN/f7OgA/qwiqhKr0
    AD+SbOl4OJEAP5J03K3aTgA/lsJL+aFkAD+SFPC7p4gAP5Y04S7O5wA/k4CJvuH6AD+S+DEh
    3hkAP6Omxr93f4A/kp2i9JwgAD+SFOjEYZoAP5fSMVoWZAA/lAsiPEe5AEBFUHTxm8DDQAuo
    QkpWQUhADt6gM8RRaEALME6e7mn4QAslBOxuIMBAD9jJ8p3XcEALL6Fev0U4QBA2/FFJKphA
    DScFgWbouEAMPocg7R74QA0pqq1wTGBACx8Kh2qjyEALJFoKTeIwQAsoK3EUBUBADKBEQBPC
    GEAQKtiBefwUQAs4omTXOVhAE0qKEEuTHEALhlp0sBHYQAxVS8d60KhADyJy272SKEALoVoD
    6+JIQA7iRaQt0cBADVaEo30DQEAM5OxxwadoQBHqd1LUWNxAC9Uw/+PdqEALIYmDCMJQQA+T
    LUcS2KBADUsYUnCS0D/2XcffnmqVv5oq/ZSIVXI/rc6o9F9nBr9RT6Q+S52Qv3AJF5hupNw/
    teKv62UxfL+O7vQo7zgIP8HB+bh4VvU/p+nRG8MPFz+j+bKjZAlIv6dNik1Iv/W/pGhy2+ci
    f7+iYyxrlH4Ov4shzHMAyAw/r8r08fJQBz+9B+WoAPW6v6EcmkpM9v8/sa5n3E19AL+gl0c6
    bbNbP5Z/I4ovlu4/xB68LmfXyr+KgBHuAEaAP7sysVWUlHQ/tAdcziRrAj+tf7e/7a72P6bM
    edRbec6/pOE4IxDMN79Twd0WCzSwP8WPgi1BERa/kp3azvQypgAABAIAAAH/AAAADQAAAAIA
    AAAeAAAABQAABAIAAAL/AAAAEwAAAAIAAAD+AAAAEAAAAAUABAAJAAAACGVscGRfbG9vAAQA
    CQAAAA1tY3NlX2VscGRfbG9vAAQACQAAAAVwX2xvbwAEAAkAAAAFbG9vaWMABAAJAAAAEmlu
    Zmx1ZW5jZV9wYXJldG9fawAAAP4AAAITAAAAAwAAAA4AAAAeP7Z3M/Z+3oW/mir9lIhVcj+t
    zqj0X2cGv1FPpD5LnZC/cAkXmG6k3D+14q/rZTF8v47u9CjvOAg/wcH5uHhW9T+n6dEbww8X
    P6P5sqNkCUi/p02KTUi/9b+kaHLb5yJ/v6JjLGuUfg6/iyHMcwDIDD+vyvTx8lAHP70H5agA
    9bq/oRyaSkz2/z+xrmfcTX0Av6CXRzpts1s/ln8jii+W7j/EHrwuZ9fKv4qAEe4ARoA/uzKx
    VZSUdD+0B1zOJGsCP61/t7/trvY/psx51Ft5zr+k4TgjEMw3v1PB3RYLNLA/xY+CLUERFr+S
    ndrO9DKmAAAADgAAAB5Ak/Dwg4seSECurmCYTUMgQK6GZIiNpldArq7KVLrefECurq6LsT47
    QK5xjnyab6BArq3SHuGPKUCuZrTzmGSJQK6kDS9Qm5FArqvKHptSZ0CuoDIHBkg3QK6uB1vl
    FFtArq3uNy14MkCurq95Y/PBQK6pD98CEa1ArmWF3fh0P0CuraLKskjiQK1cwbtJXeNArqxg
    HKDKmkCuqx85Cm9rQK6FgdGWcINArq598JLLRECuil5ipQqpQK6iHFktv3tArqazEBI7tkCt
    65ndUr6PQK6q4cDxAOlArq6k0f++pUCufAQQaRkWQK6e1Q/LzlsAAAAOAAAAHj/wAAAAAAAA
    P/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAA
    AAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAA
    AAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/w
    AAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAA
    P/AAAAAAAAA/8AAAAAAAAAAABAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAADAAQACQAAAAhw
    YXJldG9fawAEAAkAAAAFbl9lZmYABAAJAAAABXJfZWZmAAAA/gAAAP4AAAAOAAAAAcBSrf3T
    JgyVAAAADgAAAAFAKNZHY+jUZAAAAA4AAAABQGKt/dMmDJUAAAAOAAAAAUAzfso1V3jtAAAA
    DgAAAAFAJ45qhpBmBgAAAA4AAAABQEN+yjVXeO0AAAQCAAAD/wAAABAAAAAKAAQACQAAAAll
    c3RpbWF0ZXMABAAJAAAACXBvaW50d2lzZQAEAAkAAAALZGlhZ25vc3RpY3MABAAJAAAAC3Bz
    aXNfb2JqZWN0AAQACQAAAAhlbHBkX2xvbwAEAAkAAAAFcF9sb28ABAAJAAAABWxvb2ljAAQA
    CQAAAAtzZV9lbHBkX2xvbwAEAAkAAAAIc2VfcF9sb28ABAAJAAAACHNlX2xvb2ljAAAEAgAA
    AAEABAAJAAAABGRpbXMAAAANAAAAAgAAD6AAAAAeAAAEAgAAAAEABAAJAAAABWNsYXNzAAAA
    EAAAAAMABAAJAAAACHBzaXNfbG9vAAQACQAAABdpbXBvcnRhbmNlX3NhbXBsaW5nX2xvbwAE
    AAkAAAADbG9vAAAA/g==

# variance and covariance transformations work

    WAoAAAACAAQFAAACAwAAAAMTAAAACgAAAg4AAAAGwFKLcIr2U6lAJ7mJXVQ1mEBii3CK9lOp
    QDMBFEV6xQ1AJnj9Bj3oV0BDARRFesUNAAAEAgAAAAEABAAJAAAAA2RpbQAAAA0AAAACAAAA
    AwAAAAIAAAQCAAAAAQAEAAkAAAAIZGltbmFtZXMAAAATAAAAAgAAABAAAAADAAQACQAAAAhl
    bHBkX2xvbwAEAAkAAAAFcF9sb28ABAAJAAAABWxvb2ljAAAAEAAAAAIABAAJAAAACEVzdGlt
    YXRlAAQACQAAAAJTRQAAAP4AAAIOAAAAlsA00hUKopC/v/uiIMLHLuy//ssyhve2vL/7Jr8K
    SbUAv/sa1cItPhC//8POqw85FL/7JoUvY15YwAA7D8sD/c6//SUayysMjL/8O2Quh638v/0W
    mUgAlIC/+xKZXXIQsL/7F40nuPLAv/seLsVhXcC//J23rmyzmMAAIARDWDzAv/srA2aD7MjA
    A1R95vJwGL/7djZ68aYYv/xSR/sGCki//yevfroaBL/7mxSAeStQv/7m3aUwM2i//VTP19jb
    WL/84rONukOUwAHx9fU/2Fa/+8QvunAEmL/7Fx4TRHXUv/+ZdSn1rdS//Tf3NN2ngD+W4GjR
    aryTP5P9k5w+sG4/ayNFdQtzdT+UNR65X0TEP5QzA295mvs/bFTXaJrIOz+U3CtVdiGyP5RR
    Ot5s8NI/aT33KWiSfT9pDE2N/1sTP5S2SpsPGnU/lHsSPZvdgj+UhiM2HUpuP5QsiYrA7e4/
    aRVFdDks8D9tDHOUss4tP5Sd+c05JPQ/lYiPg3JzFD+UxOlt8wDqP2kMYnIOue0/lBcHGRWx
    ez+T/tJ6LYPNP5QNq0mErcY/aVODDQzbTz9pJSlA+yZNP5UHY/IZmJA/lNKlqN7G3T+UO0Qc
    JW5GP5QpKO2aIqQ/lK/I/7EmaEAmg6Qr+gHKP5I5hg6AAwA/lTqBKHTFAD+SIcNl86MAP5If
    omEDgQA/lvi9WmUhAD+THZI/2f4AP5mPg19ObgA/ktItKsZkAD+SjsiR3esAP5KkmoaHBgA/
    khZKaRj3AD+SFGHT2/EAP5Ig3ZVNGAA/kpwrkaAHAD+YCFgsLCIAP5IPS9yzsgA/qlFcWiEQ
    gD+RznOE6nMAP5KPYOKjMwA/lmRIfQIhAD+SN2h/dggAP5Xg9zNtrgA/ku5c/6I+AD+SsXhR
    Iu0AP6MUf55iiAA/kd0sR7uFAD+SH4L0gRUAP5dfNWv4+gA/kshKi4uJAEBE0hUKopC/QAui
    IMLHLuxADssyhve2vEALJr8KSbUAQAsa1cItPhBAD8POqw85FEALJoUvY15YQBA7D8sD/c5A
    DSUayysMjEAMO2Quh638QA0WmUgAlIBACxKZXXIQsEALF40nuPLAQAseLsVhXcBADJ23rmyz
    mEAQIARDWDzAQAsrA2aD7MhAE1R95vJwGEALdjZ68aYYQAxSR/sGCkhADyevfroaBEALmxSA
    eStQQA7m3aUwM2hADVTP19jbWEAM4rONukOUQBHx9fU/2FZAC8QvunAEmEALFx4TRHXUQA+Z
    dSn1rdRADTf3NN2ngD/5Fjgy7IjZP7bHI1nATQa/p51R7PmwGD+sHH7uZvEqP6/qoVFau3K/
    qA5Hs2wKJj+VYzdWUCsiP7mzNB+LFFy/vdDQDBpFx7+VFeEkxRxeP7F7m5KVODU/pVACJbjF
    Jj+jdMHsaW1mP65HVhP0fJm/s2b06ISqEb+kQJCWKzfqP6COIa9BZSQ/zvjmbLoMiD9TIc+d
    D8Ugv5o1IjVvHMo/s1oN2T4FLD+3OGyWxNwrP7EIIW+h0fm/uyMfiLLHcb+0y/cCrI9DP8sJ
    wPl6FQA/o6KWFvbG6j+0vX2EvGQYP7O2T6sntQo/qCNltsfvAgAABAIAAAH/AAAADQAAAAIA
    AAAeAAAABQAABAIAAAL/AAAAEwAAAAIAAAD+AAAAEAAAAAUABAAJAAAACGVscGRfbG9vAAQA
    CQAAAA1tY3NlX2VscGRfbG9vAAQACQAAAAVwX2xvbwAEAAkAAAAFbG9vaWMABAAJAAAAEmlu
    Zmx1ZW5jZV9wYXJldG9fawAAAP4AAAITAAAAAwAAAA4AAAAev+M6GYBVASq/vGuV8SI2X7+n
    nVHs+bAYv6TW5Y0JFQu/kWww4vFX5L+oDkezbAomP5VjN1ZQKyI/lNht5+qt4r+90NAMGkXH
    v5UV4STFHF6/oHuKtB6KiL9jFZ1qwuogP3lA+cNvwvi/nu9VS24Wgr+zZvTohKoRv6RAkJYr
    N+o/myfZhL4yPT+5EKUE8xZgv4E3FDJaGYy/mjUiNW8cyr8zDtdI8zuAv7tWWF8xxVc/hs8u
    D/wJuL+7Ix+Issdxv7TL9wKsj0M/sGq450lfB7+1y0lRJZXnv6J/sQ9joN0/n+DuBpUSPj9T
    FE1IlvogAAAADgAAAB5AA4tpeQZcIkCdJ778gk1wQJ6TSnvdXnBAnD6Pn+iHVUCcPOYri38u
    QJ6D/vF10JpAmzztybeSzECefRV6RttJQJ6p/nbR4JNAnqxBiqhjm0Cb0Jv6ZpKMQJtvTMQm
    Lw9Am1O/s5Mn+0CcUUc/BUohQJ6ryQO6wvNAnnp5yf77kUCbHdP+dPZgQJxwFCHevvhAmt1+
    OmGOAECerENzxOK5QJ6kbKmimTZAnSAFYnwMUUCepBjV4PmdQJ6pF6rZX3pAnqslCMgZy0Cd
    UOL9AXOaQJr484ME1q1AnCOQo84P50CendeyiBCCQJvx7is3eLQAAAAOAAAAHj/wAAAAAAAA
    P/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAA
    AAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAA
    AAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/w
    AAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAAP/AAAAAAAAA/8AAAAAAAAD/wAAAAAAAA
    P/AAAAAAAAA/8AAAAAAAAAAABAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAADAAQACQAAAAhw
    YXJldG9fawAEAAkAAAAFbl9lZmYABAAJAAAABXJfZWZmAAAA/gAAAP4AAAAOAAAAAcBSi3CK
    9lOpAAAADgAAAAFAJ7mJXVQ1mAAAAA4AAAABQGKLcIr2U6kAAAAOAAAAAUAzARRFesUNAAAA
    DgAAAAFAJnj9Bj3oVwAAAA4AAAABQEMBFEV6xQ0AAAQCAAAD/wAAABAAAAAKAAQACQAAAAll
    c3RpbWF0ZXMABAAJAAAACXBvaW50d2lzZQAEAAkAAAALZGlhZ25vc3RpY3MABAAJAAAAC3Bz
    aXNfb2JqZWN0AAQACQAAAAhlbHBkX2xvbwAEAAkAAAAFcF9sb28ABAAJAAAABWxvb2ljAAQA
    CQAAAAtzZV9lbHBkX2xvbwAEAAkAAAAIc2VfcF9sb28ABAAJAAAACHNlX2xvb2ljAAAEAgAA
    AAEABAAJAAAABGRpbXMAAAANAAAAAgAAB9AAAAAeAAAEAgAAAAEABAAJAAAABWNsYXNzAAAA
    EAAAAAMABAAJAAAACHBzaXNfbG9vAAQACQAAABdpbXBvcnRhbmNlX3NhbXBsaW5nX2xvbwAE
    AAkAAAADbG9vAAAA/g==

# loo_moment_match_split works

    WAoAAAACAAQFAAACAwAAAAITAAAABAAAAA4AAA+gwCSZfNDtrTLAJQg4rmPuN8AlzHES4hut
    wCIGBetDhYLAI0awRsFdk8AjINcUGsscwCXMUMW16IzAI/L2nOOyq8AqdhFsysKGwCYWOgEw
    BijAI3cQGt4njsAgq4sZEnDQwChE2LymruHAJNOCW9qoWMAmI8sb5ckXwCGzOpKcK2zAJv7I
    b1OkycAmN8XjSDXswCSq47v/hGrAKkuVjS2+qsArvf0xkquowCcn6arojHLAJVSkdZTYNMAi
    8ZtFHnYcwCfz3i1lkVTAGruuA85pIMApgBCZyu8iwCUkWt9LdRvAIzGhBL5XTcAl9Pty6V89
    wBqIpxRnqFjAFgbuAJyecMAmya15QZBDwCc2ql7TNUfAIp0rBV9IRMAhsA13A9yQwCapB7rS
    uL7AJ0PI+WyTfsAnx+KQwqBMwCRSVuhxWM3AI8tC38n9fcAl97qfkuyWwCdQWbnatcvAI8pe
    OXmmyMAZmkxX53LYwCOZ5Tc5mXjAJjwBmcXHFsAk20Jx4OgdwCgui+gwshjAJE7PfC0Gf8An
    ZVYIRUlpwBuEbwPsBzDAJqL/tJDae8AYoDbV73zowCfVnRKjGNzAI/OcDmz1PsAdVgKQKmbw
    wCVxSq5R/DfAHU0HcNNxBMAnaW9kiy0zwCOGcok3LVDAJBKkL7QL6MAiYpU7q08QwCTHX32M
    /JjAJ2cdmc2M0sAj28o1eHjbwCWJik13j/HAIjdWtmTwWMAkSvcYglIWwCSh2lNthc3AKEde
    wh6wH8Agvy/nPanSwCHuyIm2S7jAJXIQ3UCUf8Ah1ssyi808wCNqIDtFybLAHHTmbfpl8MAi
    YlLVr9/GwCZzHZCcdrbAImQaC/EEPMAl8+ZWWIMIwCC0WMI0YFDAIehpg2r7gMAd4fdRXfwE
    wCXjHaCcWd/AJWlkJvQIGcAhpT0vbj1SwCpDNVYyoYvAI1lRmeI5WcAlj8TPjDwowCN68hoz
    oM/AJdsSfkyfcsAm/RXopWr1wCtg9wqh6qjAKVYMy7eUZsAij8BC1RLYwBRfEnD1NzTAKGna
    8nd6scAtiVLMf6c/wCTvDr+czUvAKWpvU1q2nMAj500KMo8cwCJ70kpAuVzAKB5ZVU2ZnsAl
    GFdW3GWgwCGLXQGVL2bAKOEIyYEzg8AS0yQ2jPuowCHOzNTHqZTAIljvriP2AMAphMZc6DM0
    wCO0YV1Au4bAGUv1GpcOaMApOYvNs+iwwCNQyFNwPj/AJtQH7+BllsApp654oevZwCe9vYJo
    orLAI6u0gnKjHsAlxLkAvL2dwCjDXVbJ8KTAI2paDwqxQMAsgLLdDbriwCZn3BJvAIrAJxvC
    /2/D0MAhDwVNC7I+wCTnU+UNFefAIn25rJ2ydsAidWUsBIGywCDSDRHnBALAJrX+IyYg9cAY
    QQ4yJrcswCSLb2Cu8EPAKMNAwtx97sAqlNfdwfRawBm4ckNPiOTAH0IhKSJ/EMAnph7seVdi
    wCgCKZ/zn4rAJSMSEdwiFMAmSvKrLKC+wBeBF/LkuZTAJfPQIf0kjsAkEMaKN5cCwCdpTR8o
    fdHAJ6QoKHcntsAgLABDShSswCOFxcXUV8HAFC7XtYEGYMAlDWZrU91vwCNpP5QNKc/AJPMI
    yHI0FcAmkZXjW4DewCKeZOo9s6jAJiikV4Dr+cAovKmPSmcNwCgJwX4v0YHAIpY5Y6rTmMAo
    FelCsGomwCSQYoY5++7AJhQoo16efcAlMIN0+sxnwCILfNky+bLAFzm5Gkm1eMAjqAvlb6QB
    wCjS6jd28pXAIiHL9EPDssAkCiEgSGN3wCgoiBrQNyfAJxydTbIgWcAprxlOZ6HGwCqEGGv7
    iZLAKuJklh/46sAm0vUTlmYfwCkyPE1TGanAJgaPt3LALMAcm/bC5LP4wCf6FJ/GGb3AJafK
    RPIEa8Aljse0Q93GwCEFAPf1eZLAJwPox0FLRMAlHs6OL9snwCD/Yhs8zxzAGHr+8gu24MAb
    +ySMytAYwCZoFI1VQ23AIPJg0i5xDMAh6VIhSSTOwCgUCFSWfjbAHdDTCJY78MAiGgqEKLaW
    wCk2/3dJ/1jAImOXcBnIxMAkM1Li2HJVwCY2Bf4iQ0rAIUdduOKdTsAjQCE0MH7TwCpZ1UCP
    4AbAJ6g+Vfo2xsAmA1lP/ka5wCD04BVZsU7AH7cnqfLj4MAn+mGbTBjewCUaFlLPiKnAIvkG
    LEWc2MAdKMu1vtfMwCCvNFesHhDAJhh6Bwaq9cAldL7yn/CzwCopcMbG1GbAKcEY3BvwAMAj
    uli8I+qwwCb83GYYg97AJieTQnpO3cAmTsFnavpzwCSvCoe1Cj7AKx+lbYII7sAjPGBmpPxH
    wCNTWP0STGbALBxmZLNHCcAjtHCEeJ6lwCq2OKwWc6LAInh9cI1NKMAmSiY2+XtHwCahAzV7
    APfAJZMQoP/2pcAg/VbJd6FgwCU9+K6+1s3AIe+3ZVig8MApiKIdeM6WwCk4w7a1ebDAI2mJ
    8U9SR8AlEa62OFekwCFHGtAY02zAKvfGtTLgz8Aif2DPpS/WwCdgS00zLPDAKVXcNx70HcAk
    eTN33AdOwBkJ5DtEHUjAJZvbwJjQW8AmxmbPVAxswCFJqy7i1FLAIeXisi/lasAf6qFfxayk
    wCPhh2xmYc3AIsBkM4emFMAju6xOQRU4wCovGsQSEmHAJ9F0HeRNM8Am8eCFBYuUwC4FirQm
    HETAIPQDsdicMMAmw8L+f2SjwCM2OQ6TD9/AI5rDKSNK08AnPV4RTZBPwBI/brq7H1jAI7Gk
    FJRu18AinxnnCQmcwCHtvWcFuOLAIhGGzRCy+sAlYq566uQgwCBYuBEnECDAJCS6B2aKRcAo
    GsI8Bkm5wCtxr8KWGDLAJeHSQckEBcAq7VNVn50VwClIrCuQBUbAKL/b1gxkx8AlJDsgTsAN
    wCqPDUMmntvAIwQoX+PmXMAlfk35D5bgwCkrMvuKFc7AIz1GIeI7LcAiYe+D6xNUwCGUggRY
    zqDAKCeShKz+gcAjLql/oy1QwCQ7LHQfEgXAJJovqbWyWMAmk9cFICMawCcCMdyGT4nAI3SH
    Eu5IDsAkcQ9vv6B6wCc9ZTXkaeLAJOk7BuzapsAqfR1Qp4L7wBMVv0wxnAjAJd/JPfsRScAq
    SvYs+JCbwB0ECnmNVoDAKEd5iaZ1y8AkDBIo/taHwCSRzALxE0TAJiE8D8vvUsAkTv4ovDa/
    wCcYYFaNUnzAJFJWqS5qIMAjcLIEckZxwCJwHfvoMY7AJKlcfB4PgMAhGsA+v13QwChAnqe6
    xq7AIskIPDJ8FMAi7syKqKokwCGdM+nrTH7AJ/7CmIJvesAjVghWpvzpwCMfU8WCK0LAJqhm
    coYZdsArQVDJ6VlewCnNcuYQTnDAIG02fnWZ+MAmhD+4YS//wCPzUgIWK4jAKCVcb+2GXMAi
    cPV0LNc0wC5MNQzOfvfAJLHiMKBOAcAgqQX+qngmwChHBV8zmmnAJAV4/QuhocAmIyDxdfht
    wCFX6M9yVizAH56o0F4sDMAR7LxXEwO0wCWi9XbXQkjAImqZ9XS4PsAlqbCu9L2owCGyfTGw
    J37AGq8TkZ1N/MAluH6syzi7wCxsEMj1P6/AI0spn3JhacAlJnY5TRv5wB/SAx4rK+jAJ9U3
    r9w0jsAbeUBymI/EwCVMdTYknUjAIV8lUMHxPsAcwloqbhDwwCan2sTJWV7AKcbxDIyhO8Al
    ze/Qnu+4wCl1uziDQArAJbqzk0RDu8AqMhCWpH/xwCUQs9Ulf1TAIkM1BuVnzsAcQwuS58Ww
    wCRAFcy9BJbAID2p8hXvosAifrZJZSfywCKxLiH2LvrAJYs/e2+FZ8Aue1hMSHaZwCLWmyh6
    E+7AJPYBuLG9FsAiLhv+EehMwCpp7xX7nk3AKbp1DbEArcAnx2G+UwGewCQ976M+2TjAIvEo
    yiRLasAoM48ofO+JwCHr4nz60MDAJGwEqlr0H8ArSyheIE5QwCcc4IyOyJbALT2pCgTrRsAj
    W8Z9wWoEwCbUC2Tz72vAJUjOn5IQ5cAjDZVkUXAUwCS7OllAgCDAJc2CuQnFhsAliBY5JfFT
    wCj7maBnZdTAKfvCwqaLs8AizXhYfFE+wCRKM0kp8yHAGh0/XBr6QMAciIT5Px4kwCbmoSVc
    wKjAImYjI7TLyMAnYy16jzRkwCVU9brHYdvALPluD1u8vsAprOuCC5EvwCGOn88YAjbAIS9p
    t5foSMAi0Pjk1DSOwCQrtl7cwkjAJIppTRaYSsAlKBr2HfjjwChdjbI7WlbAIpPsweUyTsAh
    ilc/3mqswCirwbFE9iPAI7mp/Hth8cAiclEkC3XawCiOIR49/g3AJiSz00QfAMAlnID7ofgK
    wCBju8sIeATAIw/3VN2prMAiSX6qlzXwwCIFsb5XFfDAJhDBXxKvVcAk28t5wXkUwChryiru
    jgHAKkY1eN0oyMAjBsbqozdMwCXjyv2KgijAIFkUlDZsbsAhtWEf6l0qwCRX8px28LTAJ9RC
    h1jh1MAie5y0ThfUwCr5RODN1JjAJYJZc/7TrcAnpe/XBLcuwCWZ5HuZZTPAGjjSSzY3vMAk
    TiwfKa1lwCPlTfPxujLAJq98q+p6h8AjQupl4tRiwCMsZa7M6nfAJiz3ddJs8MAlWH9gztgk
    wCT8JrwDSWvAJy0PNBxXQ8AvjpvC4DDQwCDW+MnVk8bAKrEMEVh3FsAiDD5a3Oj+wCZy+GhG
    swHAIaoXDyiPOsAm2hct0KhiwBxrBeAoDdDAIM2j4HTacsAi1YEXD2cQwBbTVSdrjnjAKSTD
    1mj4jsAjkNCKt4hqwCYPlvYAZxTAIrTJc6YfGMAjZCUz4enWwB0WfFtzo+jAJV1Lv32dHcAe
    qilThSTMwCrkso3JfhDAHw0XfU8QHMAfqjVrtPS0wCjm5TSn4KvAH9E+th+DSMAo3T5zWj4m
    wCPoph+lOefAJeF9T4A/3sAoBM243nUTwCVbJWI2tgbAMJ5UV3BWmsAquPByiUKHwCXT6g6t
    76/AJCl33lQrXsAqI3ltlfd6wB8b/gJC9ejAIG1b90VulsAk/9kMxF+fwCTRlSnjLm/AJCul
    iIcjr8AlQhcsRTgBwCNBTbOOdW3AJawdOadECsAiq7njbZlewCeoij7RYwfAJEoPYCyTNcAf
    lkO7uddYwChb+B0ibb/AIXFU66KRlsAt3sLfi4bewB8YaFjKXHDAJQUb3YK4XcAoM3d6217Q
    wCojwNbo4OrAKPRd1QUBM8AjCSC281pSwCPuk2q6hjjAIwfDAiL89sAgulqGH+9GwBUM05/I
    ICjAJmQZthPKaMAkwwJjA5XWwCZB8FlPWpnAJ8jaswcVr8Ajv1ESd53twCQHjP8d1wXAJm4w
    qN4hp8AkNTPyer/MwCaK0wVoDCnAGsgvOrx6RMAk5Lo/LTirwCfUPpoxrVPAJfSUdWWmU8Ai
    A+GhpYt0wCesGes/FxvAJcfbyXOX3sAn8rn3JeM2wCQtyqWrqZvAIRXnPdFOasAc6BxCnBMg
    wCNIsE7VFiDAIxR9UuV2tsAhaiw8pqXewCibXICTUy/AJkONM0JdgsAmFaBUfoPhwCGuoUTp
    Uz7AJOdRiXr/CsAoKNcqZ4rwwCWQC9XazHPAI9GbTnQ5r8Aolny6OI3uwCBoX/IytNTAJsIf
    E8owMsAjRaMaqJvEwCO51nyTTTHAIX0eex3r0MAlFpJp3x/awCeQnRGDNO7AIHXDRa9+SMAg
    OYhk/LZewCJJ/Gbf/1zAJ0nMHsrUrMAg40wm7cvKwCYEFCaZBQvAIkoadaxLaMAiCe7qxdhw
    wCrXMNow5qnAKHAl1Xxk6sAiZwKAGJ/SwCQURzrhUrHAKJGTEYDsPsAoTmeaX3hcwCArOjhI
    kZ7AKuJKbF4/KcAoZIMwu3f7wCChElNYOsLAI8xtOxCwK8AmBkMNcu1jwAyoDXUVdmDAIwWN
    o0yZhsAqe8U5/hpZwBhUqgFEafDAIk++uq20GsAljSBpCPIjwCtEFRBiTX7AI0ABIMSmC8Aj
    UVPWDty0wCpOQFIKuHLAH0Co9a3D1MAmRE0lUJt/wCLmgY2AC0bAIw5g8jZOtMAlZt1s5W6i
    wCfhcr6sj9jALFnU0r83Z8AmqdJfF1JMwCD4jf0E6jjAI+z224x4UMAsgFVmUriUwCdLtKdY
    8OHAIbD+95fNmMAlxLrqkSw0wCig7tLTaEfAKe0zfGP/E8AlTcQQrMWawCfBFFjnG+7AJHnW
    A6uJ/sAdeaPJ1uJAwCTfE/FfxDzAIyHhD2o8ssAldBKZnPMowCeGMFgFGPrAIca9SHBNhsAm
    PKjFFkD2wCEm2SRYnkLALDgbZ1W7s8AXrmxal3A8wCmq9Lr1J+rAIoqX/RCxFMAkRte7Apan
    wCZyiX1EB4vAIZhmKrlGMMAiGofsXgqKwCjxQGJALvPAI8tvnw3m9cAfp/iQ8ce0wCKg0W1F
    b5jAKzlJQ7u1vsAkLes/vN1UwCQYNr9+a5rAJjzHn8gWZMAnO5UW+RfnwCan8kqAWQvALAdY
    eocLsMAlP0Gz/51gwChWhvNQwbjAHDjt41iIIMAkIBbr/NvpwCFlBorZSwDAJ3E+AcTKz8An
    JCkYLFRUwCAr44xFp+DAH4agb2vQPMAgDGqlkToIwCljnbf4OLjAJkotjiWYzsAkdSGNAxVE
    wCB6deIpTPzAIiaoJEIfVMAsZ/6Bn+PVwCNTWROKAt/AH/kvFy1MMMAncys63XxzwCVEzFFu
    WbLAIzF9wDgon8Akmy375A5UwCO96bqvEKHAJ8NOWOz6CsAVNDqHEzncwCYzx/cH1r7AJVF/
    cBoYBMAdcMaD5Qp4wCii9QfOUzDAImngolyRnsAe6GcmYMAgwCSyWJskFlPALCuWGrwBmsAb
    24ETdeQMwCjLe6W8D1/AI80hxn8U1sAqJb7ImUffwCZOBJvIHVzAJrSFHwSphsAlzKu4OSq1
    wCWso8q2ax7AKKRIRDYRk8Aks1sLGw7IwCckIKYDqKTALQZCCSJQ7cAj56kna944wCEeE01F
    FajAHpae0GSqXMAYBFMFmjfQwBa4RNSrJjzAHu+ytaJHLMAbu1Sn1BRUwCI060m1zDrAHfuC
    i+PykMAc8XTbk21EwCYeQxXCo5PALh8qT+at9MAmPepsXsp6wCYeAeifoMLAIs7ysatzhsAm
    PldXo5D0wCLVSnu0K87AJIFkeWpL6MAn0O/S14mYwCGNiYVT1jjAKuOKFuBEaMAkXA1KjOOW
    wCOmusGZyxzAIwAnRax5gMAn2P+kkyQWwCI00XbUUWDAJyfQSCZ66MAj1xEEfy8FwCthTVl+
    +AnAHmvt2XdQJMAmGQBJpxBIwCWPMqbyZ4XAJltg9W4yTMAliTILtpp+wBDCYcld7UzAKItM
    k66A6sAc+sQ1yMF0wCE5yZuxgkLAInXsgj1Z0MAnJnrc910BwCczlDPpZdPAI7DVeTtBfcAb
    BU2GPKDAwCalPMMN1vDAIUKJ4lFVFsAcYRmGtzhkwCAZmPst1EjAI/vbYe3XNsAjFxqQb8BW
    wCcF6y+cE5jAJY7mGivSSMAq1hNzfOBdwCd5AXPMA9vAJS/Lvn7nI8AgwYgcXmKWwCZ3CRrS
    PDzAJtnqUuGRdsAkz0S/qvfawCdYvIFV2/jAIWKsbGes3MAmcFycSB5ZwCJw9BJTclLAJ2GM
    n3UPKsAl4jMiav/iwCGwet2+qQLAIhhyz7spFMAo4J415SgowCNOrH6XrqDAIsX1LwE/UsAi
    +5WDPcEkwCk5sFGQMxbAJDTSHbBu48AqRwNXLzzEwCE0H3dgfbrAIzXGxVbcrcAX72eKFYmQ
    wCJ52aCAqYLAJaTT94bX9sAio6+MRlv4wCSzpN0UFC3AIvTATHLYkMAg16D48XTYwCPKPV1L
    OMzAJtaptj7z+MAq/RQe78U3wCOod3hiLfXAJcnmUvkQuMAoYYsyQPkEwCQQrvSc9UDAJX6g
    IcB9R8AZK1Plth78wClAgOSXgILAJv7SR0+tKsAfyixShnYowCkWbD8fOtPAIrPGytl9xMAi
    UbxP7FV8wCaaBPoxIJfAIq0OdDBAUMAmz7cuyONuwCXVTSOW1GrAI1nziWtoM8Am/ZJAO9C/
    wCs/OcUMC5HAJ7mPLkwTeMAiwey5C0YKwCl7GZM0n2bAI02xQ8HVJ8Aqz+0XABDBwCIZCx/8
    sOrAJcF3CYNYLMAWJnm+3pDowCGs3Nl6TxLAJ2NwV8wniMAmedgT7HmFwCpiWBq66FvAJZTg
    6rK5OsAjVVbFGiYtwCL3Qq2W+RzAJ6puQQUxTsAhB3Tc4gZwwB6zosyGIyDAJisXdcLTcMAj
    XpifRb9KwCBsslG9CQ7AI7J3u7GhAMAjavgrRKbawB9aECA6WqDAHuHoGzE/KMAkiylpq6Ai
    wCEg0OkJ04rAKWir0qifAsAg0K+n4gv2wBWilTAtGEzAKIsXOBuxbMAnpjI/ZNpXwCUTOviv
    wj/AHKWfWuCkUMAqfWR+Q4WowCjDXtCOJaPAJn7Y/dANpcAlgFoXJe3bwCqwwdwSiazAIJbf
    bCkiMsAlorfA+ZE5wCUXNPV4kQrAJxYZUXfbNcAi/VapgkHKwCOl1mjIDqzAJH2AAVbWiMAm
    uHQSD+UGwCh1Sru15mbAIC5Jd/DXfsAd8wRtTNBcwCjsX8wC3v7ALL4qxhKVa8Al+jEm71Nr
    wCcRJcRZYw/AIUKcAMrUFsAlTYgsfg3GwCV55z9Sx3XAKXvwELCkZ8AnMC1xaknswCY9gim8
    WJ3AJ0otrHAXy8AtCddzawqAwCcAiMbLv0vAIYzAiN1aKMAj+GwfZ9fEwCSGJ5oIpMHALDvT
    HT1ilcAkhhuYKCOxwCk8viI+LiPAKcfK4oDPmMArIckmslJzwCRY4fRwnLHAJG+/Dh18I8Ah
    kj9XTvnowCjvNWGjLXDAJ3EusJQHSsAe6Z6BhP3gwCQ302/SpKHAJKf9rLwrmcAiwd+O1n1+
    wC2FQ3xkGPzAKoETkyJyLcAoSlwxYlxkwCdABFpCPcLAKNNcaTKmscAjj+FLZjFbwCbNJuqR
    rXHAJV9WFPb2MsAhOcic6yc0wCIqxSDpb57AJd7RsQJi6MAjZJl70kz/wC3KK7T+6W3AI6g/
    1lbJTcAY55kqnED8wCGADdguv9LAIiS0oSCqRMAmlx8yL+XRwCUK6zrpf+zAKTww+5sTPsAp
    sPYPcBVrwCIdfJFqzazAEEdn3N++XMAn/eePEa5NwCWoBogo2kvAH5kub3RzGMAdgnnqPGi4
    wCa+bwmEraTAHoUZWO3Z2MAmoVMz14QDwCfmPc6TZPTAIyRrZqfzNsAmYzVWRwGzwCXmlGMR
    3ZTAKMzu0JhC/cAg9Z8kohCWwCBmNVwtP1zAJEgviOBS3sAn9wgsArurwB2uQZA175DAJbet
    2TsVVsAkKi1eS00ZwCBU/LpvH6bAIxGsnH8eBsAj8pht+oAGwCXvBbB5iLTAImb7J/ZQcMAn
    JU5r7MJvwCakHxASAGvAKaCZU877fcAjtXjR4VCZwCfFsFjNT+/AIFJAwjgUhsAmVMPSzb5E
    wCbqEy9fJNTAJUp0xPbwjsAhHXZLZFUwwCcrT3xh9ePAKcp+K+Bo28An8uqfK+MQwCRTA045
    /zrAIY8/aukL5sAjkoHyuRZlwCKoPG6JoOLAIWW7K16HBMApLy+tbum3wCZjV6Bljw7ALXLj
    t1vLZ8AwPWGIZ0TawBiNvUxOx1DAJj5tq+e9dcAkHzC+7HogwDBUOezBZyzAFVpCFfQmcMAp
    FhDj8bROwCFd8VdbI4bAI5rjyYnLm8AfBLfch1a4wCcLjn2SlfrAIyAzdWufisAhcBzPP0G2
    wCoBk6GGlZrAKa5HnJt9d8AgTh4MyL6ewCTgNcCz937AJfCS3XCP3cAmBAHFDvdhwCQpXIHu
    +0PAJZJkRi3aXcAnybGOEM2mwCEFN6bu8HDAHi4HYep+CMAnSJvN06u7wCjPEimoUJDALG4v
    4+3O+cAhHz04chScwCi2BHEeaHbAGalz2S0lBMAnjHk1EngswChcxW7/qbjAJtPogOvW4MAp
    oEqjTATnwCvkiK3qMGbAJb4TF5LtP8AlFC2tMDOPwCA4+ylxhCjAJo2/zD2lzsAjN8xs5eEc
    wCTndKdwgzPAIIF3v3SbKsAm0BQJnr6AwCGFIaRUmWjAIFFDDyeC+sAjzywwVA9OwCCcScFl
    WUDAICPAf9MPvMAioAunnFRCwCS1sbVsqWPAION33aJg+MAn3gPIzzKdwCOC+CUSjEPAJLs2
    shNJYcAg/ixbelDOwCPkfPTGqHvAJelhrACrp8ArtGlwV84RwCENHu4AI+jAJVG/EDrQb8Ac
    uNDgOPIUwCYALwLIL3DAIFbZ5m7Z9MApYLciEnS1wCF+dPsDa1DAKDjb4ThQp8AoOrBmutlp
    wCQt5nxP2ujAH5VyyNxVRMAi4ZKkAr6gwCRo9g3bn9rAIpbntLkA/MAhtXtjcSvuwCf0ymti
    WybAJbNNkkXJNMAiDiocxYfSwCx4pnEK42XAI+6GxDwWbsAq14d66t7fwCb0uQTkeUvAIlfS
    HHNUOsAfCxoJ/5MwwCSxycmn5RDAKCXEmbACDMAesCS0H9cYwCaMayMiubTAJRRCv4QnJ8An
    rCKifc9wwCQuJgPPBoXAIf6gn4htGMAoJqAzjGnwwCBiKTYXM57AJIqnoqq/zsAf5YH5Eq4w
    wCkvW6ajqLzAI7rVW2ytmsAjZeA55rQAwCscieHp+7jAJR4k18kt48AqRVAYj3euwCP0rBGT
    nGHAJUNT4JbQHMAnf/ZAlcxBwAq+7P80xADAI5hT0WzNMcAeFNxb7afAwCGmX6SaTfjAKL3r
    3eyr/8AiUYQj7PegwCUtAWN3EzLAJrpYToY3e8Aj9C3UIVuhwCRHOPjUQdnAKTARUxY10MAi
    x8sN7eyEwCqR2+3mhqrAInA9BHw8XMAlxSComn0kwCAz9ganNNDAKCV+yY3E48AeDG5qJK3k
    wBS5c5BIIiDAJiO4yKh1G8AhgGVt6GV0wCflmzzjPwnAGTu9ivzGSMAhY07JN0OIwCKiYiUu
    H9jAHAWTPceVNMAhO0BGK200wCc5xfn3ThTAJCK8R/PxUcAeh8DKk68swCawAoTh2gjAF5f4
    em4qbMAlxCi887xmwCCOV5jYf3rAI39KPx89e8Ap03aFKL18wCNEAAH8xYXAIzkrr9H0fMAg
    kJ0urggywCOHyRj9IPnAJhdr6mvcuMAsp1XD/J4GwCnD7pUhnC7AJl2x4w2R6sAoTp26t45A
    wCXkySTHOFDAIQ/6i7juqMAln5xnGYeZwCPwuBh/9jfAIXSzYnyqNMAj+yVZ1SgNwCQBjwP3
    08TALUldGwYU7sAkB8oO732WwCPaN6slT8vAKJEtkMcDoMAkchEu5R2jwBs0oGUuMbDAJNef
    okGqtMAwX+3FLf4KwCrOk5lu1PjAJd4eC1ZYEMAjVD4Bjzc4wCfzUqHpnPjAJW17ZeCNCsAp
    Q7P7ZcTlwCYaFaZTC7TAHUQEh7EfuMAjPK/gq2B+wCDCkyuDj+LAKMaYV/gZq8AjllK7/fsb
    wCNts+0zY0PAKkHmgiWSz8Aoo2jvZ283wCXV2J4tmTPAIW+BWEp5hsAmejVtkneDwCSlDseQ
    HnHAIamO7JyI7MAlNh9QBZ2WwCSSSlPT3TbAIP7NQbvJcMAlk2BTRcEmwCgB5cGksrTAF2nF
    /mRuVMAk+4eVziRpwCIuVRBZJCDAKKVIOm3I7cAj88EY4a4iwCGCorRxZJTAMMBa2sJloMAq
    3XYQjNdIwCQX/BxKZPTAJQGuxyfgDMAfh2w3frUMwCSDOH6WknvAJU5vVcBvbsAbj4vazn9Q
    wCsC+QaiZF/AKT5L95ColcAmAF4Aqle0wCLtw587jTTAJggb+Bvw+MAk7asuBzKYwCcs1Tgt
    5NPAIVwseyKbiMAlg7JHSg/TwCZ41RGYHc/AIQVDANt7wMAgP69Ab+WwwCACnBnz48LAKRsQ
    ed+emsAm+zsCXAuuwCSc7Y7Nr7bAJAU95FBo0cAlo52irws7wCbEgW+vmjfAI0cNzuZRLcAl
    jhMFoswPwCK6BW1/FE7AKF8PulmAFMAmO1b2z5JowCazN6L3MsvAIK03yalZ7sAnD9kU9GJd
    wCCiMuZ6QyTAKdfSJSybisAiTal4VWi6wCJ15ibZby7AIXLJsK76vsAmRaQDrWMvwCNwiKE2
    vA/AJfVMi3Yx9cAYZ/lcrBe0wB2ljV9fwYzAIV4h4CxwsMAp4NK6KXGUwBtirLZCo/TAJ5CW
    EfdYfcAj5MWhWKWIwB8DcDGma4jAIf4e+BsWrMAXxHgOTs6QwCSepK1M+BDAHh1FGwcIFMAk
    Sif7JPeLwCP/KWMvhe/AJTcojIHd9sAbEUJZ837owCVbE92v1MbAKi9V+gKGpMAjj2AJyHAa
    wCSZbS95ZF3AJHYc0Tbn4cAgbL/fN7S2wCZbRuqUjf7AJeXlfiBdo8AiaznOo8hGwCUzVs8j
    AjTAH9NhJNAIgMAT/FEr6WscwCT2e23OO3XAIMQ7XZStFMAnYHFgE7KswCd+2b2vKwDAJEkg
    hAVAasAaRmuoY3PgwCaKWo395mLAJmXuMuPXnsAePrX9TO3IwCWbi/iNcabAJbHLBSkmW8Ah
    tmw6oB9WwCPxllr89tHAIg4mp6MJmMApr+EYgJIFwCZD0KjqIbzAJYTOFx0gbcAaASbGJfVY
    wCJ34HzpsubAJMoCb65aQMAqnnH4kg8mwCRBjm5JonLAI3AQYCFMDsAj0Kw7YwxRwCi2B5Gd
    aijAKdhXx6i8o8AffnHLmX9YwBZFKf2ONbDAGYr6nRDG2MAigfAWfxRCwCgUljYARLfAJif6
    kk+DkcAkEdY0uCw3wCYtBLc4jyDAJKR8rkGh3cAlrBY3F89lwCJZDxUyqwzAGnuh2oGAQMAm
    s0CyzT6VwCd9fG1nRGDAHR+oRrEYOMAlmxeP3ORKwCGkHVnk2JTAJtkUVQX0wcAmhF/g9D92
    wCTZLPF4i7PAJxJDINO4d8AqzMMHK/54wCNqbr6t+grAJ43bLMzqfcAnWV/5AvwWwBjWHPyi
    l8DALqHif9naZMAkFYNCJz9ywCazfSETt+3AIpL1mUbvKMAg7MV9++RcwCFB+Hiy8ODAIljj
    Pwt/osAfSfKt0qhgwCdTL3jAmuLAIxRvb5dB/MAlVshNpNoCwCHwUWBofUTAJCxU2AGiY8As
    4tPxVz2fwCDi7jD84WTAH1jtbTtT2MAjFTu4dCVIwB3IN/CSeMzAIsqRfXSEfMAjUiOszdAa
    wCc2mj7vSBjAJhR9tym0TsAnb5RWmNW8wCE7oxhR7JjAHbbvUFzSEMAlKwUM2zTRwCdzCNIK
    YHrAIO7ZPwUizsAitxgzGHxSwCR4hfVNspnAIv9lb4sfysAYsm3bF034wC5dYf5wIYDAI95f
    8NSOm8Aa7SDnecoIwCU9AH81WDDAJkorFAda/cAm1Yol7OwxwCfzN5wNVH/AJL5koZu7lsAl
    sRdjURguwCa23JZMOxLAIOjAINNwdMAor5b0PpuxwCg5gWiZKi/AHK89VQdewMAlvVKOzdk/
    wCe46pJdUPfAH0WW4/ANqMAgHuYEGW9uwCbU+B72V8jAImsGqeeqjsAia5utkjGCwCiR9Vmk
    2FvALUzCSlqyBsAl/GqyYx/awCHndAcrs8DALA7nmWKmNcAmIeRqKEMfwCowJc2tNZ7AIoIa
    hheREMAj9wSjHTtJwB/loQa4z3jAKFD8Un+ZLMAkLL5cRuGnwCkCE3hBBOjAJpguWqlWIcAi
    wVFV43j+wCbAA0ZMt8zAIkfJFK058sAieTew9fkSwCYWU/KTOY7AJemkZqgio8AmfqjLtR3c
    wCLp61lEKqLAGvlCiJA93MAlnMOVlJ7VwCavaqbmBpzAKDyDECtWHsAouUeMxDLIwCS2EeGH
    ltbAJTv9BGkYzcAfsOrNcrAEwCOs/mJMmZHAJQrIjtj0LMAn/elGX3+CwDD79rCv7qvAJr+C
    N3T788Anll5Y5I44wCEk/1gkbHjAJOl7P3CayMAl82DDteWIwCk1SgsYSNHAJmzt67FoBcAl
    9KlDnjBGwCVkw5QT26bAJjd+vMhkh8Aid8aydCXwwCeUanNrpm/AJK748YPba8Aim+eHvWi4
    wCRh3SATDMHAJvHqNV53DMAl2gu5ED23wCHZkK+koXrAJ801aIl+DsAjlgpueiKZwCAncp8R
    SOrALBoM8HBlSMAef486kQtQwCcgZEjWaNPAJt7xPifGh8AjNibzYscCwCTlcrNNHUbAJJaq
    c0UvJ8AnDXug8h1cwCTs965eAKjAGmFB+eFWkMAimESjNqgkwCgmzofs1Z/AIr4aKQRKqMAh
    PCUZ76sywCMn4lQfWirAKppkVxdC1sAinKhScjm0wCg8g8bK2T3AK2mH2/xo9cAgXgzoSQyw
    wCSJiFN5O4bAJY7tJLTrqsAlJ/rWmyUSwChkqUq3djbAKiJC4aq7YMAiWKoehazKwBPHQNzc
    4UzAJ4RakCEsysAk5USKwtxgwCTJXHjFWlDAJhGAH/HUR8Ag6k/nmz9mwCOIIz2ey/zAHC7D
    Tlky4MAjBU82v5x2wCEjhZ/ZOXrAJlvJ/YodpsAlIqBSsRh2wCx8pJIh4GXAIt1NTQnAKMAm
    ZRq2vZZUwBcHpLxnSgzAGuDoJfKa/MAit2OSBtxOwCN2uqD0AgvAFX8DdFnf0MAhq0LLWjew
    wClH6zx+9F/AI9yN/zQHNsAqVKoi27NjwCbEEjdT4yfAJJZi4keMmcAlsbwgEiH1wCUIQh1e
    MUDAJh35Eu08JcAcfrtlEiF0wCStShPCIA/AJASsacr19sAl4YJHnB0owCBM84W+MfzAHV72
    B6HYVMAj5SBxomj1wCMjserQYdzAJ3ekeuRMF8AisuhgDrscwCQlaC5KYrvAI4/1DcYJMsAi
    tA6fbHEmwCVaeIhVxjvAJmRmVGxXd8AkKncMFW6HwCVzrA/FGjXAIZ/EEAbt9sApU1S593QD
    wCbRE++Qz4bAJ4NJEkk68cAilnPKMvZmwClsPx6JIArAH1Bp7H93jMAm+B7w+RS5wCYa44PW
    6DDAIvfh19AFgMAeiCMG7odcwCaJkwnc6ZHAJZxjPwZPIMAkWzaAsm9SwB5PUacIECzAHlN1
    sQ/ayMAoD4F6ljinwB7EsBiBnbDAIKi7Ls6e2MApxzWJIqWgwBtAQQzixfDAKk0DiuAwW8Am
    QSoragKTwCT6rqWf5YDADi+Qol+CQMAo801ssV72wCZKh2HaRfjAJb/E9hVLy8AX2h/L0uJg
    wCN5nVAfZaTAJC3cwGtMgMAtgEwA3izHwCZyjG4XoyTAKUUsil7q68AlWn1ovIQ0wCGAObfb
    uajAKcuZxcTLAcAd6oC3wxK8wCcdVNsxAl7AIU0QQ8T85sAY+Nr6epPowCTBdXyMD+XAKCGK
    IwVlscAmW6jTH/2RwBSNNuHKcejAK7SY0XwnY8AZa9O147qUwCWQOUFGAyXAJZGUF4gm2sAb
    V0bQFSNAwCoVA8krFCjAKv6SXnBXJ8AphSb+KQ+gwCqifaH3JYjAIrYA+kSr6sAiW+upJo8q
    wCU7UuiFRlTAJa1Mkb3JIsAiTIAjqc9UwCS8qzLJeeHAJ7M++dTIE8AhlfjHNrTswCTKkJHm
    Z+HAL2/sY6zNyMAuZ0EOE+6OwCkhUmBGDUDAJZcUo65K2sAmor1SbrVUwCPVMUcwmXLAI1aV
    r7PR8cAmuwTl5nBLwCWEcjJ9r9rAJwTYEz5IZsAqr8VHEhk6wC6NyXVq7G3AKQs6sJBkicAj
    dy7k11edwCaC2Ch184LAJ3/NhLgwVMAmH794oybRwCOjzMSSrlvAJ6OBoBYSisAji5Es3CSX
    wCDbhbFbtp7AIqW+EZ1tfsAhJDz0SoD+wB2c0qDV0PzAIYd7CieV+MAloHGHV0rSwBorGR97
    u5jAKaiBfndfY8AhHj513O7mwCTGHP8S32PAKOzhRk2aXsAmuK9RR1sZwCQrhCaqaaHAJOvj
    NbxvAMAlx23deMiOwCNmcHHwNIzAJJKq4w/JBMAhvPxIj4JMwCOt7pk55f3AI3rXlgZwLcAt
    HhUX8/QwwCeoR0eMfD7AE1RU+KDu9MAkZKcv6MYjwCohuQ45j0LAGyjrWEpbLMAuG2a8p7Ua
    wB5HBiTumAjAEZK4ojVuXMAhLTHoBG1wwCag5f9kHg/AJKduzjQ7RsAXIPLqoDwYwCgNpwFj
    bhLAIVUKNSTWosAqc8t99YYRwCbTjdyG7o3AIyJ3khvqSMAb5hu9O0R4wChHDgrE2yDAJb5r
    rMww3cAnKJAeCoVXwCE0TdlTfCrAIj838ul+TsAa1JfDPNYQwCT2sLcxA7/AI3jOH1ySYcAn
    i4vdqPIuwChYEH6rYNfAJ75P4hVC6MAlun1g2QCUwCKcVJdBQ2LAIawErda6aMAlSdyG1VLD
    wB+QeO52fnzAI4Gm6hyaMcAh00i2wz+QwCTcR4L4llPAIcuOW8V0csAlcFZMRjm9wCKj9p1Y
    CQ7AJJ1JQjhka8Ajp1iueXZowCRT7XUIbQbAJdPuoSLOLcAiWTJiKjcqwCNawHOuSnnAIjGa
    /W/X0MAkwOXexjg9wCOON975hKXAJJqaCdZ8AMAmuRjTk8ZowCKwcqwta1bAJyN5riF9+cAc
    kkNdRVXcwCjQRWDT70fAJyD6kXk8jMAiVfCNDN32wCG30E6us9zAKDf4pBnEo8AisuUadWR8
    wCrUAey5wQfAIdZQNZBf8sAnGw0KovXpwC1yAgvn0lDAKTX2Lq49EMAh5GNMO1uQwCPa1hDK
    x3/AJf4Bu9gGI8AltcYFkop/wBvGHdN3hQTAJTajCwXWqcAmXyExL975wCLrz+rqGVLAF1H9
    hguO6MAlDKoro6wDwCbAao6/x+nAIdLAUOqejsAlw3obflhdwCSJS9ASu5vAHFchK6eWLMAq
    T7QibjeBwCAVDQYOPZbAIdvztWB3MMAsYp4aJ8t+wCOWVM/K7JjAFmMLVSPf1MAjae/iRzrb
    wCBudpAXIebAIq75bA4v7MAmbcQ/H9tZwCtTHjLVpUbAIysa7QizI8Akb9658hbLwCL3GjVU
    tWLAHb+WvGxOzMAnbLdHpWbfwCKq4+wIEt7AIMZIoa4IGsAW7cfgcre8wCHCUsQNZD7AKFtQ
    WLIb08AmSFESIasQwCd7GkJvkv7AJaNTWhpZe8Ape2pCsy4JwCynp2B2PVLAJsE/YCXmwcAi
    bRMJwOD+wCi+E82mqoDAJFmLgDQ7acAmK2ZReXPvwCUXIjr+c4bAJOvPd42SpsAZGranW3c4
    wCsy/sF6kgzAJV/xOL9buMAh97sxe9p4wCcQGHA7dG7AJasrZ8N8dcAlxbqHEtSJwCQoCt1K
    k6bAHCSLl9F5AMAl6pcEcBT6wCCM+jZShozAJnA2zZ/F0sAh/nqhrRsWwCTdr/T5mp3AKc39
    VPLhVsAhs5maYQh0wCJ5joMBdtzAIdafv3iEzsAlztdP6RkZwCCOnJVX7FbAJYa5UQIE5sAm
    WGOJ5V2fwCb4O+JusSnAInE9vjlhlMAleL2cxS98wCIjlLd5fkDAJbvo60GcjsAhrBcotvcA
    wCtQyFs2plDAIpxUsYnKwMAeNmEZUxVgwCPf07jQHDzAJPsHU6qh8MAkFHdf0qEqwCPJn7rk
    YqTAI4k9d8w1HMAnrRQRN6YWwCU/DCUjqCC/+BVXxjIL0MAbHSF5upFYwCNVP3tu2YjAKFRt
    M2J8/8AmRFzOToibwCm3tFYVxGTAJRi4BPnzvcAoWefMRbpuwCEH/aNotzDAKMTy/4P38MAh
    V0tX+o7KwCci/ym7rgfAIwILwCgo+sAnE2SoHYuuwCZ58lGAjC7AIrmDM0Qo6sAnByfjsq2j
    wCIPEzj3E6TAJdkjaseIJMAiIRhtDP0wwB06+bJG/hjAIQUKEAicosAg2bkMXC7YwCaPfcq6
    cELAIeHA+vQAbMAkU5+/05D3wCFiIrtH8mzAKG5L7raJCcAg/G6PzUaCwCSWhJb6chrAI4wn
    MVazcMAm0XO8lhUewCZM1Zmw/svAJmmUVP37/sAhyCpJF9l+wCU22Fuf13rAK0f4swwtP8Ah
    2lquOB4IwBl7fXqLDBzAJkdXMBBkVcAkaHzvZccxwCP+S91TgKfAI7mYWmGXmcAoTCaObIDt
    wCdwEPHJo1LAKemSVZi+ssAleGG/qkWHwCMUz1obFNzAKGiC/Vl5BsAqg0CHu2XYwCdrSgkX
    wfLAJvm5B1+v+MAmkpVUpBRjwCIw9eI3DVLAKNb8Ab8xxMAnuR0SEyZ1wCU3mmFKO5rAJWgO
    dFfhMMAdZ+H4mMbIwCb1HlEVDRnAJZsfMBf148Aq5vaaeKnPwCzq9nKCX0DAIdv5643sssAj
    NsFkJjjmwCEtMfvI0qrAJDJTLRy0+sAi4OpCiM/kwCg4C1NaZVLAJytEpD05x8AjTdH8+zFF
    wCS4UI0k8b3AI4IwnEsPF8Ajg/UZdFa9wCK400hJ0czALAuwTqFvgsAle0eHlli9wBTj8TOE
    X6zAJHm/ahBYkMArD4XmIQNdwCT11P4k+i7AJhQQRncDTcAkZ8IZJsT3wCOkBbzHXwHAJ6jA
    CbLrk8Akd5sNx4WFwCLN+azYEj7AIXkEeIAo8MAkqQYM9jJQwCbK1Sx1J9DAJZtY9i6ub8Ar
    kUm/TbqOwCbCWc0r/CbAJ4hc7kve58Alpi07J1bIwCkGNf5bPHrAJNXhVrdTesAlZebPgL+m
    wBzVTq9oIHjAIE10k36HRMAnaGNuf9/LwCI8i5XONkrAHlR43IhvTMArew/wHRJZwB6KaNFD
    NcDAJGLR8VsIs8Al5QX8CNb6wB/+f9ISQJzAKGEYdDCh6cAj4kwdxGknwCyw3yZggdvAI7hB
    FccJzMAkckEXxeHfwCousRV3TGzAJ0Lm9w1qKcAl2oTk/IqiwCDbDtmswdbAJWO/4sgDS8Ah
    kA0U83+kwCcK6ti7DjLAI8+LlAzVYsApUviSwhqwwCG53jKod3LAI69q/R5oc8AkD7BtmTgd
    wCHSoxT3U3DAJ1JXS4KBO8ApzXKmn286wCPMSXfz2k7AIY8/atyoSsAYxGSLk5oowCg5Lams
    Wk/AIayD4nvsUsAlu8xH2h0cwCNxPXo1vt7AIB/H8bNJbMAk9BHOp81QwCTGih0nPxXAIQJv
    0RExVsAk3xhCtd2ZwCW4T+nL+zvAKJ1ARS8F7cAlblQufGURwCX50tba1AbAJukCqhIrcMAd
    MebNhBJgwCF6Jlt1Z+zAJxt6rBUDGMAoB76jQv4IwCNXDBfz0MHAIofSzH2T4MAtHoBY5ixE
    wCfvpr2NiWjAJ/iFOF+NasAfynAti5X0wCJt6bm+EPbAJqamvjyuEsAnA4pIxFjKwBuwe2Mb
    uKjAJkwO3x0goMAaD0Qw4FuowB/qurXuaEDAIYMtKRCK3MAjvJJWwYxWwBqVkLIPBDTAIggx
    o705eMAlWebBgahhwCMJDLcdEQLAK1znFhtyPsAk/lPuTVm3wCLQKNbaRArAG9DXM3oKKMAi
    MYzpeGv2wCmJ4wKMP+jAJcUQl36pL8Aje0uDijrSwCrN+MW6hs/AJmqpVoPgYsAkC4yfIRTy
    wCKuq+myNF7AKsAqH2jMK8Akvv3a2BoVwCPMDWRc32DAJiPWg4ejh8Ao5N0aQkouwCTekKH/
    vzvAJY2O7YcKUMAYLSMcW364wB4lqL+0NEDAI5ucHeRtz8AmsGUDob4vwCoXAOklp/7AJwRt
    dbxQAcAjKmT01RBGwBual074rbjAJ5g7mNl2XMAom5ObO+bfwCLQw5vyNb7AKeoiYZurL8Aq
    BKmaoaJVwCfsRHlpAjvAI8OAKtbvyMAmwNMOM2LgwCUnXFvv2WTAKfgRXecpY8AocMUAueBU
    wCSt0UDmsOrAJA7RZigs3cApn3CILqKtwClOnSs6nMLAJWMw8Q1KbsAPdkbOunN4wCXOAI5Z
    qCrAJfSP8g+7scAkihL9ETw2wCFbg5po1TLALVg5orJTV8AqdWaSgKHpwCeDDJcT1TLAH7NA
    A2DJLMAleXPMMyGTwC0m5rkBQFjAHxgl0wdN1MAlubwbInCKwCSAx0isEJXALGb+sCUxvMAi
    lWrwZk4MwBtt/8zUMNjAJRccyzmvSMAegLvK/ELIwCUo2EVY2ybAI3jrjCf10MAky/x8C/Yw
    wCM07ikVYebAJiSuNSi3T8Am2VDqUkNzwCIaNw6wQZjAKD9IlG0iSsAjgJHJBSwdwCJ8Am76
    ovzAHgP7LRijKMAibEI+8WkywCclHNXK4vTAGdX4JP56UMAoQALhBk8OwChzBWV0jlbAJEMb
    WhhXH8AmjO+NDs61wBulkbXwV9jAIwCj7sP1tMAljIcbg0wawCbc5u0ZSfzAFcULliM7NMAZ
    5IGsKjx0wCPWwoLtbETAIaXekuzaHsApJFnmWUt1wCUfao3Hbb7AJXvPvI3wrcAmPEc7LBOu
    wCA146Iex6DAIPcT7lknYMAlXEtBGVVIwCkphv62NaDAJamkabjhNMAjtyrRIPPxwCjqE43F
    f/vAIssJ42F6BMAlFJU0qg/pwBacjo9bfjjAFoApUZT/VMAokqw/Uh0swBpugF+xIyzAI//a
    3LLB+cAig0iOyhJiwCZ1VpsSPEjAIlfO0xueNMAjeXqMkP1owCc/r5dGEgPAIrEihr9gjsAh
    f9jNYOF4wCFZnxM/VqrAHN66QdOukMAnA6pMNbtlwCYlxFmRm/TAJsKpI7niFcAV5nj8PGTw
    wCpFnd+/sUzAJ+CHskmMCcAgceQUqhNkwCdCgU33ZS7ALebcwbTBNcAld8dlxXMbwCVzQaKK
    NmTAIWJsGqn29MAsUWnSXFNtwBqiX1L0QYDALe+Cq9rIM8AjBWCU+vg4wCWO+IfdTd/AIu9s
    80v1psAhWyn1om5qwCI+KnnPS9rAIrUZjC2ZGsAugUTlgOVnwCMeSGTdM+jAJoSh7iGymMAk
    6Q+0yf42wCNBP3Bg8aLAKY+uquRMe8AmkHnw2SpIwCCU1YYZ7JbAJDY5EGEoMcAdDUfOeeZU
    wCNg7vZfWJ3AI0Xx16Z+YMAitUZE+o0GwCGX4JAKzRbAJZXKc0R6NMAhlUcd4OiWwCLWI3MB
    xOzAKR9ySpdokMAiWaoXwAggwChBVoqJnTXAHdloH9n1sMAmbMafYxSswCYqo7frHNDAJdXb
    reIXWMAkQQhZ+v2kwCJ6cRVvTbTAI/ihkYZ5V8AkvYbNj3BhwCWUM0Q0RE7AKNuDiRkkRMAm
    TQ8la50pwCb8M8z+FCPAKgKWGOZMZsAb8Kd431yQwCq1IDQ2d+zAIYwJ/gqYCMAtqbNbMNCV
    wCrnylPkn9PAJsoqF312lcAmdZ4CWQytwC/zrzL5OO7AMRUwd6pMeMAl1J/fYbfawCtDV76n
    jU7ALjOxewcI58AyHGs+ajz+wCncuRfwWsjAJhT0TlmJ9cAs8P6FhSS+wC72FuHHqZ7AKHhf
    M+94BsAr8OAeAwvawCSAOy4XkCfAM4/xWvfSKMAtK8eY8rPkwDHKFRnhXL/ALC9+U9qtfsAm
    aBpEFK38wDEYFPXbaWjAK9CRfNSAYcAphFP1sfx5wCyUiI1RZV3ALHtAlr1GQ8AtEU4qRbyS
    wDCbEXEyeT/AMKBn4mED2sAtegUtjQu+wCNc2citGqzAJ7sUscJt0cAs/9veXAAnwDBAzZKI
    t9XAMOcPn8+i1cAqF2B3pk8WwC5JgDQmVpLAK3CwLKeyk8AvlJD2sBJDwDCi7DGgyqvAKqo5
    swz5zMAqe52qXLaAwCqjmYwl93PAMNK8qvrXvsAuq3MlVKAOwCtZ225VuYbALIKAlXNfw8Aq
    0qdo6nICwCvUHdbDBb3AKB4YR74ye8AtS0Bhlm28wCrnIL+i1Z7ALd9uMeSUyMAv3NWReqDF
    wDF9fXfg0B7ALby/lnCI38Arpra39bzUwC2imKaro5LAMDGugHgz/sAtozeJpYkcwDEmsyQq
    EX/AL8AfY5NW7MAwiuedhvKzwDB1qf0nfPjAMAlXOQu5vcAwOnxpcas4wDDhzH5t1MnAKMpr
    KxvXEcAnTJysUAyEwCn5JF6H9JXAJZbTxffXYsAx5U7FJI3GwCDHSsgjK0jAK0Z6WtrKeMAt
    GJ5edI2rwC1+uKxvqyzAMHkWGju6JsAq2zjoNjpxwDAGXRSDiOrALNbuKTNXssAtd6jNU+xa
    wC8/tvx1TtfAKQ6d1M6FvMAqn3nUeIpYwDOw38ZPEDTALc/Xaek8fMArkdnUGoyewC8Biigo
    OvHALuaD0t9DT8At8KFRA9k7wCzh80snLSfAK+oS060qPcAwnlK+lJp6wCd8llinLmDAK8/U
    Z5pc/cAnD3xCqYkzwDBwGXFPs2HAMCk3A+7NjMAvpkcaO784wDMcjVod6zrAJ9WjYdnfkcAw
    YsGdepC6wCzgtlUb9mjAMV453mDlv8Ai6PgSi23MwCjHoJh1DNXAK9WOCTnG1sAq4AkaPd5s
    wCfuZua8cNbAKcInR12So8AkyhHBN14dwCpCmjmJSyPALMOdCnb3EcArTp320F/5wCiz9AKG
    4hrAKIiEgOcDNMAt/icoT4CYwCL1ztv1W3rALbdC7P4KksArF4r9gvtywCm+yIwQuhLAJNy8
    qFQRTMArr3YQEZsewCwoMPD29CbAKFBtO9b6OsApo2wxANcOwCg6p/dMZhDAKSu+tKFtrMAr
    58a4OXMewCeBXy8xmTfALLvLpnZEkMAiEu+IWpYEwCYnKdYG0VDAMAMZkCNkcMAt0h+OIOuu
    wCwzKTiSLKXAK7BtI1q1x8AvaezOWRGswC0JJ7TwDArALlPS54VcLsAs4zG/s5xEwC/UvQxJ
    Z0PAJjjiLAFluMAtmtPlw0pbwC5ssXNm8x3ALPcDvfW2RMApzoQ6yvqTwC6NQmyiMffAI3KW
    C/Y9lMAjGdNM8oKowCieVJYcMknAKmmHM8IbAsAnBDYfHU2SwC4+EFlXXl3AKi4IDOeUfsAw
    iI0g34n2wCqJUXAqjE3ALB+ftCmtesAtOA694QvMwDDCAnhhhlPAMbeUT8gZh8An7IBkKA8G
    wCpcKsrnv/jAKsgGnVqt/8Av3CmELiKuwC2afbGob3/AJwJf5vzk1cAsozaM/2d6wChvZ2Oq
    E2bALBgJsDBO8cAndw6lDpsSwDApmSY13I7AL3r92H+Ia8AvXL6ucIPcwC/kkFBjQQ/AMKRq
    R/FANsAtu7IDZriWwC5fCbXXrtfAEovwdH78QMAuf7vg2wggwDCxyLUJiNPAIK55YdrzWMAs
    LQk8LzaQwCyjltq6F/HAMbAvOfx5GMAoyrxUVBmCwBtLzbjzRRDAKmwP7UqIBMAxIpVyZP4Z
    wC00IfweVJLALL+hlLjBncApL9U8IsPlwCoFb692K0HAKX/ndk+FHsAwhodufdJcwDBV6xSi
    51jAKfQs1JXe6MAtXyBm+A7owCd4Wy7gORHAMCUh/GJ5NcAjfyIl/1UDwC86Wr0SWcrALL/E
    4SxAP8AqXq0fimrQwCtmhetX0+7ALei0RuLy98AwDFG73NqzwDHmQ8F+BvbAK8yeDo4sFsAq
    +WVCjI6VwCQvY041+f7AK4zqd+iKl8AuhesJBQSjwC+rM/Yj9q3AMiu1kq0CbsAsmXMexYi2
    wCgpXF7jJjvAMeTFCfZxwMAuT6DSoGe4wCzOeYVq7wLALQ3rJ8TiY8Au9DWEDw9DwC3QNb4+
    CerAKj9u6cfB0cAr063kYu5vwCl8hd4vnybAJc34jduNDMAu6YPnd+7gwCrsWXmTpqrAMRZy
    djH+PMAupvrjhmpmwCwUv4gzc0rAKD/CCtXlLcAvaUHz+Sh1wDNQUrXSbQbAIAz9mLY9rMAs
    tKmSM1KHwCQYQb6xoFvALRZ6UTmE18Ao+a4IY/SAwDBOQ+UEJsbAMA2vfJv80MApLtKU2+jy
    wC57+K3RNpnALSp1NkYcJMAwFTVWGq4EwC1wFie/zZPAMLX+sw13zsAsLrgr/sIcwDBXZx9v
    pBLAKxYZ9SWKYsAkEiLemVwvwCZ8GrN9DpjANB8jkBdmesAtXkJ/n1HNwCIMuzqmFsbAMbJI
    WqAjG8Al3cSGTHU9wC0HoS2TUfvAKGxlza3ldMAxa9znwY8mwCphjIrF8EvALQXNlW297MAt
    edQUGeRmwCotC1GRr37AKqczCUcIm8AmnUrMD/a6wCXjjglGzOXAL39vkntXqsAoWkiiuBkb
    wC0aeNhZeknAKlFGCGSs4cAtYhaqNuumwCYoh6qrAOvAMNTTgRBvKsAvAqMfoa/rwCw/fbci
    bDjAJHAgfRq7+8Aq1hiORzE9wCsT0jqpctTAKQaI/q5oqsAqSM2zVhZAwCGDNIfEpnDAMU84
    FFa/IMAs2i/r4Wy6wCpNy0wn0lvAMBQotmRwfcAujDk8arJawC7uGONGiIrALW0mLrP4asAw
    +xfaHeF7wCo+P5I3UxPALjPIHgAFxsApWPncWL6YwCeItqZWCO7ALmcBMNyH48ARL+wh4GZQ
    wCFWN4TArbzALjaSkpTVCMAoegeVbFWrwCVB9p127qLAMYPD44Ng3cAtB92E3obSwCxVbmcj
    qVrAJ5l/1exaRcAx+8DyqLnhwCie5fCOBvnAMJLIEbCtfMAlgES4iUlNwC2C2e5AM+/AJ2LJ
    zfEp38Ao4TH2Wn0awCLWS9oqPULALRsPBBR4zcAwm1uNZyQKwCjw0SYV2rjAMOHyMCroWsAq
    NsxINvPcwC67g+NR1tLAL1IUUUiV9MAotPmhJCnYwCz7AlMqO3DAJC9Am5zygMAsDw40pnmM
    wC5IsqprjwrAKfepTEGIH8AnHiSMuF44wCxLv4bD8rHALTayCzVmfcAjUCfvO8KIwDBCyuLP
    IOXAKdxw5gRwMsAx8XLeu8JuwCsh6ro+Q7DAMGEl8gA80sAp69Y5fohEwCkHnwwGpP3ALqLx
    nZGQmcAr0mfZpZ6nwCy3RlfgKGvAJtucb52+fcAro/pWSW/FwDGO4UD5UhPALX4yTHrwBMAs
    H8x+WXL9wBlb/AekanDAKCx0GrbFlMAsTddy6u40wC4qVypJ1knAMMweizNGBMAtl88lTvEk
    wC/CpfGfm3DAHYtJBGTvuMAqdPxkMD99wDHuNyg704TAKqJxRLjIhsAs+JMTkTIvwCp7xkYY
    82PALVEr0iSFZMAucRbRvjlywCwVRnK6yx7AMGmQV8o+1MAwcK87zCXSwCzWdxGrqwfAJ+mf
    oNmPrcAnOtui3RvEwCPXYQxKXCXALFJi/S6ofMAvdgLctp2kwCdVnrKtzBHALhu07zikfsAo
    0ZCIKzcTwCbYBltBzEzAMH5t+x64eMAhI/+RpWSYwC/S+ShZtMDAJkmyyu5+P8AppnACFN5+
    wCwv8XfRVj7AKHx/wSUgecAuA1p8q430wCew1KwxIK7AKhg0HyfOAcAtRusdQSd6wC0JVFuG
    oODAMUgUead2PMAtbDGvX7NawCdPQaiRc2nALQoDi1QwcsAqj2Y3iWSnwDCegKQo61nALZ8b
    1je0rMAmeCEpqKZ1wCdBLV8t9eLAMSgO0gf/dsAyA1Ss/b6BwC1ScmArQLbAKcXRn6WYjcAc
    y9lgvqOUwCS+8L4AinbAKj+s8w/jEMApXTqO5RQCwDJjJYVbUSnALPN+VU6x9cAnDAWpyhmi
    wC8PlvTyPf7AL42idVWZgMAnQqskV1g7wC0JaRKAa/7ALS6hT5y0F8AvBPcz6Ms8wDH0mYGI
    JWTALLDi3iW8dcAvteXnFlsowDIzVlVDWpvAMBqjfdEdB8AwdzDpsa/wwDF/ocJ8duTAMROs
    rPlZNMAtTi6tM1PNwCdl/stcXVDALWgeCF9cdMAviK5cvYefwDCJS2DwsUzALksnIgc6CMAv
    O6nlynRSwDB2AP/aSDrALFIcz9nLd8AuvAdEacXYwCz7X1eVU6vAI7gPrEkFqcArjtVTx3XM
    wCzUSj/6NzTALjR6T+KFWMAwfK1wrALwwCxwFzyVZx7ALAp/q8UABMAs1tTi2x/1wDJbRfLp
    4gjALWk1QJkBYsApMHAWwmzCwCX/g8eRRVDAL9gARkCD28AxjSOvPrKSwCrN1+8U0ebAMJfb
    VL4i9MApcKqb6hkswDJljBzvTvbAKFiOhrqV6cAxGijt8KS1wCpq7LoRwjbAJK9TM7NzqsAu
    qYUWjv3+wB2UETckQqjAKpPg9lRd9cAqFYpYhsQbwCS8D7UZH4vAK6NAf0fuvsAsqcr6wlKM
    wCTRbXrnf6HAM8hv5XRAoMAwyl2OZa9swCxMrTffckHALdakWo/7ncAs7bXUzenEwC9ib3lW
    ITDAKCzzbRys/cAs5eOlJF0BwCueJoLNo1PALthBRILFzsAxKcl4p96/wCeYbWvLUsDALhrL
    2atYFMAqLWgQZAEcwC+vs5YMOtbALHWFnFGZyMAq09MgsYx0wCtsCgoRusPAHtbo5s9boMAo
    fmpHA/P5wCvj9P72SuTAIhyhLPqd6MArd3m0fq7cwCpmiulZXozAJ3aSkDnJocAsMSvUBmes
    wCmlK/VgzyXALqkrgWuh6MAorDJ58KtUwC0KSC8kHvnALj+fcTdbm8AwDz3iky02wCKMqENA
    NnbAL05JZIQWrMAtw4JUsHeRwC9EGd3hLzvAKr67PxSqisATj1578rHowCw082/9+u3AKSJs
    p8OEtcAqGc+u9WAlwC6tZNWK7KbALJWn7PJqxcAoLyv3tWo/wCwgrUHr3u7AMVunlhh4uMAy
    0t3NwcxawC46ohrZfqzALla0avZAMMAuqKGNt2zvwCp3rReyv5LALTmVT63pFsAvUA8fw1R0
    wC7exW4/ozPALCjhzCWxEcAqNhjwXMDcwCtHMKpOvlTALlkpG6mY2cAtZ7BNTrHqwCuFSa+0
    DXTAKH0FZEqMFsArBjsdFBw6wDDC6sEM2P7AL50rtiJiisAuzJq1DpwEwC/MLOYuyQLAJ++s
    0IHQQcAqu83nBiYkwCjlY9MoBX/AMAg3OmdaF8AlgXtBgI0qwCemUlviY/XAK0Nw7eitkcAx
    UqMMiIrWwC9F6amM8F7AL8YEX98dGMApWf9Y2CDdwC4IyQ8A/kfALHYXG/avhsAvnd85dmKF
    wCxO/V/dCefAK6fQhXHmNsAlftj41xUpwCXDalBKFmbAKov7PKmxwsAwCkFw4cSBwCo46ek3
    znnAK2gRIo8q7MAor6ndM3FvwCqWotocFKzAMT21PXYEdsAnfkmSB82ywC1JhaW2BqjALQTc
    TWR3KMAtpmOtO2qHwCqciJncZCLAJ2Lqja1qYMAwXBrWpfCDwCyKTHYYXBDAKS/Hsl+TxcAv
    2SLVnGCZwCjL1FfIM6DAKpMOAzHiosAnA+n2tFF6wCf4IhbTvqXAJid0lVEfmMAtC47Orenv
    wCuEa6Mzxe3AKG7VJ9Fi+sAmBZ0IrmdewDFXimhdX07ALWv9UOK6acAnLawgi/z4wCq4VetU
    AiLAL+2ESY0UrcAkz1YuU+48wCx+AZwHIl7AKsXO/Z1PJ8AthRipNmolwCtzjptq5C/AKFpP
    wwPtEsAp+W0n214VwCshBQVcacXAKM5krp+bXMAnuyDIhFA/wCr4+8G6x2PALZIBNvsVmsAs
    WQpS2xHawCwjyP2jA1LAJjfurm2P+sAqX/l/8L4vwC8UTLPB+yfAKbFkhFZ5m8AtTgm93Q4X
    wCqvUDw71o3ALLs8Tw5Gj8AwW0jy2PgwwC5ySFYQgKDAKi1Qwur4BcAmhiSsEE4awC29teZm
    DkfAL6cBVxvdTcArF5iu5820wC9tHLKHnTLALjwCqY2ZdMAqrCnAX+zRwDBTtFhBeP7AKCsc
    hd61yMApUAo4+EfAwDBWy2Lu7pzAKptOZysj2cAtru+H5i8swC+UZ8YN7eDAK2/DZvZbU8Aq
    dpvg0oItwC0jDn71yDTAJa51NH/+08ApmyncBwC4wClhvytdp13AMPP1zna5OMAxBTZKWZUU
    wDGV1cktLh3AKY1pHRKKDsAqr5rl69ErwC0XWBxOjPXAMDpouuMCycAveeC4qFZYwDCJRJMg
    eS7ALY2ZQGkfvsAw6BrSY1K8wDFnpwh/OVzALyvYLywQ+sAqNDg5HSppwCnHFBO2Zc3AMGhz
    tAHnIMArh9xE1ZVUwCYKIZtae/zALeNc16XKgsAsU7rTYXaBwCoPWVfL5wfAKc19q7i/CsAv
    5EQaFM8qwCyfUhzdvSPALrNLljwCVcAvjH3ZfBpTwC5xu2NL5FzAJg9Dl59EFMAkokh3rZQG
    wDCGXPkY5OvALmhK1GzeHMAuUJ5p2NfwwDCFLxgh7uTAL4aDHOU71MArQyxCieplwCmboso3
    OMXALnbbwGCIQcAiaJIZdat8wCyzSqlR9frAMIKaTdfLOMAl2tjZsSoAwCegxOuM/wLAK5ci
    liL9Q8Akzps5UGwrwDDe9O+ggAzAMQqFFL8I78AsPBo3V6U2wCrgyJhQbkvAKOL2n5ncdcAo
    2pBZtN9bwDBHjVwqFo/ALUUHb9JP3MAmaUUGRonxwCTgsya/XcjALrkVKOn8U8ArjvYFEnHD
    wDHmtk0AqLXALOE+MnuqPsAwyhI0v+62wC4AXAEio9/ALB89j9DhycAuBsy64uvwwCe9pxSA
    T93AMy9aP1ilFsAsXjkd6+0WwCpXt9+KFQLALqXW7+bkM8AvK/cbO6g/wCvBJc4PzCjAKDVl
    a3xoAMAqa2Zpt0O9wCZNrGKZ4wTAMBCqp2uuusAgiPNht9nswDFLvAoT3BTAMFDH428CbMAv
    4Hgr93ShwC5QZQVdQd3AJ0cuHWYt/sAoanQRS5TUwDC6WYdDP87ALA2d1OWWC8AjbAecJPZT
    wC3ItYGo8p3AJpx6P8n1nsAs3CM2hcjPwDNjYhc7IlPAKUhRAAWQtsAzKZxUSslnwC+yzlLH
    xx3AKKpPbTpymsArhmHgFFrrwCWQCxjZ+LPALgYxsZlrpMAw8vnEgeEJwC2Wd9OFfZbAMnmT
    YVI8GsAvZBwGK82TwC9Utg+A6xLAMEfBKbNaHsAtvss1HElzwDFsaICLhm3AIOhZmrWnusAs
    9sQt+qEQwCmVcK3PqqbAKaY2mQzQ5cAmuTX2c367wDBmtUOzV3/AJkzFGw4QKMAsCCdNGN5Y
    wCzNCir/31vAKa+92CM7vsAlcmQFVD1MwCQoIkMh2uTAKL2bR0Hto8AyxELDc4PqwCisLHJ/
    OIbAKxo0GOBQB8AxAFPAo4aywDCZ+4penDjALtYPNbW478AEVHBv0CUIwC5IB0uxxyLAL6Gu
    NnRNUsApztf0gqMOwCw1ws1TBnvAKEG1LZv6ssAwcyQJPjLGwCVhsc/G8SXALLWF8QtRUsAu
    GchyGPppwC0AvjL5yRzAKdPhUMKN7MAmpQXKodTxwCwpPZVlGCHALmkNhBR1+cAuEk9rk+5N
    wCSID8xTyqDAMNRGjyzb/MAoohSSOP8mwC5IUvjC0j/AKu/MWdSivMAwyieX7zY6wC1/OdNY
    vA7AIXWocBqAUMAsLISoqDw8wDA8aFeTEkHALna2AMOTi8Atc7tOtPLSwCJJys9w8zTALwAf
    wCDeq8AqOca5EaGqwCnrGleUfXfAK7HXk15my8Ark2GZXMh8wCJiQqq6jvTAJ8x52nXbE8Ar
    1X8j8z/0wC6A1ounaEfALaqHvR8TvMAnL0W+FRXYwC3p0vHb1y/AJ7SOGmCfIcAq5Cpx0qSU
    wCi3VDUgi+3AMKTwtETQYMApUX66SVOswCWlBh5Y4GnAKxghowqoAcAts4FSe7IowC0YwHcq
    o17ALucfYqVXvsAryxot7DybwCcn8iu8dh3ALR1PNnni5sAnzGntonqAwCghmtDw4HDALVRD
    DHqQVsAogptst/lrwCkMWIVmi7fAMuteM4P0XcAwH/2PAOx1wC1erGA8ZSTALIXQ5gLmWsAn
    hUvbXgaHwC/y0H7irTbAK7siaE3LpsAnBKOwtoA8wCvQiVJDw8vAK2he1UOZAcAzd8uFTDcf
    wCsR29EoOfDALCvXoAbwFMAq52it7OFuwDHmgaxkWK3AJtnnOEwzFMAtZXET/gaMwDA9jFPr
    cQTALtRVfV8mncAtriBgUbu4wCUK9oPBiH7AL1wWR1pj7cAw4udpTwU7wC+dYWJ4ztTAJgq8
    AzMzAsAqXwU/aUU8wCwGPEgx7M/AKorOkA+0EMAxhXJhfV3cwCyhJhGc9WvALNCOK/M8XcAp
    j5lS8Pl1wCho8B0FfRTAJU/gX2GRLcAq+RYgvBc5wC5s0llyJ1TAJ3BOasYYesAwDJeljVy8
    wC6EVXu4JKDALyYTHqWa2cAxfO4KR/evwCyHYWSO2bfALhNds0augMAvw6SucHYUwDA7bdee
    bAjALIodoUZx2cArbohpxzfowCq8Wim8OsfAK4Wyh1TaFcApQJm6coLIwC2zE0r+hoPAJ7lB
    hIoMAcAshaW8BRTHwCxmPPKlVH3AK96cIukahsAshc3At7z2wDD0JQRqcWnALHuttAGglMAw
    BXXg5SaewCwx2NQE1f/AMZNP2XlH6sAq/3oJKz7MwCg31ecclYPALPw0bqZ5FMAyjRtT2hEO
    wC8+iREdvXDAKTZ8gfI6/8AvnACzNcf8wDFBkXLseFzAJJax2glgdsApwMCHTCIZwCImsQJr
    HYzALOWsP4LZzMAtDvpH32VBwC+wxDxOzgfALtNqdcyAc8AuC74+tjUxwCcziqujF6XALmBL
    UdFyT8Ani70yG2fOwCtf3zuPPc7ALEcAQ0N4j8AxrMKYsm/SwDIVRY9KW07ALWV6AmtQkMAs
    bWuLwTd2wCpoxZw+KuHAKJl+Y9cxrMAm4uvXh39OwDD0e018+LnAKVUkxoxaR8AqY5KJSE3q
    wC3pWNP+ZDbALBWhqlhmaMAnF0Y/VjzcwDDCw0qWEBLAJcaOFWdDrsAsNT7YfTP6wCXK98KS
    Ur3AMaDFBAELkMApazwODF6VwCws2PE+0izALKiI/pvf08AphV4VKJ+4wCxE4OkyE2HAK1uA
    qd3PR8Ax/ZekxB0EwC1l34m8mULAMCPbM3oXK8ArwK/lu6NqwC/xX5JQIFjAK75kmFLw5cAs
    XNpRgpYywCfOg8vFpYTAMAUQaQQ29MAh3KMvBbniwDA8KnJ2PY7AMpJfUQYizMArq17OcMRs
    wDFRBfaHuHbALEct11/LaMAzeGYh2zy0wDD/GBa8XTrALCs8SBd5ZcArkSk36qUcwCucQOJE
    CZLALOKuv+I2asAtdv8s/227wCrybbBDzeXAKetJ6qvITcAr0sKc73SQwCZjkHCYR+/AL1HR
    k/GXMsAoSz7PnopKwDDoNQKYU+7AKqgjBmde5sAuRHsvbA1SwC10R+TYvKrAK2z0WTYifsAt
    9kEKEeHewDAMsEnV5InAKAbqqALRrcAqTjHRBFvYwCSrb7cPpDXALDIIc8qoJ8ApbNDziF20
    wDEgIEhvSDnAK+N/PSsU+8AsO9YNHXR9wDEXGa/8HVjALHl2EmDXBsAioutfidncwCdmyJoh
    1nzAKPo++ojY4cAwvAWvRs0ewC5vJZF3JwfAMOqd8596hsAvPleew4nWwC+s8aMCbb/AMOVh
    s5vPc8As0o9ZsAL7wC4VKke84yDAK0uw1npiuMAu4LWkOw/lwCgP/Alipv7AKd+0sWTSg8Ay
    jsUB1VS/wCQxhWrGfW/AH2hEBWdGTMAuLba3f40GwC10K1lN5iHAMERWTpu42MAoVFf7ThgG
    wC4LzxaBM6vALH6WeDFLeMAjNUpeAj42wCwEgzW17fLAJOfJthnL9cAqSomO73yNwCt6e87N
    XWLAL++6LOPiP8Axzu7UoKLowDJZx28ArXjALX4qy2p3iMApSw1WhuokwCzbM/G/HL7AKD/i
    MAUKlMAxMSfjum3XwC+VmfSdE4TAKjFy5E44GsAoXykMryCGwDIH19YUKdLAKIv9ZS+/WsAr
    a5e1STVZwC6hWR900vbALS6lLPdbgsAwvzb7rDZEwCtyA98aGGDAMHFHetHeCsAwXKO4il5S
    wCyYiiiExXvALDWXoNrgfMApxbGMLLG5wC5WgWWJSIzAKUcuiq89X8ArLHWjYzygwCrKn+qq
    eN3AK7zNhKaoOMAyLwNVK9m6wCp3TgPdWsrAK5fwizsIAsAt1Msjzu7FwC/6WhOiEgDALuaJ
    NQ0l7MAjrUe/BkaXwCjzN27ET1nAMJz3OPfYucAqi3O35nfJwDAlVcF0NmXAKgaNdwreBMAz
    OUsDuC1gwCrAKklMZ/bAMC0KPq6KMsAxWq+X3bRgwC/QCiyVASrAMUGDcagXfMAwmtwarENd
    wCub9nIVPRjAKIzxiYOR6cAkolPX0ugbwCakqBpxNJTAK6ouPHo5ZsArLC9kSHndwDCrv/9K
    7xXAJrkhAKcm48AouUSCYdAawDCRl+J0vJHAL2I+kajLxMApQN9wjOjSwClH4dsKOjHAKaZo
    TH9SVMAreYDbVArJwCvmEL/FkDvAMAGHRAA9msAp82gUT0ODwCSZ5156arrAKsriDd6RRsAu
    aLqQE1Y9wCJROYI8S8rALCm/jonDNcAqeJuQocfnwCkzzM45JJbALXI2GPvdZMAxlHIDW5bS
    wC166DStZyzALJWaL08yKsAwZam7/8X8wCk8jSkGNxHAKHo0rDW8z8AkR1/7TbeBwDCKinUc
    kKTAMU8Aoh9qqsAsz0i8uPRgwDBKS2kLtbbAKEdt5fYOK8Ap758Q1x1DwCmcG6UQ8ZrAK6Os
    qB3TpMAhRLmcwkrwwDDFc6vKmlDALLDKvW6rUsApoKVf78cuwCYfQ+HUn+TAJnbAZSbLdcAw
    HyZxqDIewCy4i55xC8nALT+5T17eHcAwIl9K2d9wwCShmmBQ+ObAMR/8u81/3sAkq+WJPB0S
    wDL9KcBCZmfALpmLNHHxaMAx6wb5Tf+0wDBBwD+1MwTAMEyN3oSZYMAvU3hAkWKCwClaGTpI
    vzvAKQ22++eR+cAgiegccNEcwDBBUlfcN7zALo/uHtnYWsAxEZ4CJykdwCZiabhOUWvAIP6o
    DBW0AMAoCiD81neXwCmv+kAmKqbAKpk4MkmMXsAsuGlh2ul8wCcwxMwo2IvAIVk43izeRMAw
    uuX3CJaEwDHPT0wOsV/AK/0FIm1+pMAvjqWGkBkDwDAFOxm49ZDALhZO+0ENPMAtmqWS/QrY
    wBwaRoJS+JDAMUxTdqwyMcAwKbHiaykywCl7MeuN51jAL0klf9wmg8Aw99LWOw2awCRE3jHG
    nAvALmaAvqVvOsAoCfz+Q7kzwCw55JJ3jSHAJKkV089CDcAuGLDJpqewwDCdwwMgQr7ALRJ7
    qUQAdcAt2NCDXNPEwC9unk0T+bfAJ9u/G3BGjMAwHDS5Ie90wCagpx6ElF3AJIgJy5B28MAp
    fQrFQPWDwCb2Z32eBE3AMeBy01K0zsAnY1pMt/XWwCWbGDKOqQLALoeOtIxMScAuQzKSyW12
    wDHYSDI1jzrALpSiZrkb18AsiY4swUahwDNNqqc1wzTAI/5q6nvTCsAtjwgKiosKwClSfURg
    jiLAMO6+eEzjYMAuTx0ai8XGwC0tmmMTLr7AKaefslCK6cAvGBAJxr6HwCML5+ivzOTAKVS/
    kYJjwsAxU/mJWxmuwC5uK34QprHAHE0cl9OskMAwIhFOs7vvwCek1a8rCFHALMJUMBJmeMAq
    PlJBYhJmwCk+PidFLanALhX9yjcCmMAo8/ti66RXwDELvr165tHALgETKjj+FsAuefSRXmba
    wCxTFIkr3XvAKwNe6Af3YsAIMa0j0eYQwCbzOd5GX9nALPyJg+g48MAx0NkArcKLwDEHQYAJ
    TjvAKMOzgoEoNMAuorR08EaSwCzbR96hF9jAK6lTvY46MMAwjC1jjM2wwC+wFPY9XcPAJkyO
    r98UKsAuLi3rBu1hwC10tsZddtDAKjIg/uAom8Aq0N2vuO5kwDAKnpu3sw3AKkNBpWDJMcAh
    6LYJ59CwwCj8H9hLHm7ALaPWJufDYsAu5an1aM+gwC579UjAlM3AKUbt2IoMAMAvgI/X91Xp
    wC392AUGdujALHC2FVgPfcAqfUfvZ68WwCyNWjLFtnfALAuUOdsS8MAsAYne+MKDwC8PRyzH
    SwzALX5MMpJh18AqS8v4/7dgwC9ydATrv8TALIZ39XTlpcAw6IxAWh67wDG0BGYkfu7ALlWH
    ZJOpjsAvqBCJ9ceCwC5rop7VrQ7ALxhNzGlb88AwzKyYXJmgwC4erKrVVSjALx/Yj/P/HMAZ
    x0itpWfkwCzLvJmaLXjAKSxROWhwr8ArQ8t/t37nwDHk2eOkEH7ALHHJPGauE8Aq/6+97l/D
    wCoeDWEOpjvAKb89rHLEJ8AsOY0Vg9nzwCgmkjmgiFPAMDVRjDp3QcAuH1xS219gwCxeELKs
    267AKBKm1g3gjcAp7NKKODjCwDAwp/wPUiLALiRb2aA0+8AtslvjE9rFwC1+fXVZXIDAMA8d
    njLWFsAqSLwENFcTwCyC8uhaRa3AK364mMaZHsAoWFdgkv+awC78jgKLoyDAKLkMUkQJG8Aw
    MbFvWn6QwC/2+iA6jsLAJkqogH1w/8Al1NgIoUYEwCfXKASJV47ALAJ1iU/YUMArdrAgre1t
    wDNPlKaXemDALsmyl3uc7sAp7ko0pXHEwCjDIXucs6HAKzYWIdXAicAwMbxgKq/TwDM6NL/J
    aaLAL6xiRYTbRcAqGvGQiOocwCnvZu2NJCLALM4Zw4A8XMApp2lGP+WzwDCEHCVj5o7ALFZD
    pNpbzMAr0M4j2PU8wCca1GauNKfAManFkTQmgMAtVp9n9WnhwDC/RLc6Ud/ALKClA2pivMAt
    6ywDV4/ywCgNOCuYWnDAK1ewNQ1qY8Am5yro/jE7wC2L0+xKQDjAL1SA6gz1w8AtYHm774Oh
    wCx5UEvwAkvALpmX24V/Z8AwDIsp6qv4wDAJaGTBv77ALCTuPmLcHMApZhlLW6s4wCWJLyng
    +MnAK3y+tmVXCsAqy89UWf58wCzj9a6sjcHALca7m855wMAl4BZc+qiGwC6llBI/FcHALgVU
    6tlbNMAyUXML4f7CwCPdH2wzELDALxZ2T/QmK8AjADkUVIRowCuSvcL4Xa3AKoR43xFPJsAs
    b+WLIA4rwC5jFCU/HkzAL/sVx3bywMAwPujAAQC2wCrmX7VveIHALA41brictcAoctqxy37C
    wCRbCDNOggfAKm6eXCnyzcAyYYa9lmGgwChNfy/GCkTAKLg7KXwVvsArOUjG+5RewC8u8P4T
    SZTALMgF9GIpbcAwDyQQuPCQwCtR/GtZ08DAJV3y/l2a5MAwM0c78ExwwDEz/KzI9L7AI+ZN
    4ljAxcAppqNmwh0vwC/vQEywYcTAJ7znEHzkm8ApiJG7iD3BwCrgVcHne9HAKO5SE5hmYMAu
    owXDFqZDwCWpZNIailjAK9S9Ch3IYsArZfFMw6LPwDFdqWiMUh3ALQNr057APcAsA6dyJ569
    wDEk9vE4U5LAKK23bkgt2MAqM+4h01qjwC6FVxmO1bjAL2CfaQF6ycAvHc8gg2KKwCZ8xgoR
    xrXALbT5KWxVusAqNRHpIaUAwC9PZ7V0MWTAKxVQHz3W28AqoSkkfkTmwCJP64EQOmDAJKEM
    4odSZcAmhyLylvpTwC3G1B0i9jrAJqZRt86HpsAuspIaqGXywCf+mc7gh3LALXjmXqO16MAg
    7vXJAYDGwCqgQZkjGAnAKs2Kypq6+MAuLDhE8g1jwCalu2VgHm/AMKokvk+BwMAte8deNl1E
    wCj4Nf/qr9TAK/2qJ1e8x8AuDz+g5SUOwCga5yLieIzAKd3EQ2jy2sApKuLtqCpgwCnVB6ug
    bbPALewSNxIqAcArVwaa5zp/wCkht2NHHrvAKvQnmtusMsAoVdViOo/twCrQevYClSvAKb0W
    N28YbMAuXeyvMmcwwC1xjscEqA3AL2eZMyTVxcArfjGXXgNCwCyIX60AF1HAKS10dwngeMAq
    rw33DI6cwBgY5cTeuujAK7u+7d1N8sAtmqTJ+3CHwCJQCWtMMKbALWqaOEBXGMAr6vWrkxV1
    wDBKo4I1RabALrjWtQu5+cAsdzo1cHlkwDCmog6SCgTALqwQmHgKAMAmnnKeS1lGwCbINReb
    9F7AMtDqajNwoMAwasRu9PlbwCtW58IYRkzAMQu9QSxhB8AwGcmG3jqTwCsqeWgacwPAMAEP
    xSmqGMAw3NefNJeywC0AfT+zmDrAKXOeK9wOnMApIx/8zAVgwCuV4YjIQ9zAJhrgE72XDsAs
    fxFGT9edwCdc9e0zjBrAKurwii3iScAwm5XoFeYuwCn0uiPOL5HALZoK+h7utsAu/Yy+ROAq
    wC3sJsNwHWvAJyAL0Oa99MAmzhvluL0swCy0oiOZ2zjALDfZc9/OtcAtIu7ReyH2wCf0m8Zz
    lUPALD2RA9Lu7MAtXHILDrwmwCXQ8kB1ZZfALOC7wi24Z8ArUt4km1MKwC0L5G3tIU/AKU51
    bPCzWsAtLRbw6bozwC3+vNE/K5zAL479g2H0SMAseRuibokbwCfiSGL6bo7AMJKov+KBl8Ac
    D/PPDkgMwC2doDwzKn7AMb5EUQI+1sAw2N3GegXUwC/yay2cXyfAK/t+udrONsAxgbzv2BJE
    wCxmpBWBywHAJO+1zuZ/icAw0So/L/NiwDCxPHvifpTAKzIwhtGO1sAxTfFruYC1wCjcF5/v
    7AHAImnI4oaJXsAwulK14xQ6wC5kN9d5asjAI769ziyBdMAsU9EIeHW6wCxQl0O2H9rAKDP6
    XkNrHsArbJLniDu/wCvfjHl+kS/AMO9IHkKYacAsM/exlZAAwCXke/R0cz7AMIk02o1pAsAo
    mQ9bDq+7wDGnZ708u+LALsCcaPqEQsAsTzYWsUxmwCsbQZfYwQbAKLAHNU0Hd8At9o7slzeE
    wDAA+jLPcxHAMDxQT5kKf8AhKl6jF0BcwDDKm9DO6urAMBk3pJ1mVcAwuwhpzMVMwC5YecUh
    bGfAMQr+Q6ZkvMAvhMlLgKL1wDG/mmVJ5cjAMJ4JDGy7c8Ao3MBecdUrwCx5tocy6K7AJ5b5
    0oW+eMAo2bibqhCuwCjigOsHGHLALWYI4QYLS8AvD1gp8BKmwDCA3omNWYDAKuBwW16JB8Al
    xJLEAYO4wChYhqs1YpDALzWCPitmk8ArR/p3Fi7GwCmpiXM88ljALVeckiNqusAqYZW+XFhT
    wDArJDzretTALVkhBUaog8AkKba9MyJ2wCmO9W7Er2HALrINaFcQaMAq7QOil6SjwCr//yH+
    ogDAK1TWFBw2zMAtF2gFVYN5wC1r+DldOX/AJuym+gxp3sArp2GFgi4ZwC7D7ijHgEjAKhDA
    VtMmTMAs/V1tPpwgwCrfAo5HmX/AJwZw7luQSsAnQGsNMqc4wCyJO+4YkXjAML0S+E1erMAx
    0sflt5l6wCmF/0+K1E7AMSBEKb9Zr8Amryc0cYf7wCzdP/jZAnrAMILGp2ZgBsAyOyujxW4g
    wCY5V9hiJLHALK9ytXNP+cAp2V2Rgo+rwC+VrNhUmebAMBSsTOPo3cAwR1NyckPIwDD7o9fs
    mSXALMAOJYAiAsAqInfHJlBMwC2jWvM3NmDAKov8ou2k/8AlADnPeo0QwC/aEfV9z+HAMWID
    /dRP9sAoi+uikS6pwCbIShxUIf3AJQzKyiIca8Ak/ZvIp4kTwCoTlBCZcGfAKzuo28am7cAq
    fyqtOHkzwDAwNbm6Uw7ALEg2dlT8bcAvT2TtCyIbwCrCyruQQYTAMe8gTPgckMAsYhDe7ZmM
    wCI291sBwDbAL4zQmhqnyMAopSNoFpniwCo3ZUU4HazAKeqocpjfpcAwzL9CArxowDE/6pLt
    QkDAL68iQoE7RsArRVVKb0+gwCwah2U6MybAKgvzBGtpScAvxwY5psV2wC8+T/NIdA3AKkIH
    jVWpHsAsc+gBOCwUwCaZsnXZHiDAMWj0VWHoTMAo4i8yXJ9vwC1fYqrqkUDALlptHIQP/cAw
    dTSc4oiRwCb0dqQtF87AKyPvP3xXM8AsfWxutj1twCGPr0dGuz7ALlNvZK20qMAoAtjY63xx
    wC3x4JFEYq7AIfkQz/dRkMAvwd/Y+h0SwDHVzvgL8vrAKR5mQIeRFMAZ8uY8dnJIwCsbd6x0
    mH3AKfHfMsABksAw947L206QwCrcfk/qqCnALeKnKZ1aXMAm8WHQlC/7wCpYIFYV+SHAMHp9
    iZtT8sAqhgIUDNTtwC/KCsgy/bDAL469E42iw8Ax37Bz9YV5wCxlHVaVkjjAL3CoNEnhZsAx
    OjxhXNCjwDDvg8bhOJbAMEgKg1DPDcAuN/JNa+UZwCwQ5n+glzjAK7OzrQYmj8AvWOKi6wZU
    wDBwxSv2NhzALftRJrqSCsAn7Dp6G3XewC7ntvDAB+/AKWgLGhfmPsAqPNUyIxs4wCz/6Ume
    Cg7AK94vFqZpbMAntRp70gY3wClVWPus6fbAJ+3cC/wZiMAt5ouxt7rqwCgU2o/3J7bALHUi
    9U8TbsAtclP3tMM3wC0iBoJBOUbAK/K330UfqsAoo7alSg2UwCzJPVcGeMTALgOZWJnoL8Aw
    gJz4FEfUwCihMIwqg9jAK42gFH/phMAltcJasqr3wDDJEHGI8HvAMCzBwDLTLMAp7LsdBaBW
    wCq3OaEFMybAJJsQ7lpIG8Aloeex9iO7wDFG7yxifETAK/0wohGh/cAwTNvRoTCcwCs/8Mf7
    Vi7AKsXG9izvPMAxVByIPGrkwComIb3c7dzAKGyK99ndTcArcmSTCHenwC1s/BZkEwHAMJti
    RWICfcAwirHm4XJLwDE0VZybZjDAMHs14HnsFMAvsxPBwppGwCetA9iIhn/AL3MpTzdUzsAw
    mMThA06nwC2kIGhsi1jAL3uNKrRaasApHNwRT5huwCwSaMKXvDbAJyCjpKH12sAjPw5kU4Gt
    wCq8B8APTtzAKJ6xJLuy7MAnhhUvHZNfwCgR5jRiX+DAHrcS7XarbMAqcAMjB8N1wCwyqFoX
    oubAKfR0RD92usAo3wgi0MQZwClpl7hm80HAKSwp7YyAtsArwde+t88vwCnTDAjpOPzAMd/W
    woW8ZMArzliEtz2kwC77C0J0LdvAI5UE2AMt/cAtbMlbml0SwDAzDwXX/YDALPx/GGZbC8Aw
    ytRM9pm+wCnH8EgllLPAKk1AuPIKPMAroQ5aGRTFwCfkV8JUgaXAMAD5kpBHpsAu1Kj4KCEO
    wCtLGR2u9b7ALrKv/pFq4sAv5i/LeEpiwChBSyEzchXAKNszzeUaPsAkramdj4rvwCx8LJ9x
    qefAKdbOFzcSP8AwidwEjvOiwCm2cWX+XgHAMYA5Dy3KpcAxAtybJhPfwC3IeMQPpijALIiR
    S+FgtsAu/g6xn8vqwC9JDKRyN/HAK67W670sq8Asr+WGvyZQwCBuUE+YJebAKdaFsk/L2cAs
    avrr22l9wCozWBCdMM/ALGzs+JTkv8AwJdAoAWv5wCUEUhAou0zALxYorbNN7sApeym1XETx
    wC1PETpIqYjAJmjmap/gK8As2TIYo5lLwC6Fek00DtXAMSezg2LfrsAmRpKEt/AswC8rX/74
    ox7AMS5zSBYbfsAtqwm8rZHbwC8DxFFnMzrAK5kr1GmxbsAwKXRd/CCJwCsCXR9GjMnAL54f
    /rK/+8AuysGi7+cwwCq/xCjOSPfAJn++BvzVNsAwh6M540U6wC//FSqZA9bAKZ3e/7hM/cAv
    2KtQkGRywCjS8+TDYNXALZoN0G1wFsApA9vor4sswCevXTYwpm/AMDOApBtaosApqAdPYJkb
    wCw+Z9BRObjAI1KKOj2m6sAmxxVc8zh+wCmz/CvnTTjAK0MOJxMyE8Aqq8+RPuipwC8x3BN/
    16zAL0XcHhacc8Ats4hrdLBywCjiW5QRkDfAJcmBhROFx8AwVBIVxwb7wC76xOS4mpjAJUOK
    dOvQHMAlr9PxqO5OwCyCd3TEGJTALozON0DOyMAn333bG2TKwC7YvnPyV7TALgh9QA5/qMAv
    VmFQgMzawCS1GuipiHHAMtXnAmo2rsAoEwBbgnNgwDKIcY2pH2DALBWH9hcUqMAtkuZ74/9s
    wCoiMolI5wrALnFuq3RPx8AlzkuhUQqmwCQ0tlOhMjLAMD38V0dm58AlmK9QHz6wwCuBsuDw
    Ad7AK0AvgsSA/cAtd411y/zjwC3XSInsyxDAKUtj16d6A8ArJv3sG2HBwCwi45Tf/lLALZPr
    tgf+q8Aw+afA8V3fwCuW3UWQJ5jAKPB4EhLt2sArus6b1MKiwDCvENwc4MbALKCyk6sRcMAx
    dmuyXHx1wCo4swy19yvAKdxZLQI3bMAo2KfwSuDhwClps1hD6hjAJ8sw/bjl1sAsk34yNGtt
    wC318tH+xKTAJi2f6QyWuMAtaiW7OhpbwC5nWgn2+UPAK/mLtBPai8AxzS1wVMrtwCf7he8g
    GerALTS92jqCLcAyCcPlUUcRwCpDhFKTQprAKOPO9ruUZcAwJh8+YEgAwCwiDger6FnAJE1Z
    bL2URsAlTWfbU1CnwCkP2fUVkXnALrJ9FaNQDcAmf+TwOKaqwCssv44BNYbALrBIdA80EcAt
    4z7vl1m8wCqqavJDXu3AL0SsGDRDu8AthEnL7vEGwCyfT6XaA2TALVSBDop1gsAvu0y7s50u
    wCmjop8Xvd7AKBut3eV8wsAtinIs8oJzwCwaHwegl3rAKRD23RwhFcAjVVGIr9rOwC/gStJX
    IVDALDN/+Q5sUMAxzuJaUVYKwCbfJkKTvzLAKDPnmH62TsAmTMDYGE7FwC/DHNo9+ETALDlS
    gVHagMAtMbWfi/NCwCw1ukBOBj7AJdgFuKDcn8AxQ9ALiDKTwC2Qcibp5QLALwHQCi5l9sAt
    6IpuIlhNwC0wn3U3ri/AJlqKlORPyMApyq0LhMv7wCZ55WPiufrAMXB7hr57ccAqDCVeQWyK
    wCY+AXJUpnjAL5Dnxm+9sMAq+wL/XCPNwCZkgEMNhJHAKwiOubscU8AsyUeLL0sowDDg5X1N
    DN7ALVfENnr5c8ApM7isyB6cwCx4GuqflJPALoE57fWB7cAwkU1Oaku5wCR4q2AkcTvAL68L
    d3Z9esAwK7+ewFk2wDCakDvtpW7ALKi41LBPncAsjy1qIbccwCxnVqcGF4/AKB5fn0XO68Ai
    sTVz1ppEwC1+95xdyNfAJXfrf/0Id8Ath3BjPmx4wDAgVS/rZ+LAL/KAA5WcNcAskbklozW7
    wCd1/JFB9E7AJ9G/rjlwcMApuiYQg1cmwCvFUjfb0GXAK74FFF+jPMAuuUOCOY/nwCmOkdVQ
    LjLALsmVpZyZbsAqGDLmqbZHwCt9dYMOBcbAL0jsGcM+s8AwGwTJ+nAOwC1YWtSKk9vAK6BR
    I6bynMAvPBamD6rzwDDyjuwB/p/AMX+k7SEbIMAwMBAPE389wCZVTeGCKZbAK+j3+6ixXcAn
    7myo6g7gwDG8iH2yBHDAKLXaGH7MycAvp9KST6+IwCxWpY1NPWDAMQfpHduEGcAqZNGUvaUb
    wCixXSZ6n57AJ68lJabRocAlXqKxyj9kwCnoOi+D+ZzAKVRHCLH+FsAuB491J2h8wC6Ptwdr
    0XfAKsy8kuourcAr+9+S/UfCwDIuHfufJ5jAKk6zL84QqsAr8cfuuuGawDCBu54XCsHAKgUV
    P1UwfsAsDwagnHETwCfrXeCi+nzAKUVUnKSjaMAiC5s2B8r6wDDeTpzr2nbALBXkQV6qh8Av
    6Tchp6NqwCzX2Y+DrKbALEO7s2HD+8AjyikZWzOuwCssNnVqZzXAMIRby75Mc8Ar6PDflBis
    wCpJHXFf0bPAJDXlBCqE1MAqo5IxJxUMwC/+JtnwPWvAMLmRODLh8MAnN06Ue2EawCwdzQDu
    aDfAKDirOZYiRMApkro/NK/2wCpT/hi+WWLALtD64VIKSMAugxzaUr7fwC72mQ36GR7AMB67
    E557J8Axul/GUzGrwC+Lzval9VTALEsfreg/OMApNAeVqZPzwCyjDOjzI1zALAGYagpkvcAq
    tzSXUKd5wCttnrc0h5jALVvr82ejDMAupn5HwF8nwCrKsu015TTAKmbxBYP/ssAm/BxCDIq+
    wCqwRLTvIkbAIPef5tqFNsAvSx+A2zziwCxJ0+QH3/DAJy5jeAzwNsApwL3dsmt6wCstMRnQ
    T2fAKnOsXbjiU8AvKnTyg/0fwDBOJaIHVbLAMDl7+pMcBsArFRDwhTmswCwCPnotb7rALeTn
    +UvM7sAo+UW3xWTtwC65z8WWXKLAL9fhi1VXq8AvqqrImb6kwCitbJ4DkWrALoauL6zasMAp
    yVSo/cA6wDDNYlR9/EPAK06puN4PisAott7rvyVHwCGgIQFH6MDALaHExKLvcMApwixz1mKo
    wCVpifHUBibAHr0SPTJB4MAwRHVn0OBNwC48tG7hT/nALRSv135cisAvDP8g6FOAwCql9J83
    cvjALXqLeD/1J8AsTlbu35Z8wCq/udC43KDAITrOf3Sx2MAq78TVXsd4wC4EqoDj/MTAL5rK
    Ah2RWMAaU+X5VKaYwCrGps91vdrALny9HXKALMAnGscN5JUbwC++MN0RV7/AJ4Pe4SqM8cAn
    A0qpXwK/wDM1sg7Cm57AK5zEV0wqzMAuUO4oZrgIwCxy1kx6X6/ALJpqt3VuXsArQDdTaExD
    wCaEwoLWRpsAAAAOAAAPoMAilgyZPR0gwCHHu8VA02jAIdslPnAZtMAkHm1aBQlSwCN3GZFH
    VVDAI4Y0ayEZWMAh5uRelDIqwCL4lOxP7wTAH5NlnXFawMAhq29x9cRUwCIX8uLAtHDAJMZE
    jMsa/MAgIimfc9/cwCJe5V4mi/zAIJVvoBuD2MAiFGdtp75mwCEg4ZsQthDAIaJB5TqL6MAi
    pNbogHyiwB+3FgLQhGDAHxnRZHe2UMAgRN0MmZyQwCIp+AHfuUbAI1F2KCYiIMAgrcuI5l6q
    wCUxxjEDTTjAH6xHpKp4YMAiOHqWRIPwwCNivOXs+iDAIcXSTJlAHMAj8WZ35jUQwCYbN1/P
    8PzAISPvAFPkEsAggm6YbXKAwCClTCDSs2zAI92nuvFQIsAhKapVCBxcwCAP3cjSYbDAIJNb
    PoHv0MAicjZCFQiIwCMhOMvkgeDAIYE+xB+8YMAgrf2BPDsUwCMS+hDcmYDAJEVmkDs5IsAj
    P0BlSpImwCF++XIzNLzAIoT4hwK6tMAgjnPQmZkwwCLL/PwF6RjAINNt7YhSzMAk+R1Fd5e2
    wCE25lcchk7AJTdPBIu8oMAggvpjq10uwCLSPuqjZ1rAJXUmSbGuxsAiInwaDPwgwCTygL7r
    817AIOH5+ef3oMAiYhKujk/wwCMDm0l6QjDAI6VsecBNDsAh7tLdnE+swCDfQ3J2UWjAIyWf
    +TkUpMAhwoA1dg5gwCOlTKTnNKjAItrWdfXYCsAiiPygp3jkwCCGMV85in7AJIhlJuxLpMAk
    LC9TaYaowCEkam2KGfjAIzwHm4cdBMAjaO6rG2sgwCV2vm8ok/jAI8X0q9CqHsAgyE4rpKgQ
    wCKeTm+G/QTAIHMxZ3vwCMAkuYEcw38gwCOYuitcJMbAJX7IB6+sBMAhkPpS80VOwCGNmr1a
    ZKDAJDk7I0UMaMAfZNoxZLuIwCNyCMuWxYDAIf5RqXs3RMAjQ0pMlg/owCGwJgnfUCbAIS8v
    VjY2kMAfHgj47FVcwB/JX62wGAjAI9WjYafkhsAmBj0eklJMwB/mmM38vkjAHqVj7Cd7iMAi
    ZMvx6nYAwCAOdjqfDQDAIm9eiueh5sAjsfS2eWmGwCCBF1jS7rbAIObvj7OlKsAkYpjcmlIk
    wCAhG3P0NJDAJpIscwmRVsAjIFZ5RLLmwCPUUk35TebAH7LzYGKPrMAjIv5ppGK6wCXq/E6X
    yAjAH/ZdW2j2qMAiZ1e5cZy0wCFHDm8ZTjTAH/AfJvNc0MAgWVvVl0F8wCM4nHSrw8LAIWQR
    C0MmdMAgRyc5/tUkwCGgMQUNhCrAHsBBTDL81MAhhuXbKXn0wCDtpPLGRYDAJCUwEZE5vsAh
    WyahEYYcwCPgKxiQH+7AI+hsw7UFKMAkbyiCDjnEwCEMn9PCtwDAJfp0XOdzhMAh3PUknD10
    wB/ta+5BV+jAH4EoIssGeMAlckDo/0UCwCUdnemoKtDAILYAdirMIMAft8G8+RLYwCFm/5In
    JKrAIYljX/hHwMAlYJJjU3nQwCHRU9RRxYTAIwCzL3o3BsAg8IqQDuJIwCDR+JJh1uzAJMc5
    HUJkdsAi1iQPohIAwCWqQ8PCH47AIldQ2nQ1aMAiaRFu695kwCExzcIfRvjAIG77GeSf+MAj
    1D+/zWO0wCF4SqFjjaTAIB9RxDLGnMAgkbVQ4TfQwCPOmaoWa7LAIB20X3fohsAiO2l90GDa
    wCEVkBw9+LTAITIK5rq2RMAkJesxikHSwCKdu/x1XWzAIilrx1KILMAgS+ytOjmWwCQa9rnW
    BPbAIwXGxB37rMAgD5cZAtwMwCEIunXZOgTAH3PrcyvmmMAfc6pUDjzowB8h8Tl+DPzAITyk
    5TI7TMAfms/57cHYwCGnmhsfb1DAJaM00iW/rMAgDbawW5nEwCG8YE3xFszAIfbnlRs7asAk
    Z/vAWHuowCEqTkRY8yrAIktNdyspqMAkc3O0dYaowCV1DJPHHarAJX2CSEGcAsAhiGEZcRNS
    wCPkFVD5oNTAJDbT/SbNyMAgnRZXvG4swCR5I8xsGqDAI/EmyTZCAMAfnmBugAwMwCGivHks
    9EjAIsBf8g2MYMAhe7kGI2UgwCNmp4V0cWDAIorj2FRl5MAfKMSiAqwwwCCJgweS6BzAIaL+
    JtIBPMAko70wxCI+wCUScLYoMazAIE26ahKb4MAh6rvJecjgwCJoO9B/PgTAJKLl9BjyVsAk
    v6+k4ZnGwCEkHZujpsTAIgzfG2sioMAfTnq2gH34wB+behvXj6jAIXvgUAO1HMAgzc0XDLwi
    wCFPc7jq3bzAIE3gaOsCmMAhv27mlaQswB7WZi/zYHjAI1znAgwbcMAjbMVRPyfQwB74xyKO
    puDAIrVun53h2MAfE5ndecNgwCNyW06/SOrAITlabpXpxMAhWtPn1JLuwCDLiWgH73zAJG6Q
    dCDQtsAiOK+iVnX0wCMaqlKVIPLAH6k8qQpb7MAfqJ/ZVj4YwCNbqWNP15DAIi8UsCiIksAk
    guBFXlSSwB8t3yvoAYDAI4TgTbUMEMAgGhvp9+wAwB+JQ5JoKNDAIsF35lRaasAkQCK1Tg5W
    wCHagONGTYzAIKV5u/ivZsAj35sZO8XwwCQ4V3Fxd/TAIraz6sUIjMAimDOP5M4kwCJRj3jW
    SKbAIxtdkGC/HMAfwByIb33AwCBZuKnHg3DAITBX6rLgkMAeuTDHNrDgwCMcSRly+6LAIUyK
    D1ES8MAjM6t9SQqowCM5FusGGAjAIJQ+I1KmkMAmUuBtZaEEwCMiraF8exzAIx0W7/0yDsAk
    KIUtKfZQwCQZ9bMMqcTAIeGj+6QU6sAkuydqPzfEwCFfdVhP60zAIJYArX9tWMAfPmTxGMro
    wCFdv9/9cxLAH0VHAHOpmMAgD83u7nFOwB+LJ8F0ZRDAIjj3imKtiMAfgPN0ZPXUwCLskxY3
    zHjAIg96jq1uEsAgE8kmJY/swCNMSO+B6SjAI/WZJr695sAkSDB9I/VawCCPDlF3BPLAIzo/
    6d52IsAg/jg0WHDwwCKg9HRfqtDAIWOOF9WREMAg/Gat1ttowCK6gQXf7c7AIl5zu3HvBMAg
    Aup4D0NswCEJ7sb+pNLAH5jq+j/YBMAlbm6+cl1mwCHc8wWjBl7AHy5pfaAIqMAk9Kj7/oL4
    wCAsv+STHnjAIt2WHm1DBMAidIgOGVBKwCDPGKW7XVTAIgdPrdGK9MAg9e2pmyhowCI8OuCx
    npLAIzQaRyElEMAjvMsjibSEwCKaZDv6nlbAI1FoLsIbLMAf5Mx9HQh4wCMj/b3z1r7AIsad
    jCm3/MAjhWg6utuUwCCUoSPJ57TAI3SBfrOKiMAjidTMcrp2wCFe1ekwk6DAHz79EyabKMAf
    IN+fU18YwCLn7c77o6rAIGFH6qBx4MAitjfTdlhQwCBUQnRyWhjAIx7LdCBh7sAeqJaR9b8A
    wCKbWIGyjGDAJHNB2edwDsAgOBnAuTgUwCIG4HLRwLDAIM5JcGTyhMAjb26letCEwCRakoSa
    errAJf85Uz+L+sAh/5pgBP/4wCL1HQDX/DTAIgD1X+xlVsAiJBG2umuSwCXuEMGtJn7AIayi
    gbYNDMAe7b8z34UAwCNvDcctNZzAIZQdDAb0YMAkI1ufZJCkwCCEswaMvMzAJaeljBN7YsAi
    OTVVNJmkwCRyCyjKidTAJJnFtRCn/MAhXd0Pw8vwwB/3oa0uW0TAIcPQgUTGjMAgC7wWdtAK
    wCGfrinm+ZDAH07A4/mFWMAiSemWHsGIwCMyw+tZoXTAJSQEczNexMAiqKZQ7bCGwCP+NSez
    dtTAI+odRNpiiMAjpDFag9IUwCEwAug4J2DAHpvfvUK9iMAiMDOiuIk2wCJg3GS3S1DAI/Io
    gv5XMMAfqrci5uzgwB+aJgY5tLjAIJaHsIkiXsAi6LCx+fzgwCOU6onGGBzAII/8TdghVMAk
    ICw82xPGwCJRGinhL5DAHzGR9zY6GMAhGBVNNQc8wB68l2qck4TAIzl2TQoWcMAhA1awPlo6
    wCI0y7D4yDDAIz+CMDAx+MAiliTCBBwQwCEX3W3xCBTAIhRsdLzzfMAgDaTj6nagwB/Z499v
    yTzAIh4nZtn2csAiOP6gxLzgwCWsnxK/1KDAJQIbTj5q6MAgVjLpqiS0wCNqeZ0yqiDAIOQU
    QLQBQsAgdnzT7YKywB7r16kkfVjAH+2D9JrsUMAkMnOGzpPywCSHpyYj9YTAI0pDSmvTBMAh
    Ezl6XTA8wCGjse7AmUjAIWrfXlQXEMAgfg2zO5aQwCPWingquWLAJEj24SpYjsAgBWGqPWa0
    wCL4WlO7y7jAIzYZJ0jZCsAgWHnPudbowCDFHwYlc1LAIgiTvYA3kMAkF9sEEckYwCNcdWnf
    cZTAI/uQ5Qm50MAhnSan4JPawCGv/7TB+kjAIoE5CHiMsMAgW/y9i1nMwB+2K4HMAijAIuIf
    oH1IpMAhprlwj/LQwCR1ipO5LsjAIsdKLv91jMAhlBcQ+8EWwCC6D96UYfjAI8II+B6IDsAf
    SNkrQ3zwwCHPAFYUr7jAINRePawtXMAhxq94LdmgwCU2zQGk+fjAIbzCu6KqNMAiiCSwGNIK
    wCFVnh16CtzAIMJwomsSwMAjdD8C7gQgwCBS9y23rULAIdMjuFQA8MAgwLQlyfHkwCEPPiE5
    uZzAHo6j61EXYMAkrnWgzqwAwB+LmFqa/AjAI58Vjytc6sAgJdyamIJ8wCQQ4ReMeDbAIJod
    l8i3XsAlfNWfekRqwCSqH+FSrGrAImPI3qimWMAmQRWMmnXWwCAdLM/F7xzAIfwB+1845MAh
    tunXZqRgwCOu4JYueDLAIr/3/SGLPMAlmIT1lI/WwCIvJhnVFhzAJP2cQEvTKsAfDBGv1bHc
    wCTAJsgKhj7AJHU597NozsAfKWOqXzK4wCTXC3aALADAH56Cjr7HSMAiJBK2dGkYwCGe/F9U
    XnjAIG7yVJK/lMAiLEoUKQ8swB6I1vYValTAH2K4TKkSkMAho5twG5wIwCLnWmiaH7DAH8fM
    jxFmGMAiuhBzBuVwwCSovkO4zVrAIm01mrML9MAif8XzrGYIwCLrEKv8MdzAIctqFwd9iMAj
    FIquZhcAwCHo3tvi24TAI76Y4yF7RMAg0KuGBB3owCKPK0HVpNTAJKt9I4sJXMAgeTr9wgds
    wCRsCk5hpILAHr6w3Y0+xMAkjOPLgUwewCJnjTd2nlTAICtMSUu7HsAfv5VgBHSAwCBCDUHh
    roDAI5PnfQvulMAituhExkYAwCNdjFWkIHjAJH4xCb8axsAl3igPf+zcwCF54nrhQGTAIoV+
    8KTkisAhTJTJOW5WwCBhSKdRI1TAIwi51idwNMAio4jU5I4wwCGCdHV/MdTAIu7jgU+4tsAh
    HpK8ProUwCHo7+dV26bAIhVHIJeyXMAgfXEQVQoEwCHQ7mS9LszAJCOPV2OiTMAgqAmdil5g
    wCGjdzHd6dTAIGXdBbwAjMAh7//gpVnswCRyJBf5BrjAJZvXbQobDsAjGr4QWBrSwCGHSxCu
    wF7AJE98DdKyTMAgXEyTiJ/2wCDeu/BZsHjAIbRqTJ/FyMAkPIHTeBSMwCIvXyVz48DAIIa3
    BvT57MAiDPAR23o0wCK5sy9xcujAH7ubGxM4gMAkt98Tb64SwCBmXoKs/qDAIv2DP2dNQMAi
    uGj5GNV8wCRph4NPyajAIe4mk37T3MAf9SbzF9okwCSWqaEhPUzAJNHsi55H0MAkBfzFV+QW
    wCD/n3Lwz3TAJKd79Ed60sAhu65l81JowCMiR6DlkQTAIlrAtXUIMsAfPpv2bRdgwCB2DEU8
    PxTAI7VUsImx4MAizYWRUc70wCBk8rGtytrAIBFwJKWMeMAkYbGsfByqwB7kOGxpw3DAICfC
    aBrpIsAjkK5a7PxuwCMUn58UuOjAIcRAXNVCasAma4ukMR88wCLZtOPLpTjAH5KMRRcdDMAk
    jjUWXxXYwCPf7rgwtcTAIfNo6UqPnMAeutOx2964wCNy0S6yL/TAIps5S+laJMAfqDtmj9fw
    wCUZ94oM9OrAIZ7ULOp3/MAi0Cx+ThpMwCINdKx3P3DAIgHuzhQdPsAgkquo6NC0wB7I5YkP
    GDDAIJeOf5AsQMAj1htMbOf6wCLyr1Df+ybAHu+w4ng3kMAgrUI/OfDEwCMmuXe2c5rAIcMa
    6KzoJMAgX8Gg1phowB8vmpWzWMjAIjwrzOlyzsAgvdWbO3pgwCLA2mKC34TAJMHrGQayFMAi
    UXUEyKyUwCFJKkx/X8zAIcDtYrvj1sAg4VZmXfQUwCJUsJ9m2eTAIBzCTWjqVsAkkXJc9Mp4
    wB7Y9ynTJPjAJObg9NEeesAflKPtdTvowCPX/yabhXjAIuFDHFcmmMAhgj3G97nIwCQGzG9i
    trLAIv5W2m2faMAflHSCBZXAwCMHV7q4MlzAIu5vEWBZvsAgw1C1jIaiwB8R7Tg3/qDAIuHI
    6XfDzMAiul+XRSP0wCGiSSCT6qDAIQbOT/JgkMAhXth7QXZiwB7TlrsoILDAIhwrfBP/9MAg
    W+fu847cwCPZjWW6BPrAIgpGR9sGbMAkMaoJbipywCDKU53PG4zAIA1WT8SamMAkHekkYR7E
    wCT2k/AQwk7AJN7FFtK06sAfxJtaU7PgwCFtngh2pJzAIoGHaMSxZsAk1veFJD1EwCQM1o4f
    mxzAHueBAyElmMAjNuqCPoy0wCTa28fJvozAIMK56sydnMAiQcXZ4+DewCOJJGzt+4TAIp4J
    3yqXwsAjGZFP/tRkwCC/4T+w8x7AI8pofKmN3MAhUprtBQZUwCIihLUOXTjAJXxo2zMlKMAf
    sb42mbtIwCOD6Tpa+57AJROHNR8EJsAiUcE8GRb4wB8UpLolDmTAJE+3OTLZcMAgDHpmm8TI
    wCMFhlQ46N7AH1X15VmhRMAhcekhas0gwCEBJIRg8nLAIAWmXmovzMAh7dwbYXwswB+uuNEm
    PajAIoY3WYhLCMAgllL1fW58wB7TA0mCx9DAIvGJt97QgsAkXPjJSwnswCSrOUDFyF7AJdxW
    oUHK3sAmVAj8BgYkwCU62RlDNC7AJWJufEpGYsAkCkJi7hbywCSMRALAVizAJQ75zIcBTsAh
    tBPHGIeOwB6x4DIzVGjAIZ2xBJHsZMAhbtv7beaMwCN5GxJ1UQbAIFOWiU+P7MAiMAXQAiAW
    wCK3u+KNdBjAIFSgn/pZ4MAjEgR4r1YUwB9Iig5WWujAIsbTDuCQOsAgz6ZwDwx4wCKA6Amf
    wfbAILg4gHKR7MAjBglNdg7wwCBZcIKgw4rAIyWicLKioMAfFgS1PDD0wCS/PYxB/V7AIQ6s
    LJzQzsAgVS5QP19qwCGNWdl8jX7AIdZPAhHBDMAmPJw91moswCALfdt3zn7AJOjJU0j6esAk
    fGFLaoquwCPol7xKYP7AIJC7C2qXiMAgorDBDwOMwCMxhyUktTjAJZxOH9d8nMAhRJQ8cqxU
    wCQnLvhuF/7AJU9mF8E85MAlABl+Kl3WwCJDzhcKCo7AIvIsyb1hAsAhH1LUvJCcwCH6GY+F
    BOrAH2VHyA4EUMAgqTBFh+WMwCI12rQ8NIDAJG/IJ/bg+sAhaSxVSs/4wCFCTFDEr+zAIZc/
    VAlUBsAg/KJd9Ig2wCQx7Y7vP4zAIXhIdnwYcMAisIekelT2wCDLJ2I5A1zAIbNNPPd4NMAk
    TkHuz5kYwCKdbnMpOnrAIEIqOsm/XsAi12PrQfi2wCMF7D0o6QjAI0Tqo2/10sAfRyY059iY
    wCLpdGNnbzjAH3fzYYarWMAkdVn7wHliwCNg2spAttzAJbp3nd8PFMAi06sI3rdKwCIECJZ6
    kFDAIrAqInDLuMAid7wKJKkowCOIHbX0yAbAJFvx9lpMOsAi5o1bX3H+wCEMT8cot2bAHz+5
    U/rgsMAjMbrpH9VswCEHUGqB8KDAIGsC4K78bMAiVBYr7N3uwCIWLEi4DKzAIrtP+F5HssAg
    BEBwjSF4wCErl/Wm0XzAJPLaHZ7aLsAgNf2melikwCK3gVVZ90LAI1mlyDy0bMAhOXSliDWS
    wCPJEqG+rVTAIEVT/N9jFMAgO3/gZK0wwCMlJikT28zAIP+3lIGnVMAfTa1wX+cYwCDIQ/11
    OuzAIh42ecqgpsAgB0oxIarUwCMfBFLjU2zAHzEFfApk9MAj+meKU9Q4wCGNBdpCtELAJiTE
    l0Z90MAj/i+foh0GwCD0kfcYCSDAIX2tDyIwLMAfmYYRA2KEwCHw+RAqHf7AI3T8/SmYlMAi
    7+4sZ+HYwCCfEN8bpNTAIvu6fnmtlMAkgCSL5FrAwCGpyKbDOnTAI1pZB4L8psAk209OT6nE
    wCCDrfxQ6tzAIvs/MIEQzsAkowYvK7HGwCU6h34bZZ7AIqeSy9O4+MAkFA2qIF70wB/jvx9c
    QCjAJI1fipWLnsAmajumKsgiwCAX3GhytmzAILFGs1eI4MAiUd7JuMokwCWdLBPf8irAH6IK
    TfeaOMAf9Rk2x+qYwCF0EbKS9xTAIV/gog0dFMAfczdDT1qowCPhyARCSdzAIdD6Z+yHvMAi
    OHhwZ+imwCERVPCEXMDAI5gXoN/sLsAgiVJBW1HywCIiIfIE+SjAILINcgImksAgaYvJUId8
    wCOT/gxLDzDAJXpUD54oSMAgPkSI2dU+wB7MfrawX8DAIcDE7uVTesAhJIrC6oW8wCN5aUHt
    u2TAIhnODYlpIsAiHdvYIkqAwCASqDqowQzAIFTSlEvlbsAhiagtbMNQwCD/VEn8/oTAHsPl
    i92ldMAhBKrLvj4kwCGZifCQ1CjAIwO3dsH2QsAis38Ug+C0wB73HiRRTrDAIhlqNCd3YMAg
    JNtB3BjUwB850asJ7bTAHwdeVeXo8MAi1NbKpXMkwCJBc9fcRkjAJGCd3RT1BsAgPw5Zr79A
    wCBs2oSUtRTAJKyAo28/dsAi4QX997lswCHpWdj9PkDAI6fQbcTbBMAeuDAODWdAwB9qlJcl
    isjAIIYxOYxSssAgYprQmhtwwB/lNWXAVGDAItXH0EuZ7MAg9qqkF/GgwCFwbXvWjpbAJIdz
    jFTxUsAkEkUM/oPawCHcMVLPcsbAIkiIMAsuZMAeu2l3fQtQwCMwioUlepjAJcbtH3mKZsAk
    XaO2enaIwCQK923/AazAIRpwkob7UMAiNNcuuoZ+wCAmvKmnPubAH1LEyyRHGMAjyfnEF6do
    wCcH3ivv8ezAIEXrqtYnzMAhV9szZK0EwCUSoKRXIijAJYUlPX0bUsAhTkGAmyQqwCUWBR9F
    YGbAISlFgboHAMAgXZgSJud0wCMgvss7G0jAINHx2PnnLMAhwFmj/guQwCBDqt6ug0rAJJ7e
    PQHuUsAk4eiFgRR6wCLdRqE+sPTAIHDhgwpz2MAlZ/NFJEX6wCHrSjVxieDAIbjvbDsurMAk
    crVSpRFawCNfWd0FUbrAIuD1o7pK3MAhxjHzgwlQwCOh0T6lzi7AIL/5/3UvAMAg85lieoyQ
    wCAH2khedkDAIqqBnSe+hMAf8UeaU1uwwCTNvzUtwZ7AIF+3q/63rMAgch/5snDIwCGonWy/
    nnjAJCxwhUk2YsAg8OmdBonWwB9mE6/A5PDAIIsVywVzBsAij7YEbOiEwCPO1JAGeMzAIPPQ
    Ia8+kMAjuOIo1+EMwCRfhy7F0cbAICy4OUaFNsAg7irorF8AwB6wTdNquQTAHpWQW6YJ2MAl
    23IMEEr+wCEvAfjWGSzAIITxgqNj6MAelpI803AcwCStNlJYGizAIDU05UCvsMAj+6gink6O
    wCMLEbdYVPTAJTFMRhH6xMAhEFBS697wwCOSH+45pPrAJGzC5br/1MAf2+PnOwDgwB+337Hk
    ypjAItzvzMCzkMAigdILzPxSwCFride/BHzAIcUKROCFRMAiK1ghI9TwwCG71png2/TAH4Ys
    hBTtIMAjcLLrAFBuwCQwakADuVrAIMfaj7cS2sAeww1+JVxQwB77ZRAgd1DAJFUlrEuVEMAg
    GytTdM3SwCTrF8Tf6/DAIMBwdWTyCMAgYmEnG64MwCCgX+qO2mzAHwanX+M05MAfA1hM9yyg
    wCGlU33UltTAIj582JlC2MAko8uOb6B0wCDXTFeT6AzAINi3ukQoqsAgvzGYJyZQwCOZesLq
    Z9rAIPITaTlGZMAjtolZGHIewCTTg6+MIPTAIyAjBY49cMAkrOxTTAaiwCO+9+svKPzAI3Nm
    H8DTCsAiie3MqhewwCSThKp/1VLAIE+XDGk/1MAgw7Ip0ArgwCIncm26NizAIPmUmuRvOsAi
    P4LsBVwcwCGr4HoDXhTAHyqeOmduoMAklylMBMnmwCFgkk09iTzAJWtC42pKVMAhx4dN8kT4
    wCTogutoOIbAIA7ztmr65MAjY9JO2lOmwCBKecHy5HTAIGhP+A+DMMAii+kkZx5uwCR9PXIp
    24bAIi8CMLS29MAg0vdl60iSwCISveTNvKTAJE+BtTa68MAgrTzKB/RswCHxlYyovQbAIg6c
    Bo8yMMAezRiiPOkowCKEwZnRBPbAH2yjJdQu+MAg6c+/9flOwCP3R1G+2B7AIP4sobNK/MAi
    nVCbTjHiwCBp9lpLK1jAI/mK+i2rAsAhJJGYA4nkwCG5Ih505GDAIFRRAIe9xMAi8rGKEEXI
    wCPmYL04YYDAIAkIv1i/AMAj1r/9WdnWwCKvL745zILAJN44ssmtGsAgGBBhVWmIwCLiAisr
    6UDAIKv7HB3F6MAfEpDWeW74wCD5g94m3Q7AH5G47GVUFMAjC3c5zJVQwCEs+qP+ZBDAIHJp
    TQEhSMAlZewSGNOEwCLrsWidbXDAJDFN7g7RxsAj2uuk/yWAwCBWk9u52UDAI7JEjFoUVsAh
    0eIHmURowCEBy+PY4ZDAIc+bOejP5MAioM2WUVvAwCADoDHUAFTAI7ULq7ZDkMAfXTx9NjwA
    wCNotkZ9QpbAIduzweDkiMAkzbc4YTP2wCCWkSP/WvTAJXPviWpZ1sAmXdo/rybGwCFQK8jz
    6oTAI1+rkdj1ksAgd8o0Xmk4wCWUHLJ4pnLAI9AUaAWXAMAjPpYecDdwwCVDBahrHCDAJIdU
    gDFu8sAg+jUJUtg0wCJzVah5VzTAJSygZmN0oMAhHas8K6ncwCXS08Ia1PjAIeIBk/ak0sAk
    x24jR1yawCMJw+eK0qjAH89lMYFChMAjXVor/ldswCIFlx3pDm7AI2QY61jzCMAjTxCXMD7Y
    wCFJ/7zrSoLAHre2XoYXbMAfg2AZF5AwwCFOXokN/vjAIBCQZ87y5MAhNRO6F0x4wCNqqMVD
    eO7AIfUA/eHL2MAjFtWHroaEwCN0oG85/GjAIwEvKhBhhMAiw35P1+SgwB7FAoiluuDAIew+
    yPzcBsAimeR+uLJUwCAKb7JgePzAIfWDZi0ZMMAli+llzLc6wCKHbuf4FHLAHpVzxR8OlMAf
    eJ8mcS4MwCHanwXcGhzAIt2Tn0yi7sAgrKmD3W16wCIkbYFVAojAH4dBpQK4+MAhYivwYAx2
    wCWWc+t3VXDAIcOz0Ngy1sAkgtffIfMUwB8uXA2uVETAI08KQu5qHMAhyHi1uYq8wB+sYSd5
    eADAIFodJZtQ5MAhWs8siDzswCOdVpu+0frAIXqfvT7KmMAibcNNZdiYwCRTtu+PthLAId2j
    PogiLsAitECBPwXwwCP1r8IX3aDAISWlvQu7HMAgo2wTWB6MwCMMSTYUJxTAIiT0DCanxMAj
    5R9cD2HMwCBWNlXE/njAIxaqjNof+MAjVpwjxq3OwB6Nx5nMZszAH2xBRdP1oMAigQBlPlSk
    wCFz7wnhtBTAI1JQ/HU6fsAie3rsl5c0wCIWcDumnBbAJTcqxPsvWMAfEi2Q/L1QwB/0m4vz
    rEjAIHflAVSBCMAjlZoy2efowCHEDEHYE/LAInVKBCsIisAg8VThXiCowCOQPffqTcrAIVWt
    RHrPZsAhfnCkzNw+wCQZgzCiCTLAJB2yE61XmsAjKVVGFqiEwCAv4FHMoWzAIM0Jv8j3JMAh
    RUAU0RCSwCML6/B7dXDAIdSBt1dmbMAf3xFY32cIwCLgO4Xf2AzAIK7FqXuapMAhIlelhJ8q
    wB9zQtIW+tDAIaMvTaKtRMAhQu5OHbwAwCS4C9FICKLAIPgHSvpdQMAkMDq47drOwB+F7eqA
    h+jAI9+ffWxZWsAjejpAOQzEwCRLyAD0R9LAIMihEvVeSsAhxygVPv3IwCHOpg5nCVDAJee3
    t48JfMAkfPpDHwGowCQpj4L3fjzAH9kO4SdmuMAlYjtzYd4KwCDcNgazaczAIuH8ZSSV6MAk
    VlfyZkh8wCQdpZHnD2jAI+ecN8zHvMAiFTl5w6PWwCV0COCkiUDAIt47UtoddMAjDkEwvKPM
    wCIzOLNCJpjAJIBR4S+GDMAhsF9UgYjCwB+/+rbR22DAIgilg0wDzMAh7cVniKX0wCF7pv3e
    fmTAJMuipr+LxMAhfHT4kc1kwCHAtqeDfWDAI98ETLUPKMAiETwBuYckwCUC1qj//9zAJrTp
    42FQQsAhvm0s2crowCRfJQwE5XDAIG5PAE1jOMAfypTtPv5AwCKUD5fDWm7AI7770HlcKsAg
    RmImO/YcwCFzIofKfEjAJXWmnIHo7MAh4cRiXIqewCHWLUbtw+zAJD71FoSSzsAjDxx+Fp5U
    wCOe9aovogTAH6u4m4sScMAhnWB2oSxMwCHITpMRqpjAJYv1jzc1qsAiQoHNILWGwCKPEw2j
    pizAHzDwTssycMAiZtd/Pe6cwCNkvbLS1rLAIp4RfpQPUMAgEkgFZV7gwB+oJDA+/oTAIzb2
    KEDt/MAmD5kcw3amwCV0fepD9jDAIt1x9NYO8MAgkO/6aTwgwCGbSnELtKTAIvMf4xnQ7MAh
    OQ2bBKvSwCJtuntBHWjAIMWBv4swwsAhxPGRQbpAwCJ1CnU0xBLAITs308B2TMAgq+tBi6HI
    wCVmEsBSrLDAIghG1EVz7sAkV+LokW3kwCDYpzgOQiDAIWFHWaEV3MAiJVKkC35EwCEh5w3Y
    BdjAH1btczLHuMAijVtr7ghewCB96qFc9rzAILV8MzpDpsAlxx8hYcY6wB6qKXSgOPDAIYZ/
    whW3FsAgLdVuElJ0wCPXzb31EDjAI4PWfbUJlsAj+DNzIqPKwCPvqUpDx+rAIgCMfuMVGMAg
    nbVfDxRQwCOZmtewASzAIXBMt4qk3sAjMmDDBPuuwCLg1x4jTYzAHuN3tSADiMAkjkrUNNkE
    wCR2ipIFIljAIvsOZbcAfsAk6sV38koCwCOOBU3kvqDAI2IfXdnR0sAg7AHgrvc6wCFWjEAG
    0mzAIPChF+DFRMAkYkzXeBSGwCSpkLqwLp7AIk5gdO7EGMAg4LoN7uQIwCP54r9p7DbAIxlz
    LyTCsMAivhTdQzaYwCOkzJyfcgbAJanw25R1kMAer00UfKCIwCMfPLGypKjAJdxVTqYkIsAh
    NSCPT0iywCFmZiDEgvzAIQdfOOu4WsAga7nRtI2AwCI5WfenIibAIV4sJ9GNSMAhWAa2hQi0
    wCSpDlsWbGrAIFW/b/kbhMAf6B22Xz9IwCRSoJX9fVLAIa1YraQrkMAgyhbDcHqkwCUr5lYQ
    FwTAI8wT9WDPnsAgqakgaUW0wCPvPwyU5IzAI1kGrU2dTsAgZ/mo6fUowB7HSml7HXDAIcb6
    PF/SSsAjwovLaU+awB8BDp0AXWjAIZtfKOm1zMAfpsKFOqA0wCNv9hh1x/LAIxR88wYTUMAk
    fCQdeGy6wCB/EZBNxFzAIgxs8YZurMAgOkbjh9E+wCFrT0ozvCTAI6khEkMfbsAhIOu2dZGe
    wCNgeWmvoXrAIt6sQOh6dMAhqn5A6sEiwCGQtJq2EITAIXNEPZXIGMAjqQbAVIkqwCXbJFW/
    dprAIbjI/bQHEMAg+EDJPeFQwB/+KhiNt0DAIEpUyA3LGMAiKRS9RqsuwCFLRPB0rbTAJRPn
    T6Mc5MAjBQzJ4TUswCJJPFjT5ELAIIfTOnOOIMAeiIhsJ6TkwCFPikIr2dTAIDVDAevf/sAk
    kegfp2tEwCJz/i6Pl/7AITCZHm97VMAgDz8GTv4iwCExvjCixBTAIIMoIFGESMAiK2Qrcgvo
    wCFmcL4E+4zAIzN2htGGWsAguHEluUaswCJ/rO6IiLDAIxSBxwtMZsAivVV5rhQQwCEvmJtB
    2LLAIZKGAPuaZMAi8J4mo6CWwCCtlPCuAvzAIrj1umpJsMAkAkIJ/EB6wB8PNba3gLjAJSut
    tZJ2PsAg+Hc/8q9AwCEKwniF8sTAI3bhx/Y1MsAibmHxSMKiwCGy3xkskvDAIN0fkV9ZesAh
    6WEU3+w8wCXwn6dVZvzAI9D5N38bssAgXrjhjfrcwCLK6EeqDuzAJDVMjYSKlsAjONIOMdnm
    wB9PYZNb9HzAI8pLbB6KbsAf8zV0j3xQwB64pTG4VYDAJEIe1U05EMAio2tj7sQQwCH4rVQ5
    OOjAIQYAlfydsMAgWm2rAUPQwB8wlHCc4CDAIixDOG1kMsAmpKnUyGk0wCAQPDQjEojAIfXc
    cvfkusAh5gbGhvDAwCG67CyMGxTAJH/HsyH6FsAitr6fjY3ewCVeKljQWRjAI5Yxn/Me2sAj
    1FHo8FVAwCGQOAcqKyLAIjODSWTOwsAewCTQcwFgwCMS2aJWQtzAITdhyliFrMAl+5G5O3X+
    wCXXvbLqHbLAI70jiexFfMAjOXXnfzqYwCVbtXaZhWrAJBBHqG4IHsAfvxIvY19owCMFffgf
    uQbAH7IP7BN0YMAhBVJx94+MwCKeCI1VXojAIeVZYaz/FMAiY6uafrHgwCGnt3IefhLAJapO
    IiGVwsAikacL21SYwCL6sJj8eoTAIdw3vVsXFsAk6v0HdFW8wCWPAoYtdTLAIwbmKegFIMAj
    S4XCkRRQwCDTbhV35RDAIvl6x5VnzMAiqZe+OnFIwCNEl5lI3bjAIYPyQy8D0MAhpd2CxIDw
    wCF8LZyTE5DAIaOtvScCnMAiCqMmp1PwwCNx+ZzQjfDAH9grdkK3kMAhRoZIfDtUwCCJO951
    0U7AI5XJKzYyKMAgCzhn5qlowCRaOBKccAbAISC7P2vuEMAhsFcbVSlEwCOm+uRjADjAJUit
    bMlUDsAhTDGgOxSSwCIBuig+c6bAIf+HEYRa7sAlTRZkg+fUwCVFF+RiLljAICCUhjwRwMAl
    P3RIeJ5iwCTHIpC4yTbAHy6YrOdUwMAj2FP3zbdywB+q5O7EpNTAIM7ANapOGMAiVSGdvH40
    wCcwclJ/pVbAICmojYPX6MAhT0U42Je8wCHyjHINyETAJQoZmDm28MAjX1Ccc9n6wCLMGhMw
    NLLAHs40AtDRGMAhEwHB/ufawB+Zcgk7FBjAIFsDxK0lBsAkaAfYJMJwwB/R8qcO9pDAJYNM
    bxoU+MAgg07R0o10wCR0fEX+EhrAJXu5Fy1PzMAiksMJiIEQwCCYRn2l1DzAIT9B7GphrMAm
    VWS1PGlAwB7ZGi52TqDAJaTlgeUz/sAgyBaiZTLgwCE2Hbv4o1zAJPWEQ8CihMAfTZ/gRmbg
    wB8Vr63LAvjAH7wgyFP13MAfJnlnY8yUwCHQmcL2b/DAIuhfetUSbsAh8h0KxGsAwCHwB36d
    DqDAJAU0p0zUmMAhjUFxoq8owCDNnm3T89TAJF6lhd6BfMAiObtYJ9agwB6UrAc7gLDAHqpj
    zTv0AMAgHrR+b8A4wCGbpa70q1rAIFM5gUAdyMAjCYV2e+/CwCKNMHAlzrzAIKSNwALe5MAh
    +fACNzc2wCCZkoXKFLrAHxrpegeYuMAeoTY1I5HwwCA5yO33UszAIona6ANQKMAheDH/6xLw
    wCCijcDf+0bAIYVKR19/FMAjFZPR3FNgwCCNkO3IZYDAIr9prmPOcMAjzMAv3YzIwCMSUVdG
    XhjAJHN0mYV/ssAk/1sXokuQwCRe9ZdYbprAIXZqZn/XMMAjuREdL99owCAEJgju2tDAJHjq
    cRPaSMAiDmt7nyr4wCAqn9rWGVzAIPBD5571tsAi3A+02H+swCB1FAttoJjAIdl3FthNzsAj
    WrMyu5SuwCG0zqDyqrrAJErIirZbPsAi/UIjSIB4wCNYfvyauijAHr5MN2MHUMAg0urhWBgQ
    wCaFvhwajtTAIpX5/NDKVMAfqeuzYmKYwCUe4sDwmgbAHretmFK0AMAk8bbRmAocwCYj5yl9
    v1zAJIxkteRRPsAgxM1yy96swCKe2tbk/AbAJjNSqxcF4sAgn7mth4RSwCN1jQtUUlTAH6Lf
    pjVKAMAg5Dv2vj3gwCMLzXoJO9DAJUT38SbLRMAgh7/knePewCBxhxn6pW7AIOh2fM1IqMAk
    VBAW8WpSwCOYnUM/c0LAJd6czQHiBMAiWmjnO3zMwCMkAEDuvhTAIDcadJjtcMAgdyQeW5YU
    wCC/Vbumv1zAIZJFSPFo8MAjHzZ9Acf+wCOw7PxrxwjAIW21Jalj8MAkiYYtFecKwCIjffk9
    WjjAITomIOtO1MAhCoU3VcSYwCPgEsaZf7rAIdZdgLhR2MAjxvnSsX4owCKKADuaFNzAItgN
    TNBcdsAh2UZKsArmwCHQ7ZmJMATAIkdVI/LUKsAjRJ8dlxw4wCOFcpbDmLzAIpajVaL1DMAi
    dgIqE4AswCKV/Kd04W7AIRpoQAhwFsAi9uKlmCt6wCED8R/XpMzAIRFWF45pNsAgEibr1Btw
    wCCdF/c0I5bAI8Gm5+IBosAkMm1CIv/uwCBdZZC2Z3jAIzMP1aNTNsAffbcx1Z04wCQhU7Hj
    kLLAIR7i2Hr5IsAe0Cj9ZBlYwB88YYPFNtjAI6OKE3suOsAjJhcDoim8wCFCZoou5ETAIfdI
    SqOUVMAloiLE05NkwCJIx83/6iDAICLRy/7iTMAjlLiHP8HUwCUaqSfnpyzAIkFwJAqvMMAh
    R6850TvGwCPTZvVqK97AIaejE9dOUsAisxthzYqswCSo4CWDjELAH7a2CK5SuMAk8LbSI5S+
    wCQo9Rb4GqLAHuswkc/GuMAiGakEAUwywCYtGuO/hbLAI2kLiPITGMAhOGzyKB5awCOeQT9H
    mZLAIM9Us6n3kMAfGXOWQwIcwCM0fo2T1lzAIr9X2VyEnMAjGPYZOCWOwCWKJAn7jQbAIB1v
    OPCaGMAjD1UwrQhkwCHIRyjAZYLAJjosgGWLNsAkK7uponBywCBW7hfDYXTAIJi3TbwhnMAg
    Z2Zt9UTYwCHqTfvPUwzAIAa+531UesAe101zllSUwCDjxIQVo6jAI+1OiPR9fMAfZiOCHidA
    wCLE79cHfQDAIaP/K0AytsAiQx+RE6TwwCJwawO8C5DAI8qd4osgTMAfHh/BMVFcwCHRXakZ
    YfjAIy8xRuKEXMAguPV6P74wwCHpxDNC8NDAIPhrb3c22MAi5IjlC8OMwCUznlQt3nLAIMQs
    5HtudMAjSA1ISO/YwCGDvc9ZEBzAIq+zoF/ISMAiQQ2oQy7EwB+bVDtxGcDAJDrRvD2yiMAj
    hhHTOk7MwCI+tp7w9ILAIUyVqcZz2MAkdRLT9DDmwCH7p9Jtq3zAIYZWsE0nvMAhJQ+1PCl0
    wCNqzPwcsRLAIgXRPdJZNsAh7Um/nLUEwCHw2I3mLbjAI1VVFzq71sAe3IepeLT4wCPBsYPu
    AjDAI4balC3qmsAi/IFK8s94wCJiNe/TKuLAIoAquuiEEsAhoVSeRcDYwCMkgCcaw6LAIEHL
    0kbdIsAiPhGCPF4UwCi8oJr4zWzAJajRRrXCgsAjdG65GieIwCBrdmuFx9DAIY7LaRj2YMAf
    /aYvmDxIwCG6FV4M0DDAIHyjVao9BsAjqje1DSAswCAiPpu2L3bAImY64O/vpsAhBkr1xD+i
    wCK+7Kp939bAINmLKSPtgMAhZURBNAaQwCJUH86YE5TAISoziWf1UMAjsRNKr2FcwCHSDAab
    ANjAIx4DOG0xXMAkva8B8/X+wCNo+4fQP4jAJJblPNaYPsAhJnXqaEtkwCQzb4Z/yfzAIssZ
    Q6QSBMAkSSoAJNKiwCB2yajj+qTAJI8564QS9sAhGZcAcsrCwCNBKWITYFDAIUIVZX3aiMAh
    ck9ofLxcwCGAXeFWviLAI3FhbTTrMsAh7p6qRXhCwB9C9sQ143TAJAK1Gd8cusAko3rVq1Ni
    wCFMqSOrCLTAIpYPc4PRQMAjCUUfw/YYwCMxJYx8dPjAIH++dlsw4MAg3nI9mPbQwB/d5/hl
    i4DAIiAFhWxDpMAjUTL1/SVMwCB4hfa6BgTAHu2dDXe2VMAg8plagAR+wCBfJPlo/UDAIWui
    CBtSOsAkCgSdL0I0wCARCuD1sLzAIJQy33VUwsAhiPibOOuewCIUVVXid/jAJW4oLGJ5SsAg
    +VhoDJxmwCHL0lnaoSjAH3NhrhMNMMAe40UF7ETowCQ9QD08fiDAI4Prp+6xvMAkjoH1JlEm
    wCGBabfoYULAIe//SoiLLsAgOAa8xU20wCC4jlCvcETAIvIb1dv5oMAhvtUAFUYQwCNJb0Hm
    tnjAImH5Vf3NmMAjoaKnqShYwB8ObBLYSgDAIc06NDFRkMAmGBBglpUewCGSZzTn3NbAHymC
    aD1qKMAh1SkbNSXMwCAsvcF3SIzAIl49243xcMAjD5sb5ZkIwCBSx5lsxoLAIrTBtavwrMAj
    pBniG5JuwCNKbQOfTSLAIJxBUTE6qsAgYFF2ntbOwCHfTEHpUXzAHteHOcQlCMAgjaRLd0YI
    wCCsWclq1ArAIOkg1wz4pMAgN8EfVdSAwCH5M+0QuJLAIemMJAq7LMAlZf/IMAlGwCTAkDHP
    NkrAINtGfOv49sAj3MM/4jLQwCOjmZlN5wzAHwJVkyqp1MAlJk1wNVmkwCLSCd71a1TAIcxQ
    ANKv1MAkvzB7Vm/mwB/ZS4O+uKDAIdpC1Ed/6MAe4Tt1NayQwCLnQmyFrNzAIgzA0BAIVMAf
    W3lVqpMowCCtgwZZL+DAITfeHbN+uMAi9tqbzagmwCFvOW4cvibAI9YvHIDAIMAhIU/Qid4k
    wCI7Qnd8srDAH9yIE2eYSMAkTMJFW1r0wCLJ/+zHDVTAIdUattYoGMAkKyQYZc7awCBnssp5
    nETAH0xSb3lONMAjLukSnNB0wCQ8Qi4p+YjAIbXcTRIXSMAgNR3Oga2EwCRTH56rdBLAIbzm
    saGPnMAjY3bHlL7awCTpYWxXlSLAImBHIC4k1MAiProvpd+wwCSB2EFfo97AIm8KegbHfMAh
    1+qRY4bkwB/0lH1p2lTAIetoTRlTtMAhEjfjg1wAwCEOIlzZz7DAI+GN6wLW8MAkGnRoEEts
    wCEZihijvGjAII3KWlzO9MAjUqlaigSUwCIx8zWwXlzAHsvwcj0JDMAgfRmNE+FMwCAqIt9N
    aGjAJCoHr/BBVMAju2hGJyREwCFUJrRNK+7AISrcOcAn3MAllUghT9lYwCE6BLtNcVDAJLVw
    j4dPjsAlC9GgWyIGwCKpSIT6wxDAIzF7cNRuJsAl28VEs9nMwCPbkkm4g9zAIgmsTMeL5MAh
    799KW4UiwB8YMuZq+QTAIawL29rM/MAjrTy1G/hYwCPo4aQLhM7AJAcR2QyHjsAfiFlH3UDQ
    wCHVskvMHFjAI1n9HwN9JMAfZFpKuO5UwCE6E0QIQ9DAIvBfxL7GDMAj0YVW+AgYwB8vKspS
    jQjAIo0LtevsLMAi66avlH0AwCDIy3a0pyDAIEUyQ1N+kMAhzpGFBwbcwCIG009w0vjAJPDp
    VHA4rMAlV4XuZrguwCGHNpu6cBzAIS03/TFBmsAfuSCXoUJ8wCEiOG7GCETAIiZf0xLmVMAl
    IKIUihwmwCAMT5T0dRbAIFi+xZrRgMAis1xkiN52wB96Zgd/PnjAH6qq5WOi+MAgR4e/Ypm4
    wCMUuS35XJTAIU0nfFDvsMAiUcE8aWyowB/iC7YBccjAH/lNbAAiTMAid7myu95awCKdN8dt
    06DAH/LFLyv08MAf2WN5A+BkwCH+zEN3FwbAJyGLFnXmRsAh3f8zRQySwCHEQmuj8yTAIkfI
    W6EWmMAjpAvCp0xmwB6wV6arMEjAH5vQ+VheCMAgvN0kDpXywCT6YFiGGtjAIeR141pPTMAe
    4YzUjjvYwCUowBVtDRDAIPIhlE0eUMAivkcYguRMwB8ETKfxfvDAI2Zxelhe1sAlqKWJ3NA+
    wCItwytae8bAJUoG/fWc/MAh7LZG0JeOwCNcl5B2MVTAIfGqlsW+OMAjcXr2xEw0wCGyfHkr
    3OjAIRgwgQCdhMAj+DM02knQwB9t1gst1XjAIrH39Foh+MAj3Sw3eeAywCRp9iWH0R7AI/TY
    U0DcgsAhGUcX4l1owCX5Vn5rR2zAH0XyXcjVkMAgGmMBZoc4wCGQfJwBKzDAIQ3tAoo/XMAl
    BzhzHTFewCFOfKyXjpjAIg2cZK0ZEsAhJB5zbb9wwCXwMhKMy3zAJV3GdaLxzsAh9Ah/ZalS
    wCQiI++lDIjAIDDZIQm3jMAiVtqp+4gowCEmi5CeoRrAIZ38KieiIsAkBfDv9z0MwCNfuHt/
    7LDAIeLXFTG2ZsAgLpj8Ha1wwCD6Tv3/zqLAIym0ghxIwMAftZ9R73bQwCPAknL3yHrAIWe3
    aeWtaMAl+NPgnI8KwCYgsZCFBGrAIFYC2ZsoZMAlzA7Xzp40wCLcDOxW5/zAI2978xZ+KMAh
    Nta6qH7owCP2XJJ9H0rAIzMYPL8QZMAgttiQsZ32wCPPdBxcDTTAJAN/gB+hjMAkOtOHekLG
    wCPK2J2CSKLAIOTJHv4+9sAhreFfkmlEwCEvi6aGOJjAJi4Ax0nVcMAfg2WFHISYwCBt9JWT
    mWjAJH77xkZShMAgvEIq7qYUwB69wUev0+DAIiEFmLUhbMAiH0BpQmYcwCRyb+wewVLAHudv
    CGHhIMAkZAhHshw+wB6iJcui6RTAIiGi5ZlmGsAhn20iSnPIwCJWP/meShbAI8a5i8Cx2sAj
    rD8V4ij0wCOMjMiX/KDAHq62ZkqzdMAjbbq21Hh+wCEFcIKhP4jAIGEZpSZ1EsAjWR9uJusM
    wB938vKUs0zAILhOU8jMMMAkqtZZ0zoiwCHVT8jP5ATAJNeg0BoSXsAikJph+Nj8wCMuBfh7
    0YLAIvyaG3XtusAkR2HVgPbmwCIMtlJMN57AJE5utgQAgMAhjB/mMVj+wCAxjFGM+97AI/5/
    SCyQYsAgiZMCRc6MwCU4HQJHzvDAIQCapBMOZsAhTesRZFbkwCHkpl2aQtzAIaw/edu7bMAj
    skeWq+ykwCMMOU/b0djAIm3hoiisWMAhJE474E8cwCBJ2OAjSHLAIU5CHAEWgMAg1aDFuzuE
    wB+4oG5nTdjAJQ+bRPeQfMAe8gsjs57AwCQV2ui1cXLAHuUIg5AW5MAgZcaF7O+ywCHTR5PF
    1JDAIb16Tz5z/MAe2t5RSzfIwB6YJWwk6HDAIXnnt4AMDsAgSSzFdnH4wB7+IDizF7jAHomp
    7kausMAg07jfsjDAwCGw4D248ZTAH3eIpdhI+MAe45/NHdTwwCEvOtQLMKzAHtVx4YdseMAi
    l01YyloAwB6F/eq0AZTAH5043qbbiMAekJfsVb6swCAMFtPPqm7AIh2or8mpsMAej/CgaOJA
    wB+zGeTNa9DAIMJmkuDK2MAfWTlB4FzowB/VYgcgirDAH6jjhVxfgMAepcG1oADQwB6hvtww
    LhzAHyKJ1AobyMAiUOG+gIlgwCGXKoUbfATAHxTtz4c22MAesN7Wn0xgwB6XPrubE4DAIIuO
    ldaAYsAfMB2yvH/gwCAJQOjQJrTAHsvWL8vkKMAep+KxyaKYwB/gG/F2b0jAIH92gzmjCsAg
    e1mbEQpgwB6fi0GluEzAHwsOU5Kv2MAgQdrwWp7AwB/hsEyVPdjAH/EhSSVdnMAgGnhEAI7Y
    wCA0vHOBRADAH5GyUAyB4MAgbnoqoWqswB8nPWxlaJjAHt03pS03TMAeicRythIIwB9e49jH
    p5jAIDLKoSiU7MAfEEyWpjikwB7MXPJ1A5DAH27sPZjoaMAel7H6uiEowB7W85CL2FTAHrHx
    RszmwMAetHhSiA7gwB7XfoyAMfDAHqqgNfdWWMAeqk8REY0gwCE7oT0I6HjAINFTNfSduMAg
    bcYCn2UkwCFHiUINX4DAHojxNqiXoMAjGeJ81bcMwCAiKDDHLKTAH6U5iF+VZMAfBl/Aks2Y
    wB64vlnZ2JDAIHGrEGr3rMAez2yyZtZIwB/COj64h4jAHxoXLwkkBMAe4+Y/TwEowB8KI8U2
    dKzAIFRX8Cb1CMAehf3qtAGUwB9gLMN8DzzAIDpat241osAe76ZMYppYwB8K96tTDEjAHtnG
    uLbecMAfNCf4ueO8wB9nr9DcwBDAHqUePXrtVMAhwkY74XJywCAkdlLRIe7AIc9GaPwRssAe
    u+vdqWnQwB7NKAh74RDAHpyGuIzq6MAehf3qtAGUwCFVTr+P+3zAHsDsZNMbuMAfv0L2pIX4
    wB6YQlBGbjzAIlZadMeIQsAhQdyeN664wB/OJyCYzSDAH+94E01S4MAg6+0obA4wwCDHYoz/
    AxjAImqRIfBRLMAgl/lsDJUAwB9hW4kAldjAICamF7gtasAgSNRJ1I5ewCFWZDxVWMzAH0wV
    y4EaXMAiQxLvsuUawB9WGlBUTrzAH8GtgP3uAMAg3tVB5/EGwCJ6iBtRNubAH4cHxP2wYMAf
    jrMDLlvowCER7ZM6UnzAINwf7TDsCMAhbbWL1uYUwCDcgztfof7AIBytF7U4vMAhsuZqZ2m0
    wB+y/U5kk1jAImTIKJUs9sAiIOKJevIgwB7KVvVy5gjAHxu9eYSI0MAfiLKgOArowCAF9HWo
    AzTAHu/acmf7GMAfjETlA3UowB8tgvvn64jAH7Ikb0l2xMAe2DCZOJGowCHDHscoW0DAH1lP
    vQM1qMAfHv/uSZvYwB9XpvWcqKjAILhUyVccPsAfGCDt7mOQwCLLXnlfKYzAIre+hBPVXsAg
    0Im7N/uYwB+2k2IadDDAIFfLrXJqAMAe81c708CQwCBtQ1Z3hajAHrjSW2whPMAfWA0kb4R4
    wCABzxkAQWTAHzZwxjptXMAepPxBcfKAwB6PEoh/PYzAIRV29uCnzMAghJ/Ltpn6wCBVnVVD
    FjTAHp8fj9/hqMAfTQAL+VAAwCHWn5nQ5QTAH0z71wR75MAgsiYAM7uqwB/7haZ+iUzAIcQe
    7vjshsAeoNgVtyWUwB7t149YC6DAHqTC+xyxxMAet8H4wCbAwB6TrDQmWSDAH2dWi3dOYMAf
    M5H11D5cwCFJ+W+xbu7AHthQ+1q6UMAeiVqRddvQwCBVvXWRC5bAH/mRbc9ukMAf03+4Q+SQ
    wB6Kzz16GZDAIBVLq/TDOMAij/79F9v2wCCG36zaEXzAHpWs9Y4DiMAfFQRhYcB0wB7nCJeN
    W4DAIP27YtR6VsAgNFzGb1B4wCDQlpa/pDzAHrGBk/oRtMAel1uGobfwwB/kiQpHnlDAHzop
    ZB7DaMAhjfWmxWf0wB7NY/yvZjDAItfBtKiBSMAe7ANoMg/4wB/JyuOwDKDAH9u4aO5FEMAf
    x2TLTx2MwB8nuCsEk9jAHrdawNAf6MAejmy5YROkwB8AOfMYmtTAIFmq2f6G8MAiugYzSNnS
    wCAVeFbWndTAHtnwH4eZiMAew1QXp/bowB6IBkvugfjAHtMcMXcM8MAhBQmBCb5IwB6K6hKG
    pFDAHwJDBsSEuMAfoNzMVd5kwB9U/OcvNmjAHw2FBectYMAfMH9FVQ+AwCCZsJrKZPbAICXe
    P+s47MAfd1jfnWTowCIEXLxPlSzAHxAtIjVuaMAf81Y4EjDAwB6UcxTI30jAHtbe4LOlKMAg
    D47+M+oowCF4qJJIRHDAHuqDikNymMAeh2uPc65UwCNDPSpKehzAH9CSvuAv2MAiSEvDLbLM
    wB8dN5LBBujAIMeCG/TjBMAeqKb+fX0cwB7DMDIWqIDAIKnWzRkbsMAe9xCuNCX4wB78G0hR
    ZSDAHsg5BYEGwMAe86G9jl/IwB6XIfPEvfDAH6DzQQZKAMAerOOoLQEwwCAj3fh7F2bAIfMC
    KdDy4MAiB2gKyWB4wB6F/eq0AZTAH1luhSpKwMAiObsVec9uwB6Tj9lZVMDAIbVB5c5BVMAf
    ZVDkM2q8wCCCQF5oxejAHplOgr49SMAgnH5HSllswB93SbxKMozAH3HCUqE06MAe5JeWV0Zg
    wCAmyNJgp0DAIeaAnCZKeMAiQzLEuzKgwB7UGAOptyDAIOmi8Id1osAfiANni7OUwCCSf6F4
    HHzAH3MTm4fe3MAh7Cupp9imwB6sA6yQH4jAHvfvNPzroMAfWw3H96TIwCDwUbQzrwzAH7XM
    1O9NWMAgY5paC3eUwCAjapfFyLzAIC4eOUVc3sAi2j35CCvKwB6Qz9KE2qzAH2xuaoR2uMAg
    S3sjefh4wB7QFlz8sMDAHyY+AEQeqMAe5VQcAeHswB9/wjUK+ejAHo41mvAG0MAgTEWjy4tW
    wB8UeJTTi1DAIO/qA7Ox3MAhUACxiq2gwB8c4kOv0cjAIuELqIxqTsAhmEhnF6ckwB72/m5Z
    eSjAITo8N48DRMAidslbFkc8wB6WAtdo3gzAH6se/MN5KMAf+Di5btb0wCEpgD1rbXDAHoYi
    81DJMMAg00fAq0bAwB6gTrAtQiDAIUeJ4bwJFMAfDrU5vxXIwCGwLZyq51TAIG3Ge9ezFMAh
    KIY+OiwKwB9KMwM0rYjAHpF36MnxbMAhMoQyJtKgwB6XeFTVt0DAIIFKW9u2LMAfGrWgUkNw
    wB7xd9lzqejAIDAQYxlqfMAe/USFPwx4wCKtllZeI4zAIA5i3GeXnMAe4izB/kmAwB/67tB9
    5FDAIc/8nDSS8MAgA4gr/MOkwB8shtnCgyDAIBhEAo/4psAelcmS03U8wCADvfZz18jAHobS
    nQBr+MAgIWOz2ZPQwB69E8GzNwDAH8nNDMyPLMAgz6scl5fYwB7JX2126SDAH/Sxp0vfcMAe
    4sNZXBYcwCCuaxtOjRzAICTV6u5QDMAejEkuRYtYwB8QhqD6h8zAH3RdSXz4aMAg3ULD/AQY
    wCF8WvrZfbLAH8D2dKIcdMAfQsCk+ob4wB6T5PSo/NDAH3OdTN/10MAezI0eUA0AwCJEWIPa
    7sbAHzwAc3wPIMAejVnqi/NUwCB+sFdBL7DAH5v9ckD26MAgg6P05wMIwB7qtRURbpDAHvjD
    wVQfeMAgEraB8f+swB65Q8tPagDAHr1ZqdltEMAfiSLMg37owCB/8keI3ZzAIW10D1QKNMAg
    CcC2fUqYwB/uLcnBmBzAHsymQQGO1MAgzohi7CrawB8j2XSsEGjAIFkBMTdBZsAhifbPJaLE
    wB660MVpKKjAH+M5aI1/BMAezIMLjYG8wCIPTtWOVnjAIItMeoRzPMAf9uDWHHngwCEk63Xl
    qADAHvQ4CHECvMAglV7mDK4AwCAtXqiI12DAH5N3N900IMAe6BUFd9JQwB6NIzKFB7zAH4Gz
    2gU73MAhl2U4wFyMwB9/T3e2PNDAIIkvaYg3asAetSox/kLcwB71Jd3FUEjAIap9AW5PsMAh
    l0RoVWm+wB6PgXA5dCjAHoggPZ1VvMAfeUJhdqjIwCDZ0q+5F3DAI22Z6x66qsAik8zNOycA
    wCCudZ4XqXDAIQSU2ZFrvMAehf3qtAGUwB9pMQH97pTAIFLZtAeOqsAewt+kmHIEwB7MwepV
    56jAIUNMEv5cBMAe7Ir0jC6gwB7dEj9Y6GzAHwnpK4YqNMAei3CfZfD4wB8MW5mRo7DAHuYn
    gmNYWMAeipmn+uoQwB7QkYuiPGzAHpFb2tVeWMAelOUHqYmkwB6UyIW1DNzAHrOK/fo+2MAh
    TM4TVvWWwB+FX1Kv7WDAHugRLZNI5MAejXVOh4BkwB8NP7BxGiDAHpbobbT6fMAetpvH36vI
    wCAA0qxiNQzAHwl3/sGS0MAfYMwjtFDIwCJntqEqBAjAH72MSuIDSMAfEb/D5zwQwB9ARiPj
    /pzAHrXUconYAMAf5whoFoP4wCAUYRVxdq7AH7E/xnOZfMAeh4U7yho4wB9l7Z4bMRzAICjl
    7tLjZsAgYN44h/aowB6Z0DqcUWDAHpUeFfP8XMAftQn30qDIwB6q1gjPF5DAINwksADh5MAe
    hrlR7YWcwCFpi+I8sijAHpnVEuazjMAgI3AWqXoYwCKWg5Ej57rAHuG4NJj8TMAjZWbql1m6
    wCB1xwJuyErAILjty+hQhsAhhYjc4PqswB8zt32v4bjAH2bO27KQSMAilHFLwNkswB6F/eq0
    AZTAHqOzrb1RcMAftjDFXrzAwB9Aphduo4DAHvhlv1oVaMAe4lqisQxIwCFpdZbptprAH35e
    ++NsIMAf/UmpjAJowB62VJsODOzAHpqavCCc9MAgK4jDCcB4wB8AIJZnYNTAILCBnd7yosAe
    mOY4VBKUwB/knXlA1rzAIDAy+i6JSsAgRURY+NScwCISK6o5zMbAIUPfrVq7FMAgDfIgzzK6
    wCJcJcGCS1DAIDfaSSQF2MAgn+g2Y7rawCHBTZ0gKEDAH/ulTg5pEMAg3HBwm+cMwB8YNhNd
    PQjAIU5GqRHZWMAffqvr+oJUwB7wDapxDdzAHtDavTTYgMAi3JxckJFOwB7eS6EXlIjAHzbg
    x4VwcMAe3JY6ob84wCBIoZ8VYB7AIfSySxCQmMAezg/6OIhwwCEM5d9Ne+bAILaTc+njNMAe
    yoFsdGS4wB/iH9cFEXjAILJ//cUwvMAfpG+tCibgwB6GO9UaZxTAHoZUylxAsMAfPj3H0sRY
    wB8wcKTxRDjAHx/+W9D8sMAgfCBQl9s4wB9GSvk8tLDAHvimfhoNmMAexDXYjH2owCAKh9SK
    E8DAIDXtPPVP/MAf/C35O1towB7XXViQonjAH4a9ElnpaMAgIlWSlCm8wCFKRfLWkIjAHz0u
    x4Jm2MAejr+LgLVMwB7rUFNPfvjAHw1FKBchpMAewj0CvfEMwCC4Wmh5fIjAHuM/1+TIIMAf
    9fUIQXY4wB6iiy4OKrzAIWH4v0tXssAgaq64VZsowCBTGmj+z67AHodR78WrEMAe6rfdZKLY
    wB7RDSTEBwzAIQj0yJyl1MAfQAZPFVt4wB/zNv8DsTDAHqqKlbIQfMAfO8PFxVfYwB/SgDck
    5ETAIjtD1MacfMAhr/42bgdQwCB7XaF/B0LAHtC9147reMAgXpa1twYIwCAsVhQZHgjAIUzr
    0fDdnsAfVuwDBkkIwB6PSfMtM0TAINzpx5AN0MAfRIPbiu+gwB+oBWhbStDAH2nNnrYpwMAg
    B53hED9uwCDnv/zU1MTAHsF75Fa1RMAex4JCyA04wCBVA8V1W2LAHrRYxSy3tMAhN73GvW/o
    wCAxnHHS2iTAIN08K1XhhMAhjmBv4wZYwCFFDxd3RM7AHzzlXMUnEMAgPTkmA7aUwCD4sWM9
    7ujAIVlhjvXC9MAelnWIXuTIwB82W2g58mzAIFdc1PN1bMAfVjdfho34wB7UAOK6GxjAIE6u
    7muuHMAf7yKfwV44wCBtUzub6BzAHzVSprjX0MAgLErBBqzcwB+KQj7vsFDAIK8zAr9xHsAg
    QPLr8kugwCENVTTGjH7AIZT+5oxaYsAgbFhfzgEYwB9blTJM3LDAH81+uxqynMAf9Fr8+1go
    wCAmqyiv/ozAIHDya9uIyMAe63Bof4IQwCCr2Uvm7TjAHzEgs+ZqgMAgU6ZVPYhIwB/RJuVv
    QpDAHsGYVrMKaMAfHHx1yNNgwCCFNpF8jWjAIgEwZ4xbssAfFy280mU4wB6/OL6zAHDAH7ea
    5ozsyMAe1fhQEy6owB7wlvAI9NDAH+HYI/TZ2MAeqglafvoYwCFrtSFlxzLAIPbZFXC33MAe
    o1ZeVFnwwCBkStZsCODAH2vAqnZELMAe5K0aaEV8wCAsS+zI20DAIBACWLChOMAfdV9kiiyQ
    wCF2QzqOkhDAIKJ4yKLhTMAhBKc6HsuWwB6m4c2SFWTAHqE7SCEBjMAejvbaIMo0wCCxPw2i
    z87AIGlpcuXzZMAe7HweEwvYwB6pGul95SDAHvGadg7Z2MAepMHgS58cwB9ViNLbygzAHpIg
    kJ0CMMAemHwhy3OAwB7zJApzm7TAIC5g3JFhIsAg3mQAioewwB68/+sJiBzAIA7aX8tSesAi
    O9HPB0VUwB8L1ugGpyjAH9t1k7Uw0MAgvFWzx5TswB8UzmIK+KjAHpBgFxxrcMAfpX+oVA0Y
    wB8bbAwzDvTAHr/6rrabEMAfAAQO5hrkwCI8Rb0zrS7AIh2dnHRmJsAemdu2qKUwwB704UCl
    t6DAHxJpQUBiQMAepUCnFdqcwB7t5sVz0kzAH+aD1Ik/eMAgq6/eVZpcwB8Aag/3jSDAIo2s
    1O5ebsAfuPV9RHO0wB6oKlwfYqTAIX1zpoJs7MAgw163Bkm0wCAbuOOAhzzAImcRYvIx7MAe
    octGLgGQwB6OiE/uD8TAH9PnsKdbCMAgdTiHw1VcwCC9u8JOAsDAIPPd3A9yNsAes9uMDMUg
    wB8T9A/m6hjAIhS0JyJI/MAgs6YOdeeUwB8br1ZvhbDAICTG9zOAyMAekT/Q8Y3cwB8QQ9XZ
    X3zAHqTZW/yTKMAfRG2Jhe7owB/P/egmHbDAH008kzleyMAhpoxx2x7GwB6G6/dbdODAH2/j
    lQww4MAgoetyWCZgwB66K2khQODAHtYovXgruMAfeaOl3Ag4wCBLgKEHygjAIGvrMfYqfMAh
    9moG2nNIwB7UgmPd00jAIg+bUIxp1MAemPiw5jIAwB6a0Nr3gUjAHsXmIk4O9MAetW08kkEA
    wCGafYc7hGTAIQdtQx5MEMAeoP8km+IMwB+GUQBKfFDAIlVm3kaXvMAfHe2jIuOIwCIHiVHI
    5sDAH1YioOOifMAehqAWsdTAwB+oouOODZjAHoX96rQBlMAe3RMGpqpswCDEO4Rt3rrAHwa4
    G46sIMAhkfWTrtFYwB8offJv/XDAHqaVVVKLCMAfca9ko4u4wB6GbdI5OZzAHqihA2NRiMAe
    suROzcTIwB6toSpzK/DAHzv1dcWUyMAejLYDlrY4wCHN5XlSBazAH5yQXAvLPMAgz/mCwOtI
    wB+39FxazrjAIH3c9cLtzMAeqlgQdDMQwCGvJUDMV4jAHuCTgvnAsMAfiYXQbQBYwCDiwVL5
    aVDAIYYtXu7FtMAiTh8ciTKQwCFAcTU5QwTAHoce1+EziMAgK/IUUVt+wCA94Am6SJDAHqYs
    3SToDMAelI+MQMskwB7jjl9bYgDAIaFRVMeHRsAen4YODSF4wB7qkU9tSBjAHsZUqPEZrMAf
    75gckaUgwCC0kBr8K7TAHp/WC9fhrMAiRKnjCPUIwB+HoRJE0EjAHw4Z9SZ5CMAfpdisuG7Y
    wCCtSftj4ZDAIUErpFNYmMAfOaT5FXtkwB8BIdAG8zTAHp9E33hEwMAiTVyoWvr6wB6aslw3
    lYzAIRoqG+hB1MAes8Tozm/IwCAeswMOgZzAHoyaxbjlCMAfRD5Iu3VkwCMHRPCbxWrAH7h+
    E3UIXMAex/tQkwnAwB8P0LoZovDAHq+A/pykqMAiUCz6b9cswB8K/ZWDOlTAIKiji6RSPMAg
    Lhh7BeBowCAvn6Ml4rzAH1mwoSZd+MAjBpthcjv0wCGWn0Mz+QDAICNkz+HMYMAfI5WtirwM
    wB9k0GDXD7TAIcNxmMUOZsAfV9xdMh88wCGQa+oD+tTAH7HULGFUQMAhQbzJq+M+wB6kgsR1
    H4DAIMfuQshLnMAh0Cn80dY8wCBgVDkRCgzAHw9P0kG90MAfpOTxWsWYwB8QwDN2wezAICSK
    gCBVAMAhNXrh/chQwB8LonCne7DAIVh4r4nIUsAhc01rfBSewB8OUeKBcWjAIV4qbzFdHMAg
    woE3UwHIwB6F/eq0AZTAHtEjATNZcMAfYNXnPePYwB/VpqRQYPzAIRpkrfMd1sAe0aHrw6yQ
    wB74HPvOYIzAINxNwox8CMAgI03CNNBQwB+dBhy16qDAHoX96rQBlMAfbpnmBxGwwB/r6Yt+
    H0jAHy+fHctriMAeiwTuU3b0wCBuI2muw9LAH23qABhG/MAerg4gvkbkwB8TrV8Zn7jAHvoG
    Eq2zoMAg6ZMZVlAwwB7CAb+OSEjAHpAoUw/XoMAerklvn6y4wCFNzq48FBbAIKNSYvi+8sAg
    DNP165A4wCCTHOm8nkbAHpgIixz4UMAe1GrYrxl0wB+x1fh8OUDAIM3g18dJhMAhPt9odGaM
    wCI7/Pjng9zAIEYKqSp+uMAfL2c/o99QwCFOhijp3Q7AHq5Oi0YIMMAfF13S92dIwB7xW/sh
    Q0DAHpjDVj4LaMAfjMlUSosAwB8Nl6UHEBDAHpPIkpeyrMAeygvqK97AwB/pJDKhkRTAIDWd
    c0sZOMAf3fksiUJgwB/7Q9Rzk0DAIKcnasdBBMAfUvIXZiYwwCGp9zF24NDAH+lPr0GfUMAf
    2q46X65gwB95SYNgyeDAH6IY18nSkMAekHv/lve0wB8lXzWkb6DAHrUMMEXLUMAfUTl624Ug
    wB6Li5u+omjAIGKYIqsBrMAhRc/RATTowB8WptSeO6jAHoc4XR4X+MAe16UJ69CMwCDnjbL+
    QCzAHurUS5KMsMAekuYGBAAkwCJhYaAUxqzAIAkvPDpiKsAiucuk5jBMwB8oTeWUKeTAH1rT
    uOVKsMAe2fFQIRTowB6O2zBzo5zAHs8EKC5kqMAg7RxMPU0gwB7Db+9kBijAIFl5YYIpLMAg
    Sk7gsWv4wCAAmTq5xHTAHo0+i9xm6MAejBLmVuz0wB9R/aIqhwDAH7zZpoeYZMAf5Q/ujYlQ
    wCDvoCZ1/QbAIZQ/68699MAekbAQzY3cwCBzc83BFZjAIJvG28EEUsAfHXLRVrXwwCAFiK/o
    t1DAIKDBshLtJMAeirRvQ+6cwCIXAfIeM3rAH1JAK0AHyMAhEin2hHLUwB6N/pBNgSTAIOzv
    513hgMAf5eGa3ZTYwB/U5Nf6QXTAIPUUlQNYksAf3rKRs+bIwCAPD/iTtKjAHoiikKaEuMAf
    b4XYDn3AwB6+j+6fQ2DAICoj77AqOsAepDqakQFMwCAroQumy1jAHx0qovrITMAg16QMq3FM
    wB7Y5dpUKzzAIqNjwhfK9MAekOvL5+Z8wB6F/eq0AZTAH7slb0pdUMAekjy6ZJsswB+1FSer
    8rDAHoaG6+hIhMAepEez38MYwB7NsZjpsvDAIDYGmevvTMAf8lHJzzFcwB++54D5pqjAHulB
    HgUCSMAgIOaSaXl4wCDIUiZnzHzAICP/xM4A5MAgshcQJkEEwB7CBACYRgjAH6/SdZhOeMAe
    ng1iK0mowCAlLXBiux7AHzRqYpDBrMAfeoVyndnAwB+55qu+1KzAHz6o/pUoUMAe1jNcLLv0
    wCF5qo4DfVTAH/ibU3/FZMAgwCjki9rIwCABDSX9XOjAIJASxMFn7MAeoFXbtUcAwB97YOnJ
    GvjAIAjl0wsRZMAenWKIP9RYwB/wpyVVMpjAIbj6QomersAglgb8XMAEwCDr15y9PGbAHquO
    IA7lyMAfLXnBRgm4wB6pA9LJlvDAHq6ftodpRMAercCpMstwwB6WrvSKvKDAH6+x92BOrMAe
    pHkNOingwB9oUawBSRzAHwLnSTS6QMAhRLfEu8ZMwB8i5FG9q3jAHog6OehKYMAic2IZcEMg
    wCNGSZvW+RjAHxPkDVHaUMAffbT0wha4wB61ySxAOtTAIS/HT2r56sAfOoH5jRhwwB/jrKzE
    BsjAIenU4WmRPsAgA7eTp7mowCAwE6hq/PjAIIMsIyu5LMAfr1Mal9TowB6g2iL5qUzAHoee
    9HCNtMAeii7R+5y8wB7pQN2TNdjAIHWeYPyb9sAfXcY5+/7AwCFQGlzt/W7AHpQBWlXiKMAe
    yEn+lLp4wCCYKd9GMLrAITRiSLgRkMAejOyQEOkowCFGArQukMjAIEYfblVebMAe/daoY5MM
    wB7LBCcFXBDAHpEHyXSZLMAgLcAic1iuwB68HdlodSDAHq+yxRWYmMAfvlK3GpnAwB9dXORF
    5iDAILHe5kcSEMAfNfNw8N7cwCDl9ku5BCzAIFUBPpV/2MAgUbnMEqW0wCAjCDQO49TAHooU
    LsQ/CMAgmIqdfkqswB+rPjQDg7jAH0vVeEFJLMAeqrtKnRd0wB7dOzmw/9TAIob8ukdubMAh
    KoKMP6u4wB6uoyUksKTAIGDpOyiCcMAexvcHt1vcwCCu0VzaJzjAHoX96rQBlMAfVT13E19k
    wB7EmxIS+gjAHop+57aJiMAesXZPs00MwB6THpLql1jAHqhL2+ojaMAf4K21hNhcwCCjAF4l
    XADAInFpgfixSMAhHHfIJsjywCAdZDw9/YjAIFoEfQUqNMAenM1lWnxkwCFdsA64HmDAIQgd
    P9z3PMAeo/D7WodYwB6XlSY0JgjAIN7Apw0kdMAg5ha4HQxmwCC8d3VXIMTAIC11ofYbDMAg
    HY/iAmRgwB7Z45pgFPzAILW5wH9xiMAifxVwyKpAwB+lIOw3YODAHxpcnc/IoMAjD0i0aw20
    wCANtSpsRwDAIBVMVC/zasAg8jhwyEZqwB7WuNjXBmDAHo/Uza3UeMAfOOVFSAkowB7aZ8bu
    ADjAHp2L+fWEeMAgkCpbvC9MwCAND2ZmHxTAIELoZknuJsAeuNbTriO4wB6SkU768ojAH8e/
    pGM+WMAewXzvdQz4wCDqt53ErljAIGbVcFrAvMAgb2sk3TBqwCAhkDqxUH7AIr2j3T+E8MAe
    oDBbPDKMwB88PG8EePDAH8ruUzdmoMAhfuvSUtxEwCHhu4gV0CDAHrnhVNFPfMAfL4ZrNUuQ
    wB8uH+c8jYzAHrYs7Kq0KMAif0W6joSYwB6VVzH8o1DAH/9HyIvT1MAehf3qtAGUwB8ZNe/r
    mfjAHo5RJ7PItMAeyLwhJa/YwB7AA52+ULjAHvfg2a0NfMAgzkevuSnowCDZdEnAoNTAIx4g
    NX6XBsAeyH65/OdowB6ah0pu38DAHpXmM3LjnMAiIYTvld08wCKr8M2vIb7AITrGtNLSHMAf
    +ybCmZPcwCCAXl47jE7AH8aVN91GwMAh3IBzESyMwCMuhYCyNQDAHqGpRcKZ4MAehwVgehd8
    wCAW2IImuVrAHsrH1zJP8MAek1cuviuMwB8qOW7Qi6jAHxzrTwb77MAhmDL0WMiywB6Mf42C
    epDAHqIYQXuT5MAgq6nigQ0EwB744ZlRYKDAHqMRSdrK8MAitfb2Tjk4wB8xL+LaRMDAIENi
    eimY8MAf9WQDbbBUwCKDZWIE5SDAHx0FDUtx8MAejy47hj1MwB7wi7jGPojAHvm0YiNF6MAe
    4xWuLfSwwCGZoqYJ0UzAHtIAIJLagMAhkUY16akswCKlMzeVoQDAILbBAqSRKMAh6orr7UxQ
    wB6Lwac3YDzAIGzzEuu7WMAgmHfp20/0wB8MWS1HUzDAHzPR9yC8XMAekwJKl0sEwB64tI9B
    3yjAH8f62GLnVMAehgol88iwwCLFEhDv2ULAH2utMNMbEMAg6hKCjarcwB6UVqDXvVzAHzW1
    i0kCwMAfl5D5wN9IwCBrSz4HgIrAHwWxPSRWbMAiTycDdHwowCDO54uZfVzAHone/wRjHMAf
    JmSv7FbgwCKhRkFMCwbAHs/qYQq3PMAhnmRbtmXUwB/M+By6dlTAIEH8SRlWLMAhBSQvPH1U
    wB9JV32UpGjAIKctX37T0MAei1WpaEzgwB8T2TqYiWDAHvgcUM1XiMAf0vKMFoO8wB9LZh/v
    8JDAIvCE1HhUiMAhaI+UJdrUwB8ZrnIrAyDAHpGT+slMwMAeon3Vk0NYwCClQ4hEHurAHrje
    mTGYEMAfZIdh2tN8wB7f0ao4JYTAHqrRNp05WMAewnbZrZMAwCCk4oGoCSTAHwHFx5WUrMAf
    S2g0Wn2kwCCcp4Z6gSrAIGLAS5IFFMAeo3zRg4fgwCBL1GLiHCjAITTP+TlX0sAhCX1Pln7i
    wB9YIAjLvUjAHv3FSDm/2MAfGt8U5ezowB//j4WXMIjAHuda32Lw4MAfTlnHDa14wB9SAoAI
    6KzAH6YJac9h1MAfEMYmDaAMwCATqz+I0OTAH6lFUKzYWMAfB1xdIxfIwB7D4xoo53DAIKuB
    OBK68MAe31tgJrvQwB/nzYM+zDDAHqLoz19JzMAeix/QyyCgwB6ogrcHTmTAHtO6669FRMAf
    G/h892AYwB6YipgYgqDAHqnHjmvz7MAfIMlyySo8wB6q21c2rJDAIs4zQPiSKsAfxZuweG4Q
    wCEKllf1EWTAIBSzBXbwwMAekzre+R6gwB+oEdhXJDTAIGiDnpXABMAguxbclRpcwCCw5X/a
    E4zAIAeHhSxWKMAhf5Ha09aYwB6rxaSpg3TAHsikBKrTxMAf9v0NJQ1gwCFRJ2Y//jrAIJCR
    IPWWlMAexsm7I7JgwB8br1+03DTAH2DBR3QGKMAfAKZO4+IQwB7UTreKT7zAH8V9pbMOBMAf
    R1k6z5gAwCAZyHasM7zAIFjPi5O19MAesJbbUuO4wB+ViorTs5jAHqp+TH6JDMAewrYVRiPg
    wCHuc3+2raTAIk4mV1OtbMAgn7yM7KCSwCAPCwl3FjjAIB2TwdefuMAehf3qtAGUwB7wR4tq
    jKDAIJ+f3pIloMAgaU5bzZfiwCA44tgUQcTAHpBEMu0HaMAeh+xlGv4cwB7nu1y1DADAH48R
    nbR72MAgR4xV0yVkwB+Ob7aVLBzAICQEkvPQuMAesM477bLgwB/VJgqqbeDAICNYLvUEaMAh
    wG9mghUKwB6VOqJCZQjAH2ywQsA6CMAeq4DO0PoAwB9Y89GDFITAH1Ys72x8rMAg5Wbb1hts
    wB/+qq2cksjAIfI4A2beosAfcaqhBvqAwB7cHLXg8FjAHya9oyyTwMAfIsaWLs8UwB7K1Ccs
    czjAHsJEbgnxTMAexIcl3gzAwB+/MLk6M2jAIJvxTWf0/MAgaBPARGkAwCA+qgPpJyDAIEFc
    WqA6eMAfhamsucKYwB8nAm6XKcDAIFtDqfn9eMAe8H3/DHr0wB8n70MG71DAHoxkWwHMfMAh
    s/8Q81vowB7gQYQS7CjAIPzq+FE5SsAf0MzEW0fwwCAF1RKrcLjAH/R0e1kG2MAe7buPL3YQ
    wB6ul1Agz4zAHqtdZGj8UMAfmg2geuF8wCADX5UEFwzAINxadm3+KMAhx342L2mcwCAfpKbF
    wyDAHokLgD5LbMAgtm2shyrowCCkHxld+fTAHtEoB13s6MAe8xDF4JtgwB+iDTkw7TTAHquS
    u9+0WMAgQtmBCrmKwCJrmLTx17DAHr+jdul9IMAejZC3wyY8wCKtc8rPHbjAIJ6rM3CH5sAe
    13FhrauowCE9hremxVzAIPR1wjKyNsAgVn9zFHmcwCCjLeoLTnTAHv9wpVR4cMAiTXyvOIbW
    wB/fvN6NzFjAH8f7hSjTiMAem2Dy48oAwB8qokEycdjAH/74XmiQqMAej7j/YmxIwCD4z9sd
    tJzAIIM0mwIcdMAfDtZrcw2owB7Ti2FpplDAHt0Mup9QYMAhhycPqa5wwB9FyfN5khDAILMr
    puTKSMAe1kwiaCaQwCAn+WocY7TAIIn4ZbHzkMAi5m9M8zvswCHkT+7z0azAIXTuiiSpdMAf
    JcqOXbx8wCFihzY/k+bAHtUBweTwgMAhT+Cc5OmswB9vH2gm34jAImIrHxLCEMAgLNgWuXa0
    wCB0nMyCzCjAHwJaAOEyCMAgdhNHWVVswB6x/Lw+s2jAH3m0vynz6MAhJSi4qfd6wB/VUzhy
    b3jAH0ZYL9m/EMAhNhPZl7ygwB88RIGTDJDAIKZtfitOFMAgiv18QcWSwB9T5h23iFDAHw6v
    vUQlfMAercAr+IjswCBQC3iVasDAIE9HHMz99MAgQBtW0JKgwCDiMYT5xxzAHuFbfm5XkMAf
    VAJb678YwB6x/7mwnFDAH25pK8aolMAflcUy6pUMwCDNLsuzptjAH8jDiwjaaMAhtDBX9Cta
    wB/FYYFuC4DAHsvvsfVMKMAiXe8V36wGwB+ENCk12vTAIAuCqxiCSMAeul681B/QwB7vHxZ4
    aGjAH6051/U28MAelsuvgaBMwB7vQ61rQwDAIgBJRD774MAhnotlBmiiwB6F/eq0AZTAHrSI
    d28SoMAfPHrINMDAwB6M0UcN8BzAHr4qSY/miMAgU5i2/qAkwB6lP4x0YNjAHqmvgv505MAf
    sYQyJG2cwCD62xdh62LAIR6IR0A8yMAgCcxgeWoQwCEukzwvsTzAH+IJ8CaN8MAgDwe5oQR4
    wCBgO9gtU4DAHp4dvji8oMAgJg3V1DmMwB8eONbmBajAHv9kplzWwMAe41C5k6oAwCEvTlBd
    y/zAId/LKotmuMAfzgfLTsqowCAJ61BgvQzAH10UUjAroMAhPwC4vUsqwCAFh0PmA2rAHuSf
    1b9SMMAgTllQYA4owB++vRY6GmDAIB3+pMZqHsAfo+JTRJAIwCD/u+WTbvzAHxQ1sHRQCMAe
    5Vu1VHSgwB6k80CN8vTAHtj/asD1XMAhiOx8DLSUwB6iQWl0+kDAIyVjXNVHpMAfTz1MbjOs
    wB6IblFlQdjAHquaHg6juMAeyRqZo9/MwCAAjaD2vg7AHofSiWVaKMAfMRn8eo9YwCJPAgWN
    u7zAHp/sbc4KuMAeq666aQA4wCAG69LjRhDAHon5kxL6AMAgyWHemzACwCKhkigY3oLAHp0R
    vQvfgMAfMK+orQS0wCAggw+st0jAH8cqcGNwaMAf2UhNzLDIwCEDrLW7gaTAICZSq6ZDnMAg
    IIn4r7bMwB6a0XvJEdDAIARC/OlixMAgGIPfhgikwB6aCAxcvITAIULtAoGJ5MAekehIzawA
    wB8CSz9T84zAIANgI+iUjsAgVZ9WG3ECwCAzc29VP/DAH0k5Np31WMAey6PVrCwUwB6UOjBx
    cyzAIVYBVVhQHsAenUBabGqgwB7DOl2cXWDAHpKtiDDgzMAfNcGAokCAwB6JdP2E/NDAHuIU
    yWZpqMAejawmMz9owB609fRd2zDAIQWCPrJiJMAf4vv38BtwwCGlnwbgNMDAIKbUX4QJhMAg
    yy1eY4kIwB9seKSpdKDAHuvaJ63p4MAeknUZmNLkwB+AKU11SgDAH17jwPZ+wMAg8ugMTgOc
    wB7559AnBLjAIEp8aIk/oMAgqzdTy/0iwB9cS8BxAAjAIGTedMcaKMAetEKbeiJswB8kQ43b
    XpjAIjXMazGFIsAg316W1ObUwB7hAMaqA3DAIEs43JFhKMAgMeTvlyHwwCBLJ8XhgnTAHwFX
    JyhN+MAfcD7wEKLAwCHHwyn73M7AIAhjwGtewMAe0f3TJ4ZIwCC4JajEujrAH6fCGqRFeMAf
    8MZdt+OwwCGzxJDAyRbAIdQa126SkMAe4Q8JeghMwB6ZNP+IfTzAHpLJxTWOwMAgxM01/1Ze
    wB6SWOgPo+jAIU2q0Al10MAfY3bEcajAwB6tAbHDBZjAHomPcdYAJMAiJniK2NWywB+EWm6/
    S7jAIBJCvdAU/MAeyM6SiKqowB6/+hbXJVjAHqLBf5LLUMAeoOGucJ3owB+0xtW8KFjAICKq
    Vbrg8MAfG+iZ4BCgwCAzpMVmrajAIbKGb7e/ssAenXOd4/C4wB6Lpp5fzbDAIOieWL/DpsAh
    /H5uZD9YwCC+TugG1BTAIZADGexMmMAgV3I9nKgswCAcvaPoO7TAIJYJmO1JIMAexcnjc3W4
    wB/OWtR100DAHvobzyQf8MAgclWKPZu0wB6RzCrQ6bTAH/iWc5FnKMAgPa2MBwK2wB7MC6aO
    8vjAIHrq1YkkCMAgfpccycbawCCiOVkoLvLAHqCzIODa4MAenpaoXZpIwB7htBsaWDDAHzal
    durznMAfzM4U9VRgwCCyAeQYmYjAHr55bGbuWMAeuIT4DiXwwCCceZXzu1bAHwSt9ag7EMAg
    AXokaprcwB6JQC3MJ5zAITD6Aq/kYMAe3DaiwrqAwB6Y6qHg3pDAHrENCLNboMAhjHhWG0PA
    wCBMcu1KslDAHq2jN8CjYMAiKZPJA6fywB7K3uzSRGTAIOTuz6ylcMAfKyiwLPewwCLcONs5
    PiLAHs5WQq9sHMAeiLy+sEGkwCEIRSFXK/bAIAuhuo2BhsAgXe+FWiAQwCC0y0QBjdDAHpNz
    gjUZmMAfm+1n5P0YwB84svhLDoDAINIurcpW9sAf+I3RxiZIwB67uOOoAizAH208/FcSeMAe
    0nWlAPiAwB7lEu2fdtTAHpIEar424MAf9URpKJ90wB7Gvz07RNzAHp8G3jLR2MAelh9+sf6c
    wB6htkkPU0jAHuWlvLUJ0MAfmqY458JgwCAuoDH4LKDAHuSLUJJiVMAepmefsFPAwB6pK9Ou
    9uzAIMt1/t2vjMAelB3DmhxAwCAr6XpqkPLAH/bMB8rUyMAfg6G/5nugwB+QAYegkcjAIObs
    8tjafsAg9sKfXwWUwCGMCzm3uhzAH1hNq/9XXMAhgDXYfiWgwB/XgOmq06DAH2rUju1aMMAe
    /o6qzqXwwCAaI6/AeITAIUsxRvHhEMAfUhKICTOUwB8PVzEph1DAHqLYRg2bfMAhM4dq4cMy
    wCApu+LsezjAIkRcOQx6eMAei/fLQ9a8wB6r0v692HDAILk0KF0kPMAgSmTA+TcSwCKadyUu
    RhjAIlpeDq2x9MAekAx3jAdwwCAYuVsnPIjAHsYmKDD6NMAgKn/VwMoewCBQt4PdpbDAHpzb
    azuDEMAgezT6dr74wCEeaKZz/djAIERE58BDqMAfRKJ0sUmowB6oAoilA2jAHrBtNLbRmMAe
    mtgUXSLUwB66NisMukzAHuQezXwsSMAhpp9vcaSAwB69d9+KPzjAHqBHdd5uEMAe34ss0yhE
    wB7rzON0GAjAIQrpCsp6KMAf9SyX6Jm4wCE+cAa31JjAIuqUnOGzMMAf5UMhVDzgwCE+FRE4
    O+rAIbpKcC5VNMAhOuo0Hzc4wCLcdDH2pwDAID8z7IEzpMAf13HkFGkYwCAwGBSsMIDAITOx
    YIGEEMAg9qTZVoZ0wCA4AsoPoVDAICmPD5nTFMAgrN+DPVzwwB6MLgdcxezAICJ6WOJoQsAe
    tMCDYrLowCKPTmU0fBzAH3SfPdxtmMAew25D6ReAwB9a0/YE4HjAHqLGoWPvmMAgDzimrhmS
    wCCm3KTTCETAH/EkTYwj7MAhVHia7gEEwB6a6JxCULzAHvdq2Sa/EMAgURa43zZIwB79eYnn
    qKzAHsrAvIiAKMAha+s2+SBCwCEwnCZyZkjAIlGFx7plmMAfncYiQO8owCDIujZdDCzAHrfy
    v9nC7MAfOyOMbwggwB6Y8pTBSYDAHo4aEx/ZaMAfYa3qsnx0wB+ReREkR+DAHtPJp+i8EMAe
    z2aCRhNwwB68r9dGpSjAHs/ohD/EgMAi5bx6BUcgwCDJUVgbn1DAH2rn/uEu2MAgRCcqDz8S
    wB/I+Y+7s6jAHshvk3pqBMAhctQQV1OkwB7OXvvjFDjAIBBcyA2iMsAfGvUdphTAwCCK9TXH
    i8DAHtagL7+/PMAfEhyDxyjAwB6N4xKDdZDAINi2m36eAMAeqTmWSuz4wB6VAY0VHbzAHyFS
    xDC+oMAej2Wva7IUwB/9JeaVu+jAHs68s0NkgMAgMdZoRcpEwB7pD5lTk+DAHsai5HD4MMAg
    eEDrU6KwwCIJkNy+anTAHrNSX0NbQMAekSPLJJLUwCDtp6R5RKzAHuE3CqhaAMAgs8UmAUA4
    wB9MZKbnqCDAIPa7tRfdesAhbyQ48iC4wB6XBS8hs8zAIOEAL+jdYMAfK1PItHcAwCH8clo3
    hMrAIfi0BUMCNMAfhX0VqctowB+2L48gpszAH76WFsLYMMAe2/3aE07QwB734shxEDjAHrfz
    m2agaMAfWW9XWapUwCGRRxN2b07AHqwYSQd2/MAfA5UO9jwAwCGG+UVP5vjAIDO75AbhhMAf
    0scaCt9kwB69wyR940DAIZio++6UasAe+XVUVYDYwB7ndbekcNDAHrIwinXi1MAilDNqXwtU
    wB6F/eq0AZTAID8PXxeWhsAeikl8nXLowB/qYQo6VIDAHyQ6Mpf6IMAfa3crcnY4wB8utQk5
    CujAIB9FWOmYmMAiJRvlJWo6wB6yEMBKBwjAITqDU6qihsAf7n5cGDbEwCBTnSigs2jAHsiq
    m0D4qMAe0jQ/C/sgwCDfheBVH6TAIA09ZGE6hsAf786T0sZUwB85W1qOfeTAHoe4uRPO4MAf
    Q5Dq4MvUwCEyyaGcou7AICzC8cBj8MAesafdroMYwB/bwc5uRTTAHo0H3pJJcMAgrF+jkc1E
    wCAfKN/8okbAH0K9PgUq6MAg4YpeeboSwCGOcrDK37DAHzm9lRI4QMAenRTtH9tkwCEmxvIm
    h5jAH2gqb/fkIMAevW4tPtoQwCABAiiIkaDAHohUQJNRRMAgX4YosNqYwB93MNbra+TAHpCz
    3VHwlMAgf164PqC+wCAQk9RRHHbAHrWkEjjRKMAgDY5KXAs0wCKTvjYRsg7AImpljc1jcMAh
    IZGA/VNAwB6bd4MakgDAIB3W/SXiqMAf6vR7j15QwB8MaOuwqdTAH1f8aI4SHMAe2bDaXID4
    wB7L/iC2fmzAHu+TGptmqMAfVnmQ0QkIwB+MvRQwNJjAHth7IecsQMAgvM4uGN+UwCGGCCdg
    vSrAHrXAQnI66MAe5RpmpY2QwCEdYdbQ+vTAIjDg5Jg9jMAez1GpqBgwwB8ZB4xa4xDAHopk
    Lo8KVMAg9oYorRLEwCF9PInmW5TAIhCxCnqL0MAeortNXuKAwB8IqHlXyQTAH3AqXaYxQMAf
    7+UezB64wCJFmuBpJRjAHp6Bp8+I2MAfCFz8PqxkwB7FB2ZX18jAHtbFbbVyMMAe2paeVs7o
    wCHXhvDfRm7AIGzf3AeiJMAhO+2UpdjMwB6X660xWRjAIMOtuTUU+MAhyur8TGZ2wB7WxK5B
    P4TAIEZJ4wg/8MAhf+WTaKjcwCBhTbK719TAH1+7+tF9aMAeiSXSrNJQwB7y7fpG5LDAIFQZ
    EYNWLMAf8TSkNaY0wB7h2oxSgLDAHqnv6QkYoMAh/iVzyoiUwB7juHgSwqDAHssjaSfOYMAe
    mF83ftX8wB97F15fucjAH05/BvwX7MAfz1+six0YwCFYTj4oIpTAIvR2cT2gbsAfDflFXDQw
    wCJmuiYYMlDAH2MjvNkE4MAem8AJDj8YwB7ZLCGXv5TAHwwXWx7XeMAhXOp2CAiSwCFGrRbu
    bzLAIE1s1QBVfsAgFL8SSWqQwCAT5JAnYbrAHv5ZX8Ym8MAgaIVdThfgwB8S5+cEeUDAIKmP
    LydwOsAfjOpKs/HcwB7h+P+WU3DAHqnSy28d3MAfjMJsYQ3swB9cfHvfs/DAHrgXpxlYaMAe
    niQEXIswwB6WWNcwTjjAHpfO0mTNxMAiAz9aOvkMwB/912cRaUjAIStYi7LLcMAej501jlqk
    wCFJr/kFE5rAHubfaCYxCMAfwXN+sVQEwB6eruz/SPjAIKHh5JDT1MAgOtXhBtrgwCCSahnJ
    rcTAImcVkrZkAMAfa8XY5rKIwCBttsdDgrLAH0kVD6LrWMAfH76p3S5QwCA16Qqge9TAIAWN
    6dtwUMAehf3qtAGUwCCQu3UCQ3DAIBhX9lKVJsAeuS/If+6QwB/warJNAgzAH754rOKDkMAh
    j6z4wTVUwCEQi3YWdQDAIn1475Yj7MAemdLYaheEwCASPPi8BVDAHtzkk8aiLMAe/lCj1P0E
    wCAGEkLj8PDAIs+lFloResAfHFyUJX1kwB64AB9lpfDAH1ZljKl4YMAgQgN1mxvswCK4fY9g
    CQDAIIlztj+y5MAev6xrr6JQwB6LOrnY0BDAIXyYTWxouMAgESBTyLmEwCDkgAdkyQjAIPJj
    Kc7fAMAgn8ww8jxOwB8UiSpoxujAHyDlX7UUmMAe934BohaYwB7O8kOIVODAHovctjPDQMAe
    qzzWGQ58wB+ibjqN8VjAIK7Lf/Lu4MAfPixeZqswwCAD5YevOPDAIGVQDQ+uTsAgQw+5nk90
    wB9Zx52kNFjAHvUjMk3NqMAgLxueFhvgwCCgYlA8jTzAIN42nCOEkMAftxhK+D9QwCJfzByw
    DOjAHrDtc/QgAMAgBKM1eFxcwCHdVbL+2V7AIOAgq8DdQMAgMVvQNAaEwCBSh5yL5cDAHvh/
    xpdl4MAeoZRt+ScAwB64xnRXK1DAIF3ad4ScesAgAw2v+u8qwB9Q15WWNzTAIQnMxgbNjMAe
    o3oECPEUwB67esRz8STAHs4JW+e4hMAgWSqF7qVkwB65jacEZBjAIIZi1/Id1MAeoXsim58o
    wCAI3unv4mTAIK9YKPrF7MAiQDV2cf1swB6ycpTtSlTAIMwDp8LXDsAh06VHFdlIwCLZvt1K
    VQjAHpY8KUrpxMAfFY71H8EwwB85auqoy6DAHvrlIBLwBMAgUT2TnJUUwB8CHQwJa1jAH38b
    FiV8zMAgJBYmMgkAwCIU8R1oEqjAH5f6dKMvCMAfS5YPuEewwB7B/D5BnPDAIPwK62ATOMAg
    dCa2BBKwwB8dZ8allyDAIeVR+CKOhMAe4R7P17dQwCE87/AHrJjAIevMX9Pi5MAehf3qtAGU
    wCATKYL7URrAHwYfPOXLGMAf17wlMxIcwB8d9BrVX5DAH8dFocWObMAh/HRXv0ywAAAADgAA
    D6DAMO+bohXS9cAwURVJXI2WwC/xW6dqNATAMv2HdU7c5sAyCYhjMRbcwDIkAmlxQhzAL/07
    FLp/msAxdCLlpDkrwCk2SN3KINbAL3fc7KH0J8AxQcUh32FvwDP+sHfKcBTAK7/4Xqlm9sAw
    twU/FAzQwC5UTAAR8L3AMiHqK3Pke8AuBMCnmUdEwC9NI33Oi/nAMO5NVC6XGsApcpzwFrmB
    wCexkvyFZX3ALP+a3Y1GGcAwW/2EE4uHwDIhQS9x8QDALJyU110DUsA120tVfCdSwCo4urRm
    gwrAMHtjmWqiacAyCeGuhWxowC+zflWMFtrANUfdNMdLcMA3fTPtruvgwC486QLuicvALS5r
    tXZzNMAw9WRLp9CSwDMIIN/k1MfALmNKFhGZmsAsrrxLQgQvwCyuICmbhYDAMQFDar/y3MAx
    nE6z+10wwC9sK6BpBcfALUBLQz27RsAxlaGpn5RawDWtc/AR2tnAMcQBVPaXVcAvJZ9USaOi
    wDDGLsh/BEnALEKPZEUdFMAxL+p92oxLwC1Qv2EfP2DANYzGn67lDcAudo4eZ+HPwDZk7Yq4
    GhTALJAEzOR6TsAxYKUsCVQMwDVWZj68WKXAMEnsc8ua8sA1F1JBLzhswC1bMhE5AGnAMV8j
    0JmsTsAxac9K0TYiwDKSv1z4mf3AMIUNbfXEiMAtWs1UhPqSwDGWPp/OaOLAMA3Ose1aNcAy
    qE61Lz0mwDE5Q2yn3fjAMOTk5IsUisAsIXoY9xBbwDPV7l3Fa+fAMxAHIse4dsAvlQEMJbt2
    wDKj8fJrwuLAMfC69djrtcA1j3laA8t+wDKjJKj+gCrALjfYFuRnV8AyDm3vuRdiwC5h8oz/
    ov3AM/Pn6zWqZsAyyXwR5q+hwDU4Oe1ucf/AL5CELjMha8AwA28JIUlBwDM7UrfZgonAKVHf
    PlvyNsAx/a9WyGESwDAomirlmIzAMdV/1x9SisAvt7sHbuawwC4UwOltAZfAKBC07bB2AsAq
    cUqG/K2awDKURU1Xg9XAN9ytsPn2V8ArbBjwYxpvwCWsBqVwTIDAMKwyVxTvWcAqhq5jIIxg
    wDE1XH5IpGPAMoxk9ApzE8AsRWV/YYsUwC+xP7SzdYbAM1zxq3CsXcArIromTzcIwDiFoOnP
    pL/AMpoYkCyfp8AyrwUN2MbxwCo3Ws8lSp7AMaiiRB/umMA2k9SelDtowCqkSlvcyJ/AMXyb
    cO7KOMAuVa37FR6awCozCJaz+IrALH5FzwrUxcAxt8e3CqtQwC+B/4ZintLAK2ZxXxEae8Ax
    DD8474RzwCbCFUTn+YTALwGxRJavZcAttIlvMreswDN8aSAw3r7AMCs9G/BTGcAyooxz51G6
    wDKq14nGXLnAM7/hdgG138AuOUkseMwHwDbeSl/YJvXAMJoWn+TBlsArFhywIGQDwCkOY69/
    xN/ANjxXoZnbRsA0r5poeZCiwCzyiQWNqrrAK7xeumUf3cAwE0p+E5xJwC8hGDCn3P3ANsFW
    8t6pgcAvwCsuMNbywDFpShCPawDALWnk7MKacsAtEHflxuUxwDQ+8CrqQuPAMZmC4tT4HsA3
    ur+ybuktwDCWSPV+RvvAMXE8q111ScAwELY6xKRwwC3ADLJlVRfAMoxBKLXzBMAvMk3Fvteo
    wCtFT7DElYrALGqbTo2cTMAyjYPhI+cLwCvqcpijtFzAMMbXOblNdMAu5A70u5AzwC/kLu2c
    H9nAMv6K6hm/DsA1ccN1llxWwDEyA67fjRPAK1up8Z98/cAy7ekgtzugwDFvJo/Y5xjAK8m2
    eg7a4sAtzsSkA0+mwCntg+cKh4HAKRhkOefK3cAokTuCe0OQwC5MV014CynAKn3TK3/9PsAv
    g7HfiOUgwDWb8HZHzdbAK/ZJjHG2BMAv9z2E20hcwDAlY65ZydDAM6LRIh+cCcAuCQz4893i
    wDCHkzJrwj7AM6tciop2xMA2jRpLTrwbwDWxS77cNPnALwL0B/gF4cAzai39U7LiwDMYFKvc
    73vALGu1fwIl8cA0ubDh/plSwDLc4eB04LPAKnrYO9I8qsAxkOZCd7DAwDE32kWIqATALyha
    g91X0cAzAPikNwUHwDGWtRAADobAKR00jE2sD8Asw+wtdOdTwC+CTFKv8H7AM8jCS6NTdsA0
    jMIuhXrcwCw2AEqiuP7AMFmmeUM7GsAxqO6QCuuUwDT4k8qK3jbAM/mRZIjY2cAu7ksQeTHL
    wDA9Y9JTs/XAKWB0EFWgkcAp70utrA3QwDDSF4feADTALbOYLNBuQMAvCofyTMTcwC3hxn1c
    PiDAMHmF7V5n9cAoLjUmU91KwDIBlwuhqpLAMf4J6ASIs8AnQqSocEJiwDFx0suAvJfAKLY7
    voKkCsAybkKtBxjfwC7R27N4pHjALpx4LjXH88AvGyBC5C7TwDOp8JNCsqnAMG6vN7nqksAy
    hs00jFr/wCouo7LolVzAKn4zsdHbV8Ax6mN27l2iwDCABrrmM3XAM482eJDbkcAogdBcnVXu
    wDJ0E3z2CRvALJx4GKD1DcAqUW0N8VZIwDEVdfUqRIzANc7sCcQa18AwEKZPRNmXwC3BumiA
    2PbAMzxLsxqTzcAzGo4djuRDwDNSBVtfNBvAMUypz61RKsAxuelglWxHwDGhLF797/DAKZOa
    /AHiesAsauwHv2w5wC4hHuGJivnAJTm1K1FyKMAzBXZxu0q3wC5rboyt5EnAMfAM9UkYY8Ax
    wH2e34GZwC05h43hTDzAOIroRfIjqsAxqdiEYiEgwDIwUkJoLzfAMw63oQA5tcAy9Ysw7BZj
    wDAwzn5Ks2PANCKLanou0MAwjrFmYsuCwCxd5e1VWZzAKBAqMdKDPsAvXpUaEKUJwCiX96Z2
    bbTAKqnJPzqiBMAq6F+GigO+wDB7sfL4EbvAKRQT8ugSC8Ax5YkZGA4MwDA56gi9BpfAKss9
    pnewGsAx+NUkvfH7wDK7KI9X8EfAM0sq+lOuW8AsSiNIpjxtwDH3HvMLv2fAMFLZngrKc8Aw
    9LYjQxc6wC6yXo6Ro/LALdzcTSzB3MAxlFC3Zu3ewDDoBePHQkPALKgsvgcPhcAwAa2d9wAU
    wCkx/6hUnwPAN+MbShrir8Av39FDhCsRwCku5g2zqbbANSqlnYoG2sArx+3WyN6pwDFaFbil
    UTzAMOKxw4I5gcAukIQRy6P9wDDNfIB4xRnALcA0zuoL6MAw5kXZr7U3wDHTB99Fik7AMpeq
    Ub7cecAw6ded3GJpwDMMp7XveazAK5RvEq/zi8AyHs5+zshTwDHdPD6uoerAMuVt5lXiicAs
    eIYHI643wDIAkFH0Yc3AMiaUQWZimMAumRbyhrAmwChA1TuGKjPAKaWkZXWXGcAzLq9mMR/X
    wC2/r64bd97AMVLGpp4xYsAsEY2AYQm4wDJIPr3n4FvAJOq9uAiWhsAw5g7mdzouwDPWcauM
    lvLAK9O73WHTqMAw8gd40SqGwC6Nz/rLMBTAMv0WqPJYKsA0NvLMI81YwDh1wVHJIA7AMB+m
    MoT51sAyNpVDn7z5wDAc9hZp7tXAMioeAHM9CMA2PJc6XVq+wC/Wy1DHCk7AJu12TNa4zsAy
    A0XRy4UXwDAoJydLBzHANA6AxhWYVsAskiLSjL46wDXm1mdRtL7AMGezzXYZLMAzesap8mdJ
    wDUNoA3a6sDALpipxtaojcAqF4dF5sJjwC/YiCyCDM/AKnioWc/F/MAvx6ISfuvQwClX91c0
    eLfAMI3unmq8GMAyaRswKDfRwDVykxLN2PTAMSWcAAZw9sAz0ZlYvN6XwDKnBzuouEnAMmrV
    WjTsi8Avh2ropNf0wCS1Pw41HijAMZ4f+w1VosAwpsET8OIbwDLTWgBkUnDAKU4T91QOHsAp
    9UVxSA+rwCyxzW4SVrzAMUa0RUus0sAyQzSdvwFXwCw/FKE3Z8bAMwt4nd48gcAw4959sTi2
    wCgwSBlXBLfALd3cPIJ0osAmA0onJZR4wDHgK6WScTTALhHyxyagy8AwZ1JGoXakwDIKSiPd
    e/DAMN7I8k/o9sAvLQIww3iJwDA3ftu5nBPAKvSyv19Gx8Ap09ao7Y7nwDGZq0Uc7ZjAMOi5
    abt/3cA2QFNwR0a+wDVQQCa9iOnALVI5QCmaCMAyc376rQoqwC1jjkIBAtrALwQulQJW08Am
    XyVBErfrwCosffQeGvXAM0M9mclj3MAznXJ1NCGcwDIt+PC56jnAMGUVS65R+MAwffgOwxt9
    wDAStfIJKhTALAMnfNxyNsAykqKZEN6IwDNQo46UEe/AKzxHdNSmjMAxkKvpjk/iwDJTN7+M
    zJbAK60ALVgO2MAugxKuvYpOwDAnXR7dOsHAM8tjWnLDiMAyF5LIbv7ywDLKXNsnXO7AMb0O
    MrLZ88AvgeXRi4DuwDDECoVJpMzAK9LaDnkBxsApd4fD5Q5HwDHfABjbI6rAL6WV7uGmpcAz
    /469r3wrwDJ6SEV4py/AMI9l+DCDL8AsyHTTF7YgwDKUid/WUxvAKI3PMLAf28AwF6cu+QkD
    wC0RFeKDrCvAMAe5PDhVNMA1/oWr8woLwDConwwqmWbAMUK/HAGm6sAuiMjta8ZQwDCxFtwy
    Oi3AMhVAZ/6n0sAuCKczwXZOwDAupemwr2TAL6c05aLedsAtxNZo+ZhUwCObXSjc2bzAM90S
    KWqnG8Ao92eX0TzqwDK6v1gVVPTALZWLri4FeMAzJLjCIA98wC2ireXURPjANZT9FaE5v8Az
    35G+XQP6wDG4d6G6uqLAN10JOmByS8Aq2xB1OSyJwDEm7HZB8zvAL4n6XUJzR8Aybl9PMkeL
    wDGfPSKN67HANXf3Idt578AwWkDrGdd9wDTFl4kyu2DAKIP9xf2Q2sA0jiFCn5oWwDRBY17a
    kjjAKpB0HGPurMA0aInLplAswCrUqk/hW3rAMQ8KCVWyl8AvoCaLsFSWwCxMzBeQgHzAMFnm
    FudHkcAh6t+TywdfwCjbEy+nfL7AL7JY3UniVsAxUEUDERUnwCmjFFXO8Y7AM4dcduDQPMA0
    DwTkJ8pgwDCoAgTlcSnAMMhsItK2y8AxUQlPqKIVwDA1/TNPPcHAMdryO1nrx8AwD7SPC+a7
    wDJ6wz3IC/HALQrIww7w3MAxE+GuwqPOwDRhgWDFKdbAK//qXHvPqMAzbq5vTaR0wCVjPQsX
    ToDANHGrjXwp8cAwooxq6A36wCvafEpMkkrAKZ6xVPWPUsArMFbouONJwDI2tyD6ZR/AMVV+
    KvP64sAyHDhnrqy/wDPTPv+9sL7AN50y3bwJYsAu+HBAqav5wDDSkgS+wljALu1L68ZJucAs
    exVwJkOiwDGWCB/GBCLAMT9RqNF2k8Au9utIfUYowDFOK4VYl3PALnZnMrLj5sA0M7/i6epA
    wDCJmi6jV9fALIvZ8f+SrsAvvwFrM752wDMBKpjNJmrALN6XLid9QsAvvkLkRofxwCxVyopy
    U1PAMNJuW2rzJsAzn3IrAfclwDWFOGPMI73AMdpanq+dV8AxKrqc0r/SwDNj+6aEITXAK6OX
    jtGCw8AufdY484jzwC+BcXP9d+TAMzhEBTV7pcAwlVqL6o1ZwCxAh1hppPfAMC/F2+5x38Ax
    ZV+ubLebwCsp+E8tRE7ANBkTToyXncAthubqvwRqwDHNQ9BNc7zAMXCc/DDfI8AzZ4hCBwnq
    wDBdHdK99P/ALEyd4+TuIMA0AcbrpvqAwDQ9hdE+47fAMs9T7SoNW8AtmHrQAjDDwDPTa6Sa
    8oLAL5pBuzaDWsAyXWpTir3MwDIZvKNFst/AKKrEnOHbA8Ar6I3rnBAnwDKYfNYmpAXAMU3y
    6SZZH8ArtgccCRSYwCulsAYiShjANAyPeAfghMAocnlFstiLwCul5rM7pyPAM2khwbh71MAx
    lWzv8B9cwC+gpMs+iwPAOZIX4WP70MAx22deLaDXwCkwKGRpqinANiNDyMyLbsAyuWu8r5vT
    wDAkd/4O6brAJ/v8RGfX28AyCrvE5N/zwDGWRnjbWbbAKWiE3RlpgsA0riVFiSR+wC89LoN2
    EnjAMeYpNlUigcAxcN2bDpNcwDA+3G6FckzALJPgZhh22MAm7UVtpIquwC3QY5xVD/DAM2Aa
    ZaIZ38AxdC/4l9xpwCbaKobFmTHALUQ1E7013sAyrDD9/W3/wC/hB3n38ezAK6F6Sd9mHsAp
    jUFKUeNOwDBoh5wMcZjALN9ovjCUb8AxFNXtWcXBwDTz4Ff7u3jAMKqER6KPKsAxBPhceKyL
    wDAXwSJ9k1XALT3NijURFsAyOE1paWEtwC3CwQQu31zAM6agWjwxGcAnFwepcAzGwDZ5KSGw
    zizAKgIEt6GsBcAymAdSs4UwwDE+iW6YYvbALvJbxY/oOMAzKIbgQtM/wDJjOzT15W3AKruh
    Wp7R6sAxj0fLw0CxwDN+jSJh1fDAMQKTYhGmg8AoMlTUPH+OwDFLQpLLjjrAMUJoKdF3K8Av
    SCj8qAo4wC2t4LTVfqTALpmNrJ1TU8AnRRpe6TqjwDBfyKH4TEjAK+gId38DH8A0z9739Ptz
    wDDma2vdMD/AM1elfTiKt8AtO70X5oa6wCzL1LN0fD/AM+pWifvWcMA0ivWaG4gWwDRagPaO
    2G/AKmFXcQ3XNMAvBhf2LUHLwDD3hqvO6Q/ANB+Uj2uTIsAy5Gry3NjiwCbuaXvM5PTAMeMc
    dUhf6cA0YHXcB6c4wC0yNivLVyTAMG/QgijelMAyHScUSQRwwDDywa+RX7XAMZ8niJX84MAs
    3zpioC8QwDaJeVp+E3XALwF6cdllksAwWdZgaD2YwDVTVoqOavTAKxiRj1rAb8AyflgJ7U/+
    wDTA/Y7lbQnAMMEIDmibUMAnQWO+MruUwDUiTxWqDrPAKyOmPLvrZMAxjYYEywUCwCln46Xv
    vr/ALwaMAX7lwcAuL0bhOH7owC4boiINOxTAMBHv5kOjhcArFbugOUM9wDDaweUkuR7ALVTZ
    y1X708AmRecXe0j3wDF2RAYnlCPAM5DGe/EVIMA0oUiqN9SWwDbeak0ocnnAN21HBsZUgcA0
    0tOdJyNKwDWztdIeORrAMtv/SopAWsA0uJUcVUlwwDU8c21MwFTAL3h4LTIZ98AlHG1FDzI7
    wC9CbhQPV+bALzOBjqp7xcAyRmfuUwm+wC335q2INPXAMZ6xaBUVIsAxDH9yf68WwCxmWEj/
    BkTAMrORN5va7MAoo2JsJx8JwDEmtqAX8VDAMIXJlSi7rMAxsbQf5785wCzB4Fe7o9LAMlnv
    qT75xsAtFEe2Vn6ewDGYnHQH1MzAKAxcfPtWbcA0tfcNsUWkwC7YU17R9oLALqijJSkt4cAv
    FKBf6pEuwDAX4jkbrkXAON8JaoHUwcArYtjDpYOQwDUnB1ogZ97AM5KflcqfNMAyqqla9J6V
    wC1M56pPcILALVHECQHTtcAxsayT4tTcwDX+J2xKsRzALoH+9UELX8AzY6ZI/HxywDWAwGgg
    61fANGST/2xfxcAxFU0YfDSqwDHe3NqU61TALfwPIPyy/8AwJu14mrRPwCi/N+xmV8fALRLW
    TZgXrMAwdFs4zMGswDPIc8O6WjDALtTKtlTJuMAuSwl5v1RywDBVUQgdSRTALYaNWHriOsAz
    WPRPMeRWwC7qk1YQMBPAMhEdhwGMUMAtTEI+oCovwC+zwZZork7AM0A3RnaTCcAyM9GPpSOx
    wCtEM4DAzTLAMbWvdENACcAyEU9FAe/ZwDIV/k4HNVXAKkyKRL/vMsAxS6TgyZspwClXndVw
    TuPAM5HxAB4Y0sAyBt3AYwgVwDbStapYQCTAMh48ch0h4sAwIO4NZ/crwDH3kQkDUt7AMNNf
    VHZle8AyOwJyrxK5wDOzfDyihq/AMX97vPg3l8AuGE2MxflqwCiFcAbp4R3AMbX1dkzuusAv
    IBGTZRXlwCvsHypKOWXAMRMHWZYPVcAwPRnRaeKwwDUEJsCvtxjAKqZnB9HW8cAuD20qM1pO
    wDR4Nbgb6ovAKwI44zdTzMAx8zEDLle9wDJ1SHoWSnbALoIXJzNK98Ayf1XUtVGAwC1YREny
    taHALkjaOKoOwsAx1u0NwlTLwC3kzNAiDJDAKEpEbwAd9sAs8VxLBV1wwDGfeJ5NyEzAKm7Y
    GclBacAx2f1FftohwCirPSLhV7XAMuIB8xmspcAvrjZMm5ISwDd6F5nZtazAMxn9IQIB+MAt
    c8kbKBeTwC7mfHcR7KLAKU0SaaL+48AwH1/Qqc1gwDIBJtn11DHAMe2pfVaPXMAs10oZ8qmD
    wDLrdo657pDANIR9UL6/lsAvYVis3J0BwDHvM/IMuazANCiiPDdrWcAwWe7ePb/swDG5d0CM
    T/jANGxSzXVdOcA01h12L34DwDD/iG8CJ2nAM2ryHnlgs8Aqa9s44bcOwDPPq69H2tLAN73M
    RPg4/MArb2ysMzr9wCztu+/O5ITAMJClpnKe8cA1loHxJer/wCk2SCSUfW/AKxnVRrIFpcAu
    1+Awnx9qwC/CLgbDZTbAKOuBQXFZpMAzlsgJ+q7TwDAIdRFnlj/AMIH1e2XGzMAt3eMa6LeH
    wDI+tDmc8DDAMGMRqje8ocAww6S2RSxOwC3cQNvOd4jAK9boiXbXE8AzpC4IGzbXwDUxvKpp
    +wvAKzSMOLMsPMAmirwRIdBywC+pO0PSNgvALfYMem1YqsAzDLpef46lwDBXdq5zyKzAMENO
    ClXchMAqeV+l1FKgwC0HTJ690X7ALy7Nf4ygrsAtl84ZaRy2wCY6ws5f/jbALebJgM601MAx
    97hxx9f+wDF2+WmbKj3AMQf/eyu498AnImNwx3q+wDC6+wvtxNbAKsrEm3ogrsApt8Vu4F0+
    wChEjYAc2ALAMS9OKQiGN8Aw2i4izYARwDNYgwDRGI3AKzKAc+jHzMAs3lNP3OPFwDSNLG9E
    e0HAMUXtBQClY8AwkgHUDqRSwDJkTC1lScHAJbl8Bn7QocApFt40TIk0wCwefIQGLErALQU9
    8jQTq8ArAeXFibl8wDGURwBgz0fALgwrNWJ6LMAv877iu85gwDOYKTWjAA3AMuUTs/ilHMAv
    4AcdqUXawDFjSxgKi7DAJXYwgpvSOMAxtXkVVXOjwDaa5AMDz/LAM2AerRP2WcAy5HUkXUay
    wC5l+NwzS3vAMIZJt9aeR8AqzTMp6GGkwCnbE9H+RB3AMseSV0SH3MA5Y2jcriRdwCwqq5eg
    r3rAL5J8JxgItMA0lFh0PI9MwDVTR+IdjnnALnJ58vKsgsA02w/3VVS7wC5qmcm+uPrALFoB
    v2+4e8Ax731wN68HwC5RY/6PG3TAL7xsvMhj+MArWWOJ8nZJwDPF80oeCdzANC8tUpgFjcAx
    O99KHUoJwCxcgNLj7irANTm8/HLCF8AwCyHsCVVDwDC4tMTmC8jANAAwCgkT2MAyGCpeMTTY
    wDFogljOAGnAL7nTvuW2mcAyjr7JRdndwC19Uw9koo7ALjIhzkTCIcAqSehwa7C+wDFr2COR
    UfTALBWa8DiT5MA0LxL3aPGKwC3tm1UNL2XALWq0Ri+B8MAwIGgR0nHzwDN40Nrgi5fALahB
    nIDJ78ApyzMn3D+awCx60qe1xfLAMQ+tGQePo8AzER5QfNFxwDCh+tVpLxTAMnmmmxU7E8Az
    bjm/ocBfwCrgMAez0XvALm16xCMF7cAlx+quNccXwCKywi0BkBHANrudcOKOqcAu0zvIypGz
    wDAkNB/Jj+LAIoWEQNZEScA28V5hnR54wCsBy30rMV3AM0AvI4+wgsAxqWq01V+qwDTIy+nV
    QrLALedpUTV+88AyKkn6VR22wDNvpskr+g3AKc8FzfMg08AqEE+4Mx3SwDM4vJ3qFXfAMMIh
    43qdaMAvXZ52KqqawC+jr/utw9/AMPJRjYiH1cAwBgznx5vKwCvcDC/V3ufAMycRX/bK/cA0
    fgcFdVgpwC1h5j2/nRvAKnUcEUaTlMAm8iof/qKswDOMR/fa2zjAK0fOXjKbWMA1/IKqEse1
    wC0Wnrwur9jAK+hDM/g6UcAtrx7lfzmIwCnFsIiBy4fAJ3/K9G2b5sAvyefiHd+RwDCGe1Oi
    oqPANCa78G0pJMAuLDQHMng6wDDByWSdPsXAL7pkbJLZGMAzfVU/qQFWwC4Eptt2vd/AMwoH
    mFAHWcA0MnQOIGn7wDGZzyiLMg/AM/mlBuFxr8Azvu9znCeewDJbAPoAWmLAMNtxyYzSJMAz
    yVokXNUrwCxUOr92QzTAMJGwwEzaTMAwp3GbwZFkwDHvB92jKjTAMR7WuY10zsAvpSZJ3uhq
    wCfDjSi4HzzAM7ZY7PBt/cAv8Xq43u7JwDV4wPeVA6PAL6n/xwZLhMA0OihAaspHwCqQ5BA0
    vCzAMuQCZ9mPKcAr9EVclsnJwCwQRw0w38PAMSBVEfm8wcA0SpXEy/NwwDGYC4RHFyjAMCZU
    afXvWsAxrz7V+HjSwDM+VubQ4n/ALJsZ2oHPQsAwEHe7H5TnwDHxjLLS8C3AJtCNW+/HKsAx
    PHEouJJCwCjBcZPbbprALde+Nu21/8AywQ5Yk9zwwDKto4xH27DAMOcXJsFBZ8AsJtk8d19I
    wDRCEA38+rnALnrN8L0GLMAwQ8NtZnmawCyK1dnmJFDAMVOZgQ66oMAy5TPMxhUywCvFEAeo
    iwzAM6ufIY9uGsAxA5fLtaFYwDRnD5kORf/AKstcNo32yMAxhOolzbjRwDCUYS8Jo/LAKE9m
    BS7xwMAvvgaCOeUnwClmM9l/aFjAMXy5UgqXdcAvzE4/Q8nwwCzVGohHiwTAOUxsJxPsQMAx
    mwKJhmsewDSEw536GfHAMwuZviCGwsAre095qWM8wDKhs/IkqVnAMEPED/8zmcAuKhsRLuAQ
    wDDfCnDR1R/AMR4eDKyn8cAqtjZamgCAwDJn9AzSRoTAKP9pzJDNUcAybZBe7p4bwC/5OpUi
    nV/AND40VssakcAsU7nWTcwMwDUoL+gaHHDAN/Hj+bOl2cAvDxp8J6tkwDLg9s/mYw3ALHTW
    c1dgLMA2bHK0azylwDMnto1VRLrAMj9tuo8mysA1kXHCscPBwDOXXdrxG93ALaMWizfAHcAx
    GaBuMM3vwDTls7566YPALlBQMyYF0cA29L+AX/rfwDAAQClvjzTANA3fAyWJjsAxtpCSI+WU
    wCn24490GcLAMf4A0u7j8sAxV4l0+af3wDNbEZxDkGnAMdT3fQep7sAvFTtOW6PGwCaXLOci
    o6zAKeBo80Zh5cAu01Qh3KMLwCukmijzmqDALzLyESxKI8AzHqras2AhwDAcBglSPR3AMYRi
    dYVjJcAy8UpETMQYwDF0WKYLt7rAMVJLY94jbMAl+8ulKP5/wDDjjhr0yjbAMVEqJ7fMQ8Ar
    W+mddatXwDCzDNmSGMTANeogV4jqL8AwyTtgyU/dwCJtn2kOWnzAKNBjdaX4CsAv3yh2YfgH
    wDG1/ozM0NnALJv+XdAGfsAwTMzLqFW9wCpilFL3zZPALyq9xek2vsA1a4yRvX3IwDE01bYE
    hCrAM9F2F71Ml8Aqsz0qu0ZzwDHNr4FmUn7AMR62IjEuusApdvGNc18uwCuZW7IQF6rAL2ee
    CjbZtMAzCD5fqEc4wC7jEcuIiRLAMNWuANj4EcAzRme/Z7GRwDBFFbUvXUrAMQJO1KOvW8Az
    bMT+HCUWwC907OWiL/LALIQtzY+h1MA1nQbZXxLzwDCGCfkaXKzAMsy448k51MArk5WXM2uG
    wDGCyHfqU+nAMttQdZi/m8AhqM/5ayXewCi7Ug45WYTAMSXV4mgS1sAwKnPfSwUCwDO4oS5J
    CvrAMO109O6dW8AwVVQw4TFSwDWpBim4EtbAKGjFPbgwRMAqnqlKRWOMwC5aLnyGX1HAMkU/
    B71IWMAvnpfFmFj2wDC1IykABffALacnJQxx0cAzC1x8UfQfwC+0onkM9Y/ALuhDDxD0a8Az
    e3PV0WG3wDPgVSeM0/PAM4SwU/99X8Aq93dTyTjNwC20djlJIXPAMEV9AO/LbMAxdKrEA6FO
    wDAJxchCSJfALQ2uuJxPSMAxveqZat5twC8DWh+1BJLAMSV82fDgbMArPTkqjjNPwC9Kf9Kv
    UNfALnJeJwK/McAz9r3BvXJYwC3K1bHiMODAM7hXpyfm08ApzcxL795nwDK6TsB5k07AMnN9
    yp3pycAzXdLmEMGIwC5lpIskMRfAMRyjd/I72sAvvAD+zQ1YwDbLMUKKmc/ANMZth6Wrb8Az
    VwqPU6HEwCnuXDJGd8XANcnGSg5hBsAtLkdwmEdLwDFv7x/UEy7ANFujqrekWsAzARcK1Bdc
    wDX0A9ZAyzjAMKyeJClw4cA1JAbnfp2ZwDE7XWnIrfLAMXjfpLSp7MAwb1vRTj9PwDVtLBgI
    /krAMBv5eVb0/MApk07dQp0HwDEt9nqv5NfAMJt/2fW7ycAwdBjUQeZAwDQgxSGyBoXALwPV
    idl1YcAvvXilP1W4wDKrOPz2vm/AMGBGVzlddsA0febJOhjMwDhMtGSkaFjAMFVMnXPiuMAz
    vsiVJjcswCzwhRwV5ofALEkYNMyKG8AxFstHzSgAwDU/NrwR7BvALZ6vFBpFt8Au79vQwtqn
    wDUceYzb1ALAMBRv8tWnesAwA4Te0GnHwDM1mCvgVLrAMYAWz3ruv8Ayubs/NGc0wCoIorEh
    LS7ALzw3SZNAjMAwExP76GATwDY3BNQAOH3AMdakZgmcTsAw09wM6MD0wCjcraqvwA/AMQP4
    RmhBE8Ax66pnRuBQwDFYBl+GnH7AKz7n76QqtcAp3mHMUvibwDOtMl8oMibAN2fVzOxI5cA2
    SNQLy+RgwDIfFK0ZmFXALF8BQEUtZsAvVfdamGcQwDFh+JUe7VjALu6wX6hSrsAw1fKkbdjD
    wC78EwRPl1nAMadE+/WimMA0jPCB6Bz3wC5qnpzPbbTALREWUACTZMA1XHMMaytIwDAn62Ai
    YtDAM0s2hURlpsAt4jpe5INcwC6/jvSJDGLAMJdmlzeURsAt8kto4INcwCjBWy5Jm2DAMYLK
    FI4iKMAs0rbwbEI8wC0+w7YTfYzANp9cD3ZYK8Akldm2UngQwDCp0f3lVtDALVz/yNrQhMAy
    k7/QRSuGwDM83D3KrZvAM0xxOyX0c8AyvLbDij8iwDMfHVHq+3LALS0tYiqvacAyM+lx+nqW
    wC/8K+XCANjAMpJbbzxaM8AxS5Tg/vCTwCZxj2UU+iDAM8cCD4oWzsA0Vl2rodc0wDHkPRSP
    iJnANPSofcKhzMAyUw2mJjgQwDH5UZZ0G9rALZgPHZvlHsAvJLYEuVQawC1jtD0kJYXAM4So
    nYEu9cA02GBHLv3JwDCDAXH34qHALVBYt8C5i8Azdth+IH+ywDIigTv0Pi3AMRQbMejc/sAy
    RAdUeEQcwDaZsLTygkjAJNzsB6pkwcAxkcIeXSYFwDYkNitiuo3AL9rHi/YmfsAu/uKImV38
    wC4UfI7bAiXALFspsYNu/sAwrs5o885GwC+PvEBcqxfALoPRnBUDn8Az0XrbD5j5wCuIz/eW
    tc7AK5007nKrcsA07tSzqwH3wC/SrZqyiEzALPPTrO9fqMA0teEv+iMWwDPH6raRyxbALbdY
    fU8j6cAys2/vRLf9wDJoCT3L0OTAK7iryyFSyMAl+YpmPxKtwC+tNwXY6GzAMt7foAzo68An
    VEcw+b58wC9cIjqdqKnAKYXi8MxQeMAyaEGHHTZvwDGAD+XihwLANDX9ivwdfcAsELy5qmEr
    wDDhKwiN4YDAKxra5yMCUsAutchrZpwAwDJlO5wd7jbALkOP7AUPzsAyfavob07CwDIkDgXn
    W6/AL3bRyjO9kMAvibev6iPdwC7XQu284DjAMlDhcXZKQsA2IJVGqcbUwC/+rOP7njbALit9
    njQQsMArpTl397t+wCtztLclzkvAMKrVK82lKsAv8e9n58rkwDSPDLJi/W/AMZ1a8bhozMAw
    kI2i65MJwCxskW/wRJrAIS9jmMjsNMAucq+GkxPcwCyBjCTjh8LAM6fIIa+aZMAwtpU1fZmZ
    wC8f39aVy8fAKrycdxLrTcAup3fAzZILwC5xJliPif7AMFSkCZ0zH8AvEZl9GM0AwDJPK6gc
    yzPALQauLinWOsAw2a28cHGhwDItoN2VDNXAMR8P6ruepsAuIFXhv5eiwC+bIcPHkqrAMnza
    eW2ajMAswwcEALrqwDGCyWPmLonAM967c2OWxsAnUDVmx5ERwDTnRsoTE0nALbq6cvh8asAu
    Dni2OmI5wDIRsSg30hbAMLXLXOvtrMAwf24Q4cziwC2yS2xJchrAMG+IcS8QyMA2UVMTIHjY
    wDKNrggSVMXALBqR1X1bOsAx97rNQP0gwDNt53e4irDAMfnLmvda3MAo7/Pucu1kwDKIJUrE
    Q1vAK5++b1ka58An1XI4u/fGwDPjXLRwMS7AMP5FRijfQ8AwJjPVsEGdwC/ArTs9rprAK9hr
    3CYDlsApWK7Sf+qtwDHbIErh9rLAOFHYcRsXRcAsbokf3hu7wDB5n7IInyvAMH+o5M7mNsAv
    jBOIdnzJwDO8D6OxeFbAMYihbuV778A1lLgWv/rSwDI5xPKH3DDAM0m54nmo4cAvFxWFfEN4
    wDB5xTlH9iTAJsYVUfPWRsAyDBnolFxYwC607o93JVXANy0za3IDesA2JPiN5oMYwDJ0M7ng
    z5XAMdKxYTO3RMA3P22cJGW/wDMj1ix4AzXAKnpFVw7xUsAxhcu6Y/PmwClnBU8KPMjALiPn
    tn/iYsAw9SaTdQP2wDALIl67iY3AMJ8IfH5bTsAvbGXbDXfpwDWmy/W6XYLAMOOCOfq1QsAx
    bFXVht1FwC/dXPGbL+rANEBYfsks3sA1YRd/HF+CwDGCNpoQ6RTAMgU9qc50OMAtPnEWb872
    wDIUnPGxcVbAMTNrheYiRMAxy6UDr4VBwDFZRY/PZFPAMBcGOyV4WMAu+m7EAvIVwDCt7xZ2
    5QnAMDzPSV8328Ay2m6EUur7wCp7aH0GHcLALlgZ1MehysAs6JpICMxZwDJw/m5vuN/AKoGg
    xTm/W8A0SlVMHHUewC4LQ8pPD1TAL3gbE1p3EMAySOBEN5hawDTzobKXIy7ALqVGEjpg/cAw
    I/8yii1BwDDDfAZXEMzANQQKhm4K3cA0/wJD2zt4wCvzuoeCDxbANN/h3AoCw8A0AIdu4zAt
    wCmyvkktOr3ANQ1tdpxFO8ApaxZoXlgLwC5wPYYcgYDAMJ6NOfxnWMA5w5rS4f1hwCsZApyu
    ru3ALudlUtqHwMAwCrd76lk7wDZ/2JcWPd7AMeQtZBhVKcAxQHJnUI8XwCXJdXxmccLALoMc
    z8N6ssAqajP2GtUcwC7jLdfM1s7AM2U6zhKfYsAqAAcJnuZEwDU4WceKYMvALUihcn3BEsAz
    hQm/CqWYwDZw+YrmHerAMNn6hGxTlMAsWWPWfKSIwC7GQJUmmhjAN/i4YBmzJMAnmpvBmzXq
    wDZo0ZFnxljALxqE3Ptlt8AvhzEgTLJ+wDWWRCvJI3DAKXRzotRVRMAobuz0UWBRwCo7kOHd
    IUrAKNNmjZb2vsAxfqAiRv0BwDI3jabFXKDAMEy4zw2tVMAwErE0Xb29wDLNrf+/naDAMFme
    3Vq1osAs/Qbv22G+wDNVqh1CAUbAMKjpIQ7SXsAjvKf/M/5qwCTQmFRmQW/AKuAJmgXo88Av
    5ziHIpZ8wC2TI6qtnm/AMYt91ZPGJsAxjKEeJxlkwC3MMFX4pJTAMCwSpcrerMAtd2HuaAJQ
    wCjAVvHN6R7AJKV5IQMShsArETW5QyQ+wDF6qb+EF0PALtgBU1FVacAtBWe4BADuwC9IMkqY
    jj7AMao3RJLtgMAszLbJjojywDGLP/6x7+vAM2nw/S8GE8AyJ51gwpNLwDOY75CLmljANQnM
    oYnMh8AzXREEhodPwC+4oFsEwlrANUMWBKcbzMAqPkwGU7FowDOeqbuJkK/AMJV6/DRAyMAr
    IGYQZLT7wC4aPBIz0JnAMUmZhQUmBMAva9hRjWeUwC/0sLU7uzzAMet1HlPLD8AwgmWc34vZ
    wDM4Od8Bh3fAMZj9gvVoPMAx4CdxOD/7wCYjuH+ZxXPALQ1LFafRzsA4Xx2N0yarwDEJ/SRi
    HRbAKZXkR1PYBcA1topIU9E2wCUjF4td2uLANNhtnX56CsA4npkqH58VwDOg7STeDOXALgaO
    70P2mMAw7QnCRntewDdCwFjRjunALHS6KABMPMAzAZUpBdjXwClAS9EBVOrALfNVlhOFT8Ax
    5f6x5MPCwDWaSMcyr4LALCNZVbU+usAulcLpCqqNwC2ijdqe+U3AM4E03L0SEsAyngZmGRV4
    wDYrfDOf1nzAMKMv1fNXhMAxxuzOtzDXwCyONhLMMT3ALAG7G4xrOsAs461VbbJxwC+6b2P0
    nlnAMjLEsM5dTMAy88flOKFOwDADQA1YI5TANFH4mNtu5MAxQj9FfnsBwDGkwnMCIqDAMAhy
    mByyIcAy+5XzWCCiwDAkV1gnJwzAMoLVWJrVi8Aw5686nvM3wDGJrg0ZjgXAMLQAKMHp7sAv
    36Z0QpfUwDHoZR7SaX7AMeZDEuKD3cAymz+Kl/t0wDDcMnlceWXAMWU443sYwcAw7wUMvU21
    wC5D9uhQ36rAMhSLuqN7EMAtwx7tklzOwDNVbfJj+iLAKySJBtxiJMAtXsThlx0GwDKnLutY
    rNTAMy6iN6hBB8AsCBRoeNjRwDIxaRuFElvAKM2BKA1DksAzFtV8F7NewC3mfUm0OTXAJdi5
    7qZwWMAqSuIPEJRYwDLQ5yGOBFPAMZb0N1nMHMAvJwxKMxQdwDASFOB2n+jANdDdq3oDb8Aw
    emYfayS6wC2mWBarOU7AMkXIDBjvP8A2qijwXwrawDCLtrohnJTALmnsJu2p2cAy8acQLeGm
    wC/G0HQ1K/HAMQY7hsuChsA1L/uFxfuUwCluTl3FJ9fANF8oo/jGksAzF9Ruuey3wCb1oaqc
    TdrAMTL92AlKy8A3bx5ahOXiwDHw4ZFDhx3AMlZO7vaZOMAyaPenis/RwC5EN/BmUjTAKBxD
    FCgRxMAx9gWOM6ybwDEZEE2jUebAMgJBr9/TFMA1RoAT0M3OwCyTX20naTbAMiOMYECVwcAy
    clMBd0mywDdS+AYEMqrAMyYIMLihGMAr3kU67XuewC4zDbd2rInALM7zp2Hn1cAwFNEOyJfG
    wCpt/CCmXG3AJqam1TEi9MAuBSyfy/LjwDKxcX2H6T3AKtelb0SfHMAxJwXpV7vKwC9bQFWi
    9MPAMIdSaPizs8Aws6GEBVdzwDWP9QVczVbAKD64mvpMnsAwKgn2Gx4ewDKNDsihb/DALYuE
    heB/v8AwEKAjrdUswC8VWGRAmErAMU+Swc6y8cA1ggACEKv3wC68PVvnj3bAM07dRulPpMAu
    9i59lYBFwDJJ8D1HcZfAMKMCl5LlEcAp4lREoeGFwDM0787ccAjAMneVZgqG9sAyJV8tqlLY
    wC9gZdW5kLzAM+SO3Tw9RsAwK8r+o+5JwC8QmqJEABnALg97TqmuRsAybhtc38K9wDA33Y50
    r9vAMdYuQf+2YMAwC8uPQGOTwDLF8rUv/WnAKAAi9WHqKMAyhAInIDa2wDQnKMGwSvPAMX+q
    hv90nMAwpOsMAl93wDEnLWt5DHLAMN0uL57KGMAxvvUVlWJBwCx3XzzrbQjAMHDWbHp1+MA8
    zk6PB2D3wDX+dALaV+nAMgDrXMPB/sAr+bCz/4DMwC8tFhamo8HAKinGPZKPu8AwQgJqd4k4
    wCwFYwVAuJTAM0JwxsBPfMArP/MYDm2CwDJ4y4Joy2zALcXzR+THl8Axz8QzGPZswC2ozfzi
    l83ALs35a4+wXsAxvqILmBBTwC4FsyGRfajAMsJTxspB2sAv25AXr66xwDJvySOeNRTANQFs
    0lZWd8AzI0x50exxwDPP6dYrT7HALnmfm4oRHcAzGisDs//GwDEtEH/WW4TAM2TXYFyLGcAr
    6yU2CaeYwDO6uWvJgTjAMDLc8qpHUsAxy9TWTHFuwC5TSSTD+2bALwghSqfzjsAu+XEINPgg
    wDLF70/8o9jAME025UDrYsAoPCoq6vp3wDMFgPPBmlfANeQxyiEBqMAu5/lvdtpbwDEIHP/9
    IAXAMXbQXyZVt8AxrRpW+4mtwCwWP2PK5e/ALVEIx6uJesAp6Akidj0KwDBFJaDPGgzAMg+F
    i98jNsAr8qp1PML5wCjWNXrcq07ALWn2zUR4iMAtSBNt5YNFwC67tC9Tc9PAMt3bG2o1b8Ar
    HLZbErTzwCy9vUk+ZEnAMBoC2uVzAMAwR3cus2ZiwDVOb1X5JfHALebhktPFScAwCa1Sz3Cg
    wCi1YbhtEsbAJmlTjE/5MMAzIfbmxWO1wDIX6N/SV2nAM6H7upzaPMAwmN8DU/EiwDF43kHt
    +KPAK+Ki5UceX8Atb/EoTmx5wDHDeKpefyvAMHSV92ZFKMAx1PMQu+6uwDFgVdwy1mzAMmW7
    bZ3GRMAnXi02put7wDAaTRQ7l2fAN8RfoVhNosAwfaejWd0hwChn4snZ57LAMGD9zHYwzcAt
    +1T23Hs6wDDskZ8hsTvAMacebX04AcAsjK8LlhDrwDEP5xHgUJHAMlxj2I/bFsAy2ggDfa0X
    wC/V4sAXPlbALXgjxgXk+sAwE01jy2yFwCe9IVlwjfLALa3x+id/3sAtBqRW+ysfwC8lmxfB
    19fAKxQynNbOA8Awgv0JGs2KwDAzJmgzGMHANW7/9iwXg8A0KuGNFnKBwC1ViopITyfAMsFv
    kvgZQcA0LgJTcvKxwCfowlVUeI3ANOHgQbf6YMAxKO+0u0xPwC/J8YCmDtbANFFMBxTCysAr
    bjTJivBiwDDtTxkvpl7AJqJmEBaKaMAxiNRpTWyGwDC+k5oTLjnAKWGzETozJMAtTUOLJ/u0
    wC9AALSTKhHAMv85nv6OJsAv7iEHMPDXwDMUZMG0uzzALfkMc6sF7sAxJy8vpgmlwCp98vLN
    53DAMzrFx0eMv8Axfp41wm1uwDDUCOKMkvzAMx2UP6VYs8As+AL601EEwCm7Xgz5bdzAMaKj
    i0KWEcAzR9UflMOdwDSbKMGSQBjAK96XoLGJMcAzRKGcBd7ewC/jweWjqHzAMepwZJ2a/MA0
    ViB7QEDZwDCnbmaxRsDAMK1rxy1rTMAzsQf2FVRCwDC5TNmWj+/AMAEhEbng0sArP7F1Yh05
    wDAv3c08kk/ALvsMiIS99sAuB8cuo9o8wDSVoQAOgd7AM0F6xDuMvsAt4Lboau9NwCxoszL2
    BujAMe8iXzk058AxxmPyh4A8wCYqH1wUjj7ALHAaS2KN4MAsFEUiyhD7wDQTu4qDVivAMpgT
    BCKkpcAukCdx7LPYwC4J+WzYBQ/ANc/Y9c8ZgMAu0J1YDIaswDXIOvl5q9vANHyN4KAR8cAy
    hGFr4zcYwDGryEr3i+bANjnSM8RG18Ay2wQQ68AwwDBJNoORDL/AMWS9B41VDsAoEdnY9kBA
    wDBIL7S01KHAMl/drQ71JcA08Y7DFVrbwDLcFjW4KMrAKhzxHT6WfMAv80kwKakkwDHgrIuq
    vCfAKMbb234mV8AushFpYJlqwDFjvVC884vAMoLAdJEE28AouhLBnLBUwDDYWqt4BAnAMYEg
    Y4npzsAuh5xvCTmUwCtC/KTtal3AMGlUL3G+z8AwLfXu4v9SwDZef6EPV6bANROshTRqBcAw
    5yD82RwkwC5fenVruWfAKag23ocvPMAuAHJ05e4/wDFvUS0NBgXANZr+9HT9o8AsVrt39zS2
    wCuf0qY7IJ3AMeKgIjlvWsAptbgeACoKwCmzU1PsZSTALD3qwdXNesAxmfA/f1FkwC5u++n5
    wsvAMIaGLirkoMAp25v49cV+wCtuiTEiZs7AMNZH9tixtsAxOIbukO5gwCo8mYtDjcjAKoC8
    DSOJbMAwPyFnIwFKwDmTUG9Rv7LAL/KmIMeaZMAvsln1cG1uwDDQLm02CC/AMxWX0g1WmMAl
    4pmsf3rJwCk7KWYHwxbALRx4CNb2vMA0gbPpWPYfwDAm1MmBsdrAJiyHLSITkMA0v6pT4s4R
    wC8bDPUG48PAMRATpdmE2sAm/c8fr8O5wDJZ1wLnI2PANeomj6d258AwfKbt/oE9wDT2KEop
    2MrAMFNCvqn5MsAx4ynAFTjAwDCEKstK/wLAMg+aJMWQJcAvcHW/31uVwC4hhxKKkAzAMuBR
    0QMfGsArWkntBf5vwDGKBtOYlevAMqHooi25mcA0pVAFa9rDwDK1nsgV1KbALdbRvfOwcMA2
    eIDz5CAgwCtFncm6UbbAK4oFF84u3sAwmARe4oUGwC5jpPFXpqLANYuLigCdt8AxGEAc1+dw
    wDAx3mKDAXrALinfAjCrcMA3eCnhq7HvwDYnFo21BMjAMP/2vCo5hcAzL3ZsSjQzwCrvJraM
    ohPAMI0LzAgoM8AvjWNP7OZpwC9EXGrXxHDAM9laZNpVtMAzJaYEgX2mwDA0maf6S43AKue5
    eUOtzcAvM1IQIyNqwDGqmJZrxWbAKtNjlw5xaMAybBgFuUI5wDAa5NiL6b7AN0aaCmWC9cA3
    YaIxy11ewCul/hYlQTPANju7EekhTcAxX2zFwC4AwDJnbXAUUOHALqQnm3J4nMAywJqdntuJ
    wDHOIpYFJHzALVnQdUfB78AygHyIvHFRwDMzJxdNewjAM2Ht+AuRDMA0nxF8OlOrwC3Dxk6k
    uY3AL2rEgd0DTMAuT4n+qIyAwDeOteKD7HrAKV68XqrG/MAscBRfJkNawDP335a8Oo7ALVxo
    WNN24sAlWqtd/164wDBF8tdl8ibAMEdTIUoy2sAzeVWmqIAtwCcE9S2w0yDANXrADQoYvcAl
    RDe10uJTwDF/dOY9Ue/AL/McFklb5cAxpL1BF0U2wDMnG4j9PLbAMqheC/eJi8AyXQ1cI0zB
    wCS4vcmAqk/AMhkM5um9ScAuY3YQW8LswC9asWw4rNjAMf1DvNEXs8AqDvJKQkMnwC4Ke97L
    1+TAM/xUJ8rBxMAwwN8aJXjnwDUZ0jJcqpjAMYkpc7rbLsAx5V3OWMSPwDIU/akry1jAM0kU
    YKkv5sAwLMmtcfmzwDNN54n/pvPAMUxR94XlB8Aq9MGC0clKwDLDvlYkXx/ALCrj85hnUsA1
    Fwg3G4UKwC52e4CML7bALwXu1VVwEMAv8XIrlGGAwDCm703eeeLAMo0+/oxqdsAxex+dGMc+
    wDDJgSg6uPnAL3LCc4hAysArUPzS5loqwC7j2nJxr1PALbwUdJldXMApvGGaKZCBwDV894Iy
    DBjAKKaM2X+NcMAzNjwzQ4ezwCWreGJzcNnAKWCjreSF28Au68T4JJP2wC8qg8jBnUzAI1xn
    cYiY88AhBGVg6FZIwC+H71P6ijDAKOh8gqsap8AlLgYdLrjwwB3dw/DtOQ7AKtmnQ54L9MAv
    fpNrO52bwCatbUlDNbrAJF5ggKN218AsmYMb9+6hwCdcgE6c4F/AMPzc00d/68AYC6vbwQVq
    wCaFfFI879vAHy0dUDK83cAnvz/70TLswC+YNeeRMbDAIPpQ5VUdIMAn66LxbmuCwCsguhkL
    BFzAJvq7j3r/FMAnUhforzURwCalyxREqSrAIf9ldEdD5sAh8rclMkVWwCX55zhUOCPAMWtX
    uNfSWMAtvr1PNUQuwCZtQoVD0ULAIrl7whpsgsAhYCQHs88uwCpW1ZoMZ0jAJTE2IRQfWcAo
    ezg4BKodwCO0AZ0SFczAIfDAcX9x8sApKHvBinPUwCnmgFS5IobAKbpnisdI6cAhjPPGuWKm
    wCS8u4BQ7drAKMqm/eEbN8AnUP8Ms3UmwCkIkLeEcsjAKCkB6Rm/F8Ar+Uunn0eAwCZgQEJM
    CTDAKWoA5trLCsAlltgAKlWBwCN0bbz4MN3AICzW+28+vMAl1VnRz4DqwChuu2UPDhPAJcg1
    IIOuvMAi5Xj0Jk/IwCX25hEDIRPAIOEr/BxE68AjjgHgjss6wCIl0OQ0w/bAIlGPqtFDe8Aj
    O7hQBNt5wCLC/sP0irjAIXQ2B4lS+cAsU92NyUdiwC1nXgWAxzDAKldJH/OmisAvk1z38b4a
    wB652nl6OMDAMxqfmEdg4MAovlVRyJgpwCacpeGXcwPAJecer7XxncAiTNp0Ua34wCl5GaQQ
    8zjAIz2jrAiPTcAm7NZyBSIOwCX4CkYM26TAJBTjnw5nuMAqWRuJqOqWwCmXhZeKoKvAF4f3
    z8xBFMAlwuZzsQEewCiLKF8v3wDAJFjweeVIOMAkgZ9+pnjQwCVe6Yczy/jAJprILRH6s8An
    rGyQnWvHwCH4kR1wd7LALihXXxZ6DsAoN0lnEvrtwC6icaIuvnvAImBqiBGEIcAi9s14PIts
    wCOKo73m7DnAGdmUxRtlpsAtYlLZklHowCJ9mnNQomXAJuGSohKCj8AgcmBQX1NHwDGoBO8M
    KDnALFzjgZ7X38An9C0C7tW3wCj6WmtFAP/ALOAtvYvTV8Aq5+LBfaZywDDBk25KlIbAKjgG
    rl9/2cAmz7g15YnXwCi6r5zEA23AK3eHwyniQMAssIc3SouTwCWKizlNQpLAMZf1x8zfzsAl
    1nG3CFLIwCir8z7YMYrAKwK0MbNs8MAwwDl3bK3LwCf2tU5Jcw/AJ4HQDHxvysAspCfTP44+
    wCsbWzgMSvbALRW1EGa1/8Ark2wCmmpOwCgXjdtX+5nALhQutxIGeMAnAFp8mDsYwDIaQA4L
    ZnfAL9xgL1BWzcAjQZ/WTuAhwCWeZqp9jrXAJ3PXk2YOysAoOC7OKYNqwCPwp+a3Id3AJp+i
    OW3kh8AlJZYSSs+SwCbYh/PNVRrAI3oCvC8XjMAvbOQXAyuEwCX0e3SahnTAJQV1/5oQysAm
    l3c4tNQMwCrMeApoV6fAJOF1hjE1zMAxnbf0opD6wDHASVl+xFnALBTcoPf/S8ApVGn5J1UR
    wC02PQoxUmvAJR5CwG635sAqIeLFbCcmwCIt9mfTMq7AKQVcnelr6sAnxNbgssnmwCZF0SEY
    YN7AIbEgq9IilsAfdfxXHX+SwC0Lng6UzsPAKgscfKsP/sApcD4zxJ4wwCNWDb+eBCLAJe6p
    0DBufMAutucusDYswCbl7tpfDHTALCVmGGXeQMAnyGCe6yuxwC4vt8XGh3DAIt/hOkwPqsAj
    3pVrCLNgwCPYSkr6CwLAI1n4J9kITsAh424rJRQawCXaoL4xJJbAJR1mwO6mU8A181RYpxNl
    wCTPFBiuiwTAIcQQw+KFrsAxxPXHyScdwCeyZvaUtrTAJyjQfUQQUsAfj8+dH7RAwCstNtN8
    37LANGZfzj03tcAp/Xc7a790wCDoYxS5QjXAJjkHsG7BpMAmlooy6iIgwCuwjaKN7G3AKhGU
    ktVbNMArM1acTFUbwCIuWWjdmh7AIoKFw9NeAcAp4L8saiY8wCYgm8bziMjALfhB88Fk38Ai
    /xWBbvarwDGdo4VCsSHAJB5OcuLkL8AnB8gMh/wMwClx1pDI7bTAKF/T9ivw1MAljc9Ke4zw
    wCMlsWSKkInAHrp5Im605sAnliZm2ldQwClC7RNOLlbAMTalMHeK6MAoazVaykk4wCTJtIKa
    /h7AI5kdkYw6w8AdnyYqS9gVwCaywnXSM77ALL5UngLOCsAevZ/s7ftOwCUUKCyeEKDAJuSc
    XJw2LMAmfzrHru7NwCR1NHrAvWrAJaqxYEiz08AqPOks3tkhwCg019dkgHjAKiHODXtJSsAw
    DIXVKB8OwCSBOiV+/lDAKO/5HlGnscAg/52lHxEywCSnHAivnirAJ9128dys2cAtG44DTpVA
    wCPup00ExtTAGQwozRNN5MAzjHOGuDk2wCcWR0kY+2DAMQlYwCwkN8AmWsj0AzSYwCuwe49t
    JIDAIppzMRKm/sAjKOCbr5CcwCtdq7QZaLrAJOI3JSUSXsAmNj/pvsxowCMcWVJnXVTAJexi
    MuOYTcAhwi7zbGsywCeEaPBgmN/AIopLERNucMAo8Gt/McMAwDDhw2OJ5lbAL2300yiH3cAV
    zr6iDtOkwCYxHD7SCZDAMgfTq1f3UsAfj1R3U9g0wC+6JNteAhLAJo2uwGKZX8Ar+IIMlxZv
    wCBXlO24NlPAKh2ZOGCfHMAmmH7Ek5FWwCYhtJES7AvAKSfn9XYpr8ApYj1E9dShwC8r3Uvy
    ibnAMCEmG6hN3MAjzUPrNbnhwCxyAcmrkoPAJowwV0iVfcAqI+EU76WYwCY6Gp9pOcTAL6ZL
    etkNt8AhjwJQA2dswCRb+/a4++HAJ1CwqLWcJ8AwMWxZepSGwCjndVgMq2rAKTJvmz46vcAq
    /4kU85YOwCnH+AHLfJrAMpzYdo/dq8AgjGiTnheuwCa+rsU9BJ3AKeBXUy5cGMAjImE9ka1j
    wCTpjT+TkvbAJGc4ppaeaMAmNWJnrbqGwCEzc+82ymLAKfCtjXBuP8AlORuoRfXdwCt5l6M3
    KUDALanxhxDarsAlChds15b+wDcV3om8NpHAMhJcLxmXssAlJ5QgdB2HwCyi3B3+45TAMIu9
    HL3HS8AgJkF7OxLGwCawWXVfa73AJ4lVcW/4HMAtcqfjW0kowB5dVRjc6dPALBcJS/l1wsAi
    DT6wkXwUwC+p7KUO9cPAJecoKnuM8MAuMAtKlfNwwCtvPAFZa/XAMRpw7/YSYsAmbLH5YhPy
    wCH0hzTZofnALCRah+0t5MAhaoXvN/kQwCotJY+A+EzAJLR+aLOA4sAkCU8XTXT7wCtdvj3R
    dp/AJmZHa1GAyMAxMH6bTrOEwCfh/COdVA3AJQsLMm/Ls8Ap6HWX2aAEwC6Uf4tYarPAJ5pw
    IRUG7sAmQjjdiBEPwDBVYceYNg3AIqf4yapFosAqCfSMS52RwB6HUEaVPSXAKOIgdXeGHMAi
    fuV4tVfYwCnbt8jD9U7AK6qzjG0o2MAkpGWVBhnywCgKmHXchw7AJpzC0KoYn8AttXYnjQSc
    wChjgxCBFkTAIAssN2KvQ8Al7Lh/3onewCd9CaJBPzLANAj2HgMCbsAtMo5b/u4awCd1S0NC
    VgLAJVmwpA+jLsAhlApXpLG6wCYEpvz9P7/AI4ZIGWShC8AzsK2+wlZzwCkLq1Fp/hDAHpn3
    evyQzsApvuaOZJ0mwCa4EyFrf0LAKeqFKqpFosAmBtY0QGfhwCTt8orIDEbAJ+AXixNqi8Ai
    bCix721VwCJf9dkwoN/AJtDB0HJKacAsePoii4PswC4VP+hTJGzAMAqDkweSOMAnh1tjjlmO
    wCPS97+mX8LALVuRLBqUxcAlWN9G+ZmxwCtqGCToQE/ALpSX78AMc8AiQzPoU1lgwDFYIk8+
    yHPAI3Xv2UlCGsAvqEOGfA41wCrHg/RLyrrAJ64mbxkcrcAsixMwnL2DwCVZaQNpKWbALMcx
    tbfDTsAp99IFPT9cwCZld/qJqJLAJk1doxF+RMAgmTvAjw88wCY3T7l/IJDALirLDAsfIMAm
    mEusYyPzwCnccK3bCL/AIgA7TImAuMAlvh6Uhyl0wC8VA1Oh3zbALji+hQOp2MAg2jS9TjPI
    wB5A0g6XnHzAJkzWTGxJqsAq9qiL77TgwDR1KltNz27AMNvBxYtpQ8AqUXAm4/xdwCuKAcaI
    jbbAHL8J7XUZyMAmo8GnjHtRwC0pe4YZqwTAJDSAWTYxAMAju2X7sZBQwC3jSGqDOcTAJk+D
    46HhTsAmIo9L6/YbwCRipN22f9rAHn6DUvnaAsAmt/Jqf0tewCOf1VX3hwHAHYLwjNN2j8Aj
    FalGCxokwCI8q9AMUZHAIC2WlS8UrcAhBWW4nWVLwCXuPk2mAZrALcl2w9bOQsAmPTkc1NA4
    wCPOAbXoUs/AIhba0brFeMAlHiAyDYkDwCPyZsNiIojAIlHzYBd7bMAnkV1YZJ+RwCSrXDbT
    OYzAJpeuNiEKtsAxSSc4XpouwCgymE2FwdTAJpc9HdWcz8AlTlA966/ywCJENtPJHBvAJ2YU
    c1IQ2cAn7IjliKymwCbkcnw64sXAHOCLoCBHPcAmLGkKUM0owCrbHVPsrKDALkQB7NLnU8Aj
    V49S6drQwCASxNYDgOjAKO9UiLC0esAiCFvWx3vcwCtOIY/y/rXAHLbcU2fBmcAs86TXXlI8
    wCD7QCluRljAKZsq2HPt3cAw5OvsplUGwCSp/n+ZtiTAND8C5XC3McApxI2H9qBRwCqGCu89
    wmfAMFYQUdIIjsAn2UK7bDgbwCbsQ+7zK5TAMNLVplrHxMAXKZtJ7GISwCGfxjXvf93AJ3ES
    pqwiGsAlrFYtA4wgwCZxJIa7VuvAI/FlU96a78AtHymlqT+ZwCa781SpjwvAKEMlzdST3MAk
    ZZCE4HakwCDcYeicxvjALHXC0xqjtcAlR+vtZI5SwCplwQlXJ4LAI39nAfoEcMAnX3CcKweS
    wCk/B1VZMtLAKLvhysNP1sAzRK9ZVyp5wCyoHOIy/RbAKAyknbUd0sAyERYIMfGywCijCBCB
    jPnAKhwEyOaSSsAuLWKIwpSawCevTk7dAtnAKxnr9xdN5MAkxZcEHzKYwCyEu6r9Y//AJpe1
    QrVYLcAlGw7f3WFPwCMsmRVQR9DAMhlNypZIasAkA4Pn4+mVwCW6lYrudqPAJAzYu0vmXsAp
    bI3b3OuQwDYH1UR5ttDAJxS8CPp/RsArzSCzZi0twCp/a0DQuQvAJJqDXIt7ssAnPg96bFP0
    wCxl+4Hr/HjAJ5QyEHVqfsAgbt9JCrLbwBsBN9ttZ/fAJUckROwZfMAlJCtjXpfowCTKBRwN
    R2bAKecatMFRo8AmTDeozKc+wCQO65sl6FTAJGX8+eLRnsAnxE2EQJiswCnie8h0xRvAKJmN
    zislW8Ak9S0Meu5gwCY+Vbe6eMbAKH+zXrxSRMAsr+gKaDpuwCh7A8KJTS/AIaQwEMldCcAj
    uyPvYZLvwCScr1rZKsrAI3eZFwxlgMAsq1UT0+JEwCiYeYDIc+fAKvg+LNTrmcAjI36eFJcs
    wC/DJPmnAITALKcD2E9tL8Ao8lD28lgZwCCBbAWtngLAJBIZwQGXCcAjhSmuXxxqwCuRnOug
    uvLAJXnhlGXlccAnZivfZ18NwCOaDY0+27XAJzGL/uHYAcAoJBcR/MHowDBPiSvl3qfAL887
    YgAm5sAp0gngsYt8wCM2g4XgIrbAKghUSFttisAopuxtZikXwCx/6XCZoivAKPd6o0NF08Ag
    rsFtl2u+wC1BR7FkdhvAJjtjw+unI8Amsc3ipWQ7wCXxKp374FTAKU28wxARSMAtZ3zrA6Bg
    wCKLL8C7r5jAJrwcJyfgicArB+OO8f2ZwCNjsQjWMT3ALE6Q6tFyRMApgTXqfS1+wC27+bB9
    xgfALXjl1Ot9r8AvAEH+AlsywCZ1i1uQ35TAKJt0/qwmpMAsbIO3SMHpwC82bAIjkZHAIH7h
    IwyI3MAmEdffFnTJwC0MWDBDrnDAKNVtQEt61sAjXyOjrC7awC9iADvz9dzAJ1w3L7XCu8Ap
    iiu52s7wwCX4OCYCN7/AKJtjoXf+qMArTXjYUCETwCqYbVbASQXAKQKVYnIX2MAsIZgCAyce
    wC28hZnkQB/AKVYEGe9vssAl/nDeB467wCdwXIaOfXDAJ7kL/LbevsAt0WP2HqSNwCnzoGfH
    AJbAJEQS/Fn73cAq3RxDbKmZwCYtLhfyXSXAKYb9lN3nuMAnD/6fhZC0wCKM4cGDytDAJP6d
    YLAfDcAqOo1KbctgwC9dszdYQ5TAJbCIc99aUsAjm0KEGdjmwCim3EA63q3AI+CG8V4wH8Al
    HvBKUxbwwCknac12thfAIpBDeJjBC8AtI0AXY0dmwCuJdlhUphjAIoa75SiFvcApq6PrHRsE
    wCXpmEkxKObAI8CWQwJq2sAonzABrrXqwCl8DfO6VQjAJnpIryuEEMAvqnWB6sk5wCrp9mh4
    FpDAK4WPip1aNcAhTizFt84+wCEo2Is5jJrAH/zFxhQ7McArBn1sbHu8wCmcdgjWWDTAJkGN
    bpcu9MAiwmN61SL7wCPhk/47TJHAIiJ/RcETLsAl/9Kk4PtEwCFbUoR+fFXAIF+S1K1LOMAk
    MGFR6fLcwCnc0B9QbLXAKvn3aLBX3sAicEAJXSvKwChppZbR8yLAMAor18R/qsAlhTYYOb8O
    wCd8p3JVV+LAKo+j19fj4MApn5EBKPNGwCNGbK0IBHjAJxYVMyh/ZcAkvRHrubshwCO2Jvm7
    aTLAJPDuIANfEsAwB9TQuE+LwDCu/lBRhA7AIiLbZP6+v8Ak9M1HwjOvwCUbPbKjjy3AIirp
    nyNFgsAj0xfBsONOwCiSvSOW61LAKvK0j/qXk8Ak7ADDd3RKwDID4RuqdHfAJwvXkSx53MAi
    MYgOPFDewC+FQkiteOnALQVBR1WArsAoZz3JOb/0wDC9jtK+/d/AIXWjP7I2q8AhFOuKdrjB
    wCeQgRzYPknAKXcXa08dDMArvWyekFxGwCv79P42yNfAIq16iY5rbcAmJ5oT/VsrwC+OFpy3
    9QbAL7WaY5K/ycAkt2n+KfyCwCh4eG39RQDAHrszyxfGbsAmiYs0TTt8wCGg78BaoiPAJYSC
    P3yJkMAnq2jgHmMKwCWCeQqV+XDALcuM2TcE5cAZj7ksqd+GwCc8YCh2YVfAKizbDqpHWcAk
    meZAhfI4wCQhxL9co5rAJ95TgLpt8cAr+MKxZ5gDwCnjLEQbHLzAL4tlIBzGP8AjK5Ne88Is
    wDK0p7VYYvLAIJerwCeW1cAijoAiefHIwCNlImELyNXAJOz5FMgUoMAuNfblsYxiwCx/oK2u
    7TfAIb5z/6OnZsAnmDInG94YwDFmA17+67PAJajoy8S1JMAvTbaN2ycdwCaxlZXIPmvAGL93
    BnYRFMAqbqftnawTwBmlZ1r/jILAI55irGfEFcAr/JOTD6IcwCffoamPMSLAL+SR9rEOosAl
    cLTDeskPwCFP/p2BuW/AJgUHWqh+QsAcZnis/w8VwCPS2/diES7AI+djk8ItTsAiqfW9rxe3
    wCXB1wGitu7AIFBIY+b6XMAyZBmtPEn3wCa6K3vneorAKx0wUM12nsAqGGsQ/MxzwC2nTnsr
    pQzAImpo/K+ghsAvRQehmn1bwCdKyfBAN/vAJtpgORLWzMArFar2smOPwC/2cNV2vmTAMQRS
    KqHG1MAsZX1p04tewBs8UE4vf6nAK2JtHa5Y9MApBlNsti6FwCE1FmknnJ7AIfidDJJqSsAk
    fl911C4MwDg3blpX2gDAJOpjNzD/l8AjtkHtHoy1wCl2+dvSH8TAJ6SwvNICEsAsVYJpPGb+
    wCJMSm9LwUbAMGLPx48c78Am8PIT80zNwCVP7ARWeBbAJrTVnz6kTMAqvBAmfYmhwC5+zVWN
    uaTAJ1Y8YwHbjcAk+irfyzmdwCUf+oAEahDAMNP6K/GzK8Ahh3OLnkjKwCxavQWLeKrAJPQ2
    94CboMApEY4lFhTcwCGUwi9H5inAJgWMzOE0oMAyuiH+Lr2LwCeSYdzufe7AIs3UdP+WW8Ak
    89nYJXPowCXGrKx1lX7AMfSE022M+sAkaAaGfPR7wCpRhE5u5o/AKiWln02Y7sAoYG+Lo7Hs
    wCf8HjMSnHzAMkOAGUnxfsAtrMzkmlPqwCgwjSfKwmjAJPObxvoru8Al6ofvKKoawC5201aM
    LorAJaTCuJlua8AtvoVLf5GuwCjXZyA6O4nALG0QEGeNTcAh6wd1jST9wCtZFwRbLezAMAbl
    rSqV6MApKtoR4pgHwCW2zhKBYrvAJpxZfV71acAkg+gy8j80wCg8F84QTmDALfAwMh2IMMAm
    Syl9thDvwC1utj3Dg87ALTRaFmdqKsAmFY1gol5ZwCy+Nn5Vma7AK5jQLcisDsAanrFfLvrA
    wCMLPd50CcrAJjRmDz7CxcAnR6noAYAgwC13wE5xTUvAI1in8ttfD8Ano5ORdZqcwC26UY2y
    McfAKDVr681CgMAoSMu085JMwBhsSGL9JxHAKIgYnbeE48AnrMShlFWNwCiTDlzVClHAHra2
    xEC2JMAtduOtPsa6wCY0K2fqUu7AIr6V5GR3ZsAkmCiuCd86wCWxiiThVBPAL8FEEXD9rcAj
    55IUSPYywCFkzOBYkirAI5xq0TM9hMAvJbom5RcQwCom9J9rr7LAJ+k/KZXZZcAp6vXViSAy
    wCAj4UrAdpjAJqu21pbNS8Am6wRMJxY/wCsg7wCyhgvALLiWx0sfdcAwZ2IKsRRWwCkvnARK
    nXrAJQ2Iwjv+UMAtwN85//qQwCMgn3ZkgJvAJOoA6Z/FAMAkNUJaxzzCwCA1LRJrTFPAJyGq
    wXKhxsAlVhWbGQ+DwCNoooK6KvTAItDRwbVNTcAnTRvz5oytwCipvIVgF0zAKRVJ6GScZsAo
    WpbewSWIwCtJNSww9DnAJdkNPJDCksAt010oyQrKwCdRqZd38N7AJ2nBpma4rsAnwLAao4Bl
    wCct5icJYk7AIUJqhr8uS8Am+alirM05wCMyQdI0zmnAJ1lrZUUijsAgAhQ5kSECwClFxZVb
    +N3ALPChZcDVYcAmccZ3hNq9wBwY8bU4F+vAJA/w77Rg0sArk7is6DsqwCO8EO5vtFnAIKim
    DSpibsAw1qug884ZwCorFjDKdg3AMjrhDyukXsAmkSIvI3EiwCaBFxBvdhPAI57b553yasAk
    VptoypdrwCU+a1E9MyDALZw5HHZrd8Ak5BQhvMbCwCywY6tC91rAKM0XIP5kJ8AnnEBzUoHh
    wB+fwBbczJvAHfyyFcYDXMAmJixKhijtwCdTqMNeyrjAKWxp1uTPxMAsOMk+ewFWwC6T+5Aj
    dKPAIUJxyscNGMArAPaDEPFOwCoa285U7GTAJYgIEIksvsAn0o6BbIbkwC1sIu6Y5kPAIaK6
    E4FZyMAwGY2sSZLkwCdWiLj/BebALynZr85WEsAf0BNmpgrCwCtkW1UtuObAJ6i/WAwuPMAn
    JJDpPXbjwCtSXfu27tbAJ40f24QV/8AoljbKkhtcwB5YagKjYR7AJjSK3ibbmcAi+jkMN6lV
    wChMG4XQvMzAI0NlNtSWSsAoT+PvMBBwwCcUYnvXA/DALOvHvMIBxMAjRPmW/d2ywDJUtAd3
    I4fAIrKJ9VWVEcAcAml5aJN6wCgU22UQoDfAIImA0Ul9RsAndgQ4UmPswBhrW/Q4K4PAITab
    KFNdFMAnHkQAOZYOwCiHhN3dgC3AKD+Pfn/FGMAm32x8dtLlwCXgSN3fSWbAKREgXgHhjsAq
    v6+3mDoqwCgz5KO6wk/ALjEuG2ovEsAj8dfoNsHNwCtvUecJ0u7AIWFEJ8Ey9MApX7Hl15I0
    wCU4YX24iYDAJiuiUFJmMsAoUqZ4hX3UwCWLuvEU6EXAIzRglkbK5MAtVWdh3OGkwCmQw1SX
    vNbAL/dgqVhsjsAnsawuDuq+wCsF6U0VQDXAIPKR2NhJC8AnvNizla5+wCevt0HJ0uTAIQMl
    YAPleMAnYYT8JfhDwDF8Wy9t/WfALRHl3hcfhcAr1EAeEJmBwCHAYy1WDqTAJQo+ywgT0cAh
    Ye1+AgxowCP7n7hcYMjAI4yWLXMt9cAhYzgOsb5zwCbn8R3cWlfAJR+5urxny8AoSx97YnfS
    wCSDZXw7gzfALRdjNzVVScAplGTzVjk1wBwTFriOQDrAMRJCFUL91sAzumeKf8X3wCU+4ssF
    lh/AJi1WnO9bOMAitN90xOG2wCy+Fs/5F+DAJXQZYiGOicAnVedaDO3pwDFLmP+hxILAJ+Hb
    2c4BscAvKvFuLWcAwCobShAYcpzAKD/VOlrDDsAjQ1pgdShjwB8R7iz6cPTAHOjWHwq2qMAl
    2R0fO1lfwCsNOIZR587AJrZWpxsYnsAs8t+oxSjWwCDJ/9Keb2jAI7Eyhol/tMAqSV521C6c
    wCy34LflJwXAHjL2FUoM3sAsnKzK2wdpwCi9LzToXw7AJMA5sJksjMAmGYRiZ4iCwCGsgkr9
    V4fAKJ5jvzV2SsAiXidy7LR5wCKBOW1SRaPAJylGruS9YsAnW75NJEiRwCrO1NX2llLAJScf
    zstc3sArgW885fzKwCkLMxcOeTPAKWnBXURi0sAoSOIrRHGYwB2TvNLawyzAKgPkFX0l3sAo
    IFYKou/XwCWzxxQt683AIz2rDYivtsAkarvjp4/6wDFeLjuOrunALBnymVeSXMAiAAqcftzc
    wCm4HP8eQKLAIvt3fM93IMAqiuthq38xwBlmjtdT+jbAKM0b7hl9uMAi6uCHiJ6dwCByux5O
    OInAI2tYdyDbWMAgqNCXzayUwCIBFTR4wPbAKDcH5IllEsAr+LZQfgASwDDY3pMA/5TALlp3
    KZHKWsAoVd17n/ofwCkQfJSY5lLAIdmOL/OWBMAuhzaJ7S15wCwxgDlXXR3AIhFwNKAAh8Aj
    zDytPxiJwCuAiLJccZ3AK4DcWO8IMcAq+LaktARswCiWnEJ+Rj/AKBomnhkKIMAjTIrBC8VG
    wCqk+SgMZADAMOPqxxU6wcAo6lXkGVUlwCUHGzqwxA7AMlBbVwV788Anxp0XvrnIwCl/WD9q
    YX/AK6ETHmtX0MAl283PS9vJwCABhIYkPzHAJgQx6dLTZMAmukEwBAPuwCJmGgDXbEDAKzZE
    rpIuNsArdYI2DJhCwC/eL+bYbKHAIin9+3omj8AgjaRZL1nkwCb3PpFU4MjAIq7PIX9RC8As
    hfEzqtYowCpZ3dtf2XXAKrX2+6h0zMAoYIsOb7LWwDKtyN4st/7AIafYUeUaosAmz/r178ch
    wConeUWIIh3AL0JPbFpyXcAvTaKeyzqmwCMBS0L0eX7AJsHfEwXP/MAmOf4gG56lwCL4/1x9
    0TDAMOApawzg18Ag7X+UXryVwC82ZdbmAtTAGlc+5mDV68Ak1bc/YBGRwB6nZuYaJIHAIsOF
    DQSn4MAiqY2NsiuYwCQLH6ghWjjAK1bV8UygqMArrmTJtUTXwDM7b8p0/fPAIsRCKSI6OcAk
    n/0COc2BwCEKePs/4QPAL6HCsyPBzMAyx/geutHdwC0TTTPYkIHAKjBAnQLVRMApyc2nzjXs
    wCcNiLXv7+DALo5jIsSJ/sAy2/oPMMZcwCG9sDCsVePAHxAb81A+aMAn/HrblXCywCO5ZeDl
    RPLAIyF/25AdjcAlYXU4A26TwCXWd5BiqRrAM7bbl4XBM8AgkFYt51ZVwCLgT9fDrYrAKxMf
    cs9bqMAkFfLIqL/JwCFEinRTgEHAMSngIDHplcAlFL6uo+khwCwcDPfCFbjAJ6N06xuBBcAw
    3nuFCOyHwCVYeTjbR0XAIe6jtu3CRMAmSHGu+1TLwCWGsSmRBSzAI+WUBd82nsAtoIsGdcC8
    wCMTPhnhxFTALtNGk0FKzMAw/+hz8LAGwCscXbk/0aHALtbK6it9/sAez91HPyR2wCzsQEIP
    +3/ALuAHMyjc7sAk4UVd85NKwCU5XeSjJrTAHvaALgam4MAkql9cxAm6wCc9FrtMYwXAGRWT
    2k8zjMAxVKdRKB4awCYJdgm7OHvAK3o8uglStsAhTv5ciuFdwCUuZSb08ZfAJoDVlal24sAq
    plMHkyudwCRNcBCnoqvAMZLzS1ByoMArXM9180+VwCB/6prufxnAJQeuVcG6vMA0Lq+4nzVd
    wCMGeg76GbzALdw2KGeTfsAnBs9aJwquwCnmUYOTecPAK6mNg9OFqMAlcVVwb4WXwCuV2Xhv
    ZXbAIRD00Qt9XMAla4Du73yWwCTkwRLkeufAJ3kMOLuaX8AohPujzDbiwDhjYIOwCIDALlf9
    Mbuw98AmcvUxCX6bwB8TOddY1HjAISVjZpM7MsArxDeBnyyywCScYlOEu3PAJrmjTiiH4sAn
    qTyTag6OwCIftVARN0jAI5PN8nWhucAuOvtNpSr3wCU1XHSgEvHAJhOkz6v9/sAqTS4Ddo6L
    wCl0ihe1TKzAIx8orS6T08Ap6zo5XYjywDGXYLWW3o/AK/AE8yeWcMAl6uFZWlE+wCR74CqQ
    RkfAJPQhvY6Xo8Aqm4FmHcJAwCPVxROWWILAJYv8WlyV0MAnGvKmiJrVwCk4ZEFcN9DAJt2w
    XB1Pi8An6r6BifPvwCe1wEU536bAJFcOfaZ21MAlxkzWXkfcwCpCXLrvOYvAI9/hJwPUIcAn
    UBZIBrZvwCFjA2LXnWzAH4DKtNDgf8Ak4WFyzDOgwCOkdGe+ERzAJQUBG4I4+sAkFp77fxtY
    wCGeMhJY/LLAJVRfimt18sAkGDyXg40pwDTmmzMBChrAJvm4un4/i8ArwOyaaNawwCizjwGb
    p9TAHsRlO51daMAnROcroRoDwClLe1yDlj7AKn+w92KqHsAq1E9PQ4ViwCewoeuEsjLALTun
    HQ+EQsAizec1vAk0wCUnnStWQH7AJ4AVT8Hg/cAtISgMDlOpwCqGZhKZk87AIuS8YU9q58Al
    TCNSFm8bwCXgrDyCXkzAJeR9LfTKg8AjLpObO7GuwCl8qkqBZevAJwNhMOm8TsAofbdZwdCa
    wCvjH6bc7FfAJD5k5voEuMAq9GBvAgatwCLUg8NmfWLAI00IZkS5KsAvhnJ7FXKhwDAt+uVH
    TrLALKs8BD9/AMAn7zz8A3PjwCiJix0F6EbAGQ02XBCa7MAkkRiqFd9fwCqT/SXI6dnAK4jU
    XA0aPcAo5XQyGrc2wCLHKLoSm4LAGWURTPkcmsAjqiLkseC2wCmPProticvAKjrM5CI3PsAm
    28WTpo+uwCpfQsiQIQLAIjLWTwtCUcAndvbcVxEhwCg1MYb4RSjALohCe7AWX8AfsyWRU9sQ
    wCZCYDVG6R7AIbnedNAPPcAm7nxhM12CwCWikfA65GDALLrWLBn298AoikydnRT9wC7ttJZE
    42PAJg+o4BVzA8Aj/DTsv7hmwCYVjJGC/DzAJvq6ewObO8Akrnmz7PAxwCMqs18L1rLAIzIa
    RUe84MAnnVGaFnOTwCsYf33of8HALsGMEj+mNMAopJLJYAYRwClYNIIicfnAJsGGo4yJh8Al
    r20XWVEcwC5d1Mjbiu7AJLVSaSNdtcAlcUoyhlJwwB0MLBESkDTAMNzDkE5AmsAkPFHt8YXm
    wDDvrK/sdW/AKDhQGxF8SMApZAOvdleNwCds/C5oqz7AJPZxHjTSucAjPt1cdasCwCK6hK4O
    srnAKMlOlqouOcAn19GiJ7BTwCxMJ0B+tWLAMKeOv16OyMApk63GeAZOwBzJACzlGKvALEuV
    +J1WocArzotrvhoxwCgR8riPmBLAJC0+4Lk6F8Am66gkEoMpwCMaKLhaLwfAKNOEkY0bxsAw
    eCaZODlkwCLb6r9wW6zAIMGEIDHOqMAxVOayKUl4wCrar0iKoLPAI18f4AKqDcAtY0cjBha8
    wCtOi4KGqnHAKVjRLQkzyMArl4NSTx4RwCS/Wgtvy/LAMENfrH0ZPcAn/cjhBVPGwChgs/Gs
    /PDAIHUFJDV2wsAmdIzI1q6rwCfefDjo35TAIOBwWlvFksAsLb/osby/wCox7fUK987AJOS7
    mAbnF8Aj683Dj45bwCQzXrioe6LALu0IgXQduMAl0JNMLKlKwCpgwTmfW0XAI/5l15wX4MAo
    9VDGusLWwCnLdr0P5KfAMjyVo9+bxMAwkvVEJFqhwC7QcxNp5RzAJa64pegeAMAuntz6TUI8
    wCSalkImSEvALTPuSeCYN8AmIVDRS+/XwDKq7mj2u6PAKW89+XKUqMApibl9xEcswCU3nDda
    wZ3ALbL/XdVs+cAh51xdXIwvwCYjun060qvALA+aNJt9osAnz6bwvbDywCV2k/Lj8HXALP3U
    MpF6EMApowV5PMlrwCteMgxfWa/AKpidTH2N28AloIhTpdAjwCgS+L+XDjvAKhfQLpFbt8Ap
    PotZlfSKwCvcGTZupALAKVJH3KozcMArB8LJZuStwCT1aIvg+pPAJhsZ4s1tesAj1A4lj65g
    wCgbqnphhwTAJyUqaFFpMcArgmHQhfxdwCkX+0pUFJPANMUyeLCB8cAoCZlOte3KwCWt+orb
    a4jAMfhGkzfYrsAmOidYNsxewCgDNHthos7AIqqP1du6l8AkoWBSDLA2wCdCCjJmWBHAIeDH
    LeTm/cAkrji6Gc19wC9EfiHP2JXALrj9yUaqQMAbCANAUysAwCJnYtmpzJfAKCn9Hd5QD8Ah
    Ea+YkFeBwCMOKZLntBrAKQvGysBjHcAjMye3wxI4wCF9z/7yQQrAJrrsVTrUkMAraeRnYhLC
    wCveD8ZQbWPAKFaSU41cMMAu9lqkTlApwCdUmy2fpVfALJS5SEmuWsApV/LJ26cywCH6iorM
    x+/AKhP7LeI/+MAl17jtMEoawCRkzRDFwTHAJWgpFTXtksAt8en7U0QEwC70VsCu34jAJxUJ
    PenAGcAntLlYXSRSwCZuQtN5KdbALS0MbiXr48Anqp2770p6wCX4hVutIu7ALmAOi8bejMAm
    4UpEy4rGwCitx/wHTRDAJqi0N5FcsMArk+30fvGewCY/q2Mso83AJVaYhUdEsMAjpiOYwTsu
    wCbWC47OJ4/ALYlLlO58AcAiDnaw0a/twDSACHiVLM3AJeyl5eAlVMAfVXUnBxR0wCGGuP3v
    fDDAI1TJmxHGu8An57Zi+CXUwCAjdBbcNofAJxSQZJeypsAwoPnZQbkYwCGQSTRjVJXAIdYF
    4Uu48MAot2LH7e01wCCL+0Ko1sLAK8/xuod5/cAyDThgt0WQwCG8iu6b/UnAJRbHeLlNjsAw
    IjZerjXowCdya6uVeHXAJ360Xwxuh8AsslnTVEyBwCicZz/6PdrAKCOk+w1bmsAhUX/9O44T
    wCey8scwCMHALhavZu3LYsAiHUHM78I6wCyMhSNPECbAH7lqVVsRTMAkozCyi6uAwCeW0YkT
    fiTAKR0FOh7l+MArZhO15G50wCWQtSqT+STAI0aFARNl5MAis9bexXoBwDIHJRcOot/AIZwQ
    B3SVd8AjEdVhb5gDwCG1nSXNNbfAJSUOdwvp1sAhEam/1gxHwCPO6JUOx9vAH1SWDiBmgcAi
    ARBdMayvwCwLaVwcwvTAJ1pu8KFbB8At8UywNqxDwCuvwz+2LtHAK8tT7zimksAmMtrtKuUC
    wCRJPGXDGEbAIinbXdYZRMAowkvHOFH0wC3NhphV8aTALH0I3PTXB8AkKhklxFHGwCjlKW1P
    RtbAKuRVXGtAxsAmOTDJ8UtGwCnl8DJG99LAIuaAT8JRisAmG6g9gzzFwDD3XpTtTFTAKzMQ
    o+xtbsAkoRp22idMwClA3LXV8oLAKRSNSXS168Ao2PktoYGlwCZL6woa2YDAJi7OuodN3MAu
    vcOry6jswChDqbbFZqTAJIe4PKh42cAqigzNzcnqwCa5KxvvvJjAKPwIHHCOVMAuj/seQW7I
    wC52V0YYIVTAJsnzEoCoqsAhtRwLBbdCwB8MSZhT7AjAKyF1YlC4DMAg6xAB8v3cwC6BKxd0
    I9LAJrci5TwH4sAiM5sF8Pi7wB1io/yBIsXAL8/ILlLm/cAm9WH9yIvfwCobjKgpu07AI7Fh
    7MvxasAjGUvtf/buwCKlYVbBFBXAITvQoztSpsAm/PzBOignwCni2gpwxqDAJc1A1ZUH68Ap
    ik+eVT6kwDBKeg4MtE/AI1dPVVBedsAgZLYeEYa3wCw/WjIKyvnALxbbzexTV8AvlCuZwO2k
    wDA6h2aQfMHAKiaFqN9twsAow7xD/crCwCn5hmeRBenAIuUg+iFKvcAngZ5vwiMuwCQQUHZj
    I9jAKZIySomQLMAemnCBwZQGwCd84da3UATAMPSu1nC8PsAju9y1CQevwCu4bulOwCLAKinZ
    U23fKsAqmjhia4VJwCGZgohHKpvAILIdqjB+ocAjpF9G6CbOwCg4pOziYCrAJ66HIRytB8Aq
    iLZbiWY7wCN63fho57PAJACaBJrU5sAqPRmEekg0wCbxFnV4J2/ALUpvKm2yucAgVbj0w1Pc
    wCwxckwveuzAJfFgIlMB+8Ak1K+wSJVIwCJQxMZw0qvALnqpLcph7sApCyspqpEYwCa8DKkG
    Sj/AMj5F/syRWMAk9KeNl6OGwCzEvXKdXvvAJYZbQq5PJ8AyYufDjxFHwCOH8sQ5zvjAHve+
    oQiXkcArzIZcq9DewDN6awwXPy/AKSUfVMG9jsAqpZONHcI5wCE9CPj7B5XAKNQf3+QMXsAl
    nFnOZGLfwC3DdFkSXPfAKYbODqlP/sAiS4jaeY8uwCkTQ+X66kzAI4HXhim0jcAjxnPfHk6j
    wB7YTN+c8EXAJ3gsWdrzfsAj1V7mL/cEwCC9sig7/aHAIU7HdLBtd8Aio22ZwkGGwCUdiAzK
    1czAJ58UGK9/9cAoXZQAzjwOwCP8CoE6YNLAIlRQ88fzpcAlO+w++R9owCzB4wCeb6rAJES1
    lryc4sAqpoXcLuCwwCmhOE2ehSjAJqSPEjFpvsAnzHkpBhV1wC0UefLjCkPAK4QRH45Rm8At
    gNapl9aRwCWoQqAkJsDALU4CxGMz58AnWUT7YoxewCYlvcueH9zAJj/oTwJPr8AoChNMV47X
    wCyKIh2ECXfAJsJzaNpXAsAlZrm71xF1wCIy2a66dBLALHT+WpN1VsAofsNKSMevwDA4oK0b
    Ar/AIZa/262STsAi3w161XvdwCqvIIczueLAKXXSm9A56MAw8QbZWBn8wDBNjuxJ4hrAIJyz
    8g8hLcAn/jA08dCGwCKsAuyyUd7AKM02iaGp7MApbZgJjOxwwCCI3CEBIbzAKje6uHYHF8As
    lIUqdlaHwCi0h9CUAf3AJhf8n9DHzsAh/+Q1arK3wCIlekh0ujHAIMdoTNP7BsAiSVbQbrr6
    wCOhoyDXsdvALdxDEsVT/sAjzjocagDJwCIBQXTEz7XAJa5Mqdk+xsAj3QDC4eeVwCvQtHVX
    F7bAJ8rVBTjGosAuAHPd8hS6wDHHFto1M7/AKRlBTHcFkcAsggtoWL76wC4W3Lzs99DALQur
    e5kNVcAzsckbi8OjwCmx2EVVpivAJ5u4E87Ho8AqHktMSO/BwCw3ULmM9fTAK2+0nMvJLsAq
    7oBYX1aXwChKXsy+OeHAKrx69jBZ8MAe0nGc22ycwCg2yVAHYJrAJEH8exlhlMAxbniEhsIN
    wCYwLb8wD7bAIt5AkiDGvMAmk5JeeEsswCGeYjKg+kvAKinv2mS628AqPENnvTQFwCg6K0iJ
    My3ALVLIVHW1W8AjLiik3M8OwCSJs/BHdHXAKOilFwx2h8AkrrRCPp9wwCNh2A6oK6/ALQ1H
    kaHkKcAsOA/UaYIGwDDDQdMDiFLAJzVd7YsDqcAq1JObAi/owCIq6NKrMC7AKcnH3BVcDMAg
    Lq6n4UVywCEj6J8nDBDAJcsFrSXODsAnItK4jPk2wCROfZ4wyBrAJAFOGI0HwsAnkih7wlvl
    wCaatjc88evAMy0J0ySrm8Aq1XMhqAlywCctII9xY+zAKfN2lU5EP8AnWjdLJSsQwCL7PvWW
    kwzAMCiUvgVnKsAkM65MGnIqwCp32o6Nkz3AJiEQ0GaW1cAuBLZHA+GQwCZ0xXsYfE/AJOY7
    cIu7iMAg2iC5GCGCwC50y5Ki49DAJAvkSAkJWcAg0BhfV/VgwCXIRyFHA3LAJCZyxdccusAo
    SA6avWKCwCL3HRmFpynAKRIgxNtzdsAjuQ9J0z/xwCR7N0skyuXAKZskPmGPtMAvbHpRncs7
    wCItCje3WSfAIyvujMPjEsArMnAgnS2swCN6l7Cf/orAK8N4vRoVX8Al7sv+4pn2wCvVh0hE
    iErALaJufp2wRcAixxiKOYbcwCsboFxkekLAJznpj+U3xMAxRkfN6wnuwC8URiQr/7LAKfFp
    2snOecAoerEcWVdPwCkWIvX+uWvAJB7KVWYFuMAkGLzB/iGmwCWLGN4a1b7AKq0Dk3d678Av
    qm0KPx+DwCKQj3TR44TAJGmtHp65ZMAwEwsmICZswC5mj246KTPAJ0mTlB2NGsAktLrW2ljT
    wC2b0pyvZZzAJIajshSetcAlTeUXn+68wCPlXnCWWozAMODf/sjccMAa9HaKgHSwwCwOtn9x
    WSLAHC4657XZcMAnwlAK4kuUwCXh3hlEM6HAKXYwiEyKD8AlCJNVBGupwC4zoTN0w+7AMOmG
    hrA3AsAiv7ctcmuywC+Ee39nmdLAKFgzyPhPgMAo9hUhuGhnwCXPb1OwtWzAJXR5EXVofcAr
    dsmEiduewCjI5vQiDsHAJ7erMOWa1MAl62lzG3ZDwCEzkTss/yjAJ+2Sq7x0TsAsJPkLZesQ
    wChUm9HH10nAId1Zsnm1/MAnL9XPaEcmwCA8cC9oTarAKlZUErgMFMAqJXcu1qDWwCqrXiqT
    6pDAK1p+ghIF9sAtpeku7i/WwCbsCBQw5rDAJTs/IG1fCsAu286E9ibcwCYslvieDbDAJNoE
    iISpwcAn6h3wUO0RwB8Zt5iHBY7ALEantWz2qcAmaYINF2nBwB4uc2LVF9TAKh6B4YeUIMAr
    D2xZcb4NwCLxOwg4Do/AJ84nvoxY1sAxFIYimCniwDB/0pcrJGLAK/RfB8P3xMAkfeYnxi7u
    wC2AmYjJcfvAKKtiK6Kvn8AkuJN9pVbVwCWrZsCL5U7AKKUU9scXi8AkA/p0AzF3wCXWJz06
    +ErAJu6Unmq3HMAmVIT3adrGwCOTmFEcLu7AKvvTCt1XscAtTQHFV3ZkwCWzFXAi0P7AJzsV
    p45lSsAr7xJ1kQ/awDFfG2viTF3AI2oFflkgxMAnO6tI+zs0wB8UqLSCrgLALfoHYfWJjsAt
    K/xtQ9tCwC+ml64+cwbAI3DoSE2u+MAnLak3Nj/+wCZpBwsjW1vAJ6TfyvQ/GsAwKB5R0j86
    wCCqSDizlUHAJdZj0xGnLMAkQ1sk2bvrwCVlf8SUlsbAJh9TVc/vQcAvX6PX1yyiwCqE2kxf
    DCXALqSvrJ9Uz8AgTcRDlUiywCqaL9bP3mnAL2+RBdP1+sAjvSIMjRgOwCkt7l+IUh/ALv4M
    zDdaRsApO2Z03PF8wCbJPe4VqYjAIWXMxSoaNMAmBFpChK7hwCsDB+CXbYvAJ2Mm41d0g8Ak
    0lrUD/RmwCIVBNOMKtvAMLQQx8EmqsAjpXhAbxnSwCLwufLvasDAIfm8iY9WFMAm93pWW8NE
    wCb6uZU4itbAJ2MAqxus+sAtHJYavommwDIS9DyhnhPAJeqsgiyHPsAwaLsQ+6/qwCYMyPcK
    S/XAIu/dIIyFxMAjXL2JEnmRwCbW+gPIa/3ALcmVYKJKQMAtV5TkkTS+wCp17kBZNFTAKDIU
    VknQJ8AoOIb3o/R6wCSokKmFuYzAKrybA9ofq8AkooXJwdkuwCp0A8RZ7+/AKCunHigpJMAk
    Crfh5CEAwCMBh02e5M7AJlCt3YIpF8An8JSWJR1ZwCQCnKlZNz7AIUybpgZ+V8AgLpd1Ir0X
    wCLOedTpgXPAL5CY9JUFc8An+JszvDlDwC0fk16k8ozAH2KKJtELwsAsdn1cYnzNwCOuRJ2f
    nvjAJ2y7reeinsAhIiy2pNJFwCoft8uvZLbAK2wgNmhxPcAsxexv/xIfwDB1jS5kLUzAKbBQ
    OMuVpMAq/Bc6bbqYwCV/oo6GQyvAJOLPyV77rsApS9PzkoMkwCfsVdK6XorAHZN3pl6s+sAq
    JK/BEGjBwCgJN4Nz6YjAIjvII+4XwsAp1ceVrYaEwCey3TGxBrLALYb2k/pw1MArrd5VTgeV
    wDIqQpq1R3fAIXLzrjmM0sAn3wAzOZDGwCNn4qQX46jAJon2PkMH2MAnpP4LXmLwwDF0Ebxt
    ieTAKESfUISNecAiNe/0EnAPwCek6WKc2YHAKduNgBeANsAxMqADiN0UwCnIiQD009XAI0RW
    18PJuMAhtUxBZs2mwC4n8TTNPZrAJ9X6zraHSMAsjnxJqtzBwCtCUGZ2ZQbAKi51lBAY6MAk
    m/Evvo8owCTv/VFkAWnAJGfNbrMoKcAjDKp2Y2oewB9oME70E8DAI6x28ELH5sAnaL7rOu9x
    wCtda2YlkOrAJt6wwhxoOMAn5PSZgQouwCmQwvGbPNHAKLgYfkX92cAmM59XRq0dwCS2us1C
    vajAKUcQLLxsqcAqHBjGlMOGwC3EwdXzL83AKQ3u7GkzXcAypWnY2N7XwCPv/rT7CRrAJ512
    zUyyZ8AukZm2zh8kwCsCCknqp8PAKObSMj/tGMApwYK6rzlpwCQ0cmyj687AIpcmbsoeF8Ai
    zBLA4ZOXwCkrcQLbmMrAJ+N2sam1bMAlpitNW4SowCvzLoodnprAJHqUuEpR5MAjaINSwNbj
    wCOfAWE2U5rAK45lY8dJ9cAkuMAfsY1ZwCqftarQk5fAIZigZC4NCcAonNys7gjVwCvbILkX
    1qHAMkFd+IMlVMAlmhwBr+u2wCrsfq/IqmLAMCZhaI8Ej8Azru6dRrUKwCKk8XdB6XrAJTC6
    h4rGnMAmaq0Zsj9BwCRTGur9Wn7AKY3wcEFYGMAl6SqJoPaCwCdT3hgPXebAKUcD0VViXMAy
    XmUM58tmwCi+3+DPBgnAJYPIAtRdEMAjqNuY33MdwDPaX7VI+vTAKZAnYmqK0cAk9J5BvIFf
    wC6tMmYaL2bAI5UGBra55MAtm7iKuVWkwC7LKTJRFiHAGXU3OUNcIMAoWQyni1xKwCUUyPHo
    Y3/AJ1uvQftfW8Am1zbR0XdmwCiGEvlWsO/AL1pZUMU8EAAAAA4AAAABP+71T2guFhoAAAQC
    AAAAAQAEAAkAAAAFbmFtZXMAAAAQAAAABAAEAAkAAAADbHdpAAQACQAAAARsd2ZpAAQACQAA
    AAhsb2dfbGlraQAEAAkAAAAHcl9lZmZfaQAAAP4=

---
title: "Using the loo package (version >= 2.0.0)"
author: "Aki Vehtari and Jonah Gabry"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Using the loo package}
-->
```{r, child="children/SETTINGS-knitr.txt"}
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

This vignette demonstrates how to use the __loo__ package to carry out 
Pareto smoothed importance-sampling leave-one-out cross-validation (PSIS-LOO) 
for purposes of model checking and model comparison. 

In this vignette we can't provide all necessary background information on
PSIS-LOO and its diagnostics (Pareto $k$ and effective sample size), so we
encourage readers to refer to the following papers for more details:

* Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. _Statistics and Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. Links: [published](https://link.springer.com/article/10.1007/s11222-016-9696-4) | [preprint arXiv](https://arxiv.org/abs/1507.04544).

* Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)
 

# Setup

In addition to the __loo__ package, we'll also be using __rstanarm__ and 
__bayesplot__:

```{r setup, message=FALSE}
library("rstanarm")
library("bayesplot")
library("loo")
```


# Example: Poisson vs negative binomial for the roaches dataset

## Background and model fitting 

The Poisson and negative binomial regression models used below in our example, 
as well as the `stan_glm` function used to fit the models,
are covered in more depth in the __rstanarm__ vignette 
[_Estimating Generalized Linear Models for Count Data with rstanarm_](http://mc-stan.org/rstanarm/articles/count.html). In the rest of 
this vignette we will assume the reader is already familiar with these kinds of
models.


### Roaches data

The example data we'll use comes from Chapter 8.3 of Gelman and Hill (2007). We
want to make inferences about the efficacy of a certain pest management system
at reducing the number of roaches in urban apartments. Here is how Gelman and
Hill describe the experiment and data (pg. 161):

> the treatment and control were applied to 160 and 104 apartments, respectively, and the outcome measurement $y_i$ in each apartment $i$ was the number of roaches caught in a set of traps. Different apartments had traps for different numbers of days

In addition to an intercept, the regression predictors for the model are
`roach1`, the pre-treatment number of roaches (rescaled above to be in units of
hundreds), the treatment indicator `treatment`, and a variable indicating
whether the apartment is in a building restricted to elderly residents `senior`.
Because the number of days for which the roach traps were used is not the same
for all apartments in the sample, we use the `offset` argument to specify that
`log(exposure2)` should be added to the linear predictor.

```{r data}
# the 'roaches' data frame is included with the rstanarm package
data(roaches)
str(roaches)

# rescale to units of hundreds of roaches
roaches$roach1 <- roaches$roach1 / 100
```

### Fit Poisson model

We'll fit a simple Poisson regression model using the `stan_glm` function from
the __rstanarm__ package.

```{r count-roaches-mcmc, results="hide"}
fit1 <-
  stan_glm(
    formula = y ~ roach1 + treatment + senior,
    offset = log(exposure2),
    data = roaches,
    family = poisson(link = "log"),
    prior = normal(0, 2.5, autoscale = TRUE),
    prior_intercept = normal(0, 5, autoscale = TRUE),
    seed = 12345
  )
```

Usually we would also run posterior predictive checks as shown in the 
__rstanarm__ vignette 
[Estimating Generalized Linear Models for Count Data with rstanarm](http://mc-stan.org/rstanarm/articles/count.html), but 
here we focus only on methods provided by the __loo__ package.

<br>

## Using the __loo__ package for model checking and comparison

_Although cross-validation is mostly used for model comparison, it is also useful for model checking._

### Computing PSIS-LOO and checking diagnostics

We start by computing PSIS-LOO with the `loo` function. Since we fit our model
using __rstanarm__ we can use the `loo` method for `stanreg` objects (fitted
model objects from __rstanarm__), which doesn't require us to first extract the
pointwise log-likelihood values. If we had written our own Stan program instead
of using __rstanarm__ we would pass an array or matrix of log-likelihood values
to the `loo` function (see, e.g. `help("loo.array", package = "loo")`). We'll
also use the argument `save_psis = TRUE` to save some intermediate results to be
re-used later.

```{r loo1}
loo1 <- loo(fit1, save_psis = TRUE)
```

`loo` gives us warnings about the Pareto diagnostics, which indicate that for
some observations the leave-one-out posteriors are different enough from the
full posterior that importance-sampling is not able to correct the difference.
We can see more details by printing the `loo` object.

```{r print-loo1}
print(loo1)
```

The table shows us a summary of Pareto $k$ diagnostic, which is used to assess
the reliability of the estimates. In addition to the proportion of leave-one-out
folds with $k$ values in different intervals, the minimum of the effective
sample sizes in that category is shown to give idea why higher $k$ values are
bad. Since we have some $k>1$, we are not able to compute an estimate for the
Monte Carlo standard error (SE) of the expected log predictive density
(`elpd_loo`) and `NA` is displayed. (Full details on the interpretation of 
the Pareto $k$ diagnostics are available in the Vehtari, Gelman, and Gabry
(2017) and Vehtari, Simpson, Gelman, Yao, and Gabry (2024) papers referenced
at the top of this vignette.)

In this case the `elpd_loo` estimate should not be considered reliable. If we
had a well-specified model we would expect the estimated effective number
of parameters (`p_loo`) to be smaller than or similar to the total number of
parameters in the model. Here `p_loo` is almost 300, which is about 70 times
the total number of parameters in the model, indicating severe model
misspecification.

### Plotting Pareto $k$ diagnostics

Using the `plot` method on our `loo1` object produces a plot of the $k$ values
(in the same order as the observations in the dataset used to fit the model) 
with horizontal lines corresponding to the same categories as in the
printed output above.

```{r plot-loo1, out.width = "70%"}
plot(loo1)
```

This plot is useful to quickly see the distribution of $k$ values, but it's
often also possible to see structure with respect to data ordering. In our case
this is mild, but there seems to be a block of data that is somewhat easier to
predict (indices around 90--150). Unfortunately even for these data points we
see some high $k$ values.

### Marginal posterior predictive checks

The `loo` package can be used in combination with the `bayesplot` package for
leave-one-out cross-validation marginal posterior predictive checks [Gabry et al
(2018)](https://arxiv.org/abs/1709.01449). LOO-PIT values are cumulative
probabilities for $y_i$ computed using the LOO marginal predictive distributions
$p(y_i|y_{-i})$. For a good model, the distribution of LOO-PIT values should be
uniform. In the following QQ-plot the LOO-PIT values for our model (y-axi) is
compared to standard uniform distribution (x-axis).

```{r ppc_loo_pit_overlay}
yrep <- posterior_predict(fit1)

ppc_loo_pit_qq(
  y = roaches$y,
  yrep = yrep,
  lw = weights(loo1$psis_object)
)
```

The excessive number of LOO-PIT values close to 0 indicates that the model is
under-dispersed compared to the data, and we should consider a model that allows
for greater dispersion.

## Try alternative model with more flexibility

Here we will try [negative binomial](https://en.wikipedia.org/wiki/Negative_binomial_distribution)
regression, which is commonly used for overdispersed count data.  
Unlike the Poisson distribution, the negative binomial distribution 
allows the conditional mean and variance of $y$ to differ.

```{r count-roaches-negbin, results="hide"}
fit2 <- update(fit1, family = neg_binomial_2)
```

```{r loo2}
loo2 <- loo(fit2, save_psis = TRUE, cores = 2)
print(loo2)
```


```{r plot-loo2}
plot(loo2, label_points = TRUE)
```

Using the `label_points` argument will label any $k$ values larger than the
diagnostic threshold with
the index of the corresponding data point. These high values are often the
result of model misspecification and frequently correspond to data points that
would be considered ``outliers'' in the data and surprising according to the
model [Gabry et al (2019)](https://arxiv.org/abs/1709.01449). Unfortunately,
while large $k$ values are a useful indicator of model misspecification, small
$k$ values are not a guarantee that a model is well-specified.

If there are a small number of problematic $k$
values then we can use a feature in __rstanarm__ that lets us refit the model
once for each of these problematic observations. Each time the model is refit,
one of the observations with a high $k$ value is omitted and the LOO
calculations are performed exactly for that observation. The results are then
recombined with the approximate LOO calculations already carried out for the
observations without problematic $k$ values:

```{r reloo}
if (any(pareto_k_values(loo2) > 0.7)) {
  loo2 <- loo(fit2, save_psis = TRUE, k_threshold = 0.7)
}
print(loo2)
```

In the print output we can see that the Monte Carlo SE is small compared to
the other uncertainties. 

On the other hand, `p_loo` is about 7 and still a bit higher than the total
number of parameters in the model. This indicates that there is almost certainly
still some degree of model misspecification, but this is much better than the
`p_loo` estimate for the Poisson model.

For further model checking we again examine the LOO-PIT values.
```{r ppc_loo_pit_overlay-negbin}
yrep <- posterior_predict(fit2)
ppc_loo_pit_qq(roaches$y, yrep, lw = weights(loo2$psis_object))
```

The plot for the negative binomial model looks better than the Poisson plot, but
we still see that this model is not capturing all of the essential features in
the data.


## Comparing the models on expected log predictive density

We can use the `loo_compare` function to compare our two models on 
expected log predictive density (ELPD) for new data:

```{r loo_compare}
loo_compare(loo1, loo2)
```

The difference in ELPD is much larger than several times the estimated standard
error of the difference again indicating that the negative-binomial model is
xpected to have better predictive performance than the Poisson model. However,
according to the LOO-PIT checks there is still some misspecification, and a
reasonable guess is that a hurdle or zero-inflated model would be an improvement
(we leave that for another case study).


<br>

# References

Gabry, J., Simpson, D., Vehtari, A., Betancourt, M. and Gelman, A. (2019),
Visualization in Bayesian workflow. _J. R. Stat. Soc. A_, 182: 389-402. 
\doi:10.1111/rssa.12378. ([journal version](https://rss.onlinelibrary.wiley.com/doi/full/10.1111/rssa.12378), 
[arXiv preprint](https://arxiv.org/abs/1709.01449), 
[code on GitHub](https://github.com/jgabry/bayes-vis-paper))
<a id="gabry2019"></a>

Gelman, A. and Hill, J. (2007). _Data Analysis Using Regression and
Multilevel/Hierarchical Models._ Cambridge University Press, Cambridge, UK.

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC. _Statistics and
Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4.
[online](https://link.springer.com/article/10.1007/s11222-016-9696-4), 
[arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

# extracting estimates by name is deprecated for loo objects

    Code
      loo1$elpd_loo
    Condition
      Warning:
      Accessing elpd_loo using '$' is deprecated and will be removed in a future release. Please extract the elpd_loo estimate from the 'estimates' component instead.
    Output
      [1] -83.58926

---

    Code
      loo1$se_elpd_loo
    Condition
      Warning:
      Accessing se_elpd_loo using '$' is deprecated and will be removed in a future release. Please extract the se_elpd_loo estimate from the 'estimates' component instead.
    Output
      [1] 4.283835

---

    Code
      loo1$p_loo
    Condition
      Warning:
      Accessing p_loo using '$' is deprecated and will be removed in a future release. Please extract the p_loo estimate from the 'estimates' component instead.
    Output
      [1] 3.328834

---

    Code
      loo1$se_p_loo
    Condition
      Warning:
      Accessing se_p_loo using '$' is deprecated and will be removed in a future release. Please extract the se_p_loo estimate from the 'estimates' component instead.
    Output
      [1] 1.152103

---

    Code
      loo1$looic
    Condition
      Warning:
      Accessing looic using '$' is deprecated and will be removed in a future release. Please extract the looic estimate from the 'estimates' component instead.
    Output
      [1] 167.1785

---

    Code
      loo1$se_looic
    Condition
      Warning:
      Accessing se_looic using '$' is deprecated and will be removed in a future release. Please extract the se_looic estimate from the 'estimates' component instead.
    Output
      [1] 8.567671

---

    Code
      loo1["elpd_loo"]
    Condition
      Warning:
      Accessing elpd_loo using '[' is deprecated and will be removed in a future release. Please extract the elpd_loo estimate from the 'estimates' component instead.
    Output
      $elpd_loo
      [1] -83.58926
      

---

    Code
      loo1["se_elpd_loo"]
    Condition
      Warning:
      Accessing se_elpd_loo using '[' is deprecated and will be removed in a future release. Please extract the se_elpd_loo estimate from the 'estimates' component instead.
    Output
      $se_elpd_loo
      [1] 4.283835
      

---

    Code
      loo1["p_loo"]
    Condition
      Warning:
      Accessing p_loo using '[' is deprecated and will be removed in a future release. Please extract the p_loo estimate from the 'estimates' component instead.
    Output
      $p_loo
      [1] 3.328834
      

---

    Code
      loo1["se_p_loo"]
    Condition
      Warning:
      Accessing se_p_loo using '[' is deprecated and will be removed in a future release. Please extract the se_p_loo estimate from the 'estimates' component instead.
    Output
      $se_p_loo
      [1] 1.152103
      

---

    Code
      loo1["looic"]
    Condition
      Warning:
      Accessing looic using '[' is deprecated and will be removed in a future release. Please extract the looic estimate from the 'estimates' component instead.
    Output
      $looic
      [1] 167.1785
      

---

    Code
      loo1["se_looic"]
    Condition
      Warning:
      Accessing se_looic using '[' is deprecated and will be removed in a future release. Please extract the se_looic estimate from the 'estimates' component instead.
    Output
      $se_looic
      [1] 8.567671
      

---

    Code
      loo1[["elpd_loo"]]
    Condition
      Warning:
      Accessing elpd_loo using '[[' is deprecated and will be removed in a future release. Please extract the elpd_loo estimate from the 'estimates' component instead.
    Output
      [1] -83.58926

---

    Code
      loo1[["se_elpd_loo"]]
    Condition
      Warning:
      Accessing se_elpd_loo using '[[' is deprecated and will be removed in a future release. Please extract the se_elpd_loo estimate from the 'estimates' component instead.
    Output
      [1] 4.283835

---

    Code
      loo1[["p_loo"]]
    Condition
      Warning:
      Accessing p_loo using '[[' is deprecated and will be removed in a future release. Please extract the p_loo estimate from the 'estimates' component instead.
    Output
      [1] 3.328834

---

    Code
      loo1[["se_p_loo"]]
    Condition
      Warning:
      Accessing se_p_loo using '[[' is deprecated and will be removed in a future release. Please extract the se_p_loo estimate from the 'estimates' component instead.
    Output
      [1] 1.152103

---

    Code
      loo1[["looic"]]
    Condition
      Warning:
      Accessing looic using '[[' is deprecated and will be removed in a future release. Please extract the looic estimate from the 'estimates' component instead.
    Output
      [1] 167.1785

---

    Code
      loo1[["se_looic"]]
    Condition
      Warning:
      Accessing se_looic using '[[' is deprecated and will be removed in a future release. Please extract the se_looic estimate from the 'estimates' component instead.
    Output
      [1] 8.567671

# extracting estimates by name is deprecated for waic objects

    Code
      waic1$elpd_waic
    Condition
      Warning:
      Accessing elpd_waic using '$' is deprecated and will be removed in a future release. Please extract the elpd_waic estimate from the 'estimates' component instead.
    Output
      [1] -83.53033

---

    Code
      waic1$se_elpd_waic
    Condition
      Warning:
      Accessing se_elpd_waic using '$' is deprecated and will be removed in a future release. Please extract the se_elpd_waic estimate from the 'estimates' component instead.
    Output
      [1] 4.258048

---

    Code
      waic1$p_waic
    Condition
      Warning:
      Accessing p_waic using '$' is deprecated and will be removed in a future release. Please extract the p_waic estimate from the 'estimates' component instead.
    Output
      [1] 3.269898

---

    Code
      waic1$se_p_waic
    Condition
      Warning:
      Accessing se_p_waic using '$' is deprecated and will be removed in a future release. Please extract the se_p_waic estimate from the 'estimates' component instead.
    Output
      [1] 1.124365

---

    Code
      waic1$waic
    Condition
      Warning:
      Accessing waic using '$' is deprecated and will be removed in a future release. Please extract the waic estimate from the 'estimates' component instead.
    Output
      [1] 167.0607

---

    Code
      waic1$se_waic
    Condition
      Warning:
      Accessing se_waic using '$' is deprecated and will be removed in a future release. Please extract the se_waic estimate from the 'estimates' component instead.
    Output
      [1] 8.516096

---

    Code
      waic1["elpd_waic"]
    Condition
      Warning:
      Accessing elpd_waic using '[' is deprecated and will be removed in a future release. Please extract the elpd_waic estimate from the 'estimates' component instead.
    Output
      $elpd_waic
      [1] -83.53033
      

---

    Code
      waic1["se_elpd_waic"]
    Condition
      Warning:
      Accessing se_elpd_waic using '[' is deprecated and will be removed in a future release. Please extract the se_elpd_waic estimate from the 'estimates' component instead.
    Output
      $se_elpd_waic
      [1] 4.258048
      

---

    Code
      waic1["p_waic"]
    Condition
      Warning:
      Accessing p_waic using '[' is deprecated and will be removed in a future release. Please extract the p_waic estimate from the 'estimates' component instead.
    Output
      $p_waic
      [1] 3.269898
      

---

    Code
      waic1["se_p_waic"]
    Condition
      Warning:
      Accessing se_p_waic using '[' is deprecated and will be removed in a future release. Please extract the se_p_waic estimate from the 'estimates' component instead.
    Output
      $se_p_waic
      [1] 1.124365
      

---

    Code
      waic1["waic"]
    Condition
      Warning:
      Accessing waic using '[' is deprecated and will be removed in a future release. Please extract the waic estimate from the 'estimates' component instead.
    Output
      $waic
      [1] 167.0607
      

---

    Code
      waic1["se_waic"]
    Condition
      Warning:
      Accessing se_waic using '[' is deprecated and will be removed in a future release. Please extract the se_waic estimate from the 'estimates' component instead.
    Output
      $se_waic
      [1] 8.516096
      

---

    Code
      waic1[["elpd_waic"]]
    Condition
      Warning:
      Accessing elpd_waic using '[[' is deprecated and will be removed in a future release. Please extract the elpd_waic estimate from the 'estimates' component instead.
    Output
      [1] -83.53033

---

    Code
      waic1[["se_elpd_waic"]]
    Condition
      Warning:
      Accessing se_elpd_waic using '[[' is deprecated and will be removed in a future release. Please extract the se_elpd_waic estimate from the 'estimates' component instead.
    Output
      [1] 4.258048

---

    Code
      waic1[["p_waic"]]
    Condition
      Warning:
      Accessing p_waic using '[[' is deprecated and will be removed in a future release. Please extract the p_waic estimate from the 'estimates' component instead.
    Output
      [1] 3.269898

---

    Code
      waic1[["se_p_waic"]]
    Condition
      Warning:
      Accessing se_p_waic using '[[' is deprecated and will be removed in a future release. Please extract the se_p_waic estimate from the 'estimates' component instead.
    Output
      [1] 1.124365

---

    Code
      waic1[["waic"]]
    Condition
      Warning:
      Accessing waic using '[[' is deprecated and will be removed in a future release. Please extract the waic estimate from the 'estimates' component instead.
    Output
      [1] 167.0607

---

    Code
      waic1[["se_waic"]]
    Condition
      Warning:
      Accessing se_waic using '[[' is deprecated and will be removed in a future release. Please extract the se_waic estimate from the 'estimates' component instead.
    Output
      [1] 8.516096

---
title: "Approximate leave-future-out cross-validation for Bayesian time series models"
author: "Paul Bürkner, Jonah Gabry, Aki Vehtari"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
encoding: "UTF-8"
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---

<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Approximate leave-future-out cross-validation for Bayesian time series models}
-->

```{r settings, child="children/SETTINGS-knitr.txt"}
```

```{r more-knitr-ops, include=FALSE}
knitr::opts_chunk$set(
  cache = TRUE,
  message = FALSE, 
  warning = FALSE
)
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

## Introduction

One of the most common goals of a time series analysis is to use the observed
series to inform predictions for future observations. We will refer to this task
of predicting a sequence of $M$ future observations as $M$-step-ahead prediction
($M$-SAP). Fortunately, once we have fit a model and can sample from the
posterior predictive distribution, it is straightforward to generate predictions
as far into the future as we want. It is also straightforward to evaluate the
$M$-SAP performance of a time series model by comparing the predictions to the
observed sequence of $M$ future data points once they become available.

Unfortunately, we are often in the position of having to use a model to inform
decisions _before_ we can collect the future observations required for assessing
the predictive performance. If we have many competing models we may also need to
first decide which of the models (or which combination of the models) we should
rely on for predictions. In these situations the best we can do is to use
methods for approximating the expected predictive performance of our models
using only the observations of the time series we already have.

If there were no time dependence in the data or if the focus is to assess
the non-time-dependent part of the model, we could use methods like
leave-one-out cross-validation (LOO-CV). For a data set with $N$ observations,
we refit the model $N$ times, each time leaving out one of the $N$ observations
and assessing how well the model predicts the left-out observation. LOO-CV is
very expensive computationally in most realistic settings, but the Pareto
smoothed importance sampling (PSIS, Vehtari et al, 2017, 2024) algorithm provided by
the *loo* package allows for approximating exact LOO-CV with PSIS-LOO-CV.
PSIS-LOO-CV requires only a single fit of the full model and comes with
diagnostics for assessing the validity of the approximation.

With a time series we can do something similar to LOO-CV but, except in a few
cases, it does not make sense to leave out observations one at a time because
then we are allowing information from the future to influence predictions of the
past (i.e., times $t + 1, t+2, \ldots$ should not be used to predict for time
$t$). To apply the idea of cross-validation to the $M$-SAP case, instead of
leave-*one*-out cross-validation we need some form of leave-*future*-out
cross-validation (LFO-CV). As we will demonstrate in this case study, LFO-CV
does not refer to one particular prediction task but rather to various possible
cross-validation approaches that all involve some form of prediction for new
time series data. Like exact LOO-CV, exact LFO-CV requires refitting the model
many times to different subsets of the data, which is computationally very
costly for most
nontrivial examples, in particular for Bayesian analyses where refitting the
model means estimating a new posterior distribution rather than a point
estimate.

Although PSIS-LOO-CV provides an efficient approximation to exact LOO-CV, until
now there has not been an analogous approximation to exact LFO-CV that
drastically reduces the computational burden while also providing informative
diagnostics about the quality of the approximation. In this case study we
present PSIS-LFO-CV, an algorithm that typically only requires refitting the
time-series model a small number times and will make LFO-CV tractable for many
more realistic applications than previously possible. 

More details can be found in our paper about approximate LFO-CV
(Bürkner, Gabry, & Vehtari, 2020), which is available as a preprint on arXiv (https://arxiv.org/abs/1902.06281).


## $M$-step-ahead predictions

Assume we have a time series of observations 
$y = (y_1, y_2, \ldots, y_N)$ 
and let $L$ be the _minimum_ number of observations from the series that
we will require before making predictions for future data. Depending on the
application and how informative the data is, it may not be possible to make
reasonable predictions for $y_{i+1}$ based on $(y_1, \dots, y_{i})$ until $i$ is
large enough so that we can learn enough about the time series to predict future
observations. Setting $L=10$, for example, means that we will only assess
predictive performance starting with observation $y_{11}$, so that we
always have at least 10 previous observations to condition on.

In order to assess $M$-SAP performance we would like to compute the 
predictive densities

$$
p(y_{i+1:M} \,|\, y_{1:i}) = 
  p(y_{i+1}, \ldots, y_{i + M} \,|\, y_{1},...,y_{i}) 
$$ 

for each $i \in \{L, \ldots, N - M\}$. The quantities 
$p(y_{i+1:M} \,|\, y_{1:i})$ can be computed with the help of the posterior
distribution $p(\theta \,|\, y_{1:i})$ of the parameters $\theta$ conditional on
only the first $i$ observations of the time-series:

$$
p(y_{i+1:M} \,| \, y_{1:i}) = 
  \int p(y_{i+1:M} \,| \, y_{1:i}, \theta) \, p(\theta\,|\,y_{1:i}) \,d\theta. 
$$

Having obtained $S$ draws $(\theta_{1:i}^{(1)}, \ldots, \theta_{1:i}^{(S)})$ 
from the posterior distribution $p(\theta\,|\,y_{1:i})$, we can estimate 
$p(y_{i+1:M} | y_{1:i})$ as

$$
p(y_{i+1:M} \,|\, y_{1:i}) \approx \frac{1}{S}\sum_{s=1}^S p(y_{i+1:M} \,|\, y_{1:i}, \theta_{1:i}^{(s)}).
$$

## Approximate $M$-SAP using importance-sampling {#approximate_MSAP}

Unfortunately, the math above makes use of the posterior distributions from many
different fits of the model to different subsets of the data. That is, to obtain
the predictive density $p(y_{i+1:M} \,|\, y_{1:i})$ requires fitting a model to
only the first $i$ data points, and we will need to do this for every value of
$i$ under consideration (all $i \in \{L, \ldots, N - M\}$).

To reduce the number of models that need to be fit for the purpose of obtaining
each of the densities $p(y_{i+1:M} \,|\, y_{1:i})$, we propose the following
algorithm. First, we refit the model using the first $L$ observations
of the time series and then perform a single exact $M$-step-ahead prediction 
step for $p(y_{L+1:M} \,|\, y_{1:L})$.
Recall that $L$ is the minimum number of observations we have deemed
acceptable for making predictions (setting $L=0$ means the first data point will
be predicted only based on the prior). We define $i^\star = L$ as the current
point of refit. Next, starting with $i = i^\star + 1$, we
approximate each $p(y_{i+1:M} \,|\, y_{1:i})$ via

$$
 p(y_{i+1:M} \,|\, y_{1:i}) \approx
   \frac{ \sum_{s=1}^S w_i^{(s)}\, p(y_{i+1:M} \,|\, y_{1:i}, \theta^{(s)})}
        { \sum_{s=1}^S w_i^{(s)}},
$$

where $\theta^{(s)} = \theta^{(s)}_{1:i^\star}$ are draws from the
posterior distribution based on the first $i^\star$ observations
and $w_i^{(s)}$ are the PSIS weights obtained in two steps.
First, we compute the raw importance ratios

$$
r_i^{(s)} =
\frac{f_{1:i}(\theta^{(s)})}{f_{1:i^\star}(\theta^{(s)})} 
\propto \prod_{j \in (i^\star + 1):i} p(y_j \,|\, y_{1:(j-1)}, \theta^{(s)}),
$$

and then stabilize them using PSIS. The
function $f_{1:i}$ denotes the posterior distribution based on the first $i$
observations, that is, $f_{1:i} = p(\theta \,|\, y_{1:i})$, with $f_{1:i^\star}$
defined analogously. The index set $(i^\star + 1):i$ indicates all observations
which are part of the data for the model $f_{1:i}$ whose predictive performance
we are trying to approximate but not for the actually fitted model
$f_{1:i^\star}$. The proportional statement arises from the fact that
we ignore the normalizing constants $p(y_{1:i})$ and $p(y_{1:i^\star})$
of the compared posteriors, which leads to a self-normalized variant of
PSIS (see Vehtari et al, 2017).

Continuing with the next observation, we gradually increase $i$ by $1$ (we move
forward in time) and repeat the process. At some observation $i$, the
variability of the importance ratios $r_i^{(s)}$ will become too large and
importance sampling will fail. We will refer to this particular value of $i$ as
$i^\star_1$. To identify the value of $i^\star_1$, we check for which value of
$i$ does the estimated shape parameter $k$ of the generalized Pareto
distribution first cross a certain threshold $\tau$ (Vehtari et al, 2024). Only
then do we refit the model using the observations up to $i^\star_1$ and restart
the process from there by setting $\theta^{(s)} = \theta^{(s)}_{1:i^\star_1}$
and $i^\star = i^\star_1$ until the next refit.

In some cases we may only need to refit once and in other cases we will find a
value $i^\star_2$ that requires a second refitting, maybe an $i^\star_3$ that
requires a third refitting, and so on. We refit as many times as is required
(only when $k > \tau$) until we arrive at observation $i = N - M$. 
For LOO, assuming posterior sample size is 4000 or larger, we recommend to use a threshold of $\tau = 0.7$ (Vehtari et al, 2017, 2024)
and it turns out this is a reasonable threshold for LFO as well (Bürkner et al. 2020).

## Autoregressive models

Autoregressive (AR) models are some of the most commonly used time-series models. 
An AR(p) model ---an autoregressive model of order $p$--- can be defined as

$$
y_i = \eta_i + \sum_{k = 1}^p \varphi_k y_{i - k} + \varepsilon_i,
$$

where $\eta_i$ is the linear predictor for the $i$th observation, $\phi_k$ are
the autoregressive parameters and $\varepsilon_i$ are pairwise independent
errors, which are usually assumed to be normally distributed with equal variance
$\sigma^2$. The model implies a recursive formula that allows for computing the
right-hand side of the above equation for observation $i$ based on the values of
the equations for previous observations.

## Case Study: Annual measurements of the level of Lake Huron

To illustrate the application of PSIS-LFO-CV for estimating expected $M$-SAP
performance, we will fit a model for 98 annual measurements of the water level
(in feet) of [Lake Huron](https://en.wikipedia.org/wiki/Lake_Huron) from the
years 1875--1972. This data set is found in the **datasets** R package, which is
installed automatically with **R**. 

In addition to the **loo** package, for this analysis we will use the **brms**
interface to Stan to generate a Stan program and fit the model, and also the
**bayesplot** and **ggplot2** packages for plotting.

```{r pkgs, cache=FALSE}
library("brms")
library("loo")
library("bayesplot")
library("ggplot2")
color_scheme_set("brightblue")
theme_set(theme_default())

CHAINS <- 4
SEED <- 5838296
set.seed(SEED)
```

Before fitting a model, we will first put the data into a data frame and then
look at the time series.

```{r hurondata}
N <- length(LakeHuron)
df <- data.frame(
  y = as.numeric(LakeHuron),
  year = as.numeric(time(LakeHuron)),
  time = 1:N
)

ggplot(df, aes(x = year, y = y)) + 
  geom_point(size = 1) +
  labs(
    y = "Water Level (ft)", 
    x = "Year",
    title = "Water Level in Lake Huron (1875-1972)"
  ) 
```

The above plot shows rather strong autocorrelation of the time-series as well as
some trend towards lower levels for later points in time. 

We can specify an AR(4) model for these data using the **brms** package as
follows:

```{r fit, results = "hide"}
fit <- brm(
  y ~ ar(time, p = 4), 
  data = df, 
  prior = prior(normal(0, 0.5), class = "ar"),
  control = list(adapt_delta = 0.99), 
  seed = SEED, 
  chains = CHAINS
)
```

The model implied predictions along with the observed values can be plotted,
which reveals a rather good fit to the data.

```{r plotpreds, cache = FALSE}
preds <- posterior_predict(fit)
preds <- cbind(
  Estimate = colMeans(preds), 
  Q5 = apply(preds, 2, quantile, probs = 0.05),
  Q95 = apply(preds, 2, quantile, probs = 0.95)
)

ggplot(cbind(df, preds), aes(x = year, y = Estimate)) +
  geom_smooth(aes(ymin = Q5, ymax = Q95), stat = "identity", linewidth = 0.5) +
  geom_point(aes(y = y)) + 
  labs(
    y = "Water Level (ft)", 
    x = "Year",
    title = "Water Level in Lake Huron (1875-1972)",
    subtitle = "Mean (blue) and 90% predictive intervals (gray) vs. observed data (black)"
  ) 
```

To allow for reasonable predictions of future values, we will require at least
$L = 20$ historical observations (20 years) to make predictions.

```{r setL}
L <- 20
```

We first perform approximate leave-one-out cross-validation (LOO-CV) for the
purpose of later comparison with exact and approximate LFO-CV for the 1-SAP
case.

```{r loo1sap, cache = FALSE}
loo_cv <- loo(log_lik(fit)[, (L + 1):N])
print(loo_cv)
```

## 1-step-ahead predictions leaving out all future values

The most basic version of $M$-SAP is 1-SAP, in which we predict only one step
ahead. In this case, $y_{i+1:M}$ simplifies to $y_{i}$ and the LFO-CV algorithm
becomes considerably simpler than for larger values of $M$.

### Exact 1-step-ahead predictions

Before we compute approximate LFO-CV using PSIS we will first compute 
exact LFO-CV for the 1-SAP case so we can use it as a benchmark later. 
The initial step for the exact computation is to calculate the log-predictive 
densities by refitting the model many times:

```{r exact_loglik, results="hide"}
loglik_exact <- matrix(nrow = ndraws(fit), ncol = N)
for (i in L:(N - 1)) {
  past <- 1:i
  oos <- i + 1
  df_past <- df[past, , drop = FALSE]
  df_oos <- df[c(past, oos), , drop = FALSE]
  fit_i <- update(fit, newdata = df_past, recompile = FALSE)
  loglik_exact[, i + 1] <- log_lik(fit_i, newdata = df_oos, oos = oos)[, oos]
}
```

Then we compute the exact expected log predictive density (ELPD):

```{r helpers}
# some helper functions we'll use throughout

# more stable than log(sum(exp(x))) 
log_sum_exp <- function(x) {
  max_x <- max(x)  
  max_x + log(sum(exp(x - max_x)))
}

# more stable than log(mean(exp(x)))
log_mean_exp <- function(x) {
  log_sum_exp(x) - log(length(x))
}

# compute log of raw importance ratios
# sums over observations *not* over posterior samples
sum_log_ratios <- function(loglik, ids = NULL) {
  if (!is.null(ids)) loglik <- loglik[, ids, drop = FALSE]
  rowSums(loglik)
}

# for printing comparisons later
rbind_print <- function(...) {
  round(rbind(...), digits = 2)
}
```


```{r exact1sap, cache = FALSE}
exact_elpds_1sap <- apply(loglik_exact, 2, log_mean_exp)
exact_elpd_1sap <- c(ELPD = sum(exact_elpds_1sap[-(1:L)]))

rbind_print(
  "LOO" = loo_cv$estimates["elpd_loo", "Estimate"],
  "LFO" = exact_elpd_1sap
)
```

We see that the ELPD from LFO-CV for 1-step-ahead predictions is lower than the
ELPD estimate from LOO-CV, which should be expected since LOO-CV is making use
of more of the time series. That is, since the LFO-CV approach only uses
observations from before the left-out data point but LOO-CV uses _all_ data
points other than the left-out observation, we should expect to see the larger
ELPD from LOO-CV.

### Approximate 1-step-ahead predictions

We compute approximate 1-SAP with refit at observations where 
the Pareto $k$ estimate exceeds the threshold of $0.7$.

```{r setkthresh}
k_thres <- 0.7
```

The code becomes a little bit more involved as compared to the exact LFO-CV.
Note that we can compute exact 1-SAP at the refitting points, which comes with
no additional computational costs since we had to refit the model anyway.

```{r refit_loglik, results="hide"}
approx_elpds_1sap <- rep(NA, N)

# initialize the process for i = L
past <- 1:L
oos <- L + 1
df_past <- df[past, , drop = FALSE]
df_oos <- df[c(past, oos), , drop = FALSE]
fit_past <- update(fit, newdata = df_past, recompile = FALSE)
loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
approx_elpds_1sap[L + 1] <- log_mean_exp(loglik[, oos])

# iterate over i > L
i_refit <- L
refits <- L
ks <- NULL
for (i in (L + 1):(N - 1)) {
  past <- 1:i
  oos <- i + 1
  df_past <- df[past, , drop = FALSE]
  df_oos <- df[c(past, oos), , drop = FALSE]
  loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
  
  logratio <- sum_log_ratios(loglik, (i_refit + 1):i)
  psis_obj <- suppressWarnings(psis(logratio))
  k <- pareto_k_values(psis_obj)
  ks <- c(ks, k)
  if (k > k_thres) {
    # refit the model based on the first i observations
    i_refit <- i
    refits <- c(refits, i)
    fit_past <- update(fit_past, newdata = df_past, recompile = FALSE)
    loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
    approx_elpds_1sap[i + 1] <- log_mean_exp(loglik[, oos])
  } else {
    lw <- weights(psis_obj, normalize = TRUE)[, 1]
    approx_elpds_1sap[i + 1] <- log_sum_exp(lw + loglik[, oos])
  }
} 
```

We see that the final Pareto-$k$-estimates are mostly well below the threshold
and that we only needed to refit the model a few times:

```{r plot_ks}
plot_ks <- function(ks, ids, thres = 0.6) {
  dat_ks <- data.frame(ks = ks, ids = ids)
  ggplot(dat_ks, aes(x = ids, y = ks)) + 
    geom_point(aes(color = ks > thres), shape = 3, show.legend = FALSE) + 
    geom_hline(yintercept = thres, linetype = 2, color = "red2") + 
    scale_color_manual(values = c("cornflowerblue", "darkblue")) + 
    labs(x = "Data point", y = "Pareto k") + 
    ylim(-0.5, 1.5)
}
```

```{r refitsummary1sap, cache=FALSE}
cat("Using threshold ", k_thres, 
    ", model was refit ", length(refits), 
    " times, at observations", refits)

plot_ks(ks, (L + 1):(N - 1))
```

The approximate 1-SAP ELPD is remarkably similar to the exact 1-SAP ELPD
computed above, which indicates our algorithm to compute approximate 1-SAP
worked well for the present data and model.

```{r lfosummary1sap, cache = FALSE}
approx_elpd_1sap <- sum(approx_elpds_1sap, na.rm = TRUE)
rbind_print(
  "approx LFO" = approx_elpd_1sap,
  "exact LFO" = exact_elpd_1sap
)
```

Plotting exact against approximate predictions, we see that no approximation
value deviates far from its exact counterpart, providing further evidence for
the good quality of our approximation.

```{r plot1sap, cache = FALSE}
dat_elpd <- data.frame(
  approx_elpd = approx_elpds_1sap,
  exact_elpd = exact_elpds_1sap
)

ggplot(dat_elpd, aes(x = approx_elpd, y = exact_elpd)) +
  geom_abline(color = "gray30") +
  geom_point(size = 2) +
  labs(x = "Approximate ELPDs", y = "Exact ELPDs")
```

We can also look at the maximum difference and average difference between
the approximate and exact ELPD calculations, which also indicate a ver close 
approximation:

```{r diffs1sap, cache=FALSE}
max_diff <- with(dat_elpd, max(abs(approx_elpd - exact_elpd), na.rm = TRUE))
mean_diff <- with(dat_elpd, mean(abs(approx_elpd - exact_elpd), na.rm = TRUE))

rbind_print(
  "Max diff" = round(max_diff, 2), 
  "Mean diff" =  round(mean_diff, 3)
)
```

## $M$-step-ahead predictions leaving out all future values

To illustrate the application of $M$-SAP for $M > 1$, we next compute exact and
approximate LFO-CV for the 4-SAP case.

### Exact $M$-step-ahead predictions

The necessary steps are the same as for 1-SAP with the exception that the
log-density values of interest are now the sums of the log predictive densities
of four consecutive observations. Further, the stability of the PSIS
approximation actually stays the same for all $M$ as it only depends on the
number of observations we leave out, not on the number of observations we
predict.

```{r exact_loglikm, results="hide"}
M <- 4
loglikm <- matrix(nrow = ndraws(fit), ncol = N)
for (i in L:(N - M)) {
  past <- 1:i
  oos <- (i + 1):(i + M)
  df_past <- df[past, , drop = FALSE]
  df_oos <- df[c(past, oos), , drop = FALSE]
  fit_past <- update(fit, newdata = df_past, recompile = FALSE)
  loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
  loglikm[, i + 1] <- rowSums(loglik[, oos])
}
```

```{r exact4sap, cache = FALSE}
exact_elpds_4sap <- apply(loglikm, 2, log_mean_exp)
(exact_elpd_4sap <- c(ELPD = sum(exact_elpds_4sap, na.rm = TRUE)))
```

### Approximate $M$-step-ahead predictions

Computing the approximate PSIS-LFO-CV for the 4-SAP case is a little bit more
involved than the approximate version for the 1-SAP case, although the
underlying principles remain the same.

```{r refit_loglikm, results="hide"}
approx_elpds_4sap <- rep(NA, N)

# initialize the process for i = L
past <- 1:L
oos <- (L + 1):(L + M)
df_past <- df[past, , drop = FALSE]
df_oos <- df[c(past, oos), , drop = FALSE]
fit_past <- update(fit, newdata = df_past, recompile = FALSE)
loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
loglikm <- rowSums(loglik[, oos])
approx_elpds_4sap[L + 1] <- log_mean_exp(loglikm)

# iterate over i > L
i_refit <- L
refits <- L
ks <- NULL
for (i in (L + 1):(N - M)) {
  past <- 1:i
  oos <- (i + 1):(i + M)
  df_past <- df[past, , drop = FALSE]
  df_oos <- df[c(past, oos), , drop = FALSE]
  loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
  
  logratio <- sum_log_ratios(loglik, (i_refit + 1):i)
  psis_obj <- suppressWarnings(psis(logratio))
  k <- pareto_k_values(psis_obj)
  ks <- c(ks, k)
  if (k > k_thres) {
    # refit the model based on the first i observations
    i_refit <- i
    refits <- c(refits, i)
    fit_past <- update(fit_past, newdata = df_past, recompile = FALSE)
    loglik <- log_lik(fit_past, newdata = df_oos, oos = oos)
    loglikm <- rowSums(loglik[, oos])
    approx_elpds_4sap[i + 1] <- log_mean_exp(loglikm)
  } else {
    lw <- weights(psis_obj, normalize = TRUE)[, 1]
    loglikm <- rowSums(loglik[, oos])
    approx_elpds_4sap[i + 1] <- log_sum_exp(lw + loglikm)
  }
} 
```

Again, we see that the final Pareto-$k$-estimates are mostly well below the
threshold and that we only needed to refit the model a few times: 

```{r refitsummary4sap, cache = FALSE}
cat("Using threshold ", k_thres, 
    ", model was refit ", length(refits), 
    " times, at observations", refits)

plot_ks(ks, (L + 1):(N - M))
```

The approximate ELPD computed for the 4-SAP case is not as close to its exact
counterpart as in the 1-SAP case. In general, the larger $M$, the larger the
variation of the approximate ELPD around the exact ELPD. It turns out that the
ELPD estimates of AR-models with $M>1$ show particular variation due to their
predictions' dependency on other predicted values. In Bürkner et al. (2020) we
provide further explanation and simulations for these cases.

```{r lfosummary4sap, cache = FALSE}
approx_elpd_4sap <- sum(approx_elpds_4sap, na.rm = TRUE)
rbind_print(
  "Approx LFO" = approx_elpd_4sap,
  "Exact LFO" = exact_elpd_4sap
)
```

Plotting exact against approximate pointwise predictions confirms that, for a 
few specific data points, the approximate predictions underestimate the exact
predictions.

```{r plot4sap, cache = FALSE}
dat_elpd_4sap <- data.frame(
  approx_elpd = approx_elpds_4sap,
  exact_elpd = exact_elpds_4sap
)

ggplot(dat_elpd_4sap, aes(x = approx_elpd, y = exact_elpd)) +
  geom_abline(color = "gray30") +
  geom_point(size = 2) +
  labs(x = "Approximate ELPDs", y = "Exact ELPDs")
```

## Conclusion

In this case study we have shown how to do carry out exact and approximate
leave-future-out cross-validation for $M$-step-ahead prediction tasks. For the
data and model used in our example, the PSIS-LFO-CV algorithm provides reasonably
stable and accurate results despite not requiring us to refit the model nearly as
many times. For more details on approximate LFO-CV, we refer to Bürkner et al. (2020).

<br />

## References

Bürkner P. C., Gabry J., & Vehtari A. (2020). Approximate leave-future-out cross-validation for time series models. *Journal of Statistical Computation and Simulation*, 90(14):2499-2523. \doi:/10.1080/00949655.2020.1783262. 
[Online](https://www.tandfonline.com/doi/full/10.1080/00949655.2020.1783262). [arXiv preprint](https://arxiv.org/abs/1902.06281). 

Vehtari A., Gelman A., & Gabry J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. *Statistics and Computing*, 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. [Online](https://link.springer.com/article/10.1007/s11222-016-9696-4). [arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

<br />

## Appendix

### Appendix: Session information

```{r sessioninfo}
sessionInfo()
```

### Appendix: Licenses

* Code &copy; 2018, Paul Bürkner, Jonah Gabry, Aki Vehtari (licensed under BSD-3).
* Text &copy; 2018, Paul Bürkner, Jonah Gabry, Aki Vehtari (licensed under CC-BY-NC 4.0).

# psislw handles special cases, throws appropriate errors/warnings

    Code
      psis <- psislw(x[, 1], wcp = 0.01)
    Condition
      Warning in `psislw()`:
      'psislw' is deprecated.
      Use 'psis' instead.
      See help("Deprecated")
      Warning:
      All tail values are the same. Weights are truncated but not smoothed.
      Warning:
      Some Pareto k diagnostic values are too high. See help('pareto-k-diagnostic') for details.

# loo_compare returns expected results (2 models)

    WAoAAAACAAQEAgACAwAAAAMTAAAADAAAABAAAAACAAQACQAAAAZtb2RlbDEABAAJAAAABm1v
    ZGVsMgAAAA4AAAACAAAAAAAAAAAAAAAAAAAAAAAAAA4AAAACAAAAAAAAAAAAAAAAAAAAAAAA
    AA4AAAACf/AAAAAAB6J/8AAAAAAHogAAABAAAAACAAQACQAAAAAABAAJAAAAAAAAABAAAAAC
    AAQACQAAAAAABAAJAAAAAAAAAA4AAAACwFTh8N3JQljAVOHw3clCWAAAAA4AAAACQBEIPbMR
    cF9AEQg9sxFwXwAAAA4AAAACQAoowGHVuVNACijAYdW5UwAAAA4AAAACP/H9Zexy814/8f1l
    7HLzXgAAAA4AAAACQGTh8N3JQlhAZOHw3clCWAAAAA4AAAACQCEIPbMRcF9AIQg9sxFwXwAA
    BAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAAMAAQACQAAAAVtb2RlbAAEAAkAAAAJZWxwZF9k
    aWZmAAQACQAAAAdzZV9kaWZmAAQACQAAAAdwX3dvcnNlAAQACQAAAAlkaWFnX2RpZmYABAAJ
    AAAACWRpYWdfZWxwZAAEAAkAAAAJZWxwZF93YWljAAQACQAAAAxzZV9lbHBkX3dhaWMABAAJ
    AAAABnBfd2FpYwAEAAkAAAAJc2VfcF93YWljAAQACQAAAAR3YWljAAQACQAAAAdzZV93YWlj
    AAAEAgAAAAEABAAJAAAABWNsYXNzAAAAEAAAAAIABAAJAAAAC2NvbXBhcmUubG9vAAQACQAA
    AApkYXRhLmZyYW1lAAAEAgAAAAEABAAJAAAACXJvdy5uYW1lcwAAAA0AAAACgAAAAP////4A
    AAD+

---

    Code
      print(comp1)
    Output
        model elpd_diff se_diff p_worse diag_diff diag_elpd
       model1       0.0     0.0      NA                    
       model2       0.0     0.0      NA                    

---

    WAoAAAACAAQEAgACAwAAAAMTAAAADAAAABAAAAACAAQACQAAAAZtb2RlbDEABAAJAAAABm1v
    ZGVsMgAAAA4AAAACAAAAAAAAAADAEDpTX5xF7gAAAA4AAAACAAAAAAAAAAA/tmpHtC8TAQAA
    AA4AAAACf/AAAAAAB6I/8AAAAAAAAAAAABAAAAACAAQACQAAAAAABAAJAAAAB04gPCAxMDAA
    AAAQAAAAAgAEAAkAAAAAAAQACQAAAAAAAAAOAAAAAsBU4fDdyUJYwFXllhPDBrkAAAAOAAAA
    AkARCD2zEXBfQBEalRIN2T8AAAAOAAAAAkAKKMBh1blTQCZnlesA0IoAAAAOAAAAAj/x/WXs
    cvNeP/GbYJxtZ8cAAAAOAAAAAkBk4fDdyUJYQGXllhPDBrkAAAAOAAAAAkAhCD2zEXBfQCEa
    lRIN2T8AAAQCAAAAAQAEAAkAAAAFbmFtZXMAAAAQAAAADAAEAAkAAAAFbW9kZWwABAAJAAAA
    CWVscGRfZGlmZgAEAAkAAAAHc2VfZGlmZgAEAAkAAAAHcF93b3JzZQAEAAkAAAAJZGlhZ19k
    aWZmAAQACQAAAAlkaWFnX2VscGQABAAJAAAACWVscGRfd2FpYwAEAAkAAAAMc2VfZWxwZF93
    YWljAAQACQAAAAZwX3dhaWMABAAJAAAACXNlX3Bfd2FpYwAEAAkAAAAEd2FpYwAEAAkAAAAH
    c2Vfd2FpYwAABAIAAAABAAQACQAAAAVjbGFzcwAAABAAAAACAAQACQAAAAtjb21wYXJlLmxv
    bwAEAAkAAAAKZGF0YS5mcmFtZQAABAIAAAABAAQACQAAAAlyb3cubmFtZXMAAAANAAAAAoAA
    AAD////+AAAA/g==

---

    Code
      print(comp2)
    Output
        model elpd_diff se_diff p_worse diag_diff diag_elpd
       model1       0.0     0.0      NA                    
       model2      -4.1     0.1    1.00   N < 100          
    Message
      
      Diagnostic flags present.
      See ?`loo-glossary` (sections `diag_diff` and `diag_elpd`)
      or https://mc-stan.org/loo/reference/loo-glossary.html.

---

    Code
      print(comp2, p_worse = FALSE)
    Output
        model elpd_diff se_diff
       model1       0.0     0.0
       model2      -4.1     0.1

---

    Code
      print(comp2, simplify = FALSE)
    Output
        model elpd_diff se_diff p_worse diag_diff diag_elpd elpd_waic se_elpd_waic
       model1       0.0     0.0      NA                         -83.5          4.3
       model2      -4.1     0.1    1.00   N < 100               -87.6          4.3
       p_waic se_p_waic  waic se_waic
          3.3       1.1 167.1     8.5
         11.2       1.1 175.2     8.6
    Message
      
      Diagnostic flags present.
      See ?`loo-glossary` (sections `diag_diff` and `diag_elpd`)
      or https://mc-stan.org/loo/reference/loo-glossary.html.

---

    Code
      print(comp2, simplify = FALSE, p_worse = FALSE)
    Output
        model elpd_diff se_diff elpd_waic se_elpd_waic p_waic se_p_waic  waic se_waic
       model1       0.0     0.0     -83.5          4.3    3.3       1.1 167.1     8.5
       model2      -4.1     0.1     -87.6          4.3   11.2       1.1 175.2     8.6

# loo_compare returns expected result (3 models)

    WAoAAAACAAQEAgACAwAAAAMTAAAADAAAABAAAAADAAQACQAAAAZtb2RlbDEABAAJAAAABm1v
    ZGVsMgAEAAkAAAAGbW9kZWwzAAAADgAAAAMAAAAAAAAAAMAQOlNfnEXuwDANypG2BBgAAAAO
    AAAAAwAAAAAAAAAAP7ZqR7QvEwE/y6/t4TTtXgAAAA4AAAADf/AAAAAAB6I/8AAAAAAAAD/w
    AAAAAAAAAAAAEAAAAAMABAAJAAAAAAAEAAkAAAAHTiA8IDEwMAAEAAkAAAAHTiA8IDEwMAAA
    ABAAAAADAAQACQAAAAAABAAJAAAAAAAEAAkAAAAAAAAADgAAAAPAVOHw3clCWMBV5ZYTwwa5
    wFjlY4I2w2IAAAAOAAAAA0ARCD2zEXBfQBEalRIN2T9AEPIF3GigEwAAAA4AAAADQAoowGHV
    uVNAJmeV6wDQikBByNhSGt0KAAAADgAAAAM/8f1l7HLzXj/xm2CcbWfHP/GA0JJnyV8AAAAO
    AAAAA0Bk4fDdyUJYQGXllhPDBrlAaOVjgjbDYgAAAA4AAAADQCEIPbMRcF9AIRqVEg3ZP0Ag
    8gXcaKATAAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAwABAAJAAAABW1vZGVsAAQACQAA
    AAllbHBkX2RpZmYABAAJAAAAB3NlX2RpZmYABAAJAAAAB3Bfd29yc2UABAAJAAAACWRpYWdf
    ZGlmZgAEAAkAAAAJZGlhZ19lbHBkAAQACQAAAAllbHBkX3dhaWMABAAJAAAADHNlX2VscGRf
    d2FpYwAEAAkAAAAGcF93YWljAAQACQAAAAlzZV9wX3dhaWMABAAJAAAABHdhaWMABAAJAAAA
    B3NlX3dhaWMAAAQCAAAAAQAEAAkAAAAFY2xhc3MAAAAQAAAAAgAEAAkAAAALY29tcGFyZS5s
    b28ABAAJAAAACmRhdGEuZnJhbWUAAAQCAAAAAQAEAAkAAAAJcm93Lm5hbWVzAAAADQAAAAKA
    AAAA/////QAAAP4=

---

    Code
      print(comp1)
    Output
        model elpd_diff se_diff p_worse diag_diff diag_elpd
       model1       0.0     0.0      NA                    
       model2      -4.1     0.1    1.00   N < 100          
       model3     -16.1     0.2    1.00   N < 100          
    Message
      
      Diagnostic flags present.
      See ?`loo-glossary` (sections `diag_diff` and `diag_elpd`)
      or https://mc-stan.org/loo/reference/loo-glossary.html.

# loo_compare with simplify=FALSE returns expected result

    Code
      print(comp, simplify = FALSE)
    Output
        model elpd_diff se_diff p_worse diag_diff diag_elpd elpd_loo se_elpd_loo
       model3       0.0     0.0      NA                        -19.6         4.3
       model2     -32.0     0.0    1.00   N < 100              -51.6         4.3
       model1     -64.0     0.0    1.00   N < 100              -83.6         4.3
       p_loo se_p_loo looic se_looic
         3.3      1.2  39.2      8.6
         3.3      1.2 103.2      8.6
         3.3      1.2 167.2      8.6
    Message
      
      Diagnostic flags present.
      See ?`loo-glossary` (sections `diag_diff` and `diag_elpd`)
      or https://mc-stan.org/loo/reference/loo-glossary.html.

# compare returns expected result (3 models)

    WAoAAAACAAQFAAACAwAAAAMOAAAAGAAAAAAAAAAAwBA6U1+cRe7AMA3KkbYEGAAAAAAAAAAA
    P7ZqR7QvEwE/y6/t4TTtXsBU4fDdyUJYwFXllhPDBrnAWOVjgjbDYkARCD2zEXBfQBEalRIN
    2T9AEPIF3GigE0AKKMBh1blTQCZnlesA0IpAQcjYUhrdCj/x/WXscvNeP/GbYJxtZ8c/8YDQ
    kmfJX0Bk4fDdyUJYQGXllhPDBrlAaOVjgjbDYkAhCD2zEXBfQCEalRIN2T9AIPIF3GigEwAA
    BAIAAAABAAQACQAAAANkaW0AAAANAAAAAgAAAAMAAAAIAAAEAgAAAAEABAAJAAAACGRpbW5h
    bWVzAAAAEwAAAAIAAAAQAAAAAwAEAAkAAAACdzEABAAJAAAAAncyAAQACQAAAAJ3MwAAABAA
    AAAIAAQACQAAAAllbHBkX2RpZmYABAAJAAAAB3NlX2RpZmYABAAJAAAACWVscGRfd2FpYwAE
    AAkAAAAMc2VfZWxwZF93YWljAAQACQAAAAZwX3dhaWMABAAJAAAACXNlX3Bfd2FpYwAEAAkA
    AAAEd2FpYwAEAAkAAAAHc2Vfd2FpYwAABAIAAAABAAQACQAAAAVjbGFzcwAAABAAAAAEAAQA
    CQAAAAtjb21wYXJlLmxvbwAEAAkAAAAGbWF0cml4AAQACQAAAAVhcnJheQAEAAkAAAAPb2xk
    X2NvbXBhcmUubG9vAAAA/g==

# gpdfit returns correct result

    WAoAAAACAAQFAAACAwAAAAIOAAAAAj+cD4qKVTgaP/BK8xJCK3sAAAQCAAAAAQAEAAkAAAAF
    bmFtZXMAAAAQAAAAAgAEAAkAAAABawAEAAkAAAAFc2lnbWEAAAD+

---

    WAoAAAACAAQFAAACAwAAAAIOAAAAAj+yA4g2tkbvP/BK8xJCK3sAAAQCAAAAAQAEAAkAAAAF
    bmFtZXMAAAAQAAAAAgAEAAkAAAABawAEAAkAAAAFc2lnbWEAAAD+

---

    WAoAAAACAAQFAAACAwAAAAIOAAAAAj+yA5BUlFrHP/BK8oexSVIAAAQCAAAAAQAEAAkAAAAF
    bmFtZXMAAAAQAAAAAgAEAAkAAAABawAEAAkAAAAFc2lnbWEAAAD+

---
title: "Bayesian Stacking and Pseudo-BMA weights using the loo package"
author: "Aki Vehtari and Jonah Gabry"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Bayesian Stacking and Pseudo-BMA weights}
-->
```{r, child="children/SETTINGS-knitr.txt"}
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

This vignette demonstrates the new functionality in __loo__ v2.0.0 for 
Bayesian stacking and Pseudo-BMA weighting. In this vignette we can't
provide all of the necessary background on this topic, so we encourage
readers to refer to the paper

* Yao, Y., Vehtari, A., Simpson, D., and Gelman, A.  (2018). Using stacking to average Bayesian predictive distributions. In Bayesian Analysis, \doi:10.1214/17-BA1091. [Online](https://projecteuclid.org/euclid.ba/1516093227)

which provides important details on the methods demonstrated in this vignette.
Here we just quote from the abstract of the paper: 

> **Abstract**: Bayesian model averaging is flawed in the $\mathcal{M}$-open setting in which the true data-generating process is not one of the candidate models being fit. We take the idea of stacking from the point estimation literature and generalize to the combination of predictive distributions. We extend the utility function to any proper scoring rule and use Pareto smoothed importance sampling to efficiently compute the required leave-one-out posterior distributions. We compare stacking of predictive distributions to several alternatives: stacking of means, Bayesian model averaging (BMA), Pseudo-BMA, and a variant of Pseudo-BMA that is stabilized using the Bayesian bootstrap. Based on simulations and real-data applications, we recommend stacking of predictive distributions, with bootstrapped-Pseudo-BMA as an approximate alternative when computation cost is an issue.

Ideally, we would avoid the Bayesian model combination problem by
extending the model to include the separate models as special cases,
and preferably as a continuous expansion of the model space.  For
example, instead of model averaging over different covariate
combinations, all potentially relevant covariates should be included
in a predictive model (for causal analysis more  care is needed) 
and a prior assumption that only some of the covariates are
relevant can be presented with regularized horseshoe prior (Piironen
and Vehtari, 2017a). For variable selection we recommend projective
predictive variable selection (Piironen and Vehtari, 2017a;
[__projpred__ package](https://cran.r-project.org/package=projpred)).

To demonstrate how to use __loo__ package to compute Bayesian stacking and
Pseudo-BMA weights, we repeat two simple model averaging examples from Chapters
6 and 10 of _Statistical Rethinking_ by Richard McElreath. In _Statistical
Rethinking_ WAIC is used to form weights which are similar to classical "Akaike
weights". Pseudo-BMA weighting using PSIS-LOO for computation is close to these
WAIC weights, but named after the Pseudo Bayes Factor by Geisser and Eddy
(1979). As discussed below, in general we prefer using stacking rather than WAIC
weights or the similar pseudo-BMA weights.


# Setup

In addition to the __loo__ package we will also load the __rstanarm__ package
for fitting the models.

```{r setup, message=FALSE}
library(rstanarm)
library(loo)
```

# Example: Primate milk 

In _Statistical Rethinking_, McElreath describes the data for the primate milk
example as follows:

> A popular hypothesis has it that primates with larger brains produce more energetic milk, so that brains can grow quickly. ... The question here is to what extent energy content of milk, measured here by kilocalories, is related to the percent of the brain mass that is neocortex. ... We'll end up needing female body mass as well, to see the masking that hides the relationships among the variables.

```{r data}
data(milk)
d <- milk[complete.cases(milk),]
d$neocortex <- d$neocortex.perc /100
str(d)
```

We repeat the analysis in Chapter 6 of _Statistical Rethinking_ using the
following four models (here we use the default weakly informative priors in __rstanarm__, while flat priors were used in _Statistical Rethinking_).

```{r fits, results="hide"}
fit1 <- stan_glm(kcal.per.g ~ 1, data = d, seed = 2030)
fit2 <- update(fit1, formula = kcal.per.g ~ neocortex)
fit3 <- update(fit1, formula = kcal.per.g ~ log(mass))
fit4 <- update(fit1, formula = kcal.per.g ~ neocortex + log(mass))
```

McElreath uses WAIC for model comparison and averaging, so we'll start by also
computing WAIC for these models so we can compare the results to the 
other options presented later in the vignette. The __loo__ package
provides `waic` methods for log-likelihood arrays, matrices and functions.
Since we fit our model with rstanarm we can use the `waic` method provided
by the __rstanarm__ package (a wrapper around `waic` from the __loo__ package), 
which allows us to just pass in our fitted model objects instead of first 
extracting the log-likelihood values.

```{r waic}
waic1 <- waic(fit1)
waic2 <- waic(fit2)
waic3 <- waic(fit3)
waic4 <- waic(fit4)
waics <- c(
  waic1$estimates["elpd_waic", 1],
  waic2$estimates["elpd_waic", 1],
  waic3$estimates["elpd_waic", 1],
  waic4$estimates["elpd_waic", 1]
)
```

We get some warnings when computing WAIC for models 3 and 4, indicating that we
shouldn't trust the WAIC weights we will compute later. Following the
recommendation in the warning, we next use the `loo` methods to compute PSIS-LOO
instead. The __loo__ package provides `loo` methods for log-likelihood arrays,
matrices, and functions, but since we fit our model with __rstanarm__ we can
just pass the fitted model objects directly and __rstanarm__ will extract the
needed values to pass to the __loo__ package. (Like __rstanarm__, some other R
packages for fitting Stan models, e.g. __brms__, also provide similar methods
for interfacing with the __loo__ package.)

```{r loo}
# note: the loo function accepts a 'cores' argument that we recommend specifying
# when working with bigger datasets

loo1 <- loo(fit1)
loo2 <- loo(fit2)
loo3 <- loo(fit3)
loo4 <- loo(fit4)
lpd_point <- cbind(
  loo1$pointwise[,"elpd_loo"], 
  loo2$pointwise[,"elpd_loo"],
  loo3$pointwise[,"elpd_loo"], 
  loo4$pointwise[,"elpd_loo"]
)
```

With `loo` we don't get any warnings for models 3 and 4, but for illustration of
good results, we display the diagnostic details for these models anyway.

```{r print-loo}
print(loo3)
print(loo4)
```

One benefit of PSIS-LOO over WAIC is better diagnostics. Here for both models 3
and 4 all $k<0.7$ and the Monte Carlo SE of `elpd_loo` is 0.1 or less, and we can expect
the model comparison to be reliable.

Next we compute and compare 1) WAIC weights, 2) Pseudo-BMA weights without
Bayesian bootstrap, 3) Pseudo-BMA+ weights with Bayesian bootstrap, and 4)
Bayesian stacking weights.

```{r weights}
waic_wts <- exp(waics) / sum(exp(waics))
pbma_wts <- pseudobma_weights(lpd_point, BB=FALSE)
pbma_BB_wts <- pseudobma_weights(lpd_point) # default is BB=TRUE
stacking_wts <- stacking_weights(lpd_point)
round(cbind(waic_wts, pbma_wts, pbma_BB_wts, stacking_wts), 2)
```

With all approaches Model 4 with `neocortex` and `log(mass)` gets most of the
weight. Based on theory, Pseudo-BMA weights without Bayesian bootstrap should be
close to WAIC weights, and we can also see that here. Pseudo-BMA+ weights with
Bayesian bootstrap provide more cautious weights further away from 0 and 1 (see
Yao et al. (2018) for a discussion of why this can be beneficial and results
from related experiments). In this particular example, the Bayesian stacking
weights are not much different from the other weights.

One of the benefits of stacking is that it manages well if there are many
similar models. Consider for example that there could be many irrelevant
covariates that when included would produce a similar model to one of the
existing models. To emulate this situation here we simply copy the first model a
bunch of times, but you can imagine that instead we would have ten alternative
models with about the same predictive performance. WAIC weights for such a
scenario would be close to the following:

```{r waic_wts_demo}
waic_wts_demo <- 
  exp(waics[c(1,1,1,1,1,1,1,1,1,1,2,3,4)]) /
  sum(exp(waics[c(1,1,1,1,1,1,1,1,1,1,2,3,4)]))
round(waic_wts_demo, 3)
```

Notice how much the weight for model 4 is lowered now that more models similar
to model 1 (or in this case identical) have been added. Both WAIC weights and
Pseudo-BMA approaches first estimate the predictive performance separately for
each model and then compute weights based on estimated relative predictive
performances. Similar models share similar weights so the weights of other
models must be reduced for the total sum of the weights to remain the same.

On the other hand, stacking optimizes the weights _jointly_, allowing for the
very similar models (in this toy example repeated models) to share their weight
while more unique models keep their original weights. In our example 
we can see this difference clearly: 

```{r stacking_weights}
stacking_weights(lpd_point[,c(1,1,1,1,1,1,1,1,1,1,2,3,4)])
```
Using stacking, the weight for the best model stays essentially unchanged.


# Example: Oceanic tool complexity

Another example we consider is the Kline oceanic tool complexity data, which
McElreath describes as follows:

>Different historical island populations possessed tool kits of different size. These kits include fish hooks, axes, boats, hand plows, and many other types of tools. A number of theories predict that larger populations will both develop and sustain more complex tool kits. ... It's also suggested that contact rates among populations effectively increases population [sic, probably should be tool kit] size, as it's relevant to technological evolution.

We build models predicting the total number of tools given the log
population size and the contact rate (high vs. low).

```{r Kline}
data(Kline)
d <- Kline
d$log_pop <- log(d$population)
d$contact_high <- ifelse(d$contact=="high", 1, 0)
str(d)
```

We start with a Poisson regression model with the log population size,
the contact rate, and an interaction term between them (priors are
informative priors as in _Statistical Rethinking_).
```{r fit10, results="hide"}
fit10 <-
  stan_glm(
    total_tools ~ log_pop + contact_high + log_pop * contact_high,
    family = poisson(link = "log"),
    data = d,
    prior = normal(0, 1, autoscale = FALSE),
    prior_intercept = normal(0, 100, autoscale = FALSE),
    seed = 2030
  )
```

Before running other models, we check whether Poisson is good choice
as the conditional observation model.

```{r loo10}
loo10 <- loo(fit10)
print(loo10)
```

We get at least one observation with $k>0.7$ and the estimated effective
number of parameters `p_loo` is larger than the total number of
parameters in the model. This indicates that Poisson might be too
narrow. A negative binomial model might be better, but with so few
observations it is not so clear.

We can compute LOO more accurately by running Stan again for the
leave-one-out folds with high $k$ estimates. When using __rstanarm__
this can be done by specifying the `k_threshold` argument:

```{r loo10-threshold}
loo10 <- loo(fit10, k_threshold=0.7)
print(loo10)
```

In this case we see that there is not much difference, and thus it is relatively
safe to continue.

As a comparison we also compute WAIC:
```{r waic10}
waic10 <- waic(fit10)
print(waic10)
```

The WAIC computation is giving warnings and the estimated ELPD is slightly more
optimistic. We recommend using the PSIS-LOO results instead.

To assess whether the contact rate and interaction term are useful, we can make
a comparison to models without these terms.

```{r contact_high, results="hide"}
fit11 <- update(fit10, formula = total_tools ~ log_pop + contact_high)
fit12 <- update(fit10, formula = total_tools ~ log_pop)
```
```{r loo-contact_high}
(loo11 <- loo(fit11))
(loo12 <- loo(fit12))
```
```{r relo-contact_high}
loo11 <- loo(fit11, k_threshold=0.7)
loo12 <- loo(fit12, k_threshold=0.7)
lpd_point <- cbind(
  loo10$pointwise[, "elpd_loo"], 
  loo11$pointwise[, "elpd_loo"], 
  loo12$pointwise[, "elpd_loo"]
)
```

For comparison we'll also compute WAIC values for these additional models:
```{r waic-contact_high}
waic11 <- waic(fit11)
waic12 <- waic(fit12)
waics <- c(
  waic10$estimates["elpd_waic", 1], 
  waic11$estimates["elpd_waic", 1], 
  waic12$estimates["elpd_waic", 1]
)
```
The WAIC computation again gives warnings, and we recommend using PSIS-LOO
instead.

Finally, we compute 1) WAIC weights, 2) Pseudo-BMA weights without
Bayesian bootstrap, 3) Pseudo-BMA+ weights with Bayesian bootstrap, and
4) Bayesian stacking weights.

```{r weights-contact_high}
waic_wts <- exp(waics) / sum(exp(waics))
pbma_wts <- pseudobma_weights(lpd_point, BB=FALSE)
pbma_BB_wts <- pseudobma_weights(lpd_point) # default is BB=TRUE
stacking_wts <- stacking_weights(lpd_point)
round(cbind(waic_wts, pbma_wts, pbma_BB_wts, stacking_wts), 2)
```

All weights favor the second model with the log population and the contact rate.
WAIC weights and Pseudo-BMA weights (without Bayesian bootstrap) are similar, while Pseudo-BMA+ is more cautious and closer to stacking weights. 

It may seem surprising that Bayesian stacking is giving zero weight to the first
model, but this is likely due to the fact that the estimated effect for the
interaction term is close to zero and thus models 1 and 2 give very similar
predictions. In other words, incorporating the model with the interaction (model
1) into the model average doesn't improve the predictions at all and so model 1
is given a weight of 0. On the other hand, models 2 and 3 are giving slightly
different predictions and thus their combination may be slightly better than
either alone. This behavior is related to the repeated similar model illustration
in the milk example above.

# Simpler coding using `loo_model_weights` function

Although in the examples above we called the `stacking_weights` and 
`pseudobma_weights` functions directly, we can also use the 
`loo_model_weights` wrapper, which takes as its input either a list
of pointwise log-likelihood matrices or a list of precomputed loo objects.
There are also `loo_model_weights` methods for stanreg objects (fitted
model objects from __rstanarm__) as well as fitted model objects from 
other packages (e.g. __brms__) that do the preparation work for the user
(see, e.g., the examples at `help("loo_model_weights", package = "rstanarm")`).

```{r loo_model_weights}
# using list of loo objects
loo_list <- list(loo10, loo11, loo12)
loo_model_weights(loo_list)
loo_model_weights(loo_list, method = "pseudobma")
loo_model_weights(loo_list, method = "pseudobma", BB = FALSE)
```

# References

McElreath, R. (2016). _Statistical rethinking: A Bayesian course with examples in R and Stan_. Chapman & Hall/CRC. http://xcelab.net/rm/statistical-rethinking/

Piironen, J. and Vehtari, A. (2017a). Sparsity information and
regularization in the horseshoe and other shrinkage priors. In
Electronic Journal of Statistics,
11(2):5018-5051. [Online](https://projecteuclid.org/euclid.ejs/1513306866).
	  
Piironen, J. and Vehtari, A. (2017b). Comparison of Bayesian predictive
methods for model selection. Statistics and Computing,
27(3):711-735. \doi:10.1007/s11222-016-9649-y. [Online](https://link.springer.com/article/10.1007/s11222-016-9649-y).

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC. _Statistics and
Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4.
[online](https://link.springer.com/article/10.1007/s11222-016-9696-4), 
[arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

Yao, Y., Vehtari, A., Simpson, D., and Gelman, A. (2018). Using
stacking to average Bayesian predictive distributions. In Bayesian
Analysis, \doi:10.1214/17-BA1091. [Online](https://projecteuclid.org/euclid.ba/1516093227).

---
title: "Using Leave-one-out cross-validation for large data"
author: "Mans Magnusson, Paul Bürkner, Aki Vehtari and Jonah Gabry"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Using Leave-one-out cross-validation for large data}
-->

```{r settings, child="children/SETTINGS-knitr.txt"}
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

This vignette demonstrates how to do leave-one-out cross-validation for large
data using the __loo__ package and Stan. There are two approaches covered:
LOO with subsampling and LOO using approximations to posterior distributions. 
Some sections from this vignette are excerpted from the papers

* Magnusson, M., Riis Andersen, M., Jonasson, J. and Vehtari, A. (2020). Leave-One-Out Cross-Validation for Model Comparison in Large Data.
Proceedings of the 23rd International Conference on Artificial Intelligence and Statistics (AISTATS), in PMLR 108. [arXiv preprint arXiv:2001.00980](https://arxiv.org/abs/2001.00980).

* Magnusson, M., Andersen, M., Jonasson, J. & Vehtari, A. (2019). Bayesian leave-one-out cross-validation for large data. Proceedings of the 36th International Conference on Machine Learning, in PMLR 97:4244-4253
[online](http://proceedings.mlr.press/v97/magnusson19a.html), 
[arXiv preprint arXiv:1904.10679](https://arxiv.org/abs/1904.10679).

* Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. _Statistics and Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. Links: [published](https://link.springer.com/article/10.1007/s11222-016-9696-4) | [arXiv preprint](https://arxiv.org/abs/1507.04544).

* Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

which provide important background for understanding the methods implemented in
the package.

# Setup 

In addition to the __loo__ package, we'll also be using __rstan__:

```{r setup, message=FALSE}
library("rstan")
library("loo")
set.seed(4711)
```


# Example: Well water in Bangladesh

We will use the same example as in the vignette [_Writing Stan programs for use
with the loo package_](http://mc-stan.org/loo/articles/loo2-with-rstan.html).
See that vignette for a description of the problem and data.

The sample size in this example is only $N=3020$, which is not large enough to
_require_ the special methods for large data described in this vignette, but
is sufficient for demonstration purposes in this tutorial.

## Coding the Stan model

Here is the Stan code for fitting the logistic regression model, which 
we save in a file called `logistic.stan`:

```
// Note: some syntax used in this program requires RStan >= 2.26 (or CmdStanR)
// To use an older version of RStan change the line declaring `y` to:
//    int<lower=0,upper=1> y[N];
data {
  int<lower=0> N;             // number of data points
  int<lower=0> P;             // number of predictors (including intercept)
  matrix[N,P] X;              // predictors (including 1s for intercept)
  array[N] int<lower=0,upper=1> y;  // binary outcome
}
parameters {
  vector[P] beta;
}
model {
  beta ~ normal(0, 1);
  y ~ bernoulli_logit(X * beta);
}
```

Importantly, unlike the general approach recommended in 
[_Writing Stan programs for use with the loo package_](http://mc-stan.org/loo/articles/loo2-with-rstan.html), 
we do _not_ compute the log-likelihood for each observation in the 
`generated quantities` block of the Stan program. Here we are assuming we have a
large data set (larger than the one we're actually using in this demonstration)
and so it is preferable to instead define a function in R to compute the
log-likelihood for each data point when needed rather than storing all of the
log-likelihood values in memory.

The log-likelihood in R can be coded as follows:
```{r llfun_logistic}
# we'll add an argument log to toggle whether this is a log-likelihood or 
# likelihood function. this will be useful later in the vignette.
llfun_logistic <- function(data_i, draws, log = TRUE) {
  x_i <- as.matrix(data_i[, which(grepl(colnames(data_i), pattern = "X")), drop=FALSE])
  logit_pred <- draws %*% t(x_i)
  dbinom(x = data_i$y, size = 1, prob = 1/(1 + exp(-logit_pred)), log = log)
}
```

The function `llfun_logistic()` needs to have arguments `data_i` and `draws`. 
Below we will test that the function is working by using the `loo_i()` function.

## Fitting the model with RStan

Next we fit the model in Stan using the **rstan** package:

```{r, eval=FALSE}
# Prepare data
url <- "http://stat.columbia.edu/~gelman/arm/examples/arsenic/wells.dat"
wells <- read.table(url)
wells$dist100 <- with(wells, dist / 100)
X <- model.matrix(~ dist100 + arsenic, wells)
standata <- list(y = wells$switch, X = X, N = nrow(X), P = ncol(X))

# Compile
stan_mod <- stan_model("logistic.stan")

# Fit model
fit_1 <- sampling(stan_mod, data = standata, seed = 4711)
print(fit_1, pars = "beta")
```

```
         mean se_mean   sd  2.5%   25%   50%   75% 97.5% n_eff Rhat
beta[1]  0.00       0 0.08 -0.15 -0.05  0.00  0.06  0.16  1933    1
beta[2] -0.89       0 0.10 -1.09 -0.96 -0.89 -0.82 -0.69  2332    1
beta[3]  0.46       0 0.04  0.38  0.43  0.46  0.49  0.54  2051    1
```

Before we move on to computing LOO we can now test that the log-likelihood
function we wrote is working as it should. The `loo_i()` function is a helper
function that can be used to test a log-likelihood function on a single observation.

```{r, eval=FALSE}
# used for draws argument to loo_i
parameter_draws_1 <- extract(fit_1)$beta

# used for data argument to loo_i
stan_df_1 <- as.data.frame(standata)

# compute relative efficiency (this is slow and optional but is recommended to allow 
# for adjusting PSIS effective sample size based on MCMC effective sample size)
r_eff <- relative_eff(llfun_logistic, 
                      log = FALSE, # relative_eff wants likelihood not log-likelihood values
                      chain_id = rep(1:4, each = 1000), 
                      data = stan_df_1, 
                      draws = parameter_draws_1, 
                      cores = 2)

loo_i(i = 1, llfun_logistic, r_eff = r_eff, data = stan_df_1, draws = parameter_draws_1)
```

```
$pointwise
    elpd_loo mcse_elpd_loo        p_loo     looic influence_pareto_k
1 -0.3314552  0.0002887608 0.0003361772 0.6629103        -0.05679886
...
```

# Approximate LOO-CV using PSIS-LOO and subsampling

We can then use the `loo_subsample()` function to compute the efficient PSIS-LOO
approximation to exact LOO-CV using subsampling:

```{r, eval=FALSE}
set.seed(4711)
loo_ss_1 <-
  loo_subsample(
    llfun_logistic,
    observations = 100, # take a subsample of size 100
    cores = 2,
    # these next objects were computed above
    r_eff = r_eff, 
    draws = parameter_draws_1,
    data = stan_df_1
  )
print(loo_ss_1)
```

```
Computed from 4000 by 100 subsampled log-likelihood
values from 3020 total observations.

         Estimate   SE subsampling SE
elpd_loo  -1968.5 15.6            0.3
p_loo         3.1  0.1            0.4
looic      3936.9 31.2            0.6
------
Monte Carlo SE of elpd_loo is 0.0.
MCSE and ESS estimates assume MCMC draws (r_eff in [0.9, 1.0]).

All Pareto k estimates are good (k < 0.7).
See help('pareto-k-diagnostic') for details.
```

The `loo_subsample()` function creates an object of class `psis_loo_ss`, that
inherits from `psis_loo, loo` (the classes of regular `loo` objects).

The printed output above shows the estimates
$\widehat{\mbox{elpd}}_{\rm loo}$ (expected log predictive density),
$\widehat{p}_{\rm loo}$ (effective number of parameters), and ${\rm looic} =-2\,
\widehat{\mbox{elpd}}_{\rm loo}$ (the LOO information criterion). Unlike when 
using `loo()`, when using `loo_subsample()` there is an additional column 
giving the "subsampling SE", which reflects the additional uncertainty due 
to the subsampling used. 

The line at the bottom of the printed output provides information about the
reliability of the LOO approximation (the interpretation of the $k$ parameter is
explained in `help('pareto-k-diagnostic')` and in greater detail in Vehtari,
Simpson, Gelman, Yao, and Gabry (2019)). In this case, the message tells us that all of the
estimates for $k$ are fine _for this given subsample_.


## Adding additional subsamples

If we are not satisfied with the subsample size (i.e., the accuracy) we can
simply add more samples until we are satisfied using the `update()` method.

```{r, eval=FALSE}
set.seed(4711)
loo_ss_1b <-
  update(
    loo_ss_1,
    observations = 200, # subsample 200 instead of 100
    r_eff = r_eff,
    draws = parameter_draws_1,
    data = stan_df_1
  ) 
print(loo_ss_1b)
```

```
Computed from 4000 by 200 subsampled log-likelihood
values from 3020 total observations.

         Estimate   SE subsampling SE
elpd_loo  -1968.3 15.6            0.2
p_loo         3.2  0.1            0.4
looic      3936.7 31.2            0.5
------
Monte Carlo SE of elpd_loo is 0.0.
MCSE and ESS estimates assume MCMC draws (r_eff in [0.9, 1.0]).

All Pareto k estimates are good (k < 0.7).
See help('pareto-k-diagnostic') for details.
```

## Specifying estimator and sampling method

The performance relies on two components: the estimation method and the
approximation used for the elpd. See the documentation for `loo_subsample()`
more information on which estimators and approximations are implemented. The
default implementation is using the point log predictive density evaluated at
the mean of the posterior (`loo_approximation="plpd"`) and the difference
estimator (`estimator="diff_srs"`). This combination has a focus on fast
inference. But we can easily use other estimators as well as other elpd
approximations, for example:

```{r, eval=FALSE}
set.seed(4711)
loo_ss_1c <-
  loo_subsample(
    x = llfun_logistic,
    r_eff = r_eff,
    draws = parameter_draws_1,
    data = stan_df_1,
    observations = 100,
    estimator = "hh_pps", # use Hansen-Hurwitz
    loo_approximation = "lpd", # use lpd instead of plpd
    loo_approximation_draws = 100,
    cores = 2
  )
print(loo_ss_1c)
```


```
Computed from 4000 by 100 subsampled log-likelihood
values from 3020 total observations.

         Estimate   SE subsampling SE
elpd_loo  -1968.9 15.4            0.5
p_loo         3.5  0.2            0.5
looic      3937.9 30.7            1.1
------
Monte Carlo SE of elpd_loo is 0.0.
MCSE and ESS estimates assume MCMC draws (r_eff in [0.9, 1.0]).

All Pareto k estimates are good (k < 0.7).
See help('pareto-k-diagnostic') for details.
```

See the documentation and references for `loo_subsample()` for details on
the implemented approximations.


# Approximate LOO-CV using PSIS-LOO with posterior approximations

Using posterior approximations, such as variational inference and Laplace
approximations, can further speed-up LOO-CV for large data. Here we demonstrate
using a Laplace approximation in Stan.

```{r, eval=FALSE}
fit_laplace <- optimizing(stan_mod, data = standata, draws = 2000, 
                          importance_resampling = TRUE)
parameter_draws_laplace <- fit_laplace$theta_tilde # draws from approximate posterior
log_p <- fit_laplace$log_p # log density of the posterior
log_g <- fit_laplace$log_g # log density of the approximation
```

Using the posterior approximation we can then do LOO-CV by correcting for the
posterior approximation when we compute the elpd. To do this we use the 
`loo_approximate_posterior()` function.

```{r, eval=FALSE}
set.seed(4711)
loo_ap_1 <-
  loo_approximate_posterior(
    x = llfun_logistic,
    draws = parameter_draws_laplace,
    data = stan_df_1,
    log_p = log_p,
    log_g = log_g,
    cores = 2
  )
print(loo_ap_1)
```

The function creates a class, `psis_loo_ap` that inherits from `psis_loo, loo`.

```
Computed from 2000 by 3020 log-likelihood matrix

         Estimate   SE
elpd_loo  -1968.4 15.6
p_loo         3.2  0.2
looic      3936.8 31.2
------
Posterior approximation correction used.
Monte Carlo SE of elpd_loo is 0.0.
MCSE and ESS estimates assume independent draws (r_eff=1).

All Pareto k estimates are good (k < 0.7).
See help('pareto-k-diagnostic') for details.
```

## Combining the posterior approximation method with subsampling

The posterior approximation correction can also be used together with
subsampling:

```{r, eval=FALSE}
set.seed(4711)
loo_ap_ss_1 <-
  loo_subsample(
    x = llfun_logistic,
    draws = parameter_draws_laplace,
    data = stan_df_1,
    log_p = log_p,
    log_g = log_g,
    observations = 100,
    cores = 2
  )
print(loo_ap_ss_1)
```

```
Computed from 2000 by 100 subsampled log-likelihood
values from 3020 total observations.

         Estimate   SE subsampling SE
elpd_loo  -1968.2 15.6            0.4
p_loo         2.9  0.1            0.5
looic      3936.4 31.1            0.8
------
Posterior approximation correction used.
Monte Carlo SE of elpd_loo is 0.0.
MCSE and ESS estimates assume independent draws (r_eff=1).

All Pareto k estimates are good (k < 0.7).
See help('pareto-k-diagnostic') for details.
```

The object created is of class `psis_loo_ss`, which inherits from the
`psis_loo_ap` class previously described.


## Comparing models

To compare this model to an alternative model for the same data we can use the
`loo_compare()` function just as we would if using `loo()` instead of
`loo_subsample()` or `loo_approximate_posterior()`. First we'll fit a second
model to the well-switching data, using `log(arsenic)` instead of `arsenic` as a
predictor:

```{r, eval=FALSE}
standata$X[, "arsenic"] <- log(standata$X[, "arsenic"])
fit_2 <- sampling(stan_mod, data = standata) 
parameter_draws_2 <- extract(fit_2)$beta
stan_df_2 <- as.data.frame(standata)

# recompute subsampling loo for first model for demonstration purposes

# compute relative efficiency (this is slow and optional but is recommended to allow 
# for adjusting PSIS effective sample size based on MCMC effective sample size)
r_eff_1 <- relative_eff(
  llfun_logistic,
  log = FALSE, # relative_eff wants likelihood not log-likelihood values
  chain_id = rep(1:4, each = 1000),
  data = stan_df_1,
  draws = parameter_draws_1,
  cores = 2
)

set.seed(4711)
loo_ss_1 <- loo_subsample(
  x = llfun_logistic,
  r_eff = r_eff_1,
  draws = parameter_draws_1,
  data = stan_df_1,
  observations = 200,
  cores = 2
)

# compute subsampling loo for a second model (with log-arsenic)

r_eff_2 <- relative_eff(
  llfun_logistic,
  log = FALSE, # relative_eff wants likelihood not log-likelihood values
  chain_id = rep(1:4, each = 1000),
  data = stan_df_2,
  draws = parameter_draws_2,
  cores = 2
)
loo_ss_2 <- loo_subsample(
  x = llfun_logistic,
  r_eff = r_eff_2, 
  draws = parameter_draws_2,
  data = stan_df_2,
  observations = 200,
  cores = 2
)

print(loo_ss_2)
```

```
Computed from 4000 by 100 subsampled log-likelihood
values from 3020 total observations.

         Estimate   SE subsampling SE
elpd_loo  -1952.0 16.2            0.2
p_loo         2.6  0.1            0.3
looic      3903.9 32.4            0.4
------
Monte Carlo SE of elpd_loo is 0.0.
MCSE and ESS estimates assume MCMC draws (r_eff in [1.0, 1.1]).

All Pareto k estimates are good (k < 0.7).
See help('pareto-k-diagnostic') for details.
```

We can now compare the models on LOO using the `loo_compare` function:

```{r, eval=FALSE}
# Compare
comp <- loo_compare(loo_ss_1, loo_ss_2)
print(comp)
```

```
Warning: Different subsamples in 'model2' and 'model1'. Naive diff SE is used.

       elpd_diff se_diff subsampling_se_diff
model2  0.0       0.0     0.0               
model1 16.5      22.5     0.4               
```

This new object `comp` contains the estimated difference of expected
leave-one-out prediction errors between the two models, along with the standard
error. As the warning indicates, because different subsamples were used the
comparison will not take the correlations between different observations into
account. Here we see that the naive SE is 22.5 and we cannot see any difference
in performance between the models.

To force subsampling to use the same observations for each of the models 
we can simply extract the observations used in `loo_ss_1` and use them in 
`loo_ss_2` by supplying the `loo_ss_1` object to the `observations` argument.


```{r, eval=FALSE}
loo_ss_2 <-
  loo_subsample(
    x = llfun_logistic,
    r_eff = r_eff_2,
    draws = parameter_draws_2,
    data = stan_df_2,
    observations = loo_ss_1,
    cores = 2
  )
```

We could also supply the subsampling indices using the `obs_idx()`
helper function:

```{r, eval=FALSE}
idx <- obs_idx(loo_ss_1)
loo_ss_2 <- loo_subsample(
  x = llfun_logistic,
  r_eff = r_eff_2, 
  draws = parameter_draws_2,
  data = stan_df_2,
  observations = idx,
  cores = 2
)
```

```
Simple random sampling with replacement assumed.
```

This results in a message indicating that we assume these observations to have
been sampled with simple random sampling, which is true because we had used the
default `"diff_srs"` estimator for `loo_ss_1`.

We can now compare the models and estimate the difference based on the same
subsampled observations.

```{r, eval=FALSE}
comp <- loo_compare(loo_ss_1, loo_ss_2)
print(comp) 
```

```
       elpd_diff se_diff subsampling_se_diff
model2  0.0       0.0     0.0               
model1 16.1       4.4     0.1               
```

First, notice that now the `se_diff` is now around 4 (as opposed to 20 when using
different subsamples). The first column shows the difference in ELPD relative to
the model with the largest ELPD. In this case, the difference in `elpd` and its
scale relative to the approximate standard error of the difference) indicates a
preference for the second model (`model2`). Since the subsampling uncertainty is
so small in this case it can effectively be ignored. If we need larger
subsamples we can simply add samples using the `update()` method demonstrated
earlier.

It is also possible to compare a subsampled loo computation with a full loo object.

```{r, eval=FALSE}
# use loo() instead of loo_subsample() to compute full PSIS-LOO for model 2
loo_full_2 <- loo(
  x = llfun_logistic,
  r_eff = r_eff_2,
  draws = parameter_draws_2,
  data = stan_df_2,
  cores = 2
)
loo_compare(loo_ss_1, loo_full_2)
```

```
Estimated elpd_diff using observations included in loo calculations for all models.
```

Because we are comparing a non-subsampled loo calculation to a subsampled
calculation we get the message that only the observations that are included in
the loo calculations for both `model1` and `model2` are included in the
computations for the comparison.

```
       elpd_diff se_diff subsampling_se_diff
model2  0.0       0.0     0.0               
model1 16.3       4.4     0.3   
```

Here we actually see an increase in `subsampling_se_diff`, but this is due to a
technical detail not elaborated here. In general, the difference should be
better or negligible.

# References

Gelman, A., and Hill, J. (2007).  *Data Analysis Using Regression and Multilevel Hierarchical Models.*  Cambridge University Press.

Stan Development Team (2017). _The Stan C++ Library, Version 2.17.0._   https://mc-stan.org/

Stan Development Team (2018) _RStan: the R interface to Stan, Version 2.17.3._   https://mc-stan.org/

Magnusson, M., Riis Andersen, M., Jonasson, J. and Vehtari, A. (2020). Leave-One-Out Cross-Validation for Model Comparison in Large Data.
Proceedings of the 23rd International Conference on Artificial Intelligence and Statistics (AISTATS), in PMLR 108. [arXiv preprint arXiv:2001.00980](https://arxiv.org/abs/2001.00980).

Magnusson, M., Andersen, M., Jonasson, J. & Vehtari, A. (2019). Bayesian leave-one-out cross-validation for large data. Proceedings of the 36th International Conference on Machine Learning, in PMLR 97:4244-4253
[online](http://proceedings.mlr.press/v97/magnusson19a.html), 
[arXiv preprint arXiv:1904.10679](https://arxiv.org/abs/1904.10679).

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC. _Statistics and
Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4.
[online](https://link.springer.com/article/10.1007/s11222-016-9696-4), 
[arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

---
title: "Avoiding model refits in leave-one-out cross-validation with moment matching"
author: "Topi Paananen, Paul Bürkner, Aki Vehtari and Jonah Gabry"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Avoiding model refits in leave-one-out cross-validation with moment matching}
-->
```{r, child="children/SETTINGS-knitr.txt"}
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

This vignette demonstrates how to improve the Monte Carlo sampling accuracy of
leave-one-out cross-validation with the __loo__ package and Stan. The __loo__
package automatically monitors the sampling accuracy using Pareto $k$
diagnostics for each observation. Here, we present a method for quickly
improving the accuracy when the Pareto diagnostics indicate problems. This is
done by performing some additional computations using the existing posterior
sample. If successful, this will decrease the Pareto $k$ values, making the
model assessment more reliable. __loo__ also stores the original Pareto $k$
values with the name `influence_pareto_k` which are not changed. They can be
used as a diagnostic of how much each observation influences the posterior
distribution. 

The methodology presented is based on the paper

* Paananen, T., Piironen, J., Buerkner, P.-C., Vehtari, A. (2020).
Implicitly Adaptive Importance Sampling.
[arXiv preprint arXiv:1906.08850](https://arxiv.org/abs/1906.08850).

More information about the Pareto $k$ diagnostics is given in the following
papers

* Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. _Statistics and Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. Links: [published](https://link.springer.com/article/10.1007/s11222-016-9696-4) | [arXiv preprint](https://arxiv.org/abs/1507.04544).

* Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

# Example: Eradication of Roaches

We will use the same example as in the vignette
[_Using the loo package (version >= 2.0.0)_](https://mc-stan.org/loo/articles/loo2-example.html).
See the demo for a description of the problem and data. We will use the same Poisson
regression model as in the case study.

## Coding the Stan model

Here is the Stan code for fitting the Poisson regression model, which 
we will use for modeling the number of roaches.

```{r stancode}
# Note: some syntax used in this Stan program requires RStan >= 2.26 (or CmdStanR)
# To use an older version of RStan change the line declaring `y` to: int y[N];
stancode <- "
data {
  int<lower=1> K;
  int<lower=1> N;
  matrix[N,K] x;
  array[N] int y;
  vector[N] offset_; // offset is reserved keyword in Stan so use offset_

  real beta_prior_scale;
  real alpha_prior_scale;
}
parameters {
  vector[K] beta;
  real intercept;
}
model {
  y ~ poisson(exp(x * beta + intercept + offset_));
  beta ~ normal(0,beta_prior_scale);
  intercept ~ normal(0,alpha_prior_scale);
}
generated quantities {
  vector[N] log_lik;
  for (n in 1:N) {
    log_lik[n] = poisson_lpmf(y[n] | exp(x[n] * beta + intercept + offset_[n]));
  }
}
"
```

Following the usual approach recommended in 
[_Writing Stan programs for use with the loo package_](http://mc-stan.org/loo/articles/loo2-with-rstan.html), 
we compute the log-likelihood for each observation in the 
`generated quantities` block of the Stan program.


## Setup

In addition to __loo__, we load the __rstan__ package for fitting the model,
and the __rstanarm__ package for the data.

```{r setup, message=FALSE}
library("rstan")
library("loo")
seed <- 9547
set.seed(seed)
```

## Fitting the model with RStan

Next we fit the model in Stan using the __rstan__ package:

```{r modelfit, message=FALSE}
# Prepare data
data(roaches, package = "rstanarm")
roaches$roach1 <- sqrt(roaches$roach1)
y <- roaches$y
x <- roaches[, c("roach1", "treatment", "senior")]
offset <- log(roaches[, "exposure2"])
n <- dim(x)[1]
k <- dim(x)[2]

standata <- list(
  N = n,
  K = k,
  x = as.matrix(x),
  y = y,
  offset_ = offset,
  beta_prior_scale = 2.5,
  alpha_prior_scale = 5.0
)

# Compile
stanmodel <- stan_model(model_code = stancode)

# Fit model
fit <- sampling(stanmodel, data = standata, seed = seed, refresh = 0)
print(fit, pars = "beta")
```

Let us now evaluate the predictive performance of the model using `loo()`.

```{r loo1}
loo1 <- loo(fit)
loo1
```

The `loo()` function output warnings that there are some observations which are
highly influential, and thus the accuracy of importance sampling is compromised
as indicated by the large Pareto $k$ diagnostic values (> 0.7). As discussed in
the vignette
[_Using the loo package (version >= 2.0.0)_](https://mc-stan.org/loo/articles/loo2-example.html),
this may be an indication of model misspecification. Despite that, it is still
beneficial to be able to evaluate the predictive performance of the model
accurately.

## Moment matching correction for importance sampling

To improve the accuracy of the `loo()` result above, we could perform
leave-one-out cross-validation by explicitly leaving out single observations and
refitting the model using MCMC repeatedly. However, the Pareto $k$ diagnostics
indicate that there are 19 observations which are problematic. This would
require 19 model refits which may require a lot of computation time.

Instead of refitting with MCMC, we can perform a faster moment matching
correction to the importance sampling for the problematic observations. This can
be done with the `loo_moment_match()` function in the __loo__ package, which
takes our existing `loo` object as input and modifies it. The moment matching
requires some evaluations of the model posterior density. For models fitted with
__rstan__, this can be conveniently done by using the existing `stanfit` object.

First, we show how the moment matching can be used for a model fitted using
__rstan__. It only requires setting the argument `moment_match` to `TRUE` in the
`loo()` function. Optionally, you can also set the argument `k_threshold` which
determines the Pareto $k$ threshold, above which moment matching is used. By
default, it operates on all observations whose Pareto $k$ value is larger than
the sample size ($S$) specific threshold $\min(1 - 1 / \log_{10}(S), 0.7)$ (which is $0.7$ for $S>2200$).

```{r loo_moment_match}
# available in rstan >= 2.21
loo2 <- loo(fit, moment_match = TRUE)
loo2
```

After the moment matching, all observations have the diagnostic
Pareto $k$ less than 0.7, meaning that the estimates are now reliable.
The total `elpd_loo` estimate also changed from `-5457.8` to `-5478.5`, showing
that before moment matching, `loo()` overestimated the predictive
performance of the model.

The updated Pareto $k$ values stored in
`loo2$diagnostics$pareto_k` are considered algorithmic diagnostic values that
indicate the sampling accuracy. The original Pareto $k$ values are
stored in `loo2$pointwise[,"influence_pareto_k"]` and these are not modified
by the moment matching. These can be considered as diagnostics
for how big influence each observation has on the posterior distribution.
In addition to the Pareto $k$ diagnostics, moment matching also updates
the effective sample size estimates. 

# Using `loo_moment_match()` directly

The moment matching can also be performed by explicitly calling the function
`loo_moment_match()`. This enables its use also for models that are not using
__rstan__ or another package with built-in support for `loo_moment_match()`. To
use `loo_moment_match()`, the user must give the model object `x`, the `loo`
object, and 5 helper functions as arguments to `loo_moment_match()`. The helper
functions are

* `post_draws`	
    + A function the takes `x` as the first argument and returns a matrix of
    posterior draws of the model parameters, `pars`.
* `log_lik_i`	
    + A function that takes `x` and `i` and returns a matrix
    (one column per chain) or a vector (all chains stacked) of log-likeliood
    draws of the ith observation based on the model `x`. If the draws are
    obtained using MCMC, the matrix with MCMC chains separated is preferred.
* `unconstrain_pars`	
    + A function that takes arguments `x` and `pars`, and returns posterior
    draws on the unconstrained space based on the posterior draws on the
    constrained space passed via `pars`.
* `log_prob_upars`	
    + A function that takes arguments `x` and `upars`, and returns a matrix of
    log-posterior density values of the unconstrained posterior draws passed
    via `upars`.
* `log_lik_i_upars`	
    + A function that takes arguments `x`, `upars`, and `i` and returns a
    vector of log-likelihood draws of the `i`th observation based on the
    unconstrained posterior draws passed via `upars`.

Next, we show how the helper functions look like for RStan objects, and show an
example of using `loo_moment_match()` directly. For stanfit objects from 
__rstan__ objects, the functions look like this:

```{r stanfitfuns}
# create a named list of draws for use with rstan methods
.rstan_relist <- function(x, skeleton) {
  out <- utils::relist(x, skeleton)
  for (i in seq_along(skeleton)) {
    dim(out[[i]]) <- dim(skeleton[[i]])
  }
  out
}

# rstan helper function to get dims of parameters right
.create_skeleton <- function(pars, dims) {
  out <- lapply(seq_along(pars), function(i) {
    len_dims <- length(dims[[i]])
    if (len_dims < 1) {
      return(0)
    }
    return(array(0, dim = dims[[i]]))
  })
  names(out) <- pars
  out
}

# extract original posterior draws
post_draws_stanfit <- function(x, ...) {
  as.matrix(x)
}

# compute a matrix of log-likelihood values for the ith observation
# matrix contains information about the number of MCMC chains
log_lik_i_stanfit <- function(x, i, parameter_name = "log_lik", ...) {
  loo::extract_log_lik(x, parameter_name, merge_chains = FALSE)[,, i]
}

# transform parameters to the unconstraint space
unconstrain_pars_stanfit <- function(x, pars, ...) {
  skeleton <- .create_skeleton(x@sim$pars_oi, x@par_dims[x@sim$pars_oi])
  upars <- apply(pars, 1, FUN = function(theta) {
    rstan::unconstrain_pars(x, .rstan_relist(theta, skeleton))
  })
  # for one parameter models
  if (is.null(dim(upars))) {
    dim(upars) <- c(1, length(upars))
  }
  t(upars)
}

# compute log_prob for each posterior draws on the unconstrained space
log_prob_upars_stanfit <- function(x, upars, ...) {
  apply(
    upars,
    1,
    rstan::log_prob,
    object = x,
    adjust_transform = TRUE,
    gradient = FALSE
  )
}

# compute log_lik values based on the unconstrained parameters
log_lik_i_upars_stanfit <- function(
  x,
  upars,
  i,
  parameter_name = "log_lik",
  ...
) {
  S <- nrow(upars)
  out <- numeric(S)
  for (s in seq_len(S)) {
    out[s] <- rstan::constrain_pars(x, upars = upars[s, ])[[parameter_name]][i]
  }
  out
}
```

Using these function, we can call `loo_moment_match()` to update the existing
`loo` object.

```{r loo_moment_match.default, message=FALSE}
loo3 <- loo::loo_moment_match.default(
  x = fit,
  loo = loo1,
  post_draws = post_draws_stanfit,
  log_lik_i = log_lik_i_stanfit,
  unconstrain_pars = unconstrain_pars_stanfit,
  log_prob_upars = log_prob_upars_stanfit,
  log_lik_i_upars = log_lik_i_upars_stanfit
)
loo3
```

As expected, the result is identical to the previous result of
`loo2 <- loo(fit, moment_match = TRUE)`.

# References

Gelman, A., and Hill, J. (2007).  *Data Analysis Using Regression and Multilevel Hierarchical Models.*  Cambridge University Press.

Stan Development Team (2020) _RStan: the R interface to Stan, Version 2.21.1_   https://mc-stan.org

Paananen, T., Piironen, J., Buerkner, P.-C., Vehtari, A. (2021).
Implicitly adaptive importance sampling. _Statistics and Computing_, 31, 16.
\doi:10.1007/s11222-020-09982-2. arXiv preprint arXiv:1906.08850.

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. _Statistics and Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. Links: [published](https://link.springer.com/article/10.1007/s11222-016-9696-4) | [arXiv preprint](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

---
title: "Writing Stan programs for use with the loo package"
author: "Aki Vehtari and Jonah Gabry"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Writing Stan programs for use with the loo package}
-->

```{r settings, child="children/SETTINGS-knitr.txt"}
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

This vignette demonstrates how to write a Stan program that computes and stores
the pointwise log-likelihood required for using the __loo__ package. The other
vignettes included with the package demonstrate additional functionality.

Some sections from this vignette are excerpted from our papers

* Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. _Statistics and Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. Links: [published](https://link.springer.com/article/10.1007/s11222-016-9696-4) | [arXiv preprint](https://arxiv.org/abs/1507.04544).

* Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

which provide important background for understanding the methods implemented in
the package.


# Example: Well water in Bangladesh

This example comes from a survey of residents from a small area in Bangladesh
that was affected by arsenic in drinking water. Respondents with elevated
arsenic levels in their wells were asked if they were interested in getting
water from a neighbor's well, and a series of logistic regressions were fit to
predict this binary response given various information about the households
(Gelman and Hill, 2007). Here we fit a model for the well-switching response
given two predictors: the arsenic level of the water in the resident's home, and
the distance of the house from the nearest safe well.

The sample size in this example is $N=3020$, which is not huge but is large
enough that it is important to have a computational method for LOO that is fast
for each data point. On the plus side, with such a large dataset, the influence
of any given observation is small, and so the computations should be stable.

## Coding the Stan model

Here is the Stan code for fitting the logistic regression model, which 
we save in a file called `logistic.stan`:

```
// Note: some syntax used in this program requires RStan >= 2.26 (or CmdStanR)
// To use an older version of RStan change the line declaring `y` to:
//    int<lower=0,upper=1> y[N];
data {
  int<lower=0> N;                   // number of data points
  int<lower=0> P;                   // number of predictors (including intercept)
  matrix[N,P] X;                    // predictors (including 1s for intercept)
  array[N] int<lower=0,upper=1> y;  // binary outcome
}
parameters {
  vector[P] beta;
}
model {
  beta ~ normal(0, 1);
  y ~ bernoulli_logit(X * beta);
}
generated quantities {
  vector[N] log_lik;
  for (n in 1:N) {
    log_lik[n] = bernoulli_logit_lpmf(y[n] | X[n] * beta);
  }
}
```

We have defined the log likelihood as a vector named `log_lik` in the generated 
quantities block so that the individual terms will be saved by Stan. After 
running Stan, `log_lik` can be extracted (using the `extract_log_lik` function
provided in the **loo** package) as an $S \times N$ matrix, where $S$ is the 
number of simulations (posterior draws) and $N$ is the number of data points.

## Fitting the model with RStan

Next we fit the model in Stan using the **rstan** package:

```{r, eval=FALSE}
library("rstan")

# Prepare data 
url <- "http://stat.columbia.edu/~gelman/arm/examples/arsenic/wells.dat"
wells <- read.table(url)
wells$dist100 <- with(wells, dist / 100)
X <- model.matrix(~ dist100 + arsenic, wells)
standata <- list(y = wells$switch, X = X, N = nrow(X), P = ncol(X))

# Fit model
fit_1 <- stan("logistic.stan", data = standata)
print(fit_1, pars = "beta")
```

```
         mean se_mean   sd  2.5%   25%   50%   75% 97.5% n_eff Rhat
beta[1]  0.00       0 0.08 -0.16 -0.05  0.00  0.05  0.15  1964    1
beta[2] -0.89       0 0.10 -1.09 -0.96 -0.89 -0.82 -0.68  2048    1
beta[3]  0.46       0 0.04  0.38  0.43  0.46  0.49  0.54  2198    1
```

## Computing approximate leave-one-out cross-validation using PSIS-LOO

We can then use the **loo** package to compute the efficient PSIS-LOO
approximation to exact LOO-CV:

```{r, eval=FALSE}
library("loo")

# Extract pointwise log-likelihood
# using merge_chains=FALSE returns an array, which is easier to 
# use with relative_eff()
log_lik_1 <- extract_log_lik(fit_1, merge_chains = FALSE)

# as of loo v2.0.0 we can optionally provide relative effective sample sizes
# when calling loo, which allows for better estimates of the PSIS effective
# sample sizes and Monte Carlo error
r_eff <- relative_eff(exp(log_lik_1), cores = 2) 

# preferably use more than 2 cores (as many cores as possible)
# will use value of 'mc.cores' option if cores is not specified
loo_1 <- loo(log_lik_1, r_eff = r_eff, cores = 2)
print(loo_1)
```

```
Computed from 4000 by 3020 log-likelihood matrix

         Estimate   SE
elpd_loo  -1968.5 15.6
p_loo         3.2  0.1
looic      3937.0 31.2
------
Monte Carlo SE of elpd_loo is 0.0.
MCSE and ESS estimates assume MCMC draws (r_eff in [0.5, 1.3]).

All Pareto k estimates are good (k < 0.7).
See help('pareto-k-diagnostic') for details.
```

The printed output from the `loo` function shows the estimates
$\widehat{\mbox{elpd}}_{\rm loo}$ (expected log predictive density),
$\widehat{p}_{\rm loo}$ (effective number of parameters), and ${\rm looic} =-2\,
\widehat{\mbox{elpd}}_{\rm loo}$ (the LOO information criterion).

The line at the bottom of the printed output provides information about the
reliability of the LOO approximation (the interpretation of the $k$ parameter is
explained in `help('pareto-k-diagnostic')` and in greater detail
in Vehtari, Simpson, Gelman, Yao, and Gabry (2019)). In this case the message tells us that
all of the estimates for $k$ are fine.

## Comparing models

To compare this model to an alternative model for the same data we can use the 
`loo_compare` function in the **loo** package. First we'll fit a second model to the
well-switching data, using `log(arsenic)` instead of `arsenic` as a predictor:

```{r, eval=FALSE}
standata$X[, "arsenic"] <- log(standata$X[, "arsenic"])
fit_2 <- stan(fit = fit_1, data = standata) 

log_lik_2 <- extract_log_lik(fit_2, merge_chains = FALSE)
r_eff_2 <- relative_eff(exp(log_lik_2))
loo_2 <- loo(log_lik_2, r_eff = r_eff_2, cores = 2)
print(loo_2)
```

```
Computed from 4000 by 3020 log-likelihood matrix

         Estimate   SE
elpd_loo  -1952.3 16.2
p_loo         3.1  0.1
looic      3904.6 32.4
------
Monte Carlo SE of elpd_loo is 0.0.
MCSE and ESS estimates assume MCMC draws (r_eff in [0.4, 1.2]).

All Pareto k estimates are good (k < 0.7).
See help('pareto-k-diagnostic') for details.
```

We can now compare the models on LOO using the `loo_compare` function:

```{r, eval=FALSE}
# Compare
comp <- loo_compare(loo_1, loo_2)
```

This new object, `comp`, contains the estimated difference of expected 
leave-one-out prediction errors between the two models, along with the standard 
error:

```{r, eval=FALSE}
print(comp) # can set simplify=FALSE for more detailed print output
```
```
       elpd_diff se_diff
model2   0.0       0.0  
model1 -16.3       4.4  
```

The first column shows the difference in ELPD relative to the model with 
the largest ELPD. In this case, the difference in `elpd` and its scale relative 
to the approximate standard error of the difference) indicates a 
preference for the second model (`model2`).


# References

Gelman, A., and Hill, J. (2007).  *Data Analysis Using Regression and Multilevel Hierarchical Models.*  Cambridge University Press.

Stan Development Team (2017). _The Stan C++ Library, Version 2.17.0._   https://mc-stan.org/

Stan Development Team (2018) _RStan: the R interface to Stan, Version 2.17.3._   https://mc-stan.org/

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC. _Statistics and
Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4.
[online](https://link.springer.com/article/10.1007/s11222-016-9696-4), 
[arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

---
title: "Approximate leave-future-out cross-validation for Bayesian time series models"
author: "Paul Bürkner, Jonah Gabry, Aki Vehtari"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
encoding: "UTF-8"
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---

<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Approximate leave-future-out cross-validation for Bayesian time series models}
-->

```{r settings, child="children/SETTINGS-knitr.txt"}
```

```{r more-knitr-ops, include=FALSE}
knitr::opts_chunk$set(
  cache = TRUE,
  message = FALSE, 
  warning = FALSE
)
```

## Introduction

One of the most common goals of a time series analysis is to use the observed
series to inform predictions for future observations. We will refer to this task
of predicting a sequence of $M$ future observations as $M$-step-ahead prediction
($M$-SAP). Fortunately, once we have fit a model and can sample from the
posterior predictive distribution, it is straightforward to generate predictions
as far into the future as we want. It is also straightforward to evaluate the
$M$-SAP performance of a time series model by comparing the predictions to the
observed sequence of $M$ future data points once they become available.

Unfortunately, we are often in the position of having to use a model to inform
decisions _before_ we can collect the future observations required for assessing
the predictive performance. If we have many competing models we may also need to
first decide which of the models (or which combination of the models) we should
rely on for predictions. In these situations the best we can do is to use
methods for approximating the expected predictive performance of our models
using only the observations of the time series we already have.

If there were no time dependence in the data or if the focus is to assess
the non-time-dependent part of the model, we could use methods like
leave-one-out cross-validation (LOO-CV). For a data set with $N$ observations,
we refit the model $N$ times, each time leaving out one of the $N$ observations
and assessing how well the model predicts the left-out observation. LOO-CV is
very expensive computationally in most realistic settings, but the Pareto
smoothed importance sampling (PSIS, Vehtari et al, 2017ab) algorithm provided by
the *loo* package allows for approximating exact LOO-CV with PSIS-LOO-CV.
PSIS-LOO-CV requires only a single fit of the full model and comes with
diagnostics for assessing the validity of the approximation.

With a time series we can do something similar to LOO-CV but, except in a few
cases, it does not make sense to leave out observations one at a time because
then we are allowing information from the future to influence predictions of the
past (i.e., times $t + 1, t+2, \ldots$ should not be used to predict for time
$t$). To apply the idea of cross-validation to the $M$-SAP case, instead of
leave-*one*-out cross-validation we need some form of leave-*future*-out
cross-validation (LFO-CV). As we will demonstrate in this case study, LFO-CV
does not refer to one particular prediction task but rather to various possible
cross-validation approaches that all involve some form of prediction for new
time series data. Like exact LOO-CV, exact LFO-CV requires refitting the model
many times to different subsets of the data, which is computationally very
costly for most
nontrivial examples, in particular for Bayesian analyses where refitting the
model means estimating a new posterior distribution rather than a point
estimate.

Although PSIS-LOO-CV provides an efficient approximation to exact LOO-CV, until
now there has not been an analogous approximation to exact LFO-CV that
drastically reduces the computational burden while also providing informative
diagnostics about the quality of the approximation. In this case study we
present PSIS-LFO-CV, an algorithm that typically only requires refitting the
time-series model a small number times and will make LFO-CV tractable for many
more realistic applications than previously possible. 

More details can be found in our paper about approximate LFO-CV
(Bürkner, Gabry, & Vehtari, 2019), which is available as a preprint on arXiv (https://arxiv.org/abs/1902.06281).


## $M$-step-ahead predictions

Assume we have a time series of observations 
$y = (y_1, y_2, \ldots, y_N)$ 
and let $L$ be the _minimum_ number of observations from the series that
we will require before making predictions for future data. Depending on the
application and how informative the data is, it may not be possible to make
reasonable predictions for $y_{i+1}$ based on $(y_1, \dots, y_{i})$ until $i$ is
large enough so that we can learn enough about the time series to predict future
observations. Setting $L=10$, for example, means that we will only assess
predictive performance starting with observation $y_{11}$, so that we
always have at least 10 previous observations to condition on.

In order to assess $M$-SAP performance we would like to compute the 
predictive densities

$$
p(y_{i+1:M} \,|\, y_{1:i}) = 
  p(y_{i+1}, \ldots, y_{i + M} \,|\, y_{1},...,y_{i}) 
$$ 

for each $i \in \{L, \ldots, N - M\}$. The quantities 
$p(y_{i+1:M} \,|\, y_{1:i})$ can be computed with the help of the posterior
distribution $p(\theta \,|\, y_{1:i})$ of the parameters $\theta$ conditional on
only the first $i$ observations of the time-series:

$$
p(y_{i+1:M} \,| \, y_{1:i}) = 
  \int p(y_{i+1:M} \,| \, y_{1:i}, \theta) \, p(\theta\,|\,y_{1:i}) \,d\theta. 
$$

Having obtained $S$ draws $(\theta_{1:i}^{(1)}, \ldots, \theta_{1:i}^{(S)})$ 
from the posterior distribution $p(\theta\,|\,y_{1:i})$, we can estimate 
$p(y_{i+1:M} | y_{1:i})$ as

$$
p(y_{i+1:M} \,|\, y_{1:i}) \approx \frac{1}{S}\sum_{s=1}^S p(y_{i+1:M} \,|\, y_{1:i}, \theta_{1:i}^{(s)}).
$$

In the following, we consider factorizable models in which the response values
are conditionally independent given the parameters and the likelihood can be
written in the familiar form

$$
p(y \,|\, \theta) = \prod_{i=1}^N p(y_i \,|\, \theta).
$$

In this case, $p(y_{i+1:M} \,|\, y_{1:i}, \theta)$ reduces to 
$$
p(y_{i+1:M} \,|\, \theta) = \prod_{j = i+1}^{i + M} p(y_j \,|\, \theta),
$$
due to the assumption of conditional independence between $y_{i+1:M}$ and
$y_{1:i}$ given $\theta$.

Non-factorizable models, which do not make this assumption, are discussed in a
separate vignette [_Leave-one-out cross-validation for non-factorizable models_](http://mc-stan.org/loo/articles/loo2-non-factorizable.html).

## Approximate $M$-SAP using importance-sampling {#approximate_MSAP}

Unfortunately, the math above makes use of the posterior distributions from many
different fits of the model to different subsets of the data. That is, to obtain
the predictive density $p(y_{i+1:M} \,|\, y_{1:i})$ requires fitting a model to
only the first $i$ data points, and we will need to do this for every value of
$i$ under consideration (all $i \in \{L, \ldots, N - M\}$).

To reduce the number of models that need to be fit for the purpose of obtaining
each of the densities $p(y_{i+1:M} \,|\, y_{1:i})$, we propose the following
algorithm. Starting with $i = N - M$, we approximate each 
$p(y_{i+1:M} \,|\, y_{1:i})$ using Pareto smoothed importance sampling (PSIS,
Vehtari et al, 2017ab):

$$
 p(y_{i+1:M} \,|\, y_{1:i}) \approx
   \frac{ \sum_{s=1}^S w_i^{(s)}\, p(y_{i+1:M} \,|\, \theta^{(s)})}{ \sum_{s=1}^S w_i^{(s)}},
$$

where $w_i^{(s)}$ are importance weights and $\theta^{(s)}$ are draws from the
posterior distribution based on _all_ observations. To obtain $w_i^{(s)}$, we
first compute the raw importance ratios

$$
r_i^{(s)} \propto \frac{1}{\prod_{j = i}^N p(y_j \,|\, \,\theta^{(s)})},
$$

and then stabilize them using PSIS.

We then _decrease_ $i$ by $1$ (i.e., we move backwards in time) and repeat the
process. At some observation $i$, the variability of importance ratios
$r_i^{(s)}$ will become too large and importance sampling fails. We will refer
to this particular value of $i$ as $i^\star_1$. To identify the value of
$i^\star_1$ we check for which value of $i$ does the estimated shape parameter
$k$ of the generalized Pareto distribution first crosses a certain threshold.
Only then do we refit the model using only observations before $i^\star_1$ and
then restart the process. In some cases we may only need to refit once and in
other cases we will find a value $i^\star_2$ that requires a second refitting,
maybe an $i^\star_3$ that requires a third refitting, and so on. We repeat the
refitting as few times as is required (only if $k > \text{threshold}$) until we
arrive at $i = L$. Recall that $L$ is the minimum number of observations we
have deemed acceptable for making predictions (setting $L=0$ means predictions
of all observations should be computed).

For LOO, we recommend to use a threshold of $0.7$ (Vehtari et al, 2017ab), but
for LFO, the threshold should be smaller as influential observations will affect
all pointwise LFO approximations until the next refit and thus much stronger
than in LOO. For the present case study, we choose a threshold of $0.6$, 
which also seems to be generally reasonable threshold for LFO (Bürkner et al. 2019).


## Autoregressive models

Autoregressive (AR) models are some of the most commonly used time-series models. 
An AR(p) model ---an autoregressive model of order $p$--- can be defined as

$$
y_i = \eta_i + \sum_{k = 1}^p \varphi_k y_{i - k} + \varepsilon_i,
$$

where $\eta_i$ is the linear predictor for the $i$th observation, $\phi_k$ are
the autoregressive parameters and $\varepsilon_i$ are pairwise independent
errors, which are usually assumed to be normally distributed with equal variance
$\sigma^2$. The model implies a recursive formula that allows for computing the
right-hand side of the above equation for observation $i$ based on the values of
the equations for previous observations.

## Case Study: Annual measurements of the level of Lake Huron

To illustrate the application of PSIS-LFO-CV for estimating expected $M$-SAP
performance, we will fit a model for 98 annual measurements of the water level
(in feet) of [Lake Huron](https://en.wikipedia.org/wiki/Lake_Huron) from the
years 1875--1972. This data set is found in the **datasets** R package, which is
installed automatically with **R**. 

In addition to the **loo** package, for this analysis we will use the **brms**
interface to Stan to generate a Stan program and fit the model, and also the
**bayesplot** and **ggplot2** packages for plotting.

```{r pkgs, cache=FALSE}
library("loo")
library("brms")
library("bayesplot")
library("ggplot2")
color_scheme_set("brightblue")
theme_set(theme_default())

CHAINS <- 4
SEED <- 1234
set.seed(SEED)
```

Before fitting a model, we will first put the data into a data frame and then
look at the time series.

```{r hurondata}
N <- length(LakeHuron)
df <- data.frame(
  y = as.numeric(LakeHuron),
  year = as.numeric(time(LakeHuron)),
  time = 1:N
)

# save plot labels to reuse them
plot_labs <- labs(
  y = "Water Level (ft)", 
  x = "Year",
  title = "Water Level in Lake Huron (1875-1972)"
) 

ggplot(df, aes(x = year, y = y)) + 
  geom_point(size = 1) +
  plot_labs
```

The above plot shows rather strong autocorrelation of the time-series as well as
some trend towards lower levels for later points in time. 

We can specify an AR(4) model for these data using the **brms** package as
follows:

```{r fit, results = "hide"}
control <- list(adapt_delta = 0.95)
fit <- brm(
  y ~ 1, 
  data = df, 
  autocor = cor_ar(~time, p = 4), 
  prior = prior(normal(0, 0.5), class = "ar"),
  control = control, 
  seed = SEED, 
  chains = CHAINS
)
```

The model implied predictions along with the observed values can be plotted,
which reveals a rather good fit to the data.

```{r plotpreds, cache = FALSE}
preds <- posterior_predict(fit)
preds <- cbind(
  Estimate = colMeans(preds), 
  Q5 = apply(preds, 2, quantile, probs = 0.05),
  Q95 = apply(preds, 2, quantile, probs = 0.95)
)

ggplot(cbind(df, preds), aes(x = year, y = Estimate)) +
  geom_smooth(aes(ymin = Q5, ymax = Q95), stat = "identity", size = 0.5) +
  geom_point(aes(y = y)) + 
  labs(subtitle = "Mean (blue) and 90% predictive intervals (gray) vs. observed data (black)") +
  plot_labs
```

To allow for reasonable predictions of future values, we will require at least
$L = 20$ historical observations (20 years) to make predictions.

```{r setL}
L <- 20
```

We first perform approximate leave-one-out cross-validation (LOO-CV) for the
purpose of later comparison with exact and approximate LFO-CV for the 1-SAP
case.

```{r loo1sap, cache = FALSE}
loo_cv <- loo(log_lik(fit)[, (L + 1):N])
print(loo_cv)
```

## 1-step-ahead predictions leaving out all future values

The most basic version of $M$-SAP is 1-SAP, in which we predict only one step
ahead. In this case, $y_{i+1:M}$ simplifies to $y_{i}$ and the LFO-CV algorithm
becomes considerably simpler than for larger values of $M$.

### Exact 1-step-ahead predictions

Before we compute approximate LFO-CV using PSIS we will first compute 
exact LFO-CV for the 1-SAP case so we can use it as a benchmark later. 
The initial step for the exact computation is to calculate the log-predictive 
densities by refitting the model many times:

```{r exact_loglik, results="hide"}
loglik_exact <- matrix(nrow = nsamples(fit), ncol = N)
for (i in (N - 1):L) {
  past <- 1:i
  oos <- i + 1
  df_past <- df[past, , drop = FALSE]
  df_oos <- df[c(past, oos), , drop = FALSE]
  fit_i <- update(fit, newdata = df_past, recompile = FALSE)
  loglik_exact[, i + 1] <- log_lik(fit_i, newdata = df_oos, oos = oos)[, oos]
}
```

Then we compute the exact expected log predictive density (ELPD):

```{r helpers}
# some helper functions we'll use throughout

# more stable than log(sum(exp(x))) 
log_sum_exp <- function(x) {
  max_x <- max(x)  
  max_x + log(sum(exp(x - max_x)))
}

# more stable than log(mean(exp(x)))
log_mean_exp <- function(x) {
  log_sum_exp(x) - log(length(x))
}

# compute log of raw importance ratios
# sums over observations *not* over posterior samples
sum_log_ratios <- function(ll, ids = NULL) {
  if (!is.null(ids)) ll <- ll[, ids, drop = FALSE]
  - rowSums(ll)
}

# for printing comparisons later
rbind_print <- function(...) {
  round(rbind(...), digits = 2)
}
```


```{r exact1sap, cache = FALSE}
exact_elpds_1sap <- apply(loglik_exact, 2, log_mean_exp)
exact_elpd_1sap <- c(ELPD = sum(exact_elpds_1sap[-(1:L)]))

rbind_print(
  "LOO" = loo_cv$estimates["elpd_loo", "Estimate"],
  "LFO" = exact_elpd_1sap
)
```

We see that the ELPD from LFO-CV for 1-step-ahead predictions is lower than the
ELPD estimate from LOO-CV, which should be expected since LOO-CV is making use
of more of the time series. That is, since the LFO-CV approach only uses
observations from before the left-out data point but LOO-CV uses _all_ data
points other than the left-out observation, we should expect to see the larger
ELPD from LOO-CV.

### Approximate 1-step-ahead predictions

For illustrative purposes, we will first compute approximate 1-SAP without any
refitting, even when the Pareto $k$ estimate is too large. This will of course
be a poor approximation to exact 1-SAP, in particular if the Pareto $k$
estimates increase rather quickly beyond the threshold up to which PSIS tends to
produce stable results.

```{r loglik}
approx_elpds_1sap_no_refit <- rep(NA, N)
loglik <- log_lik(fit)
ks <- NULL

for (i in (N - 1):L) {
  logratio <- sum_log_ratios(loglik, (i + 1):N)
  psis_obj <- suppressWarnings(psis(logratio))
  ks <- c(ks, pareto_k_values(psis_obj))
  lw <- weights(psis_obj, normalize = TRUE)[, 1]
  approx_elpds_1sap_no_refit[i + 1] <- log_sum_exp(lw + loglik[, i + 1])
}
```

If we plot the Pareto $k$ estimates we can see that they do start to increase 
quickly and that once we work backwards to about halfway through the data points 
we get $k > 0.6$. This is indicated by the vertical line in the plot below, which 
corresponds to the point $i^\star_1$ discussed in Section
[Approximate $M$-SAP using importance-sampling](#approximate_MSAP) above.

```{r plot_ks}
plot_ks <- function(ks, ids, thres = 0.6) {
  dat_ks <- data.frame(ks = ks, ids = ids)
  ggplot(dat_ks, aes(x = ids, y = ks)) + 
    geom_point(aes(color = ks > thres), shape = 3, show.legend = FALSE) + 
    geom_hline(yintercept = thres, linetype = 2, color = "red2") + 
    scale_color_manual(values = c("cornflowerblue", "darkblue")) + 
    labs(x = "Data point", y = "Pareto k") + 
    ylim(-0.5, 1.5)
}
```

```{r plotk1sap, cache = FALSE}
plot_ks(ks, N:(L + 1))
```

Consequently, it is not surprising that the resulting approximate ELPD for 1-SAP 
is far away from the exact ELPD we computed above:

```{r compare1sap, cache = FALSE}
approx_elpd_1sap_no_refit <- 
  c(ELPD = sum(approx_elpds_1sap_no_refit, na.rm = TRUE))
rbind_print(
  "approx LFO (no refit)" = approx_elpd_1sap_no_refit,
  "exact LFO" = exact_elpd_1sap
)
```

Next, we compute approximate 1-SAP with refit at observations where 
the Pareto $k$ estimate exceeds the threshold of $0.6$.

```{r setkthresh}
k_thres <- 0.6
```

The code becomes a little bit more involved to handle the refitting procedure.
Note that we can compute exact 1-SAP at the refitting points, which comes with
no additional computational costs since we had to refit the model anyway.

```{r refit_loglik, results="hide"}
loglik <- matrix(nrow = nsamples(fit), ncol = N)
approx_elpds_1sap <- rep(NA, N)
fit_past <- fit
i_refit <- N
refits <- ks <- NULL

for (i in (N - 1):L) {
  loglik[, i + 1] <- log_lik(fit_past)[, i + 1]
  logratio <- sum_log_ratios(loglik, (i + 1):i_refit)
  psis_obj <- suppressWarnings(psis(logratio))
  k <- pareto_k_values(psis_obj)
  ks <- c(ks, k)
  if (k > k_thres) {
    # refit the model based on the first i observations
    i_refit <- i
    refits <- c(refits, i)
    past <- 1:i
    oos <- i + 1
    df_past <- df[past, , drop = FALSE]
    df_oos <- df[c(past, oos), , drop = FALSE]
    fit_past <- update(fit_past, newdata = df_past, recompile = FALSE)
    loglik[, i + 1] <- log_lik(fit_past, newdata = df_oos)[, oos]
    approx_elpds_1sap[i + 1] <- log_mean_exp(loglik[, i + 1])
  } else {
    lw <- weights(psis_obj, normalize = TRUE)[, 1]
    approx_elpds_1sap[i + 1] <- log_sum_exp(lw + loglik[, i + 1])
  }
} 
```

We see that the final Pareto-$k$-estimates are mostly well below the threshold
and that we only needed to refit the model a few times:

```{r refitsummary1sap, cache=FALSE}
cat("Using threshold ", k_thres, 
    ", model was refit ", length(refits), 
    " times, at observations", refits)

plot_ks(ks, N:(L + 1))
```

The approximate 1-SAP ELPD is remarkably similar to the exact 1-SAP ELPD
computed above, which indicates our algorithm to compute approximate 1-SAP
worked well for the present data and model.

```{r lfosummary1sap, cache = FALSE}
approx_elpd_1sap <- sum(approx_elpds_1sap, na.rm = TRUE)
rbind_print(
  "approx LFO (no refits)" = approx_elpd_1sap_no_refit,
  "approx LFO (with refits)" = approx_elpd_1sap,
  "exact LFO" = exact_elpd_1sap
)
```

Plotting exact against approximate predictions, we see that no approximation
value deviates far from its exact counterpart, providing further evidence for
the good quality of our approximation.

```{r plot1sap, cache = FALSE}
dat_elpd <- data.frame(
  approx_elpd = approx_elpds_1sap,
  exact_elpd = exact_elpds_1sap
)

ggplot(dat_elpd, aes(x = approx_elpd, y = exact_elpd)) +
  geom_abline(color = "gray30") +
  geom_point(size = 2) +
  labs(x = "Approximate ELPDs", y = "Exact ELPDs")
```

We can also look at the maximum difference and average difference between
the approximate and exact ELPD calculations, which also indicate a ver close approximation:

```{r diffs1sap, cache=FALSE}
max_diff <- with(dat_elpd, max(abs(approx_elpd - exact_elpd), na.rm = TRUE))
mean_diff <- with(dat_elpd, mean(abs(approx_elpd - exact_elpd), na.rm = TRUE))

rbind_print(
  "Max diff" = round(max_diff, 2), 
  "Mean diff" =  round(mean_diff, 3)
)
```

## $M$-step-ahead predictions leaving out all future values

To illustrate the application of $M$-SAP for $M > 1$, we next compute exact and
approximate LFO-CV for the 4-SAP case.

### Exact $M$-step-ahead predictions

The necessary steps are the same as for 1-SAP with the exception that the
log-density values of interest are now the sums of the log predictive densities
of four consecutive observations. Further, the stability of the PSIS
approximation actually stays the same for all $M$ as it only depends on the
number of observations we leave out, not on the number of observations we
predict.

```{r exact_loglikm, results="hide"}
M <- 4
loglikm <- matrix(nrow = nsamples(fit), ncol = N)
for (i in (N - M):L) {
  past <- 1:i
  oos <- (i + 1):(i + M)
  df_past <- df[past, , drop = FALSE]
  df_oos <- df[c(past, oos), , drop = FALSE]
  fit_past <- update(fit, newdata = df_past, recompile = FALSE)
  ll <- log_lik(fit_past, newdata = df_oos, oos = oos)
  loglikm[, i + 1] <- rowSums(ll[, oos])
}
```

```{r exact4sap, cache = FALSE}
exact_elpds_4sap <- apply(loglikm, 2, log_mean_exp)
(exact_elpd_4sap <- c(ELPD = sum(exact_elpds_4sap, na.rm = TRUE)))
```

### Approximate $M$-step-ahead predictions

Computing the approximate PSIS-LFO-CV for the 4-SAP case is a little bit more
involved than the approximate version for the 1-SAP case, although the
underlying principles remain the same.

```{r refit_loglikm, results="hide"}
loglikm <- loglik <- matrix(nrow = nsamples(fit), ncol = N)
approx_elpds_4sap <- rep(NA, N)
fit_past <- fit
i_refit <- N
refits <- ks <- NULL

loglik[, (N - M + 2):N] <- log_lik(fit_past)[, (N - M + 2):N]
for (i in (N - M):L) {
  past <- 1:i
  oos <- (i + 1):(i + M)
  df_oos <- df[c(past, oos), , drop = FALSE]
  ll <- log_lik(fit_past, newdata = df_oos, oos = oos)
  loglik[, i + 1] <- ll[, i + 1]
  logratio <- sum_log_ratios(loglik, (i + 1):i_refit)
  psis_obj <- suppressWarnings(psis(logratio))
  k <- pareto_k_values(psis_obj)
  ks <- c(ks, k)
  if (k > k_thres) {
    # refit the model based on the first i observations
    i_refit <- i
    refits <- c(refits, i)
    df_past <- df[past, , drop = FALSE]
    fit_past <- update(fit, newdata = df_past, recompile = FALSE)
    ll <- log_lik(fit_past, newdata = df_oos, oos = oos)
    loglikm[, i + 1] <- rowSums(ll[, oos])
    loglik[, i + 1] <- ll[, i + 1]
    approx_elpds_4sap[i + 1] <- log_mean_exp(loglikm[, i + 1])
  } else {
    lw <- weights(psis_obj, normalize = TRUE)[, 1]
    loglikm[, i + 1] <- rowSums(ll[, oos])
    approx_elpds_4sap[i + 1] <- log_sum_exp(lw + loglikm[, i + 1])
  }
} 
```

Again, we see that the final Pareto-$k$-estimates are mostly well below the
threshold and that we only needed to refit the model a few times: 

```{r refitsummary4sap, cache = FALSE}
cat("Using threshold ", k_thres, 
    ", model was refit ", length(refits), 
    " times, at observations", refits)

plot_ks(ks, (N - M + 1):(L + 1))
```

The approximate ELPD computed for the 4-SAP case is not as close to its exact
counterpart as in the 1-SAP case.

```{r lfosummary4sap, cache = FALSE}
approx_elpd_4sap <- sum(approx_elpds_4sap, na.rm = TRUE)
rbind_print(
  "Approx LFO" = approx_elpd_4sap,
  "Exact LFO" = exact_elpd_4sap
)
```

Plotting exact against approximate pointwise predictions confirms that, for a 
few specific data points, the approximate predictions underestimate the exact
predictions.

```{r plot4sap, cache = FALSE}
dat_elpd_4sap <- data.frame(
  approx_elpd = approx_elpds_4sap,
  exact_elpd = exact_elpds_4sap
)

ggplot(dat_elpd_4sap, aes(x = approx_elpd, y = exact_elpd)) +
  geom_abline(color = "gray30") +
  geom_point(size = 2) +
  labs(x = "Approximate ELPDs", y = "Exact ELPDs")
```

## Conclusion

In this case study we have shown how to do carry out exact and approximate
leave-future-out cross-validation for $M$-step-ahead prediction tasks. For the
data and model used in our example, the PSIS-LFO-CV algorithm provides reasonably
stable and accurate results despite not requiring us to refit the model nearly as
many times. For more details on approximate LFO-CV, we refer to Bürkner et al. (2019).

<br />

## References

Bürkner P. C., Gabry J., & Vehtari A. (2019). Approximate leave-future-out cross-validation for time series models. [arXiv preprint](https://arxiv.org/abs/1902.06281). 

Vehtari A., Gelman A., & Gabry J. (2017a). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. *Statistics and Computing*, 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. [Online](http://link.springer.com/article/10.1007/s11222-016-9696-4). [arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

Vehtari A., Gelman A., & Gabry J. (2017b). Pareto smoothed importance sampling. [arXiv preprint arXiv:1507.02646](https://arxiv.org/abs/1507.02646).

<br />

## Appendix

### Appendix: Session information

```{r sessioninfo}
sessionInfo()
```

### Appendix: Licenses

* Code &copy; 2018, Paul Bürkner, Jonah Gabry, Aki Vehtari (licensed under BSD-3).
* Text &copy; 2018, Paul Bürkner, Jonah Gabry, Aki Vehtari (licensed under CC-BY-NC 4.0).

---
title: "Holdout validation and K-fold cross-validation of Stan programs with the loo package" 
author: "Bruno Nicenboim"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Holdout validation and K-fold cross-validation of Stan programs with the loo package}
-->
```{r, child="children/SETTINGS-knitr.txt"}
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

This vignette demonstrates how to do holdout validation and K-fold cross-validation with __loo__ for a Stan program.


# Example: Eradication of Roaches using holdout validation approach

This vignette uses the same example as in the vignettes
[_Using the loo package (version >= 2.0.0)_](http://mc-stan.org/loo/articles/loo2-example.html) and [_Avoiding model refits in leave-one-out cross-validation with moment matching_](https://mc-stan.org/loo/articles/loo2-moment-matching.html).



## Coding the Stan model

Here is the Stan code for fitting a Poisson regression model:

```{r stancode}
# Note: some syntax used in this Stan program requires RStan >= 2.26 (or CmdStanR)
# To use an older version of RStan change the line declaring `y` to: int y[N];
stancode <- "
data {
  int<lower=1> K;
  int<lower=1> N;
  matrix[N,K] x;
  array[N] int y;
  vector[N] offset_; // offset is reserved keyword in Stan so use offset_

  real beta_prior_scale;
  real alpha_prior_scale;
}
parameters {
  vector[K] beta;
  real intercept;
}
model {
  y ~ poisson(exp(x * beta + intercept + offset_));
  beta ~ normal(0,beta_prior_scale);
  intercept ~ normal(0,alpha_prior_scale);
}
generated quantities {
  vector[N] log_lik;
  for (n in 1:N)
    log_lik[n] = poisson_lpmf(y[n] | exp(x[n] * beta + intercept + offset_[n]));
}
"
```

Following the usual approach recommended in 
[_Writing Stan programs for use with the loo package_](http://mc-stan.org/loo/articles/loo2-with-rstan.html), 
we compute the log-likelihood for each observation in the 
`generated quantities` block of the Stan program.


## Setup

In addition to __loo__, we load the __rstan__ package for fitting the model.
We will also need the __rstanarm__ package for the data.

```{r setup, message=FALSE}
library("rstan")
library("loo")
seed <- 9547
set.seed(seed)
```

# Holdout validation

For this approach, the model is first fit to the "train" data and then is evaluated on the held-out "test" data.

## Splitting the data between train and test 

The data is divided between train (80% of the data) and test (20%):

```{r modelfit-holdout, message=FALSE}
# Prepare data
data(roaches, package = "rstanarm")
roaches$roach1 <- sqrt(roaches$roach1)
roaches$offset <- log(roaches[,"exposure2"])
# 20% of the data goes to the test set:
roaches$test <- 0
roaches$test[sample(.2 * seq_len(nrow(roaches)))] <- 1
# data to "train" the model
data_train <- list(y = roaches$y[roaches$test == 0],
                   x = as.matrix(roaches[roaches$test == 0,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$test == 0,]),
                   K = 3,
                   offset_ = roaches$offset[roaches$test == 0],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )
# data to "test" the model
data_test <- list(y = roaches$y[roaches$test == 1],
                   x = as.matrix(roaches[roaches$test == 1,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$test == 1,]),
                   K = 3,
                   offset_ = roaches$offset[roaches$test == 1],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )
```


## Fitting the model with RStan

Next we fit the model to the "test" data in Stan using the __rstan__ package:

```{r fit-train}
# Compile
stanmodel <- stan_model(model_code = stancode)
# Fit model
fit <- sampling(stanmodel, data = data_train, seed = seed, refresh = 0)
```

We recompute the generated quantities using the posterior draws conditional on the training data, but we now pass in the held-out data to get the log predictive densities for the test data. Because we are using independent data, the log predictive density coincides with the log likelihood of the test data.

```{r gen-test}
gen_test <- gqs(stanmodel, draws = as.matrix(fit), data= data_test)
log_pd <- extract_log_lik(gen_test)
```

## Computing holdout elpd:

Now we evaluate the predictive performance of the model on the test data using `elpd()`.

```{r elpd-holdout}
(elpd_holdout <- elpd(log_pd))
```

When one wants to compare different models, the function `loo_compare()` can be used to assess the difference in performance.

# K-fold cross validation

For this approach the data is divided into folds, and each time one fold is tested while the rest of the data is used to fit the model (see Vehtari et al., 2017).

## Splitting the data in folds

We use the data that is already pre-processed and we divide it in 10 random folds using `kfold_split_random`

```{r prepare-folds, message=FALSE}
# Prepare data
roaches$fold <- kfold_split_random(K = 10, N = nrow(roaches))
```


## Fitting and extracting the log pointwise predictive densities for each fold

We now loop over the 10 folds. In each fold we do the following. First, we fit the model to all the observations except the ones belonging to the left-out fold. Second, we compute the log pointwise predictive densities for the left-out fold. Last, we store the predictive density for the observations of the left-out fold in a matrix. The output of this loop is a matrix of the log pointwise predictive densities of all the observations.

```{r}
# Prepare a matrix with the number of post-warmup iterations by number of observations:
log_pd_kfold <- matrix(nrow = 4000, ncol = nrow(roaches))
# Loop over the folds
for(k in 1:10){
  data_train <- list(y = roaches$y[roaches$fold != k],
                   x = as.matrix(roaches[roaches$fold != k,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$fold != k,]),
                   K = 3,
                   offset_ = roaches$offset[roaches$fold != k],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )
  data_test <- list(y = roaches$y[roaches$fold == k],
                   x = as.matrix(roaches[roaches$fold == k,
                                         c("roach1", "treatment", "senior")]),
                   N = nrow(roaches[roaches$fold == k,]),
                   K = 3,
                   offset_ = roaches$offset[roaches$fold == k],
                   beta_prior_scale = 2.5,
                   alpha_prior_scale = 5.0
                   )
  fit <- sampling(stanmodel, data = data_train, seed = seed, refresh = 0)
  gen_test <- gqs(stanmodel, draws = as.matrix(fit), data= data_test)
  log_pd_kfold[, roaches$fold == k] <- extract_log_lik(gen_test)
}
```

## Computing K-fold elpd:

Now we evaluate the predictive performance of the model on the 10 folds using `elpd()`.

```{r elpd-kfold}
(elpd_kfold <- elpd(log_pd_kfold))
```

If one wants to compare several models (with `loo_compare`), one should use the same folds for all the different models. 

# References

Gelman, A., and Hill, J. (2007).  *Data Analysis Using Regression and Multilevel Hierarchical Models.*  Cambridge University Press.

Stan Development Team (2020) _RStan: the R interface to Stan, Version 2.21.1_   https://mc-stan.org

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. _Statistics and Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. Links: [published](https://link.springer.com/article/10.1007/s11222-016-9696-4) | [arXiv preprint](https://arxiv.org/abs/1507.04544).

---
title: "Using the loo package (version >= 2.0.0)"
author: "Aki Vehtari and Jonah Gabry"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Using the loo package}
-->
```{r, child="children/SETTINGS-knitr.txt"}
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

This vignette demonstrates how to use the __loo__ package to carry out 
Pareto smoothed importance-sampling leave-one-out cross-validation (PSIS-LOO) 
for purposes of model checking and model comparison. 

In this vignette we can't provide all necessary background information on
PSIS-LOO and its diagnostics (Pareto $k$ and effective sample size), so we
encourage readers to refer to the following papers for more details:

* Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. _Statistics and Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. Links: [published](https://link.springer.com/article/10.1007/s11222-016-9696-4) | [preprint arXiv](https://arxiv.org/abs/1507.04544).

* Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)
 

# Setup

In addition to the __loo__ package, we'll also be using __rstanarm__ and 
__bayesplot__:

```{r setup, message=FALSE}
library("rstanarm")
library("bayesplot")
library("loo")
```


# Example: Poisson vs negative binomial for the roaches dataset

## Background and model fitting 

The Poisson and negative binomial regression models used below in our example, 
as well as the `stan_glm` function used to fit the models,
are covered in more depth in the __rstanarm__ vignette 
[_Estimating Generalized Linear Models for Count Data with rstanarm_](http://mc-stan.org/rstanarm/articles/count.html). In the rest of 
this vignette we will assume the reader is already familiar with these kinds of
models.


### Roaches data

The example data we'll use comes from Chapter 8.3 of [Gelman and Hill
(2007)](http://www.stat.columbia.edu/~gelman/arm/). We want to make inferences
about the efficacy of a certain pest management system at reducing the number of
roaches in urban apartments. Here is how Gelman and Hill describe the experiment
and data (pg. 161):

> the treatment and control were applied to 160 and 104 apartments, respectively, and the outcome measurement $y_i$ in each apartment $i$ was the number of roaches caught in a set of traps. Different apartments had traps for different numbers of days

In addition to an intercept, the regression predictors for the model are
`roach1`, the pre-treatment number of roaches (rescaled above to be in units of
hundreds), the treatment indicator `treatment`, and a variable indicating
whether the apartment is in a building restricted to elderly residents `senior`.
Because the number of days for which the roach traps were used is not the same
for all apartments in the sample, we use the `offset` argument to specify that
`log(exposure2)` should be added to the linear predictor.

```{r data}
# the 'roaches' data frame is included with the rstanarm package
data(roaches)
str(roaches)

# rescale to units of hundreds of roaches
roaches$roach1 <- roaches$roach1 / 100
```

### Fit Poisson model

We'll fit a simple Poisson regression model using the `stan_glm` function from
the __rstanarm__ package.

```{r count-roaches-mcmc, results="hide"}
fit1 <-
  stan_glm(
    formula = y ~ roach1 + treatment + senior,
    offset = log(exposure2),
    data = roaches,
    family = poisson(link = "log"),
    prior = normal(0, 2.5, autoscale = TRUE),
    prior_intercept = normal(0, 5, autoscale = TRUE),
    seed = 12345
  )
```

Usually we would also run posterior predictive checks as shown in the 
__rstanarm__ vignette 
[Estimating Generalized Linear Models for Count Data with rstanarm](http://mc-stan.org/rstanarm/articles/count.html), but 
here we focus only on methods provided by the __loo__ package.

<br>

## Using the __loo__ package for model checking and comparison

_Although cross-validation is mostly used for model comparison, it is also useful for model checking._

### Computing PSIS-LOO and checking diagnostics

We start by computing PSIS-LOO with the `loo` function. Since we fit our model
using __rstanarm__ we can use the `loo` method for `stanreg` objects (fitted
model objects from __rstanarm__), which doesn't require us to first extract the
pointwise log-likelihood values. If we had written our own Stan program instead
of using __rstanarm__ we would pass an array or matrix of log-likelihood values
to the `loo` function (see, e.g. `help("loo.array", package = "loo")`). We'll
also use the argument `save_psis = TRUE` to save some intermediate results to be
re-used later.

```{r loo1}
loo1 <- loo(fit1, save_psis = TRUE)
```

`loo` gives us warnings about the Pareto diagnostics, which indicate that for
some observations the leave-one-out posteriors are different enough from the
full posterior that importance-sampling is not able to correct the difference.
We can see more details by printing the `loo` object.

```{r print-loo1}
print(loo1)
```

The table shows us a summary of Pareto $k$ diagnostic, which is used to assess
the reliability of the estimates. In addition to the proportion of leave-one-out
folds with $k$ values in different intervals, the minimum of the effective
sample sizes in that category is shown to give idea why higher $k$ values are
bad. Since we have some $k>1$, we are not able to compute an estimate for the
Monte Carlo standard error (SE) of the expected log predictive density
(`elpd_loo`) and `NA` is displayed. (Full details on the interpretation of 
the Pareto $k$ diagnostics are available in the Vehtari, Gelman, and Gabry
(2017) and Vehtari, Simpson, Gelman, Yao, and Gabry (2024) papers referenced
at the top of this vignette.)

In this case the `elpd_loo` estimate should not be considered reliable. If we
had a well-specified model we would expect the estimated effective number
of parameters (`p_loo`) to be smaller than or similar to the total number of
parameters in the model. Here `p_loo` is almost 300, which is about 70 times
the total number of parameters in the model, indicating severe model
misspecification.

### Plotting Pareto $k$ diagnostics

Using the `plot` method on our `loo1` object produces a plot of the $k$ values
(in the same order as the observations in the dataset used to fit the model) 
with horizontal lines corresponding to the same categories as in the
printed output above.

```{r plot-loo1, out.width = "70%"}
plot(loo1)
```

This plot is useful to quickly see the distribution of $k$ values, but it's
often also possible to see structure with respect to data ordering. In our case
this is mild, but there seems to be a block of data that is somewhat easier to
predict (indices around 90--150). Unfortunately even for these data points we
see some high $k$ values.

### Marginal posterior predictive checks

The `loo` package can be used in combination with the `bayesplot` package for
leave-one-out cross-validation marginal posterior predictive checks [Gabry et al
(2018)](https://arxiv.org/abs/1709.01449). LOO-PIT values are cumulative
probabilities for $y_i$ computed using the LOO marginal predictive distributions
$p(y_i|y_{-i})$. For a good model, the distribution of LOO-PIT values should be
uniform. In the following QQ-plot the LOO-PIT values for our model (y-axi) is
compared to standard uniform distribution (x-axis).

```{r ppc_loo_pit_overlay}
yrep <- posterior_predict(fit1)

ppc_loo_pit_qq(
  y = roaches$y,
  yrep = yrep,
  lw = weights(loo1$psis_object)
)
```

The excessive number of LOO-PIT values close to 0 indicates that the model is
under-dispersed compared to the data, and we should consider a model that allows
for greater dispersion.

## Try alternative model with more flexibility

Here we will try [negative binomial](https://en.wikipedia.org/wiki/Negative_binomial_distribution)
regression, which is commonly used for overdispersed count data.  
Unlike the Poisson distribution, the negative binomial distribution 
allows the conditional mean and variance of $y$ to differ.

```{r count-roaches-negbin, results="hide"}
fit2 <- update(fit1, family = neg_binomial_2)
```

```{r loo2}
loo2 <- loo(fit2, save_psis = TRUE, cores = 2)
print(loo2)
```


```{r plot-loo2}
plot(loo2, label_points = TRUE)
```

Using the `label_points` argument will label any $k$ values larger than the
diagnostic threshold with
the index of the corresponding data point. These high values are often the
result of model misspecification and frequently correspond to data points that
would be considered ``outliers'' in the data and surprising according to the
model [Gabry et al (2019)](https://arxiv.org/abs/1709.01449). Unfortunately,
while large $k$ values are a useful indicator of model misspecification, small
$k$ values are not a guarantee that a model is well-specified.

If there are a small number of problematic $k$
values then we can use a feature in __rstanarm__ that lets us refit the model
once for each of these problematic observations. Each time the model is refit,
one of the observations with a high $k$ value is omitted and the LOO
calculations are performed exactly for that observation. The results are then
recombined with the approximate LOO calculations already carried out for the
observations without problematic $k$ values:

```{r reloo}
if (any(pareto_k_values(loo2) > 0.7)) {
  loo2 <- loo(fit2, save_psis = TRUE, k_threshold = 0.7)
}
print(loo2)
```

In the print output we can see that the Monte Carlo SE is small compared to
the other uncertainties. 

On the other hand, `p_loo` is about 7 and still a bit higher than the total
number of parameters in the model. This indicates that there is almost certainly
still some degree of model misspecification, but this is much better than the
`p_loo` estimate for the Poisson model.

For further model checking we again examine the LOO-PIT values.
```{r ppc_loo_pit_overlay-negbin}
yrep <- posterior_predict(fit2)
ppc_loo_pit_qq(roaches$y, yrep, lw = weights(loo2$psis_object))
```

The plot for the negative binomial model looks better than the Poisson plot, but
we still see that this model is not capturing all of the essential features in
the data.


## Comparing the models on expected log predictive density

We can use the `loo_compare` function to compare our two models on 
expected log predictive density (ELPD) for new data:

```{r loo_compare}
loo_compare(loo1, loo2)
```

The difference in ELPD is much larger than several times the estimated standard
error of the difference again indicating that the negative-binomial model is
xpected to have better predictive performance than the Poisson model. However,
according to the LOO-PIT checks there is still some misspecification, and a
reasonable guess is that a hurdle or zero-inflated model would be an improvement
(we leave that for another case study).


<br>

# References

Gabry, J., Simpson, D., Vehtari, A., Betancourt, M. and Gelman, A. (2019),
Visualization in Bayesian workflow. _J. R. Stat. Soc. A_, 182: 389-402. 
\doi:10.1111/rssa.12378. ([journal version](https://rss.onlinelibrary.wiley.com/doi/full/10.1111/rssa.12378), 
[arXiv preprint](https://arxiv.org/abs/1709.01449), 
[code on GitHub](https://github.com/jgabry/bayes-vis-paper))
<a id="gabry2019"></a>

Gelman, A. and Hill, J. (2007). _Data Analysis Using Regression and
Multilevel/Hierarchical Models._ Cambridge University Press, Cambridge, UK.

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC. _Statistics and
Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4.
[online](https://link.springer.com/article/10.1007/s11222-016-9696-4), 
[arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

---
title: "Writing Stan programs for use with the loo package"
author: "Aki Vehtari and Jonah Gabry"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Writing Stan programs for use with the loo package}
-->

```{r settings, child="children/SETTINGS-knitr.txt"}
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

This vignette demonstrates how to write a Stan program that computes and stores
the pointwise log-likelihood required for using the __loo__ package. The other
vignettes included with the package demonstrate additional functionality.

Some sections from this vignette are excerpted from our papers

* Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. _Statistics and Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. Links: [published](https://link.springer.com/article/10.1007/s11222-016-9696-4) | [arXiv preprint](https://arxiv.org/abs/1507.04544).

* Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2019). Pareto smoothed importance sampling. [arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

which provide important background for understanding the methods implemented in
the package.


# Example: Well water in Bangladesh

This example comes from a survey of residents from a small area in Bangladesh
that was affected by arsenic in drinking water. Respondents with elevated
arsenic levels in their wells were asked if they were interested in getting
water from a neighbor's well, and a series of logistic regressions were fit to
predict this binary response given various information about the households
(Gelman and Hill, 2007). Here we fit a model for the well-switching response
given two predictors: the arsenic level of the water in the resident's home, and
the distance of the house from the nearest safe well.

The sample size in this example is $N=3020$, which is not huge but is large
enough that it is important to have a computational method for LOO that is fast
for each data point. On the plus side, with such a large dataset, the influence
of any given observation is small, and so the computations should be stable.

## Coding the Stan model

Here is the Stan code for fitting the logistic regression model, which 
we save in a file called `logistic.stan`:

```
data {
  int<lower=0> N;             // number of data points
  int<lower=0> P;             // number of predictors (including intercept)
  matrix[N,P] X;              // predictors (including 1s for intercept)
  int<lower=0,upper=1> y[N];  // binary outcome
}
parameters {
  vector[P] beta;
}
model {
  beta ~ normal(0, 1);
  y ~ bernoulli_logit(X * beta);
}
generated quantities {
  vector[N] log_lik;
  for (n in 1:N) {
    log_lik[n] = bernoulli_logit_lpmf(y[n] | X[n] * beta);
  }
}
```

We have defined the log likelihood as a vector named `log_lik` in the generated 
quantities block so that the individual terms will be saved by Stan. After 
running Stan, `log_lik` can be extracted (using the `extract_log_lik` function
provided in the **loo** package) as an $S \times N$ matrix, where $S$ is the 
number of simulations (posterior draws) and $N$ is the number of data points.

## Fitting the model with RStan

Next we fit the model in Stan using the **rstan** package:

```{r, eval=FALSE}
library("rstan")

# Prepare data 
url <- "http://stat.columbia.edu/~gelman/arm/examples/arsenic/wells.dat"
wells <- read.table(url)
wells$dist100 <- with(wells, dist / 100)
X <- model.matrix(~ dist100 + arsenic, wells)
standata <- list(y = wells$switch, X = X, N = nrow(X), P = ncol(X))

# Fit model
fit_1 <- stan("logistic.stan", data = standata)
print(fit_1, pars = "beta")
```

```
         mean se_mean   sd  2.5%   25%   50%   75% 97.5% n_eff Rhat
beta[1]  0.00       0 0.08 -0.16 -0.05  0.00  0.05  0.15  1964    1
beta[2] -0.89       0 0.10 -1.09 -0.96 -0.89 -0.82 -0.68  2048    1
beta[3]  0.46       0 0.04  0.38  0.43  0.46  0.49  0.54  2198    1
```

## Computing approximate leave-one-out cross-validation using PSIS-LOO

We can then use the **loo** package to compute the efficient PSIS-LOO
approximation to exact LOO-CV:

```{r, eval=FALSE}
library("loo")

# Extract pointwise log-likelihood
# using merge_chains=FALSE returns an array, which is easier to 
# use with relative_eff()
log_lik_1 <- extract_log_lik(fit_1, merge_chains = FALSE)

# as of loo v2.0.0 we can optionally provide relative effective sample sizes
# when calling loo, which allows for better estimates of the PSIS effective
# sample sizes and Monte Carlo error
r_eff <- relative_eff(exp(log_lik_1), cores = 2) 

# preferably use more than 2 cores (as many cores as possible)
# will use value of 'mc.cores' option if cores is not specified
loo_1 <- loo(log_lik_1, r_eff = r_eff, cores = 2)
print(loo_1)
```

```
Computed from 4000 by 3020 log-likelihood matrix

         Estimate   SE
elpd_loo  -1968.5 15.6
p_loo         3.2  0.1
looic      3937.0 31.2
------
Monte Carlo SE of elpd_loo is 0.0.

All Pareto k estimates are good (k < 0.5).
See help('pareto-k-diagnostic') for details.
```

The printed output from the `loo` function shows the estimates
$\widehat{\mbox{elpd}}_{\rm loo}$ (expected log predictive density),
$\widehat{p}_{\rm loo}$ (effective number of parameters), and ${\rm looic} =-2\,
\widehat{\mbox{elpd}}_{\rm loo}$ (the LOO information criterion).

The line at the bottom of the printed output provides information about the
reliability of the LOO approximation (the interpretation of the $k$ parameter is
explained in `help('pareto-k-diagnostic')` and in greater detail
in Vehtari, Simpson, Gelman, Yao, and Gabry (2019)). In this case the message tells us that
all of the estimates for $k$ are fine.

## Comparing models

To compare this model to an alternative model for the same data we can use the 
`loo_compare` function in the **loo** package. First we'll fit a second model to the
well-switching data, using `log(arsenic)` instead of `arsenic` as a predictor:

```{r, eval=FALSE}
standata$X[, "arsenic"] <- log(standata$X[, "arsenic"])
fit_2 <- stan(fit = fit_1, data = standata) 

log_lik_2 <- extract_log_lik(fit_2, merge_chains = FALSE)
r_eff_2 <- relative_eff(exp(log_lik_2))
loo_2 <- loo(log_lik_2, r_eff = r_eff_2, cores = 2)
print(loo_2)
```

```
Computed from 4000 by 3020 log-likelihood matrix

         Estimate   SE
elpd_loo  -1952.3 16.2
p_loo         3.1  0.1
looic      3904.6 32.4
------
Monte Carlo SE of elpd_loo is 0.0.

All Pareto k estimates are good (k < 0.5).
See help('pareto-k-diagnostic') for details.
```

We can now compare the models on LOO using the `loo_compare` function:

```{r, eval=FALSE}
# Compare
comp <- loo_compare(loo_1, loo_2)
```

This new object, `comp`, contains the estimated difference of expected 
leave-one-out prediction errors between the two models, along with the standard 
error:

```{r, eval=FALSE}
print(comp) # can set simplify=FALSE for more detailed print output
```
```
       elpd_diff se_diff
model2   0.0       0.0  
model1 -16.3       4.4  
```

The first column shows the difference in ELPD relative to the model with 
the largest ELPD. In this case, the difference in `elpd` and its scale relative 
to the approximate standard error of the difference) indicates a 
preference for the second model (`model2`).


# References

Gelman, A., and Hill, J. (2007).  *Data Analysis Using Regression and Multilevel Hierarchical Models.*  Cambridge University Press.

Stan Development Team (2017). _The Stan C++ Library, Version 2.17.0._   https://mc-stan.org/

Stan Development Team (2018) _RStan: the R interface to Stan, Version 2.17.3._   https://mc-stan.org/

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC. _Statistics and
Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4.
[online](https://link.springer.com/article/10.1007/s11222-016-9696-4), 
[arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2019). Pareto
smoothed importance sampling.
[arXiv preprint arXiv:1507.02646](https://arxiv.org/abs/1507.02646).

---
title: "Avoiding model refits in leave-one-out cross-validation with moment matching"
author: "Topi Paananen, Paul Bürkner, Aki Vehtari and Jonah Gabry"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Avoiding model refits in leave-one-out cross-validation with moment matching}
-->
```{r, child="children/SETTINGS-knitr.txt"}
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

This vignette demonstrates how to improve the Monte Carlo sampling accuracy of
leave-one-out cross-validation with the __loo__ package and Stan. The __loo__
package automatically monitors the sampling accuracy using Pareto $k$
diagnostics for each observation. Here, we present a method for quickly
improving the accuracy when the Pareto diagnostics indicate problems. This is
done by performing some additional computations using the existing posterior
sample. If successful, this will decrease the Pareto $k$ values, making the
model assessment more reliable. __loo__ also stores the original Pareto $k$
values with the name `influence_pareto_k` which are not changed. They can be
used as a diagnostic of how much each observation influences the posterior
distribution. 

The methodology presented is based on the paper

* Paananen, T., Piironen, J., Buerkner, P.-C., Vehtari, A. (2020).
Implicitly Adaptive Importance Sampling.
[arXiv preprint arXiv:1906.08850](https://arxiv.org/abs/1906.08850).

More information about the Pareto $k$ diagnostics is given in the following
papers

* Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. _Statistics and Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. Links: [published](https://link.springer.com/article/10.1007/s11222-016-9696-4) | [arXiv preprint](https://arxiv.org/abs/1507.04544).

* Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

# Example: Eradication of Roaches

We will use the same example as in the vignette
[_Using the loo package (version >= 2.0.0)_](https://mc-stan.org/loo/articles/loo2-example.html).
See the demo for a description of the problem and data. We will use the same Poisson
regression model as in the case study.

## Coding the Stan model

Here is the Stan code for fitting the Poisson regression model, which 
we will use for modeling the number of roaches.

```{r stancode}
# Note: some syntax used in this Stan program requires RStan >= 2.26 (or CmdStanR)
# To use an older version of RStan change the line declaring `y` to: int y[N];
stancode <- "
data {
  int<lower=1> K;
  int<lower=1> N;
  matrix[N,K] x;
  array[N] int y;
  vector[N] offset;

  real beta_prior_scale;
  real alpha_prior_scale;
}
parameters {
  vector[K] beta;
  real intercept;
}
model {
  y ~ poisson(exp(x * beta + intercept + offset));
  beta ~ normal(0,beta_prior_scale);
  intercept ~ normal(0,alpha_prior_scale);
}
generated quantities {
  vector[N] log_lik;
  for (n in 1:N)
    log_lik[n] = poisson_lpmf(y[n] | exp(x[n] * beta + intercept + offset[n]));
}
"
```

Following the usual approach recommended in 
[_Writing Stan programs for use with the loo package_](http://mc-stan.org/loo/articles/loo2-with-rstan.html), 
we compute the log-likelihood for each observation in the 
`generated quantities` block of the Stan program.


## Setup

In addition to __loo__, we load the __rstan__ package for fitting the model,
and the __rstanarm__ package for the data.

```{r setup, message=FALSE}
library("rstan")
library("loo")
seed <- 9547
set.seed(seed)
```

## Fitting the model with RStan

Next we fit the model in Stan using the __rstan__ package:

```{r modelfit, message=FALSE}
# Prepare data
data(roaches, package = "rstanarm")
roaches$roach1 <- sqrt(roaches$roach1)
y <- roaches$y
x <- roaches[,c("roach1", "treatment", "senior")]
offset <- log(roaches[,"exposure2"])
n <- dim(x)[1]
k <- dim(x)[2]

standata <- list(N = n, K = k, x = as.matrix(x), y = y, offset = offset, beta_prior_scale = 2.5, alpha_prior_scale = 5.0)

# Compile
stanmodel <- stan_model(model_code = stancode)

# Fit model
fit <- sampling(stanmodel, data = standata, seed = seed, refresh = 0)
print(fit, pars = "beta")
```

Let us now evaluate the predictive performance of the model using `loo()`.

```{r loo1}
loo1 <- loo(fit)
loo1
```

The `loo()` function output warnings that there are some observations which are
highly influential, and thus the accuracy of importance sampling is compromised
as indicated by the large Pareto $k$ diagnostic values (> 0.7). As discussed in
the vignette
[_Using the loo package (version >= 2.0.0)_](https://mc-stan.org/loo/articles/loo2-example.html),
this may be an indication of model misspecification. Despite that, it is still
beneficial to be able to evaluate the predictive performance of the model
accurately.

## Moment matching correction for importance sampling

To improve the accuracy of the `loo()` result above, we could perform
leave-one-out cross-validation by explicitly leaving out single observations and
refitting the model using MCMC repeatedly. However, the Pareto $k$ diagnostics
indicate that there are 19 observations which are problematic. This would
require 19 model refits which may require a lot of computation time.

Instead of refitting with MCMC, we can perform a faster moment matching
correction to the importance sampling for the problematic observations. This can
be done with the `loo_moment_match()` function in the __loo__ package, which
takes our existing `loo` object as input and modifies it. The moment matching
requires some evaluations of the model posterior density. For models fitted with
__rstan__, this can be conveniently done by using the existing `stanfit` object.

First, we show how the moment matching can be used for a model fitted using
__rstan__. It only requires setting the argument `moment_match` to `TRUE` in the
`loo()` function. Optionally, you can also set the argument `k_threshold` which
determines the Pareto $k$ threshold, above which moment matching is used. By
default, it operates on all observations whose Pareto $k$ value is larger than
the sample size ($S$) specific threshold $\min(1 - 1 / \log_{10}(S), 0.7)$ (which is $0.7$ for $S>2200$).

```{r loo_moment_match}
# available in rstan >= 2.21
loo2 <- loo(fit, moment_match = TRUE)
loo2
```

After the moment matching, all observations have the diagnostic
Pareto $k$ less than 0.7, meaning that the estimates are now reliable.
The total `elpd_loo` estimate also changed from `-5457.8` to `-5478.5`, showing
that before moment matching, `loo()` overestimated the predictive
performance of the model.

The updated Pareto $k$ values stored in
`loo2$diagnostics$pareto_k` are considered algorithmic diagnostic values that
indicate the sampling accuracy. The original Pareto $k$ values are
stored in `loo2$pointwise[,"influence_pareto_k"]` and these are not modified
by the moment matching. These can be considered as diagnostics
for how big influence each observation has on the posterior distribution.
In addition to the Pareto $k$ diagnostics, moment matching also updates
the effective sample size estimates. 

# Using `loo_moment_match()` directly

The moment matching can also be performed by explicitly calling the function
`loo_moment_match()`. This enables its use also for models that are not using
__rstan__ or another package with built-in support for `loo_moment_match()`. To
use `loo_moment_match()`, the user must give the model object `x`, the `loo`
object, and 5 helper functions as arguments to `loo_moment_match()`. The helper
functions are

* `post_draws`	
    + A function the takes `x` as the first argument and returns a matrix of
    posterior draws of the model parameters, `pars`.
* `log_lik_i`	
    + A function that takes `x` and `i` and returns a matrix
    (one column per chain) or a vector (all chains stacked) of log-likeliood
    draws of the ith observation based on the model `x`. If the draws are
    obtained using MCMC, the matrix with MCMC chains separated is preferred.
* `unconstrain_pars`	
    + A function that takes arguments `x` and `pars`, and returns posterior
    draws on the unconstrained space based on the posterior draws on the
    constrained space passed via `pars`.
* `log_prob_upars`	
    + A function that takes arguments `x` and `upars`, and returns a matrix of
    log-posterior density values of the unconstrained posterior draws passed
    via `upars`.
* `log_lik_i_upars`	
    + A function that takes arguments `x`, `upars`, and `i` and returns a
    vector of log-likelihood draws of the `i`th observation based on the
    unconstrained posterior draws passed via `upars`.

Next, we show how the helper functions look like for RStan objects, and show an
example of using `loo_moment_match()` directly. For stanfit objects from 
__rstan__ objects, the functions look like this:

```{r stanfitfuns}
# create a named list of draws for use with rstan methods
.rstan_relist <- function(x, skeleton) {
  out <- utils::relist(x, skeleton)
  for (i in seq_along(skeleton)) {
    dim(out[[i]]) <- dim(skeleton[[i]])
  }
  out
}

# rstan helper function to get dims of parameters right
.create_skeleton <- function(pars, dims) {
  out <- lapply(seq_along(pars), function(i) {
    len_dims <- length(dims[[i]])
    if (len_dims < 1) return(0)
    return(array(0, dim = dims[[i]]))
  })
  names(out) <- pars
  out
}

# extract original posterior draws
post_draws_stanfit <- function(x, ...) {
  as.matrix(x)
}

# compute a matrix of log-likelihood values for the ith observation
# matrix contains information about the number of MCMC chains
log_lik_i_stanfit <- function(x, i, parameter_name = "log_lik", ...) {
  loo::extract_log_lik(x, parameter_name, merge_chains = FALSE)[, , i]
}

# transform parameters to the unconstraint space
unconstrain_pars_stanfit <- function(x, pars, ...) {
  skeleton <- .create_skeleton(x@sim$pars_oi, x@par_dims[x@sim$pars_oi])
  upars <- apply(pars, 1, FUN = function(theta) {
    rstan::unconstrain_pars(x, .rstan_relist(theta, skeleton))
  })
  # for one parameter models
  if (is.null(dim(upars))) {
    dim(upars) <- c(1, length(upars))
  }
  t(upars)
}

# compute log_prob for each posterior draws on the unconstrained space
log_prob_upars_stanfit <- function(x, upars, ...) {
  apply(upars, 1, rstan::log_prob, object = x,
        adjust_transform = TRUE, gradient = FALSE)
}

# compute log_lik values based on the unconstrained parameters
log_lik_i_upars_stanfit <- function(x, upars, i, parameter_name = "log_lik",
                                  ...) {
  S <- nrow(upars)
  out <- numeric(S)
  for (s in seq_len(S)) {
    out[s] <- rstan::constrain_pars(x, upars = upars[s, ])[[parameter_name]][i]
  }
  out
}
```

Using these function, we can call `loo_moment_match()` to update the existing
`loo` object.

```{r loo_moment_match.default, message=FALSE}
loo3 <- loo::loo_moment_match.default(
  x = fit,
  loo = loo1,
  post_draws = post_draws_stanfit,
  log_lik_i = log_lik_i_stanfit,
  unconstrain_pars = unconstrain_pars_stanfit,
  log_prob_upars = log_prob_upars_stanfit,
  log_lik_i_upars = log_lik_i_upars_stanfit
)
loo3
```

As expected, the result is identical to the previous result of
`loo2 <- loo(fit, moment_match = TRUE)`.

# References

Gelman, A., and Hill, J. (2007).  *Data Analysis Using Regression and Multilevel Hierarchical Models.*  Cambridge University Press.

Stan Development Team (2020) _RStan: the R interface to Stan, Version 2.21.1_   https://mc-stan.org

Paananen, T., Piironen, J., Buerkner, P.-C., Vehtari, A. (2021).
Implicitly adaptive importance sampling. _Statistics and Computing_, 31, 16.
\doi:10.1007/s11222-020-09982-2. arXiv preprint arXiv:1906.08850.

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. _Statistics and Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. Links: [published](https://link.springer.com/article/10.1007/s11222-016-9696-4) | [arXiv preprint](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

---
title: "Bayesian Stacking and Pseudo-BMA weights using the loo package"
author: "Aki Vehtari and Jonah Gabry"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Bayesian Stacking and Pseudo-BMA weights}
-->
```{r, child="children/SETTINGS-knitr.txt"}
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

This vignette demonstrates the new functionality in __loo__ v2.0.0 for 
Bayesian stacking and Pseudo-BMA weighting. In this vignette we can't
provide all of the necessary background on this topic, so we encourage
readers to refer to the paper

* Yao, Y., Vehtari, A., Simpson, D., and Gelman, A.  (2018). Using stacking to average Bayesian predictive distributions. In Bayesian Analysis, \doi:10.1214/17-BA1091. [Online](https://projecteuclid.org/euclid.ba/1516093227)

which provides important details on the methods demonstrated in this vignette.
Here we just quote from the abstract of the paper: 

> **Abstract**: Bayesian model averaging is flawed in the $\mathcal{M}$-open setting in which the true data-generating process is not one of the candidate models being fit. We take the idea of stacking from the point estimation literature and generalize to the combination of predictive distributions. We extend the utility function to any proper scoring rule and use Pareto smoothed importance sampling to efficiently compute the required leave-one-out posterior distributions. We compare stacking of predictive distributions to several alternatives: stacking of means, Bayesian model averaging (BMA), Pseudo-BMA, and a variant of Pseudo-BMA that is stabilized using the Bayesian bootstrap. Based on simulations and real-data applications, we recommend stacking of predictive distributions, with bootstrapped-Pseudo-BMA as an approximate alternative when computation cost is an issue.

Ideally, we would avoid the Bayesian model combination problem by
extending the model to include the separate models as special cases,
and preferably as a continuous expansion of the model space.  For
example, instead of model averaging over different covariate
combinations, all potentially relevant covariates should be included
in a predictive model (for causal analysis more  care is needed) 
and a prior assumption that only some of the covariates are
relevant can be presented with regularized horseshoe prior (Piironen
and Vehtari, 2017a). For variable selection we recommend projective
predictive variable selection (Piironen and Vehtari, 2017a;
[__projpred__ package](https://cran.r-project.org/package=projpred)).

To demonstrate how to use __loo__ package to compute Bayesian stacking and
Pseudo-BMA weights, we repeat two simple model averaging examples from Chapters
6 and 10 of _Statistical Rethinking_ by Richard McElreath. In _Statistical
Rethinking_ WAIC is used to form weights which are similar to classical "Akaike
weights". Pseudo-BMA weighting using PSIS-LOO for computation is close to these
WAIC weights, but named after the Pseudo Bayes Factor by Geisser and Eddy
(1979). As discussed below, in general we prefer using stacking rather than WAIC
weights or the similar pseudo-BMA weights.


# Setup

In addition to the __loo__ package we will also load the __rstanarm__ package
for fitting the models.

```{r setup, message=FALSE}
library(rstanarm)
library(loo)
```

# Example: Primate milk 

In _Statistical Rethinking_, McElreath describes the data for the primate milk
example as follows:

> A popular hypothesis has it that primates with larger brains produce more energetic milk, so that brains can grow quickly. ... The question here is to what extent energy content of milk, measured here by kilocalories, is related to the percent of the brain mass that is neocortex. ... We'll end up needing female body mass as well, to see the masking that hides the relationships among the variables.

```{r data}
data(milk)
d <- milk[complete.cases(milk),]
d$neocortex <- d$neocortex.perc /100
str(d)
```

We repeat the analysis in Chapter 6 of _Statistical Rethinking_ using the
following four models (here we use the default weakly informative priors in __rstanarm__, while flat priors were used in _Statistical Rethinking_).

```{r fits, results="hide"}
fit1 <- stan_glm(kcal.per.g ~ 1, data = d, seed = 2030)
fit2 <- update(fit1, formula = kcal.per.g ~ neocortex)
fit3 <- update(fit1, formula = kcal.per.g ~ log(mass))
fit4 <- update(fit1, formula = kcal.per.g ~ neocortex + log(mass))
```

McElreath uses WAIC for model comparison and averaging, so we'll start by also
computing WAIC for these models so we can compare the results to the 
other options presented later in the vignette. The __loo__ package
provides `waic` methods for log-likelihood arrays, matrices and functions.
Since we fit our model with rstanarm we can use the `waic` method provided
by the __rstanarm__ package (a wrapper around `waic` from the __loo__ package), 
which allows us to just pass in our fitted model objects instead of first 
extracting the log-likelihood values.

```{r waic}
waic1 <- waic(fit1)
waic2 <- waic(fit2)
waic3 <- waic(fit3)
waic4 <- waic(fit4)
waics <- c(
  waic1$estimates["elpd_waic", 1],
  waic2$estimates["elpd_waic", 1],
  waic3$estimates["elpd_waic", 1],
  waic4$estimates["elpd_waic", 1]
)
```

We get some warnings when computing WAIC for models 3 and 4, indicating that we
shouldn't trust the WAIC weights we will compute later. Following the
recommendation in the warning, we next use the `loo` methods to compute PSIS-LOO
instead. The __loo__ package provides `loo` methods for log-likelihood arrays,
matrices, and functions, but since we fit our model with __rstanarm__ we can
just pass the fitted model objects directly and __rstanarm__ will extract the
needed values to pass to the __loo__ package. (Like __rstanarm__, some other R
packages for fitting Stan models, e.g. __brms__, also provide similar methods
for interfacing with the __loo__ package.)

```{r loo}
# note: the loo function accepts a 'cores' argument that we recommend specifying
# when working with bigger datasets

loo1 <- loo(fit1)
loo2 <- loo(fit2)
loo3 <- loo(fit3)
loo4 <- loo(fit4)
lpd_point <- cbind(
  loo1$pointwise[,"elpd_loo"], 
  loo2$pointwise[,"elpd_loo"],
  loo3$pointwise[,"elpd_loo"], 
  loo4$pointwise[,"elpd_loo"]
)
```

With `loo` we don't get any warnings for models 3 and 4, but for illustration of
good results, we display the diagnostic details for these models anyway.

```{r print-loo}
print(loo3)
print(loo4)
```

One benefit of PSIS-LOO over WAIC is better diagnostics. Here for both models 3
and 4 all $k<0.7$ and the Monte Carlo SE of `elpd_loo` is 0.1 or less, and we can expect
the model comparison to be reliable.

Next we compute and compare 1) WAIC weights, 2) Pseudo-BMA weights without
Bayesian bootstrap, 3) Pseudo-BMA+ weights with Bayesian bootstrap, and 4)
Bayesian stacking weights.

```{r weights}
waic_wts <- exp(waics) / sum(exp(waics))
pbma_wts <- pseudobma_weights(lpd_point, BB=FALSE)
pbma_BB_wts <- pseudobma_weights(lpd_point) # default is BB=TRUE
stacking_wts <- stacking_weights(lpd_point)
round(cbind(waic_wts, pbma_wts, pbma_BB_wts, stacking_wts), 2)
```

With all approaches Model 4 with `neocortex` and `log(mass)` gets most of the
weight. Based on theory, Pseudo-BMA weights without Bayesian bootstrap should be
close to WAIC weights, and we can also see that here. Pseudo-BMA+ weights with
Bayesian bootstrap provide more cautious weights further away from 0 and 1 (see
Yao et al. (2018) for a discussion of why this can be beneficial and results
from related experiments). In this particular example, the Bayesian stacking
weights are not much different from the other weights.

One of the benefits of stacking is that it manages well if there are many
similar models. Consider for example that there could be many irrelevant
covariates that when included would produce a similar model to one of the
existing models. To emulate this situation here we simply copy the first model a
bunch of times, but you can imagine that instead we would have ten alternative
models with about the same predictive performance. WAIC weights for such a
scenario would be close to the following:

```{r waic_wts_demo}
waic_wts_demo <- 
  exp(waics[c(1,1,1,1,1,1,1,1,1,1,2,3,4)]) /
  sum(exp(waics[c(1,1,1,1,1,1,1,1,1,1,2,3,4)]))
round(waic_wts_demo, 3)
```

Notice how much the weight for model 4 is lowered now that more models similar
to model 1 (or in this case identical) have been added. Both WAIC weights and
Pseudo-BMA approaches first estimate the predictive performance separately for
each model and then compute weights based on estimated relative predictive
performances. Similar models share similar weights so the weights of other
models must be reduced for the total sum of the weights to remain the same.

On the other hand, stacking optimizes the weights _jointly_, allowing for the
very similar models (in this toy example repeated models) to share their weight
while more unique models keep their original weights. In our example 
we can see this difference clearly: 

```{r stacking_weights}
stacking_weights(lpd_point[,c(1,1,1,1,1,1,1,1,1,1,2,3,4)])
```
Using stacking, the weight for the best model stays essentially unchanged.


# Example: Oceanic tool complexity

Another example we consider is the Kline oceanic tool complexity data, which
McElreath describes as follows:

>Different historical island populations possessed tool kits of different size. These kits include fish hooks, axes, boats, hand plows, and many other types of tools. A number of theories predict that larger populations will both develop and sustain more complex tool kits. ... It's also suggested that contact rates among populations effectively increases population [sic, probably should be tool kit] size, as it's relevant to technological evolution.

We build models predicting the total number of tools given the log
population size and the contact rate (high vs. low).

```{r Kline}
data(Kline)
d <- Kline
d$log_pop <- log(d$population)
d$contact_high <- ifelse(d$contact=="high", 1, 0)
str(d)
```

We start with a Poisson regression model with the log population size,
the contact rate, and an interaction term between them (priors are
informative priors as in _Statistical Rethinking_).
```{r fit10, results="hide"}
fit10 <-
  stan_glm(
    total_tools ~ log_pop + contact_high + log_pop * contact_high,
    family = poisson(link = "log"),
    data = d,
    prior = normal(0, 1, autoscale = FALSE),
    prior_intercept = normal(0, 100, autoscale = FALSE),
    seed = 2030
  )
```

Before running other models, we check whether Poisson is good choice
as the conditional observation model.

```{r loo10}
loo10 <- loo(fit10)
print(loo10)
```

We get at least one observation with $k>0.7$ and the estimated effective
number of parameters `p_loo` is larger than the total number of
parameters in the model. This indicates that Poisson might be too
narrow. A negative binomial model might be better, but with so few
observations it is not so clear.

We can compute LOO more accurately by running Stan again for the
leave-one-out folds with high $k$ estimates. When using __rstanarm__
this can be done by specifying the `k_threshold` argument:

```{r loo10-threshold}
loo10 <- loo(fit10, k_threshold=0.7)
print(loo10)
```

In this case we see that there is not much difference, and thus it is relatively
safe to continue.

As a comparison we also compute WAIC:
```{r waic10}
waic10 <- waic(fit10)
print(waic10)
```

The WAIC computation is giving warnings and the estimated ELPD is slightly more
optimistic. We recommend using the PSIS-LOO results instead.

To assess whether the contact rate and interaction term are useful, we can make
a comparison to models without these terms.

```{r contact_high, results="hide"}
fit11 <- update(fit10, formula = total_tools ~ log_pop + contact_high)
fit12 <- update(fit10, formula = total_tools ~ log_pop)
```
```{r loo-contact_high}
(loo11 <- loo(fit11))
(loo12 <- loo(fit12))
```
```{r relo-contact_high}
loo11 <- loo(fit11, k_threshold=0.7)
loo12 <- loo(fit12, k_threshold=0.7)
lpd_point <- cbind(
  loo10$pointwise[, "elpd_loo"], 
  loo11$pointwise[, "elpd_loo"], 
  loo12$pointwise[, "elpd_loo"]
)
```

For comparison we'll also compute WAIC values for these additional models:
```{r waic-contact_high}
waic11 <- waic(fit11)
waic12 <- waic(fit12)
waics <- c(
  waic10$estimates["elpd_waic", 1], 
  waic11$estimates["elpd_waic", 1], 
  waic12$estimates["elpd_waic", 1]
)
```
The WAIC computation again gives warnings, and we recommend using PSIS-LOO
instead.

Finally, we compute 1) WAIC weights, 2) Pseudo-BMA weights without
Bayesian bootstrap, 3) Pseudo-BMA+ weights with Bayesian bootstrap, and
4) Bayesian stacking weights.

```{r weights-contact_high}
waic_wts <- exp(waics) / sum(exp(waics))
pbma_wts <- pseudobma_weights(lpd_point, BB=FALSE)
pbma_BB_wts <- pseudobma_weights(lpd_point) # default is BB=TRUE
stacking_wts <- stacking_weights(lpd_point)
round(cbind(waic_wts, pbma_wts, pbma_BB_wts, stacking_wts), 2)
```

All weights favor the second model with the log population and the contact rate.
WAIC weights and Pseudo-BMA weights (without Bayesian bootstrap) are similar, while Pseudo-BMA+ is more cautious and closer to stacking weights. 

It may seem surprising that Bayesian stacking is giving zero weight to the first
model, but this is likely due to the fact that the estimated effect for the
interaction term is close to zero and thus models 1 and 2 give very similar
predictions. In other words, incorporating the model with the interaction (model
1) into the model average doesn't improve the predictions at all and so model 1
is given a weight of 0. On the other hand, models 2 and 3 are giving slightly
different predictions and thus their combination may be slightly better than
either alone. This behavior is related to the repeated similar model illustration
in the milk example above.

# Simpler coding using `loo_model_weights` function

Although in the examples above we called the `stacking_weights` and 
`pseudobma_weights` functions directly, we can also use the 
`loo_model_weights` wrapper, which takes as its input either a list
of pointwise log-likelihood matrices or a list of precomputed loo objects.
There are also `loo_model_weights` methods for stanreg objects (fitted
model objects from __rstanarm__) as well as fitted model objects from 
other packages (e.g. __brms__) that do the preparation work for the user
(see, e.g., the examples at `help("loo_model_weights", package = "rstanarm")`).

```{r loo_model_weights}
# using list of loo objects
loo_list <- list(loo10, loo11, loo12)
loo_model_weights(loo_list)
loo_model_weights(loo_list, method = "pseudobma")
loo_model_weights(loo_list, method = "pseudobma", BB = FALSE)
```

# References

McElreath, R. (2016). _Statistical rethinking: A Bayesian course with examples in R and Stan_. Chapman & Hall/CRC. http://xcelab.net/rm/statistical-rethinking/

Piironen, J. and Vehtari, A. (2017a). Sparsity information and
regularization in the horseshoe and other shrinkage priors. In
Electronic Journal of Statistics,
11(2):5018-5051. [Online](https://projecteuclid.org/euclid.ejs/1513306866).
	  
Piironen, J. and Vehtari, A. (2017b). Comparison of Bayesian predictive
methods for model selection. Statistics and Computing,
27(3):711-735. \doi:10.1007/s11222-016-9649-y. [Online](https://link.springer.com/article/10.1007/s11222-016-9649-y).

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC. _Statistics and
Computing_. 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4.
[online](https://link.springer.com/article/10.1007/s11222-016-9696-4), 
[arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2019). Pareto smoothed importance sampling. [arXiv preprint arXiv:1507.02646](https://arxiv.org/abs/1507.02646).

Yao, Y., Vehtari, A., Simpson, D., and Gelman, A. (2018). Using
stacking to average Bayesian predictive distributions. In Bayesian
Analysis, \doi:10.1214/17-BA1091. [Online](https://projecteuclid.org/euclid.ba/1516093227).

---
title: "Leave-one-out cross-validation for non-factorized models"
author: "Aki Vehtari, Paul Bürkner and Jonah Gabry"
date: "`r Sys.Date()`"
output:
  html_vignette:
    toc: yes
encoding: "UTF-8"
params:
  EVAL: !r identical(Sys.getenv("NOT_CRAN"), "true")
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{Leave-one-out cross-validation for non-factorized models}
-->
```{r settings, child="children/SETTINGS-knitr.txt"}
```
```{r more-knitr-ops, include=FALSE}
knitr::opts_chunk$set(
  cache=TRUE,
  message=FALSE, 
  warning=FALSE
)
```

```{r, child="children/SEE-ONLINE.txt", eval = if (isTRUE(exists("params"))) !params$EVAL else TRUE}
```

# Introduction

When computing ELPD-based LOO-CV for a Bayesian model we need to
compute the log leave-one-out predictive densities $\log{p(y_i | y_{-i})}$ for
every response value $y_i, \: i = 1, \ldots, N$, where $y_{-i}$ denotes all
response values except observation $i$. To obtain $p(y_i | y_{-i})$, we need to
have access to the pointwise likelihood $p(y_i\,|\, y_{-i}, \theta)$ and
integrate over the model parameters $\theta$:

$$
p(y_i\,|\,y_{-i}) =
  \int p(y_i\,|\, y_{-i}, \theta) \, p(\theta\,|\, y_{-i}) \,d \theta
$$

Here, $p(\theta\,|\, y_{-i})$ is the leave-one-out posterior distribution for
$\theta$, that is, the posterior distribution for $\theta$ obtained by fitting
the model while holding out the $i$th observation (we will later show how 
refitting the model to data $y_{-i}$ can be avoided).

If the observation model is formulated directly as the product of the
pointwise observation models, we call it a *factorized*
model. In this case, the likelihood is also the product of the pointwise
likelihood contributions $p(y_i\,|\, y_{-i}, \theta)$.
To better illustrate possible structures of the observation models, we
formally divide $\theta$ into two parts, observation-specific latent
variables $f = (f_1, \ldots, f_N)$ and hyperparameters $\psi$, so that
$p(y_i\,|\, y_{-i}, \theta) = p(y_i\,|\, y_{-i}, f_i, \psi)$. 
Depending on the model, one of the two parts of $\theta$
may also be empty. In very simple models, such as linear regression models, 
latent variables are not explicitly presented and response values are 
conditionally independent given
$\psi$, so that $p(y_i\,|\, y_{-i}, f_i, \psi) = p(y_i \,|\, \psi)$. 
The full likelihood can then be written in the familiar form

$$
p(y \,|\, \psi) = \prod_{i=1}^N p(y_i \,|\, \psi),
$$

where $y = (y_1, \ldots, y_N)$ denotes the vector of all responses.
When the likelihood factorizes this way, the conditional
pointwise log-likelihood can be obtained easily by computing
$p(y_i\,|\, \psi)$ for each $i$ with computational cost $O(n)$.

Yet, there are several reasons why a *non-factorized* observation model
may be necessary or preferred. In non-factorized
models, the joint likelihood of the response values $p(y \,|\, \theta)$ is not
factorized into observation-specific components, but rather given directly as one
joint expression. For some models, an analytic factorized formulation is
simply not available in which case we speak of a *non-factorizable* model. Even
in models whose observation model can be factorized in principle, it may still be
preferable to use a non-factorized form for reasons of efficiency
and numerical stability (Bürkner et al. 2020).

Whether a non-factorized model is used by necessity or for efficiency and
stability, it comes at the cost of having no direct access to the leave-one-out
predictive densities and thus to the overall leave-one-out
predictive accuracy. In theory, we can express the observation-specific
likelihoods in terms of the joint likelihood via

$$
p(y_i \,|\, y_{i-1}, \theta) = 
  \frac{p(y \,|\, \theta)}{p(y_{-i} \,|\, \theta)} = 
  \frac{p(y \,|\, \theta)}{\int p(y \,|\, \theta) \, d y_i},
$$

but the expression on the right-hand side may not always
have an analytical solution. Computing $\log p(y_i \,|\, y_{-i},
\theta)$ for non-factorized models is therefore often impossible, or at least
inefficient and numerically unstable. However, there is a large
class of multivariate normal and Student-$t$ models for which there are
efficient analytical solutions available.

More details can be found in our paper about LOO-CV for non-factorized models
(Bürkner, Gabry, & Vehtari, 2020), which is available as a preprint on arXiv
(https://arxiv.org/abs/1810.10559).

# LOO-CV for multivariate normal models

In this vignette, we will focus on non-factorized multivariate normal
models. Based on results of Sundararajan and Keerthi (2001), Bürkner et al. (2020) 
show that, for multivariate normal models with coriance matrix $C$, 
the LOO predictive mean and standard deviation can be computed as follows:

\begin{align}
  \mu_{\tilde{y},-i} &= y_i-\bar{c}_{ii}^{-1} g_i \nonumber \\
  \sigma_{\tilde{y},-i} &= \sqrt{\bar{c}_{ii}^{-1}},
\end{align}
where $g_i$ and $\bar{c}_{ii}$ are
\begin{align}
  g_i &= \left[C^{-1} y\right]_i \nonumber \\
  \bar{c}_{ii} &= \left[C^{-1}\right]_{ii}.
\end{align}

Using these results, the log predictive density of the $i$th observation is then
computed as

$$
  \log p(y_i \,|\, y_{-i},\theta)
  = - \frac{1}{2}\log(2\pi)
  - \frac{1}{2}\log \sigma^2_{-i}
  - \frac{1}{2}\frac{(y_i-\mu_{-i})^2}{\sigma^2_{-i}}.
$$

Expressing this same equation in terms of $g_i$ and $\bar{c}_{ii}$, the log
predictive density becomes:

$$
  \log p(y_i \,|\, y_{-i},\theta)
  = - \frac{1}{2}\log(2\pi)
  + \frac{1}{2}\log \bar{c}_{ii}
  - \frac{1}{2}\frac{g_i^2}{\bar{c}_{ii}}.
$$
(Note that Vehtari et al. (2016) has a typo in the corresponding Equation 34.)

From these equations we can now derive a recipe for obtaining the conditional
pointwise log-likelihood for _all_ models that can be expressed conditionally in
terms of a multivariate normal with invertible covariance matrix $C$.

## Approximate LOO-CV using integrated importance-sampling

The above LOO equations for multivariate normal models are conditional on
parameters $\theta$. Therefore, to obtain the leave-one-out predictive density
$p(y_i \,|\, y_{-i})$ we need to integrate over $\theta$,

$$
p(y_i\,|\,y_{-i}) =
  \int p(y_i\,|\,y_{-i}, \theta) \, p(\theta\,|\,y_{-i}) \,d\theta.
$$

Here, $p(\theta\,|\,y_{-i})$ is the leave-one-out posterior distribution for
$\theta$, that is, the posterior distribution for $\theta$ obtained by fitting
the model while holding out the $i$th observation. 

To avoid the cost of sampling from $N$ leave-one-out posteriors, it is possible
to take the posterior draws $\theta^{(s)}, \, s=1,\ldots,S$, from the
\emph{full} posterior $p(\theta\,|\,y)$, and then approximate the above integral
using integrated importance sampling (Vehtari et al., 2016, Section 3.6.1):

$$
 p(y_i\,|\,y_{-i}) \approx
   \frac{ \sum_{s=1}^S p(y_i\,|\,y_{-i},\,\theta^{(s)}) \,w_i^{(s)}}{ \sum_{s=1}^S w_i^{(s)}},
$$

where $w_i^{(s)}$ are importance weights. First we compute the raw importance ratios

$$
  r_i^{(s)} \propto \frac{1}{p(y_i \,|\, y_{-i}, \,\theta^{(s)})},
$$

and then stabilize them using Pareto smoothed importance sampling (PSIS, Vehtari
et al, 2019) to obtain the weights $w_i^{(s)}$. The resulting approximation is
referred to as PSIS-LOO (Vehtari et al, 2017).

## Exact LOO-CV with re-fitting

In order to validate the approximate LOO procedure, and also in order to allow
exact computations to be made for a small number of leave-one-out folds for
which the Pareto $k$ diagnostic (Vehtari et al, 2024) indicates an unstable
approximation, we need to consider how we might to do _exact_ leave-one-out CV
for a non-factorized model. In the case of a Gaussian process that has the
marginalization property, we could just drop the one row and column of $C$
corresponding to the held out out observation. This does not hold in general for
multivariate normal models, however, and to keep the original prior we may need
to maintain the full covariance matrix $C$ even when one of the observations is
left out.

The solution is to model $y_i$ as a missing observation and estimate it along
with all of the other model parameters. For a conditional multivariate normal
model, $\log p(y_i\,|\,y_{-i})$ can be computed as follows. First, we model
$y_i$ as missing and denote the corresponding parameter $y_i^{\mathrm{mis}}$.
Then, we define

$$
y_{\mathrm{mis}(i)} = (y_1, \ldots, y_{i-1}, y_i^{\mathrm{mis}}, y_{i+1}, \ldots, y_N).
$$
to be the same as the full set of observations $y$, except replacing $y_i$ with
the parameter $y_i^{\mathrm{mis}}$.

Second, we compute the LOO predictive mean and standard deviations as above, but
replace $y$ with $y_{\mathrm{mis}(i)}$ in the computation of
$\mu_{\tilde{y},-i}$:

$$
\mu_{\tilde{y},-i} = y_{{\mathrm{mis}}(i)}-\bar{c}_{ii}^{-1}g_i,
$$

where in this case we have

$$
g_i = \left[ C^{-1} y_{\mathrm{mis}(i)} \right]_i.
$$

The conditional log predictive density is then computed with the above
$\mu_{\tilde{y},-i}$ and the left out observation $y_i$:

$$
  \log p(y_i\,|\,y_{-i},\theta)
  = - \frac{1}{2}\log(2\pi)
  - \frac{1}{2}\log \sigma^2_{\tilde{y},-i}
  - \frac{1}{2}\frac{(y_i-\mu_{\tilde{y},-i})^2}{\sigma^2_{\tilde{y},-i}}.
$$

Finally, the leave-one-out predictive distribution can then be estimated as

$$
 p(y_i\,|\,y_{-i}) \approx \sum_{s=1}^S p(y_i\,|\,y_{-i}, \theta_{-i}^{(s)}),
$$

where $\theta_{-i}^{(s)}$ are draws from the posterior distribution
$p(\theta\,|\,y_{\mathrm{mis}(i)})$.

# Lagged SAR models

A common non-factorized multivariate normal model is the simultaneously
autoregressive (SAR) model, which is frequently used for spatially correlated
data. The lagged SAR model is defined as

$$
y = \rho Wy + \eta + \epsilon
$$
or equivalently
$$
(I - \rho W)y = \eta + \epsilon,
$$
where $\rho$ is the spatial correlation parameter and $W$ is a user-defined
weight matrix. The matrix $W$ has entries $w_{ii} = 0$ along the diagonal and
the off-diagonal entries $w_{ij}$ are larger when areas $i$ and $j$ are closer
to each other. In a linear model, the predictor term $\eta$ is given by 
$\eta = X \beta$ with design matrix $X$ and regression coefficients $\beta$. 
However, since the above equation holds for arbitrary $\eta$, these results are 
not restricted to linear models.

If we have $\epsilon \sim {\mathrm N}(0, \,\sigma^2 I)$, it follows that
$$
(I - \rho W)y \sim {\mathrm N}(\eta, \sigma^2 I),
$$
which corresponds to the following log PDF coded in **Stan**:

```{r lpdf, eval=FALSE}
/** 
 * Normal log-pdf for spatially lagged responses
 * 
 * @param y Vector of response values.
 * @param mu Mean parameter vector.
 * @param sigma Positive scalar residual standard deviation.
 * @param rho Positive scalar autoregressive parameter.
 * @param W Spatial weight matrix.
 *
 * @return A scalar to be added to the log posterior.
 */
real normal_lagsar_lpdf(vector y, vector mu, real sigma, 
                        real rho, matrix W) {
  int N = rows(y);
  real inv_sigma2 = 1 / square(sigma);
  matrix[N, N] W_tilde = -rho * W;
  vector[N] half_pred;
  
  for (n in 1:N) W_tilde[n,n] += 1;
  
  half_pred = W_tilde * (y - mdivide_left(W_tilde, mu));
  
  return 0.5 * log_determinant(crossprod(W_tilde) * inv_sigma2) -
         0.5 * dot_self(half_pred) * inv_sigma2;
}
```

For the purpose of computing LOO-CV, it makes sense to rewrite the SAR model in
slightly different form. Conditional on $\rho$, $\eta$, and $\sigma$, 
if we write

\begin{align}
y-(I-\rho W)^{-1}\eta &\sim {\mathrm N}(0, \sigma^2(I-\rho W)^{-1}(I-\rho W)^{-T}),
\end{align}
or more compactly, with $\widetilde{W}=(I-\rho W)$,
\begin{align}
y-\widetilde{W}^{-1}\eta &\sim {\mathrm N}(0, \sigma^2(\widetilde{W}^{T}\widetilde{W})^{-1}),
\end{align}

then this has the same form as the zero mean Gaussian process from above.
Accordingly, we can compute the leave-one-out predictive densities with the
equations from Sundararajan and Keerthi (2001), replacing $y$ with
$(y-\widetilde{W}^{-1}\eta)$ and taking the covariance matrix $C$ to be
$\sigma^2(\widetilde{W}^{T}\widetilde{W})^{-1}$.


## Case Study: Neighborhood Crime in Columbus, Ohio

In order to demonstrate how to carry out the computations implied by these
equations, we will first fit a lagged SAR model to data on crime in 49 different
neighborhoods of Columbus, Ohio during the year 1980. The data was originally
described in Aneslin (1988) and ships with the **spdep** R package.

In addition to the **loo** package, for this analysis we will use the **brms**
interface to Stan to generate a Stan program and fit the model, and also the
**bayesplot** and **ggplot2** packages for plotting.

```{r setup, cache=FALSE}
library("loo")
library("brms")
library("bayesplot")
library("ggplot2")
color_scheme_set("brightblue")
theme_set(theme_default())


SEED <- 10001 
set.seed(SEED) # only sets seed for R (seed for Stan set later)

# loads COL.OLD data frame and COL.nb neighbor list
data(oldcol, package = "spdep") 
```

The three variables in the data set relevant to this example are:

* `CRIME`: the number of residential burglaries and vehicle thefts per thousand
households in the neighbood
* `HOVAL`: housing value in units of $1000 USD 
* `INC`: household income in units of $1000 USD

```{r data}
str(COL.OLD[, c("CRIME", "HOVAL", "INC")])
```

We will also use the object `COL.nb`, which is a list containing information
about which neighborhoods border each other. From this list we will be able to
construct the weight matrix to used to help account for the spatial dependency
among the observations.

### Fit lagged SAR model

A model predicting `CRIME` from `INC` and `HOVAL`, while accounting for the
spatial dependency via an SAR structure, can be specified in **brms** as
follows.

```{r fit, results="hide"}
fit <- brm(
  CRIME ~ INC + HOVAL + sar(COL.nb, type = "lag"), 
  data = COL.OLD,
  data2 = list(COL.nb = COL.nb),
  chains = 4,
  seed = SEED
)
```

The code above fits the model in **Stan** using a log PDF equivalent to the
`normal_lagsar_lpdf` function we defined above. In the summary output below we
see that both higher income and higher housing value predict lower crime rates
in the neighborhood. Moreover, there seems to be substantial spatial correlation
between adjacent neighborhoods, as indicated by the posterior distribution 
of the `lagsar` parameter.

```{r plot-lagsar, message=FALSE}
lagsar <- as.matrix(fit, pars = "lagsar")
estimates <- quantile(lagsar, probs = c(0.25, 0.5, 0.75))
mcmc_hist(lagsar) + 
  vline_at(estimates, linetype = 2, size = 1) +
  ggtitle("lagsar: posterior median and 50% central interval")
```

### Approximate LOO-CV

After fitting the model, the next step is to compute the pointwise
log-likelihood values needed for approximate LOO-CV. To do this we will 
use the recipe laid out in the previous sections.

```{r approx}
posterior <- as.data.frame(fit)
y <- fit$data$CRIME
N <- length(y)
S <- nrow(posterior)
loglik <- yloo <- sdloo <- matrix(nrow = S, ncol = N)

for (s in 1:S) {
  p <- posterior[s, ]
  eta <- p$b_Intercept + p$b_INC * fit$data$INC + p$b_HOVAL * fit$data$HOVAL
  W_tilde <- diag(N) - p$lagsar * spdep::nb2mat(COL.nb)
  Cinv <- t(W_tilde) %*% W_tilde / p$sigma^2
  g <- Cinv %*% (y - solve(W_tilde, eta))
  cbar <- diag(Cinv)
  yloo[s, ] <- y - g / cbar
  sdloo[s, ] <- sqrt(1 / cbar)
  loglik[s, ] <- dnorm(y, yloo[s, ], sdloo[s, ], log = TRUE)
}

# use loo for psis smoothing
log_ratios <- -loglik
psis_result <- psis(log_ratios)
```

The quality of the PSIS-LOO approximation can be investigated graphically by
plotting the Pareto-k estimate for each observation. The approximation is robust up to values
of $0.7$ (Vehtari et al, 2017, 2024). In the plot below, we see that the fourth
observation is problematic and so may reduce the accuracy of the LOO-CV
approximation.

```{r plot, cache = FALSE}
plot(psis_result, label_points = TRUE)
```

We can also check that the conditional leave-one-out predictive distribution
equations work correctly, for instance, using the last posterior draw:

```{r checklast, cache = FALSE}
yloo_sub <- yloo[S, ]
sdloo_sub <- sdloo[S, ]
df <- data.frame(
  y = y, 
  yloo = yloo_sub,
  ymin = yloo_sub - sdloo_sub * 2,
  ymax = yloo_sub + sdloo_sub * 2
)
ggplot(data=df, aes(x = y, y = yloo, ymin = ymin, ymax = ymax)) +
  geom_errorbar(
    width = 1, 
    color = "skyblue3", 
    position = position_jitter(width = 0.25)
  ) +
  geom_abline(color = "gray30", size = 1.2) +
  geom_point()
```

Finally, we use PSIS-LOO to approximate the expected log predictive density
(ELPD) for new data, which we will validate using exact LOO-CV in the upcoming
section.

```{r psisloo}
(psis_loo <- loo(loglik))
```

### Exact LOO-CV

Exact LOO-CV for the above example is somewhat more involved, as we need to
re-fit the model $N$ times and each time model the held-out data point as a
parameter. First, we create an empty dummy model that we will update below as we
loop over the observations.

```{r fit_dummy, cache = TRUE}
# see help("mi", "brms") for details on the mi() usage
fit_dummy <- brm(
  CRIME | mi() ~ INC + HOVAL + sar(COL.nb, type = "lag"), 
  data = COL.OLD,
  data2 = list(COL.nb = COL.nb),
  chains = 0
)
```

Next, we fit the model $N$ times, each time leaving out a single observation and
then computing the log predictive density for that observation. For obvious
reasons, this takes much longer than the approximation we computed above, but it
is necessary in order to validate the approximate LOO-CV method. Thanks 
to the PSIS-LOO approximation, in general doing these slow exact computations
can be avoided.

```{r exact-loo-cv, results="hide", message=FALSE, warning=FALSE, cache = TRUE}
S <- 500
res <- vector("list", N)
loglik <- matrix(nrow = S, ncol = N)
for (i in seq_len(N)) {
  dat_mi <- COL.OLD
  dat_mi$CRIME[i] <- NA
  fit_i <- update(fit_dummy, newdata = dat_mi, 
                  # just for vignette
                  chains = 1, iter = S * 2)
  posterior <- as.data.frame(fit_i)
  yloo <- sdloo <- rep(NA, S)
  for (s in seq_len(S)) {
    p <- posterior[s, ]
    y_miss_i <- y
    y_miss_i[i] <- p$Ymi
    eta <- p$b_Intercept + p$b_INC * fit_i$data$INC + p$b_HOVAL * fit_i$data$HOVAL
    W_tilde <- diag(N) - p$lagsar * spdep::nb2mat(COL.nb)
    Cinv <- t(W_tilde) %*% W_tilde / p$sigma^2
    g <- Cinv %*% (y_miss_i - solve(W_tilde, eta))
    cbar <- diag(Cinv);
    yloo[s] <- y_miss_i[i] - g[i] / cbar[i]
    sdloo[s] <- sqrt(1 / cbar[i])
    loglik[s, i] <- dnorm(y[i], yloo[s], sdloo[s], log = TRUE)
  }
  ypred <- rnorm(S, yloo, sdloo)
  res[[i]] <- data.frame(y = c(posterior$Ymi, ypred))
  res[[i]]$type <- rep(c("pp", "loo"), each = S)
  res[[i]]$obs <- i
}
res <- do.call(rbind, res)
```

A first step in the validation of the pointwise predictive density is to compare
the distribution of the implied response values for the left-out observation to
the distribution of the $y_i^{\mathrm{mis}}$ posterior-predictive values
estimated as part of the model. If the pointwise predictive density is correct,
the two distributions should match very closely (up to sampling error). In the
plot below, we overlay these two distributions for the first four observations
and see that they match very closely (as is the case for all $49$ observations
of in this example).

```{r yplots, cache = FALSE, fig.width=10, out.width="95%", fig.asp = 0.3}
res_sub <- res[res$obs %in% 1:4, ]
ggplot(res_sub, aes(y, fill = type)) +
  geom_density(alpha = 0.6) +
  facet_wrap("obs", scales = "fixed", ncol = 4)
```

In the final step, we compute the ELPD based on the exact LOO-CV and compare it
to the approximate PSIS-LOO result computed earlier.

```{r loo_exact, cache=FALSE}
log_mean_exp <- function(x) {
  # more stable than log(mean(exp(x)))
  max_x <- max(x)
  max_x + log(sum(exp(x - max_x))) - log(length(x))
}
exact_elpds <- apply(loglik, 2, log_mean_exp)
exact_elpd <- sum(exact_elpds)
round(exact_elpd, 1)
```

The results of the approximate and exact LOO-CV are similar but not as close as
we would expect if there were no problematic observations. We can investigate
this issue more closely by plotting the approximate against the exact pointwise
ELPD values.

```{r compare, fig.height=5}
df <- data.frame(
  approx_elpd = psis_loo$pointwise[, "elpd_loo"],
  exact_elpd = exact_elpds
)
ggplot(df, aes(x = approx_elpd, y = exact_elpd)) +
  geom_abline(color = "gray30") +
  geom_point(size = 2) +
  geom_point(data = df[4, ], size = 3, color = "red3") +
  xlab("Approximate elpds") +
  ylab("Exact elpds") +
  coord_fixed(xlim = c(-16, -3), ylim = c(-16, -3))
```

In the plot above the fourth data point ---the observation flagged as
problematic by the PSIS-LOO approximation--- is colored in red and is the clear
outlier. Otherwise, the correspondence between the exact and approximate values
is strong. In fact, summing over the pointwise ELPD values and leaving out the
fourth observation yields practically equivalent results for approximate and
exact LOO-CV:

```{r pt4}
without_pt_4 <- c(
  approx = sum(psis_loo$pointwise[-4, "elpd_loo"]),
  exact = sum(exact_elpds[-4])  
)
round(without_pt_4, 1)
```

From this we can conclude that the difference we found when including *all*
observations does not indicate a bug in our implementation of the approximate
LOO-CV but rather a violation of its assumptions.


# Working with Stan directly

So far, we have specified the models in brms and only used Stan implicitely
behind the scenes. This allowed us to focus on the primary purpose of validating
approximate LOO-CV for non-factorized models. However, we would also like to
show how everything can be set up in Stan directly. The Stan code brms generates
is human readable and so we can use it to learn some of the essential aspects of
Stan and the particular model we are implementing. The Stan program below is
a slightly modified version of the code extracted via `stancode(fit_dummy)`:

```{r brms-stan-code, eval=FALSE}
// generated with brms 2.2.0
functions {
/** 
 * Normal log-pdf for spatially lagged responses
 * 
 * @param y Vector of response values.
 * @param mu Mean parameter vector.
 * @param sigma Positive scalar residual standard deviation.
 * @param rho Positive scalar autoregressive parameter.
 * @param W Spatial weight matrix.
 *
 * @return A scalar to be added to the log posterior.
 */
  real normal_lagsar_lpdf(vector y, vector mu, real sigma,
                          real rho, matrix W) {
    int N = rows(y);
    real inv_sigma2 = 1 / square(sigma);
    matrix[N, N] W_tilde = -rho * W;
    vector[N] half_pred;
    for (n in 1:N) W_tilde[n, n] += 1;
    half_pred = W_tilde * (y - mdivide_left(W_tilde, mu));
    return 0.5 * log_determinant(crossprod(W_tilde) * inv_sigma2) -
           0.5 * dot_self(half_pred) * inv_sigma2;
  }
}
data {
  int<lower=1> N;  // total number of observations
  vector[N] Y;  // response variable
  int<lower=0> Nmi;  // number of missings
  int<lower=1> Jmi[Nmi];  // positions of missings
  int<lower=1> K;  // number of population-level effects
  matrix[N, K] X;  // population-level design matrix
  matrix[N, N] W;  // spatial weight matrix
  int prior_only;  // should the likelihood be ignored?
}
transformed data {
  int Kc = K - 1;
  matrix[N, K - 1] Xc;  // centered version of X
  vector[K - 1] means_X;  // column means of X before centering
  for (i in 2:K) {
    means_X[i - 1] = mean(X[, i]);
    Xc[, i - 1] = X[, i] - means_X[i - 1];
  }
}
parameters {
  vector[Nmi] Ymi;  // estimated missings
  vector[Kc] b;  // population-level effects
  real temp_Intercept;  // temporary intercept
  real<lower=0> sigma;  // residual SD
  real<lower=0,upper=1> lagsar;  // SAR parameter
}
transformed parameters {
}
model {
  vector[N] Yl = Y;
  vector[N] mu = Xc * b + temp_Intercept;
  Yl[Jmi] = Ymi;
  // priors including all constants
  target += student_t_lpdf(temp_Intercept | 3, 34, 17);
  target += student_t_lpdf(sigma | 3, 0, 17)
    - 1 * student_t_lccdf(0 | 3, 0, 17);
  // likelihood including all constants
  if (!prior_only) {
    target += normal_lagsar_lpdf(Yl | mu, sigma, lagsar, W);
  }
}
generated quantities {
  // actual population-level intercept
  real b_Intercept = temp_Intercept - dot_product(means_X, b);
}
```

Here we want to focus on two aspects of the Stan code. First, because there is
no built-in function in Stan that calculates the log-likelihood for the lag-SAR
model, we define a new `normal_lagsar_lpdf` function in the `functions` block of
the Stan program. This is the same function we showed earlier in the vignette
and it can be used to compute the log-likelihood in an efficient and numerically
stable way. The `_lpdf` suffix used in the function name informs Stan that this
is a log probability density function.

Second, this Stan program nicely illustrates how to set up missing value
imputation. Instead of just computing the log-likelihood for the observed
responses `Y`, we define a new variable `Yl` which is equal to `Y` if the
reponse is observed and equal to `Ymi` if the response is missing. The latter is
in turn defined as a parameter and thus estimated along with all other paramters
of the model. More details about missing value imputation in Stan can be found
in the *Missing Data & Partially Known Parameters* section of the
[Stan manual](https://mc-stan.org/users/documentation/index.html).

The Stan code extracted from brms is not only helpful when learning Stan, but
can also drastically speed up the specification of models that are not support
by brms. If brms can fit a model similar but not identical to the desired model,
we can let brms generate the Stan program for the similar model and then mold it
into the program that implements the model we actually want to fit. Rather than
calling `stancode()`, which requires an existing fitted model object, we
recommend using `make_stancode()` and specifying the `save_model` argument to
write the Stan program to a file. The corresponding data can be prepared with
`make_standata()` and then manually amended if needed. Once the code and data
have been edited, they can be passed to RStan's `stan()` function via the `file`
and `data` arguments.

# Conclusion

In summary, we have shown how to set up and validate approximate and exact
LOO-CV for non-factorized multivariate normal models using Stan with the
**brms** and **loo** packages. Although we focused on the particular example of
a spatial SAR model, the presented recipe applies more generally to models that
can be expressed in terms of a multivariate normal likelihood.

<br />

# References

Anselin L. (1988). *Spatial econometrics: methods and models*. Dordrecht: Kluwer Academic.

Bürkner P. C., Gabry J., & Vehtari A. (2020). Efficient leave-one-out cross-validation for Bayesian non-factorized normal and Student-t models. *Computational Statistics*, \doi:10.1007/s00180-020-01045-4. 
[ArXiv preprint](https://arxiv.org/abs/1810.10559).

Sundararajan S. & Keerthi S. S. (2001). Predictive approaches for choosing hyperparameters in Gaussian processes. *Neural Computation*, 13(5), 1103--1118.

Vehtari A., Mononen T., Tolvanen V., Sivula T., & Winther O. (2016). Bayesian leave-one-out cross-validation approximations for Gaussian latent variable models. *Journal of Machine Learning Research*, 17(103), 1--38. [Online](https://jmlr.org/papers/v17/14-540.html).

Vehtari A., Gelman A., & Gabry J. (2017). Practical Bayesian model evaluation using leave-one-out cross-validation and WAIC. *Statistics and Computing*, 27(5), 1413--1432. \doi:10.1007/s11222-016-9696-4. [Online](https://link.springer.com/article/10.1007/s11222-016-9696-4). [arXiv preprint arXiv:1507.04544](https://arxiv.org/abs/1507.04544).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling.  *Journal of Machine Learning Research*,
25(72):1-58. [PDF](https://jmlr.org/papers/v25/19-556.html)

## Test data for testing print method of `kfold` object

### Case 1: All pareto-k values are good

```{r}
set.seed(123)
dat <- dplyr::tibble(
  x = rnorm(200),
  y = 2 + 1.5 * x + rnorm(200, sd = 1)
)

fit <- brm(y ~ x, data = dat, seed = 42)
kfold1 <- kfold(fit)
saveRDS(kfold, "kfold-calibrated.Rds")
```

### Case 2: Some pareto-k values are problematic

```{r}
data(roaches, package = "rstanarm")
roaches$sqrt_roach1 <- sqrt(roaches$roach1)

fit_p <- brm(y ~ sqrt_roach1 + treatment + senior + offset(log(exposure2)),
                data = roaches,
                family = poisson,
                prior = prior(normal(0,1), class = b),
                refresh = 0)

kfold2 <- kfold(fit_p)
saveRDS(kfold2, "kfold-miscalibrated.Rds")
```

# loo, waic and elpd results haven't changed

    WAoAAAACAAQFAAACAwAAAAMTAAAACgAAAg4AAAAGwFTltnf8GWZACqFzqDCa8kBk5bZ3/Blm
    QBEipbqh/pI/8m8Dok+RiEAhIqW6of6SAAAEAgAAAAEABAAJAAAAA2RpbQAAAA0AAAACAAAA
    AwAAAAIAAAQCAAAAAQAEAAkAAAAIZGltbmFtZXMAAAATAAAAAgAAABAAAAADAAQACQAAAAhl
    bHBkX2xvbwAEAAkAAAAFcF9sb28ABAAJAAAABWxvb2ljAAAAEAAAAAIABAAJAAAACEVzdGlt
    YXRlAAQACQAAAAJTRQAAAP4AAAIOAAAAoMAC+fJnaLHwwAEP7abtCtDAArJiAQQUDsABa6HW
    xx3swAC19bRGOqbAAOVsk57EhMAHXE6oz8X6wAhOMwo4HzjAAxWJEz7wYMAAvCuQLdoawAEz
    +EIq/LDAARqnVc5fZMAAuMi8GPlmwAJBKzhjQ9zAAe5YKyVuTsADlXoqYwIGwBIRlu/H12/A
    E/WagapcWsACpJaoNucYwBLYLgxdc6fAA7rSrp1oIMAEf9XEej+SwAa4oZCQGv7ABf5oQger
    6sADXWMppHVuwAEJEltQ9nLAAOaxVhD86sAB7SPP8uDQwAnDIjSfxULABAp5a+6r9sAFJHaV
    7SH8wAEph84QBEw/eQDDaV43ET9xp+BQBCmTP3qfwKytXZ0/ctyOXeTxzz9xqi9N9aR4P3IZ
    d1RCmrs/hDyfFGZAlj+GNSCYztY2P3e6iLhToUY/cX4fUwwY4j9yNONK++tcP3MW2+Teir8/
    chrQtYH1VT94S/hkJOXvP4KGRiL5AqY/jR2tM+V2kj+xFwrjXj1kP6T55+1SS6g/h0J5DiSH
    Lz+l12L3soNYP36Eg9WZBwg/fE5degjr7j+CgjOtLdigP4MAE38XAjM/efuEjSOY2z938dyR
    QJQSP3UOviSS5E0/hJj7+xYimz+KmFtUHEseP3x0d1lSXvQ/fyjSq08iSD9x5Q1KokZ4P6BC
    0qXNAwA/kYGquy8AAD+imuIOTvsAP5NUpOQ/eAA/kQtugOBEAD+RMeghdl4AP7Jn3QxtCsA/
    tE9QFCi4gD+bSeGjWlIAP5FVoLo7XwA/kbVn03S8AD+Ui/4kS28AP5F+F0mQ2wA/mP14d7y0
    AD+pRBELXxKAP7xVzQoMgYA/6p0tx9H/CD/kuDZpjnkQP7DPsKVRKQA/5Y1g8Rh9sD+m1FBE
    ZMYAP6LmZaqjsIA/rvRkyt/1gD+wN1KCt4TAP6FZe4TNQ4A/lxkATIN0AD+UCq5AO8gAP6ku
    qlvkC4A/u837aXIIAD+joD1aCLQAP6afX0AzCAA/khQ6CvIGAEAS+fJnaLHwQBEP7abtCtBA
    ErJiAQQUDkARa6HWxx3sQBC19bRGOqZAEOVsk57EhEAXXE6oz8X6QBhOMwo4HzhAExWJEz7w
    YEAQvCuQLdoaQBEz+EIq/LBAERqnVc5fZEAQuMi8GPlmQBJBKzhjQ9xAEe5YKyVuTkATlXoq
    YwIGQCIRlu/H129AI/WagapcWkASpJaoNucYQCLYLgxdc6dAE7rSrp1oIEAUf9XEej+SQBa4
    oZCQGv5AFf5oQger6kATXWMppHVuQBEJEltQ9nJAEOaxVhD86kAR7SPP8uDQQBnDIjSfxUJA
    FAp5a+6r9kAVJHaV7SH8QBEph84QBEw/qQv9LeIWyb+uWWyqUA9QP7GREVrajU6/qkhYWDHd
    rL+9tsdoV7uMv7fBKJv14eU/t7YcNUPKsT/aTE3zYuljP6G3mVtQiba/vp1jR9XXbb+gLW87
    Hw2mv7FTEHhMvSa/sxFLk3EUuz+5A6XuSljYP9idj04idy4/0bL+IIz1RT/e15NGvsbPP9VS
    jdQVlsE/2LoPMwLTfD/U+wQrF7IQP6aKY3Lc//Q/zQFaAA69qT+4Rrgot4/8P8jynnS82Ey/
    nP2U9Wg1Vj+9sDYbFcUYP4hK+zwAbv4/1HzUh2soqj/D6wM3SdjbP7XPnkuMip4/xNfwCK9V
    Jj+1oylZVifyAAAEAgAAAf8AAAANAAAAAgAAACAAAAAFAAAEAgAAAv8AAAATAAAAAgAAAP4A
    AAAQAAAABQAEAAkAAAAIZWxwZF9sb28ABAAJAAAADW1jc2VfZWxwZF9sb28ABAAJAAAABXBf
    bG9vAAQACQAAAAVsb29pYwAEAAkAAAASaW5mbHVlbmNlX3BhcmV0b19rAAAA/gAAAhMAAAAD
    AAAADgAAACA/qQv9LeIWyb+uWWyqUA9QP7GREVrajU6/qkhYWDHdrL+9tsdoV7uMv7fBKJv1
    4eU/t7YcNUPKsT/aTE3zYuljP6G3mVtQiba/vp1jR9XXbb+gLW87Hw2mv7FTEHhMvSa/sxFL
    k3EUuz+5A6XuSljYP9idj04idy4/0bL+IIz1RT/e15NGvsbPP9VSjdQVlsE/2LoPMwLTfD/U
    +wQrF7IQP6aKY3Lc//Q/zQFaAA69qT+4Rrgot4/8P8jynnS82Ey/nP2U9Wg1Vj+9sDYbFcUY
    P4hK+zwAbv4/1HzUh2soqj/D6wM3SdjbP7XPnkuMip4/xNfwCK9VJj+1oylZVifyAAAADgAA
    ACBAjGtR4vOUSECNSIXMZdptQI1OzHDwC35AjCl3nEfMhECMV+BQfP7MQItWfnwINa1Aipoa
    +DhXG0CJ4jnKlWa8QIp1d1J1ZG1AjUahmmkFkkCL5UfgNu4DQI2abt7PHcNAi6rgACII5UCH
    F8533XLIQIiixsC750JAiRWGP5WHBUBkAZfpVM1eQHdzaroEDUlAhbbxJuuvP0B2G3GedgHa
    QIurDxL70Z1AimK5GPqZQ0CKb2AkeCpMQIqX6UZQkVlAjPSaSuWUT0CGV1Ue/HaIQIgmSdq8
    SDtAhKKy2922NECIf2FmaiksQIrvm5gPU99AioGA+OUNHUCNhdwLjHhJAAAADgAAACA/7iIj
    iprtpD/uhgpFxKUzP+9C2+HteTo/7WeuRVdLxz/thpx9PZyZP+x8NVYA5gc/7bgj+feDuD/t
    VJY/ijqGP+vk6tOVjqc/7oFMRnLCAz/tFUhNscONP+71S5Zdfro/7NXCRLX3pz/oSat+OPKu
    P+rzPc6+kK4/7rBYQ1wZPj/h+KYWRg06P/BP+OJ7Frg/6GyNUDLmvD/v8/5XaLpuP+3Om11x
    ZMc/7CaGH4VipT/tF1NNe4n9P+1mF6TWIdc/7s96p0MNpT/ndID3UIhJP+lAQASVPCI/5qSP
    laOkhD/s9t8Xe5GFP+zJwNIA1oY/7It81CYJlj/uynC4minDAAAEAgAAAAEABAAJAAAABW5h
    bWVzAAAAEAAAAAMABAAJAAAACHBhcmV0b19rAAQACQAAAAVuX2VmZgAEAAkAAAAFcl9lZmYA
    AAD+AAAA/gAAAA4AAAABwFTltnf8GWYAAAAOAAAAAUAKoXOoMJryAAAADgAAAAFAZOW2d/wZ
    ZgAAAA4AAAABQBEipbqh/pIAAAAOAAAAAT/ybwOiT5GIAAAADgAAAAFAISKluqH+kgAABAIA
    AAP/AAAAEAAAAAoABAAJAAAACWVzdGltYXRlcwAEAAkAAAAJcG9pbnR3aXNlAAQACQAAAAtk
    aWFnbm9zdGljcwAEAAkAAAALcHNpc19vYmplY3QABAAJAAAACGVscGRfbG9vAAQACQAAAAVw
    X2xvbwAEAAkAAAAFbG9vaWMABAAJAAAAC3NlX2VscGRfbG9vAAQACQAAAAhzZV9wX2xvbwAE
    AAkAAAAIc2VfbG9vaWMAAAQCAAAAAQAEAAkAAAAEZGltcwAAAA0AAAACAAAD6AAAACAAAAQC
    AAAAAQAEAAkAAAAFY2xhc3MAAAAQAAAAAwAEAAkAAAAIcHNpc19sb28ABAAJAAAAF2ltcG9y
    dGFuY2Vfc2FtcGxpbmdfbG9vAAQACQAAAANsb28AAAD+

---

    WAoAAAACAAQFAAACAwAAAAMTAAAACAAAAg4AAAAGwFTh8N3JQlhACijAYdW5U0Bk4fDdyUJY
    QBEIPbMRcF8/8f1l7HLzXkAhCD2zEXBfAAAEAgAAAAEABAAJAAAAA2RpbQAAAA0AAAACAAAA
    AwAAAAIAAAQCAAAAAQAEAAkAAAAIZGltbmFtZXMAAAATAAAAAgAAABAAAAADAAQACQAAAAll
    bHBkX3dhaWMABAAJAAAABnBfd2FpYwAEAAkAAAAEd2FpYwAAABAAAAACAAQACQAAAAhFc3Rp
    bWF0ZQAEAAkAAAACU0UAAAD+AAACDgAAAGDAAvmNuneZ+MABD8PfZfY4wAKx99lcynHAAWt8
    Ymy7fsAAtdsVao/IwADlTdu1DZXAB1udTX3k5MAITnOoOskWwAMVPYmACM/AALwIUpw5kMAB
    M9Wkr7fgwAEacHJh2rLAALipB9PpyMACQM0jnEqOwAHuKRocp5fAA5B3eimOtsAR/4j9ckNW
    wBPkxJfCGnvAAqNA9Dl1Q8ASxzduTNRhwAO6YsL8iDTABH/w24FP78AGt/B5FTffwAX9sPYm
    FK7AA1z0NJiPzcABCNYOV3oNwADmeBUiLKbAAewYWMd6msAJv/mOcrF/wAQKIFP4rJDABSRV
    zW5E/8ABKW7gHFe9P6App2mHBRU/kWzG96Sz0z+igFgkfJPZP5NB6rcOQRY/kP4fEwrVCz+R
    IowsmuZNP7JRsaIw6A4/tFdj1H30QD+bJBzD5olzP5FEAfFrGdk/kaQZFdJUCj+UcIxuCRXv
    P5FuPScJC8U/mM5uFEAM6j+pOEzJLWSxP7u1dwLeF4A/6gy+NSVeQz/kMYcaTGoXP7Ck+iWi
    7mA/5QWsAJODgj+muFVcLMsTP6LtK2xnx8I/rsge7Cctwj+wIGkGhJ02P6E9vkHT2zQ/lvrZ
    z8VBdj+T7g3I06XIP6jrzJEKfhw/u2jmo8+PpT+jifdciNpuP6aXLSB7yNY/kgfDERu+V0AS
    +Y26d5n4QBEPw99l9jhAErH32VzKcUARa3xibLt+QBC12xVqj8hAEOVN27UNlUAXW51NfeTk
    QBhOc6g6yRZAExU9iYAIz0AQvAhSnDmQQBEz1aSvt+BAERpwcmHaskAQuKkH0+nIQBJAzSOc
    So5AEe4pGhynl0ATkHd6KY62QCH/iP1yQ1ZAI+TEl8Iae0ASo0D0OXVDQCLHN25M1GFAE7pi
    wvyINEAUf/DbgU/vQBa38HkVN99AFf2w9iYUrkATXPQ0mI/NQBEI1g5Xeg1AEOZ4FSIspkAR
    7BhYx3qaQBm/+Y5ysX9AFAogU/iskEAVJFXNbkT/QBEpbuAcV70AAAQCAAAB/wAAAA0AAAAC
    AAAAIAAAAAMAAAQCAAAC/wAAABMAAAACAAAA/gAAABAAAAADAAQACQAAAAllbHBkX3dhaWMA
    BAAJAAAABnBfd2FpYwAEAAkAAAAEd2FpYwAAAP4AAAAOAAAAAcBU4fDdyUJYAAAADgAAAAFA
    CijAYdW5UwAAAA4AAAABQGTh8N3JQlgAAAAOAAAAAUARCD2zEXBfAAAADgAAAAE/8f1l7HLz
    XgAAAA4AAAABQCEIPbMRcF8AAAQCAAAAAQAEAAkAAAAFbmFtZXMAAAAQAAAACAAEAAkAAAAJ
    ZXN0aW1hdGVzAAQACQAAAAlwb2ludHdpc2UABAAJAAAACWVscGRfd2FpYwAEAAkAAAAGcF93
    YWljAAQACQAAAAR3YWljAAQACQAAAAxzZV9lbHBkX3dhaWMABAAJAAAACXNlX3Bfd2FpYwAE
    AAkAAAAHc2Vfd2FpYwAABAIAAAABAAQACQAAAARkaW1zAAAADQAAAAIAAAPoAAAAIAAABAIA
    AAABAAQACQAAAAVjbGFzcwAAABAAAAACAAQACQAAAAR3YWljAAQACQAAAANsb28AAAD+

---

    WAoAAAACAAQFAAACAwAAAAMTAAAAAgAAAg4AAAAEwFQQqtq6lJBAZBCq2rqUkEAJ4d/NRDUI
    QBnh381ENQgAAAQCAAAAAQAEAAkAAAADZGltAAAADQAAAAIAAAACAAAAAgAABAIAAAABAAQA
    CQAAAAhkaW1uYW1lcwAAABMAAAACAAAAEAAAAAIABAAJAAAABGVscGQABAAJAAAAAmljAAAA
    EAAAAAIABAAJAAAACEVzdGltYXRlAAQACQAAAAJTRQAAAP4AAAIOAAAAQMACuOcc0X3kwADs
    6lF2rNDAAmf2eMrYIsABRPiM/p78wACT3tdEeh7AAMMIw1vXyMAGyQ/AbF2kwAeruImW2XTA
    At71T/g7vMAAmYBOuWNcwAEQjXKEEzjAAPGPWYXIhsAAlcyNhdewwAIPMEdzynTAAYlH5vfy
    BMACssvCEp36wA174m2bLxzAEV6TtHiNOMACHhkjDF3QwBAmge46Y/HAA1+BbYvVCMAENDwt
    z7DQwAY8z/1kmyjABXytrfHvxMADF/07kUBgwADa4Fq374rAAL6b+ZCFWsABiGkmg1CiwAjk
    sllUNQLAA7v4doaJJsAEyfkY7FXcwAEFX1n6IEBAErjnHNF95EAQ7OpRdqzQQBJn9njK2CJA
    EUT4jP6e/EAQk97XRHoeQBDDCMNb18hAFskPwGxdpEAXq7iJltl0QBLe9U/4O7xAEJmATrlj
    XEAREI1yhBM4QBDxj1mFyIZAEJXMjYXXsEASDzBHc8p0QBGJR+b38gRAErLLwhKd+kAde+Jt
    my8cQCFek7R4jThAEh4ZIwxd0EAgJoHuOmPxQBNfgW2L1QhAFDQ8Lc+w0EAWPM/9ZJsoQBV8
    ra3x78RAExf9O5FAYEAQ2uBat++KQBC+m/mQhVpAEYhpJoNQokAY5LJZVDUCQBO7+HaGiSZA
    FMn5GOxV3EARBV9Z+iBAAAAEAgAAAf8AAAANAAAAAgAAACAAAAACAAAEAgAAAv8AAAATAAAA
    AgAAAP4AAAAQAAAAAgAEAAkAAAAEZWxwZAAEAAkAAAACaWMAAAD+AAAEAgAAAAEABAAJAAAA
    BW5hbWVzAAAAEAAAAAIABAAJAAAACWVzdGltYXRlcwAEAAkAAAAJcG9pbnR3aXNlAAAEAgAA
    AAEABAAJAAAABGRpbXMAAAANAAAAAgAAA+gAAAAgAAAEAgAAAAEABAAJAAAABWNsYXNzAAAA
    EAAAAAIABAAJAAAADGVscGRfZ2VuZXJpYwAEAAkAAAADbG9vAAAA/g==

# loo_predictive_metric is equal to snapshot

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP+b6DHwJlO4AAAAOAAAAAT+0VQV434B0
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAACGVzdGltYXRlAAQACQAAAAJz
    ZQAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP/c/cc5N6ckAAAAOAAAAAT/C02/w8SU2
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAACGVzdGltYXRlAAQACQAAAAJz
    ZQAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP+r8D5W4BOIAAAAOAAAAAT+07COl+XlU
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAACGVzdGltYXRlAAQACQAAAAJz
    ZQAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP/qvkDt3DvwAAAAOAAAAAT/BcF2FCRTJ
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAACGVzdGltYXRlAAQACQAAAAJz
    ZQAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP+bBWsjIsVUAAAAOAAAAAT/BpKrAKkv4
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAACGVzdGltYXRlAAQACQAAAAJz
    ZQAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABQAZBDZS19iwAAAAOAAAAAT/dFfIrJ59T
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAACGVzdGltYXRlAAQACQAAAAJz
    ZQAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP9gAAAAAAAAAAAAOAAAAAT+16K3SNqWP
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAACGVzdGltYXRlAAQACQAAAAJz
    ZQAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP9QAAAAAAAAAAAAOAAAAAT+0+ea7xOyz
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAACGVzdGltYXRlAAQACQAAAAJz
    ZQAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP9aw32sN9rEAAAAOAAAAAT+uapXy8Hyq
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAACGVzdGltYXRlAAQACQAAAAJz
    ZQAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP+AAAAAAAAAAAAAOAAAAAQAAAAAAAAAA
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAACGVzdGltYXRlAAQACQAAAAJz
    ZQAAAP4=

# Test the vignette

    Code
      print(looss_1)
    Output
      
      Computed from 4000 by 100 subsampled log-likelihood
      values from 3020 total observations.
      
               Estimate   SE subsampling SE
      elpd_loo  -1968.5 15.6            0.3
      p_loo         3.1  0.1            0.4
      looic      3936.9 31.2            0.6
      ------
      MCSE of elpd_loo is 0.0.
      MCSE and ESS estimates assume independent draws (r_eff=1).
      
      All Pareto k estimates are good (k < 0.7).
      See help('pareto-k-diagnostic') for details.

---

    Code
      print(looss_1b)
    Output
      
      Computed from 4000 by 200 subsampled log-likelihood
      values from 3020 total observations.
      
               Estimate   SE subsampling SE
      elpd_loo  -1968.3 15.6            0.2
      p_loo         3.2  0.1            0.4
      looic      3936.7 31.2            0.5
      ------
      MCSE of elpd_loo is 0.0.
      MCSE and ESS estimates assume independent draws (r_eff=1).
      
      All Pareto k estimates are good (k < 0.7).
      See help('pareto-k-diagnostic') for details.

---

    Code
      print(aploo_1)
    Output
      
      Computed from 2000 by 3020 log-likelihood matrix.
      
               Estimate   SE
      elpd_loo  -1968.4 15.6
      p_loo         3.2  0.2
      looic      3936.8 31.2
      ------
      Posterior approximation correction used.
      MCSE of elpd_loo is 0.0.
      MCSE and ESS estimates assume independent draws (r_eff=1).
      
      All Pareto k estimates are good (k < 0.7).
      See help('pareto-k-diagnostic') for details.

---

    Code
      print(looapss_1)
    Output
      
      Computed from 2000 by 100 subsampled log-likelihood
      values from 3020 total observations.
      
               Estimate   SE subsampling SE
      elpd_loo  -1968.2 15.6            0.4
      p_loo         2.9  0.1            0.5
      looic      3936.4 31.1            0.8
      ------
      Posterior approximation correction used.
      MCSE of elpd_loo is 0.0.
      MCSE and ESS estimates assume independent draws (r_eff=1).
      
      All Pareto k estimates are good (k < 0.7).
      See help('pareto-k-diagnostic') for details.

---

    Code
      print(looss_2)
    Output
      
      Computed from 4000 by 100 subsampled log-likelihood
      values from 3020 total observations.
      
               Estimate   SE subsampling SE
      elpd_loo  -1952.0 16.2            0.2
      p_loo         2.6  0.1            0.3
      looic      3903.9 32.4            0.4
      ------
      MCSE of elpd_loo is 0.0.
      MCSE and ESS estimates assume independent draws (r_eff=1).
      
      All Pareto k estimates are good (k < 0.7).
      See help('pareto-k-diagnostic') for details.

---

    Code
      print(comp)
    Output
             elpd_diff se_diff subsampling_se_diff
      model2  0.0       0.0     0.0               
      model1 16.5      22.5     0.4               

---

    Code
      print(comp)
    Output
             elpd_diff se_diff subsampling_se_diff
      model2  0.0       0.0     0.0               
      model1 16.1       4.4     0.1               

---

    Code
      print(comp2)
    Output
             elpd_diff se_diff subsampling_se_diff
      model2  0.0       0.0     0.0               
      model1 16.3       4.4     0.1               

---

    Code
      print(comp3)
    Output
             elpd_diff se_diff subsampling_se_diff
      model2  0.0       0.0     0.0               
      model1 16.5       4.4     0.3               

# E_loo.default equal to snapshots

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP5YkyJk2Uw4AAAAOAAAAAT/CCf5d2lYl
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAABXZhbHVlAAQACQAAAAhwYXJl
    dG9fawAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP+99QfBwyEoAAAAOAAAAAT/CwzqDy8zd
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAABXZhbHVlAAQACQAAAAhwYXJl
    dG9fawAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP+++XajbmJQAAAAOAAAAAT/CwzqDy8zd
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAABXZhbHVlAAQACQAAAAhwYXJl
    dG9fawAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAABP5Q/UH4+kokAAAAOAAAAAT+pC/0t4hZY
    AAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAIABAAJAAAABXZhbHVlAAQACQAAAAhwYXJl
    dG9fawAAAP4=

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAADv/RdTp/OKVA/lD9Qfj6SiT/0B+XkzCLI
    AAAADgAAAAE/qQv9LeIWWAAABAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAACAAQACQAAAAV2
    YWx1ZQAEAAkAAAAIcGFyZXRvX2sAAAD+

# E_loo.matrix equal to snapshots

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAAgP5YkyJk2Uw4/qWAe5+slHb+UGxtRYEj6
    v3phLSMHz52/odiRUA18Oz+j5kMZY+OZv6UiBYpZA1c/gOCbIdkALD+aGlgT7KOMv6le704r
    1Mk/lQ5ZQDf3uj+ewlXQr8iOv7Cx0vAv3x8/qBb/vlaNmD+RVknGML4AP5YmXOYjb6E/oOks
    rHS+3b+SlEsEisU6v5de/GugPA6/poyaPupPOr+GCwyWu+Dpv3vFGSbfabK/mtljUf0Mlz+K
    vNxsu3xMv6N5GBqP57a/jGhAZ3mD+L+mlNu35k+lP3ecrpcfLVC/dzGNuHKBtD+g/BS/OEnN
    v4Igw4FPLoM/WQ0Eo+5f+gAAAA4AAAAgP8IJ/l3aViU/pca2RdOM3j+/8rXkBqCrv6pIWFgx
    3Ro/zBDTTM3Ckz+e+pZvzROmP8SF8jCU81Y/2kxN82LpUD+ht5lbUInzP8PjBotEZAk/sBld
    t19woT+1Fp9rH8ZgP7mOpdZ+VZ8/uQOl7kpZID/YnY9OInbSP9Gy/iCM9TY/4pVb5TMm8D/h
    I59hPK5oP9i6DzMC02U/2d2gH1IGej/Fs7T0mf6XP80BWgAOvZg/wVDbzlrO2z/I8p50vNhw
    P82+MASgNhQ/vbA2GxXFDz+ISvs8AG4SP9R81IdrKFQ/x2BUQaDJEz+7JHLaPhQmP8TX8Aiv
    VNQ/wcnwypNNIwAABAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAACAAQACQAAAAV2YWx1ZQAE
    AAkAAAAIcGFyZXRvX2sAAAD+

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAAgP+99QfBwyEo/8GLAkD/xED/urslTY/r8
    P+9KWn8Ems8/8CFJIrn2dz/vm+wdgqz5P/AqS8QTOzQ/8Cbqm5CTOT/xSPL9dvdTP+8H16ot
    cmc/8Eke14BJfj/vYHX1dPLcP/BiQhlcUl4/78dCMDeGID/vSig0D8UoP/AOt7Yy0Nc/8QBL
    caz6WD/v2Rg/ai74P+7+Midd/Bg/7b36n0mRlT/wXcau4TfXP+/m6IR7cp4/8DQ+EMd7Xj/u
    46MOX9w+P/Gsf0/jcxw/7yZac/NVID/u8Trrn7y5P/AFJDL0n7s/72gByARvRT/wDbVLG9Ph
    P+8uXczdPuI/7yVeIEvf5gAAAA4AAAAgP8LDOoPLzN0/vC9PGiC2mD/Dw665YwtSv0/zEf9g
    gzs/wrKXeFX04D+6oCb1eqeGP9VkIEZUNbU/2kxN82LpUD+ht5lbUInzP8CPzHjviWW/oC1v
    Ox8Noz+3SaHHVF8tv5miuj3t2Pk/yY0i+C8iSD/YnY9OInbSP9Gy/iCM9TY/3teTRr7GQT/l
    5tcuLT5aP9i6DzMC02U/1PsEKxeyET+/YoKHfssaP80BWgAOvZg/uJULoaFQPT/I8p50vNhw
    P9eBi3nBuZc/x4JCmLKJjj+5nmB9gzl/P9R81IdrKFQ/w+sDN0nZGz+1z55LjIrMP8TX8Aiv
    VNQ/zWAJ7qX34gAABAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAACAAQACQAAAAV2YWx1ZQAE
    AAkAAAAIcGFyZXRvX2sAAAD+

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAAgP+++XajbmJQ/8DEU/8Lr5j/vVZ8WRY5+
    P++kqukYCOA/8BCb8nKMzD/vzc6xkMxBP/AVF/p+31Y/8BNphxvXqz/woUx0rrHkP++C95Ag
    m9k/8CRmBO/JwD/vr9aTD09nP/Aw1oOFlzI/7+OUeVFCtz/vpJF6/sziP/AHWiqornE/8H4z
    /wmU2T/v7IYye0D5P+9+EUnmiwk/7tm0BTqf0j/wLp9qBaS7P+/zccuuIkY/8BoJ2GEYeT/v
    cJAOK3HZP/DQ66mN/dI/75Jxr/yrxj/vd3o8ODlDP/ACkeSklNY/77OlzUqqMT/wBtkuWosN
    P++WgQC9Cqo/75Hxz1uVVwAAAA4AAAAgP8LDOoPLzN0/vC9PGiC2mD/Dw665YwtSv0/zEf9g
    gzs/wrKXeFX04D+6oCb1eqeGP9VkIEZUNbU/2kxN82LpUD+ht5lbUInzP8CPzHjviWW/oC1v
    Ox8Noz+3SaHHVF8tv5miuj3t2Pk/yY0i+C8iSD/YnY9OInbSP9Gy/iCM9TY/3teTRr7GQT/l
    5tcuLT5aP9i6DzMC02U/1PsEKxeyET+/YoKHfssaP80BWgAOvZg/uJULoaFQPT/I8p50vNhw
    P9eBi3nBuZc/x4JCmLKJjj+5nmB9gzl/P9R81IdrKFQ/w+sDN0nZGz+1z55LjIrMP8TX8Aiv
    VNQ/zWAJ7qX34gAABAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAACAAQACQAAAAV2YWx1ZQAE
    AAkAAAAIcGFyZXRvX2sAAAD+

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAA4AAAAgP5Q/UH4+kok/rJhyoKzLQ7+mS88H3cZO
    v4Diw77SBPK/nJmNhpoF3z+eJuXpnME8v6uKXkbg5Hc/iwobiZz0Yz+CcwHscaAbv7nU2Akt
    GX+/kI6eu9QpOj+rntL08OAEv7OnDaPGnrk/qZJuk37o6j+khJ9TfTs9v3VOr8Mp0w4/Y4WW
    2hIb6L+5piRwV8y9v51UE9Vh3d6/vDKUx9urWT+Vff+IHf9IP4PB4HUxxjK/rgu97ZVdnD91
    l2lfiMfQv6ABPz28O8C/sFJbkh7yZL+o2Wq0N3S+P4Zn3piChxE/bj6AYsnnNj93X8jkk/2u
    v5UhfPojTuu/pJCpz+t4lwAAAA4AAAAgP6kL/S3iFli/rllsqlAP3D+xkRFa2o0Dv6pIWFgx
    3Rq/vbbHaFe7lL+3wSib9eF2P7e2HDVDyo8/2kxN82LpUD+ht5lbUInzv76dY0fV19G/oC1v
    Ox8No7+xUxB4TL1Bv7MRS5NxFNQ/uQOl7kpZID/YnY9OInbSP9Gy/iCM9TY/3teTRr7GQT/V
    Uo3UFZa8P9i6DzMC02U/1PsEKxeyET+mimNy3QAxP80BWgAOvZg/uEa4KLePjD/I8p50vNhw
    v5z9lPVoNWw/vbA2GxXFDz+ISvs8AG4SP9R81IdrKFQ/w+sDN0nZGz+1z55LjIrMP8TX8Aiv
    VNQ/taMpWVYn7wAABAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAACAAQACQAAAAV2YWx1ZQAE
    AAkAAAAIcGFyZXRvX2sAAAD+

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAg4AAABAv/RdTp/OKVA/9Afl5MwiyL/z0FnOG0F+
    P/W3tESq+hG/83dZE3d+GD/zW0+Mmka2v/XZilmZkLg/9Anod0Btdb/0f6W0vqw/P/OZiJ83
    G76/9JCX02s4vD/z/c7djgcNv/TDQRow++g/86vcLJmh6L/0wjA4NlBjP/RMb7F0RzK/9HAc
    YG8N8D/1mlKBbWwFv/SagAvlYpw/88LV9Rj1xb/0mH86EFjXP/QlJsXLmXy/9FUITMCdhT/0
    YOU3NGbzv/XZiRsmTKE/8/w2ct9kJL/zwk+t/SPSP/Y2Nkb1kwi/9Jccz9A3RT/z1vUQMbyE
    v/Szng+3zeY/9MciBArOKL/0f9xw2jLPP/Vga+lxCxC/8tVQDeo7Az/1p5eybgglv/SJCtxH
    yEM/9AVPifNnSr/zjLE8SzISP/Ohe1nttQ6/9S6bPIHJgT/0hd4sDJmuv/RsPduiC1E/89vZ
    GzKrar/1FZjeT/k2P/S7KA0QeGi/83Y/x5KlBz/ztNqmF3ywv/ZCUvZXPyI/9PI83KF96L/0
    Ml3NWrX1P/Rnu9LXaiq/9WcGWiOq+D/zkg6tBDLov/T2vDTMsMs/9Og0XHuRtr/04y833CW8
    P/NqXO3hEbG/9HEUw/KPQD/1y+CeGQebv/SL1cuvxX8/9SbAzpRiIb/zufzsmkU1P/S0K91V
    63IAAAQCAAAAAQAEAAkAAAADZGltAAAADQAAAAIAAAACAAAAIAAAAP4AAAAOAAAAID+pC/0t
    4hZYv65ZbKpQD9w/sZERWtqNA7+qSFhYMd0av722x2hXu5S/t8Eom/Xhdj+3thw1Q8qPP9pM
    TfNi6VA/obeZW1CJ87++nWNH1dfRv6AtbzsfDaO/sVMQeEy9Qb+zEUuTcRTUP7kDpe5KWSA/
    2J2PTiJ20j/Rsv4gjPU2P97Xk0a+xkE/1VKN1BWWvD/Yug8zAtNlP9T7BCsXshE/popjct0A
    MT/NAVoADr2YP7hGuCi3j4w/yPKedLzYcL+c/ZT1aDVsP72wNhsVxQ8/iEr7PABuEj/UfNSH
    ayhUP8PrAzdJ2Rs/tc+eS4yKzD/E1/AIr1TUP7WjKVlWJ+8AAAQCAAAAAQAEAAkAAAAFbmFt
    ZXMAAAAQAAAAAgAEAAkAAAAFdmFsdWUABAAJAAAACHBhcmV0b19rAAAA/g==

# print.waic output is ok

    Code
      print(waic1)
    Output
      
      Computed from 1000 by 32 log-likelihood matrix.
      
                Estimate  SE
      elpd_waic    -83.5 4.3
      p_waic         3.3 1.1
      waic         167.1 8.5
      
      3 (9.4%) p_waic estimates greater than 0.4. We recommend trying loo instead. 

# print.psis_loo and print.psis output ok

    Code
      print(psis1)
    Output
      Computed from 1000 by 32 log-weights matrix.
      MCSE and ESS estimates assume independent draws (r_eff=1).
      
      All Pareto k estimates are good (k < 0.67).
      See help('pareto-k-diagnostic') for details.

---

    Code
      print(loo1)
    Output
      
      Computed from 1000 by 32 log-likelihood matrix.
      
               Estimate  SE
      elpd_loo    -83.6 4.3
      p_loo         3.3 1.2
      looic       167.2 8.6
      ------
      MCSE of elpd_loo is 0.1.
      MCSE and ESS estimates assume independent draws (r_eff=1).
      
      All Pareto k estimates are good (k < 0.67).
      See help('pareto-k-diagnostic') for details.

---

    Code
      print(loo1_r_eff)
    Output
      
      Computed from 1000 by 32 log-likelihood matrix.
      
               Estimate  SE
      elpd_loo    -83.6 4.3
      p_loo         3.3 1.2
      looic       167.2 8.6
      ------
      MCSE of elpd_loo is 0.1.
      MCSE and ESS estimates assume MCMC draws (r_eff in [0.6, 1.0]).
      
      All Pareto k estimates are good (k < 0.67).
      See help('pareto-k-diagnostic') for details.

# mcse_loo extractor gives correct value

    WAoAAAACAAQFAAACAwAAAAAOAAAAAT+2J8YDcP5s

# print.loo supports kfold with pareto-k diagnostics - calibrated

    Code
      print(kfold1)
    Output
      
      Based on 10-fold cross-validation.
      
                 Estimate   SE
      elpd_kfold   -285.0  9.2
      p_kfold         2.5  0.6
      kfoldic       570.0 18.4
      ------
      
      All Pareto k estimates are good (k < 0.7).
      See help('pareto-k-diagnostic') for details.

# print.loo supports kfold with pareto-k diagnostics - miscalibrated

    Code
      print(kfold1)
    Output
      
      Based on 10-fold cross-validation.
      
                 Estimate     SE
      elpd_kfold  -5556.6  701.0
      p_kfold       358.2  108.5
      kfoldic     11113.1 1401.9
      ------
      
      Pareto k diagnostic values:
                               Count Pct.    Min. ESS
      (-Inf, 0.7]   (good)     245   93.5%   24      
         (0.7, 1]   (bad)        8    3.1%   <NA>    
         (1, Inf)   (very bad)   9    3.4%   <NA>    
      See help('pareto-k-diagnostic') for details.

# print.loo supports kfold without pareto-k diagnostics

    Code
      print(kfold1)
    Output
      
      Based on 10-fold cross-validation.
      
                 Estimate     SE
      elpd_kfold  -5556.6  701.0
      p_kfold       358.2  108.5
      kfoldic     11113.1 1401.9

# crps matches snapshots

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAg4AAAACv+IeiUeluMc/vSY1w5IjJgAABAIAAAAB
    AAQACQAAAAVuYW1lcwAAABAAAAACAAQACQAAAAhFc3RpbWF0ZQAEAAkAAAACU0UAAAD+AAAA
    DgAAAAq/zdHHAHZD6L/K3ky9mEk4v/B2uF/xc76/38/N1JUUkL/Z4VW5uFHUv/D3XaD33ZO/
    yPJtFZYXlL/RgM+YTul0v+9F9Rv30V+/6Q2h5ndyqgAABAIAAAH/AAAAEAAAAAIABAAJAAAA
    CWVzdGltYXRlcwAEAAkAAAAJcG9pbnR3aXNlAAAA/g==

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAg4AAAACv/ELnetrJtw/uJgUmYONYAAABAIAAAAB
    AAQACQAAAAVuYW1lcwAAABAAAAACAAQACQAAAAhFc3RpbWF0ZQAEAAkAAAACU0UAAAD+AAAA
    DgAAAAq/6MKGUxChmr/pWmzfnhawv/dXJCu4gh2/7/dH2oOKIL/tR104AbV4v/fX3UkYDWm/
    5+ZWJz0Y3r/qbdDewNjFv/aQu+wUiEO/89yOK7F4DAAABAIAAAH/AAAAEAAAAAIABAAJAAAA
    CWVzdGltYXRlcwAEAAkAAAAJcG9pbnR3aXNlAAAA/g==

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAg4AAAACv+IVUtAszNU/vWbB6hSQFgAABAIAAAAB
    AAQACQAAAAVuYW1lcwAAABAAAAACAAQACQAAAAhFc3RpbWF0ZQAEAAkAAAACU0UAAAD+AAAA
    DgAAAAq/zKEONUJbrL/KuMhrCsFsv/BrsoFFUDu/31EWnT302L/ZnTwwjfZiv/E78Fc0R+6/
    yM8apxh02L/RTacKeM0kv+6ksQycKl+/6bkMJjTk9gAABAIAAAH/AAAAEAAAAAIABAAJAAAA
    CWVzdGltYXRlcwAEAAkAAAAJcG9pbnR3aXNlAAAA/g==

---

    WAoAAAACAAQFAAACAwAAAAITAAAAAgAAAg4AAAACv/EL1OkLv00/uNvlkZb58AAABAIAAAAB
    AAQACQAAAAVuYW1lcwAAABAAAAACAAQACQAAAAhFc3RpbWF0ZQAEAAkAAAACU0UAAAD+AAAA
    DgAAAAq/6JdmIec4s7/pNG4XiJ0yv/diwSdPiFi/77pjBRqbxr/tO+gNa/rMv/glDLlafB6/
    5+SScRdhOb/qdN1KnaG+v/Y6AtSCgWa/9Ca44XM7bgAABAIAAAH/AAAAEAAAAAIABAAJAAAA
    CWVzdGltYXRlcwAEAAkAAAAJcG9pbnR3aXNlAAAA/g==

# loo_compare returns expected results (2 models)

    WAoAAAACAAQEAgACAwAAAAMTAAAADAAAABAAAAACAAQACQAAAAZtb2RlbDEABAAJAAAABm1v
    ZGVsMgAAAA4AAAACAAAAAAAAAAAAAAAAAAAAAAAAAA4AAAACAAAAAAAAAAAAAAAAAAAAAAAA
    AA4AAAACf/AAAAAAB6J/8AAAAAAHogAAABAAAAACAAQACQAAAAAABAAJAAAAAAAAABAAAAAC
    AAQACQAAAAAABAAJAAAAAAAAAA4AAAACwFTh8N3JQljAVOHw3clCWAAAAA4AAAACQBEIPbMR
    cF9AEQg9sxFwXwAAAA4AAAACQAoowGHVuVNACijAYdW5UwAAAA4AAAACP/H9Zexy814/8f1l
    7HLzXgAAAA4AAAACQGTh8N3JQlhAZOHw3clCWAAAAA4AAAACQCEIPbMRcF9AIQg9sxFwXwAA
    BAIAAAABAAQACQAAAAVuYW1lcwAAABAAAAAMAAQACQAAAAVtb2RlbAAEAAkAAAAJZWxwZF9k
    aWZmAAQACQAAAAdzZV9kaWZmAAQACQAAAAdwX3dvcnNlAAQACQAAAAlkaWFnX2RpZmYABAAJ
    AAAACWRpYWdfZWxwZAAEAAkAAAAJZWxwZF93YWljAAQACQAAAAxzZV9lbHBkX3dhaWMABAAJ
    AAAABnBfd2FpYwAEAAkAAAAJc2VfcF93YWljAAQACQAAAAR3YWljAAQACQAAAAdzZV93YWlj
    AAAEAgAAAAEABAAJAAAABWNsYXNzAAAAEAAAAAIABAAJAAAAC2NvbXBhcmUubG9vAAQACQAA
    AApkYXRhLmZyYW1lAAAEAgAAAAEABAAJAAAACXJvdy5uYW1lcwAAAA0AAAACgAAAAP////4A
    AAD+

---

    Code
      print(comp1)
    Output
        model elpd_diff se_diff p_worse diag_diff diag_elpd
       model1       0.0     0.0      NA                    
       model2       0.0     0.0      NA                    

---

    WAoAAAACAAQEAgACAwAAAAMTAAAADAAAABAAAAACAAQACQAAAAZtb2RlbDEABAAJAAAABm1v
    ZGVsMgAAAA4AAAACAAAAAAAAAADAEDpTX5xF7gAAAA4AAAACAAAAAAAAAAA/tmpHtC8TAQAA
    AA4AAAACf/AAAAAAB6I/8AAAAAAAAAAAABAAAAACAAQACQAAAAAABAAJAAAAB04gPCAxMDAA
    AAAQAAAAAgAEAAkAAAAAAAQACQAAAAAAAAAOAAAAAsBU4fDdyUJYwFXllhPDBrkAAAAOAAAA
    AkARCD2zEXBfQBEalRIN2T8AAAAOAAAAAkAKKMBh1blTQCZnlesA0IoAAAAOAAAAAj/x/WXs
    cvNeP/GbYJxtZ8cAAAAOAAAAAkBk4fDdyUJYQGXllhPDBrkAAAAOAAAAAkAhCD2zEXBfQCEa
    lRIN2T8AAAQCAAAAAQAEAAkAAAAFbmFtZXMAAAAQAAAADAAEAAkAAAAFbW9kZWwABAAJAAAA
    CWVscGRfZGlmZgAEAAkAAAAHc2VfZGlmZgAEAAkAAAAHcF93b3JzZQAEAAkAAAAJZGlhZ19k
    aWZmAAQACQAAAAlkaWFnX2VscGQABAAJAAAACWVscGRfd2FpYwAEAAkAAAAMc2VfZWxwZF93
    YWljAAQACQAAAAZwX3dhaWMABAAJAAAACXNlX3Bfd2FpYwAEAAkAAAAEd2FpYwAEAAkAAAAH
    c2Vfd2FpYwAABAIAAAABAAQACQAAAAVjbGFzcwAAABAAAAACAAQACQAAAAtjb21wYXJlLmxv
    bwAEAAkAAAAKZGF0YS5mcmFtZQAABAIAAAABAAQACQAAAAlyb3cubmFtZXMAAAANAAAAAoAA
    AAD////+AAAA/g==

---

    Code
      print(comp2)
    Output
        model elpd_diff se_diff p_worse diag_diff diag_elpd
       model1       0.0     0.0      NA                    
       model2      -4.1     0.1    1.00   N < 100          
    Message
      
      Diagnostic flags present.
      See ?`loo-glossary` (sections `diag_diff` and `diag_elpd`)
      or https://mc-stan.org/loo/reference/loo-glossary.html.

---

    Code
      print(comp2, p_worse = FALSE)
    Output
        model elpd_diff se_diff
       model1       0.0     0.0
       model2      -4.1     0.1

# loo_compare returns expected result (3 models)

    WAoAAAACAAQEAgACAwAAAAMTAAAADAAAABAAAAADAAQACQAAAAZtb2RlbDEABAAJAAAABm1v
    ZGVsMgAEAAkAAAAGbW9kZWwzAAAADgAAAAMAAAAAAAAAAMAQOlNfnEXuwDANypG2BBgAAAAO
    AAAAAwAAAAAAAAAAP7ZqR7QvEwE/y6/t4TTtXgAAAA4AAAADf/AAAAAAB6I/8AAAAAAAAD/w
    AAAAAAAAAAAAEAAAAAMABAAJAAAAAAAEAAkAAAAHTiA8IDEwMAAEAAkAAAAHTiA8IDEwMAAA
    ABAAAAADAAQACQAAAAAABAAJAAAAAAAEAAkAAAAAAAAADgAAAAPAVOHw3clCWMBV5ZYTwwa5
    wFjlY4I2w2IAAAAOAAAAA0ARCD2zEXBfQBEalRIN2T9AEPIF3GigEwAAAA4AAAADQAoowGHV
    uVNAJmeV6wDQikBByNhSGt0KAAAADgAAAAM/8f1l7HLzXj/xm2CcbWfHP/GA0JJnyV8AAAAO
    AAAAA0Bk4fDdyUJYQGXllhPDBrlAaOVjgjbDYgAAAA4AAAADQCEIPbMRcF9AIRqVEg3ZP0Ag
    8gXcaKATAAAEAgAAAAEABAAJAAAABW5hbWVzAAAAEAAAAAwABAAJAAAABW1vZGVsAAQACQAA
    AAllbHBkX2RpZmYABAAJAAAAB3NlX2RpZmYABAAJAAAAB3Bfd29yc2UABAAJAAAACWRpYWdf
    ZGlmZgAEAAkAAAAJZGlhZ19lbHBkAAQACQAAAAllbHBkX3dhaWMABAAJAAAADHNlX2VscGRf
    d2FpYwAEAAkAAAAGcF93YWljAAQACQAAAAlzZV9wX3dhaWMABAAJAAAABHdhaWMABAAJAAAA
    B3NlX3dhaWMAAAQCAAAAAQAEAAkAAAAFY2xhc3MAAAAQAAAAAgAEAAkAAAALY29tcGFyZS5s
    b28ABAAJAAAACmRhdGEuZnJhbWUAAAQCAAAAAQAEAAkAAAAJcm93Lm5hbWVzAAAADQAAAAKA
    AAAA/////QAAAP4=

---

    Code
      print(comp1)
    Output
        model elpd_diff se_diff p_worse diag_diff diag_elpd
       model1       0.0     0.0      NA                    
       model2      -4.1     0.1    1.00   N < 100          
       model3     -16.1     0.2    1.00   N < 100          
    Message
      
      Diagnostic flags present.
      See ?`loo-glossary` (sections `diag_diff` and `diag_elpd`)
      or https://mc-stan.org/loo/reference/loo-glossary.html.

# compare returns expected result (3 models)

    WAoAAAACAAQFAAACAwAAAAMOAAAAGAAAAAAAAAAAwBA6U1+cRe7AMA3KkbYEGAAAAAAAAAAA
    P7ZqR7QvEwE/y6/t4TTtXsBU4fDdyUJYwFXllhPDBrnAWOVjgjbDYkARCD2zEXBfQBEalRIN
    2T9AEPIF3GigE0AKKMBh1blTQCZnlesA0IpAQcjYUhrdCj/x/WXscvNeP/GbYJxtZ8c/8YDQ
    kmfJX0Bk4fDdyUJYQGXllhPDBrlAaOVjgjbDYkAhCD2zEXBfQCEalRIN2T9AIPIF3GigEwAA
    BAIAAAABAAQACQAAAANkaW0AAAANAAAAAgAAAAMAAAAIAAAEAgAAAAEABAAJAAAACGRpbW5h
    bWVzAAAAEwAAAAIAAAAQAAAAAwAEAAkAAAACdzEABAAJAAAAAncyAAQACQAAAAJ3MwAAABAA
    AAAIAAQACQAAAAllbHBkX2RpZmYABAAJAAAAB3NlX2RpZmYABAAJAAAACWVscGRfd2FpYwAE
    AAkAAAAMc2VfZWxwZF93YWljAAQACQAAAAZwX3dhaWMABAAJAAAACXNlX3Bfd2FpYwAEAAkA
    AAAEd2FpYwAEAAkAAAAHc2Vfd2FpYwAABAIAAAABAAQACQAAAAVjbGFzcwAAABAAAAAEAAQA
    CQAAAAtjb21wYXJlLmxvbwAEAAkAAAAGbWF0cml4AAQACQAAAAVhcnJheQAEAAkAAAAPb2xk
    X2NvbXBhcmUubG9vAAAA/g==

# relative_eff results haven't changed

    WAoAAAACAAQFAAACAwAAAAAOAAAAID/uIiOKmu2kP+6GCkXEpTM/70Lb4e15Oj/tZ65FV0vH
    P+2GnH09nJk/7Hw1VgDmBz/tuCP594O4P+1Ulj+KOoY/6+Tq05WOpz/ugUxGcsIDP+0VSE2x
    w40/7vVLll1+uj/s1cJEtfenP+hJq3448q4/6vM9zr6Qrj/usFhDXBk+P+H4phZGDTo/8E/4
    4nsWuD/obI1QMua8P+/z/ldoum4/7c6bXXFkxz/sJoYfhWKlP+0XU017if0/7WYXpNYh1z/u
    z3qnQw2lP+d0gPdQiEk/6UBABJU8Ij/mpI+Vo6SEP+z23xd7kYU/7MnA0gDWhj/si3zUJgmW
    P+7KcLiaKcM=

# loo_model_weights (stacking and pseudo-BMA) gives expected result

    WAoAAAACAAQFAgACAwAAAAAOAAAAAz/KEXnf1DM9P+l7oXTYyi4+YzIpAwAAAA==

---

    Code
      print(w1)
    Output
      Method: stacking
      ------
             weight
      model1 0.204 
      model2 0.796 
      model3 0.000 

---

    WAoAAAACAAQFAgACAwAAAAAOAAAAAz+xA6UGtqDDP+3eFS5zKzY/J2MLYAsc4A==

---

    Code
      print(w2)
    Output
      Method: pseudo-BMA+ with Bayesian bootstrap
      ------
             weight
      model1 0.066 
      model2 0.933 
      model3 0.000 

---

    Code
      print(w3)
    Output
      Method: pseudo-BMA
      ------
             weight
      model1 0.000 
      model2 1.000 
      model3 0.000 

# tis throws correct errors and warnings

    Code
      psis(-LLarr, r_eff = r_eff_arr)
    Message
      Replacing NAs in `r_eff` with 1s
    Output
      Computed from 1000 by 32 log-weights matrix.
      MCSE and ESS estimates assume MCMC draws (r_eff in [0.6, 1.0]).
      
      All Pareto k estimates are good (k < 0.67).
      See help('pareto-k-diagnostic') for details.

# tis_loo and sis_loo are returned

    Code
      print(loo_tis)
    Output
      
      Computed from 1000 by 32 log-likelihood matrix using tis_loo .
      
               Estimate  SE
      elpd_loo    -83.6 4.3
      p_loo         3.3 1.2
      looic       167.2 8.6
      ------

---

    Code
      print(loo_sis)
    Output
      
      Computed from 1000 by 32 log-likelihood matrix using sis_loo .
      
               Estimate  SE
      elpd_loo    -83.6 4.3
      p_loo         3.3 1.2
      looic       167.2 8.6
      ------

# loo_compare_subsample

    Code
      lcss <- loo:::loo_compare.psis_loo_ss_list(x = list(lss1, lss2, lss3))
    Condition
      Warning:
      Different subsamples in 'model3' and 'model2'. Naive diff SE is used.
      Warning:
      Different subsamples in 'model3' and 'model1'. Naive diff SE is used.

---

    Code
      lcssapi <- loo_compare(lss1, lss2, lss3)
    Condition
      Warning:
      Different subsamples in 'model3' and 'model2'. Naive diff SE is used.
      Warning:
      Different subsamples in 'model3' and 'model1'. Naive diff SE is used.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_subsample.R
\name{.compute_point_estimate}
\alias{.compute_point_estimate}
\alias{.compute_point_estimate.matrix}
\alias{.compute_point_estimate.default}
\title{Compute a point estimate from a draws object}
\usage{
.compute_point_estimate(draws)

\method{.compute_point_estimate}{matrix}(draws)

\method{.compute_point_estimate}{default}(draws)
}
\arguments{
\item{draws}{A draws object with draws from the posterior.}
}
\value{
A 1 by P matrix with point estimates from a draws object.
}
\description{
Compute a point estimate from a draws object
}
\details{
This is a generic function to compute point estimates from draws
objects. The function is internal and should only be used by developers to
enable \code{\link[=loo_subsample]{loo_subsample()}} for arbitrary draws objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_subsample.R
\name{obs_idx}
\alias{obs_idx}
\title{Get observation indices used in subsampling}
\usage{
obs_idx(x, rep = TRUE)
}
\arguments{
\item{x}{A \code{psis_loo_ss} object.}

\item{rep}{If sampling with replacement is used, an observation can have
multiple samples and these are then repeated in the returned object if
\code{rep=TRUE} (e.g., a vector \code{c(1,1,2)} indicates that observation 1 has been
subsampled two times). If \code{rep=FALSE} only the unique indices are returned.}
}
\value{
An integer vector.
}
\description{
Get observation indices used in subsampling
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effective_sample_sizes.R
\name{relative_eff}
\alias{relative_eff}
\alias{relative_eff.default}
\alias{relative_eff.matrix}
\alias{relative_eff.array}
\alias{relative_eff.function}
\alias{relative_eff.importance_sampling}
\title{Convenience function for computing relative efficiencies}
\usage{
relative_eff(x, ...)

\method{relative_eff}{default}(x, chain_id, ...)

\method{relative_eff}{matrix}(x, chain_id, ..., cores = getOption("mc.cores", 1))

\method{relative_eff}{array}(x, ..., cores = getOption("mc.cores", 1))

\method{relative_eff}{`function`}(
  x,
  chain_id,
  ...,
  cores = getOption("mc.cores", 1),
  data = NULL,
  draws = NULL
)

\method{relative_eff}{importance_sampling}(x, ...)
}
\arguments{
\item{x}{A vector, matrix, 3-D array, or function. See the \strong{Methods (by
class)} section below for details on specifying \code{x}, but where
"log-likelihood" is mentioned replace it with one of the following
depending on the use case:
\itemize{
\item For use with the \code{\link[=loo]{loo()}} function, the values in \code{x} (or generated by
\code{x}, if a function) should be \strong{likelihood} values
(i.e., \code{exp(log_lik)}), not on the log scale.
\item For generic use with \code{\link[=psis]{psis()}}, the values in \code{x} should be the reciprocal
of the importance ratios (i.e., \code{exp(-log_ratios)}).
}}

\item{chain_id}{A vector of length \code{NROW(x)} containing MCMC chain
indexes for each each row of \code{x} (if a matrix) or each value in
\code{x} (if a vector). No \code{chain_id} is needed if \code{x} is a 3-D
array. If there are \code{C} chains then valid chain indexes are values
in \code{1:C}.}

\item{cores}{The number of cores to use for parallelization.}

\item{data, draws, ...}{Same as for the \code{\link[=loo]{loo()}} function method.}
}
\value{
A vector of relative effective sample sizes.
}
\description{
\code{relative_eff()} computes the the MCMC effective sample size divided by
the total sample size.
}
\section{Methods (by class)}{
\itemize{
\item \code{relative_eff(default)}: A vector of length \eqn{S} (posterior sample size).

\item \code{relative_eff(matrix)}: An \eqn{S} by \eqn{N} matrix, where \eqn{S} is the size
of the posterior sample (with all chains merged) and \eqn{N} is the number
of data points.

\item \code{relative_eff(array)}: An \eqn{I} by \eqn{C} by \eqn{N} array, where \eqn{I}
is the number of MCMC iterations per chain, \eqn{C} is the number of
chains, and \eqn{N} is the number of data points.

\item \code{relative_eff(`function`)}: A function \code{f()} that takes arguments \code{data_i} and \code{draws} and returns a
vector containing the log-likelihood for a single observation \code{i} evaluated
at each posterior draw. The function should be written such that, for each
observation \code{i} in \code{1:N}, evaluating

\if{html}{\out{<div class="sourceCode">}}\preformatted{f(data_i = data[i,, drop=FALSE], draws = draws)
}\if{html}{\out{</div>}}

results in a vector of length \code{S} (size of posterior sample). The
log-likelihood function can also have additional arguments but \code{data_i} and
\code{draws} are required.

If using the function method then the arguments \code{data} and \code{draws} must also
be specified in the call to \code{loo()}:
\itemize{
\item \code{data}: A data frame or matrix containing the data (e.g.
observed outcome and predictors) needed to compute the pointwise
log-likelihood. For each observation \code{i}, the \code{i}th row of
\code{data} will be passed to the \code{data_i} argument of the
log-likelihood function.
\item \code{draws}: An object containing the posterior draws for any
parameters needed to compute the pointwise log-likelihood. Unlike
\code{data}, which is indexed by observation, for each observation the
entire object \code{draws} will be passed to the \code{draws} argument of
the log-likelihood function.
\item The \code{...} can be used if your log-likelihood function takes additional
arguments. These arguments are used like the \code{draws} argument in that they
are recycled for each observation.
}

\item \code{relative_eff(importance_sampling)}: If \code{x} is an object of class \code{"psis"}, \code{relative_eff()} simply returns
the \code{r_eff} attribute of \code{x}.

}}
\examples{
LLarr <- example_loglik_array()
LLmat <- example_loglik_matrix()
dim(LLarr)
dim(LLmat)

rel_n_eff_1 <- relative_eff(exp(LLarr))
rel_n_eff_2 <- relative_eff(exp(LLmat), chain_id = rep(1:2, each = 500))
all.equal(rel_n_eff_1, rel_n_eff_2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_moment_matching.R
\name{loo_moment_match}
\alias{loo_moment_match}
\alias{loo_moment_match.default}
\title{Moment matching for efficient approximate leave-one-out cross-validation (LOO)}
\usage{
loo_moment_match(x, ...)

\method{loo_moment_match}{default}(
  x,
  loo,
  post_draws,
  log_lik_i,
  unconstrain_pars,
  log_prob_upars,
  log_lik_i_upars,
  max_iters = 30L,
  k_threshold = NULL,
  split = TRUE,
  cov = TRUE,
  cores = getOption("mc.cores", 1),
  ...
)
}
\arguments{
\item{x}{A fitted model object.}

\item{...}{Further arguments passed to the custom functions documented above.}

\item{loo}{A loo object to be modified.}

\item{post_draws}{A function the takes \code{x} as the first argument and returns
a matrix of posterior draws of the model parameters.}

\item{log_lik_i}{A function that takes \code{x} and \code{i} and returns a matrix (one
column per chain) or a vector (all chains stacked) of log-likelihood draws
of the \code{i}th observation based on the model \code{x}. If the draws are obtained
using MCMC, the matrix with MCMC chains separated is preferred.}

\item{unconstrain_pars}{A function that takes arguments \code{x}, and \code{pars} and
returns posterior draws on the unconstrained space based on the posterior
draws on the constrained space passed via \code{pars}.}

\item{log_prob_upars}{A function that takes arguments \code{x} and \code{upars} and
returns a matrix of log-posterior density values of the unconstrained
posterior draws passed via \code{upars}.}

\item{log_lik_i_upars}{A function that takes arguments \code{x}, \code{upars}, and \code{i}
and returns a vector of log-likelihood draws of the \code{i}th observation based
on the unconstrained posterior draws passed via \code{upars}.}

\item{max_iters}{Maximum number of moment matching iterations. Usually this
does not need to be modified. If the maximum number of iterations is
reached, there will be a warning, and increasing \code{max_iters} may improve
accuracy.}

\item{k_threshold}{Threshold value for Pareto k values above which the moment
matching algorithm is used. The default value is \code{min(1 - 1/log10(S), 0.7)},
where \code{S} is the sample size.}

\item{split}{Logical; Indicate whether to do the split transformation or not
at the end of moment matching for each LOO fold.}

\item{cov}{Logical; Indicate whether to match the covariance matrix of the
samples or not. If \code{FALSE}, only the mean and marginal variances are
matched.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}
}
\value{
The \code{loo_moment_match()} methods return an updated \code{loo} object. The
structure of the updated \code{loo} object is similar, but the method also
stores the original Pareto k diagnostic values in the diagnostics field.
}
\description{
Moment matching algorithm for updating a loo object when Pareto k estimates
are large.
}
\details{
The \code{loo_moment_match()} function is an S3 generic and we provide a
default method that takes as arguments user-specified functions
\code{post_draws}, \code{log_lik_i}, \code{unconstrain_pars}, \code{log_prob_upars}, and
\code{log_lik_i_upars}. All of these functions should take \code{...}. as an argument
in addition to those specified for each function.
}
\section{Methods (by class)}{
\itemize{
\item \code{loo_moment_match(default)}: A default method that takes as arguments a
user-specified model object \code{x}, a \code{loo} object and user-specified
functions \code{post_draws}, \code{log_lik_i}, \code{unconstrain_pars}, \code{log_prob_upars},
and \code{log_lik_i_upars}.

}}
\examples{
# See the vignette for loo_moment_match()
}
\references{
Paananen, T., Piironen, J., Buerkner, P.-C., Vehtari, A. (2021).
Implicitly adaptive importance sampling. \emph{Statistics and Computing}, 31, 16.
doi:10.1007/s11222-020-09982-2. arXiv preprint arXiv:1906.08850.
}
\seealso{
\code{\link[=loo]{loo()}}, \code{\link[=loo_moment_match_split]{loo_moment_match_split()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psislw.R
\name{psislw}
\alias{psislw}
\title{Pareto smoothed importance sampling (deprecated, old version)}
\usage{
psislw(
  lw,
  wcp = 0.2,
  wtrunc = 3/4,
  cores = getOption("mc.cores", 1),
  llfun = NULL,
  llargs = NULL,
  ...
)
}
\arguments{
\item{lw}{A matrix or vector of log weights. For computing LOO, \code{lw = -log_lik}, the \emph{negative} of an \eqn{S} (simulations) by \eqn{N} (data
points) pointwise log-likelihood matrix.}

\item{wcp}{The proportion of importance weights to use for the generalized
Pareto fit. The \code{100*wcp}\\% largest weights are used as the sample
from which to estimate the parameters of the generalized Pareto
distribution.}

\item{wtrunc}{For truncating very large weights to \eqn{S}^\code{wtrunc}. Set
to zero for no truncation.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}, the old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until it is
removed. \strong{As of version 2.0.0, the default is now 1 core if
\code{mc.cores} is not set, but we recommend using as many (or close to as
many) cores as possible.}}

\item{llfun, llargs}{See \code{\link[=loo.function]{loo.function()}}.}

\item{...}{Ignored when \code{psislw()} is called directly. The \code{...} is
only used internally when \code{psislw()} is called by the \code{\link[=loo]{loo()}}
function.}
}
\value{
A named list with components \code{lw_smooth} (modified log weights) and
\code{pareto_k} (estimated generalized Pareto shape parameter(s) k).
}
\description{
As of version \verb{2.0.0} this function is \strong{deprecated}. Please use the
\code{\link[=psis]{psis()}} function for the new PSIS algorithm.
}
\references{
Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}
\seealso{
\link{pareto-k-diagnostic} for PSIS diagnostics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/waic.R
\name{waic}
\alias{waic}
\alias{waic.array}
\alias{waic.matrix}
\alias{waic.function}
\alias{is.waic}
\title{Widely applicable information criterion (WAIC)}
\usage{
waic(x, ...)

\method{waic}{array}(x, ...)

\method{waic}{matrix}(x, ...)

\method{waic}{`function`}(x, ..., data = NULL, draws = NULL)

is.waic(x)
}
\arguments{
\item{x}{A log-likelihood array, matrix, or function. The \strong{Methods (by class)}
section, below, has detailed descriptions of how to specify the inputs for
each method.}

\item{draws, data, ...}{For the function method only. See the
\strong{Methods (by class)} section below for details on these arguments.}
}
\value{
A named list (of class \code{c("waic", "loo")}) with components:

\describe{
\item{\code{estimates}}{
A matrix with two columns (\code{"Estimate"}, \code{"SE"}) and three
rows (\code{"elpd_waic"}, \code{"p_waic"}, \code{"waic"}). This contains
point estimates and standard errors of the expected log pointwise predictive
density (\code{elpd_waic}), the effective number of parameters
(\code{p_waic}) and the information criterion \code{waic} (which is just
\code{-2 * elpd_waic}, i.e., converted to deviance scale).
}
\item{\code{pointwise}}{
A matrix with three columns (and number of rows equal to the number of
observations) containing the pointwise contributions of each of the above
measures (\code{elpd_waic}, \code{p_waic}, \code{waic}).
}
}
}
\description{
The \code{waic()} methods can be used to compute WAIC from the pointwise
log-likelihood. However, we recommend LOO-CV using PSIS (as implemented by
the \code{\link[=loo]{loo()}} function) because PSIS provides useful diagnostics as well as
effective sample size and Monte Carlo estimates.
}
\section{Methods (by class)}{
\itemize{
\item \code{waic(array)}: An \eqn{I} by \eqn{C} by \eqn{N} array, where \eqn{I}
is the number of MCMC iterations per chain, \eqn{C} is the number of
chains, and \eqn{N} is the number of data points.

\item \code{waic(matrix)}: An \eqn{S} by \eqn{N} matrix, where \eqn{S} is the size
of the posterior sample (with all chains merged) and \eqn{N} is the number
of data points.

\item \code{waic(`function`)}: A function \code{f()} that takes arguments \code{data_i} and \code{draws} and returns a
vector containing the log-likelihood for a single observation \code{i} evaluated
at each posterior draw. The function should be written such that, for each
observation \code{i} in \code{1:N}, evaluating

\if{html}{\out{<div class="sourceCode">}}\preformatted{f(data_i = data[i,, drop=FALSE], draws = draws)
}\if{html}{\out{</div>}}

results in a vector of length \code{S} (size of posterior sample). The
log-likelihood function can also have additional arguments but \code{data_i} and
\code{draws} are required.

If using the function method then the arguments \code{data} and \code{draws} must also
be specified in the call to \code{loo()}:
\itemize{
\item \code{data}: A data frame or matrix containing the data (e.g.
observed outcome and predictors) needed to compute the pointwise
log-likelihood. For each observation \code{i}, the \code{i}th row of
\code{data} will be passed to the \code{data_i} argument of the
log-likelihood function.
\item \code{draws}: An object containing the posterior draws for any
parameters needed to compute the pointwise log-likelihood. Unlike
\code{data}, which is indexed by observation, for each observation the
entire object \code{draws} will be passed to the \code{draws} argument of
the log-likelihood function.
\item The \code{...} can be used if your log-likelihood function takes additional
arguments. These arguments are used like the \code{draws} argument in that they
are recycled for each observation.
}

}}
\examples{
### Array and matrix methods
LLarr <- example_loglik_array()
dim(LLarr)

LLmat <- example_loglik_matrix()
dim(LLmat)

waic_arr <- waic(LLarr)
waic_mat <- waic(LLmat)
identical(waic_arr, waic_mat)


\dontrun{
log_lik1 <- extract_log_lik(stanfit1)
log_lik2 <- extract_log_lik(stanfit2)
(waic1 <- waic(log_lik1))
(waic2 <- waic(log_lik2))
print(compare(waic1, waic2), digits = 2)
}

}
\references{
Watanabe, S. (2010). Asymptotic equivalence of Bayes cross validation and
widely applicable information criterion in singular learning theory.
\emph{Journal of Machine Learning Research} \strong{11}, 3571-3594.

Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}
\seealso{
\itemize{
\item The \strong{loo} package \href{https://mc-stan.org/loo/articles/}{vignettes} and
Vehtari, Gelman, and Gabry (2017) and Vehtari, Simpson, Gelman, Yao,
and Gabry (2024) for more details on why we prefer \code{loo()} to \code{waic()}.
\item \code{\link[=loo_compare]{loo_compare()}} for comparing models on approximate LOO-CV or WAIC.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.loo}
\alias{print.loo}
\alias{print.waic}
\alias{print.psis_loo}
\alias{print.importance_sampling_loo}
\alias{print.psis_loo_ap}
\alias{print.psis}
\alias{print.importance_sampling}
\alias{print.kfold}
\title{Print methods}
\usage{
\method{print}{loo}(x, digits = 1, ...)

\method{print}{waic}(x, digits = 1, ...)

\method{print}{psis_loo}(x, digits = 1, plot_k = FALSE, ...)

\method{print}{importance_sampling_loo}(x, digits = 1, plot_k = FALSE, ...)

\method{print}{psis_loo_ap}(x, digits = 1, plot_k = FALSE, ...)

\method{print}{psis}(x, digits = 1, plot_k = FALSE, ...)

\method{print}{importance_sampling}(x, digits = 1, plot_k = FALSE, ...)

\method{print}{kfold}(x, digits = 1, plot_k = FALSE, ...)
}
\arguments{
\item{x}{An object returned by \code{\link[=loo]{loo()}}, \code{\link[=psis]{psis()}}, or \code{\link[=waic]{waic()}}.}

\item{digits}{An integer passed to \code{\link[base:round]{base::round()}}.}

\item{...}{Arguments passed to \code{\link[=plot.psis_loo]{plot.psis_loo()}} if \code{plot_k} is
\code{TRUE}.}

\item{plot_k}{Logical. If \code{TRUE} the estimates of the Pareto shape
parameter \eqn{k} are plotted. Ignored if \code{x} was generated by
\code{\link[=waic]{waic()}}. To just plot \eqn{k} without printing use the
\link[=pareto-k-diagnostic]{plot()} method for 'loo' objects.}
}
\value{
\code{x}, invisibly.
}
\description{
Print methods
}
\seealso{
\link{pareto-k-diagnostic}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/E_loo.R
\name{E_loo}
\alias{E_loo}
\alias{E_loo.default}
\alias{E_loo.matrix}
\title{Compute weighted expectations}
\usage{
E_loo(x, psis_object, ...)

\method{E_loo}{default}(
  x,
  psis_object,
  ...,
  type = c("mean", "variance", "sd", "quantile"),
  probs = NULL,
  log_ratios = NULL
)

\method{E_loo}{matrix}(
  x,
  psis_object,
  ...,
  type = c("mean", "variance", "sd", "quantile"),
  probs = NULL,
  log_ratios = NULL
)
}
\arguments{
\item{x}{A numeric vector or matrix.}

\item{psis_object}{An object returned by \code{\link[=psis]{psis()}}.}

\item{...}{Arguments passed to individual methods.}

\item{type}{The type of expectation to compute. The options are
\code{"mean"}, \code{"variance"}, \code{"sd"}, and \code{"quantile"}.}

\item{probs}{For computing quantiles, a vector of probabilities.}

\item{log_ratios}{Optionally, a vector or matrix (the same dimensions as \code{x})
of raw (not smoothed) log ratios. If working with log-likelihood values,
the log ratios are the \strong{negative} of those values. If \code{log_ratios} is
specified we are able to compute more accurate \link[=pareto-k-diagnostic]{Pareto k}
diagnostics specific to \code{E_loo()}.}
}
\value{
A named list with the following components:
\describe{
\item{\code{value}}{
The result of the computation.

For the matrix method, \code{value} is a vector with \code{ncol(x)}
elements, with one exception: when \code{type="quantile"} and
multiple values are specified in \code{probs} the \code{value} component of
the returned object is a \code{length(probs)} by \code{ncol(x)} matrix.

For the default/vector method the \code{value} component is scalar, with
one exception: when \code{type="quantile"} and multiple values
are specified in \code{probs} the \code{value} component is a vector with
\code{length(probs)} elements.
}
\item{\code{pareto_k}}{
Function-specific diagnostic.

For the matrix method it will be a vector of length \code{ncol(x)}
containing estimates of the shape parameter \eqn{k} of the
generalized Pareto distribution. For the default/vector method,
the estimate is a scalar. If \code{log_ratios} is not specified when
calling \code{E_loo()}, the smoothed log-weights are used to estimate
Pareto-k's, which may produce optimistic estimates.

For \code{type="mean"}, \code{type="var"}, and \code{type="sd"}, the returned Pareto-k is
usually the maximum of the Pareto-k's for the left and right tail of \eqn{hr}
and the right tail of \eqn{r}, where \eqn{r} is the importance ratio and
\eqn{h=x} for \code{type="mean"} and \eqn{h=x^2} for \code{type="var"} and \code{type="sd"}.
If \eqn{h} is binary, constant, or not finite, or if \code{type="quantile"}, the
returned Pareto-k is the Pareto-k for the right tail of \eqn{r}.
}
}
}
\description{
The \code{E_loo()} function computes weighted expectations (means, variances,
quantiles) using the importance weights obtained from the \link[=psis]{PSIS}
smoothing procedure. The expectations estimated by the \code{E_loo()} function
assume that the PSIS approximation is working well.
\strong{A small \link[=pareto-k-diagnostic]{Pareto k} estimate is necessary,
but not sufficient, for \code{E_loo()} to give reliable estimates}. If the
\code{log_ratios} argument is provided, \code{E_loo()} also computes a function
specific Pareto k diagnostic, which must also be small for a reliable
estimate. See more details below.
}
\examples{
\donttest{
if (requireNamespace("rstanarm", quietly = TRUE)) {
# Use rstanarm package to quickly fit a model and get both a log-likelihood
# matrix and draws from the posterior predictive distribution
library("rstanarm")

# data from help("lm")
ctl <- c(4.17,5.58,5.18,6.11,4.50,4.61,5.17,4.53,5.33,5.14)
trt <- c(4.81,4.17,4.41,3.59,5.87,3.83,6.03,4.89,4.32,4.69)
d <- data.frame(
  weight = c(ctl, trt),
  group = gl(2, 10, 20, labels = c("Ctl","Trt"))
)
fit <- stan_glm(weight ~ group, data = d, refresh = 0)
yrep <- posterior_predict(fit)
dim(yrep)

log_ratios <- -1 * log_lik(fit)
dim(log_ratios)

r_eff <- relative_eff(exp(-log_ratios), chain_id = rep(1:4, each = 1000))
psis_object <- psis(log_ratios, r_eff = r_eff, cores = 2)

E_loo(yrep, psis_object, type = "mean")
E_loo(yrep, psis_object, type = "var")
E_loo(yrep, psis_object, type = "sd")
E_loo(yrep, psis_object, type = "quantile", probs = 0.5) # median
E_loo(yrep, psis_object, type = "quantile", probs = c(0.1, 0.9))

# We can get more accurate Pareto k diagnostic if we also provide
# the log_ratios argument
E_loo(yrep, psis_object, type = "mean", log_ratios = log_ratios)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_subsample.R
\name{loo_subsample}
\alias{loo_subsample}
\alias{loo_subsample.function}
\title{Efficient approximate LOO-CV using subsampling}
\usage{
loo_subsample(x, ...)

\method{loo_subsample}{`function`}(
  x,
  ...,
  data = NULL,
  draws = NULL,
  observations = 400,
  log_p = NULL,
  log_g = NULL,
  r_eff = 1,
  save_psis = FALSE,
  cores = getOption("mc.cores", 1),
  loo_approximation = "plpd",
  loo_approximation_draws = NULL,
  estimator = "diff_srs",
  llgrad = NULL,
  llhess = NULL
)
}
\arguments{
\item{x}{A function. The \strong{Methods (by class)} section, below, has detailed
descriptions of how to specify the inputs.}

\item{data, draws, ...}{For \code{loo_subsample.function()}, these are the data,
posterior draws, and other arguments to pass to the log-likelihood
function. Note that for some \code{loo_approximation}s, the draws will be replaced
by the posteriors summary statistics to compute loo approximations. See
argument \code{loo_approximation} for details.}

\item{observations}{The subsample observations to use. The argument can take
four (4) types of arguments:
\itemize{
\item \code{NULL} to use all observations. The algorithm then just uses
standard \code{loo()} or \code{loo_approximate_posterior()}.
\item A single integer to specify the number of observations to be subsampled.
\item A vector of integers to provide the indices used to subset the data.
\emph{These observations need to be subsampled with the same scheme as given by
the \code{estimator} argument}.
\item A \code{psis_loo_ss} object to use the same observations that were used in a
previous call to \code{loo_subsample()}.
}}

\item{log_p, log_g}{Should be supplied only if approximate posterior draws are
used. The default (\code{NULL}) indicates draws are from "true" posterior (i.e.
using MCMC). If not \code{NULL} then they should be specified as described in
\code{\link[=loo_approximate_posterior]{loo_approximate_posterior()}}.}

\item{r_eff}{Vector of relative effective sample size estimates for the
likelihood (\code{exp(log_lik)}) of each observation. This is related to
the relative efficiency of estimating the normalizing term in
self-normalized importance sampling when using posterior draws obtained
with MCMC. If MCMC draws are used and \code{r_eff} is not provided then
the reported PSIS effective sample sizes and Monte Carlo error estimates
can be over-optimistic. If the posterior draws are (near) independent then
\code{r_eff=1} can be used. \code{r_eff} has to be a scalar (same value is used
for all observations) or a vector with length equal to the number of
observations. The default value is 1. See the \code{\link[=relative_eff]{relative_eff()}} helper
functions for help computing \code{r_eff}.}

\item{save_psis}{Should the \code{"psis"} object created internally by
\code{loo_subsample()} be saved in the returned object? See \code{\link[=loo]{loo()}} for details.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}

\item{loo_approximation}{What type of approximation of the loo_i's should be used?
The default is \code{"plpd"} (the log predictive density using the posterior expectation).
There are six different methods implemented to approximate loo_i's
(see the references for more details):
\itemize{
\item \code{"plpd"}: uses the lpd based on point estimates (i.e., \eqn{p(y_i|\hat{\theta})}).
\item \code{"lpd"}: uses the lpds (i,e., \eqn{p(y_i|y)}).
\item \code{"tis"}: uses truncated importance sampling to approximate PSIS-LOO.
\item \code{"waic"}: uses waic (i.e., \eqn{p(y_i|y) - p_{waic}}).
\item \code{"waic_grad_marginal"}: uses waic approximation using first order delta
method and posterior marginal variances to approximate \eqn{p_{waic}} (ie.
\eqn{p(y_i|\hat{\theta})}-p_waic_grad_marginal). Requires gradient of
likelihood function.
\item \code{"waic_grad"}: uses waic approximation using first order delta method and
posterior covariance to approximate \eqn{p_{waic}} (ie.
\eqn{p(y_i|\hat{\theta})}-p_waic_grad). Requires gradient of likelihood
function.
\item \code{"waic_hess"}: uses waic approximation using second order delta method and
posterior covariance to approximate \eqn{p_{waic}} (ie.
\eqn{p(y_i|\hat{\theta})}-p_waic_grad). Requires gradient and Hessian of
likelihood function.
}

As point estimates of \eqn{\hat{\theta}}, the posterior expectations
of the parameters are used.}

\item{loo_approximation_draws}{The number of posterior draws used when
integrating over the posterior. This is used if \code{loo_approximation} is set
to \code{"lpd"}, \code{"waic"}, or \code{"tis"}.}

\item{estimator}{How should \code{elpd_loo}, \code{p_loo} and \code{looic} be estimated?
The default is \code{"diff_srs"}.
\itemize{
\item \code{"diff_srs"}: uses the difference estimator with simple random sampling
without replacement (srs). \code{p_loo} is estimated using standard srs.
(Magnusson et al., 2020)
\item \code{"hh"}: uses the Hansen-Hurwitz estimator with sampling with replacement
proportional to size, where \code{abs} of loo_approximation is used as size.
(Magnusson et al., 2019)
\item \code{"srs"}: uses simple random sampling and ordinary estimation.
}}

\item{llgrad}{The gradient of the log-likelihood. This
is only used when \code{loo_approximation} is \code{"waic_grad"},
\code{"waic_grad_marginal"}, or \code{"waic_hess"}. The default is \code{NULL}.}

\item{llhess}{The Hessian of the log-likelihood. This is only used
with \code{loo_approximation = "waic_hess"}. The default is \code{NULL}.}
}
\value{
\code{loo_subsample()} returns a named list with class \code{c("psis_loo_ss", "psis_loo", "loo")}. This has the same structure as objects returned by
\code{\link[=loo]{loo()}} but with the additional slot:
\itemize{
\item \code{loo_subsampling}: A list with two vectors, \code{log_p} and \code{log_g}, of the
same length containing the posterior density and the approximation density
for the individual draws.
}
}
\description{
Efficiently approximates leave-one-out cross-validation (LOO-CV) via subsampling.
Less costly and more approximate computation is made for all LOO-fold,
and more costly and accurate computations are made only for m<N LOO-folds.
}
\details{
The \code{loo_subsample()} function is an S3 generic and a methods is
currently provided for log-likelihood functions. The implementation works
for both MCMC and for posterior approximations where it is possible to
compute the log density for the approximation.
}
\section{Methods (by class)}{
\itemize{
\item \code{loo_subsample(`function`)}: A function \code{f()} that takes arguments \code{data_i} and \code{draws} and returns a
vector containing the log-likelihood for a single observation \code{i} evaluated
at each posterior draw. The function should be written such that, for each
observation \code{i} in \code{1:N}, evaluating

\if{html}{\out{<div class="sourceCode">}}\preformatted{f(data_i = data[i,, drop=FALSE], draws = draws)
}\if{html}{\out{</div>}}

results in a vector of length \code{S} (size of posterior sample). The
log-likelihood function can also have additional arguments but \code{data_i} and
\code{draws} are required.

If using the function method then the arguments \code{data} and \code{draws} must also
be specified in the call to \code{loo()}:
\itemize{
\item \code{data}: A data frame or matrix containing the data (e.g.
observed outcome and predictors) needed to compute the pointwise
log-likelihood. For each observation \code{i}, the \code{i}th row of
\code{data} will be passed to the \code{data_i} argument of the
log-likelihood function.
\item \code{draws}: An object containing the posterior draws for any
parameters needed to compute the pointwise log-likelihood. Unlike
\code{data}, which is indexed by observation, for each observation the
entire object \code{draws} will be passed to the \code{draws} argument of
the log-likelihood function.
\item The \code{...} can be used if your log-likelihood function takes additional
arguments. These arguments are used like the \code{draws} argument in that they
are recycled for each observation.
}

}}
\references{
Magnusson, M., Riis Andersen, M., Jonasson, J. and Vehtari, A. (2019).
Leave-One-Out Cross-Validation for Large Data.
In \emph{Thirty-sixth International Conference on Machine Learning},
PMLR 97:4244-4253.

Magnusson, M., Riis Andersen, M., Jonasson, J. and Vehtari, A. (2020).
Leave-One-Out Cross-Validation for Model Comparison in Large Data.
In \emph{Proceedings of the 23rd International Conference on Artificial
Intelligence and Statistics (AISTATS)}, PMLR 108:341-351.
}
\seealso{
\code{\link[=loo]{loo()}}, \code{\link[=psis]{psis()}}, \code{\link[=loo_compare]{loo_compare()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_subsample.R
\name{.ndraws}
\alias{.ndraws}
\alias{.ndraws.matrix}
\alias{.ndraws.default}
\title{The number of posterior draws in a draws object.}
\usage{
.ndraws(x)

\method{.ndraws}{matrix}(x)

\method{.ndraws}{default}(x)
}
\arguments{
\item{x}{A draws object with posterior draws.}
}
\value{
An integer with the number of draws.
}
\description{
The number of posterior draws in a draws object.
}
\details{
This is a generic function to return the total number of draws from
an arbitrary draws objects. The function is internal and should only be
used by developers to enable \code{\link[=loo_subsample]{loo_subsample()}} for arbitrary draws objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elpd.R
\name{elpd}
\alias{elpd}
\alias{elpd.array}
\alias{elpd.matrix}
\title{Generic (expected) log-predictive density}
\usage{
elpd(x, ...)

\method{elpd}{array}(x, ...)

\method{elpd}{matrix}(x, ...)
}
\arguments{
\item{x}{A log-likelihood array or matrix. The \strong{Methods (by class)}
section, below, has detailed descriptions of how to specify the inputs for
each method.}

\item{...}{Currently ignored.}
}
\description{
The \code{elpd()} methods for arrays and matrices can compute the expected log
pointwise predictive density for a new dataset or the log pointwise
predictive density of the observed data (an overestimate of the elpd).
}
\details{
The \code{elpd()} function is an S3 generic and methods are provided for
3-D pointwise log-likelihood arrays and matrices.
}
\section{Methods (by class)}{
\itemize{
\item \code{elpd(array)}: An \eqn{I} by \eqn{C} by \eqn{N} array, where \eqn{I}
is the number of MCMC iterations per chain, \eqn{C} is the number of
chains, and \eqn{N} is the number of data points.

\item \code{elpd(matrix)}: An \eqn{S} by \eqn{N} matrix, where \eqn{S} is the size
of the posterior sample (with all chains merged) and \eqn{N} is the number
of data points.

}}
\examples{
# Calculate the lpd of the observed data
LLarr <- example_loglik_array()
elpd(LLarr)

}
\seealso{
The vignette \emph{Holdout validation and K-fold cross-validation of Stan
programs with the loo package} for demonstrations of using the \code{elpd()}
methods.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pointwise.R
\name{pointwise}
\alias{pointwise}
\alias{pointwise.loo}
\title{Convenience function for extracting pointwise estimates}
\usage{
pointwise(x, estimate, ...)

\method{pointwise}{loo}(x, estimate, ...)
}
\arguments{
\item{x}{A \code{loo} object, for example one returned by \code{\link[=loo]{loo()}},
\code{\link[=loo_subsample]{loo_subsample()}}, \code{\link[=loo_approximate_posterior]{loo_approximate_posterior()}}, \code{\link[=loo_moment_match]{loo_moment_match()}}, etc.}

\item{estimate}{Which pointwise estimate to return. By default all are
returned. The objects returned by the different functions (\code{\link[=loo]{loo()}},
\code{\link[=loo_subsample]{loo_subsample()}}, etc.) have slightly different estimates available.
Typically at a minimum the estimates \code{elpd_loo}, \code{looic}, \code{mcse_elpd_loo},
\code{p_loo}, and \code{influence_pareto_k} will be available, but there may be
others.}

\item{...}{Currently ignored.}
}
\value{
A vector of length equal to the number of observations.
}
\description{
Convenience function for extracting pointwise estimates
}
\examples{
x <- loo(example_loglik_array())
pointwise(x, "elpd_loo")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{old-extractors}
\alias{old-extractors}
\alias{[.loo}
\alias{[[.loo}
\alias{$.loo}
\title{Extractor methods}
\usage{
\method{[}{loo}(x, i)

\method{[[}{loo}(x, i, exact = TRUE)

\method{$}{loo}(x, name)
}
\arguments{
\item{x, i, exact, name}{See \link{Extract}.}
}
\description{
These are only defined in order to deprecate with a warning (rather than
remove and break backwards compatibility) the old way of accessing the point
estimates in a \code{"psis_loo"} or \code{"psis"} object. The new way as of
v2.0.0 is to get them from the \code{"estimates"} component of the object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example_log_lik_array.R
\name{example_loglik_array}
\alias{example_loglik_array}
\alias{example_loglik_matrix}
\title{Objects to use in examples and tests}
\usage{
example_loglik_array()

example_loglik_matrix()
}
\value{
\code{example_loglik_array()} returns a 500 (draws) x 2 (chains) x 32
(observations) pointwise log-likelihood array.

\code{example_loglik_matrix()} returns the same pointwise log-likelihood values
as \code{example_loglik_array()} but reshaped into a 1000 (draws*chains) x 32
(observations) matrix.
}
\description{
Example pointwise log-likelihood objects to use in demonstrations and tests.
See the \strong{Value} and \strong{Examples} sections below.
}
\examples{
LLarr <- example_loglik_array()
(dim_arr <- dim(LLarr))
LLmat <- example_loglik_matrix()
(dim_mat <- dim(LLmat))

all.equal(dim_mat[1], dim_arr[1] * dim_arr[2])
all.equal(dim_mat[2], dim_arr[3])

all.equal(LLarr[, 1, ], LLmat[1:500, ])
all.equal(LLarr[, 2, ], LLmat[501:1000, ])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psis_approximate_posterior.R
\name{psis_approximate_posterior}
\alias{psis_approximate_posterior}
\title{Diagnostics for Laplace and ADVI approximations and Laplace-loo and ADVI-loo}
\usage{
psis_approximate_posterior(
  log_p = NULL,
  log_g = NULL,
  log_liks = NULL,
  cores,
  save_psis,
  ...,
  log_q = NULL
)
}
\arguments{
\item{log_p}{The log-posterior (target) evaluated at S samples from the
proposal distribution (g). A vector of length S.}

\item{log_g}{The log-density (proposal) evaluated at S samples from the
proposal distribution (g). A vector of length S.}

\item{log_liks}{A log-likelihood matrix of size S * N, where N is the number
of observations and S is the number of samples from q. See
\code{\link[=loo.matrix]{loo.matrix()}} for details. Default is \code{NULL}. Then only the
posterior is evaluated using the k_hat diagnostic.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}

\item{save_psis}{Should the \code{psis} object created internally by \code{loo()} be
saved in the returned object? The \code{loo()} function calls \code{\link[=psis]{psis()}}
internally but by default discards the (potentially large) \code{psis} object
after using it to compute the LOO-CV summaries. Setting \code{save_psis=TRUE}
will add a \code{psis_object} component to the list returned by \code{loo}.
This is useful if you plan to use the \code{\link[=E_loo]{E_loo()}} function to compute
weighted expectations after running \code{loo}. Several functions in the
\pkg{bayesplot} package also accept \code{psis} objects.}

\item{log_q}{Deprecated argument name (the same as log_g).}
}
\value{
If log likelihoods are supplied, the function returns a \code{"loo"} object,
otherwise the function returns a \code{"psis"} object.
}
\description{
Diagnostics for Laplace and ADVI approximations and Laplace-loo and ADVI-loo
}
\references{
Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}
\seealso{
\code{\link[=loo]{loo()}} and \code{\link[=psis]{psis()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kfold-helpers.R
\name{kfold-helpers}
\alias{kfold-helpers}
\alias{kfold_split_random}
\alias{kfold_split_stratified}
\alias{kfold_split_grouped}
\title{Helper functions for K-fold cross-validation}
\usage{
kfold_split_random(K = 10, N = NULL)

kfold_split_stratified(K = 10, x = NULL)

kfold_split_grouped(K = 10, x = NULL)
}
\arguments{
\item{K}{The number of folds to use.}

\item{N}{The number of observations in the data.}

\item{x}{A discrete variable of length \code{N} with at least \code{K} levels (unique
values). Will be coerced to a \link{factor}.}
}
\value{
An integer vector of length \code{N} where each element is an index in \code{1:K}.
}
\description{
These functions can be used to generate indexes for use with
K-fold cross-validation. See the \strong{Details} section for explanations.
}
\details{
\code{kfold_split_random()} splits the data into \code{K} groups
of equal size (or roughly equal size).

For a categorical variable \code{x} \code{kfold_split_stratified()}
splits the observations into \code{K} groups ensuring that relative
category frequencies are approximately preserved.

For a grouping variable \code{x}, \code{kfold_split_grouped()} places
all observations in \code{x} from the same group/level together in
the same fold. The selection of which groups/levels go into which
fold (relevant when when there are more groups than folds) is
randomized.
}
\examples{
ids <- kfold_split_random(K = 5, N = 20)
print(ids)
table(ids)


x <- sample(c(0, 1), size = 200, replace = TRUE, prob = c(0.05, 0.95))
table(x)
ids <- kfold_split_stratified(K = 5, x = x)
print(ids)
table(ids, x)

grp <- gl(n = 50, k = 15, labels = state.name)
length(grp)
head(table(grp))

ids_10 <- kfold_split_grouped(K = 10, x = grp)
(tab_10 <- table(grp, ids_10))
colSums(tab_10)

ids_9 <- kfold_split_grouped(K = 9, x = grp)
(tab_9 <- table(grp, ids_9))
colSums(tab_9)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{loo}
\alias{loo}
\alias{loo.array}
\alias{loo.matrix}
\alias{loo.function}
\alias{loo_i}
\alias{is.loo}
\alias{is.psis_loo}
\title{Efficient approximate leave-one-out cross-validation (LOO)}
\usage{
loo(x, ...)

\method{loo}{array}(
  x,
  ...,
  r_eff = 1,
  save_psis = FALSE,
  cores = getOption("mc.cores", 1),
  is_method = c("psis", "tis", "sis")
)

\method{loo}{matrix}(
  x,
  ...,
  r_eff = 1,
  save_psis = FALSE,
  cores = getOption("mc.cores", 1),
  is_method = c("psis", "tis", "sis")
)

\method{loo}{`function`}(
  x,
  ...,
  data = NULL,
  draws = NULL,
  r_eff = 1,
  save_psis = FALSE,
  cores = getOption("mc.cores", 1),
  is_method = c("psis", "tis", "sis")
)

loo_i(i, llfun, ..., data = NULL, draws = NULL, r_eff = 1, is_method = "psis")

is.loo(x)

is.psis_loo(x)
}
\arguments{
\item{x}{A log-likelihood array, matrix, or function. The \strong{Methods (by class)}
section, below, has detailed descriptions of how to specify the inputs for
each method.}

\item{r_eff}{Vector of relative effective sample size estimates for the
likelihood (\code{exp(log_lik)}) of each observation. This is related to
the relative efficiency of estimating the normalizing term in
self-normalized importance sampling when using posterior draws obtained
with MCMC. If MCMC draws are used and \code{r_eff} is not provided then
the reported PSIS effective sample sizes and Monte Carlo error estimates
can be over-optimistic. If the posterior draws are (near) independent then
\code{r_eff=1} can be used. \code{r_eff} has to be a scalar (same value is used
for all observations) or a vector with length equal to the number of
observations. The default value is 1. See the \code{\link[=relative_eff]{relative_eff()}} helper
functions for help computing \code{r_eff}.}

\item{save_psis}{Should the \code{psis} object created internally by \code{loo()} be
saved in the returned object? The \code{loo()} function calls \code{\link[=psis]{psis()}}
internally but by default discards the (potentially large) \code{psis} object
after using it to compute the LOO-CV summaries. Setting \code{save_psis=TRUE}
will add a \code{psis_object} component to the list returned by \code{loo}.
This is useful if you plan to use the \code{\link[=E_loo]{E_loo()}} function to compute
weighted expectations after running \code{loo}. Several functions in the
\pkg{bayesplot} package also accept \code{psis} objects.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}

\item{is_method}{The importance sampling method to use. The following methods
are implemented:
\itemize{
\item \code{\link[=psis]{"psis"}}: Pareto-Smoothed Importance Sampling (PSIS). Default method.
\item \code{\link[=tis]{"tis"}}: Truncated Importance Sampling (TIS) with truncation at
\code{sqrt(S)}, where \code{S} is the number of posterior draws.
\item \code{\link[=sis]{"sis"}}: Standard Importance Sampling (SIS).
}}

\item{data, draws, ...}{For the \code{loo.function()} method and the \code{loo_i()}
function, these are the data, posterior draws, and other arguments to pass
to the log-likelihood function. See the \strong{Methods (by class)} section
below for details on how to specify these arguments.}

\item{i}{For \code{loo_i()}, an integer in \code{1:N}.}

\item{llfun}{For \code{loo_i()}, the same as \code{x} for the
\code{loo.function()} method. A log-likelihood function as described in the
\strong{Methods (by class)} section.}
}
\value{
The \code{loo()} methods return a named list with class
\code{c("psis_loo", "loo")} and components:
\describe{
\item{\code{estimates}}{
A matrix with two columns (\code{Estimate}, \code{SE}) and three rows (\code{elpd_loo},
\code{p_loo}, \code{looic}). This contains point estimates and standard errors of the
expected log pointwise predictive density (\code{\link[=loo-glossary]{elpd_loo}}), the
effective number of parameters (\code{\link[=loo-glossary]{p_loo}}) and the LOO
information criterion \code{looic} (which is just \code{-2 * elpd_loo}, i.e.,
converted to deviance scale).
}

\item{\code{pointwise}}{
A matrix with five columns (and number of rows equal to the number of
observations) containing the pointwise contributions of the measures
(\code{elpd_loo}, \code{mcse_elpd_loo}, \code{p_loo}, \code{looic}, \code{influence_pareto_k}).
in addition to the three measures in \code{estimates}, we also report
pointwise values of the Monte Carlo standard error of \code{\link[=loo-glossary]{elpd_loo}}
(\code{\link[=loo-glossary]{mcse_elpd_loo}}), and statistics describing the influence of
each observation on the posterior distribution (\code{influence_pareto_k}).
These are the estimates of the shape parameter \eqn{k} of the
generalized Pareto fit to the importance ratios for each leave-one-out
distribution (see the \link{pareto-k-diagnostic} page for details).
}

\item{\code{diagnostics}}{
A named list containing two vectors:
\itemize{
\item \code{pareto_k}: Importance sampling reliability diagnostics. By default,
these are equal to the \code{influence_pareto_k} in \code{pointwise}.
Some algorithms can improve importance sampling reliability and
modify these diagnostics. See the \link{pareto-k-diagnostic} page for details.
\item \code{n_eff}: PSIS effective sample size estimates.
}
}

\item{\code{psis_object}}{
This component will be \code{NULL} unless the \code{save_psis} argument is set to
\code{TRUE} when calling \code{loo()}. In that case \code{psis_object} will be the object
of class \code{"psis"} that is created when the \code{loo()} function calls \code{\link[=psis]{psis()}}
internally to do the PSIS procedure.
}
}

The \code{loo_i()} function returns a named list with components
\code{pointwise} and \code{diagnostics}. These components have the same
structure as the \code{pointwise} and \code{diagnostics} components of the
object returned by \code{loo()} except they contain results for only a single
observation.
}
\description{
The \code{loo()} methods for arrays, matrices, and functions compute PSIS-LOO
CV, efficient approximate leave-one-out (LOO) cross-validation for Bayesian
models using Pareto smoothed importance sampling (\link[=psis]{PSIS}). This is
an implementation of the methods described in Vehtari, Gelman, and Gabry
(2017) and Vehtari, Simpson, Gelman, Yao, and Gabry (2024).

The \code{loo_i()} function enables testing log-likelihood
functions for use with the \code{loo.function()} method.
}
\details{
The \code{loo()} function is an S3 generic and methods are provided for
3-D pointwise log-likelihood arrays, pointwise log-likelihood matrices, and
log-likelihood functions. The array and matrix methods are the most
convenient, but for models fit to very large datasets the \code{loo.function()}
method is more memory efficient and may be preferable.
}
\section{Methods (by class)}{
\itemize{
\item \code{loo(array)}: An \eqn{I} by \eqn{C} by \eqn{N} array, where \eqn{I}
is the number of MCMC iterations per chain, \eqn{C} is the number of
chains, and \eqn{N} is the number of data points.

\item \code{loo(matrix)}: An \eqn{S} by \eqn{N} matrix, where \eqn{S} is the size
of the posterior sample (with all chains merged) and \eqn{N} is the number
of data points.

\item \code{loo(`function`)}: A function \code{f()} that takes arguments \code{data_i} and \code{draws} and returns a
vector containing the log-likelihood for a single observation \code{i} evaluated
at each posterior draw. The function should be written such that, for each
observation \code{i} in \code{1:N}, evaluating

\if{html}{\out{<div class="sourceCode">}}\preformatted{f(data_i = data[i,, drop=FALSE], draws = draws)
}\if{html}{\out{</div>}}

results in a vector of length \code{S} (size of posterior sample). The
log-likelihood function can also have additional arguments but \code{data_i} and
\code{draws} are required.

If using the function method then the arguments \code{data} and \code{draws} must also
be specified in the call to \code{loo()}:
\itemize{
\item \code{data}: A data frame or matrix containing the data (e.g.
observed outcome and predictors) needed to compute the pointwise
log-likelihood. For each observation \code{i}, the \code{i}th row of
\code{data} will be passed to the \code{data_i} argument of the
log-likelihood function.
\item \code{draws}: An object containing the posterior draws for any
parameters needed to compute the pointwise log-likelihood. Unlike
\code{data}, which is indexed by observation, for each observation the
entire object \code{draws} will be passed to the \code{draws} argument of
the log-likelihood function.
\item The \code{...} can be used if your log-likelihood function takes additional
arguments. These arguments are used like the \code{draws} argument in that they
are recycled for each observation.
}

}}
\section{Defining \code{loo()} methods in a package}{
 Package developers can define
\code{loo()} methods for fitted models objects. See the example \code{loo.stanfit()}
method in the \strong{Examples} section below for an example of defining a
method that calls \code{loo.array()}. The \code{loo.stanreg()} method in the
\strong{rstanarm} package is an example of defining a method that calls
\code{loo.function()}.
}

\examples{
### Array and matrix methods (using example objects included with loo package)
# Array method
LLarr <- example_loglik_array()
rel_n_eff <- relative_eff(exp(LLarr))
loo(LLarr, r_eff = rel_n_eff, cores = 2)

# Matrix method
LLmat <- example_loglik_matrix()
rel_n_eff <- relative_eff(exp(LLmat), chain_id = rep(1:2, each = 500))
loo(LLmat, r_eff = rel_n_eff, cores = 2)


### Using log-likelihood function instead of array or matrix
set.seed(124)

# Simulate data and draw from posterior
N <- 50; K <- 10; S <- 100; a0 <- 3; b0 <- 2
p <- rbeta(1, a0, b0)
y <- rbinom(N, size = K, prob = p)
a <- a0 + sum(y); b <- b0 + N * K - sum(y)
fake_posterior <- as.matrix(rbeta(S, a, b))
dim(fake_posterior) # S x 1
fake_data <- data.frame(y,K)
dim(fake_data) # N x 2

llfun <- function(data_i, draws) {
  # each time called internally within loo the arguments will be equal to:
  # data_i: ith row of fake_data (fake_data[i,, drop=FALSE])
  # draws: entire fake_posterior matrix
  dbinom(data_i$y, size = data_i$K, prob = draws, log = TRUE)
}

# Use the loo_i function to check that llfun works on a single observation
# before running on all obs. For example, using the 3rd obs in the data:
loo_3 <- loo_i(i = 3, llfun = llfun, data = fake_data, draws = fake_posterior)
print(loo_3$pointwise[, "elpd_loo"])

# Use loo.function method (default r_eff=1 is used as this posterior not obtained via MCMC)
loo_with_fn <- loo(llfun, draws = fake_posterior, data = fake_data)

# If we look at the elpd_loo contribution from the 3rd obs it should be the
# same as what we got above with the loo_i function and i=3:
print(loo_with_fn$pointwise[3, "elpd_loo"])
print(loo_3$pointwise[, "elpd_loo"])

# Check that the loo.matrix method gives same answer as loo.function method
log_lik_matrix <- sapply(1:N, function(i) {
  llfun(data_i = fake_data[i,, drop=FALSE], draws = fake_posterior)
})
loo_with_mat <- loo(log_lik_matrix)
all.equal(loo_with_mat$estimates, loo_with_fn$estimates) # should be TRUE!


\dontrun{
### For package developers: defining loo methods

# An example of a possible loo method for 'stanfit' objects (rstan package).
# A similar method is included in the rstan package.
# In order for users to be able to call loo(stanfit) instead of
# loo.stanfit(stanfit) the NAMESPACE needs to be handled appropriately
# (roxygen2 and devtools packages are good for that).
#
loo.stanfit <-
 function(x,
         pars = "log_lik",
         ...,
         save_psis = FALSE,
         cores = getOption("mc.cores", 1)) {
  stopifnot(length(pars) == 1L)
  LLarray <- loo::extract_log_lik(stanfit = x,
                                  parameter_name = pars,
                                  merge_chains = FALSE)
  r_eff <- loo::relative_eff(x = exp(LLarray), cores = cores)
  loo::loo.array(LLarray,
                 r_eff = r_eff,
                 cores = cores,
                 save_psis = save_psis)
}
}


}
\references{
Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}
\seealso{
\itemize{
\item The \strong{loo} package \href{https://mc-stan.org/loo/articles/index.html}{vignettes}
for demonstrations.
\item The \href{https://mc-stan.org/loo/articles/online-only/faq.html}{FAQ page} on
the \strong{loo} website for answers to frequently asked questions.
\item \code{\link[=psis]{psis()}} for the underlying Pareto Smoothed Importance Sampling (PSIS)
procedure used in the LOO-CV approximation.
\item \link{pareto-k-diagnostic} for convenience functions for looking at diagnostics.
\item \code{\link[=loo_compare]{loo_compare()}} for model comparison.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tis.R
\name{tis}
\alias{tis}
\alias{tis.array}
\alias{tis.matrix}
\alias{tis.default}
\title{Truncated importance sampling (TIS)}
\usage{
tis(log_ratios, ...)

\method{tis}{array}(log_ratios, ..., r_eff = 1, cores = getOption("mc.cores", 1))

\method{tis}{matrix}(log_ratios, ..., r_eff = 1, cores = getOption("mc.cores", 1))

\method{tis}{default}(log_ratios, ..., r_eff = 1)
}
\arguments{
\item{log_ratios}{An array, matrix, or vector of importance ratios on the log
scale (for Importance sampling LOO, these are \emph{negative} log-likelihood
values). See the \strong{Methods (by class)} section below for a detailed
description of how to specify the inputs for each method.}

\item{...}{Arguments passed on to the various methods.}

\item{r_eff}{Vector of relative effective sample size estimates containing
one element per observation. The values provided should be the relative
effective sample sizes of \code{1/exp(log_ratios)} (i.e., \code{1/ratios}).
This is related to the relative efficiency of estimating the normalizing
term in self-normalizing importance sampling. If \code{r_eff} is not
provided then the reported (T)IS effective sample sizes and Monte Carlo
error estimates can be over-optimistic. If the posterior draws are (near)
independent then \code{r_eff=1} can be used. \code{r_eff} has to be a scalar (same
value is used for all observations) or a vector with length equal to the
number of observations. The default value is 1. See the \code{\link[=relative_eff]{relative_eff()}}
helper function for computing \code{r_eff}.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}
}
\value{
The \code{tis()} methods return an object of class \code{"tis"},
which is a named list with the following components:

\describe{
\item{\code{log_weights}}{
Vector or matrix of smoothed (and truncated) but \emph{unnormalized} log
weights. To get normalized weights use the
\code{\link[=weights.importance_sampling]{weights()}} method provided for objects of
class \code{tis}.
}
\item{\code{diagnostics}}{
A named list containing one vector:
\itemize{
\item \code{pareto_k}: Not used in \code{tis}, all set to 0.
\item \code{n_eff}: Effective sample size estimates.
}
}
}

Objects of class \code{"tis"} also have the following \link{attributes}:
\describe{
\item{\code{norm_const_log}}{
Vector of precomputed values of \code{colLogSumExps(log_weights)} that are
used internally by the \code{\link[=weights]{weights()}}method to normalize the log weights.
}
\item{\code{r_eff}}{
If specified, the user's \code{r_eff} argument.
}
\item{\code{tail_len}}{
Not used for \code{tis}.
}
\item{\code{dims}}{
Integer vector of length 2 containing \code{S} (posterior sample size)
and \code{N} (number of observations).
}
\item{\code{method}}{
Method used for importance sampling, here \code{tis}.
}
}
}
\description{
Implementation of truncated (self-normalized) importance sampling (TIS),
truncated at S^(1/2) as recommended by Ionides (2008).
}
\section{Methods (by class)}{
\itemize{
\item \code{tis(array)}: An \eqn{I} by \eqn{C} by \eqn{N} array, where \eqn{I}
is the number of MCMC iterations per chain, \eqn{C} is the number of
chains, and \eqn{N} is the number of data points.

\item \code{tis(matrix)}: An \eqn{S} by \eqn{N} matrix, where \eqn{S} is the size
of the posterior sample (with all chains merged) and \eqn{N} is the number
of data points.

\item \code{tis(default)}: A vector of length \eqn{S} (posterior sample size).

}}
\examples{
log_ratios <- -1 * example_loglik_array()
r_eff <- relative_eff(exp(-log_ratios))
tis_result <- tis(log_ratios, r_eff = r_eff)
str(tis_result)

# extract smoothed weights
lw <- weights(tis_result) # default args are log=TRUE, normalize=TRUE
ulw <- weights(tis_result, normalize=FALSE) # unnormalized log-weights

w <- weights(tis_result, log=FALSE) # normalized weights (not log-weights)
uw <- weights(tis_result, log=FALSE, normalize = FALSE) # unnormalized weights

}
\references{
Ionides, Edward L. (2008). Truncated importance sampling.
\emph{Journal of Computational and Graphical Statistics} 17(2): 295--311.
}
\seealso{
\itemize{
\item \code{\link[=psis]{psis()}} for approximate LOO-CV using PSIS.
\item \code{\link[=loo]{loo()}} for approximate LOO-CV.
\item \link{pareto-k-diagnostic} for PSIS diagnostics.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo-package.R
\docType{package}
\name{loo-package}
\alias{loo-package}
\title{Efficient LOO-CV and WAIC for Bayesian models}
\description{
\if{html}{
  \figure{stanlogo.png}{options: width="50" alt="mc-stan.org"}
}
\emph{Stan Development Team}

This package implements the methods described in Vehtari, Gelman, and
Gabry (2017), Vehtari, Simpson, Gelman, Yao, and Gabry (2024), and
Yao et al. (2018). To get started see the \strong{loo} package
\href{https://mc-stan.org/loo/articles/index.html}{vignettes}, the
\code{\link[=loo]{loo()}} function for efficient approximate leave-one-out
cross-validation (LOO-CV), the \code{\link[=psis]{psis()}} function for the Pareto
smoothed importance sampling (PSIS) algorithm, or
\code{\link[=loo_model_weights]{loo_model_weights()}} for an implementation of Bayesian stacking of
predictive distributions from multiple models.
}
\details{
Leave-one-out cross-validation (LOO-CV) and the widely applicable
information criterion (WAIC) are methods for estimating pointwise
out-of-sample prediction accuracy from a fitted Bayesian model using the
log-likelihood evaluated at the posterior simulations of the parameter
values. LOO-CV and WAIC have various advantages over simpler estimates of
predictive error such as AIC and DIC but are less used in practice because
they involve additional computational steps. This package implements the
fast and stable computations for approximate LOO-CV laid out in Vehtari,
Gelman, and Gabry (2017). From existing posterior simulation draws, we
compute LOO-CV using Pareto smoothed importance sampling (PSIS; Vehtari,
Simpson, Gelman, Yao, and Gabry, 2024), a new procedure for stabilizing
and diagnosing importance weights. As a byproduct of our calculations,
we also obtain approximate standard errors for estimated predictive
errors and for comparing of predictive errors between two models.

We recommend PSIS-LOO-CV instead of WAIC, because PSIS provides useful
diagnostics and effective sample size and Monte Carlo standard error
estimates.
}
\references{
Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}

Yao, Y., Vehtari, A., Simpson, D., and Gelman, A. (2018) Using
stacking to average Bayesian predictive distributions.
\emph{Bayesian Analysis}, advance publication,  doi:10.1214/17-BA1091.
(\href{https://projecteuclid.org/euclid.ba/1516093227}{online}).

Magnusson, M., Riis Andersen, M., Jonasson, J. and Vehtari, A. (2019).
Leave-One-Out Cross-Validation for Large Data.
In \emph{Thirty-sixth International Conference on Machine Learning},
PMLR 97:4244-4253.

Magnusson, M., Riis Andersen, M., Jonasson, J. and Vehtari, A. (2020).
Leave-One-Out Cross-Validation for Model Comparison in Large Data.
In \emph{Proceedings of the 23rd International Conference on Artificial
Intelligence and Statistics (AISTATS)}, PMLR 108:341-351.

Epifani, I., MacEachern, S. N., and Peruggia, M. (2008). Case-deletion
importance sampling estimators: Central limit theorems and related results.
\emph{Electronic Journal of Statistics} \strong{2}, 774-806.

Gelfand, A. E. (1996). Model determination using sampling-based methods. In
\emph{Markov Chain Monte Carlo in Practice}, ed. W. R. Gilks, S. Richardson,
D. J. Spiegelhalter, 145-162. London: Chapman and Hall.

Gelfand, A. E., Dey, D. K., and Chang, H. (1992). Model determination using
predictive distributions with implementation via sampling-based methods. In
\emph{Bayesian Statistics 4}, ed. J. M. Bernardo, J. O. Berger, A. P. Dawid,
and A. F. M. Smith, 147-167. Oxford University Press.

Gelman, A., Hwang, J., and Vehtari, A. (2014). Understanding predictive
information criteria for Bayesian models. \emph{Statistics and Computing}
\strong{24}, 997-1016.

Ionides, E. L. (2008). Truncated importance sampling. \emph{Journal of
Computational and Graphical Statistics} \strong{17}, 295-311.

Koopman, S. J., Shephard, N., and Creal, D. (2009). Testing the assumptions
behind importance sampling. \emph{Journal of Econometrics} \strong{149}, 2-11.

Peruggia, M. (1997). On the variability of case-deletion importance sampling
weights in the Bayesian linear model. \emph{Journal of the American
Statistical Association} \strong{92}, 199-207.

Stan Development Team (2017). The Stan C++ Library, Version 2.17.0.
\url{https://mc-stan.org}.

Stan Development Team (2018). RStan: the R interface to Stan, Version 2.17.3.
\url{https://mc-stan.org}.

Watanabe, S. (2010). Asymptotic equivalence of Bayes cross validation and
widely applicable information criterion in singular learning theory.
\emph{Journal of Machine Learning Research} \strong{11}, 3571-3594.

Zhang, J., and Stephens, M. A. (2009). A new and efficient estimation method
for the generalized Pareto distribution. \emph{Technometrics} \strong{51},
316-325.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://mc-stan.org/loo/}
  \item \url{https://discourse.mc-stan.org}
  \item Report bugs at \url{https://github.com/stan-dev/loo/issues}
}

}
\author{
\strong{Maintainer}: Jonah Gabry \email{jgabry@gmail.com}

Authors:
\itemize{
  \item Jonah Gabry \email{jgabry@gmail.com}
  \item Aki Vehtari \email{Aki.Vehtari@aalto.fi}
  \item Måns Magnusson
  \item Yuling Yao
  \item Paul-Christian Bürkner
  \item Topi Paananen
  \item Andrew Gelman
}

Other contributors:
\itemize{
  \item Ben Goodrich [contributor]
  \item Juho Piironen [contributor]
  \item Bruno Nicenboim [contributor]
  \item Leevi Lindgren [contributor]
  \item Visruth Srimath Kandali [contributor]
  \item Florence Bockting [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_subsample.R
\name{update.psis_loo_ss}
\alias{update.psis_loo_ss}
\title{Update \code{psis_loo_ss} objects}
\usage{
\method{update}{psis_loo_ss}(
  object,
  ...,
  data = NULL,
  draws = NULL,
  observations = NULL,
  r_eff = 1,
  cores = getOption("mc.cores", 1),
  loo_approximation = NULL,
  loo_approximation_draws = NULL,
  llgrad = NULL,
  llhess = NULL
)
}
\arguments{
\item{object}{A \code{psis_loo_ss} object to update.}

\item{...}{Currently not used.}

\item{data, draws}{See \code{\link[=loo_subsample.function]{loo_subsample.function()}}.}

\item{observations}{The subsample observations to use. The argument can take
four (4) types of arguments:
\itemize{
\item \code{NULL} to use all observations. The algorithm then just uses
standard \code{loo()} or \code{loo_approximate_posterior()}.
\item A single integer to specify the number of observations to be subsampled.
\item A vector of integers to provide the indices used to subset the data.
\emph{These observations need to be subsampled with the same scheme as given by
the \code{estimator} argument}.
\item A \code{psis_loo_ss} object to use the same observations that were used in a
previous call to \code{loo_subsample()}.
}}

\item{r_eff}{Vector of relative effective sample size estimates for the
likelihood (\code{exp(log_lik)}) of each observation. This is related to
the relative efficiency of estimating the normalizing term in
self-normalized importance sampling when using posterior draws obtained
with MCMC. If MCMC draws are used and \code{r_eff} is not provided then
the reported PSIS effective sample sizes and Monte Carlo error estimates
can be over-optimistic. If the posterior draws are (near) independent then
\code{r_eff=1} can be used. \code{r_eff} has to be a scalar (same value is used
for all observations) or a vector with length equal to the number of
observations. The default value is 1. See the \code{\link[=relative_eff]{relative_eff()}} helper
functions for help computing \code{r_eff}.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}

\item{loo_approximation}{What type of approximation of the loo_i's should be used?
The default is \code{"plpd"} (the log predictive density using the posterior expectation).
There are six different methods implemented to approximate loo_i's
(see the references for more details):
\itemize{
\item \code{"plpd"}: uses the lpd based on point estimates (i.e., \eqn{p(y_i|\hat{\theta})}).
\item \code{"lpd"}: uses the lpds (i,e., \eqn{p(y_i|y)}).
\item \code{"tis"}: uses truncated importance sampling to approximate PSIS-LOO.
\item \code{"waic"}: uses waic (i.e., \eqn{p(y_i|y) - p_{waic}}).
\item \code{"waic_grad_marginal"}: uses waic approximation using first order delta
method and posterior marginal variances to approximate \eqn{p_{waic}} (ie.
\eqn{p(y_i|\hat{\theta})}-p_waic_grad_marginal). Requires gradient of
likelihood function.
\item \code{"waic_grad"}: uses waic approximation using first order delta method and
posterior covariance to approximate \eqn{p_{waic}} (ie.
\eqn{p(y_i|\hat{\theta})}-p_waic_grad). Requires gradient of likelihood
function.
\item \code{"waic_hess"}: uses waic approximation using second order delta method and
posterior covariance to approximate \eqn{p_{waic}} (ie.
\eqn{p(y_i|\hat{\theta})}-p_waic_grad). Requires gradient and Hessian of
likelihood function.
}

As point estimates of \eqn{\hat{\theta}}, the posterior expectations
of the parameters are used.}

\item{loo_approximation_draws}{The number of posterior draws used when
integrating over the posterior. This is used if \code{loo_approximation} is set
to \code{"lpd"}, \code{"waic"}, or \code{"tis"}.}

\item{llgrad}{The gradient of the log-likelihood. This
is only used when \code{loo_approximation} is \code{"waic_grad"},
\code{"waic_grad_marginal"}, or \code{"waic_hess"}. The default is \code{NULL}.}

\item{llhess}{The Hessian of the log-likelihood. This is only used
with \code{loo_approximation = "waic_hess"}. The default is \code{NULL}.}
}
\value{
A \code{psis_loo_ss} object.
}
\description{
Update \code{psis_loo_ss} objects
}
\details{
If \code{observations} is updated then if a vector of indices or a \code{psis_loo_ss}
object is supplied the updated object will have exactly the observations
indicated by the vector or \code{psis_loo_ss} object. If a single integer is
supplied, new observations will be sampled to reach the supplied sample size.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print_dims}
\alias{print_dims}
\alias{print_dims.importance_sampling}
\alias{print_dims.psis_loo}
\alias{print_dims.importance_sampling_loo}
\alias{print_dims.waic}
\alias{print_dims.kfold}
\alias{print_dims.psis_loo_ss}
\title{Print dimensions of log-likelihood or log-weights matrix}
\usage{
print_dims(x, ...)

\method{print_dims}{importance_sampling}(x, ...)

\method{print_dims}{psis_loo}(x, ...)

\method{print_dims}{importance_sampling_loo}(x, ...)

\method{print_dims}{waic}(x, ...)

\method{print_dims}{kfold}(x, ...)

\method{print_dims}{psis_loo_ss}(x, ...)
}
\arguments{
\item{x}{The object returned by \code{\link[=psis]{psis()}}, \code{\link[=loo]{loo()}}, or \code{\link[=waic]{waic()}}.}

\item{...}{Ignored.}
}
\description{
Print dimensions of log-likelihood or log-weights matrix
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_subsample.R
\name{.thin_draws}
\alias{.thin_draws}
\alias{.thin_draws.matrix}
\alias{.thin_draws.numeric}
\alias{.thin_draws.default}
\title{Thin a draws object}
\usage{
.thin_draws(draws, loo_approximation_draws)

\method{.thin_draws}{matrix}(draws, loo_approximation_draws)

\method{.thin_draws}{numeric}(draws, loo_approximation_draws)

\method{.thin_draws}{default}(draws, loo_approximation_draws)
}
\arguments{
\item{draws}{A draws object with posterior draws.}

\item{loo_approximation_draws}{The number of posterior draws to return (ie after thinning).}
}
\value{
A thinned draws object.
}
\description{
Thin a draws object
}
\details{
This is a generic function to thin draws from arbitrary draws
objects. The function is internal and should only be used by developers to
enable \code{\link[=loo_subsample]{loo_subsample()}} for arbitrary draws objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psis.R, R/sis.R, R/tis.R
\name{psis}
\alias{psis}
\alias{psis.array}
\alias{psis.matrix}
\alias{psis.default}
\alias{is.psis}
\alias{is.sis}
\alias{is.tis}
\title{Pareto smoothed importance sampling (PSIS)}
\usage{
psis(log_ratios, ...)

\method{psis}{array}(log_ratios, ..., r_eff = 1, cores = getOption("mc.cores", 1))

\method{psis}{matrix}(log_ratios, ..., r_eff = 1, cores = getOption("mc.cores", 1))

\method{psis}{default}(log_ratios, ..., r_eff = 1)

is.psis(x)

is.sis(x)

is.tis(x)
}
\arguments{
\item{log_ratios}{An array, matrix, or vector of importance ratios on the log
scale (for PSIS-LOO these are \emph{negative} log-likelihood values). See the
\strong{Methods (by class)} section below for a detailed description of how
to specify the inputs for each method.}

\item{...}{Arguments passed on to the various methods.}

\item{r_eff}{Vector of relative effective sample size estimates containing
one element per observation. The values provided should be the relative
effective sample sizes of \code{1/exp(log_ratios)} (i.e., \code{1/ratios}).
This is related to the relative efficiency of estimating the normalizing
term in self-normalizing importance sampling. If \code{r_eff} is not
provided then the reported PSIS effective sample sizes and Monte Carlo
error estimates can be over-optimistic. If the posterior draws are (near)
independent then \code{r_eff=1} can be used. \code{r_eff} has to be a scalar (same
value is used for all observations) or a vector with length equal to the
number of observations. The default value is 1. See the \code{\link[=relative_eff]{relative_eff()}}
helper function for computing \code{r_eff}.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}

\item{x}{For \code{is.psis()}, an object to check.}
}
\value{
The \code{psis()} methods return an object of class \code{"psis"},
which is a named list with the following components:

\describe{
\item{\code{log_weights}}{
Vector or matrix of smoothed (and truncated) but \emph{unnormalized} log
weights. To get normalized weights use the
\code{\link[=weights.importance_sampling]{weights()}} method provided for objects of
class \code{"psis"}.
}
\item{\code{diagnostics}}{
A named list containing two vectors:
\itemize{
\item \code{pareto_k}: Estimates of the shape parameter \eqn{k} of the
generalized Pareto distribution. See the \link{pareto-k-diagnostic}
page for details.
\item \code{n_eff}: PSIS effective sample size estimates.
}
}
}

Objects of class \code{"psis"} also have the following \link{attributes}:
\describe{
\item{\code{norm_const_log}}{
Vector of precomputed values of \code{colLogSumExps(log_weights)} that are
used internally by the \code{weights} method to normalize the log weights.
}
\item{\code{tail_len}}{
Vector of tail lengths used for fitting the generalized Pareto distribution.
}
\item{\code{r_eff}}{
If specified, the user's \code{r_eff} argument.
}
\item{\code{dims}}{
Integer vector of length 2 containing \code{S} (posterior sample size)
and \code{N} (number of observations).
}
\item{\code{method}}{
Method used for importance sampling, here \code{psis}.
}
}
}
\description{
Implementation of Pareto smoothed importance sampling (PSIS), a method for
stabilizing importance ratios. The version of PSIS implemented here
corresponds to the algorithm presented in Vehtari, Simpson, Gelman, Yao,
and Gabry (2024).
For PSIS diagnostics see the \link{pareto-k-diagnostic} page.
}
\section{Methods (by class)}{
\itemize{
\item \code{psis(array)}: An \eqn{I} by \eqn{C} by \eqn{N} array, where \eqn{I}
is the number of MCMC iterations per chain, \eqn{C} is the number of
chains, and \eqn{N} is the number of data points.

\item \code{psis(matrix)}: An \eqn{S} by \eqn{N} matrix, where \eqn{S} is the size
of the posterior sample (with all chains merged) and \eqn{N} is the number
of data points.

\item \code{psis(default)}: A vector of length \eqn{S} (posterior sample size).

}}
\examples{
log_ratios <- -1 * example_loglik_array()
r_eff <- relative_eff(exp(-log_ratios))
psis_result <- psis(log_ratios, r_eff = r_eff)
str(psis_result)
plot(psis_result)

# extract smoothed weights
lw <- weights(psis_result) # default args are log=TRUE, normalize=TRUE
ulw <- weights(psis_result, normalize=FALSE) # unnormalized log-weights

w <- weights(psis_result, log=FALSE) # normalized weights (not log-weights)
uw <- weights(psis_result, log=FALSE, normalize = FALSE) # unnormalized weights



}
\references{
Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}
\seealso{
\itemize{
\item \code{\link[=loo]{loo()}} for approximate LOO-CV using PSIS.
\item \link{pareto-k-diagnostic} for PSIS diagnostics.
\item The \strong{loo} package \href{https://mc-stan.org/loo/articles/index.html}{vignettes}
for demonstrations.
\item The \href{https://mc-stan.org/loo/articles/online-only/faq.html}{FAQ page} on
the \strong{loo} website for answers to frequently asked questions.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_model_weights.R
\name{loo_model_weights}
\alias{loo_model_weights}
\alias{loo_model_weights.default}
\alias{stacking_weights}
\alias{pseudobma_weights}
\title{Model averaging/weighting via stacking or pseudo-BMA weighting}
\usage{
loo_model_weights(x, ...)

\method{loo_model_weights}{default}(
  x,
  ...,
  method = c("stacking", "pseudobma"),
  optim_method = "BFGS",
  optim_control = list(),
  BB = TRUE,
  BB_n = 1000,
  alpha = 1,
  r_eff_list = NULL,
  cores = getOption("mc.cores", 1)
)

stacking_weights(lpd_point, optim_method = "BFGS", optim_control = list())

pseudobma_weights(lpd_point, BB = TRUE, BB_n = 1000, alpha = 1)
}
\arguments{
\item{x}{A list of \code{"psis_loo"} objects (objects returned by \code{\link[=loo]{loo()}}) or
pointwise log-likelihood matrices, one for each model. If the list
elements are named the names will be used to label the models in the
results. Each matrix/object should have dimensions \eqn{S} by \eqn{N},
where \eqn{S} is the size of the posterior sample (with all chains merged)
and \eqn{N} is the number of data points. If \code{x} is a list of
log-likelihood matrices then \code{\link[=loo]{loo()}} is called internally on each matrix.
Currently the \code{loo_model_weights()} function is not implemented to be used
with results from K-fold CV, but you can still obtain weights using K-fold
CV results by calling the \code{stacking_weights()} or \code{pseudobma_weights()}
function directly.}

\item{...}{Unused, except for the generic to pass arguments to individual
methods.}

\item{method}{Either \code{"stacking"} (the default) or \code{"pseudobma"}, indicating which method
to use for obtaining the weights. \code{"stacking"} refers to stacking of
predictive distributions and  \code{"pseudobma"} refers to pseudo-BMA+ weighting
(or plain pseudo-BMA weighting if argument \code{BB} is \code{FALSE}).}

\item{optim_method}{If \code{method="stacking"}, a string passed to the \code{method}
argument of \code{\link[stats:constrOptim]{stats::constrOptim()}} to specify the optimization algorithm.
The default is \code{optim_method="BFGS"}, but other options are available (see
\code{\link[stats:optim]{stats::optim()}}).}

\item{optim_control}{If \code{method="stacking"}, a list of control parameters for
optimization passed to the \code{control} argument of \code{\link[stats:constrOptim]{stats::constrOptim()}}.}

\item{BB}{Logical used when \code{"method"}=\code{"pseudobma"}. If
\code{TRUE} (the default), the Bayesian bootstrap will be used to adjust
the pseudo-BMA weighting, which is called pseudo-BMA+ weighting. It helps
regularize the weight away from 0 and 1, so as to reduce the variance.}

\item{BB_n}{For pseudo-BMA+ weighting only, the number of samples to use for
the Bayesian bootstrap. The default is \code{BB_n=1000}.}

\item{alpha}{Positive scalar shape parameter in the Dirichlet distribution
used for the Bayesian bootstrap. The default is \code{alpha=1}, which
corresponds to a uniform distribution on the simplex space.}

\item{r_eff_list}{Optionally, a list of relative effective sample size
estimates for the likelihood \code{(exp(log_lik))} of each observation in
each model. See \code{\link[=psis]{psis()}} and  \code{\link[=relative_eff]{relative_eff()}} helper
function for computing \code{r_eff}. If \code{x} is a list of \code{"psis_loo"}
objects then \code{r_eff_list} is ignored.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}

\item{lpd_point}{If calling \code{stacking_weights()} or \code{pseudobma_weights()}
directly, a matrix of pointwise leave-one-out (or K-fold) log likelihoods
evaluated for different models. It should be a \eqn{N} by \eqn{K}  matrix
where \eqn{N} is sample size and \eqn{K} is the number of models. Each
column corresponds to one model. These values can be calculated
approximately using \code{\link[=loo]{loo()}} or by running exact leave-one-out or K-fold
cross-validation.}
}
\value{
A numeric vector containing one weight for each model.
}
\description{
Model averaging via stacking of predictive distributions, pseudo-BMA
weighting or pseudo-BMA+ weighting with the Bayesian bootstrap. See Yao et
al. (2018), Vehtari, Gelman, and Gabry (2017), and Vehtari, Simpson,
Gelman, Yao, and Gabry (2024) for background.
}
\details{
\code{loo_model_weights()} is a wrapper around the \code{stacking_weights()} and
\code{pseudobma_weights()} functions that implements stacking, pseudo-BMA, and
pseudo-BMA+ weighting for combining multiple predictive distributions. We can
use approximate or exact leave-one-out cross-validation (LOO-CV) or K-fold CV
to estimate the expected log predictive density (ELPD).

The stacking method (\code{method="stacking"}), which is the default for
\code{loo_model_weights()}, combines all models by maximizing the leave-one-out
predictive density of the combination distribution. That is, it finds the
optimal linear combining weights for maximizing the leave-one-out log score.

The pseudo-BMA method (\code{method="pseudobma"}) finds the relative weights
proportional to the ELPD of each model. However, when
\code{method="pseudobma"}, the default is to also use the Bayesian bootstrap
(\code{BB=TRUE}), which corresponds to the pseudo-BMA+ method. The Bayesian
bootstrap  takes into account the uncertainty of finite data points and
regularizes the weights away from the extremes of 0 and 1.

In general, we recommend stacking for averaging predictive distributions,
while pseudo-BMA+ can serve as a computationally easier alternative.
}
\examples{
\dontrun{
### Demonstrating usage after fitting models with RStan
library(rstan)

# generate fake data from N(0,1).
N <- 100
y <- rnorm(N, 0, 1)

# Suppose we have three models: N(-1, sigma), N(0.5, sigma) and N(0.6,sigma).
stan_code <- "
  data {
    int N;
    vector[N] y;
    real mu_fixed;
  }
  parameters {
    real<lower=0> sigma;
  }
  model {
    sigma ~ exponential(1);
    y ~ normal(mu_fixed, sigma);
  }
  generated quantities {
    vector[N] log_lik;
    for (n in 1:N) log_lik[n] = normal_lpdf(y[n]| mu_fixed, sigma);
  }"

mod <- stan_model(model_code = stan_code)
fit1 <- sampling(mod, data=list(N=N, y=y, mu_fixed=-1))
fit2 <- sampling(mod, data=list(N=N, y=y, mu_fixed=0.5))
fit3 <- sampling(mod, data=list(N=N, y=y, mu_fixed=0.6))
model_list <- list(fit1, fit2, fit3)
log_lik_list <- lapply(model_list, extract_log_lik)

# optional but recommended
r_eff_list <- lapply(model_list, function(x) {
  ll_array <- extract_log_lik(x, merge_chains = FALSE)
  relative_eff(exp(ll_array))
})

# stacking method:
wts1 <- loo_model_weights(
  log_lik_list,
  method = "stacking",
  r_eff_list = r_eff_list,
  optim_control = list(reltol=1e-10)
)
print(wts1)

# can also pass a list of psis_loo objects to avoid recomputing loo
loo_list <- lapply(1:length(log_lik_list), function(j) {
  loo(log_lik_list[[j]], r_eff = r_eff_list[[j]])
})

wts2 <- loo_model_weights(
  loo_list,
  method = "stacking",
  optim_control = list(reltol=1e-10)
)
all.equal(wts1, wts2)

# can provide names to be used in the results
loo_model_weights(setNames(loo_list, c("A", "B", "C")))


# pseudo-BMA+ method:
set.seed(1414)
loo_model_weights(loo_list, method = "pseudobma")

# pseudo-BMA method (set BB = FALSE):
loo_model_weights(loo_list, method = "pseudobma", BB = FALSE)

# calling stacking_weights or pseudobma_weights directly
lpd1 <- loo(log_lik_list[[1]], r_eff = r_eff_list[[1]])$pointwise[,1]
lpd2 <- loo(log_lik_list[[2]], r_eff = r_eff_list[[2]])$pointwise[,1]
lpd3 <- loo(log_lik_list[[3]], r_eff = r_eff_list[[3]])$pointwise[,1]
stacking_weights(cbind(lpd1, lpd2, lpd3))
pseudobma_weights(cbind(lpd1, lpd2, lpd3))
pseudobma_weights(cbind(lpd1, lpd2, lpd3), BB = FALSE)
}

}
\references{
Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}

Yao, Y., Vehtari, A., Simpson, D., and Gelman, A. (2018) Using
stacking to average Bayesian predictive distributions.
\emph{Bayesian Analysis}, advance publication,  doi:10.1214/17-BA1091.
(\href{https://projecteuclid.org/euclid.ba/1516093227}{online}).
}
\seealso{
\itemize{
\item The \strong{loo} package \href{https://mc-stan.org/loo/articles/}{vignettes}, particularly
\href{https://mc-stan.org/loo/articles/loo2-weights.html}{Bayesian Stacking and Pseudo-BMA weights using the \strong{loo} package}.
\item \code{\link[=loo]{loo()}} for details on leave-one-out ELPD estimation.
\item \code{\link[=constrOptim]{constrOptim()}} for the choice of optimization methods and control-parameters.
\item \code{\link[=relative_eff]{relative_eff()}} for computing \code{r_eff}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_predictive_metric.R
\name{loo_predictive_metric}
\alias{loo_predictive_metric}
\alias{loo_predictive_metric.matrix}
\title{Estimate leave-one-out predictive performance..}
\usage{
loo_predictive_metric(x, ...)

\method{loo_predictive_metric}{matrix}(
  x,
  y,
  log_lik,
  ...,
  metric = c("mae", "rmse", "mse", "acc", "balanced_acc"),
  r_eff = 1,
  cores = getOption("mc.cores", 1)
)
}
\arguments{
\item{x}{A numeric matrix of predictions.}

\item{...}{Additional arguments passed on to \code{\link[=E_loo]{E_loo()}}}

\item{y}{A numeric vector of observations. Length should be equal to the
number of rows in \code{x}.}

\item{log_lik}{A matrix of pointwise log-likelihoods. Should be of same
dimension as \code{x}.}

\item{metric}{The type of predictive metric to be used. Currently
supported options are \code{"mae"}, \code{"rmse"} and \code{"mse"} for regression and
for binary classification \code{"acc"} and \code{"balanced_acc"}.
\describe{
\item{\code{"mae"}}{
Mean absolute error.
}
\item{\code{"mse"}}{
Mean squared error.
}
\item{\code{"rmse"}}{
Root mean squared error, given by as the square root of \code{MSE}.
}
\item{\code{"acc"}}{
The proportion of predictions indicating the correct outcome.
}
\item{\code{"balanced_acc"}}{
Balanced accuracy is given by the average of true positive and true
negative rates.
}
}}

\item{r_eff}{A Vector of relative effective sample size estimates containing
one element per observation. See \code{\link[=psis]{psis()}} for more details.}

\item{cores}{The number of cores to use for parallelization of \verb{[psis()]}.
See \code{\link[=psis]{psis()}} for details.}
}
\value{
A list with the following components:
\describe{
\item{\code{estimate}}{
Estimate of the given metric.
}
\item{\code{se}}{
Standard error of the estimate.
}
}
}
\description{
The \code{loo_predictive_metric()} function computes estimates of leave-one-out
predictive metrics given a set of predictions and observations. Currently
supported metrics are mean absolute error, mean squared error and root mean
squared error for continuous predictions and accuracy and balanced accuracy
for binary classification. Predictions are passed on to the \code{\link[=E_loo]{E_loo()}}
function, so this function assumes that the PSIS approximation is working
well.
}
\examples{
\donttest{
if (requireNamespace("rstanarm", quietly = TRUE)) {
# Use rstanarm package to quickly fit a model and get both a log-likelihood
# matrix and draws from the posterior predictive distribution
library("rstanarm")

# data from help("lm")
ctl <- c(4.17,5.58,5.18,6.11,4.50,4.61,5.17,4.53,5.33,5.14)
trt <- c(4.81,4.17,4.41,3.59,5.87,3.83,6.03,4.89,4.32,4.69)
d <- data.frame(
  weight = c(ctl, trt),
  group = gl(2, 10, 20, labels = c("Ctl","Trt"))
)
fit <- stan_glm(weight ~ group, data = d, refresh = 0)
ll <- log_lik(fit)
r_eff <- relative_eff(exp(-ll), chain_id = rep(1:4, each = 1000))

mu_pred <- posterior_epred(fit)
# Leave-one-out mean absolute error of predictions
mae <- loo_predictive_metric(x = mu_pred, y = d$weight, log_lik = ll,
                            pred_error = 'mae', r_eff = r_eff)
# Leave-one-out 90\%-quantile of mean absolute error
mae_90q <- loo_predictive_metric(x = mu_pred, y = d$weight, log_lik = ll,
                                pred_error = 'mae', r_eff = r_eff,
                                type = 'quantile', probs = 0.9)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{compare}
\alias{compare}
\title{Model comparison (deprecated, old version)}
\usage{
compare(..., x = list())
}
\arguments{
\item{...}{At least two objects returned by \code{\link[=loo]{loo()}} (or \code{\link[=waic]{waic()}}).}

\item{x}{A list of at least two objects returned by \code{\link[=loo]{loo()}} (or
\code{\link[=waic]{waic()}}). This argument can be used as an alternative to
specifying the objects in \code{...}.}
}
\value{
A vector or matrix with class \code{'compare.loo'} that has its own
print method. If exactly two objects are provided in \code{...} or
\code{x}, then the difference in expected predictive accuracy and the
standard error of the difference are returned. If more than two objects are
provided then a matrix of summary information is returned (see \strong{Details}).
}
\description{
\strong{This function is deprecated}. Please use the new \code{\link[=loo_compare]{loo_compare()}} function
instead.
}
\details{
When comparing two fitted models, we can estimate the difference in their
expected predictive accuracy by the difference in \code{elpd_loo} or
\code{elpd_waic} (or multiplied by -2, if desired, to be on the
deviance scale).

\emph{When that difference, \code{elpd_diff}, is positive then the expected
predictive accuracy for the second model is higher. A negative
\code{elpd_diff} favors the first model.}

When using \code{compare()} with more than two models, the values in the
\code{elpd_diff} and \code{se_diff} columns of the returned matrix are
computed by making pairwise comparisons between each model and the model
with the best ELPD (i.e., the model in the first row).
Although the \code{elpd_diff} column is equal to the difference in
\code{elpd_loo}, do not expect the \code{se_diff} column to be equal to the
the difference in \code{se_elpd_loo}.

To compute the standard error of the difference in ELPD we use a
paired estimate to take advantage of the fact that the same set of \emph{N}
data points was used to fit both models. These calculations should be most
useful when \emph{N} is large, because then non-normality of the
distribution is not such an issue when estimating the uncertainty in these
sums. These standard errors, for all their flaws, should give a better
sense of uncertainty than what is obtained using the current standard
approach of comparing differences of deviances to a Chi-squared
distribution, a practice derived for Gaussian linear models or
asymptotically, and which only applies to nested models in any case.
}
\examples{
\dontrun{
loo1 <- loo(log_lik1)
loo2 <- loo(log_lik2)
print(compare(loo1, loo2), digits = 3)
print(compare(x = list(loo1, loo2)))

waic1 <- waic(log_lik1)
waic2 <- waic(log_lik2)
compare(waic1, waic2)
}

}
\references{
Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importance_sampling.R
\name{weights.importance_sampling}
\alias{weights.importance_sampling}
\title{Extract importance sampling weights}
\usage{
\method{weights}{importance_sampling}(object, ..., log = TRUE, normalize = TRUE)
}
\arguments{
\item{object}{An object returned by \code{\link[=psis]{psis()}}, \code{\link[=tis]{tis()}}, or \code{\link[=sis]{sis()}}.}

\item{...}{Ignored.}

\item{log}{Should the weights be returned on the log scale? Defaults to
\code{TRUE}.}

\item{normalize}{Should the weights be normalized? Defaults to \code{TRUE}.}
}
\value{
The \code{weights()} method returns an object with the same dimensions as
the \code{log_weights} component of \code{object}. The \code{normalize} and \code{log}
arguments control whether the returned weights are normalized and whether
or not to return them on the log scale.
}
\description{
Extract importance sampling weights
}
\examples{
# See the examples at help("psis")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psis_approximate_posterior.R
\name{ap_psis}
\alias{ap_psis}
\alias{ap_psis.array}
\alias{ap_psis.matrix}
\alias{ap_psis.default}
\title{Pareto smoothed importance sampling (PSIS)
using approximate posteriors}
\usage{
ap_psis(log_ratios, log_p, log_g, ...)

\method{ap_psis}{array}(log_ratios, log_p, log_g, ..., cores = getOption("mc.cores", 1))

\method{ap_psis}{matrix}(log_ratios, log_p, log_g, ..., cores = getOption("mc.cores", 1))

\method{ap_psis}{default}(log_ratios, log_p, log_g, ...)
}
\arguments{
\item{log_ratios}{The log-likelihood ratios (ie -log_liks)}

\item{log_p}{The log-posterior (target) evaluated at S samples from the
proposal distribution (g). A vector of length S.}

\item{log_g}{The log-density (proposal) evaluated at S samples from the
proposal distribution (g). A vector of length S.}

\item{...}{Currently not in use.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}
}
\description{
Pareto smoothed importance sampling (PSIS)
using approximate posteriors
}
\section{Methods (by class)}{
\itemize{
\item \code{ap_psis(array)}: An \eqn{I} by \eqn{C} by \eqn{N} array, where \eqn{I}
is the number of MCMC iterations per chain, \eqn{C} is the number of
chains, and \eqn{N} is the number of data points.

\item \code{ap_psis(matrix)}: An \eqn{S} by \eqn{N} matrix, where \eqn{S} is the size
of the posterior sample (with all chains merged) and \eqn{N} is the number
of data points.

\item \code{ap_psis(default)}: A vector of length \eqn{S} (posterior sample size).

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_approximate_posterior.R
\name{loo_approximate_posterior}
\alias{loo_approximate_posterior}
\alias{loo_approximate_posterior.array}
\alias{loo_approximate_posterior.matrix}
\alias{loo_approximate_posterior.function}
\title{Efficient approximate leave-one-out cross-validation (LOO) for posterior
approximations}
\usage{
loo_approximate_posterior(x, log_p, log_g, ...)

\method{loo_approximate_posterior}{array}(
  x,
  log_p,
  log_g,
  ...,
  save_psis = FALSE,
  cores = getOption("mc.cores", 1)
)

\method{loo_approximate_posterior}{matrix}(
  x,
  log_p,
  log_g,
  ...,
  save_psis = FALSE,
  cores = getOption("mc.cores", 1)
)

\method{loo_approximate_posterior}{`function`}(
  x,
  ...,
  data = NULL,
  draws = NULL,
  log_p = NULL,
  log_g = NULL,
  save_psis = FALSE,
  cores = getOption("mc.cores", 1)
)
}
\arguments{
\item{x}{A log-likelihood array, matrix, or function.
The \strong{Methods (by class)} section, below, has detailed descriptions of how
to specify the inputs for each method.}

\item{log_p}{The log-posterior (target) evaluated at S samples from the
proposal distribution (g). A vector of length S.}

\item{log_g}{The log-density (proposal) evaluated at S samples from the
proposal distribution (g). A vector of length S.}

\item{save_psis}{Should the \code{"psis"} object created internally by
\code{loo_approximate_posterior()} be saved in the returned object? See
\code{\link[=loo]{loo()}} for details.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}

\item{data, draws, ...}{For the \code{loo_approximate_posterior.function()} method,
these are the data, posterior draws, and other arguments to pass to the
log-likelihood function. See the \strong{Methods (by class)} section below for
details on how to specify these arguments.}
}
\value{
The \code{loo_approximate_posterior()} methods return a named list with
class \code{c("psis_loo_ap", "psis_loo", "loo")}. It has the same structure
as the objects returned by \code{\link[=loo]{loo()}} but with the additional slot:
\describe{
\item{\code{posterior_approximation}}{
A list with two vectors, \code{log_p} and \code{log_g} of the same length
containing the posterior density and the approximation density
for the individual draws.
}
}
}
\description{
Efficient approximate leave-one-out cross-validation (LOO) for posterior
approximations
}
\details{
The \code{loo_approximate_posterior()} function is an S3 generic and
methods are provided for 3-D pointwise log-likelihood arrays, pointwise
log-likelihood matrices, and log-likelihood functions. The implementation
works for posterior approximations where it is possible to compute the log
density for the posterior approximation.
}
\section{Methods (by class)}{
\itemize{
\item \code{loo_approximate_posterior(array)}: An \eqn{I} by \eqn{C} by \eqn{N} array, where \eqn{I}
is the number of MCMC iterations per chain, \eqn{C} is the number of
chains, and \eqn{N} is the number of data points.

\item \code{loo_approximate_posterior(matrix)}: An \eqn{S} by \eqn{N} matrix, where \eqn{S} is the size
of the posterior sample (with all chains merged) and \eqn{N} is the number
of data points.

\item \code{loo_approximate_posterior(`function`)}: A function \code{f()} that takes arguments \code{data_i} and \code{draws} and returns a
vector containing the log-likelihood for a single observation \code{i} evaluated
at each posterior draw. The function should be written such that, for each
observation \code{i} in \code{1:N}, evaluating

\if{html}{\out{<div class="sourceCode">}}\preformatted{f(data_i = data[i,, drop=FALSE], draws = draws)
}\if{html}{\out{</div>}}

results in a vector of length \code{S} (size of posterior sample). The
log-likelihood function can also have additional arguments but \code{data_i} and
\code{draws} are required.

If using the function method then the arguments \code{data} and \code{draws} must also
be specified in the call to \code{loo()}:
\itemize{
\item \code{data}: A data frame or matrix containing the data (e.g.
observed outcome and predictors) needed to compute the pointwise
log-likelihood. For each observation \code{i}, the \code{i}th row of
\code{data} will be passed to the \code{data_i} argument of the
log-likelihood function.
\item \code{draws}: An object containing the posterior draws for any
parameters needed to compute the pointwise log-likelihood. Unlike
\code{data}, which is indexed by observation, for each observation the
entire object \code{draws} will be passed to the \code{draws} argument of
the log-likelihood function.
\item The \code{...} can be used if your log-likelihood function takes additional
arguments. These arguments are used like the \code{draws} argument in that they
are recycled for each observation.
}

}}
\references{
Magnusson, M., Riis Andersen, M., Jonasson, J. and Vehtari, A. (2019).
Leave-One-Out Cross-Validation for Large Data.
In \emph{Thirty-sixth International Conference on Machine Learning},
PMLR 97:4244-4253.

Magnusson, M., Riis Andersen, M., Jonasson, J. and Vehtari, A. (2020).
Leave-One-Out Cross-Validation for Model Comparison in Large Data.
In \emph{Proceedings of the 23rd International Conference on Artificial
Intelligence and Statistics (AISTATS)}, PMLR 108:341-351.
}
\seealso{
\code{\link[=loo]{loo()}}, \code{\link[=psis]{psis()}}, \code{\link[=loo_compare]{loo_compare()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{pareto-k-diagnostic}
\alias{pareto-k-diagnostic}
\alias{pareto_k_table}
\alias{pareto_k_ids}
\alias{pareto_k_values}
\alias{pareto_k_influence_values}
\alias{psis_n_eff_values}
\alias{mcse_loo}
\alias{plot.psis_loo}
\alias{plot.loo}
\alias{plot.psis}
\title{Diagnostics for Pareto smoothed importance sampling (PSIS)}
\usage{
pareto_k_table(x)

pareto_k_ids(x, threshold = NULL)

pareto_k_values(x)

pareto_k_influence_values(x)

psis_n_eff_values(x)

mcse_loo(x, threshold = NULL)

\method{plot}{psis_loo}(
  x,
  diagnostic = c("k", "ESS", "n_eff"),
  ...,
  label_points = FALSE,
  main = "PSIS diagnostic plot"
)

\method{plot}{psis}(
  x,
  diagnostic = c("k", "ESS", "n_eff"),
  ...,
  label_points = FALSE,
  main = "PSIS diagnostic plot"
)
}
\arguments{
\item{x}{An object created by \code{\link[=loo]{loo()}} or \code{\link[=psis]{psis()}}.}

\item{threshold}{For \code{pareto_k_ids()}, \code{threshold} is the minimum \eqn{k}
value to flag (default is a sample size \code{S} dependent threshold
\code{1 - 1 / log10(S)}). For \code{mcse_loo()}, if any \eqn{k} estimates are
greater than \code{threshold} the MCSE estimate is returned as \code{NA}
See \strong{Details} for the motivation behind these defaults.}

\item{diagnostic}{For the \code{plot} method, which diagnostic should be
plotted? The options are \code{"k"} for Pareto \eqn{k} estimates (the
default), or \code{"ESS"} or \code{"n_eff"} for PSIS effective sample size estimates.}

\item{label_points, ...}{For the \code{plot()} method, if \code{label_points} is
\code{TRUE} the observation numbers corresponding to any values of \eqn{k}
greater than the diagnostic threshold will be displayed in the plot.
Any arguments specified in \code{...} will be passed to \code{\link[graphics:text]{graphics::text()}}
and can be used to control the appearance of the labels.}

\item{main}{For the \code{plot()} method, a title for the plot.}
}
\value{
\code{pareto_k_table()} returns an object of class
\code{"pareto_k_table"}, which is a matrix with columns \code{"Count"},
\code{"Proportion"}, and \code{"Min. n_eff"}, and has its own print method.

\code{pareto_k_ids()} returns an integer vector indicating which
observations have Pareto \eqn{k} estimates above \code{threshold}.

\code{pareto_k_values()} returns a vector of the estimated Pareto
\eqn{k} parameters. These represent the reliability of sampling.

\code{pareto_k_influence_values()} returns a vector of the estimated Pareto
\eqn{k} parameters. These represent influence of the observations on the
model posterior distribution.

\code{psis_n_eff_values()} returns a vector of the estimated PSIS
effective sample sizes.

\code{mcse_loo()} returns the Monte Carlo standard error (MCSE)
estimate for PSIS-LOO. MCSE will be NA if any Pareto \eqn{k} values are
above \code{threshold}.

The \code{plot()} method is called for its side effect and does not
return anything. If \code{x} is the result of a call to \code{\link[=loo]{loo()}}
or \code{\link[=psis]{psis()}} then \code{plot(x, diagnostic)} produces a plot of
the estimates of the Pareto shape parameters (\code{diagnostic = "k"}) or
estimates of the PSIS effective sample sizes (\code{diagnostic = "ESS"}).
}
\description{
Print a diagnostic table summarizing the estimated Pareto shape parameters
and PSIS effective sample sizes, find the indexes of observations for which
the estimated Pareto shape parameter \eqn{k} is larger than some
\code{threshold} value, or plot observation indexes vs. diagnostic estimates.
The \strong{Details} section below provides a brief overview of the
diagnostics, but we recommend consulting Vehtari, Gelman, and Gabry (2017)
and Vehtari, Simpson, Gelman, Yao, and Gabry (2024) for full details.
}
\details{
The reliability and approximate convergence rate of the PSIS-based
estimates can be assessed using the estimates for the shape
parameter \eqn{k} of the generalized Pareto distribution. The
diagnostic threshold for Pareto \eqn{k} depends on sample size
\eqn{S} (sample size dependent threshold was introduced by Vehtari
et al. (2024), and before that fixed thresholds of 0.5 and 0.7 were
recommended). For simplicity, \code{loo} package uses the nominal sample
size \eqn{S} when computing the sample size specific
threshold. This provides an optimistic threshold if the effective
sample size is less than 2200, but if MCMC-ESS > S/2 the difference
is usually negligible. Thinning of MCMC draws can be used to
improve the ratio ESS/S.
\itemize{
\item If \eqn{k < min(1 - 1 / log10(S), 0.7)}, where \eqn{S} is the
sample size, the PSIS estimate and the corresponding Monte Carlo
standard error estimate are reliable.
\item If \eqn{1 - 1 / log10(S) <= k < 0.7}, the PSIS estimate and the
corresponding Monte Carlo standard error estimate are not
reliable, but increasing the (effective) sample size \eqn{S} above
2200 may help (this will increase the sample size specific
threshold \eqn{(1-1/log10(2200)>0.7} and then the bias specific
threshold 0.7 dominates).
\item If \eqn{0.7 <= k < 1}, the PSIS estimate and the corresponding Monte
Carlo standard error have large bias and are not reliable. Increasing
the sample size may reduce the variability in \eqn{k} estimate, which
may result in lower \eqn{k} estimate, too.
\item If \eqn{k \geq 1}{k >= 1}, the target distribution is estimated to
have a non-finite mean. The PSIS estimate and the corresponding Monte
Carlo standard error are not well defined. Increasing the sample size
may reduce the variability in the \eqn{k} estimate, which
may also result in a lower \eqn{k} estimate.
}

\subsection{What if the estimated tail shape parameter \eqn{k}
exceeds the diagnostic threshold?}{ Importance sampling is likely to
work less well if the marginal posterior \eqn{p(\theta^s | y)} and
LOO posterior \eqn{p(\theta^s | y_{-i})} are very different, which
is more likely to happen with a non-robust model and highly
influential observations.  If the estimated tail shape parameter
\eqn{k} exceeds the diagnostic threshold, the user should be
warned. (Note: If \eqn{k} is greater than the diagnostic threshold
then WAIC is also likely to fail, but WAIC lacks as accurate
diagnostic.)  When using PSIS in the context of approximate LOO-CV,
we recommend one of the following actions:
\itemize{
\item With some additional computations, it is possible to transform
the MCMC draws from the posterior distribution to obtain more
reliable importance sampling estimates. This results in a smaller
shape parameter \eqn{k}.  See \code{\link[=loo_moment_match]{loo_moment_match()}} and the
vignette \emph{Avoiding model refits in leave-one-out cross-validation
with moment matching} for an example of this.
\item Sampling from a leave-one-out mixture distribution (see the
vignette \emph{Mixture IS leave-one-out cross-validation for
high-dimensional Bayesian models}), directly from \eqn{p(\theta^s
  | y_{-i})} for the problematic observations \eqn{i}, or using
\eqn{K}-fold cross-validation (see the vignette \emph{Holdout
validation and K-fold cross-validation of Stan programs with the
loo package}) will generally be more stable.
\item Using a model that is more robust to anomalous observations will
generally make approximate LOO-CV more stable.
}

}

\subsection{Observation influence statistics}{ The estimated shape parameter
\eqn{k} for each observation can be used as a measure of the observation's
influence on posterior distribution of the model. These can be obtained with
\code{pareto_k_influence_values()}.
}

\subsection{Effective sample size and error estimates}{ In the case that we
obtain the samples from the proposal distribution via MCMC the \strong{loo}
package also computes estimates for the Monte Carlo error and the effective
sample size for importance sampling, which are more accurate for PSIS than
for IS and TIS (see Vehtari et al (2024) for details). However, the PSIS
effective sample size estimate will be
\strong{over-optimistic when the estimate of \eqn{k} is greater than}
\eqn{min(1-1/log10(S), 0.7)}, where \eqn{S} is the sample size.
}
}
\references{
Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}
\seealso{
\itemize{
\item \code{\link[=psis]{psis()}} for the implementation of the PSIS algorithm.
\item The \href{https://mc-stan.org/loo/articles/online-only/faq.html}{FAQ page} on
the \strong{loo} website for answers to frequently asked questions.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importance_sampling.R
\name{importance_sampling}
\alias{importance_sampling}
\alias{importance_sampling.array}
\alias{importance_sampling.matrix}
\alias{importance_sampling.default}
\title{A parent class for different importance sampling methods.}
\usage{
importance_sampling(log_ratios, method, ...)

\method{importance_sampling}{array}(
  log_ratios,
  method,
  ...,
  r_eff = 1,
  cores = getOption("mc.cores", 1)
)

\method{importance_sampling}{matrix}(
  log_ratios,
  method,
  ...,
  r_eff = 1,
  cores = getOption("mc.cores", 1)
)

\method{importance_sampling}{default}(log_ratios, method, ..., r_eff = 1)
}
\arguments{
\item{log_ratios}{An array, matrix, or vector of importance ratios on the log
scale (for PSIS-LOO these are \emph{negative} log-likelihood values). See the
\strong{Methods (by class)} section below for a detailed description of how
to specify the inputs for each method.}

\item{method}{The importance sampling method to use. The following methods
are implemented:
\itemize{
\item \code{\link[=psis]{"psis"}}: Pareto-Smoothed Importance Sampling (PSIS). Default method.
\item \code{\link[=tis]{"tis"}}: Truncated Importance Sampling (TIS) with truncation at
\code{sqrt(S)}, where \code{S} is the number of posterior draws.
\item \code{\link[=sis]{"sis"}}: Standard Importance Sampling (SIS).
}}

\item{...}{Arguments passed on to the various methods.}

\item{r_eff}{Vector of relative effective sample size estimates containing
one element per observation. The values provided should be the relative
effective sample sizes of \code{1/exp(log_ratios)} (i.e., \code{1/ratios}).
This is related to the relative efficiency of estimating the normalizing
term in self-normalizing importance sampling. If \code{r_eff} is not
provided then the reported PSIS effective sample sizes and Monte Carlo
error estimates can be over-optimistic. If the posterior draws are (near)
independent then \code{r_eff=1} can be used. \code{r_eff} has to be a scalar (same
value is used for all observations) or a vector with length equal to the
number of observations. The default value is 1. See the \code{\link[=relative_eff]{relative_eff()}}
helper function for computing \code{r_eff}.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}
}
\description{
A parent class for different importance sampling methods.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo-glossary.R
\name{loo-glossary}
\alias{loo-glossary}
\title{LOO package glossary}
\description{
This page provides definitions of key terms. Also see the
\href{https://mc-stan.org/loo/articles/online-only/faq.html}{FAQ page} on
the \strong{loo} website for answers to frequently asked questions.

Note: VGG2017 refers to Vehtari, Gelman, and Gabry (2017). See
\strong{References}, below.
}
\section{ELPD and \code{elpd_loo}}{


The ELPD is the theoretical expected log pointwise predictive density for a new
dataset (Eq 1 in VGG2017), which can be estimated, e.g., using
cross-validation. \code{elpd_loo} is the Bayesian LOO estimate of the
expected log pointwise predictive density (Eq 4 in VGG2017) and
is a sum of N individual pointwise log predictive densities. Probability
densities can be smaller or larger than 1, and thus log predictive densities
can be negative or positive. For simplicity the ELPD acronym is used also for
expected log pointwise predictive probabilities for discrete models.
Probabilities are always equal or less than 1, and thus log predictive
probabilities are 0 or negative.
}

\section{Standard error of \code{elpd_loo}}{


As \code{elpd_loo} is defined as the sum of N independent components (Eq 4 in
VGG2017), we can compute the standard error by using the standard deviation
of the N components and multiplying by \code{sqrt(N)} (Eq 23 in VGG2017).
This standard error is a coarse description of our uncertainty about the
predictive performance for unknown future data. When N is small or there is
severe model misspecification, the current SE estimate is overoptimistic and
the actual SE can even be twice as large. Even for moderate N, when the SE
estimate is an accurate estimate for the scale, it ignores the skewness. When
making model comparisons, the SE of the component-wise (pairwise) differences
should be used instead (see the \code{se_diff} section below and Eq 24 in
VGG2017). Sivula et al. (2022) discuss the conditions when the normal
approximation used for SE and \code{se_diff} is good.
}

\section{Monte Carlo SE of elpd_loo}{


The Monte Carlo standard error is the estimate for the computational accuracy
of MCMC and importance sampling used to compute \code{elpd_loo}. Usually this
is negligible compared to the standard describing the uncertainty due to
finite number of observations (Eq 23 in VGG2017).
}

\section{\code{p_loo} (effective number of parameters)}{


\code{p_loo} is the difference between \code{elpd_loo} and the non-cross-validated
log posterior predictive density. It describes how much more difficult it
is to predict future data than the observed data. Asymptotically under
certain regularity conditions, \code{p_loo} can be interpreted as the
\emph{effective number of parameters}. In well behaving cases \code{p_loo < N} and
\code{p_loo < p}, where \code{p} is the total number of parameters in the
model. \code{p_loo > N}  or \code{p_loo > p} indicates that the model has very
weak predictive capability and may indicate a severe model misspecification.
See below for more on interpreting \code{p_loo} when there are warnings
about high Pareto k diagnostic values.
}

\section{Pareto k estimates}{


The Pareto \eqn{k} estimate is a diagnostic for Pareto smoothed importance
sampling (PSIS), which is used to compute components of \code{elpd_loo}. In
importance-sampling LOO the full posterior distribution is used as the
proposal distribution. The Pareto k diagnostic estimates how far an
individual leave-one-out distribution is from the full distribution. If
leaving out an observation changes the posterior too much then importance
sampling is not able to give a reliable estimate. Pareto smoothing stabilizes
importance sampling and guarantees a finite variance estimate at the
cost of some bias.

The diagnostic threshold for Pareto \eqn{k} depends on sample size
\eqn{S} (sample size dependent threshold was introduced by Vehtari
et al., 2024, and before that fixed thresholds of 0.5 and 0.7 were
recommended). For simplicity, \code{loo} package uses the nominal sample
size \eqn{S}  when computing the sample size specific
threshold. This provides an optimistic threshold if the effective
sample size is less than 2200, but even then if ESS/S > 1/2 the difference
is usually negligible. Thinning of MCMC draws can be used to improve
the ratio ESS/S.
\itemize{
\item If \eqn{k < \min(1 - 1 / \log_{10}(S), 0.7)}, where \eqn{S} is the
sample size, the PSIS estimate and the corresponding Monte
Carlo standard error estimate are reliable.
\item If \eqn{1 - 1 / \log_{10}(S) <= k < 0.7}, the PSIS estimate and the
corresponding Monte Carlo standard error estimate are not
reliable, but increasing the (effective) sample size \eqn{S} above
2200 may help (this will increase the sample size specific
threshold \eqn{(1 - 1 / \log_{10}(2200) > 0.7} and then the bias specific
threshold 0.7 dominates).
\item If \eqn{0.7 <= k < 1}, the PSIS estimate and the corresponding Monte
Carlo standard error have large bias and are not reliable. Increasing
the sample size may reduce the variability in the \eqn{k} estimate, which
may also result in a lower \eqn{k} estimate.
\item If \eqn{k \geq 1}{k >= 1}, the target distribution is estimated to
have non-finite mean. The PSIS estimate and the corresponding Monte
Carlo standard error are not well defined. Increasing the sample size
may reduce the variability in \eqn{k} estimate, which may also result in
a lower \eqn{k} estimate.
}

Pareto \eqn{k} is also useful as a measure of influence of an
observation.  Highly influential observations have high \eqn{k}
values. Very high \eqn{k} values often indicate model
misspecification, outliers or mistakes in data processing. See
Section 6 of Gabry et al. (2019) for an example.

\subsection{Interpreting \code{p_loo} when Pareto \code{k} is large}{
If \eqn{k > 0.7} then we can also look at
the \code{p_loo} estimate for some additional information about the problem:
\itemize{
\item If \verb{p_loo << p} (the total number of parameters in the model),
then the model is likely to be misspecified. Posterior predictive checks
(PPCs) are then likely to also detect the problem. Try using an overdispersed
model, or add more structural information (nonlinearity, mixture model,
etc.).
\item If \code{p_loo < p} and the number of parameters \code{p} is relatively
large compared to the number of observations (e.g., \code{p>N/5}), it is
likely that the model is so flexible or the population prior so weak that it’s
difficult to predict the left out observation (even for the true model).
This happens, for example, in the simulated 8 schools (in VGG2017), random
effect models with a few observations per random effect, and Gaussian
processes and spatial models with short correlation lengths.
\item If \code{p_loo > p}, then the model is likely to be badly misspecified.
If the number of parameters \verb{p<<N}, then PPCs are also likely to detect the
problem. See the case study at
\url{https://avehtari.github.io/modelselection/roaches.html} for an example.
If \code{p} is relatively large compared to the number of
observations, say \code{p>N/5} (more accurately we should count number of
observations influencing each parameter as in hierarchical models some groups
may have few observations and other groups many), it is possible that PPCs won't
detect the problem.
}
}
}

\section{elpd_diff}{

\code{elpd_diff} is the difference in \code{elpd_loo} for two models. If more
than two models are compared, the difference is computed relative to the
model with highest \code{elpd_loo}.
}

\section{se_diff}{


The standard error of component-wise differences of elpd_loo (Eq 24 in
VGG2017) between two models. This SE is \emph{smaller} than the SE for
individual models due to correlation (i.e., if some observations are easier
and some more difficult to predict for all models).
}

\section{\code{p_worse} (probability of worse predictive performance)}{


\code{p_worse} is the estimated probability that a model has worse predictive
performance than the best-ranked model in the comparison, based on the normal
approximation to the uncertainty in \code{elpd_diff}. It is computed as

\if{html}{\out{<div class="sourceCode">}}\preformatted{p_worse = pnorm(0, elpd_diff, se_diff).
}\if{html}{\out{</div>}}

The best-ranked model (the first row in the \code{loo_compare()} output, where
\code{elpd_diff = 0}) always receives \code{NA}, since the comparison is defined
relative to that model.

Because models are ordered by \code{elpd_loo} before computing \code{p_worse}, all
reported values are at least 0.5 by construction. A value close to 0.5
indicates that the models are nearly indistinguishable in predictive
performance and that the ranking could easily be reversed with different
data. A value close to 1 indicates that the lower-ranked model is almost
certainly worse. \code{p_worse} inherits all the limitations of \code{se_diff} and the
normal approximation on which it is based. In particular, when \code{se_diff} is
underestimated, \code{p_worse} will be estimated too close to 1, making a model
appear more clearly worse than the data actually support. Conversely, when
\code{elpd_diff} is biased due to an unreliable LOO approximation, \code{p_worse} can
point in the wrong direction entirely. When any of these conditions are
present, \code{diag_diff} or \code{diag_elpd} will be flagged in the \code{loo_compare()}
output.
For further guidance, see the sections below and the case study on
\href{https://users.aalto.fi/~ave/casestudies/LOO_uncertainty/loo_uncertainty.html}{Uncertainty in Bayesian LOO-CV Model Comparison}.
}

\section{\code{diag_diff} (pairwise comparison diagnostics)}{


\code{diag_diff} is a diagnostic column in the \code{loo_compare()} output for each
model comparison against the current reference model. It flags conditions
under which the normal approximation behind \code{se_diff} and \code{p_worse} is likely
to be poorly calibrated. The column contains a short label when a condition
is detected, and is empty otherwise.

The column \code{diag_diff} currently flags two problems:
\subsection{\code{N < 100}}{

When the number of observations is small, we may assume \code{se_diff} to be
underestimated. As a rough heuristic one can multiply \code{se_diff} by 2 to
make a more conservative estimate.
}

\subsection{\verb{|elpd_diff| < 4}}{

When \verb{|elpd_diff|} is below 4, the models have very similar predictive
performance. In this setting, Sivula et al. (2025) show that skewness in
the error distribution can make the normal approximation for \code{se_diff}
and \code{p_worse} miscalibrated, even for large N. In practice, this usually
supports treating the models as predictively similar.
}

\subsection{Relation between \code{N < 100} and \verb{|elpd_diff| < 4}}{

The conditions flagged by \code{diag_diff} are not independent: they tend to
co-occur, and when they do, some flags carry more information than others.
\code{loo_compare()} therefore follows a priority hierarchy and shows only the
most critical flag in the table output.

The hierarchy is as follows:
\itemize{
\item \strong{\code{N < 100} takes highest priority.} A small sample size undermines the
reliability of \code{se_diff} by underestimating uncertainty. Because of this,
even if \verb{|elpd_diff| < 4} is also true for a comparison, the table will only
show \code{N < 100}. The small sample size renders the \verb{|elpd_diff| < 4}
diagnostic less meaningful.
\item \strong{\verb{|elpd_diff| < 4} takes second priority.} When N >= 100 and the
difference is small, the normal approximation is miscalibrated due to the
skewness of the error distribution (Sivula et al., 2025). In this
situation, \code{se_diff} exists and is not heavily biased in scale, but the
shape of the approximation is wrong, making \code{p_worse} unreliable.
}

For further guidance, see the case study on
\href{https://users.aalto.fi/~ave/casestudies/LOO_uncertainty/loo_uncertainty.html}{Uncertainty in Bayesian LOO-CV Model Comparison}.
}
}

\section{\code{diag_elpd}}{


\code{diag_elpd} is a diagnostic column in the \code{loo_compare()} output that flags
when the PSIS-LOO approximation for an individual model is unreliable. Unlike
\code{diag_diff}, which concerns the \emph{comparison} between models, \code{diag_elpd}
concerns the quality of the \code{elpd_loo} estimate for each model individually.
It contains a short text label when a problem is detected, and is empty
otherwise.
\subsection{\verb{K k_psis > t} (K observations with Pareto-k values > t)}{

This label indicates that K observations for this model have Pareto-k values
above the PSIS reliability threshold \code{t} used by \code{loo} for that fit. The
threshold is sample-size dependent, and in many practical cases close to
0.7. When this flag appears, the PSIS approximation can be unreliable for
those observations, and the resulting \code{elpd_loo} may be biased. Because
\code{elpd_diff} is a direct difference of two models' \code{elpd_loo} values, bias in
either model's estimate propagates directly into \code{elpd_diff} and \code{p_worse}.
This is qualitatively different from the calibration issues flagged by
\code{diag_diff}: here the estimate itself may be wrong, not just uncertain.

See for further information on Pareto-k values the "Pareto k estimates"
section.
}
}

\references{
Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}

Sivula, T, Magnusson, M., Matamoros A. A., and Vehtari, A. (2025).
Uncertainty in Bayesian leave-one-out cross-validation based model
comparison. \emph{Bayesian Analysis}. \doi{10.1214/25-BA1569}.

Gabry, J. , Simpson, D. , Vehtari, A. , Betancourt, M. and
Gelman, A. (2019), Visualization in Bayesian workflow.
\emph{J. R. Stat. Soc. A}, 182: 389-402. doi:10.1111/rssa.12378
(\href{https://rss.onlinelibrary.wiley.com/doi/full/10.1111/rssa.12378}{journal version},
\href{https://arxiv.org/abs/1709.01449}{preprint arXiv:1709.01449},
\href{https://github.com/jgabry/bayes-vis-paper}{code on GitHub})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split_moment_matching.R
\name{loo_moment_match_split}
\alias{loo_moment_match_split}
\title{Split moment matching for efficient approximate leave-one-out cross-validation (LOO)}
\usage{
loo_moment_match_split(
  x,
  upars,
  cov,
  total_shift,
  total_scaling,
  total_mapping,
  i,
  log_prob_upars,
  log_lik_i_upars,
  r_eff_i,
  cores,
  is_method,
  ...
)
}
\arguments{
\item{x}{A fitted model object.}

\item{upars}{A matrix containing the model parameters in unconstrained space
where they can have any real value.}

\item{cov}{Logical; Indicate whether to match the covariance matrix of the
samples or not. If \code{FALSE}, only the mean and marginal variances are
matched.}

\item{total_shift}{A vector representing the total shift made by the moment
matching algorithm.}

\item{total_scaling}{A vector representing the total scaling of marginal
variance made by the moment matching algorithm.}

\item{total_mapping}{A vector representing the total covariance
transformation made by the moment matching algorithm.}

\item{i}{Observation index.}

\item{log_prob_upars}{A function that takes arguments \code{x} and \code{upars} and
returns a matrix of log-posterior density values of the unconstrained
posterior draws passed via \code{upars}.}

\item{log_lik_i_upars}{A function that takes arguments \code{x}, \code{upars}, and \code{i}
and returns a vector of log-likelihood draws of the \code{i}th observation based
on the unconstrained posterior draws passed via \code{upars}.}

\item{r_eff_i}{MCMC relative effective sample size of the \code{i}'th log
likelihood draws.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}

\item{is_method}{The importance sampling method to use. The following methods
are implemented:
\itemize{
\item \code{\link[=psis]{"psis"}}: Pareto-Smoothed Importance Sampling (PSIS). Default method.
\item \code{\link[=tis]{"tis"}}: Truncated Importance Sampling (TIS) with truncation at
\code{sqrt(S)}, where \code{S} is the number of posterior draws.
\item \code{\link[=sis]{"sis"}}: Standard Importance Sampling (SIS).
}}

\item{...}{Further arguments passed to the custom functions documented above.}
}
\value{
A list containing the updated log-importance weights and
log-likelihood values. Also returns the updated MCMC effective sample size
and the integrand-specific log-importance weights.
}
\description{
A function that computes the split moment matching importance sampling loo.
Takes in the moment matching total transformation, transforms only half
of the draws, and computes a single elpd using multiple importance sampling.
}
\references{
Paananen, T., Piironen, J., Buerkner, P.-C., Vehtari, A. (2021).
Implicitly adaptive importance sampling. \emph{Statistics and Computing}, 31, 16.
doi:10.1007/s11222-020-09982-2. arXiv preprint arXiv:1906.08850.
}
\seealso{
\code{\link[=loo]{loo()}}, \code{\link[=loo_moment_match]{loo_moment_match()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{parallel_psis_list}
\alias{parallel_psis_list}
\alias{parallel_importance_sampling_list}
\title{Parallel psis list computations}
\usage{
parallel_psis_list(
  N,
  .loo_i,
  .llfun,
  data,
  draws,
  r_eff,
  save_psis,
  cores,
  ...
)

parallel_importance_sampling_list(
  N,
  .loo_i,
  .llfun,
  data,
  draws,
  r_eff,
  save_psis,
  cores,
  method,
  ...
)
}
\arguments{
\item{N}{The total number of observations (i.e. \code{nrow(data)}).}

\item{.loo_i}{The function used to compute individual loo contributions.}

\item{.llfun}{See \code{llfun} in \code{\link[=loo.function]{loo.function()}}.}

\item{data, draws, ...}{For the \code{loo.function()} method and the \code{loo_i()}
function, these are the data, posterior draws, and other arguments to pass
to the log-likelihood function. See the \strong{Methods (by class)} section
below for details on how to specify these arguments.}

\item{r_eff}{Vector of relative effective sample size estimates for the
likelihood (\code{exp(log_lik)}) of each observation. This is related to
the relative efficiency of estimating the normalizing term in
self-normalized importance sampling when using posterior draws obtained
with MCMC. If MCMC draws are used and \code{r_eff} is not provided then
the reported PSIS effective sample sizes and Monte Carlo error estimates
can be over-optimistic. If the posterior draws are (near) independent then
\code{r_eff=1} can be used. \code{r_eff} has to be a scalar (same value is used
for all observations) or a vector with length equal to the number of
observations. The default value is 1. See the \code{\link[=relative_eff]{relative_eff()}} helper
functions for help computing \code{r_eff}.}

\item{save_psis}{Should the \code{psis} object created internally by \code{loo()} be
saved in the returned object? The \code{loo()} function calls \code{\link[=psis]{psis()}}
internally but by default discards the (potentially large) \code{psis} object
after using it to compute the LOO-CV summaries. Setting \code{save_psis=TRUE}
will add a \code{psis_object} component to the list returned by \code{loo}.
This is useful if you plan to use the \code{\link[=E_loo]{E_loo()}} function to compute
weighted expectations after running \code{loo}. Several functions in the
\pkg{bayesplot} package also accept \code{psis} objects.}

\item{cores}{The number of cores to use for parallelization. This defaults to
the option \code{mc.cores} which can be set for an entire R session by
\code{options(mc.cores = NUMBER)}. The old option \code{loo.cores} is now
deprecated but will be given precedence over \code{mc.cores} until
\code{loo.cores} is removed in a future release. \strong{As of version
2.0.0 the default is now 1 core if \code{mc.cores} is not set}, but we
recommend using as many (or close to as many) cores as possible.
\itemize{
\item Note for Windows 10 users: it is \strong{strongly}
\href{https://github.com/stan-dev/loo/issues/94}{recommended} to avoid using
the \code{.Rprofile} file to set \code{mc.cores} (using the \code{cores} argument or
setting \code{mc.cores} interactively or in a script is fine).
}}

\item{method}{See \code{is_method} for \code{\link[=loo]{loo()}}}
}
\description{
Parallel psis list computations
}
\details{
Refactored function to handle parallel computations
for psis_list
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crps.R
\name{crps}
\alias{crps}
\alias{scrps}
\alias{loo_crps}
\alias{loo_scrps}
\alias{crps.matrix}
\alias{crps.numeric}
\alias{loo_crps.matrix}
\alias{scrps.matrix}
\alias{scrps.numeric}
\alias{loo_scrps.matrix}
\title{Continuously ranked probability score}
\usage{
crps(x, ...)

scrps(x, ...)

loo_crps(x, ...)

loo_scrps(x, ...)

\method{crps}{matrix}(x, x2, y, ..., permutations = 1)

\method{crps}{numeric}(x, x2, y, ..., permutations = 1)

\method{loo_crps}{matrix}(
  x,
  x2,
  y,
  log_lik,
  ...,
  permutations = 1,
  r_eff = 1,
  cores = getOption("mc.cores", 1)
)

\method{scrps}{matrix}(x, x2, y, ..., permutations = 1)

\method{scrps}{numeric}(x, x2, y, ..., permutations = 1)

\method{loo_scrps}{matrix}(
  x,
  x2,
  y,
  log_lik,
  ...,
  permutations = 1,
  r_eff = 1,
  cores = getOption("mc.cores", 1)
)
}
\arguments{
\item{x}{A \code{S} by \code{N} matrix (draws by observations), or a vector of length
\code{S} when only single observation is provided in \code{y}.}

\item{...}{Passed on to \code{\link[=E_loo]{E_loo()}} in the \verb{loo_*()} version of these
functions.}

\item{x2}{Independent draws from the same distribution as draws in \code{x}.
Should be of the identical dimension.}

\item{y}{A vector of observations or a single value.}

\item{permutations}{An integer, with default value of 1,  specifying how many
times the expected value of  |X - X'| (\verb{|x - x2|}) is computed. The row
order of \code{x2} is shuffled as elements \code{x} and \code{x2} are typically drawn
given the same values of parameters. This happens, e.g., when one calls
\code{posterior_predict()} twice for a fitted \pkg{rstanarm} or \pkg{brms}
model. Generating more permutations is expected to decrease the variance of
the computed expected value.}

\item{log_lik}{A log-likelihood matrix the same size as \code{x}.}

\item{r_eff}{An optional vector of relative effective sample size estimates
containing one element per observation. See \code{\link[=psis]{psis()}} for details.}

\item{cores}{The number of cores to use for parallelization of \verb{[psis()]}.
See \code{\link[=psis]{psis()}} for details.}
}
\value{
A list containing two elements: \code{estimates} and \code{pointwise}.
The former reports estimator and standard error and latter the pointwise
values. Following Bolin & Wallin (2023), a larger value is better.
}
\description{
The \code{crps()} and \code{scrps()} functions and their \verb{loo_*()} counterparts can be
used to compute the continuously ranked probability score (CRPS) and scaled
CRPS (SCRPS) (as defined by Bolin and Wallin, 2023). CRPS is a proper scoring rule, and
strictly proper when the first moment of the predictive distribution is
finite. Both can be expressed in terms of samples form the predictive
distribution. See, for example, a paper by Gneiting and Raftery (2007)
for a comprehensive discussion on CRPS.
}
\details{
To compute (S)CRPS, the user needs to provide two sets of draws, \code{x} and
\code{x2}, from the predictive distribution. This is due to the fact that formulas
used to compute CRPS involve an expectation of the absolute difference of \code{x}
and \code{x2}, both having the same distribution. See the \code{permutations} argument,
as well as Gneiting and Raftery (2007) for details.
}
\examples{
\dontrun{
# An example using rstanarm
library(rstanarm)
data("kidiq")
fit <- stan_glm(kid_score ~ mom_hs + mom_iq, data = kidiq)
ypred1 <- posterior_predict(fit)
ypred2 <- posterior_predict(fit)
crps(ypred1, ypred2, y = fit$y)
loo_crps(ypred1, ypred2, y = fit$y, log_lik = log_lik(fit))
}

}
\references{
Bolin, D., & Wallin, J. (2023). Local scale invariance and robustness of
proper scoring rules. Statistical Science, 38(1):140-159.

Gneiting, T., & Raftery, A. E. (2007). Strictly Proper Scoring Rules,
Prediction, and Estimation. Journal of the American Statistical Association,
102(477), 359–378.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kfold-generic.R
\name{kfold-generic}
\alias{kfold-generic}
\alias{kfold}
\alias{is.kfold}
\title{Generic function for K-fold cross-validation for developers}
\usage{
kfold(x, ...)

is.kfold(x)
}
\arguments{
\item{x}{A fitted model object.}

\item{...}{Arguments to pass to specific methods.}
}
\value{
For developers defining a \code{kfold()} method for a class
\code{"foo"}, the \code{kfold.foo()} function should return a list with class
\code{c("kfold", "loo")} with at least the following named elements:
\itemize{
\item \code{"estimates"}: A \verb{1x2} matrix containing the ELPD estimate and its
standard error. The matrix must have row name "\code{elpd_kfold}" and column
names \code{"Estimate"} and \code{"SE"}.
\item \code{"pointwise"}: A \code{Nx1} matrix with column name \code{"elpd_kfold"} containing
the pointwise contributions for each data point.
}

It is important for the object to have at least these classes and
components so that it is compatible with other functions like
\code{\link[=loo_compare]{loo_compare()}} and \code{print()} methods.
}
\description{
For developers of Bayesian modeling packages, \strong{loo} includes
a generic function \code{kfold()} so that methods may be defined for K-fold
CV without name conflicts between packages. See, for example, the
\code{kfold()} methods in the \strong{rstanarm} and \strong{brms} packages.

The \strong{Value} section below describes the objects that \code{kfold()}
methods should return in order to be compatible with
\code{\link[=loo_compare]{loo_compare()}} and the \strong{loo} package print methods.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_compare.R,
%   R/loo_compare.psis_loo_ss_list.R
\name{loo_compare}
\alias{loo_compare}
\alias{loo_compare.default}
\alias{print.compare.loo}
\alias{print.compare.loo_ss}
\title{Model comparison}
\usage{
loo_compare(x, ...)

\method{loo_compare}{default}(x, ...)

\method{print}{compare.loo}(x, ..., digits = 1, p_worse = TRUE)

\method{print}{compare.loo_ss}(x, ..., digits = 1)
}
\arguments{
\item{x}{An object of class \code{"loo"} or a list of such objects. If a list is
used then the list names will be used as the model names in the output. See
\strong{Examples}.}

\item{...}{Additional objects of class \code{"loo"}, if not passed in as a single
list.}

\item{digits}{For the print method only, the number of digits to use when
printing.}

\item{p_worse}{For the print method only, should we include the normal
approximation based probability of each model having worse performance than
the best model? The default is \code{TRUE}.}
}
\value{
A data frame with class \code{"compare.loo"} that has its own
print method. See the \strong{Details} and \strong{Examples} sections.
}
\description{
Compare fitted models based on \link[=loo-glossary]{ELPD}.
}
\details{
When comparing two fitted models, we can estimate the difference in their
expected predictive accuracy by the difference in
\code{\link[=loo-glossary]{elpd_loo}} or \code{elpd_waic} (or multiplied by \eqn{-2}, if
desired, to be on the deviance scale).
\subsection{\code{elpd_diff} and \code{se_diff}}{

When using \code{loo_compare()}, the returned data frame will have one row per
model and several columns of estimates. The values of
\code{\link[=loo-glossary]{elpd_diff}} and \code{\link[=loo-glossary]{se_diff}} are computed by
making pairwise comparisons between each model and the model with the
largest ELPD (the model listed first). Therefore, the first \code{elpd_diff}
value will always be \code{0} (i.e., the difference between the preferred model
and itself) and the rest of the values will be negative.

To compute the standard error of the difference in \link[=loo-glossary]{ELPD} ---
which should not be expected to equal the difference of the standard errors
--- we use a paired estimate to take advantage of the fact that the same
set of \eqn{N} data points was used to fit both models. These calculations
should be most useful when \eqn{N} is large, because then non-normality of
the distribution is not such an issue when estimating the uncertainty in
these sums. These standard errors, for all their flaws, should give a
better sense of uncertainty than what is obtained using the current
standard approach of comparing differences of deviances to a Chi-squared
distribution, a practice derived for Gaussian linear models or
asymptotically, and which only applies to nested models in any case.
}

\subsection{\code{p_worse}, \code{diag_diff}, and \code{diag_elpd}}{

The values in the \code{p_worse} column show the probability of each model
having worse ELPD than the best model. These probabilities are computed
with a normal approximation using the values from \code{elpd_diff} and
\code{se_diff}. Sivula et al. (2025) present the conditions when the normal
approximation used for SE and \code{se_diff} is good, and the column
\code{diag_diff} contains possible diagnostic messages:
\itemize{
\item \code{N < 100} (small data)
\item \verb{|elpd_diff| < 4} (models make similar predictions)
}

If either of these diagnostic messages is shown, the error distribution is
skewed or thick tailed and the normal approximation based on \code{elpd_diff}
and \code{se_diff} is not well calibrated. In that case, the probabilities
\code{p_worse} are likely to be too large. However, \code{elpd_diff} and \code{se_diff}
will still be indicative of the differences and uncertainties (for example,
if \verb{|elpd_diff|} is many times larger than \code{se_diff} the difference is quite
certain). In addition, if the model is not well specificed and there are
outliers, the error distribution can also be skewed or thick tailed and the
normal approximation is not well calibrated. Possible model misspecification
and outliers can be diagnosed with usual predictive checking methods.

The column \code{diag_elpd} shows the PSIS-LOO Pareto k diagnostic for the
pointwise ELPD computations for each model. If \verb{K k_psis > 0.7} is shown,
where \code{K} is the number of high Pareto k values in the PSIS
computation, then there may be significant bias in \code{elpd_diff} favoring
models with a large number of high Pareto k values.
}

\subsection{Warnings for many model comparisons}{

If more than \eqn{11} models are compared, we internally recompute the model
differences using the median model by ELPD as the baseline model. We then
estimate whether the differences in predictive performance are potentially
due to chance as described by McLatchie and Vehtari (2023). This will flag
a warning if it is deemed that there is a risk of over-fitting due to the
selection process. In that case users are recommended to avoid model
selection based on LOO-CV, and instead to favor model averaging/stacking or
projection predictive inference.
}
}
\examples{
# very artificial example, just for demonstration!
LL <- example_loglik_array()
loo1 <- loo(LL)     # should be worst model when compared
loo2 <- loo(LL + 1) # should be second best model when compared
loo3 <- loo(LL + 2) # should be best model when compared

comp <- loo_compare(loo1, loo2, loo3)
print(comp, digits = 2)

# can use a list of objects with custom names
# the names will be used in the output
loo_compare(list("apple" = loo1, "banana" = loo2, "cherry" = loo3))

\dontrun{
# works for waic (and kfold) too
loo_compare(waic(LL), waic(LL - 10))
}

}
\references{
Vehtari, A., Gelman, A., and Gabry, J. (2017). Practical Bayesian model
evaluation using leave-one-out cross-validation and WAIC.
\emph{Statistics and Computing}. 27(5), 1413--1432. doi:10.1007/s11222-016-9696-4
(\href{https://link.springer.com/article/10.1007/s11222-016-9696-4}{journal version},
\href{https://arxiv.org/abs/1507.04544}{preprint arXiv:1507.04544}).

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. \emph{Journal of Machine Learning Research},
25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}

Sivula, T, Magnusson, M., Matamoros A. A., and Vehtari, A. (2025).
Uncertainty in Bayesian leave-one-out cross-validation based model
comparison. \emph{Bayesian Analysis}. \doi{10.1214/25-BA1569}

McLatchie, Y., and Vehtari, A. (2024).  Efficient estimation and
correction of selection-induced bias with order statistics.
\emph{Statistics and Computing}. 34(132). \doi{10.1007/s11222-024-10442-4}
}
\seealso{
\itemize{
\item The \href{https://mc-stan.org/loo/articles/online-only/faq.html}{FAQ page} on
the \strong{loo} website for answers to frequently asked questions.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_log_lik.R
\name{extract_log_lik}
\alias{extract_log_lik}
\title{Extract pointwise log-likelihood from a Stan model}
\usage{
extract_log_lik(stanfit, parameter_name = "log_lik", merge_chains = TRUE)
}
\arguments{
\item{stanfit}{A \code{stanfit} object (\pkg{rstan} package).}

\item{parameter_name}{A character string naming the parameter (or generated
quantity) in the Stan model corresponding to the log-likelihood.}

\item{merge_chains}{If \code{TRUE} (the default), all Markov chains are
merged together (i.e., stacked) and a matrix is returned. If \code{FALSE}
they are kept separate and an array is returned.}
}
\value{
If \code{merge_chains=TRUE}, an \eqn{S} by \eqn{N} matrix of
(post-warmup) extracted draws, where \eqn{S} is the size of the posterior
sample and \eqn{N} is the number of data points. If
\code{merge_chains=FALSE}, an \eqn{I} by \eqn{C} by \eqn{N} array, where
\eqn{I \times C = S}{I * C = S}.
}
\description{
Convenience function for extracting the pointwise log-likelihood
matrix or array from a \code{stanfit} object from the \pkg{rstan} package.
Note: recent versions of \pkg{rstan} now include a \code{loo()} method for
\code{stanfit} objects that handles this internally.
}
\details{
Stan does not automatically compute and store the log-likelihood. It
is up to the user to incorporate it into the Stan program if it is to be
extracted after fitting the model. In a Stan model, the pointwise log
likelihood can be coded as a vector in the transformed parameters block
(and then summed up in the model block) or it can be coded entirely in the
generated quantities block. We recommend using the generated quantities
block so that the computations are carried out only once per iteration
rather than once per HMC leapfrog step.

For example, the following is the \verb{generated quantities} block for
computing and saving the log-likelihood for a linear regression model with
\code{N} data points, outcome \code{y}, predictor matrix \code{X},
coefficients \code{beta}, and standard deviation \code{sigma}:

\verb{vector[N] log_lik;}

\code{for (n in 1:N) log_lik[n] = normal_lpdf(y[n] | X[n, ] * beta, sigma);}
}
\references{
Stan Development Team (2017). The Stan C++ Library, Version 2.16.0.
\url{https://mc-stan.org/}

Stan Development Team (2017). RStan: the R interface to Stan, Version 2.16.1.
\url{https://mc-stan.org/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\name{loo-datasets}
\alias{loo-datasets}
\alias{Kline}
\alias{milk}
\alias{voice}
\alias{voice_loo}
\title{Datasets for loo examples and vignettes}
\description{
Small datasets for use in \strong{loo} examples and vignettes. The \code{Kline}
and \code{milk} datasets are also included in the \strong{rethinking} package
(McElreath, 2016a), but we include them here as \strong{rethinking} is not
on CRAN.
}
\details{
Currently the data sets included are:
\itemize{
\item \code{Kline}:
Small dataset from Kline and Boyd (2010) on tool complexity and demography
in Oceanic islands societies. This data is discussed in detail in
McElreath (2016a,2016b). \href{https://www.rdocumentation.org/packages/rethinking/versions/1.59/topics/Kline}{(Link to variable descriptions)}
\item \code{milk}:
Small dataset from Hinde and Milligan (2011) on primate milk
composition.This data is discussed in detail in McElreath (2016a,2016b).
\href{https://www.rdocumentation.org/packages/rethinking/versions/1.59/topics/milk}{(Link to variable descriptions)}
\item \code{voice}:
Voice rehabilitation data from Tsanas et al. (2014).
}
}
\examples{
str(Kline)
str(milk)

}
\references{
Hinde and Milligan. 2011. \emph{Evolutionary Anthropology} 20:9-23.

Kline, M.A. and R. Boyd. 2010. \emph{Proc R Soc B} 277:2559-2564.

McElreath, R. (2016a). rethinking: Statistical Rethinking book package.
R package version 1.59.

McElreath, R. (2016b). \emph{Statistical rethinking: A Bayesian course with
examples in R and Stan}. Chapman & Hall/CRC.

A. Tsanas, M.A. Little, C. Fox, L.O. Ramig: Objective automatic assessment of
rehabilitative speech treatment in Parkinson's disease, IEEE
Transactions on Neural Systems and Rehabilitation Engineering, Vol. 22, pp.
181-190, January 2014
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gpdfit.R
\name{gpdfit}
\alias{gpdfit}
\title{Estimate parameters of the Generalized Pareto distribution}
\usage{
gpdfit(x, wip = TRUE, min_grid_pts = 30, sort_x = TRUE)
}
\arguments{
\item{x}{A numeric vector. The sample from which to estimate the parameters.}

\item{wip}{Logical indicating whether to adjust \eqn{k} based on a weakly
informative Gaussian prior centered on 0.5. Defaults to \code{TRUE}.}

\item{min_grid_pts}{The minimum number of grid points used in the fitting
algorithm. The actual number used is \code{min_grid_pts + floor(sqrt(length(x)))}.}

\item{sort_x}{If \code{TRUE} (the default), the first step in the fitting
algorithm is to sort the elements of \code{x}. If \code{x} is already
sorted in ascending order then \code{sort_x} can be set to \code{FALSE} to
skip the initial sorting step.}
}
\value{
A named list with components \code{k} and \code{sigma}.
}
\description{
Given a sample \eqn{x}, Estimate the parameters \eqn{k} and \eqn{\sigma} of
the generalized Pareto distribution (GPD), assuming the location parameter is
0. By default the fit uses a prior for \eqn{k}, which will stabilize
estimates for very small sample sizes (and low effective sample sizes in the
case of MCMC samples). The weakly informative prior is a Gaussian prior
centered at 0.5.
}
\details{
Here the parameter \eqn{k} is the negative of \eqn{k} in Zhang &
Stephens (2009).
}
\references{
Zhang, J., and Stephens, M. A. (2009). A new and efficient estimation method
for the generalized Pareto distribution. \emph{Technometrics} \strong{51}, 316-325.
}
\seealso{
\code{\link[=psis]{psis()}}, \link{pareto-k-diagnostic}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_subsample.R
\name{nobs.psis_loo_ss}
\alias{nobs.psis_loo_ss}
\title{The number of observations in a \code{psis_loo_ss} object.}
\usage{
\method{nobs}{psis_loo_ss}(object, ...)
}
\arguments{
\item{object}{a \code{psis_loo_ss} object.}

\item{...}{Currently unused.}
}
\description{
The number of observations in a \code{psis_loo_ss} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo_compare.R
\name{find_model_names}
\alias{find_model_names}
\title{Find the model names associated with \code{"loo"} objects}
\usage{
find_model_names(x)
}
\arguments{
\item{x}{List of \code{"loo"} objects.}
}
\value{
Character vector of model names the same length as \code{x.}
}
\description{
Find the model names associated with \code{"loo"} objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{nlist}
\alias{nlist}
\title{Named lists}
\usage{
nlist(...)
}
\arguments{
\item{...}{Objects to include in the list.}
}
\value{
A named list.
}
\description{
Create a named list using specified names or, if names are omitted, using the
names of the objects in the list. The code \code{list(a = a, b = b)} becomes
\code{nlist(a,b)} and \code{list(a = a, b = 2)} becomes \code{nlist(a, b = 2)}, etc.
}
\examples{

# All variables already defined
a <- rnorm(100)
b <- mat.or.vec(10, 3)
nlist(a,b)

# Define some variables in the call and take the rest from the environment
nlist(a, b, veggies = c("lettuce", "spinach"), fruits = c("banana", "papaya"))

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importance_sampling.R
\name{importance_sampling.default}
\alias{importance_sampling.default}
\title{Importance sampling (default)}
\usage{
\method{importance_sampling}{default}(log_ratios, method, ..., r_eff = NULL)
}
\arguments{
\item{log_ratios}{An array, matrix, or vector of importance ratios on the log
scale (for PSIS-LOO these are \emph{negative} log-likelihood values). See the
\strong{Methods (by class)} section below for a detailed description of how
to specify the inputs for each method.}

\item{...}{Arguments passed on to the various methods.}

\item{r_eff}{Vector of relative effective sample size estimates containing
one element per observation. The values provided should be the relative
effective sample sizes of \code{1/exp(log_ratios)} (i.e., \code{1/ratios}).
This is related to the relative efficiency of estimating the normalizing
term in self-normalizing importance sampling. If \code{r_eff} is not
provided then the reported PSIS effective sample sizes and Monte Carlo
error estimates will be over-optimistic. See the \code{\link[=relative_eff]{relative_eff()}}
helper function for computing \code{r_eff}. If using \code{psis} with
draws of the \code{log_ratios} not obtained from MCMC then the warning
message thrown when not specifying \code{r_eff} can be disabled by
setting \code{r_eff} to \code{NA}.}

\item{is_method}{The importance sampling method to use. The following methods
are implemented:
\itemize{
\item \code{\link[=psis]{"psis"}}: Pareto-Smoothed Importance Sampling (PSIS). Default method.
\item \code{\link[=tis]{"tis"}}: Truncated Importance Sampling (TIS) with truncation at
\code{sqrt(S)}, where \code{S} is the number of posterior draws.
\item \code{\link[=sis]{"sis"}}: Standard Importance Sampling (SIS).
}}
}
\description{
Importance sampling (default)
}
\keyword{internal}
